Package: TextAnalysisR
Title: A Text Mining Workflow Tool
Version: 0.1.4
Authors@R: 
    person("Mikyung", "Shin", , "shin.mikyung@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-7907-9193"))
Description: Provides a text mining and natural language processing
    workflow for documents. Includes preprocessing
    via 'quanteda', lexical analysis (term frequency-inverse document
    frequency, log-odds ratios, lexical diversity) via 'tidytext',
    topic modeling via 'stm' and the 'BERTopic' approach, semantic
    similarity and document clustering on transformer representations,
    an interactive 'Shiny' interface with 'ggplot2' visualization,
    optional 'spaCy' preprocessing, and local 'sentence-transformers'
    or web-based ('OpenAI', 'Gemini') model
    providers for retrieval-augmented generation, as described in
    Shin et al. (2026) <doi:10.1177/07319487251412879>.
License: GPL (>= 3)
Language: en-US
Encoding: UTF-8
LazyData: true
LazyDataCompression: xz
Imports: dplyr, DT, ggplot2, htmltools, htmlwidgets, igraph, magrittr,
        Matrix, methods, quanteda, quanteda.textstats, RColorBrewer,
        rlang, scales, shiny, tibble, tidyr, tidytext, widyr, withr
Suggests: aricode, tidylo, broom, patchwork, cluster, clusterCrit,
        colourpicker, covr, dbscan, digest, httr, jsonlite, knitr,
        markdown, MASS, moments, numform, officer, openxlsx, pdftools,
        progress, proxy, pscl, purrr, readxl, reticulate (>= 1.28),
        rmarkdown, roxygen2, Rtsne, shinyBS, shinybusy, shinyjs,
        spelling, stm, stopwords, stringr, syuzhet, testthat (>=
        3.0.0), textdata, umap, plotly
Depends: R (>= 4.0)
SystemRequirements: Python (>= 3.8) with spaCy (>= 3.5) for
        lemmatization and sentence-transformers (>= 2.2) for embeddings
        (optional)
URL: https://mshin77.github.io/TextAnalysisR/,
        https://github.com/mshin77/TextAnalysisR
BugReports: https://github.com/mshin77/TextAnalysisR/issues
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-17 17:44:05 UTC; shinm
Author: Mikyung Shin [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-7907-9193>)
Maintainer: Mikyung Shin <shin.mikyung@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-27 14:40:13 UTC
Built: R 4.6.1; ; 2026-07-27 16:55:00 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(analyze_document_clustering)
export(analyze_semantic_evolution)
export(analyze_sentiment)
export(analyze_sentiment_llm)
export(analyze_similarity_gaps)
export(assess_embedding_stability)
export(auto_tune_embedding_topics)
export(calculate_assignment_consistency)
export(calculate_clustering_metrics)
export(calculate_cosine_similarity)
export(calculate_cross_similarity)
export(calculate_dispersion_metrics)
export(calculate_document_similarity)
export(calculate_keyword_stability)
export(calculate_lexical_dispersion)
export(calculate_log_odds_ratio)
export(calculate_semantic_drift)
export(calculate_similarity_robust)
export(calculate_text_readability)
export(calculate_topic_probability)
export(calculate_topic_stability)
export(calculate_weighted_log_odds)
export(calculate_word_frequency)
export(call_llm_api)
export(check_python_env)
export(check_vision_models)
export(clear_lexdiv_cache)
export(cluster_embeddings)
export(describe_image)
export(detect_multi_words)
export(detect_pdf_content_type)
export(detect_pdf_content_type_py)
export(export_document_clustering)
export(extract_cross_category_similarities)
export(extract_keywords_keyness)
export(extract_keywords_tfidf)
export(extract_morphology)
export(extract_named_entities)
export(extract_noun_chunks)
export(extract_pos_tags)
export(extract_subjects_objects)
export(extract_tables_from_pdf_py)
export(extract_topic_terms_df)
export(find_optimal_k)
export(find_similar_words)
export(find_topic_matches)
export(fit_embedding_model)
export(fit_embedding_topics)
export(fit_semantic_model)
export(fit_temporal_model)
export(generate_cluster_labels)
export(generate_cluster_labels_auto)
export(generate_embeddings)
export(generate_topic_content)
export(generate_topic_labels)
export(get_available_dfm)
export(get_available_tokens)
export(get_best_embeddings)
export(get_content_type_prompt)
export(get_content_type_user_template)
export(get_sentences)
export(get_sentiment_color)
export(get_sentiment_colors)
export(get_spacy_model_info)
export(get_topic_prevalence)
export(get_topic_terms)
export(get_topic_texts)
export(get_word_similarity)
export(identify_topic_trends)
export(import_files)
export(init_spacy_nlp)
export(lemmatize_tokens)
export(lexical_diversity_analysis)
export(lexical_frequency_analysis)
export(plot_cluster_terms)
export(plot_cross_category_heatmap)
export(plot_document_sentiment_trajectory)
export(plot_emotion_radar)
export(plot_entity_frequencies)
export(plot_keyness_keywords)
export(plot_keyword_comparison)
export(plot_lexical_dispersion)
export(plot_lexical_diversity_distribution)
export(plot_log_odds_ratio)
export(plot_model_comparison)
export(plot_morphology_feature)
export(plot_mwe_frequency)
export(plot_ngram_frequency)
export(plot_pos_frequencies)
export(plot_quality_metrics)
export(plot_readability_by_group)
export(plot_readability_distribution)
export(plot_semantic_viz)
export(plot_sentiment_boxplot)
export(plot_sentiment_by_category)
export(plot_sentiment_distribution)
export(plot_sentiment_violin)
export(plot_similarity_heatmap)
export(plot_term_trends_continuous)
export(plot_tfidf_keywords)
export(plot_top_readability_documents)
export(plot_topic_effects_categorical)
export(plot_topic_effects_continuous)
export(plot_topic_probability)
export(plot_weighted_log_odds)
export(plot_word_frequency)
export(plot_word_probability)
export(prep_texts)
export(process_pdf_unified)
export(reduce_dimensions)
export(render_displacy_dep)
export(render_displacy_ent)
export(run_app)
export(run_neural_topics_internal)
export(run_rag_search)
export(run_text_workflow)
export(semantic_document_clustering)
export(semantic_similarity_analysis)
export(sentiment_embedding_analysis)
export(sentiment_lexicon_analysis)
export(setup_python_env)
export(show_web_banner)
export(spacy_extract_entities)
export(spacy_has_vectors)
export(spacy_initialized)
export(spacy_lemmatize)
export(spacy_parse_full)
export(summarize_morphology)
export(unite_cols)
export(validate_cross_models)
export(validate_semantic_coherence)
export(word_co_occurrence_network)
export(word_correlation_network)
import(dplyr)
import(ggplot2)
import(quanteda)
import(shiny)
import(tidyr)
importFrom(DT,datatable)
importFrom(DT,formatStyle)
importFrom(Matrix,colSums)
importFrom(RColorBrewer,brewer.pal)
importFrom(dplyr,arrange)
importFrom(dplyr,case_when)
importFrom(dplyr,count)
importFrom(dplyr,desc)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,group_map)
importFrom(dplyr,inner_join)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n_distinct)
importFrom(dplyr,pull)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(dplyr,ungroup)
importFrom(ggplot2,aes)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_size_identity)
importFrom(ggplot2,theme_void)
importFrom(grDevices,colorRampPalette)
importFrom(htmltools,browsable)
importFrom(htmltools,tagList)
importFrom(htmltools,tags)
importFrom(igraph,V)
importFrom(igraph,assortativity_degree)
importFrom(igraph,betweenness)
importFrom(igraph,closeness)
importFrom(igraph,cluster_leiden)
importFrom(igraph,cluster_louvain)
importFrom(igraph,degree)
importFrom(igraph,diameter)
importFrom(igraph,distances)
importFrom(igraph,ecount)
importFrom(igraph,edge_density)
importFrom(igraph,eigen_centrality)
importFrom(igraph,graph_from_data_frame)
importFrom(igraph,induced_subgraph)
importFrom(igraph,layout_with_fr)
importFrom(igraph,modularity)
importFrom(igraph,transitivity)
importFrom(igraph,vcount)
importFrom(magrittr,"%>%")
importFrom(methods,as)
importFrom(quanteda,corpus)
importFrom(quanteda,dfm)
importFrom(quanteda,docnames)
importFrom(quanteda,tokens)
importFrom(quanteda.textstats,textstat_lexdiv)
importFrom(quanteda.textstats,textstat_readability)
importFrom(rlang,"%||%")
importFrom(rlang,sym)
importFrom(scales,rescale)
importFrom(shiny,showNotification)
importFrom(stats,as.dist)
importFrom(stats,as.formula)
importFrom(stats,binomial)
importFrom(stats,coef)
importFrom(stats,cor)
importFrom(stats,cutree)
importFrom(stats,dist)
importFrom(stats,glm)
importFrom(stats,glm.control)
importFrom(stats,hclust)
importFrom(stats,kmeans)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,poisson)
importFrom(stats,quantile)
importFrom(stats,reformulate)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,pivot_longer)
importFrom(tidyr,pivot_wider)
importFrom(tidytext,get_sentiments)
importFrom(tidytext,tidy)
importFrom(utils,head)
importFrom(utils,modifyList)
importFrom(utils,read.csv)
importFrom(utils,str)
importFrom(widyr,pairwise_cor)
importFrom(widyr,pairwise_count)
importFrom(widyr,pairwise_pmi)

<img src="man/figures/logo.png" alt="TextAnalysisR Logo" align="right" width="150px"/>

[![R-CMD-check](https://github.com/mshin77/TextAnalysisR/workflows/R-CMD-check/badge.svg)](https://github.com/mshin77/TextAnalysisR/actions)
[![Project Status:
Active](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![License:
GPL-3](https://img.shields.io/badge/License-GPL%20v3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0.html)

Text mining and natural language processing workflow for documents
(`PDF`, `DOCX`, `XLSX`, `CSV`, `TXT`). Includes preprocessing via
[quanteda](https://github.com/quanteda/quanteda), lexical analysis
(term frequency-inverse document frequency, log-odds ratios, lexical
diversity) via [tidytext](https://github.com/juliasilge/tidytext),
topic modeling via [stm](https://github.com/bstewart/stm) and
[BERTopic](https://maartengr.github.io/BERTopic/), semantic similarity
and document clustering on transformer embeddings, an interactive
[Shiny](https://shiny.posit.co/) interface with
[ggplot2](https://ggplot2.tidyverse.org/) visualization, optional
[spaCy](https://spacy.io/) lemmatization, and local
[sentence-transformers](https://www.sbert.net/) or web-based
(OpenAI,
[Gemini](https://ai.google.dev/)) model providers for
retrieval-augmented generation.

## Installation

From [R-universe](https://mshin77.r-universe.dev) (pre-built binaries
for Windows, macOS, and Linux):

    install.packages("TextAnalysisR",
      repos = c("https://mshin77.r-universe.dev", "https://cloud.r-project.org"))

Or the development version from
[GitHub](https://github.com/mshin77/TextAnalysisR):

    install.packages("remotes")
    remotes::install_github("mshin77/TextAnalysisR")

## First-Time Python Setup (Optional)

Python is optional. Core analyses (topic modeling, keywords, lexical
diversity, networks, sentiment lexicons) run in plain R. Use the local
package for large datasets.

Several functions (`lemmatize_tokens()`, `generate_embeddings()`,
`cluster_embeddings()`, PDF extraction, transformer-based analyses)
require Python packages. Run this **once** after installing
TextAnalysisR:

    library(TextAnalysisR)
    setup_python_env()

This creates a dedicated virtualenv (`textanalysisr-env`), installs the
packages listed in `inst/python/requirements.txt` (spaCy, pandas,
pdfplumber, sentence-transformers, torch, umap-learn, hdbscan,
scikit-learn, numba), and downloads the `en_core_web_sm` spaCy model.
Restart R afterward. Check status anytime with `check_python_env()`.

## Load the TextAnalysisR Package

    library(TextAnalysisR)

## Alternatively, Launch and Browse the Shiny App

Access the web app at <https://www.textanalysisr.org>.

Launch and browse the app on the local computer:

    run_app()

## Getting Started

See [Quick
Start](https://mshin77.github.io/TextAnalysisR//articles/quickstart.html)
for tutorials.

## Citation

- Shin, M. (2026). *TextAnalysisR: A text mining workflow tool* (R
  package version 0.1.4) \[Computer software\].
  <https://mshin77.github.io/TextAnalysisR/>

- Shin, M. (2026). *TextAnalysisR: A text mining workflow tool* \[Web
  application\]. <https://www.textanalysisr.org>

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(message = FALSE, warning = FALSE)

## -----------------------------------------------------------------------------
library(TextAnalysisR)
packageVersion("TextAnalysisR")

mydata <- SpecialEduTech[seq_len(20), c("title", "abstract")]
united <- unite_cols(mydata, listed_vars = c("title", "abstract"))
toks   <- prep_texts(united, text_field = "united_texts")
dfm    <- quanteda::dfm(toks)
extract_keywords_tfidf(dfm, top_n = 10)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(message = FALSE, warning = FALSE)

## -----------------------------------------------------------------------------
library(TextAnalysisR)
packageVersion("TextAnalysisR")

data("SpecialEduTech")
nrow(SpecialEduTech)
head(SpecialEduTech[, c("title", "year")], 3)

## ----eval = FALSE-------------------------------------------------------------
# install.packages("TextAnalysisR",
#   repos = c("https://mshin77.r-universe.dev", "https://cloud.r-project.org"))
# library(TextAnalysisR)
# run_app()

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(message = FALSE, warning = FALSE)

## -----------------------------------------------------------------------------
library(TextAnalysisR)

sample_text <- c(
  "Figure 1 shows the distribution of student outcomes.",
  "Table 2 reports the effect sizes for each intervention."
)
toks <- prep_texts(
  data.frame(united_texts = sample_text),
  text_field = "united_texts"
)
quanteda::ntoken(toks)

is_web <- TextAnalysisR:::check_web_deployment()
is_docker <- TextAnalysisR:::check_docker_deployment()
is_remote <- is_web || is_docker

server_gemini_key <- Sys.getenv("GEMINI_API_KEY", "")
has_server_gemini <- nzchar(server_gemini_key)
server_gemini_model <- Sys.getenv("GEMINI_DEFAULT_MODEL", "gemini-2.5-flash")

shiny::enableBookmarking("disable")

.password_input <- function(inputId, label, value = "", placeholder = NULL) {
  if (isTRUE(has_server_gemini) && grepl("gemini", inputId, ignore.case = TRUE)) {
    return(tags$p(
      style = "color: #475569; font-size: 13px; margin-bottom: 8px;",
      "Usage is free here, supported by the Google Cloud Research program."
    ))
  }
  pw <- shiny::passwordInput(inputId, label, value = value, placeholder = placeholder)
  pw$children <- lapply(pw$children, function(ch) {
    if (inherits(ch, "shiny.tag") && identical(ch$name, "input")) {
      ch$attribs$autocomplete <- "new-password"
    }
    ch
  })
  pw
}

topic_modeling_ui_content <- function() {
      sidebarLayout(
        sidebarPanel(
          width = 3,
          class = "sidebar-panel",

          tags$h5(HTML("<strong>Topic modeling approach</strong>"), style = "color: #4269BF; margin-bottom: 10px;"),
          tags$div(
            id = "topic_modeling_path",
            class = "shiny-input-radiogroup shiny-input-container",
            role = "radiogroup",
            tags$div(
              class = "radio",
              style = "display: flex; align-items: center; margin-bottom: 8px;",
              tags$label(
                style = "margin-bottom: 0; display: flex; align-items: center; width: 100%;",
                tags$input(
                  type = "radio",
                  name = "topic_modeling_path",
                  value = "probability",
                  checked = "checked"
                ),
                tags$span("Structural Topic Model", style = "margin-left: 5px;"),
                actionLink("showSTMInfo", icon("info-circle"),
                          style = "color: #337ab7; font-size: 16px; margin-left: 8px;",
                          title = "Learn about STM",
                          `aria-label` = "Learn about STM")
              )
            ),
            tags$div(
              class = "radio",
              style = "display: flex; align-items: center; margin-bottom: 8px;",
              tags$label(
                style = "margin-bottom: 0; display: flex; align-items: center; width: 100%;",
                tags$input(
                  type = "radio",
                  name = "topic_modeling_path",
                  value = "embedding"
                ),
                tags$span("Embedding-based Topic Model", style = "margin-left: 5px;"),
                actionLink("showEmbeddingTopicsInfo", icon("info-circle"),
                          style = "color: #337ab7; font-size: 16px; margin-left: 8px;",
                          title = "Learn about Embedding-based Topics",
                          `aria-label` = "Learn about Embedding-based Topics")
              )
            )
          ),
          hr(),

          conditionalPanel(
            condition = "input.topic_modeling_path == 'probability' && input.conditioned3 == 4 && input.searchKSubtabs != 'ai_rec'",
            tags$h5(
              HTML("<strong>Evaluate optimal topic number (K)</strong> <a href='https://github.com/bstewart/stm' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"),
              style = "color: #4269BF; margin-bottom: 10px;"
            ),
            sliderInput(
              "stm_K_range",
              "Range of topic numbers",
              value = c(5, 10),
              min = 2,
              max = 50
            ),
            selectizeInput(
              "stm_categorical_var",
              "Categorical covariate(s)",
              choices = NULL,
              multiple = TRUE,
              options = list(placeholder = "Optional")
            ),
            selectizeInput(
              "stm_continuous_var",
              "Continuous covariate(s)",
              choices = NULL,
              multiple = TRUE,
              options = list(placeholder = "Optional")
            ),
            selectInput("stm_init_type_search", "Initialization type",
              choices = c("Spectral", "LDA", "Random", "Custom"),
              selected = "Spectral"
            ),
            radioButtons("stm_gamma_prior_search", "Gamma prior",
              choices = c("Pooled", "L1"),
              selected = "Pooled"
            ),
            radioButtons("stm_kappa_prior_search", "Kappa prior",
              choices = c("L1", "Jeffreys"),
              selected = "L1"
            ),
            numericInput("stm_max_em_its_search", "Max EM iterations",
              value = 500, min = 100, max = 2000, step = 50
            ),
            div(
              style = "margin-bottom: 15px;",
              actionButton("stm_search", "Search K", class = "btn-primary btn-block")
            )
          ),
          conditionalPanel(
            condition = "input.topic_modeling_path == 'probability' && input.conditioned3 == 5",
            tags$h5(strong("Structural topic model"), style = "color: #4269BF; margin-bottom: 10px;"),
            uiOutput("stm_k_selector_uiOutput"),
            selectInput(
              "stm_topic_measure",
              "Topic term measure",
              choices = c(
                "FREX" = "frex",
                "Lift" = "lift",
                "Score" = "score",
                "Probability" = "beta"
              ),
              selected = "frex"
            ),
            conditionalPanel(
              condition = "input.stm_topic_measure == 'frex'",
              sliderInput(
                "stm_top_term_number_frex",
                "Top terms (FREX)",
                value = 5,
                min = 0,
                max = 50
              )
            ),
            conditionalPanel(
              condition = "input.stm_topic_measure == 'lift'",
              sliderInput(
                "stm_top_term_number_lift",
                "Top terms (Lift)",
                value = 5,
                min = 0,
                max = 50
              )
            ),
            conditionalPanel(
              condition = "input.stm_topic_measure == 'score'",
              sliderInput(
                "stm_top_term_number_score",
                "Top terms (Score)",
                value = 5,
                min = 0,
                max = 50
              )
            ),
            conditionalPanel(
              condition = "input.stm_topic_measure == 'beta'",
              sliderInput(
                "stm_top_term_number_beta",
                "Top terms (Probability)",
                value = 5,
                min = 0,
                max = 50
              )
            ),
            sliderInput(
              "stm_ncol_top_terms",
              "Column numbers",
              value = 2,
              min = 1,
              max = 10
            ),
            selectizeInput(
              "stm_categorical_var_2",
              "Categorical covariate(s)",
              choices = NULL,
              multiple = TRUE,
              options = list(placeholder = "Optional")
            ),
            selectizeInput(
              "stm_continuous_var_2",
              "Continuous covariate(s)",
              choices = NULL,
              multiple = TRUE,
              options = list(placeholder = "Optional")
            ),
            selectInput("stm_init_type_K", "Initialization type",
              choices = c("Spectral", "LDA", "Random", "Custom"),
              selected = "Spectral"
            ),
            radioButtons("stm_gamma_prior_K", "Gamma prior",
              choices = c("Pooled", "L1"),
              selected = "Pooled"
            ),
            radioButtons("stm_kappa_prior_K", "Kappa prior",
              choices = c("L1", "Jeffreys"),
              selected = "L1"
            ),
            numericInput("stm_max_em_its_K", "Max EM iterations",
              value = 500, min = 100, max = 2000, step = 50
            ),
            div(
              style = "margin-bottom: 15px;",
              actionButton("stm_run", "Display", class = "btn-primary btn-block")
            ),
            tags$hr(),
            tags$h5(strong("Generate topic labels using AI"), style = "color: #4269BF; margin-bottom: 10px;"),
            sliderInput(
              "stm_top_term_number_labeling",
              "Top terms for labeling",
              value = 7,
              min = 3,
              max = 15
            ),
            radioButtons(
              "stm_label_provider",
              "AI Provider:",
              choices = c(
                "OpenAI (API Key Required)" = "openai",
                "Gemini (API Key Required)" = "gemini"
              ),
              selected = "openai",
              inline = FALSE
            ),
            conditionalPanel(
              condition = "input.stm_label_provider == 'openai'",
              selectizeInput(
                "stm_label_openai_model",
                "OpenAI Model:",
                choices = c("GPT-4.1 Mini (Default, fast)" = "gpt-4.1-mini", "GPT-4.1 (Accurate)" = "gpt-4.1", "GPT-4" = "gpt-4"),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("stm_label_openai_api_key", "API Key:", placeholder = "sk-..."),
              conditionalPanel(
                condition = "output.has_openai_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),
            conditionalPanel(
              condition = "input.stm_label_provider == 'gemini'",
              selectizeInput(
                "stm_label_gemini_model",
                "Gemini Model:",
                choices = c("Gemini 2.5 Flash Lite (Default, economy)" = "gemini-2.5-flash-lite", "Gemini 2.5 Flash" = "gemini-2.5-flash", "Gemini 2.5 Pro (Accurate)" = "gemini-2.5-pro"),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("stm_label_gemini_api_key", "API Key:", placeholder = "AIza..."),
              conditionalPanel(
                condition = "output.has_gemini_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),
            textAreaInput(
              "stm_system_prompt",
              "System prompt",
              value = "
You are a highly skilled data scientist specializing in generating concise and descriptive topic labels based on provided top terms for each topic.
Each topic consists of a list of terms ordered from most to least significant (by beta scores).

Your objective is to create precise labels that capture the essence of each topic by following these guidelines:

1. Use Person-First Language
   - Prioritize respectful and inclusive language.
   - Avoid terms that may be considered offensive or stigmatizing.
   - For example, use 'students with learning disabilities' instead of 'disabled students'.
   - Use 'students with visual impairments' instead of 'impaired students'
   - Use 'students with blindness' instead of 'blind students'.

1. Analyze Top Terms' Significance
   - Primary Focus: Emphasize high beta-score terms as they strongly define the topic.
   - Secondary Consideration: Include lower-scoring terms if they add essential context.

2. Synthesize the Topic Label
   - Clarity: Make sure the label is clear and easily understandable.
   - Conciseness: Aim for a short phrase of about 5-7 words.
   - Relevance: Reflect the collective meaning of the most influential terms.
   - Intelligent interpretation: Use your understanding to create meaningful labels that capture the topic's essence.

3. Maintain Consistency
   - Capitalize the first word of all topic labels.
   - Keep formatting and terminology uniform across all labels.
   - Avoid ambiguity or generic wording that does not fit the provided top terms.

4. Adhere to Style Guidelines
   - Capitalization: Use title case for labels.
   - Avoid Jargon: Maintain accessibility; only use technical terms if absolutely necessary.
   - Uniqueness: Ensure each label is distinct and does not overlap significantly with others.

5. Handle Edge Cases
   - Conflicting Top Terms: If the terms suggest different directions, prioritize those with higher beta scores.
   - Low-Scoring Terms: Include them only if they add meaningful context.

6. Iterative Improvement
   - If the generated label is insufficiently representative, re-check term significance and revise accordingly.
   - Always adhere to these guidelines.

Example
----------
Top Terms (highest to lowest beta score):
virtual manipulatives (.035)
manipulatives (.022)
mathematical (.014)
app (.013)
solving (.013)
learning disability (.012)
algebra (.012)
area (.011)
tool (.010)
concrete manipulatives (.010)

Generated Topic Label:
Mathematical learning tools for students with disabilities

Focus on incorporating the most significant keywords while following the guidelines above to produce a concise, descriptive topic label.
",
              rows = 10
            ),
            textAreaInput(
              "stm_user_prompt",
              "User prompt:",
              value = "You have a topic with keywords listed from most to least significant: [terms will be inserted here]. Please create a concise and descriptive label (5-7 words) that: 1. Reflects the collective meaning of these keywords. 2. Gives higher priority to the most significant terms. 3. Adheres to the style guidelines provided in the system message.",
              rows = 5
            ),
            sliderInput(
              "stm_temperature",
              "Temperature (creativity level)",
              min = 0, max = 1, value = 0.5, step = 0.1
            ),
            actionButton("topic_generate_labels", HTML("<i class=\"fas fa-wand-magic-sparkles\"></i> Generate Labels"), class = "btn-primary btn-block"),
            tags$hr(),
            textInput(
              "stm_label_topics",
              "Manually label topics",
              value = "",
              placeholder = "Type labels, use comma to separate"
            )
          ),
          conditionalPanel(
            condition = "input.topic_modeling_path == 'probability' && input.conditioned3 == 6",
            uiOutput("topic_number_uiOutput"),
            sliderInput(
              "stm_top_term_number_2",
              "Top terms per topic",
              value = 5,
              min = 1,
              max = 20
            ),
            actionButton("stm_display", "Display", class = "btn-primary btn-block")
          ),
          conditionalPanel(
            condition = "input.topic_modeling_path == 'probability' && input.conditioned3 == 7",
            tags$h5(strong("Explore example documents"), style = "color: #4269BF; margin-bottom: 10px;"),
            uiOutput("quote_topic_number_uiOutput"),
            selectInput("stm_topic_texts", "Example quotes to display", choices = NULL),
            actionButton("stm_quote", "Quote", class = "btn-primary btn-block")
          ),
          conditionalPanel(
            condition = "input.topic_modeling_path == 'probability' && input.conditioned3 == 8",
            tags$h5(
              strong("Estimate Covariate Effects on Topics"),
              style = "color: #4269BF; margin-bottom: 10px;"
            ),
            tags$div(
              class = "status-step-purple",
              style = "margin-bottom: 15px;",
              tags$i(class = "fa fa-check-circle status-icon status-icon-purple"),
              "Choose covariates in the Word-Topic tab"
            ),
            div(
              style = "display: flex; gap: 10px; margin-bottom: 15px;",
              div(
                style = "flex: 1;",
                actionButton("stm_effect", "Estimate", class = "btn-primary btn-block")
              ),
              div(
                style = "flex: 1;",
                downloadButton("stm_effect_download_table", class = "btn-secondary btn-block")
              )
            )
          ),
          conditionalPanel(
            condition = "input.topic_modeling_path == 'probability' && input.conditioned3 == 4 && input.searchKSubtabs == 'ai_rec'",
            tags$h5(strong("AI Configuration for K Selection"), style = "color: #4269BF; margin-bottom: 10px;"),

            radioButtons(
              "k_rec_provider",
              "AI Provider:",
              choices = c(
                "OpenAI (API Key Required)" = "openai",
                "Gemini (API Key Required)" = "gemini"
              ),
              selected = "openai",
              inline = FALSE
            ),

            conditionalPanel(
              condition = "input.k_rec_provider == 'openai'",
              selectizeInput(
                "k_rec_openai_model",
                "OpenAI Model:",
                choices = c(
                  "GPT-4.1 Mini (Default, fast)" = "gpt-4.1-mini",
                  "GPT-4.1 (Accurate)" = "gpt-4.1",
                  "GPT-4" = "gpt-4"
                ),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("k_rec_openai_api_key", "API Key:", placeholder = "sk-..."),
              conditionalPanel(
                condition = "output.has_openai_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),

            conditionalPanel(
              condition = "input.k_rec_provider == 'gemini'",
              selectizeInput(
                "k_rec_gemini_model",
                "Gemini Model:",
                choices = c(
                  "Gemini 2.5 Flash Lite (Default, economy)" = "gemini-2.5-flash-lite",
                  "Gemini 2.5 Flash" = "gemini-2.5-flash",
                  "Gemini 2.5 Pro (Accurate)" = "gemini-2.5-pro"
                ),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("k_rec_gemini_api_key", "API Key:", placeholder = "AIza..."),
              conditionalPanel(
                condition = "output.has_gemini_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),

            textAreaInput(
              "ai_system_search",
              "System prompt",
              value = "You are an expert in Structural Topic Modeling (STM) and topic model selection. Analyze the provided metrics to recommend the optimal number of topics (K).\n\nEvaluation criteria in order of importance:\n1. **Held-out likelihood**: Higher values indicate better generalization (most important for model selection)\n2. **Semantic coherence**: Higher values mean topics are more interpretable\n3. **Exclusivity**: Higher values mean topics are more distinct\n4. **Residuals**: Lower values indicate better model fit\n5. **Overall score**: A balanced metric combining all factors\n\nBest practice: Look for K values where:\n- Held-out likelihood plateaus or peaks\n- Semantic coherence and exclusivity are both reasonably high (upper-right quadrant)\n- There's an \"elbow\" in the residuals plot\n- The rate of improvement diminishes (diminishing returns)\n\nProvide a specific K recommendation with brief justification based on these metrics.",
              rows = 8
            ),
            textAreaInput(
              "ai_user_search",
              "User prompt",
              value = "Based on the Search K diagnostic plots and metrics table, what is the optimal number of topics (K)?\n\nKey observations:\n- Models with high coherence but low exclusivity may be too general\n- Models with high exclusivity but low coherence may be too specific\n- The ideal model balances both metrics while maximizing held-out likelihood\n- Consider practical interpretability: too few topics may oversimplify, too many may fragment meaningful themes\n\nPlease recommend a specific K value and explain which metrics most influenced your decision.",
              rows = 6
            ),
            actionButton(
              "generate_k_recommendation",
              HTML("<i class=\"fas fa-wand-magic-sparkles\"></i> Generate Recommendation"),
              class = "btn-primary btn-block"
            )
          ),
          conditionalPanel(
            condition = "input.topic_modeling_path == 'probability' && input.conditioned3 == 9",
            tags$h5(strong("Plot Topic Effects by Categorical Covariates"), style = "color: #4269BF; margin-bottom: 10px;"),
            tags$div(
              class = "status-step-purple",
              tags$i(class = "fa fa-check-circle status-icon status-icon-purple"),
              tags$strong("Step 1:"), " Select categorical covariates in Word-Topic tab"
            ),
            tags$div(
              class = "status-step-blue",
              tags$i(class = "fa fa-check-circle status-icon status-icon-info"),
              tags$strong("Step 2:"), " Estimate coefficients in Estimated Effects tab"
            ),
            selectizeInput(
              "stm_effect_cat_btn",
              "Categorical covariate",
              choices = NULL,
              multiple = FALSE
            ),
            sliderInput(
              "stm_ncol_cat",
              "Column numbers",
              value = 2,
              min = 1,
              max = 10
            ),
            actionButton("stm_display_cat", "Display", class = "btn-primary btn-block")
          ),
          conditionalPanel(
            condition = "input.topic_modeling_path == 'probability' && input.conditioned3 == 10",
            tags$h5(strong("Plot Topic Effects by Continuous Covariates"), style = "color: #4269BF; margin-bottom: 10px;"),
            tags$div(
              class = "status-step-purple",
              tags$i(class = "fa fa-check-circle status-icon status-icon-purple"),
              tags$strong("Step 1:"), " Select continuous covariates in Word-Topic tab"
            ),
            tags$div(
              class = "status-step-blue",
              tags$i(class = "fa fa-check-circle status-icon status-icon-info"),
              tags$strong("Step 2:"), " Estimate coefficients in Estimated Effects tab"
            ),
            selectizeInput(
              "stm_effect_con_btn",
              "Continuous covariate",
              choices = NULL,
              multiple = FALSE
            ),
            sliderInput(
              "stm_ncol_con",
              "Column numbers",
              value = 2,
              min = 1,
              max = 10
            ),
            actionButton("stm_display_con", "Display", class = "btn-primary btn-block")
          ),

          conditionalPanel(
            condition = "input.conditioned3 == 'ai_content'",
            tags$h5(HTML("<strong>Generate content using AI</strong>"), style = "color: #4269BF; margin-bottom: 15px;"),
            tags$div(
              class = "status-main-info",
              tags$i(class = "fas fa-robot status-icon status-icon-info"),
              "Generate survey items, research questions, or other content from topic terms."
            ),
            selectInput(
              "content_type",
              "Content type",
              choices = c(
                "Survey Item" = "survey_item",
                "Research Question" = "research_question",
                "Theme Description" = "theme_description",
                "Policy Recommendation" = "policy_recommendation",
                "Interview Question" = "interview_question",
                "Custom" = "custom"
              ),
              selected = "survey_item"
            ),
            sliderInput(
              "content_top_terms",
              "Top terms for content",
              value = 7,
              min = 3,
              max = 15
            ),
            sliderInput(
              "content_max_tokens",
              "Max tokens",
              min = 50, max = 500, value = 150, step = 25
            ),
            radioButtons(
              "content_provider",
              "AI Provider:",
              choices = c(
                "OpenAI (API Key Required)" = "openai",
                "Gemini (API Key Required)" = "gemini"
              ),
              selected = "openai",
              inline = FALSE
            ),
            conditionalPanel(
              condition = "input.content_provider == 'openai'",
              selectizeInput(
                "content_openai_model",
                "OpenAI Model:",
                choices = c("GPT-4.1 Mini (Default, fast)" = "gpt-4.1-mini", "GPT-4.1 (Accurate)" = "gpt-4.1", "GPT-4" = "gpt-4"),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("content_openai_api_key", "API Key:", placeholder = "sk-..."),
              conditionalPanel(
                condition = "output.has_openai_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),
            conditionalPanel(
              condition = "input.content_provider == 'gemini'",
              selectizeInput(
                "content_gemini_model",
                "Gemini Model:",
                choices = c("Gemini 2.5 Flash Lite (Default, economy)" = "gemini-2.5-flash-lite", "Gemini 2.5 Flash" = "gemini-2.5-flash", "Gemini 2.5 Pro (Accurate)" = "gemini-2.5-pro"),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("content_gemini_api_key", "API Key:", placeholder = "AIza..."),
              conditionalPanel(
                condition = "output.has_gemini_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),
            textAreaInput(
              "content_system_prompt",
              "System prompt",
              value = "",
              rows = 8
            ),
            textAreaInput(
              "content_user_prompt",
              "User prompt",
              value = "",
              rows = 5
            ),
            sliderInput(
              "content_temperature",
              "Temperature (creativity level)",
              min = 0, max = 1, value = 0.5, step = 0.1
            ),
            actionButton(
              "generate_topic_content",
              HTML("<i class=\"fas fa-pen-fancy\"></i> Generate Content"),
              class = "btn-primary btn-block",
              style = "margin-top: 15px;"
            )
          ),
          conditionalPanel(
            condition = "input.topic_modeling_path == 'embedding' && input.conditioned3 == 4",
            tags$h5(HTML("<strong>Configure Embedding-based Topic Modeling</strong>"), style = "color: #4269BF; margin-bottom: 10px;"),

            selectInput(
              "embedding_backend",
              "Backend:",
              choices = c(
                "Python (BERTopic, recommended)" = "python",
                "R (no Python required)" = "r"
              ),
              selected = "python"
            ),

            uiOutput("embedding_topic_provider_status"),

            conditionalPanel(
              condition = "input.embedding_backend == 'python'",
              uiOutput("embedding_topic_backend_status")
            ),

            radioButtons(
              "topic_embedding_provider",
              "AI Provider:",
              choices = c(
                "Sentence Transformers (Python)" = "sentence-transformers",
                "OpenAI (API Key Required)" = "openai",
                "Gemini (API Key Required)" = "gemini"
              ),
              selected = "sentence-transformers",
              inline = FALSE
            ),

            conditionalPanel(
              condition = "input.topic_embedding_provider == 'sentence-transformers'",
              selectizeInput(
                "topic_embedding_st_model",
                "Model:",
                choices = c(
                  "all-MiniLM-L6-v2 (Fast)" = "all-MiniLM-L6-v2",
                  "all-mpnet-base-v2 (Higher Quality)" = "all-mpnet-base-v2",
                  "BGE Small EN v1.5 (Fast + Strong Retrieval)" = "BAAI/bge-small-en-v1.5",
                  "BGE Base EN v1.5 (BERTopic Optimized)" = "BAAI/bge-base-en-v1.5",
                  "E5 Base v2 (Instructor-tuned)" = "intfloat/e5-base-v2",
                  "Nomic Embed Text v2 (Multilingual)" = "nomic-ai/nomic-embed-text-v2-moe",
                  "GTE Multilingual Base (Fast, Multilingual)" = "Alibaba-NLP/gte-multilingual-base"
                ),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              tags$p(
                style = "font-size: 16px; color: #666;",
                "Requires Python + sentence-transformers"
              )
            ),

            conditionalPanel(
              condition = "input.topic_embedding_provider == 'openai'",
              selectizeInput(
                "topic_embedding_openai_model",
                "OpenAI Model:",
                choices = c(
                  "Text Embedding 3 Small (Default)" = "text-embedding-3-small",
                  "Text Embedding 3 Large (Higher Quality)" = "text-embedding-3-large"
                ),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("topic_embedding_openai_api_key", "API Key:", placeholder = "sk-..."),
              conditionalPanel(
                condition = "output.has_openai_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),

            conditionalPanel(
              condition = "input.topic_embedding_provider == 'gemini'",
              selectizeInput(
                "topic_embedding_gemini_model",
                "Gemini Model:",
                choices = c("Gemini Embedding 001" = "gemini-embedding-001"),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("topic_embedding_gemini_api_key", "API Key:", placeholder = "AIza..."),
              conditionalPanel(
                condition = "output.has_gemini_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),

            conditionalPanel(
              condition = "input.embedding_backend == 'python'",
              sliderInput(
                "embedding_umap_neighbors",
                "Neighbors:",
                value = 15,
                min = 5,
                max = 50
              ),
              sliderInput(
                "embedding_umap_n_components",
                "UMAP dimensions:",
                value = 5,
                min = 2,
                max = 50
              ),
              sliderInput(
                "embedding_umap_min_dist",
                "Min distance (0.0 = tight clusters for topic modeling):",
                value = 0.0,
                min = 0.0,
                max = 0.99,
                step = 0.01
              ),
              selectInput(
                "embedding_umap_metric",
                "Distance metric:",
                choices = c(
                  "Cosine (recommended for text)" = "cosine",
                  "Euclidean" = "euclidean"
                ),
                selected = "cosine"
              ),
              sliderInput(
                "embedding_min_topic_size",
                "Min cluster size:",
                value = 10,
                min = 2,
                max = 50
              ),
              selectInput(
                "embedding_cluster_selection",
                "Cluster selection method:",
                choices = c(
                  "EOM (broader topics, default)" = "eom",
                  "Leaf (smaller, focused topics)" = "leaf"
                ),
                selected = "eom"
              )
            ),

            conditionalPanel(
              condition = "input.embedding_backend == 'r'",
              selectInput(
                "embedding_dimred_method",
                "Dimensionality reduction:",
                choices = c(
                  "UMAP" = "umap",
                  "t-SNE" = "tsne",
                  "PCA" = "pca"
                ),
                selected = "umap"
              ),

              conditionalPanel(
                condition = "input.embedding_dimred_method == 'umap'",
                sliderInput(
                  "embedding_r_umap_neighbors",
                  "Neighbors:",
                  value = 15,
                  min = 5,
                  max = 50
                ),
                sliderInput(
                  "embedding_r_umap_n_components",
                  "UMAP dimensions:",
                  value = 5,
                  min = 2,
                  max = 50
                ),
                sliderInput(
                  "embedding_r_umap_min_dist",
                  "Min distance (0.0 = tight clusters for topic modeling):",
                  value = 0.0,
                  min = 0.0,
                  max = 0.5,
                  step = 0.01
                ),
                selectInput(
                  "embedding_r_umap_metric",
                  "Distance metric:",
                  choices = c(
                    "Cosine (recommended for text)" = "cosine",
                    "Euclidean" = "euclidean"
                  ),
                  selected = "cosine"
                )
              ),

              conditionalPanel(
                condition = "input.embedding_dimred_method == 'tsne'",
                sliderInput(
                  "embedding_tsne_perplexity",
                  "Perplexity:",
                  value = 30,
                  min = 5,
                  max = 100
                )
              ),

              conditionalPanel(
                condition = "input.embedding_dimred_method == 'pca'",
                sliderInput(
                  "embedding_pca_dims",
                  "Components:",
                  value = 50,
                  min = 10,
                  max = 200
                )
              ),

              selectInput(
                "embedding_method_r",
                "Clustering method:",
                choices = c(
                  "DBSCAN (density-based)" = "dbscan",
                  "K-means" = "kmeans",
                  "Hierarchical" = "hierarchical",
                  "HDBSCAN" = "hdbscan"
                ),
                selected = "dbscan"
              ),

              conditionalPanel(
                condition = "input.embedding_method_r == 'kmeans' || input.embedding_method_r == 'hierarchical'",
                numericInput(
                  "embedding_r_n_topics",
                  "Number of clusters:",
                  value = 5,
                  min = 2,
                  max = 50
                )
              ),

              conditionalPanel(
                condition = "input.embedding_method_r == 'dbscan'",
                sliderInput(
                  "embedding_dbscan_eps",
                  "Epsilon:",
                  value = 0.5,
                  min = 0.1,
                  max = 2.0,
                  step = 0.1
                ),
                sliderInput(
                  "embedding_dbscan_minpts",
                  "Min points:",
                  value = 5,
                  min = 2,
                  max = 20
                )
              ),

              conditionalPanel(
                condition = "input.embedding_method_r == 'hdbscan'",
                sliderInput(
                  "embedding_r_min_cluster_size",
                  "Min cluster size:",
                  value = 5,
                  min = 2,
                  max = 50
                )
              ),

              checkboxInput(
                "embedding_r_reduce_outliers",
                "Reduce outliers (for DBSCAN/HDBSCAN)",
                value = TRUE
              )
            ),

            actionButton("embedding_run", "Run Model", class = "btn-primary btn-block")
          ),

          conditionalPanel(
            condition = "input.topic_modeling_path == 'embedding' && input.conditioned3 == 5",
            uiOutput("embedding_topics_info"),
            actionButton("embedding_display", "Display", class = "btn-primary btn-block")
          ),

          conditionalPanel(
            condition = "input.topic_modeling_path == 'embedding' && input.conditioned3 == 6",
            selectInput(
              "embedding_viz_type",
              "Visualization type",
              choices = c(
                "Topic Distribution" = "distribution",
                "Document Clusters" = "documents",
                "Topic Similarity Heatmap" = "heatmap",
                "Intertopic Distance" = "distance"
              ),
              selected = "distribution"
            )
          ),

          conditionalPanel(
            condition = "input.topic_modeling_path == 'embedding' && input.conditioned3 == 7",
            sliderInput(
              "embedding_quote_number",
              "Quotes to display",
              value = 3,
              min = 1,
              max = 10
            ),
            selectInput(
              "embedding_quote_topic",
              "Topic",
              choices = NULL
            ),
            actionButton("embedding_quote", "Quote", class = "btn-primary btn-block")
          ),






        ),
        mainPanel(
          width = 9,
          tabsetPanel(
            id = "conditioned3",
            tabPanel(
              "1. Model Configuration",
              value = 4,
              shinyBS::bsCollapse(
                open = 0,
                shinyBS::bsCollapsePanel(
                  p(strong("Click to set plot dimensions"),
                    class = "plot-dimensions-text"
                  ),
                  value = 4,
                  style = "success",
                  p(strong("Dimensions of the plots")),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "height_search_k",
                      post = " px",
                      label = "height",
                      min = 200,
                      max = 4000,
                      step = 5,
                      value = 600
                    )
                  ),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "width_search_k",
                      post = " px",
                      label = "width",
                      min = 500,
                      max = 3000,
                      step = 5,
                      value = 1000
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_search_k_results == true",
                tabsetPanel(
                  id = "searchKSubtabs",
                  tabPanel(
                    "Diagnostic Plots",
                    value = "diagnostic",
                    br(),
                    uiOutput("topic_search_message"),
                    br(),
                    fluidRow(
                      column(6, uiOutput("quality_metrics_semcoh_uiOutput")),
                      column(6, uiOutput("quality_metrics_residual_uiOutput"))
                    ),
                    fluidRow(
                      column(6, uiOutput("quality_metrics_heldout_uiOutput")),
                      column(6, uiOutput("quality_metrics_lbound_uiOutput"))
                    )
                  ),
                  tabPanel(
                    "Quality Comparison",
                    value = "quality",
                    br(),
                    wellPanel(
                      style = "background-color: #f0f8ff; border: 1px solid #4682b4;",
                      p("Overall Score = Coherence(z) + Exclusivity(z) - Residual(z) + Heldout(z)",
                        style = "font-family: monospace; font-size: 16px; font-weight: bold;"),
                      tags$div(
                        style = "font-size: 16px; margin-top: 10px; line-height: 1.6;",
                        p(tags$b("Coherence:"), " How interpretable topics are based on co-occurring words", style = "margin: 5px 0;"),
                        p(tags$b("Exclusivity:"), " How distinctive topics are from each other", style = "margin: 5px 0;"),
                        p(tags$b("Residual:"), " Model fit to data (lower is better)", style = "margin: 5px 0;"),
                        p(tags$b("Heldout:"), " Model's ability to generalize to new data", style = "margin: 5px 0;")
                      )
                    ),
                    br(),
                    DT::dataTableOutput("quality_summary_table")
                  ),
                  tabPanel(
                    "Coherence vs Exclusivity",
                    value = "comparison",
                    br(),
                    uiOutput("model_comparison_plot_uiOutput")
                  ),
                  tabPanel(
                    "AI Recommendation",
                    value = "ai_rec",
                    br(),
                    uiOutput("ai_recommendation_output")
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_search_k_results == false && input.topic_modeling_path == 'probability'",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-search-plus", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Configure K range and click ",
                      tags$strong("'Search K'", style = "color: #4269BF;"),
                      " to find optimal topic numbers",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "input.topic_modeling_path == 'embedding'",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-cogs", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Configure settings and click ",
                      tags$strong("'Run Model'", style = "color: #4269BF;"),
                      " to discover topics",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
            ),
            tabPanel(
              "2. Word-Topic",
              value = 5,
              shinyBS::bsCollapse(
                open = 0,
                shinyBS::bsCollapsePanel(
                  p(strong("Click to set plot dimensions"),
                    class = "plot-dimensions-text"
                  ),
                  value = 1,
                  style = "success",
                  p(strong("Dimensions of the plot")),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "height",
                      post = " px",
                      label = "height",
                      min = 200,
                      max = 4000,
                      step = 5,
                      value = 1000
                    )
                  ),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "width",
                      post = " px",
                      label = "width",
                      min = 500,
                      max = 3000,
                      step = 5,
                      value = 1000
                    )
                  )
                )
              ),
              tags$style(
                HTML(
                  ".plot-container {
                                max-height: 4000px;
                                max-width: 3000px;
                                overflow: auto; }"
                )
              ),
              conditionalPanel(
                condition = "output.has_word_topic_results == true",
                uiOutput("topic_term_message"),
                uiOutput("topic_term_plot_uiOutput"),
                br(),
                uiOutput("topic_term_table_uiOutput"),
                br()
              ),
              conditionalPanel(
                condition = "output.has_word_topic_results == false && input.topic_modeling_path == 'probability'",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-project-diagram", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Search K, and then click ",
                      tags$strong("'Display'", style = "color: #4269BF;"),
                      " to view word-topic distributions",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_word_topic_results == false && input.topic_modeling_path == 'embedding'",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-project-diagram", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Run model, then click ",
                      tags$strong("'Display'", style = "color: #4269BF;"),
                      " to view topic keywords",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
            ),
            tabPanel(
              "3. Content Generation",
              value = "ai_content",
              div(
                style = "padding: 20px;",

                conditionalPanel(
                  condition = "output.has_generated_content == true",
                  DT::dataTableOutput("generated_content_table")
                ),
                conditionalPanel(
                  condition = "output.has_generated_content == false",
                  div(
                    style = "padding: 60px 40px; text-align: center;",
                    tags$i(class = "fas fa-file-alt", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Run topic modeling first, then configure settings in the sidebar and click ",
                      tags$strong("'Generate Content'", style = "color: #4269BF;"),
                      " to create content from your topics.",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),

            tabPanel(
              "4. Document-Topic",
              value = 6,
              shinyBS::bsCollapse(
                open = 0,
                shinyBS::bsCollapsePanel(
                  p(strong("Click to set plot dimensions"),
                    class = "plot-dimensions-text"
                  ),
                  value = 1,
                  style = "success",
                  p(strong("Dimensions of the plot")),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "height_topic_prevalence",
                      post = " px",
                      label = "height",
                      min = 200,
                      max = 4000,
                      step = 5,
                      value = 500
                    )
                  ),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "width_topic_prevalence",
                      post = " px",
                      label = "width",
                      min = 500,
                      max = 3000,
                      step = 5,
                      value = 1000
                    )
                  )
                )
              ),
              tags$style(
                HTML(
                  ".plot-container {
                                max-height: 4000px;
                                max-width: 3000px;
                                overflow: auto; }"
                )
              ),
              conditionalPanel(
                condition = "output.has_document_topic_results == true",
                uiOutput("topic_prevalence_plot_uiOutput"),
                br(),
                uiOutput("topic_prevalence_table_uiOutput")
              ),
              conditionalPanel(
                condition = "output.has_document_topic_results == false && input.topic_modeling_path == 'probability'",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-file-alt", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Complete ",
                      tags$strong("Word-Topic tab", style = "color: #4269BF;"),
                      " to view document-topic distributions",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_document_topic_results == false && input.topic_modeling_path == 'embedding'",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-file-alt", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Complete ",
                      tags$strong("Word-Topic tab", style = "color: #4269BF;"),
                      " to view document-topic distributions",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
            ),
            tabPanel(
              "5. Quotes",
              value = 7,
              conditionalPanel(
                condition = "output.has_quotes == true && input.topic_modeling_path == 'probability'",
                DT::dataTableOutput("quote_table")
              ),
              conditionalPanel(
                condition = "output.has_quotes == true && input.topic_modeling_path == 'embedding'",
                DT::dataTableOutput("embedding_quote_table")
              ),
              conditionalPanel(
                condition = "output.has_quotes == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-quote-right", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Run topic model and select a topic to view ",
                      tags$strong("representative quotes", style = "color: #4269BF;"),
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "6. Estimated Effects",
              value = 8,
              conditionalPanel(
                condition = "output.has_effect_estimates == true",
                DT::dataTableOutput("effect_table")
              ),
              conditionalPanel(
                condition = "output.has_effect_estimates == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-calculator", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Click ",
                      tags$strong("'Estimate'", style = "color: #4269BF;"),
                      " button to generate effect estimates",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "7. Categorical Covariates",
              value = 9,
              shinyBS::bsCollapse(
                open = 0,
                shinyBS::bsCollapsePanel(
                  p(strong("Click to set plot dimensions"),
                    class = "plot-dimensions-text"
                  ),
                  value = 1,
                  style = "success",
                  p(strong("Dimensions of the plot")),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "height_cat_plot",
                      post = " px",
                      label = "height",
                      min = 200,
                      max = 4000,
                      step = 5,
                      value = 500
                    )
                  ),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "width_cat_plot",
                      post = " px",
                      label = "width",
                      min = 500,
                      max = 3000,
                      step = 5,
                      value = 1000
                    )
                  )
                )
              ),
              tags$style(
                HTML(
                  ".plot-container {
                                max-height: 4000px;
                                max-width: 3000px;
                                overflow: auto; }"
                )
              ),
              conditionalPanel(
                condition = "output.has_categorical_plot == true",
                uiOutput("cat_plot_uiOutput"),
                br(),
                uiOutput("cat_table_uiOutput")
              ),
              conditionalPanel(
                condition = "output.has_categorical_plot == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-chart-bar", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Estimate effects, select categorical covariate, then click ",
                      tags$strong("'Display'", style = "color: #4269BF;"),
                      " to visualize topic prevalence by categories",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "8. Continuous Covariates",
              value = 10,
              shinyBS::bsCollapse(
                open = 0,
                shinyBS::bsCollapsePanel(
                  p(strong("Click to set plot dimensions"),
                    class = "plot-dimensions-text"
                  ),
                  value = 1,
                  style = "success",
                  p(strong("Dimensions of the plot")),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "height_con_plot",
                      post = " px",
                      label = "height",
                      min = 200,
                      max = 4000,
                      step = 5,
                      value = 500
                    )
                  ),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "width_con_plot",
                      post = " px",
                      label = "width",
                      min = 500,
                      max = 3000,
                      step = 5,
                      value = 1000
                    )
                  )
                )
              ),
              tags$style(
                HTML(
                  ".plot-container {
                                max-height: 4000px;
                                max-width: 3000px;
                                overflow: auto; }"
                )
              ),
              conditionalPanel(
                condition = "output.has_continuous_plot == true",
                uiOutput("con_plot_uiOutput"),
                br(),
                uiOutput("con_table_uiOutput")
              ),
              conditionalPanel(
                condition = "output.has_continuous_plot == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-chart-line", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Estimate effects, select continuous covariate, then click ",
                      tags$strong("'Display'", style = "color: #4269BF;"),
                      " to visualize topic prevalence trends",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            )
          )
        )
      )
}

lexical_analysis_ui_content <- function() {
      sidebarLayout(
        sidebarPanel(
          width = 3,
          class = "sidebar-panel",
          conditionalPanel(
            condition = "input.conditioned2 == 1 && input.linguistic_subtabs == 'lemmas'",
            tags$h5(
              HTML("<strong>Linguistic Analysis</strong> <a href='https://spacy.io/usage/linguistic-features#lemmatization' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"),
              tags$span("OPTIONAL", style = "background-color: #6c757d; color: white; padding: 2px 8px; border-radius: 3px; font-size: 13px; margin-left: 8px;"),
              style = "color: #4269BF; margin-bottom: 10px;"
            ),
            div(
              style = "display: flex; gap: 10px; margin-bottom: 15px;",
              div(
                style = "flex: 1;",
                actionButton("lemma", "Apply", class = "btn-primary btn-block")
              ),
              div(
                style = "flex: 1;",
                actionButton("skip", "Skip", class = "btn-secondary btn-block")
              )
            )
          ),
          conditionalPanel(
            condition = "input.conditioned2 == 1 && input.linguistic_subtabs == 'pos'",
            tags$div(
              style = "display: flex; align-items: center; justify-content: space-between; margin-bottom: 10px;",
              tags$h5(HTML("<strong>Part-of-Speech Configuration</strong> <a href='https://universaldependencies.org/u/pos/' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"), style = "color: #4269BF; margin: 0;"),
              actionLink("showPOSInfo", icon("info-circle"),
                        style = "color: #337ab7; font-size: 16px;",
                        title = "Click for POS tags guide",
                        `aria-label` = "POS tags guide")
            ),
            selectizeInput(
              "pos_filter",
              "Filter by POS tag",
              choices = NULL,
              selected = NULL,
              multiple = TRUE,
              options = list(
                placeholder = "All POS tags (optional)"
              )
            ),
            sliderInput(
              "pos_top_n",
              "Top tags to display",
              value = 15,
              min = 1,
              max = 20,
              step = 1
            ),
            uiOutput("spacy_mode_ui"),
            checkboxInput(
              "include_dependency",
              "Show dependency relations",
              value = FALSE
            ),
            div(
              style = "display: flex; gap: 10px; margin-bottom: 15px;",
              div(
                style = "flex: 1;",
                actionButton("apply_pos", "Apply", class = "btn-primary btn-block")
              ),
              div(
                style = "flex: 1;",
                actionButton("generate_pos_report", "Report", class = "btn-primary btn-block", icon = icon("file-alt"))
              )
            )
          ),
          # Morphological Features sidebar
          conditionalPanel(
            condition = "input.conditioned2 == 1 && input.linguistic_subtabs == 'morphology'",
            div(
              style = "display: flex; align-items: center; justify-content: space-between; margin-bottom: 10px;",
              tags$h5(HTML("<strong>Morphology Analysis</strong>"), style = "color: #4269BF; margin: 0;"),
              actionLink("showMorphInfo", icon("info-circle"),
                         style = "color: #337ab7; font-size: 16px;",
                         title = "Click for morphology guide",
                         `aria-label` = "Morphology guide")
            ),
            tags$p(
              "Run POS tagging first, then select features to analyze.",
              style = "font-size: 16px; color: #475569; margin-bottom: 10px;"
            ),
            uiOutput("morph_status_ui"),
            checkboxGroupInput(
              "morph_features",
              NULL,
              choices = c(
                "Number (Sing/Plur)" = "Number",
                "Tense (Past/Pres/Fut)" = "Tense",
                "Verb Form (Fin/Inf/Part)" = "VerbForm",
                "Person (1/2/3)" = "Person",
                "Case (Nom/Acc/Gen)" = "Case",
                "Mood (Ind/Imp/Sub)" = "Mood",
                "Aspect (Perf/Imp/Prog)" = "Aspect"
              ),
              selected = c("Number", "Tense", "VerbForm", "Person")
            ),
            actionButton(
              "analyze_morphology",
              "Analyze Morphology",
              class = "btn-primary btn-block",
              icon = icon("language")
            )
          ),
          # Dependency Parsing sidebar
          conditionalPanel(
            condition = "input.conditioned2 == 1 && input.linguistic_subtabs == 'dependencies'",
            div(
              style = "display: flex; align-items: center; justify-content: space-between; margin-bottom: 10px;",
              tags$h5(HTML("<strong>Dependency Parsing</strong> <a href='https://universaldependencies.org/u/dep/' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"), style = "color: #4269BF; margin: 0;"),
              actionLink("showDepInfo", icon("info-circle"),
                         style = "color: #337ab7; font-size: 16px;",
                         title = "Click for dependency relations guide",
                         `aria-label` = "Dependency relations guide")
            )
          ),
          conditionalPanel(
            condition = "input.conditioned2 == 2",
            tags$h5(strong("Word Frequency by Variable"), style = "color: #4269BF; margin-bottom: 10px;"),
            selectizeInput(
              "continuous_var_3",
              "Continuous variable",
              choices = NULL,
              multiple = FALSE
            ),
            selectizeInput(
              "type_terms",
              "Terms",
              choices = NULL,
              options = list(maxItems = 20, multiple = TRUE, placeholder = "Select or type terms")
            ),
            sliderInput(
              "height_line_con_var_plot",
              "Plot height",
              value = 500,
              min = 200,
              max = 2000,
              step = 50
            ),
            sliderInput(
              "width_line_con_var_plot",
              "Plot width",
              value = 1000,
              min = 500,
              max = 2000,
              step = 50
            ),
            actionButton("plot_term", "Plot Terms", class = "btn-primary btn-block")
          ),
          conditionalPanel(
            condition = "input.conditioned2 == 4",
            tags$div(
              style = "display: flex; align-items: center; justify-content: space-between; margin-bottom: 10px;",
              tags$h5(HTML("<strong>Lexical Diversity Analysis</strong> <a href='https://github.com/quanteda/quanteda.textstats' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px; color: #337ab7;'>Source</a>"), style = "color: #4269BF; margin: 0;"),
              actionLink("showLexDivMetricsInfo", icon("info-circle"),
                        style = "color: #337ab7; font-size: 16px;",
                        title = "Click for detailed metric descriptions",
                        `aria-label` = "Lexical diversity metric descriptions")
            ),
            selectizeInput(
              "lexdiv_doc_id_var",
              "Document ID variable",
              choices = NULL,
              selected = "",
              options = list(
                allowEmptyOption = TRUE,
                persist = TRUE,
                placeholder = "Optional"
              )
            ),
            br(),
            checkboxGroupInput(
              "lexdiv_metrics",
              "Metrics to calculate",
              choices = c(
                "MTLD (Most Recommended)" = "MTLD",
                "MATTR (Recommended)" = "MATTR",
                "MSTTR (Mean Segmental TTR)" = "MSTTR",
                "TTR (Type-Token Ratio)" = "TTR",
                "CTTR (Corrected TTR)" = "CTTR",
                "Herdan's C" = "C",
                "Guiraud's R" = "R",
                "Maas" = "Maas",
                "Yule's K" = "K",
                "Simpson's D" = "D"
              ),
              selected = c("MTLD", "MATTR", "TTR")
            ),
            actionButton("run_lexdiv_analysis", "Analyze", class = "btn-primary btn-block")
          ),
          conditionalPanel(
            condition = "input.conditioned2 == 5",
            tags$div(
              style = "display: flex; align-items: center; justify-content: space-between; margin-bottom: 10px;",
              tags$h5(HTML("<strong>Readability Analysis</strong> <a href='https://github.com/quanteda/quanteda.textstats' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"), style = "color: #4269BF; margin: 0;"),
              actionLink("showReadabilityMetricsInfo", icon("info-circle"),
                        style = "color: #337ab7; font-size: 16px;",
                        title = "Click for detailed metric descriptions",
                        `aria-label` = "Readability metric descriptions")
            ),
            selectizeInput(
              "readability_doc_id_var",
              "Document ID variable",
              choices = NULL,
              selected = "",
              options = list(
                allowEmptyOption = TRUE,
                persist = TRUE,
                placeholder = "Optional"
              )
            ),
            br(),
            checkboxGroupInput(
              "readability_metrics",
              "Metrics to calculate",
              choices = c(
                "Flesch Reading Ease" = "flesch",
                "Flesch-Kincaid Grade" = "flesch_kincaid",
                "Gunning Fog" = "gunning_fog",
                "SMOG" = "smog",
                "Automated Readability Index" = "ari",
                "Coleman-Liau" = "coleman_liau"
              ),
              selected = c("flesch", "flesch_kincaid", "gunning_fog")
            ),
            tags$hr(),
            tags$h5(strong("Visualization Options"), style = "color: #4269BF; margin-top: 10px; margin-bottom: 10px;"),
            radioButtons(
              "readability_view_type",
              "View:",
              choices = c(
                "Overall Distribution" = "distribution",
                "Group Comparison" = "group",
                "Top Documents" = "document"
              ),
              selected = "distribution"
            ),
            conditionalPanel(
              condition = "input.readability_view_type == 'group'",
              selectizeInput(
                "readability_group_var",
                "Group by:",
                choices = NULL,
                selected = NULL,
                multiple = FALSE
              )
            ),
            conditionalPanel(
              condition = "input.readability_view_type == 'document'",
              sliderInput(
                "readability_top_n",
                "Number of documents",
                value = 15,
                min = 1,
                max = 30,
                step = 1
              )
            ),
            actionButton("run_readability_analysis", "Analyze", class = "btn-primary btn-block")
          ),
          conditionalPanel(
            condition = "input.conditioned2 == 3",
            tags$h5(HTML("<strong>Keyword Extraction</strong> <a href='https://quanteda.io/reference/dfm_tfidf.html' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"), style = "color: #4269BF; margin-bottom: 10px;"),
            conditionalPanel(
              condition = "input.keywords_subtabs == 'tfidf'",
              sliderInput(
                "tfidf_top_n",
                "Number of keywords",
                value = 20,
                min = 1,
                max = 50,
                step = 1
              ),
              checkboxInput(
                "tfidf_normalize",
                "Normalize TF-IDF scores",
                value = TRUE
              ),
              tags$p(
                "Scales scores to 0-1 range.",
                style = "font-size: 16px; color: #475569; margin-top: -5px; margin-bottom: 10px;"
              )
            ),
            conditionalPanel(
              condition = "input.keywords_subtabs == 'textrank'",
              sliderInput(
                "textrank_top_n",
                "Number of keywords",
                value = 15,
                min = 1,
                max = 30,
                step = 1
              ),
              selectizeInput(
                "tfidf_group_var",
                "Compare groups (for keyness)",
                choices = NULL,
                multiple = FALSE,
                options = list(placeholder = "Select grouping variable")
              ),
              tags$p(
                "Finds words used significantly more in one group vs others.",
                style = "font-size: 16px; color: #475569; margin-top: -5px; margin-bottom: 10px;"
              )
            ),
            conditionalPanel(
              condition = "input.keywords_subtabs == 'comparison'",
              sliderInput(
                "comparison_top_n",
                "Keywords to display",
                value = 10,
                min = 1,
                max = 20,
                step = 1
              )
            ),
            actionButton("run_keyword_extraction", "Extract", class = "btn-primary btn-block")
          ),
          # Log Odds Ratio Analysis sidebar
          conditionalPanel(
            condition = "input.conditioned2 == 6",
            tags$h5(HTML("<strong>Log Odds Ratio Analysis</strong>"), style = "color: #4269BF; margin-bottom: 10px;"),
            tags$p(
              "Compare word frequencies between categories to identify distinctive terms.",
              style = "font-size: 16px; color: #475569; margin-bottom: 10px;"
            ),
            selectizeInput(
              "log_odds_group_var",
              "Categorical variable",
              choices = NULL,
              multiple = FALSE,
              options = list(placeholder = "Select category variable")
            ),
            radioButtons(
              "log_odds_method",
              "Method",
              choices = c(
                "Weighted (Fightin' Words)" = "weighted",
                "Simple (Laplace smoothing)" = "simple"
              ),
              selected = "weighted"
            ),
            conditionalPanel(
              condition = "input.log_odds_method == 'simple'",
              radioButtons(
                "log_odds_comparison_mode",
                "Comparison mode",
                choices = c(
                  "Binary (2 categories)" = "binary",
                  "Pairwise (all pairs)" = "pairwise"
                ),
                selected = "binary"
              )
            ),
            conditionalPanel(
              condition = "input.log_odds_method == 'simple' && input.log_odds_comparison_mode == 'binary'",
              selectizeInput(
                "log_odds_reference",
                "Reference category",
                choices = NULL,
                multiple = FALSE,
                options = list(placeholder = "First category")
              )
            ),
            sliderInput(
              "log_odds_top_n",
              "Top terms per direction",
              value = 10,
              min = 1,
              max = 25,
              step = 1
            ),
            sliderInput(
              "log_odds_min_count",
              "Minimum word count",
              value = 5,
              min = 1,
              max = 20,
              step = 1
            ),
            actionButton("run_log_odds", "Calculate", class = "btn-primary btn-block", icon = icon("balance-scale"))
          ),
          # Lexical Dispersion sidebar
          conditionalPanel(
            condition = "input.conditioned2 == 7",
            tags$h5(HTML("<strong>Lexical Dispersion</strong>"), style = "color: #4269BF; margin-bottom: 10px;"),
            tags$p(
              "Visualize where selected terms appear across documents in your corpus.",
              style = "font-size: 16px; color: #475569; margin-bottom: 10px;"
            ),
            selectizeInput(
              "dispersion_terms",
              "Terms",
              choices = NULL,
              multiple = TRUE,
              options = list(
                placeholder = "Type or select terms...",
                maxItems = 10,
                create = TRUE
              )
            ),
            radioButtons(
              "dispersion_scale",
              "Position scale",
              choices = c(
                "Relative (0-1)" = "relative",
                "Absolute (token position)" = "absolute"
              ),
              selected = "relative"
            ),
            checkboxInput(
              "dispersion_show_metrics",
              "Show dispersion metrics",
              value = TRUE
            ),
            actionButton("run_dispersion", "Analyze", class = "btn-primary btn-block", icon = icon("chart-line"))
          ),
          conditionalPanel(
            condition = "input.conditioned2 == 1 && input.linguistic_subtabs == 'dependencies'",
            selectizeInput(
              "dep_doc_id_var",
              "Document ID variable",
              choices = NULL,
              selected = "",
              options = list(
                allowEmptyOption = TRUE,
                persist = TRUE,
                placeholder = "Optional"
              )
            )
          ),
          conditionalPanel(
            condition = "input.conditioned2 == 1 && input.linguistic_subtabs == 'ner'",
            selectizeInput(
              "ner_doc_id_var",
              "Document ID variable",
              choices = NULL,
              selected = "",
              options = list(
                allowEmptyOption = TRUE,
                persist = TRUE,
                placeholder = "Optional"
              )
            )
          ),
          conditionalPanel(
            condition = "input.conditioned2 == 1 && input.linguistic_subtabs == 'ner'",
            tags$div(
              style = "margin-bottom: 15px;",
              tags$h5(
                HTML("<strong>spaCy NER Labels</strong>"),
                actionLink("showNERInfo", icon("info-circle"),
                          style = "color: #337ab7; font-size: 16px; margin-left: 8px;",
                          title = "Click for NER guide"),
                style = "color: #4269BF; margin-bottom: 10px;"
              ),
              div(
                style = "max-height: 300px; overflow-y: auto;",
                tags$details(
                  style = "margin-bottom: 6px; background: #f8f9fa; padding: 6px; border-radius: 4px;",
                  tags$summary(style = "cursor: pointer; font-weight: 600; font-size: 16px;", "People & Places"),
                  div(
                    style = "padding: 8px 0 0 8px;",
                    checkboxGroupInput("ner_named", HTML("<span class='sr-only'>Named entities</span>"), inline = FALSE,
                      choiceNames = list(
                        HTML("<span class='sidebar-color-picker' data-entity='PERSON' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#e91e63;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>PERSON</span> <span style='color:#475569;'>- People, including fictional</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='NORP' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#ff8f00;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>NORP</span> <span style='color:#475569;'>- Nationalities, religious/political groups</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='ORG' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#1565c0;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>ORG</span> <span style='color:#475569;'>- Companies, agencies, institutions</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='GPE' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#2e7d32;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>GPE</span> <span style='color:#475569;'>- Countries, cities, states</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='LOC' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#0277bd;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>LOC</span> <span style='color:#475569;'>- Non-GPE locations, mountains, water bodies</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='FAC' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#9e9d24;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>FAC</span> <span style='color:#475569;'>- Buildings, airports, highways, bridges</span>")
                      ),
                      choiceValues = c("PERSON", "NORP", "ORG", "GPE", "LOC", "FAC"),
                      selected = c("PERSON", "NORP", "ORG", "GPE", "LOC", "FAC")
                    )
                  )
                ),
                tags$details(
                  style = "margin-bottom: 6px; background: #f8f9fa; padding: 6px; border-radius: 4px;",
                  tags$summary(style = "cursor: pointer; font-weight: 600; font-size: 16px;", "Objects & Events"),
                  div(
                    style = "padding: 8px 0 0 8px;",
                    checkboxGroupInput("ner_objects", HTML("<span class='sr-only'>Object and event entities</span>"), inline = FALSE,
                      choiceNames = list(
                        HTML("<span class='sidebar-color-picker' data-entity='PRODUCT' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#5060D5;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>PRODUCT</span> <span style='color:#475569;'>- Objects, vehicles, foods, etc.</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='EVENT' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#c62828;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>EVENT</span> <span style='color:#475569;'>- Named hurricanes, battles, wars, sports</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='WORK_OF_ART' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#734FE2;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>WORK_OF_ART</span> <span style='color:#475569;'>- Titles of books, songs, etc.</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='LAW' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#03786A;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>LAW</span> <span style='color:#475569;'>- Named documents made into laws</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='LANGUAGE' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#558b2f;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>LANGUAGE</span> <span style='color:#475569;'>- Any named language</span>")
                      ),
                      choiceValues = c("PRODUCT", "EVENT", "WORK_OF_ART", "LAW", "LANGUAGE"),
                      selected = c("PRODUCT", "EVENT", "WORK_OF_ART", "LAW", "LANGUAGE")
                    )
                  )
                ),
                tags$details(
                  style = "margin-bottom: 6px; background: #f8f9fa; padding: 6px; border-radius: 4px;",
                  tags$summary(style = "cursor: pointer; font-weight: 600; font-size: 16px;", "Numeric & Temporal"),
                  div(
                    style = "padding: 8px 0 0 8px;",
                    checkboxGroupInput("ner_numeric", HTML("<span class='sr-only'>Numeric and temporal entities</span>"), inline = FALSE,
                      choiceNames = list(
                        HTML("<span class='sidebar-color-picker' data-entity='DATE' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#ef6c00;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>DATE</span> <span style='color:#475569;'>- Absolute or relative dates/periods</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='TIME' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#d84315;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>TIME</span> <span style='color:#475569;'>- Times smaller than a day</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='MONEY' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#9C3AD7;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>MONEY</span> <span style='color:#475569;'>- Monetary values including unit</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='PERCENT' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#00838f;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>PERCENT</span> <span style='color:#475569;'>- Percentage including %</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='QUANTITY' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#78909c;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>QUANTITY</span> <span style='color:#475569;'>- Measurements (weight, distance)</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='ORDINAL' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#866358;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>ORDINAL</span> <span style='color:#475569;'>- first, second, third, etc.</span>"),
                        HTML("<span class='sidebar-color-picker' data-entity='CARDINAL' data-source='spacy' style='display:inline-block;width:12px;height:12px;background:#546e7a;border-radius:2px;margin-right:6px;cursor:pointer;'></span><span style='font-weight:500;'>CARDINAL</span> <span style='color:#475569;'>- Numerals not in other category</span>")
                      ),
                      choiceValues = c("DATE", "TIME", "MONEY", "PERCENT", "QUANTITY", "ORDINAL", "CARDINAL"),
                      selected = c("DATE", "TIME", "MONEY", "PERCENT", "QUANTITY", "ORDINAL", "CARDINAL")
                    )
                  )
                )
              ),
              shinyjs::hidden(
                selectizeInput("spacy_default_entities", HTML("<span class='sr-only'>Default spaCy entities</span>"), choices = NULL)
              )
            ),
            tags$hr(style = "margin: 15px 0; border-color: #dee2e6;"),
            tags$div(
              style = "margin-bottom: 15px;",
              tags$h5(
                HTML("<strong>Domain Entities</strong>"),
                style = "color: #4269BF; margin-bottom: 10px;"
              ),
              selectizeInput(
                "ner_domain_preset",
                NULL,
                choices = c(
                  "None (Custom)" = "none",
                  "Special Education" = "special_education"
                ),
                selected = "special_education",
                multiple = FALSE,
                options = list(
                  placeholder = "Select a domain"
                )
              ),
              div(
                style = "display: flex; gap: 10px; margin-bottom: 8px;",
                div(
                  style = "flex: 1;",
                  downloadButton("export_domain_preset", "Export",
                    class = "btn-secondary btn-block", icon = icon("download"))
                ),
                div(
                  style = "flex: 1;",
                  actionButton("import_domain_trigger", "Import",
                    class = "btn-secondary btn-block", icon = icon("upload"))
                ),
                tags$input(type = "file", id = "import_domain_preset_file",
                  accept = ".xlsx,.csv", style = "display: none;"),
                tags$script(HTML("
                  $('#import_domain_trigger').on('click', function() {
                    $('#import_domain_preset_file').click();
                  });
                  $('#import_domain_preset_file').on('change', function() {
                    var file = this.files[0];
                    if (file) {
                      var reader = new FileReader();
                      reader.onload = function(e) {
                        Shiny.setInputValue('import_domain_preset_data', {
                          name: file.name,
                          dataurl: e.target.result
                        });
                      };
                      reader.readAsDataURL(file);
                      this.value = '';
                    }
                  });
                "))
              ),
              div(
                uiOutput("domain_entity_categories"),
                uiOutput("custom_entities_ui")
              )
            ),
            tags$hr(style = "margin: 15px 0; border-color: #dee2e6;"),
            tags$div(
              style = "margin-bottom: 15px;",
              tags$h5(
                HTML("<strong>Add Custom Entity</strong>"),
                style = "color: #4269BF; margin-bottom: 10px;"
              ),
              div(
                style = "display: flex; gap: 8px; align-items: flex-end;",
                div(
                  style = "flex: 0 0 36px;",
                  title = "Select the entity color",
                  if (requireNamespace("colourpicker", quietly = TRUE)) {
                    colourpicker::colourInput("custom_entity_color", HTML("<span class='sr-only'>Custom entity color</span>"),
                      value = "#607D8B", showColour = "background")
                  } else {
                    textInput("custom_entity_color", HTML("<span class='sr-only'>Custom entity color hex</span>"), value = "#607D8B", placeholder = "#HEX")
                  }
                ),
                div(
                  style = "flex: 1;",
                  textInput("custom_entity_name", HTML("<span class='sr-only'>Custom entity name</span>"), placeholder = "Type a New Entity")
                ),
                div(
                  style = "flex: 0 0 auto; margin-bottom: 15px;",
                  actionButton("apply_custom_entity", icon("plus"),
                    class = "btn-primary btn-sm",
                    style = "height: 34px; width: 34px; padding: 0;",
                    title = "Add custom entity")
                )
              ),
              selectizeInput("uncategorized_tokens", HTML("<span class='sr-only'>Uncategorized tokens to assign</span>"),
                choices = NULL, selected = NULL, multiple = TRUE,
                options = list(create = TRUE, placeholder = "Select terms for this entity")
              )
            ),
            tags$hr(style = "margin: 15px 0; border-color: #dee2e6;"),
            div(
              style = "display: flex; gap: 10px; margin-bottom: 15px;",
              div(
                style = "flex: 1;",
                actionButton("apply_ner_filter", "Apply",
                           class = "btn-primary btn-block",
                           icon = icon("play"))
              ),
              div(
                style = "flex: 1;",
                downloadButton("export_entities", "Export",
                           class = "btn-secondary btn-block",
                           icon = icon("download"))
              )
            ),
            shinyjs::hidden(
              selectizeInput("entity_type_filter", HTML("<span class='sr-only'>Entity type filter</span>"), choices = NULL),
              textAreaInput("batch_entity_text", HTML("<span class='sr-only'>Batch entity text input</span>")),
              checkboxInput("batch_include_lemmas", HTML("<span class='sr-only'>Include lemmas in batch</span>"), value = TRUE),
              actionButton("add_custom_entity_type", ""),
              actionButton("tab_type", ""),
              actionButton("tab_upload", ""),
              actionButton("add_batch_entities", ""),
              actionButton("add_batch_entities_bottom", ""),
              actionButton("process_codebook", ""),
              actionButton("process_codebook_bottom", ""),
              fileInput("codebook_upload", HTML("<span class='sr-only'>Upload codebook file</span>"))
            ),
            tags$script(HTML("
              $(document).on('click', '.sidebar-color-picker', function(e) {
                e.stopPropagation();
                e.preventDefault();
                var $el = $(this);
                var entity = $el.data('entity');
                var source = $el.data('source') || '';
                var category = $el.data('category') || '';
                var bgColor = $el.css('background-color');
                var hex = bgColor;
                if (bgColor.indexOf('rgb') === 0) {
                  var parts = bgColor.match(/\\d+/g);
                  if (parts && parts.length >= 3) {
                    hex = '#' +
                      ('0' + parseInt(parts[0]).toString(16)).slice(-2) +
                      ('0' + parseInt(parts[1]).toString(16)).slice(-2) +
                      ('0' + parseInt(parts[2]).toString(16)).slice(-2);
                  }
                }
                Shiny.setInputValue('sidebar_color_edit', {
                  entity: entity,
                  source: source,
                  category: category,
                  color: hex
                }, {priority: 'event'});
              });
            ")),
          tags$script(HTML("
            $(document).on('click', '.entity-visibility-chk', function(e) {
              e.stopPropagation();
              Shiny.setInputValue('entity_visibility_toggle', {
                category: $(this).data('category'),
                source: $(this).data('source'),
                checked: this.checked,
                time: new Date().getTime()
              }, {priority: 'event'});
            });
            $(document).on('click', '.entity-delete-btn', function(e) {
              e.stopPropagation();
              e.preventDefault();
              Shiny.setInputValue('delete_entity_category', {
                category: $(this).data('category'),
                source: $(this).data('source'),
                time: new Date().getTime()
              }, {priority: 'event'});
            });
            "))
          )
        ),
        mainPanel(
          width = 9,
          tabsetPanel(
            id = "conditioned2",
            tabPanel(
              "1. Linguistic Annotation",
              value = 1,
              br(),
              tabsetPanel(
                id = "linguistic_subtabs",
                tabPanel(
                  "Word Forms (Lemmas)",
                  value = "lemmas",
                  br(),
                  conditionalPanel(
                    condition = "output.lemma_ready == false",
                    div(
                      style = "padding: 60px 40px; text-align: center;",
                      div(
                        style = "max-width: 500px; margin: 0 auto;",
                        tags$i(class = "fa fa-info-circle", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                        tags$p(
                          "Click ",
                          tags$strong("'Apply'", style = "color: #4269BF;"),
                          " to run spaCy linguistic analysis, or ",
                          tags$strong("'Skip'", style = "color: #4269BF;"),
                          " to use standard tokenization.",
                          style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                        )
                      )
                    )
                  ),
                  conditionalPanel(
                    condition = "output.lemma_ready == true",
                    div(
                      style = "margin-bottom: 20px; overflow: visible;",
                      plotly::plotlyOutput("lemma_plot", height = 500, width = "100%")
                    ),
                    br(),
                    DT::dataTableOutput("lemma_table")
                  )
                ),
                tabPanel(
                  "Part-of-Speech Tags",
                  value = "pos",
                  br(),
                  conditionalPanel(
                    condition = "output.pos_ready == false",
                    div(
                      style = "padding: 60px 40px; text-align: center;",
                      div(
                        style = "max-width: 500px; margin: 0 auto;",
                        tags$i(class = "fa fa-info-circle", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                        tags$p(
                          "First, click ",
                          tags$strong("'Apply'", style = "color: #4269BF;"),
                          " on the Word Forms tab to run spaCy analysis. Then configure and click ",
                          tags$strong("'Apply'", style = "color: #4269BF;"),
                          " here to view POS tags.",
                          style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                        )
                      )
                    )
                  ),
                  conditionalPanel(
                    condition = "output.pos_ready == true",
                    uiOutput("pos_plot_uiOutput"),
                    br(),
                    DT::dataTableOutput("pos_table")
                  )
                ),
                tabPanel(
                  "Morphological Features",
                  value = "morphology",
                  br(),
                  conditionalPanel(
                    condition = "output.morph_ready == false",
                    div(
                      style = "padding: 60px 40px; text-align: center;",
                      div(
                        style = "max-width: 500px; margin: 0 auto;",
                        tags$i(class = "fa fa-info-circle", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                        tags$p(
                          "First, run POS tagging on the Word Forms tab. Then select morphological features in the sidebar and click ",
                          tags$strong("'Analyze Morphology'", style = "color: #4269BF;"),
                          " to extract features.",
                          style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                        )
                      )
                    )
                  ),
                  conditionalPanel(
                    condition = "output.morph_ready == true",
                    tabsetPanel(
                      id = "morph_subtabs",
                      tabPanel(
                        "Feature Distribution",
                        br(),
                        uiOutput("morph_warning_ui"),
                        fluidRow(
                          column(6, plotly::plotlyOutput("morph_number_plot", height = "300px")),
                          column(6, plotly::plotlyOutput("morph_tense_plot", height = "300px"))
                        ),
                        fluidRow(
                          column(6, plotly::plotlyOutput("morph_verbform_plot", height = "300px")),
                          column(6, plotly::plotlyOutput("morph_person_plot", height = "300px"))
                        )
                      ),
                      tabPanel(
                        "Additional Features",
                        br(),
                        conditionalPanel(
                          condition = "output.has_additional_features == true",
                          fluidRow(
                            conditionalPanel(
                              condition = "output.has_case_data == true",
                              column(4, plotly::plotlyOutput("morph_case_plot", height = "300px"))
                            ),
                            conditionalPanel(
                              condition = "output.has_mood_data == true",
                              column(4, plotly::plotlyOutput("morph_mood_plot", height = "300px"))
                            ),
                            conditionalPanel(
                              condition = "output.has_aspect_data == true",
                              column(4, plotly::plotlyOutput("morph_aspect_plot", height = "300px"))
                            )
                          )
                        ),
                        conditionalPanel(
                          condition = "output.has_additional_features == false",
                          div(style = "padding: 60px 40px; text-align: center;",
                            div(style = "max-width: 400px; margin: 0 auto;",
                              tags$i(class = "fa fa-language", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", `aria-hidden` = "true"),
                              tags$p("Select Case, Mood, or Aspect checkboxes from the sidebar to display additional morphological features.",
                                style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;")
                            )
                          )
                        )
                      ),
                      tabPanel(
                        "Summary Table",
                        br(),
                        DT::dataTableOutput("morph_summary_table")
                      )
                    )
                  )
                ),
                tabPanel(
                  "Dependency Parsing",
                  value = "dependencies",
                  br(),
                  conditionalPanel(
                    condition = "output.dependencies_ready == false",
                    div(
                      style = "padding: 60px 40px; text-align: center;",
                      div(
                        style = "max-width: 500px; margin: 0 auto;",
                        tags$i(class = "fa fa-info-circle", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                        tags$p(
                          "Click ",
                          tags$strong("'Apply'", style = "color: #4269BF;"),
                          " on the Word Forms tab to run spaCy analysis and extract dependency parsing data.",
                          style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                        )
                      )
                    )
                  ),
                  conditionalPanel(
                    condition = "output.dependencies_ready == true",
                    # Document selector for displaCy visualization
                    div(
                      style = "margin-bottom: 15px;",
                      selectInput(
                        "dep_viz_doc",
                        "Select document to visualize dependencies:",
                        choices = NULL,
                        width = "100%"
                      )
                    ),
                    # Download button for Dependency Parsing
                    uiOutput("dep_download_button"),
                    uiOutput("dep_displacy_container"),
                    br(),
                    DT::dataTableOutput("dep_table")
                  )
                ),
                tabPanel(
                  "Named Entity Recognition",
                  value = "ner",
                  br(),
                  conditionalPanel(
                    condition = "output.spacy_ready == false",
                    div(
                      style = "padding: 60px 40px; text-align: center;",
                      div(
                        style = "max-width: 500px; margin: 0 auto;",
                        tags$i(class = "fa fa-info-circle", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                        tags$p(
                          "Click ",
                          tags$strong("'Apply'", style = "color: #4269BF;"),
                          " on the Word Forms tab to run spaCy analysis.",
                          style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                        )
                      )
                    )
                  ),
                  conditionalPanel(
                    condition = "output.ner_pending == true",
                    div(
                      style = "padding: 60px 40px; text-align: center;",
                      div(
                        style = "max-width: 500px; margin: 0 auto;",
                        tags$i(class = "fa fa-info-circle", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                        tags$p(
                          "Click ",
                          tags$strong("'Apply'", style = "color: #4269BF;"),
                          " to extract named entities.",
                          style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                        )
                      )
                    )
                  ),
                  conditionalPanel(
                    condition = "output.ner_ready == true",
                    # Document selector, displaCy visualization, and download button (conditional)
                    uiOutput("ner_doc_selector"),
                    uiOutput("ner_displacy_container"),
                    uiOutput("ner_download_button"),
                    # Existing frequency plot
                    uiOutput("entity_plot_uiOutput"),
                    br(),
                    uiOutput("entity_table_uiOutput")
                  )
                )
              )
            ),
            tabPanel(
              "2. Frequency Trends",
              value = 2,
              bsCollapse(
                open = 0,
                bsCollapsePanel(
                  p(strong("Click to set plot height"),
                    class = "plot-dimensions-text"
                  ),
                  value = 2,
                  style = "success",
                  p(strong("Height of the plot")),
                  sliderInput(
                    inputId = "height_frequency_trends",
                    post = " px",
                    label = "height",
                    min = 200,
                    max = 2000,
                    step = 50,
                    value = 500
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_frequency_plot == true",
                uiOutput("line_con_var_plot_uiOutput")
              ),
              conditionalPanel(
                condition = "output.has_frequency_plot == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-chart-line", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Select terms, continuous variable, and click ",
                      tags$strong("'Plot Terms'", style = "color: #4269BF;"),
                      " to analyze frequency trends",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "3. Keywords",
              value = 3,
              br(),
              tabsetPanel(
                id = "keywords_subtabs",
                tabPanel(
                  "TF-IDF",
                  value = "tfidf",
                  uiOutput("keywords_tfidf_uiOutput")
                ),
                tabPanel(
                  "Statistical Keyness",
                  value = "textrank",
                  uiOutput("keywords_textrank_uiOutput")
                ),
                tabPanel(
                  "Comparison",
                  value = "comparison",
                  uiOutput("keywords_comparison_uiOutput")
                )
              )
            ),
            tabPanel(
              "4. Lexical Diversity",
              value = 4,
              br(),
              uiOutput("lexical_diversity_uiOutput")
            ),
            tabPanel(
              "5. Readability",
              value = 5,
              br(),
              uiOutput("readability_results_uiOutput")
            ),
            tabPanel(
              "6. Log Odds Ratio",
              value = 6,
              br(),
              conditionalPanel(
                condition = "output.log_odds_ready == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 500px; margin: 0 auto;",
                    tags$i(class = "fa fa-balance-scale", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Select a grouping variable and click ",
                      tags$strong("'Calculate'", style = "color: #4269BF;"),
                      " to compare word usage between categories.",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    ),

                  )
                )
              ),
              conditionalPanel(
                condition = "output.log_odds_ready == true",
                uiOutput("log_odds_plot_uiOutput"),
                br(),
                DT::dataTableOutput("log_odds_table")
              )
            ),
            tabPanel(
              "7. Lexical Dispersion",
              value = 7,
              br(),
              conditionalPanel(
                condition = "output.dispersion_ready == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 500px; margin: 0 auto;",
                    tags$i(class = "fa fa-chart-line", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Select terms in the sidebar and click ",
                      tags$strong("'Analyze'", style = "color: #4269BF;"),
                      " to visualize their dispersion across documents.",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    ),

                  )
                )
              ),
              conditionalPanel(
                condition = "output.dispersion_ready == true",
                uiOutput("dispersion_plot_uiOutput"),
                conditionalPanel(
                  condition = "input.dispersion_show_metrics == true",
                  br(),
                  div(
                    style = "margin-top: 20px;",
                    tags$h5("Dispersion Metrics", style = "color: #4269BF; margin-bottom: 10px;"),
                    DT::dataTableOutput("dispersion_metrics_table")
                  )
                )
              )
            )
          )
        )
      )
}

semantic_analysis_ui_content <- function() {
      sidebarLayout(
        sidebarPanel(
          width = 3,
          class = "sidebar-panel",
          # Categorical variable - only for tabs that use it (Similarity and Comparative)
          conditionalPanel(
            condition = "input.semantic_analysis_tabs == 'similarity' || input.semantic_analysis_tabs == 'comparative'",
            div(
              style = "margin-bottom: 15px; padding-bottom: 15px; border-bottom: 1px solid #e2e8f0;",
              selectizeInput(
                "doc_category_var",
                "Categorical variable",
                choices = NULL,
                selected = "",
                options = list(
                  allowEmptyOption = TRUE,
                  persist = TRUE,
                  placeholder = "Optional"
                )
              )
            )
          ),
          conditionalPanel(
            condition = "input.semantic_analysis_tabs == 'summary'",
            tags$h5(strong("Configure document metadata"), style = "color: #4269BF; margin-bottom: 10px;"),
            div(class = "checkbox-margin",
              selectizeInput(
                "doc_id_var",
                "Document ID variable",
                choices = NULL,
                selected = "",
                options = list(
                  allowEmptyOption = TRUE,
                  persist = TRUE,
                  placeholder = "Optional"
                )
              )
            ),
            br(),
            div(
              style = "margin-bottom: 15px;",
              actionButton("process_documents", "Process", class = "btn-primary btn-block")
            )
          ),
          conditionalPanel(
            condition = "input.semantic_analysis_tabs != 'summary' && input.semantic_analysis_tabs != 'sentiment'",
            div(
              style = "margin-bottom: 8px;",
              uiOutput("semantic_feature_space_selector")
            )
          ),

          conditionalPanel(
            condition = "input.semantic_analysis_tabs == 'similarity'",
            tags$h5(HTML("<strong>Document Similarity</strong> <a href='https://www.sbert.net/' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"), style = "color: #4269BF; margin-bottom: 10px;"),

            conditionalPanel(
              condition = "input.semantic_feature_space == 'embeddings'",
              tags$label("Embedding Configuration", style = "font-weight: 600; margin-bottom: 8px; display: block; color: #4269BF;"),
              tags$p(style = "font-size: 16px; color: #666; margin-bottom: 10px;", "Document-to-document similarity using vector embeddings."),
              uiOutput("embedding_status_ui"),
              radioButtons(
                "embedding_provider",
                "AI Provider:",
                choices = c(
                  "Sentence Transformers (Python)" = "sentence-transformers",
                  "OpenAI (API Key Required)" = "openai",
                  "Gemini (API Key Required)" = "gemini"
                ),
                selected = "sentence-transformers",
                inline = FALSE
              ),
              conditionalPanel(
                condition = "input.embedding_provider == 'sentence-transformers'",
                selectizeInput(
                  "embedding_st_model",
                  "Model:",
                  choices = c(
                    "all-MiniLM-L6-v2 (Fast)" = "all-MiniLM-L6-v2",
                    "all-mpnet-base-v2 (Higher Quality)" = "all-mpnet-base-v2",
                    "BGE Small EN v1.5 (Fast + Strong Retrieval)" = "BAAI/bge-small-en-v1.5",
                    "BGE Base EN v1.5 (BERTopic Optimized)" = "BAAI/bge-base-en-v1.5",
                    "E5 Base v2 (Instructor-tuned)" = "intfloat/e5-base-v2",
                    "Nomic Embed Text v2 (Multilingual)" = "nomic-ai/nomic-embed-text-v2-moe",
                    "GTE Multilingual Base (Fast, Multilingual)" = "Alibaba-NLP/gte-multilingual-base"
                  ),
                  selected = NULL,
                  options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
                ),
                tags$p(
                  style = "font-size: 16px; color: #666;",
                  "Requires Python + sentence-transformers."
                )
              ),
              conditionalPanel(
                condition = "input.embedding_provider == 'openai'",
                selectizeInput(
                  "embedding_openai_model",
                  "OpenAI Model:",
                  choices = c("Text Embedding 3 Small (Default)" = "text-embedding-3-small", "Text Embedding 3 Large (Higher Quality)" = "text-embedding-3-large"),
                  selected = NULL,
                  options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
                ),
                .password_input("embedding_openai_api_key", "API Key:", placeholder = "sk-..."),
                conditionalPanel(
                  condition = "output.has_openai_key",
                  tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                    icon("check-circle"), " Key stored. Enter new key to override.")
                )
              ),
              conditionalPanel(
                condition = "input.embedding_provider == 'gemini'",
                selectizeInput(
                  "embedding_gemini_model",
                  "Gemini Model:",
                  choices = c("Gemini Embedding 001" = "gemini-embedding-001", "Gemini Embedding 2 Preview" = "gemini-embedding-2-preview"),
                  selected = NULL,
                  options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
                ),
                .password_input("embedding_gemini_api_key", "API Key:", placeholder = "AIza..."),
                conditionalPanel(
                  condition = "output.has_gemini_key",
                  tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                    icon("check-circle"), " Key stored. Enter new key to override.")
                )
              ),
              div(
                style = "margin-bottom: 15px;",
                actionButton("generate_embeddings", "Generate Embeddings", class = "btn-primary btn-block", icon = icon("brain"))
              ),
              tags$hr(style = "margin: 10px 0; border-color: #dee2e6;")
            ),
            conditionalPanel(
              condition = "input.semantic_feature_space == 'words'"
            ),
            conditionalPanel(
              condition = "input.semantic_feature_space == 'ngrams'"
            ),
            div(
              style = "margin-bottom: 15px;",
              actionButton("calculate_similarity", "Calculate", class = "btn-primary btn-block")
            ),
            div(
              style = "margin-bottom: 8px;",
              selectizeInput(
                "heatmap_category_filter",
                "Filter by category",
                choices = c("All Categories" = "all"),
                selected = "all"
              )
            ),
            tags$div(
              id = "words_similarity_ready_status",
              class = "status-sidebar-success",
              style = "display: none;",
              tags$i(class = "fa fa-check-circle status-icon status-icon-success"),
              tags$span("Words similarity data calculated and ready")
            ),
            tags$div(
              id = "ngrams_similarity_ready_status",
              class = "status-sidebar-success",
              style = "display: none;",
              tags$i(class = "fa fa-check-circle status-icon status-icon-success"),
              tags$span("N-grams similarity data calculated and ready")
            ),
            tags$div(
              id = "topics_similarity_ready_status",
              class = "status-sidebar-success",
              style = "display: none;",
              tags$i(class = "fa fa-check-circle status-icon status-icon-success"),
              tags$span("Topics similarity data calculated and ready")
            ),
            tags$div(
              id = "embeddings_similarity_ready_status",
              class = "status-sidebar-success",
              style = "display: none;",
              tags$i(class = "fa fa-check-circle status-icon status-icon-success"),
              tags$span("Embeddings similarity data calculated and ready")
            ),
            actionButton("visualize_similarity", "Visualize Results", class = "btn-primary btn-block", icon = icon("chart-bar"))
          ),

          # Comparative Analysis Controls (new subtab)
          conditionalPanel(
            condition = "input.semantic_analysis_tabs == 'comparative'",
            tags$label(HTML("<strong>Comparative Analysis</strong>"), style = "font-weight: 700; margin-bottom: 8px; display: block;"),
            tags$div(
              class = "status-main-info",
              "Compare reference category against others to identify unique content, gaps, and cross-category opportunities."
            ),
            uiOutput("gap_reference_category_ui"),
            sliderInput(
              "gap_unique_threshold",
              "Unique threshold",
              min = 0.3, max = 0.9, value = 0.6, step = 0.05
            ),
            sliderInput(
              "gap_cross_policy_max",
              "Cross-category max",
              min = 0.6, max = 1.0, value = 0.8, step = 0.05
            ),
            actionButton("run_gap_analysis", "Run Comparative Analysis", class = "btn-info btn-block", icon = icon("search-minus")),
            tags$div(
              id = "gap_analysis_ready_status",
              class = "status-sidebar-success",
              style = "display: none; margin-top: 10px;",
              tags$i(class = "fa fa-check-circle status-icon status-icon-success"),
              tags$span("Comparative analysis complete")
            )
          ),

          conditionalPanel(
            condition = "input.semantic_analysis_tabs == 'search'",
            tags$p(style = "font-size: 16px; color: #666; margin-bottom: 10px;", "Query-to-document retrieval across your corpus."),
            selectInput(
              "search_method",
              "Search method:",
              choices = c(
                "Keyword" = "keyword",
                "Words" = "words",
                "N-grams" = "ngrams",
                "Embeddings" = "embeddings",
                "RAG Q&A (LLM)" = "rag"
              ),
              selected = "keyword"
            ),
            conditionalPanel(
              condition = "input.search_method != 'keyword' && input.search_method != 'rag' && input.search_method != 'embeddings'",
              uiOutput("search_method_status_message")
            ),
            conditionalPanel(
              condition = "input.search_method == 'embeddings'",
              radioButtons(
                "search_embedding_provider",
                "AI Provider:",
                choices = c(
                  "Sentence Transformers (Python)" = "sentence-transformers",
                  "OpenAI (API Key Required)" = "openai",
                  "Gemini (API Key Required)" = "gemini"
                ),
                selected = "sentence-transformers"
              ),
              conditionalPanel(
                condition = "input.search_embedding_provider == 'sentence-transformers'",
                selectizeInput(
                  "search_embedding_st_model",
                  "Model:",
                  choices = c(
                    "all-MiniLM-L6-v2 (Fast)" = "all-MiniLM-L6-v2",
                    "all-mpnet-base-v2 (Higher Quality)" = "all-mpnet-base-v2",
                    "BGE Small EN v1.5 (Fast + Strong Retrieval)" = "BAAI/bge-small-en-v1.5",
                    "BGE Base EN v1.5 (BERTopic Optimized)" = "BAAI/bge-base-en-v1.5",
                    "E5 Base v2 (Instructor-tuned)" = "intfloat/e5-base-v2",
                    "Nomic Embed Text v2 (Multilingual)" = "nomic-ai/nomic-embed-text-v2-moe",
                    "GTE Multilingual Base (Fast, Multilingual)" = "Alibaba-NLP/gte-multilingual-base"
                  ),
                  selected = NULL,
                  options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
                ),
                tags$p(
                  style = "font-size: 16px; color: #666;",
                  "Requires Python + sentence-transformers."
                )
              ),
              conditionalPanel(
                condition = "input.search_embedding_provider == 'openai'",
                selectizeInput(
                  "search_embedding_openai_model",
                  "OpenAI Model:",
                  choices = c(
                    "Text Embedding 3 Small (Default)" = "text-embedding-3-small",
                    "Text Embedding 3 Large (Higher Quality)" = "text-embedding-3-large"
                  ),
                  selected = NULL,
                  options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
                ),
                .password_input("search_embedding_openai_api_key", "API Key:", placeholder = "sk-..."),
                conditionalPanel(
                  condition = "output.has_openai_key",
                  tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                    icon("check-circle"), " Key stored. Enter new key to override.")
                )
              ),
              conditionalPanel(
                condition = "input.search_embedding_provider == 'gemini'",
                selectizeInput(
                  "search_embedding_gemini_model",
                  "Gemini Model:",
                  choices = c("Gemini Embedding 001" = "gemini-embedding-001", "Gemini Embedding 2 Preview" = "gemini-embedding-2-preview"),
                  selected = NULL,
                  options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
                ),
                .password_input("search_embedding_gemini_api_key", "API Key:", placeholder = "AIza..."),
                conditionalPanel(
                  condition = "output.has_gemini_key",
                  tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                    icon("check-circle"), " Key stored. Enter new key to override.")
                )
              )
            ),
            conditionalPanel(
              condition = "input.search_method == 'rag'",
              radioButtons(
                "rag_provider",
                "AI Provider:",
                choices = c(
                  "OpenAI (API Key Required)" = "openai",
                  "Gemini (API Key Required)" = "gemini"
                ),
                selected = "openai"
              ),
              conditionalPanel(
                condition = "input.rag_provider == 'openai'",
                selectizeInput(
                  "rag_openai_model",
                  "OpenAI Model:",
                  choices = c("GPT-4.1 Mini (Default, fast)" = "gpt-4.1-mini", "GPT-4.1 (Accurate)" = "gpt-4.1", "GPT-4" = "gpt-4"),
                  selected = NULL,
                  options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
                ),
                .password_input("rag_openai_api_key", "API Key:", placeholder = "sk-..."),
                conditionalPanel(
                  condition = "output.has_openai_key",
                  tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                    icon("check-circle"), " Key stored. Enter new key to override.")
                )
              ),
              conditionalPanel(
                condition = "input.rag_provider == 'gemini'",
                selectizeInput(
                  "rag_gemini_model",
                  "Gemini Model:",
                  choices = c("Gemini 2.5 Flash Lite (Default, economy)" = "gemini-2.5-flash-lite", "Gemini 2.5 Flash" = "gemini-2.5-flash", "Gemini 2.5 Pro (Accurate)" = "gemini-2.5-pro"),
                  selected = NULL,
                  options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
                ),
                .password_input("rag_gemini_api_key", "API Key:", placeholder = "AIza..."),
                conditionalPanel(
                  condition = "output.has_gemini_key",
                  tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                    icon("check-circle"), " Key stored. Enter new key to override.")
                )
              )
            ),
            div(
              uiOutput("search_query_input"),
              sliderInput(
                "semantic_search_top_k",
                "Results to return",
                value = 5,
                min = 1,
                max = 20,
                step = 1
              ),
              br(),
              div(
                style = "margin-bottom: 15px;",
                actionButton("run_semantic_search", "Search", class = "btn-primary btn-block", icon = icon("search"))
              )
            )
          ),
          conditionalPanel(
            condition = "input.semantic_analysis_tabs == 'clustering'",
            tags$div(
              style = "display: flex; align-items: center; justify-content: space-between; margin-bottom: 15px;",
              tags$h5(HTML("<strong>Find Document Groups</strong> <a href='https://scikit-learn.org/stable/modules/clustering.html' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"), style = "color: #4269BF; margin: 0;"),
              actionLink("showClusteringInfo", icon("info-circle"),
                        style = "color: #337ab7; font-size: 16px;",
                        title = "Learn about document grouping")
            ),

            uiOutput("clustering_warning"),

            wellPanel(
              style = "padding: 15px; margin-bottom: 15px;",
              tags$h5(strong("Configuration"), style = "color: #4269BF; margin-bottom: 10px;"),
              radioButtons(
                "clustering_approach",
                "Discovery mode",
                choices = list(
                  "Automatic" = "auto",
                  "Manual" = "manual",
                  "Method" = "method"
                ),
                selected = "auto"
              ),

              conditionalPanel(
                condition = "input.clustering_approach == 'manual'",
                numericInput(
                  "manual_n_clusters",
                  "Number of groups",
                  value = 4,
                  min = 2,
                  max = 15
                )
              ),

              conditionalPanel(
                condition = "input.clustering_approach == 'method'",
                selectInput(
                  "semantic_cluster_method",
                  "Algorithm",
                  choices = c(
                    "K-means" = "kmeans",
                    "Hierarchical" = "hierarchical",
                    "DBSCAN (Density-based)" = "dbscan",
                    "Neural Topic Model" = "neural",
                    "Semantic Topic Modeling (All-in-one)" = "semantic_unified"
                  ),
                  selected = "kmeans"
                ),
                tags$p(
                  tags$i(class = "fa fa-info-circle", style = "margin-right: 5px;"),
                  tags$strong("Note:"), " Neural and Semantic models use transformer embeddings.",
                  style = "font-size: 16px; color: #6b7280; margin-top: 5px;"
                ),
                conditionalPanel(
                  condition = "input.semantic_cluster_method == 'kmeans'",
                  numericInput(
                    "kmeans_n_clusters",
                    "Number of clusters",
                    value = 0,
                    min = 0,
                    max = 20,
                    step = 1
                  )
                ),
                conditionalPanel(
                  condition = "input.semantic_cluster_method == 'neural'",
                  numericInput(
                    "neural_n_topics",
                    "Number of topics",
                    value = 10,
                    min = 2,
                    max = 50,
                    step = 1
                  ),
                  numericInput(
                    "neural_hidden_size",
                    "Hidden layer size",
                    value = 100,
                    min = 50,
                    max = 500,
                    step = 50
                  )
                ),
                conditionalPanel(
                  condition = "input.semantic_cluster_method == 'semantic_unified'",
                  tags$h5("Unified Semantic Topic Modeling", style = "color: #10B981; font-weight: bold; margin-bottom: 10px;"),
                  tags$p("Combines multiple approaches for stable topic modeling",
                        style = "font-size: 16px; color: #6b7280; margin-bottom: 10px;"),
                  selectInput(
                    "unified_method",
                    "Base method",
                    choices = c(
                      "Embedding + Clustering" = "embedding_clustering",
                      "Similarity-based Grouping" = "similarity_grouping"
                    ),
                    selected = "embedding_clustering"
                  ),
                  numericInput(
                    "unified_n_topics",
                    "Number of topics",
                    value = 10,
                    min = 2,
                    max = 30,
                    step = 1
                  ),
                  sliderInput(
                    "unified_min_topic_size",
                    "Minimum topic size",
                    value = 3,
                    min = 2,
                    max = 10,
                    step = 1
                  )
                )
              )
            ),

            tags$details(
              style = "margin-top: 15px; margin-bottom: 15px;",
              tags$summary(
                style = "cursor: pointer; color: #596D80; font-weight: 600;",
                "More Settings"
              ),
              div(
                style = "margin-top: 10px; padding: 10px; border-radius: 4px;",
                sliderInput(
                  "semantic_cluster_seed",
                  "Random seed",
                  value = 123,
                  min = 1,
                  max = 1000,
                  step = 1
                ),
                selectInput(
                  "semantic_cluster_table_view",
                  "Table view",
                  choices = c(
                    "Cluster Summary" = "summary",
                    "Document Details" = "details"
                  ),
                  selected = "summary"
                )
              )
            ),

            div(
              style = "margin-bottom: 15px;",
              actionButton(
                "run_semantic_analysis_clustering",
                "Discover",
                class = "btn-primary btn-block",
                icon = icon("magnifying-glass-chart")
              )
            )
          ),
            conditionalPanel(
              condition = "output.has_clustering_results && input.semantic_analysis_tabs == 'clustering'",
            wellPanel(
              style = "padding: 12px; margin-top: 15px; background-color: #f8f9fa;",
            tags$h5(strong("AI-Powered Cluster Labels"), style = "color: #4269BF; margin-bottom: 15px;"),
            uiOutput("cluster_labeling_status"),

            radioButtons(
              "cluster_label_provider",
              "AI Provider:",
              choices = c(
                "OpenAI (API Key Required)" = "openai",
                "Gemini (API Key Required)" = "gemini"
              ),
              selected = "openai",
              inline = FALSE
            ),

            conditionalPanel(
              condition = "input.cluster_label_provider == 'openai'",
              selectizeInput(
                "cluster_openai_model",
                "OpenAI Model:",
                choices = c("GPT-4.1 Mini (Default, fast)" = "gpt-4.1-mini", "GPT-4.1 (Accurate)" = "gpt-4.1", "GPT-4" = "gpt-4"),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("cluster_openai_api_key", "API Key:", placeholder = "sk-..."),
              conditionalPanel(
                condition = "output.has_openai_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),

            conditionalPanel(
              condition = "input.cluster_label_provider == 'gemini'",
              selectizeInput(
                "cluster_gemini_model",
                "Gemini Model:",
                choices = c("Gemini 2.5 Flash Lite (Default, economy)" = "gemini-2.5-flash-lite", "Gemini 2.5 Flash" = "gemini-2.5-flash", "Gemini 2.5 Pro (Accurate)" = "gemini-2.5-pro"),
                selected = NULL,
                options = list(create = TRUE, placeholder = "Type your model...", onInitialize = I("function() { this.setValue(\"\"); }"))
              ),
              .password_input("cluster_gemini_api_key", "API Key:", placeholder = "AIza..."),
              conditionalPanel(
                condition = "output.has_gemini_key",
                tags$div(style = "color: #0C795A; font-size: 13px; margin-top: -8px; margin-bottom: 8px;",
                  icon("check-circle"), " Key stored. Enter new key to override.")
              )
            ),

            numericInput(
              "top_terms_for_labels",
              "Top terms per cluster:",
              value = 10,
              min = 5,
              max = 20,
              step = 1
            ),

            actionButton(
              "generate_cluster_labels",
              "Generate Labels with AI",
              class = "btn-primary btn-block",
              icon = icon("wand-magic-sparkles")
            )
            )
          ),
          conditionalPanel(
            condition = "input.semantic_analysis_tabs == 'sentiment'",
            tags$h5(HTML("<strong>Sentiment & Emotion Analysis</strong> <a href='https://github.com/juliasilge/tidytext' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"), style = "color: #4269BF; margin-bottom: 10px;"),

            uiOutput("sentiment_status_message"),
            conditionalPanel(
              condition = "input.sentiment_subtabs == 'overall'",
              radioButtons(
                "sentiment_method",
                "Analysis method",
                choices = c(
                  "Lexicon-based (Fast)" = "lexicon",
                  "Pre-trained Classifier (Accurate)" = "neural",
                  "LLM (with Explanations)" = "llm"
                ),
                selected = "lexicon",
                inline = FALSE
              ),
              conditionalPanel(
                condition = "input.sentiment_method == 'lexicon'",
                selectInput(
                  "sentiment_lexicon",
                  "Sentiment lexicon",
                  choices = c(
                    "AFINN" = "afinn",
                    "Bing" = "bing",
                    "NRC" = "nrc"
                  ),
                  selected = "bing"
                ),
                actionButton("run_sentiment_analysis", "Analyze Sentiment", class = "btn-primary btn-block")
              ),
              conditionalPanel(
                condition = "input.sentiment_method == 'neural'",
                selectInput(
                  "neural_sentiment_model",
                  "Sentiment model",
                  choices = c(
                    "DistilBERT SST-2 (Fast)" = "distilbert-base-uncased-finetuned-sst-2-english",
                    "RoBERTa Sentiment (Accurate)" = "cardiffnlp/twitter-roberta-base-sentiment-latest",
                    "BERT Multilingual" = "nlptown/bert-base-multilingual-uncased-sentiment"
                  ),
                  selected = "distilbert-base-uncased-finetuned-sst-2-english"
                ),
                actionButton("run_neural_sentiment", "Analyze Sentiment", class = "btn-primary btn-block")
              ),
              conditionalPanel(
                condition = "input.sentiment_method == 'llm'",
                radioButtons(
                  "llm_sentiment_provider",
                  "AI Provider:",
                  choices = c(
                    "OpenAI (API Key Required)" = "openai",
                    "Gemini (API Key Required)" = "gemini"
                  ),
                  selected = "openai",
                  inline = FALSE
                ),
                uiOutput("llm_sentiment_model_ui"),
                checkboxInput(
                  "llm_sentiment_explanation",
                  "Include sentiment explanations",
                  value = TRUE
                ),
                numericInput(
                  "llm_sentiment_batch_size",
                  "Batch size",
                  value = 5,
                  min = 1,
                  max = 20,
                  step = 1
                ),
                actionButton("run_llm_sentiment", "Analyze with LLM", class = "btn-primary btn-block", icon = icon("robot"))
              )
            ),
            conditionalPanel(
              condition = "input.sentiment_subtabs == 'category'",
              selectizeInput(
                "sentiment_category_var",
                "Categorical variable",
                choices = NULL,
                multiple = FALSE
              ),
              selectInput(
                "category_plot_type",
                "Plot type",
                choices = c(
                  "Bar Chart" = "bar",
                  "Box Plot" = "box",
                  "Violin Plot" = "violin"
                ),
                selected = "bar"
              )
            ),
            conditionalPanel(
              condition = "input.sentiment_subtabs == 'document'",
              selectizeInput(
                "sentiment_doc_id_var",
                "Document ID variable",
                choices = NULL,
                selected = "",
                options = list(
                  allowEmptyOption = TRUE,
                  persist = TRUE,
                  placeholder = "Optional"
                )
              ),
              sliderInput(
                "sentiment_top_docs",
                "Documents to show",
                value = 10,
                min = 1,
                max = 50,
                step = 1
              )
            ),
            conditionalPanel(
              condition = "input.sentiment_subtabs == 'emotion'",
              uiOutput("emotion_ready_indicator"),
              uiOutput("emotion_warning_message"),
              checkboxInput(
                "normalize_emotions",
                "Normalize emotion scores",
                value = FALSE
              ),
              selectizeInput(
                "emotion_group_var",
                "Group by variable",
                choices = NULL,
                multiple = FALSE,
                options = list(placeholder = "Optional")
              )
            )
          ),
          conditionalPanel(
            condition = "input.semantic_analysis_tabs == 'cooccurrence'",
            tags$h5(HTML("<strong>Word Co-occurrence Networks</strong> <a href='https://igraph.org/r/' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"), style = "color: #4269BF; margin-bottom: 10px;"),
            selectizeInput(
              "doc_var_co_occurrence",
              "Categorical variable",
              choices = NULL,
              multiple = FALSE,
              options = list(placeholder = "Optional")
            ),
            div(
              id = "global_cooccur_params", class = "global-params",
              sliderInput(
                "co_occurence_number_global",
                "Minimum co-occurrences",
                value = 10,
                min = 2,
                max = 150,
                step = 1
              ),
              sliderInput(
                "top_node_n_co_occurrence_global",
                "Top N nodes",
                value = 20,
                min = 2,
                max = 150,
                step = 1
              ),
              sliderInput(
                "node_label_size_cooccur",
                "Node label size (0 hides labels)",
                value = 22,
                min = 0,
                max = 40,
                step = 2
              ),
              selectInput(
                "community_method_cooccur",
                "Community detection",
                choices = c("Leiden (recommended)" = "leiden",
                            "Louvain" = "louvain"),
                selected = "leiden"
              ),
              selectInput(
                "node_size_cooccur",
                "Node size by",
                choices = c(
                  "Degree (connections)" = "degree",
                  "Betweenness (bridging)" = "betweenness",
                  "Frequency (word count)" = "frequency"
                ),
                selected = "degree"
              ),
              selectInput(
                "node_color_cooccur",
                "Node color by",
                choices = c(
                  "Community" = "community",
                  "Frequency (word count)" = "frequency"
                ),
                selected = "community"
              ),
              sliderInput(
                "nrows_co_occurrence",
                "Row numbers",
                value = 1,
                min = 1,
                max = 10,
                step = 1
              ),
              numericInput(
                "seed_cooccur",
                "Layout seed",
                value = 2026, min = 1, max = 999999, step = 1
              ),
              div(style = "font-size: 16px;",
                  checkboxInput("showlegend_cooccur", "Show legend", value = TRUE)),
              div(
                class = "category-toggle",
                div(style = "font-size: 16px;", checkboxInput("use_category_cooccur",
                  "Use category-specific analysis",
                  value = FALSE
                ))
              )
            ),
            conditionalPanel(
              condition = "input.use_category_cooccur == true && input.doc_var_co_occurrence != 'None' && input.doc_var_co_occurrence != ''",
              tags$div(
                class = "status-main-warning",
                "Analyze top categories by document count for faster processing and clearer visualization."
              ),
              uiOutput("top_n_selector_cooccur"),
              uiOutput("category_selector_cooccur"),
              uiOutput("category_cooccur_controls")
            ),
            actionButton("plot_word_co_occurrence_network", "Plot Network", class = "btn-primary btn-block"),
            conditionalPanel(
              condition = "output.has_cooccurrence_plot == true",
              tags$div(
                style = "margin-top: 8px; display: flex; gap: 6px;",
                downloadButton("download_cooccur_html", "HTML",
                               class = "btn-default btn-block",
                               style = "margin: 0; flex: 1;")
              )
            )
          ),
          conditionalPanel(
            condition = "input.semantic_analysis_tabs == 'correlation'",
            tags$h5(HTML("<strong>Word Correlation Networks</strong> <a href='https://igraph.org/r/' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"), style = "color: #4269BF; margin-bottom: 10px;"),
            selectizeInput(
              "doc_var_correlation",
              "Categorical variable",
              choices = NULL,
              multiple = FALSE,
              options = list(placeholder = "Optional")
            ),
            div(
              id = "global_corr_params", class = "global-params",
              sliderInput(
                "common_term_n_global",
                "Common terms",
                value = 20,
                min = 5,
                max = 150,
                step = 1
              ),
              sliderInput(
                "corr_n_global",
                "Minimum correlation",
                value = 0.4,
                min = 0,
                max = 1,
                step = 0.1
              ),
              sliderInput(
                "top_node_n_correlation_global",
                "Top N nodes",
                value = 20,
                min = 2,
                max = 150,
                step = 1
              ),
              sliderInput(
                "node_label_size_corr",
                "Node label size (0 hides labels)",
                value = 22,
                min = 0,
                max = 40,
                step = 2
              ),
              selectInput(
                "community_method_corr",
                "Community detection",
                choices = c("Leiden (recommended)" = "leiden",
                            "Louvain" = "louvain"),
                selected = "leiden"
              ),
              selectInput(
                "node_size_corr",
                "Node size by",
                choices = c(
                  "Degree (connections)" = "degree",
                  "Betweenness (bridging)" = "betweenness",
                  "Frequency (word count)" = "frequency"
                ),
                selected = "degree"
              ),
              selectInput(
                "node_color_corr",
                "Node color by",
                choices = c(
                  "Community" = "community",
                  "Frequency (word count)" = "frequency"
                ),
                selected = "community"
              ),
              sliderInput(
                "nrows_correlation",
                "Row numbers",
                value = 1,
                min = 1,
                max = 10,
                step = 1
              ),
              numericInput(
                "seed_corr",
                "Layout seed",
                value = 2026, min = 1, max = 999999, step = 1
              ),
              div(style = "font-size: 16px;",
                  checkboxInput("showlegend_corr", "Show legend", value = TRUE)),
              div(
                class = "category-toggle",
                div(style = "font-size: 16px;", checkboxInput("use_category_corr",
                  "Use category-specific analysis",
                  value = FALSE
                ))
              )
            ),
            conditionalPanel(
              condition = "input.use_category_corr == true && input.doc_var_correlation != 'None' && input.doc_var_correlation != ''",
              tags$div(
                class = "status-main-warning",
                tags$i(class = "fa fa-info-circle status-icon status-icon-warning"),
                "Analyze top categories by document count for faster processing and clearer visualization."
              ),
              uiOutput("top_n_selector_corr"),
              uiOutput("category_selector_corr"),
              uiOutput("category_corr_controls")
            ),
            actionButton("plot_word_correlation_network", "Plot Network", class = "btn-primary btn-block"),
            conditionalPanel(
              condition = "output.has_correlation_plot == true",
              tags$div(
                style = "margin-top: 8px; display: flex; gap: 6px;",
                downloadButton("download_corr_html", "HTML",
                               class = "btn-default btn-block",
                               style = "margin: 0; flex: 1;")
              )
            )
          )
        ),
        mainPanel(
          width = 9,
          tabsetPanel(
            id = "semantic_analysis_tabs",
            tabPanel(
              "1. Setup",
              value = "summary",
              br(),
              conditionalPanel(
                condition = "output.has_documents == true",
                DT::dataTableOutput("document_summary_table")
              ),
              conditionalPanel(
                condition = "output.has_documents == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-upload", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Load data and process documents in the ",
                      tags$strong("1. Setup", style = "color: #4269BF;"),
                      " tab first",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "2. Word Co-occurrence",
              value = "cooccurrence",
              bsCollapse(
                open = 0,
                bsCollapsePanel(
                  p(strong("Click to set plot dimensions"),
                    class = "plot-dimensions-text"
                  ),
                  value = 4,
                  style = "success",
                  p(strong("Dimensions of the plot")),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "height_word_co_occurrence_network_plot",
                      post = " px",
                      label = "height",
                      min = 200,
                      max = 4000,
                      step = 5,
                      value = 700
                    )
                  ),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "width_word_co_occurrence_network_plot",
                      post = " px",
                      label = "width",
                      min = 500,
                      max = 3000,
                      step = 5,
                      value = 1000
                    )
                  )
                )
              ),
              tags$style(
                HTML(
                  ".plot-container {
                                max-height: 4000px;
                                max-width: 3000px;
                                overflow: auto; }"
                )
              ),
              conditionalPanel(
                condition = "output.has_cooccurrence_plot == true",
                tabsetPanel(
                  id = "word_co_occur_subTab",
                  tabPanel(
                    "Plot",
                    uiOutput("word_co_occurrence_network_plot_uiOutput")
                  ),
                  tabPanel(
                    "Table",
                    uiOutput("word_co_occurrence_network_table_uiOutput")
                  ),
                  tabPanel(
                    "Summary",
                    uiOutput("word_co_occurrence_network_summary_uiOutput")
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_cooccurrence_plot == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-project-diagram", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Configure settings and click ",
                      tags$strong("'Plot Network'", style = "color: #4269BF;"),
                      " to visualize word co-occurrence",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "3. Word Correlation",
              value = "correlation",
              bsCollapse(
                open = 0,
                bsCollapsePanel(
                  p(strong("Click to set plot dimensions"),
                    class = "plot-dimensions-text"
                  ),
                  value = 4,
                  style = "success",
                  p(strong("Dimensions of the plot")),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "height_word_correlation_network_plot",
                      post = " px",
                      label = "height",
                      min = 200,
                      max = 4000,
                      step = 5,
                      value = 700
                    )
                  ),
                  div(
                    style = "display:inline-block",
                    sliderInput(
                      inputId = "width_word_correlation_network_plot",
                      post = " px",
                      label = "width",
                      min = 500,
                      max = 3000,
                      step = 5,
                      value = 1000
                    )
                  )
                )
              ),
              tags$style(
                HTML(
                  ".plot-container {
                                max-height: 4000px;
                                max-width: 3000px;
                                overflow: auto; }"
                )
              ),
              conditionalPanel(
                condition = "output.has_correlation_plot == true",
                tabsetPanel(
                  id = "word_correlation_subTab",
                  tabPanel(
                    "Plot",
                    uiOutput("word_correlation_network_plot_uiOutput")
                  ),
                  tabPanel(
                    "Table",
                    uiOutput("word_correlation_network_table_uiOutput")
                  ),
                  tabPanel(
                    "Summary",
                    uiOutput("word_correlation_network_summary_uiOutput")
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_correlation_plot == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-share-alt", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Configure settings and click ",
                      tags$strong("'Plot Network'", style = "color: #4269BF;"),
                      " to visualize word correlation",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "4. Document Similarity",
              value = "similarity",
              conditionalPanel(
                condition = "output.has_documents == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-cog", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Process documents in the ",
                      tags$strong("1. Setup", style = "color: #4269BF;"),
                      " tab first",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_documents == true && output.has_similarity_calculation == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-calculator", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Configure settings and click ",
                      tags$strong("'Calculate'", style = "color: #4269BF;"),
                      " to begin analysis",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_documents == true && output.has_similarity_calculation == true && output.has_similarity_analysis == false",
                verbatimTextOutput("similarity_calculation_summary")
              ),
              conditionalPanel(
                condition = "output.has_documents == true && output.has_similarity_analysis == true",
                plotly::plotlyOutput("semantic_similarity_plot",
                  height = "auto", width = "100%"
                ),
                br(),
                DT::dataTableOutput("semantic_similarity_stats")
              )
            ),
            tabPanel(
              "5. Comparative Analysis",
              value = "comparative",
              br(),
              conditionalPanel(
                condition = "output.has_documents == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 500px; margin: 0 auto;",
                    tags$i(class = "fa fa-balance-scale", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Upload data to enable Comparative Analysis",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_documents == true && output.has_gap_analysis == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 500px; margin: 0 auto;",
                    tags$i(class = "fa fa-balance-scale", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Run Comparative Analysis from the sidebar to compare categories and identify unique content, gaps, and cross-category opportunities.",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_documents == true && output.has_gap_analysis == true",
                tags$h5(HTML("<strong>Comparative Analysis Results</strong>"), style = "color: #4269BF; margin-bottom: 15px;"),
                uiOutput("gap_reference_category_label"),
                tabsetPanel(
                  id = "gap_analysis_tabs",
                  tabPanel(
                    "Summary",
                    br(),
                    DT::dataTableOutput("gap_summary_stats")
                  ),
                  tabPanel(
                    "Heatmap",
                    br(),
                    tags$p("Cross-category similarity heatmap comparing reference documents against other categories.",
                           style = "color: #475569; font-size: 16px; margin-bottom: 10px;"),
                    plotly::plotlyOutput("gap_cross_category_heatmap", height = "600px")
                  ),
                  tabPanel(
                    "Unique (Reference)",
                    br(),
                    tags$p("Reference items with low similarity to all comparison categories (distinctive content).",
                           style = "color: #475569; font-size: 16px; margin-bottom: 10px;"),
                    DT::dataTableOutput("gap_unique_items")
                  ),
                  tabPanel(
                    "Missing (Comparison)",
                    br(),
                    tags$p("Comparison category items not well-covered by reference category (content gaps).",
                           style = "color: #475569; font-size: 16px; margin-bottom: 10px;"),
                    DT::dataTableOutput("gap_missing_items")
                  ),
                  tabPanel(
                    "Cross-Category",
                    br(),
                    tags$p("Items with moderate similarity - potential for cross-category learning or transfer.",
                           style = "color: #475569; font-size: 16px; margin-bottom: 10px;"),
                    DT::dataTableOutput("gap_cross_policy")
                  )
                )
              )
            ),
            tabPanel(
              "6. Semantic Search",
              value = "search",
              conditionalPanel(
                condition = "output.has_documents == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-cog", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Process documents in the ",
                      tags$strong("1. Setup", style = "color: #4269BF;"),
                      " tab first",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_documents == true && output.has_search_results == false",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-search", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;"),
                    tags$p(
                      "Enter a search query and click ",
                      tags$strong("'Search'", style = "color: #4269BF;"),
                      " to see results",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_documents == true && output.has_search_results == true",
                br(),
                uiOutput("search_method_indicator"),
                br(),
                DT::dataTableOutput("semantic_search_results")
              )
            ),
            tabPanel(
              "7. Sentiment & Emotion",
              value = "sentiment",
              br(),
              tabsetPanel(
                id = "sentiment_subtabs",
                tabPanel(
                  "Overall Sentiment",
                  value = "overall",
                  uiOutput("sentiment_overall_uiOutput")
                ),
                tabPanel(
                  "By Category",
                  value = "category",
                  uiOutput("sentiment_category_uiOutput")
                ),
                tabPanel(
                  "Document-Level",
                  value = "document",
                  uiOutput("sentiment_document_uiOutput")
                ),
                tabPanel(
                  "Emotion",
                  value = "emotion",
                  uiOutput("sentiment_emotion_uiOutput")
                )
              )
            ),
            tabPanel(
              "8. Document Groups",
              value = "clustering",
              br(),
              uiOutput("clustering_warning"),
              uiOutput("clustering_quality_metrics"),
              br(),
              tabsetPanel(
                id = "clustering_subtabs",
                tabPanel(
                  "Map",
                  br(),
                  plotly::plotlyOutput("semantic_cluster_plot", height = "600px")
                ),
                tabPanel(
                  "Groups Table",
                  br(),
                  DT::dataTableOutput("semantic_cluster_table")
                ),
                tabPanel(
                  "Group Details",
                  br(),
                  selectInput("selected_cluster_group", "Select a group", choices = NULL),
                  uiOutput("cluster_group_summary"),
                  br(),
                  plotly::plotlyOutput("cluster_terms_plot", height = "400px"),
                  br(),
                  DT::dataTableOutput("cluster_sample_docs")
                )
              )
            )
          )
        )
      )
}

#' @title Lexical Analysis Functions
#'
#' @description
#' Functions for lexical analysis including:
#' - Linguistic Annotation (POS tagging, NER)
#' - Frequency Analysis (word frequency, n-grams, MWEs)
#' - Keywords (TF-IDF, keyness)
#' - Lexical Diversity (TTR, MTLD, MATTR)
#' - Readability (Flesch, Gunning Fog, etc.)
#'
#' @name lexical_analysis
#' @concept lexical
NULL


.lexdiv_cache <- new.env(hash = TRUE, parent = emptyenv())

.mtld_one_direction <- function(toks, factor_size = 0.72) {
  n_toks <- length(toks)
  seen <- new.env(hash = TRUE, size = n_toks)
  unique_count <- 0
  factors <- 0
  start_idx <- 1

  for (i in seq_len(n_toks)) {
    token <- toks[i]
    if (!exists(token, envir = seen, inherits = FALSE)) {
      assign(token, TRUE, envir = seen)
      unique_count <- unique_count + 1
    }
    current_length <- i - start_idx + 1
    current_ttr <- unique_count / current_length

    # factors under 10 tokens are textual blips, not counted
    if (current_ttr <= factor_size && current_length >= 10) {
      factors <- factors + 1
      rm(list = ls(seen), envir = seen)
      unique_count <- 0
      start_idx <- i + 1
    }
  }

  if (start_idx <= n_toks) {
    remaining_length <- n_toks - start_idx + 1
    if (remaining_length > 0 && unique_count > 0) {
      final_ttr <- unique_count / remaining_length
      factors <- factors + (1 - final_ttr) / (1 - factor_size)
    }
  }

  if (factors > 0) n_toks / factors else NA_real_
}

.calc_mtld <- function(tokens, factor_size = 0.72) {
  if (length(tokens) < 10) return(NA_real_)
  mean(c(
    .mtld_one_direction(tokens, factor_size),
    .mtld_one_direction(rev(tokens), factor_size)
  ), na.rm = TRUE)
}

.calc_hdd <- function(tokens, sample_size = 42) {
  n_toks <- length(tokens)
  if (n_toks < sample_size) return(NA_real_)
  type_counts <- as.integer(table(tokens))
  # probability each type appears in a random 42-token draw
  sum((1 - stats::dhyper(0, type_counts, n_toks - type_counts, sample_size)) / sample_size)
}

#' Clear Lexical Diversity Cache
#'
#' @description
#' Clears the internal cache used for lexical diversity calculations.
#' Call this function to free memory or force fresh calculations.
#'
#' @return Invisible NULL
#' @concept lexical
#' @importFrom Matrix colSums
#' @export
clear_lexdiv_cache <- function() {
  rm(list = ls(.lexdiv_cache), envir = .lexdiv_cache)
  invisible(NULL)
}


#' @title Detect Multi-Word Expressions
#'
#' @description
#' This function detects multi-word expressions (collocations) of specified
#' sizes that appear at least a specified number of times in the provided tokens.
#'
#' @param tokens A \code{tokens} object from the \code{quanteda} package.
#' @param size A numeric vector specifying the sizes of the collocations to detect (default: 2:5).
#' @param min_count The minimum number of occurrences for a collocation to be
#'   considered (default: 2).
#'
#' @return A character vector of detected collocations.
#'
#' @concept lexical
#' @export
#'
#' @examples
#' \donttest{
#'   mydata <- TextAnalysisR::SpecialEduTech[seq_len(50), ]
#'
#'   united_tbl <- TextAnalysisR::unite_cols(
#'     mydata,
#'     listed_vars = c("title", "keyword", "abstract")
#'   )
#'
#'   tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")
#'
#'   collocations <- TextAnalysisR::detect_multi_words(tokens, size = 2:3, min_count = 2)
#'   print(collocations)
#' }
detect_multi_words <- function(tokens, size = 2:5, min_count = 2) {
  tstat <- quanteda.textstats::textstat_collocations(tokens, size = size, min_count = min_count)
  tstat_collocation <- tstat$collocation
  return(tstat_collocation)
}


#' Extract Part-of-Speech Tags from Tokens
#'
#' @description
#' Uses spaCy to extract part-of-speech (POS) tags from tokenized text.
#' Returns a data frame with token-level POS annotations.
#'
#' @param tokens A quanteda tokens object or character vector of texts.
#' @param include_lemma Logical; include lemmatized forms (default: TRUE).
#' @param include_entity Logical; include named entity recognition (default: FALSE).
#' @param include_dependency Logical; include dependency parsing (default: FALSE).
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#'
#' @return A data frame with columns:
#' \itemize{
#'   \item \code{doc_id}: Document identifier
#'   \item \code{sentence_id}: Sentence number within document
#'   \item \code{token_id}: Token position within sentence
#'   \item \code{token}: Original token
#'   \item \code{pos}: Universal POS tag (e.g., NOUN, VERB, ADJ)
#'   \item \code{tag}: Detailed POS tag (e.g., NN, VBD, JJ)
#'   \item \code{lemma}: Lemmatized form (if include_lemma = TRUE)
#'   \item \code{entity}: Named entity type (if include_entity = TRUE)
#'   \item \code{head_token_id}: Head token in dependency tree (if include_dependency = TRUE)
#'   \item \code{dep_rel}: Dependency relation type, e.g., nsubj, dobj (if include_dependency = TRUE)
#' }
#'
#' @details
#' This function requires the Python
#' with spaCy installed. If spaCy is not initialized, this function will
#' attempt to initialize it with the specified model.
#'
#' @concept lexical
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1])
#'   pos_data <- extract_pos_tags(tokens)
#'   print(pos_data)
#' }
extract_pos_tags <- function(tokens,
                             include_lemma = TRUE,
                             include_entity = FALSE,
                             include_dependency = FALSE,
                             model = "en_core_web_sm") {

  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("POS tagging via spaCy"))
  }

  parsed <- spacy_parse_full(
    tokens,
    pos = TRUE,
    tag = TRUE,
    lemma = include_lemma,
    entity = include_entity,
    dependency = include_dependency,
    model = model
  )

  return(parsed)
}


#' Extract Morphological Features
#'
#' @description
#' Uses spaCy to extract morphological features from text.
#' Returns data with Number, Tense, VerbForm, Person, Case, Mood, Aspect, etc.
#'
#' @param tokens A quanteda tokens object or character vector of texts.
#' @param features Character vector of morphological features to extract.
#'   Default includes common Universal Dependencies features.
#' @param include_pos Logical; include POS tags (default: TRUE).
#' @param include_lemma Logical; include lemmatized forms (default: TRUE).
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#'
#' @return A data frame with token-level morphological annotations including
#'   morph_* columns for each requested feature.
#'
#' @details
#' Morphological features follow Universal Dependencies annotation.
#' Common features include:
#' \itemize{
#'   \item \code{Number}: Sing (singular), Plur (plural)
#'   \item \code{Tense}: Past, Pres (present), Fut (future)
#'   \item \code{VerbForm}: Fin (finite), Inf (infinitive), Part (participle), Ger (gerund)
#'   \item \code{Person}: 1, 2, 3 (first, second, third person)
#'   \item \code{Case}: Nom (nominative), Acc (accusative), Gen (genitive), Dat (dative)
#'   \item \code{Mood}: Ind (indicative), Imp (imperative), Sub (subjunctive)
#'   \item \code{Aspect}: Perf (perfective), Imp (imperfective), Prog (progressive)
#' }
#'
#' @concept lexical
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1])
#'   morphology_data <- extract_morphology(tokens)
#'   print(morphology_data)
#' }
extract_morphology <- function(tokens,
                               features = c("Number", "Tense", "VerbForm",
                                            "Person", "Case", "Mood", "Aspect"),
                               include_pos = TRUE,
                               include_lemma = TRUE,
                               model = "en_core_web_sm") {

  # Use spacy_parse_full with morphology enabled
  parsed <- spacy_parse_full(
    tokens,
    pos = include_pos,
    tag = include_pos,
    lemma = include_lemma,
    entity = FALSE,
    dependency = FALSE,
    morph = TRUE,
    model = model
  )

  # Parse the morph string into individual feature columns
  if ("morph" %in% names(parsed) && nrow(parsed) > 0) {
    parsed <- parse_morphology_string(parsed, features)
  }

  return(parsed)
}


#' Plot Morphology Feature Distribution
#'
#' @description
#' Creates a bar chart showing the distribution of a morphological feature
#' using consistent package styling.
#'
#' @param data Data frame with morph_* columns from extract_morphology().
#' @param feature Character; feature name (e.g., "Number", "Tense").
#' @param title Character; plot title (auto-generated if NULL).
#' @param colors Named character vector of custom colors for feature values.
#'
#' @return A plotly object.
#'
#' @concept lexical
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1])
#'   morphology_data <- extract_morphology(tokens)
#'   plot_morphology_feature(morphology_data, "Tense")
#' }
plot_morphology_feature <- function(data,
                                    feature,
                                    title = NULL,
                                    colors = NULL) {

  col_name <- paste0("morph_", feature)

  if (!col_name %in% names(data)) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = paste("Feature", feature, "not available"),
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  values <- data[[col_name]]
  values <- values[!is.na(values) & values != ""]

  if (length(values) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = paste("No", feature, "data found"),
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  freq_df <- as.data.frame(table(values), stringsAsFactors = FALSE)
  names(freq_df) <- c("Value", "Count")
  freq_df <- freq_df[order(-freq_df$Count), ]
  freq_df$Percentage <- round(freq_df$Count / sum(freq_df$Count) * 100, 1)

  if (is.null(title)) {
    title <- paste(feature, "Distribution")
  }

  if (is.null(colors)) {
    colors <- switch(feature,
      "Number" = c("Sing" = "#3B82F6", "Plur" = "#10B981"),
      "Tense" = c("Past" = "#EF4444", "Pres" = "#3B82F6", "Fut" = "#10B981"),
      "VerbForm" = c("Fin" = "#3B82F6", "Inf" = "#8B5CF6",
                     "Part" = "#F59E0B", "Ger" = "#10B981"),
      "Person" = c("1" = "#3B82F6", "2" = "#10B981", "3" = "#F59E0B"),
      "Case" = c("Nom" = "#3B82F6", "Acc" = "#10B981",
                 "Gen" = "#F59E0B", "Dat" = "#8B5CF6"),
      "Mood" = c("Ind" = "#3B82F6", "Imp" = "#EF4444", "Sub" = "#8B5CF6"),
      "Aspect" = c("Perf" = "#3B82F6", "Imp" = "#10B981", "Prog" = "#F59E0B"),
      NULL
    )
  }

  bar_colors <- if (!is.null(colors) && length(colors) > 0) {
    vapply(freq_df$Value, function(v) {
      if (v %in% names(colors)) colors[[v]] else "#6B7280"
    }, character(1))
  } else {
    rep("#337ab7", nrow(freq_df))
  }

  freq_df$fill_color <- bar_colors

  ggplot2::ggplot(freq_df, ggplot2::aes(x = Value, y = Count,
                                         text = paste0(Value, "\nCount: ", Count, "\n", Percentage, "%"))) +
    ggplot2::geom_col(fill = freq_df$fill_color) +
    ggplot2::labs(x = "", y = "Frequency", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}


#' Summarize Morphology Features
#'
#' @description
#' Creates a summary table of morphological feature distributions
#' with counts and percentages for each feature value.
#'
#' @param data Data frame with morph_* columns from extract_morphology().
#' @param features Character vector of features to summarize.
#'   If NULL, all available morph_* columns are used.
#'
#' @return A data frame with Feature, Value, Count, and Percentage columns.
#'
#' @concept lexical
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1])
#'   morphology_data <- extract_morphology(tokens)
#'   summary_table <- summarize_morphology(morphology_data)
#'   print(summary_table)
#' }
summarize_morphology <- function(data, features = NULL) {
  morph_cols <- grep("^morph_", names(data), value = TRUE)

  if (length(morph_cols) == 0) {
    return(data.frame(
      Feature = character(0),
      Value = character(0),
      Count = integer(0),
      Percentage = numeric(0),
      stringsAsFactors = FALSE
    ))
  }

  if (!is.null(features)) {
    target_cols <- paste0("morph_", features)
    morph_cols <- intersect(morph_cols, target_cols)
  }

  summary_list <- lapply(morph_cols, function(col) {
    feat_name <- gsub("morph_", "", col)
    values <- data[[col]]
    values <- values[!is.na(values) & values != ""]

    if (length(values) == 0) return(NULL)

    counts <- as.data.frame(table(values), stringsAsFactors = FALSE)
    names(counts) <- c("Value", "Count")
    counts$Feature <- feat_name
    counts$Percentage <- round(counts$Count / sum(counts$Count) * 100, 1)
    counts[, c("Feature", "Value", "Count", "Percentage")]
  })

  result <- do.call(rbind, summary_list)
  if (is.null(result)) {
    return(data.frame(
      Feature = character(0),
      Value = character(0),
      Count = integer(0),
      Percentage = numeric(0),
      stringsAsFactors = FALSE
    ))
  }

  return(result)
}


#' Extract Named Entities from Tokens
#'
#' @description
#' Uses spaCy to extract named entities (NER) from tokenized text.
#' Returns a data frame with token-level entity annotations.
#'
#' @param tokens A quanteda tokens object or character vector of texts.
#' @param include_pos Logical; include POS tags (default: TRUE).
#' @param include_lemma Logical; include lemmatized forms (default: TRUE).
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#'
#' @return A data frame with columns:
#' \itemize{
#'   \item \code{doc_id}: Document identifier
#'   \item \code{token}: Original token
#'   \item \code{entity}: Named entity type (e.g., PERSON, ORG, GPE)
#'   \item \code{pos}: Universal POS tag (if include_pos = TRUE)
#'   \item \code{lemma}: Lemmatized form (if include_lemma = TRUE)
#' }
#'
#' @details
#' This function requires the Python
#' with spaCy installed. If spaCy is not initialized, this function will
#' attempt to initialize it with the specified model.
#'
#' @concept lexical
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1])
#'   entity_data <- extract_named_entities(tokens)
#'   print(entity_data)
#' }
extract_named_entities <- function(tokens,
                                   include_pos = TRUE,
                                   include_lemma = TRUE,
                                   model = "en_core_web_sm") {

  # Use spacy_parse_full with entity enabled
  parsed <- spacy_parse_full(
    tokens,
    pos = include_pos,
    tag = include_pos,
    lemma = include_lemma,
    entity = TRUE,
    dependency = FALSE,
    model = model
  )

  return(parsed)
}


#' Lexical Diversity Analysis
#'
#' @description
#' Calculates multiple lexical diversity metrics for a document-feature matrix (DFM)
#' or tokens object. Supports all quanteda.textstats measures plus MTLD
#' (Measure of Textual Lexical Diversity), which is the most recommended measure
#' according to McCarthy & Jarvis (2010) for being independent of text length.
#'
#' @param x A quanteda DFM or tokens object. Tokens object is preferred for
#'   accurate MTLD calculation since it preserves token order.
#' @param measures Character vector of measures to calculate.
#'   Default is "all" which includes: TTR, C, R, CTTR, U, S, K, I, D, Vm, Maas, MATTR, MSTTR, MTLD, and HDD.
#'   Most recommended: "MTLD", "MATTR", or "HDD" for length-independent measures.
#' @param texts Optional character vector of original texts. Required for MTLD
#'   calculation when using DFM input (since DFM loses token order).
#' @param cache_key Optional cache key (e.g., from digest::digest) for caching
#'   expensive calculations. Use the same cache_key to retrieve cached results.
#'
#' @return A list containing:
#' \itemize{
#'   \item \code{lexical_diversity}: Data frame with per-document lexical diversity scores
#'   \item \code{summary_stats}: List of summary statistics (mean, median, sd) for each measure
#' }
#'
#' @details
#' MTLD (Measure of Textual Lexical Diversity) is calculated using the algorithm

#' from McCarthy & Jarvis (2010). It counts the number of "factors" needed to
#' reduce TTR below 0.72, then divides the number of tokens by the number of factors.
#' This provides a length-independent measure of lexical diversity.
#'
#' Important notes:
#' \itemize{
#'   \item For MTLD accuracy, pass a tokens object (not DFM) as input
#'   \item If using DFM, provide the 'texts' parameter for MTLD calculation
#'   \item MATTR and MSTTR window sizes are automatically adjusted for short documents
#'   \item Raw TTR falls mechanically as documents lengthen; compare TTR only
#'     across documents of similar length
#'   \item MTLD and MATTR are most reliable at 100+ tokens per document
#'   \item Results are cached when cache_key is provided for repeated analysis
#' }
#'
#' @references
#' McCarthy, P. M., & Jarvis, S. (2010). MTLD, vocd-D, and HD-D: A validation study
#' of sophisticated approaches to lexical diversity assessment.
#' Behavior Research Methods, 42(2), 381-392.
#'
#' @concept lexical
#' @seealso [calculate_text_readability()] for grade-level / Flesch metrics on the same input; [calculate_lexical_dispersion()] for term spread across documents; [plot_lexical_diversity_distribution()] to visualize
#' @export
#'
#' @examples
#' \donttest{
#' data(SpecialEduTech)
#' texts <- SpecialEduTech$abstract[1:10]
#' corp <- quanteda::corpus(texts)
#' toks <- quanteda::tokens(corp)
#' # Preferred: pass tokens object for accurate MTLD
#' lex_div <- lexical_diversity_analysis(toks, texts = texts)
#' # With caching for repeated analysis
#' cache_key <- digest::digest(texts)
#' lex_div <- lexical_diversity_analysis(toks, texts = texts, cache_key = cache_key)
#' # Alternative: pass DFM with texts for MTLD accuracy
#' dfm_obj <- quanteda::dfm(toks)
#' lex_div <- lexical_diversity_analysis(dfm_obj, texts = texts)
#' print(lex_div)
#' }
#'
#' @importFrom quanteda.textstats textstat_lexdiv
#' @importFrom quanteda docnames
lexical_diversity_analysis <- function(x,
                                      measures = "all",
                                      texts = NULL,
                                      cache_key = NULL) {

  # Check cache first if cache_key is provided

  if (!is.null(cache_key) && nzchar(cache_key)) {
    cache_id <- paste0("lexdiv_", cache_key)
    if (exists(cache_id, envir = .lexdiv_cache, inherits = FALSE)) {
      return(get(cache_id, envir = .lexdiv_cache, inherits = FALSE))
    }
  }

  if (!requireNamespace("quanteda.textstats", quietly = TRUE)) {
    stop("Package 'quanteda.textstats' is required. Please install it.")
  }

  quanteda_measures <- c("TTR", "C", "R", "CTTR", "U", "S", "K", "I", "D", "Vm", "Maas")
  mtld_requested <- FALSE
  mattr_requested <- FALSE
  msttr_requested <- FALSE
  hdd_requested <- FALSE

  if ("all" %in% measures) {
    measures_to_use <- quanteda_measures
    mtld_requested <- TRUE
    mattr_requested <- TRUE
    msttr_requested <- TRUE
    hdd_requested <- TRUE
  } else {
    if ("MTLD" %in% measures) {
      mtld_requested <- TRUE
      measures <- setdiff(measures, "MTLD")
    }
    if ("MATTR" %in% measures) {
      mattr_requested <- TRUE
      measures <- setdiff(measures, "MATTR")
    }
    if ("MSTTR" %in% measures) {
      msttr_requested <- TRUE
      measures <- setdiff(measures, "MSTTR")
    }
    if ("HDD" %in% measures) {
      hdd_requested <- TRUE
      measures <- setdiff(measures, "HDD")
    }
    measures_to_use <- intersect(measures, quanteda_measures)
  }

  is_tokens_input <- inherits(x, "tokens")
  seq_tokens <- NULL
  if ((mtld_requested || mattr_requested || msttr_requested || hdd_requested) && !is_tokens_input) {
    if (!is.null(texts) && length(texts) == quanteda::ndoc(x)) {
      seq_tokens <- quanteda::tokens(texts, remove_punct = TRUE)
    } else {
      message("MTLD/MATTR/MSTTR/HDD require sequential token order. DFM input loses token order. ",
              "Pass a tokens object or provide the 'texts' parameter. Skipping.")
      mtld_requested <- FALSE
      mattr_requested <- FALSE
      msttr_requested <- FALSE
      hdd_requested <- FALSE
    }
  }

  tryCatch({
    if (is_tokens_input) {
      x_dfm <- quanteda::dfm(x)
    } else {
      x_dfm <- x
    }

    doc_lengths <- quanteda::ntoken(x_dfm)
    valid_mask <- doc_lengths > 0
    min_length <- if (any(valid_mask)) min(doc_lengths[valid_mask]) else 0

    window_size <- min(100, max(10, min_length))

    if (length(measures_to_use) > 0 && any(valid_mask)) {
      if (all(valid_mask)) {
        lexdiv_results <- suppressWarnings(
          quanteda.textstats::textstat_lexdiv(
            x_dfm,
            measure = measures_to_use,
            MATTR_window = window_size,
            MSTTR_segment = window_size
          )
        )
      } else {
        x_valid <- quanteda::dfm_subset(x_dfm, valid_mask)
        valid_results <- suppressWarnings(
          quanteda.textstats::textstat_lexdiv(
            x_valid,
            measure = measures_to_use,
            MATTR_window = min(100, max(10, min(quanteda::ntoken(x_valid)))),
            MSTTR_segment = min(100, max(10, min(quanteda::ntoken(x_valid))))
          )
        )
        lexdiv_results <- data.frame(document = quanteda::docnames(x_dfm))
        for (m in measures_to_use) {
          lexdiv_results[[m]] <- NA_real_
          lexdiv_results[[m]][valid_mask] <- valid_results[[m]]
        }
      }
    } else {
      lexdiv_results <- data.frame(document = quanteda::docnames(x_dfm))
    }

    if (mtld_requested) {
      tryCatch({
        tokens_source <- if (is_tokens_input) x else seq_tokens

        mtld_values <- vapply(seq_len(quanteda::ndoc(tokens_source)), function(i) {
          .calc_mtld(as.character(tokens_source[[i]]))
        }, numeric(1))

        lexdiv_results$MTLD <- as.numeric(mtld_values)
      }, error = function(e) {
        message("MTLD calculation failed: ", e$message, ". Skipping MTLD.")
      })
    }

    if (mattr_requested) {
      tryCatch({
        tokens_source <- if (is_tokens_input) x else seq_tokens
        mattr_window <- min(window_size, min(quanteda::ntoken(tokens_source)))

        mattr_values <- vapply(seq_len(quanteda::ndoc(tokens_source)), function(i) {
          doc_tokens <- as.character(tokens_source[[i]])
          n <- length(doc_tokens)
          if (n < mattr_window) return(NA_real_)
          ttrs <- vapply(seq_len(n - mattr_window + 1), function(j) {
            length(unique(doc_tokens[j:(j + mattr_window - 1)])) / mattr_window
          }, numeric(1))
          mean(ttrs)
        }, numeric(1))

        lexdiv_results$MATTR <- as.numeric(mattr_values)
      }, error = function(e) {
        message("MATTR calculation failed: ", e$message, ". Skipping MATTR.")
      })
    }

    if (msttr_requested) {
      tryCatch({
        tokens_source <- if (is_tokens_input) x else seq_tokens
        msttr_segment <- min(50, min(quanteda::ntoken(tokens_source)))

        msttr_values <- vapply(seq_len(quanteda::ndoc(tokens_source)), function(i) {
          doc_tokens <- as.character(tokens_source[[i]])
          n <- length(doc_tokens)
          if (n < msttr_segment) return(NA_real_)
          n_segments <- n %/% msttr_segment
          if (n_segments < 1) return(NA_real_)
          ttrs <- vapply(seq_len(n_segments), function(j) {
            seg <- doc_tokens[((j - 1) * msttr_segment + 1):(j * msttr_segment)]
            length(unique(seg)) / msttr_segment
          }, numeric(1))
          mean(ttrs)
        }, numeric(1))

        lexdiv_results$MSTTR <- as.numeric(msttr_values)
      }, error = function(e) {
        message("MSTTR calculation failed: ", e$message, ". Skipping MSTTR.")
      })
    }

    if (hdd_requested) {
      tryCatch({
        tokens_source <- if (is_tokens_input) x else seq_tokens

        hdd_values <- vapply(seq_len(quanteda::ndoc(tokens_source)), function(i) {
          .calc_hdd(as.character(tokens_source[[i]]))
        }, numeric(1))

        lexdiv_results$HDD <- as.numeric(hdd_values)
      }, error = function(e) {
        message("HDD calculation failed: ", e$message, ". Skipping HDD.")
      })
    }

    if (!"document" %in% names(lexdiv_results)) {
      lexdiv_results$document <- quanteda::docnames(x)
    }

    # Standardize document names to "Doc 1, Doc 2..." format
    lexdiv_results$document <- paste0("Doc ", seq_len(nrow(lexdiv_results)))

    # Get actual column names from result (after MTLD is added)
    actual_measures <- setdiff(names(lexdiv_results), "document")

    # Select only columns that exist
    cols_to_keep <- c("document", actual_measures)
    lexdiv_results <- lexdiv_results[, cols_to_keep, drop = FALSE]

    # Add Average Sentence Length if texts provided
    if (!is.null(texts) && length(texts) == nrow(lexdiv_results)) {
      # quanteda sentence tokenizer matches calculate_text_readability()
      avg_sentence_length <- vapply(texts, function(t) {
        sents <- quanteda::tokens(t, what = "sentence")[[1]]
        sents <- sents[nzchar(trimws(sents))]
        if (length(sents) == 0) return(NA_real_)
        words <- sum(lengths(quanteda::tokens(sents, what = "word", remove_punct = TRUE)))
        words / length(sents)
      }, numeric(1))
      lexdiv_results$`Avg Sentence Length` <- avg_sentence_length
      actual_measures <- c(actual_measures, "Avg Sentence Length")
    }

    summary_stats <- list(
      n_documents = nrow(lexdiv_results),
      measures_calculated = actual_measures
    )

    for (measure in actual_measures) {
      if (measure %in% names(lexdiv_results)) {
        summary_stats[[paste0(measure, "_mean")]] <- mean(lexdiv_results[[measure]], na.rm = TRUE)
        summary_stats[[paste0(measure, "_median")]] <- median(lexdiv_results[[measure]], na.rm = TRUE)
        summary_stats[[paste0(measure, "_sd")]] <- sd(lexdiv_results[[measure]], na.rm = TRUE)
      }
    }

    result <- list(
      lexical_diversity = lexdiv_results,
      summary_stats = summary_stats
    )

    # Store in cache if cache_key provided
    if (!is.null(cache_key) && nzchar(cache_key)) {
      cache_id <- paste0("lexdiv_", cache_key)
      assign(cache_id, result, envir = .lexdiv_cache)
    }

    return(result)

  }, error = function(e) {
    stop("Error calculating lexical diversity: ", e$message)
  })
}


#' Plot Lexical Diversity Distribution
#'
#' @description
#' Creates a boxplot showing the distribution of a lexical diversity metric.
#'
#' @param lexdiv_data Data frame from lexical_diversity_analysis()
#' @param metric Metric to plot. Recommended: "MTLD" or "MATTR" (text-length independent)
#' @param title Plot title (default: auto-generated)
#'
#' @return A plotly boxplot
#'
#' @concept lexical
#' @export
#'
#' @examples
#' \donttest{
#' abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
#' tokens <- quanteda::tokens(quanteda::corpus(abstracts))
#' diversity_result <- lexical_diversity_analysis(tokens, texts = abstracts)
#' diversity_plot <- plot_lexical_diversity_distribution(
#'   diversity_result$lexical_diversity, "MTLD"
#' )
#' print(diversity_plot)
#' }
plot_lexical_diversity_distribution <- function(lexdiv_data,
                                               metric,
                                               title = NULL) {

  if (!metric %in% names(lexdiv_data)) {
    stop(paste("Metric", metric, "not found in lexical diversity data"))
  }

  if (is.null(title)) {
    title <- paste(metric, "- Overall Distribution")
  }

  metric_values <- lexdiv_data[[metric]]
  metric_values <- metric_values[is.finite(metric_values)]

  if (length(metric_values) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "No valid data for selected metric",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  plot_df <- data.frame(
    value = metric_values,
    doc = paste("Doc", seq_along(metric_values)),
    x = ""
  )
  plot_df$hover_text <- paste0("Document: ", plot_df$doc,
                               "<br>", metric, ": ", round(plot_df$value, 4))

  ggplot2::ggplot(plot_df, ggplot2::aes(x = x, y = value)) +
    ggplot2::geom_boxplot(fill = "#8B5CF6", alpha = 0.7,
                          color = "#0c1f4a", outlier.shape = NA) +
    suppressWarnings(ggplot2::geom_jitter(ggplot2::aes(text = hover_text),
                         width = 0.15, alpha = 0.5, color = "#0c1f4a", size = 1.8)) +
    ggplot2::labs(y = metric, x = "", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      axis.text.x = ggplot2::element_blank(),
      axis.ticks.x = ggplot2::element_blank()
    )

}


#' @title Lexical Frequency Analysis
#'
#' @description
#' Wrapper function for plot_word_frequency for lexical analysis.
#'
#' @param ... Arguments passed to plot_word_frequency
#'
#' @return A plotly bar chart of word frequencies
#'
#' @concept lexical
#' @export
lexical_frequency_analysis <- function(...) {
  return(plot_word_frequency(...))
}



#' @title Plot Word Frequency
#'
#' @description
#' Creates a bar plot showing the most frequent words in a document-feature matrix (dfm).
#'
#' @param dfm_object A document-feature matrix created by quanteda::dfm().
#' @param n The number of top words to display (default: 20).
#' @param height Plot height in pixels (default: 800). Kept for backward compatibility.
#' @param width Plot width in pixels (default: 1000). Kept for backward compatibility.
#' @param ... Additional arguments (kept for backward compatibility).
#'
#' @return A ggplot object showing word frequency.
#'
#' @concept visualization
#' @export
#'
#' @examples
#' \donttest{
#'   data(SpecialEduTech, package = "TextAnalysisR")
#'   texts <- SpecialEduTech$abstract[1:10]
#'   dfm <- quanteda::dfm(quanteda::tokens(texts))
#'   plot <- plot_word_frequency(dfm, n = 10)
#'   print(plot)
#' }
plot_word_frequency <- function(dfm_object,
                                n = 20,
                                height = NULL,
                                width = NULL,
                                ...) {

  if (!inherits(dfm_object, "dfm")) {
    stop("Input must be a quanteda dfm object")
  }

  freq_df <- quanteda.textstats::textstat_frequency(dfm_object, n = n) %>%
    dplyr::mutate(
      feature = stats::reorder(feature, frequency)
    )

  ggplot_obj <- ggplot2::ggplot(freq_df,
                                ggplot2::aes(x = feature, y = frequency,
                                            text = paste("Word:", feature,
                                                       "<br>Frequency:", frequency))) +
    ggplot2::geom_point(color = "#0c1f4a", size = 2.5, alpha = 0.9) +
    ggplot2::scale_x_discrete(expand = ggplot2::expansion(add = 0.5)) +
    ggplot2::coord_flip() +
    ggplot2::labs(x = "", y = "Frequency") +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      panel.grid.major.x = ggplot2::element_blank(),
      panel.grid.major.y = ggplot2::element_line(color = "#E0E0E0", linewidth = 0.3),
      panel.grid.minor = ggplot2::element_blank(),
      axis.line.x = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      axis.line.y = ggplot2::element_blank(),
      axis.ticks.x = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      axis.ticks.y = ggplot2::element_blank(),
      axis.text.x = ggplot2::element_text(size = 11, color = "#3B3B3B", margin = ggplot2::margin(t = 3)),
      axis.text.y = ggplot2::element_text(size = 11, color = "#3B3B3B", margin = ggplot2::margin(r = 3)),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      axis.title.x = ggplot2::element_text(margin = ggplot2::margin(t = 5)),
      axis.title.y = ggplot2::element_text(margin = ggplot2::margin(r = 5)),
      plot.margin = ggplot2::margin(t = 5, r = 10, b = 5, l = 5)
    )

  ggplot_obj
}


#' Plot N-gram Frequency
#'
#' @description
#' Creates a bar plot showing n-gram frequencies with optional highlighting
#' of selected n-grams. Supports both detected n-grams and selected multi-word expressions.
#'
#' @param ngram_data Data frame containing n-gram data with columns:
#'   \itemize{
#'     \item \code{collocation}: The n-gram text
#'     \item \code{count}: Frequency count
#'     \item \code{lambda}: (optional) Lambda statistic
#'     \item \code{z}: (optional) Z-score statistic
#'   }
#' @param top_n Number of top n-grams to display (default: 30)
#' @param selected Character vector of selected n-grams to highlight (default: NULL)
#' @param title Plot title (default: "N-gram Frequency")
#' @param highlight_color Color for highlighted bars (default: "#10B981")
#' @param default_color Color for non-highlighted bars (default: "#6B7280")
#' @param height Plot height in pixels (default: 500)
#' @param width Plot width in pixels (default: NULL for auto)
#' @param show_stats Whether to show lambda and z-score in hover (default: TRUE)
#'
#' @return A plotly object
#'
#' @concept visualization
#' @export
#'
#' @examples
#' \donttest{
#'   ngram_df <- data.frame(
#'     collocation = c("machine learning", "deep learning", "neural network"),
#'     count = c(150, 120, 90),
#'     lambda = c(5.2, 4.8, 4.1),
#'     z = c(12.3, 10.5, 9.2)
#'   )
#'   plot_ngram_frequency(ngram_df, selected = c("machine learning"))
#' }
plot_ngram_frequency <- function(ngram_data,
                                  top_n = 30,
                                  selected = NULL,
                                  title = "N-gram Frequency",
                                  highlight_color = "#10B981",
                                  default_color = "#6B7280",
                                  height = 500,
                                  width = NULL,
                                  show_stats = TRUE) {

  if (is.null(ngram_data) || nrow(ngram_data) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "No n-grams detected. Adjust parameters and click 'Detect N-grams'",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  top_ngrams <- utils::head(ngram_data, top_n)

  top_ngrams <- top_ngrams %>%
    dplyr::mutate(
      order_rank = dplyr::row_number(),
      collocation_ordered = factor(collocation, levels = rev(collocation))
    )

  is_selected <- if (!is.null(selected)) {
    top_ngrams$collocation %in% selected
  } else {
    rep(FALSE, nrow(top_ngrams))
  }

  top_ngrams$bar_fill <- ifelse(is_selected, highlight_color, default_color)
  top_ngrams$bar_border <- ifelse(is_selected, "#337ab7", "#4B5563")

  top_ngrams$hover_text <- if (show_stats && "lambda" %in% names(top_ngrams) && "z" %in% names(top_ngrams)) {
    paste0(top_ngrams$collocation, "\nFrequency: ", top_ngrams$count,
           "\nLambda: ", round(top_ngrams$lambda, 2),
           "\nZ-score: ", round(top_ngrams$z, 2))
  } else {
    paste0(top_ngrams$collocation, "\nFrequency: ", top_ngrams$count)
  }

  ggplot2::ggplot(top_ngrams, ggplot2::aes(x = collocation_ordered, y = count,
                                            text = hover_text)) +
    ggplot2::geom_col(fill = top_ngrams$bar_fill, color = top_ngrams$bar_border,
                      linewidth = 0.3) +
    ggplot2::labs(x = "", y = "Frequency", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text.x = ggplot2::element_text(size = 11, color = "#3B3B3B", angle = -45, hjust = 0),
      axis.text.y = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}


#' Plot Multi-Word Expression Frequency
#'
#' @description
#' Creates a bar plot showing multi-word expression frequencies with optional
#' source-based coloring to distinguish between detected and manually added expressions.
#'
#' @param mwe_data Data frame containing MWE data with columns:
#'   \itemize{
#'     \item \code{feature}: The multi-word expression text
#'     \item \code{frequency}: Frequency count
#'     \item \code{rank}: (optional) Rank of the expression
#'     \item \code{docfreq}: (optional) Document frequency
#'     \item \code{source}: (optional) Source category (e.g., "Top 20", "Manual")
#'   }
#' @param title Plot title (default: "Multi-Word Expression Frequency")
#' @param color_by_source Whether to color bars by source column (default: TRUE)
#' @param primary_color Color for primary/top expressions (default: "#10B981")
#' @param secondary_color Color for secondary/manual expressions (default: "#A855F7")
#' @param height Plot height in pixels (default: 500)
#' @param width Plot width in pixels (default: NULL for auto)
#'
#' @return A plotly object
#'
#' @concept visualization
#' @export
plot_mwe_frequency <- function(mwe_data,
                                title = "Multi-Word Expression Frequency",
                                color_by_source = TRUE,
                                primary_color = "#10B981",
                                secondary_color = "#A855F7",
                                height = 500,
                                width = NULL) {

  if (is.null(mwe_data) || nrow(mwe_data) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "No multi-word expressions found",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  if (color_by_source && "source" %in% names(mwe_data)) {
    mwe_data$bar_fill <- ifelse(mwe_data$source == "Top 20", primary_color, secondary_color)
  } else {
    mwe_data$bar_fill <- primary_color
  }

  mwe_data$hover_text <- if (all(c("rank", "docfreq", "source") %in% names(mwe_data))) {
    paste0(mwe_data$feature, "\nFrequency: ", mwe_data$frequency,
           "\nRank: ", mwe_data$rank,
           "\nDoc Frequency: ", mwe_data$docfreq,
           "\nSource: ", mwe_data$source)
  } else {
    paste0(mwe_data$feature, "\nFrequency: ", mwe_data$frequency)
  }

  mwe_data$feature_ordered <- stats::reorder(mwe_data$feature, mwe_data$frequency)

  ggplot2::ggplot(mwe_data, ggplot2::aes(x = feature_ordered, y = frequency,
                                          text = hover_text)) +
    ggplot2::geom_col(fill = mwe_data$bar_fill) +
    ggplot2::labs(x = "", y = "Frequency", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text.x = ggplot2::element_text(size = 11, color = "#3B3B3B", angle = -45, hjust = 0),
      axis.text.y = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}



# Keyword extraction

#' Extract Keywords Using TF-IDF
#'
#' @description
#' Extracts top keywords from a document-feature matrix using TF-IDF weighting.
#' Uses quanteda's default scheme (raw term counts, base-10 idf), which
#' differs from tidytext's proportion-based tf; rankings favor terms in
#' longer documents.
#'
#' @param dfm A quanteda dfm object
#' @param top_n Number of top keywords to extract (default: 20)
#' @param normalize Logical, whether to normalize TF-IDF scores to 0-1 range (default: FALSE)
#'
#' @return Data frame with columns: Keyword, TF_IDF_Score, Frequency
#'
#' @concept lexical
#' @export
#'
#' @examples
#' \donttest{
#'   library(quanteda)
#'   corp <- corpus(c("text analysis", "data mining", "text mining"))
#'   dfm_obj <- dfm(tokens(corp))
#'   keywords <- extract_keywords_tfidf(dfm_obj, top_n = 5)
#'   print(keywords)
#' }
extract_keywords_tfidf <- function(dfm,
                                   top_n = 20,
                                   normalize = FALSE) {

  if (!requireNamespace("quanteda", quietly = TRUE)) {
    stop("Package 'quanteda' is required.")
  }

  tfidf <- quanteda::dfm_tfidf(dfm)

  feature_scores <- colSums(as.matrix(tfidf))
  feature_freq <- colSums(as.matrix(dfm))

  if (normalize) {
    max_score <- max(feature_scores)
    if (max_score > 0) {
      feature_scores <- feature_scores / max_score
    }
  }

  top_features <- sort(feature_scores, decreasing = TRUE)[seq_len(min(top_n, length(feature_scores)))]

  data.frame(
    Keyword = names(top_features),
    TF_IDF_Score = unname(top_features),
    Frequency = unname(feature_freq[names(top_features)]),
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}


#' Extract Keywords Using Statistical Keyness
#'
#' @description
#' Extracts distinctive keywords by comparing document groups using log-likelihood ratio (G-squared).
#'
#' @param dfm A quanteda dfm object
#' @param target Target document indices or logical vector
#' @param top_n Number of top keywords to extract (default: 20)
#' @param measure Keyness measure: "lr" (log-likelihood G-squared), "chi2",
#'   "exact" (Fisher's exact odds ratio), or "pmi" (default: "lr")
#' @param min_count Minimum total term frequency before computing keyness
#'   (default: 0, no pruning)
#'
#' @return Data frame with columns: Keyword, Keyness_Score
#'
#' @concept lexical
#' @export
#'
#' @examples
#' \donttest{
#' abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
#' dfm_object <- quanteda::dfm(quanteda::tokens(quanteda::corpus(abstracts)))
#' keywords <- extract_keywords_keyness(dfm_object, target = 1)
#' print(keywords)
#' }
extract_keywords_keyness <- function(dfm,
                                     target,
                                     top_n = 20,
                                     measure = "lr",
                                     min_count = 0) {

  measure <- match.arg(measure, c("lr", "chi2", "exact", "pmi"))

  if (!requireNamespace("quanteda.textstats", quietly = TRUE)) {
    stop("Package 'quanteda.textstats' is required.")
  }

  if (quanteda::ndoc(dfm) < 2) {
    return(data.frame(
      Keyword = character(),
      Keyness_Score = numeric(),
      stringsAsFactors = FALSE
    ))
  }

  # keyness statistics are unstable for very low-frequency terms
  if (min_count > 0) {
    dfm <- quanteda::dfm_trim(dfm, min_termfreq = min_count)
  }

  keyness <- quanteda.textstats::textstat_keyness(
    dfm,
    target = target,
    measure = measure
  )

  score_col <- switch(measure, lr = "G2", chi2 = "chi2", exact = "exact", pmi = "pmi")
  keyness_top <- head(keyness[order(-abs(keyness[[score_col]])), ], min(top_n, nrow(keyness)))

  data.frame(
    Keyword = keyness_top$feature,
    Keyness_Score = keyness_top[[score_col]],
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}


#' Plot TF-IDF Keywords
#'
#' @description
#' Creates a horizontal bar plot of top keywords by TF-IDF score.
#'
#' @param tfidf_data Data frame from extract_keywords_tfidf()
#' @param title Plot title (default: "Top Keywords by TF-IDF Score")
#' @param normalized Logical, whether scores are normalized (for label) (default: FALSE)
#'
#' @return A plotly bar chart
#'
#' @concept lexical
#' @export
plot_tfidf_keywords <- function(tfidf_data,
                                 title = NULL,
                                 normalized = FALSE) {

  tfidf_data_sorted <- tfidf_data[order(tfidf_data$TF_IDF_Score, decreasing = FALSE), ]

  score_label <- if (normalized) "TF-IDF Score (Normalized)" else "TF-IDF Score"

  if (is.null(title)) {
    title <- paste("Top Keywords by", score_label)
  }

  tfidf_data_sorted$Keyword_ordered <- factor(tfidf_data_sorted$Keyword,
                                               levels = tfidf_data_sorted$Keyword)

  tfidf_data_sorted$hover_text <- paste0("Keyword: ", tfidf_data_sorted$Keyword,
                                          "\n", score_label, ": ",
                                          round(tfidf_data_sorted$TF_IDF_Score, 4),
                                          "\nFrequency: ", tfidf_data_sorted$Frequency)

  ggplot2::ggplot(tfidf_data_sorted, ggplot2::aes(x = TF_IDF_Score, y = Keyword_ordered,
                                                    text = hover_text)) +
    ggplot2::geom_col(fill = "#337ab7") +
    ggplot2::labs(x = score_label, y = "", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}


#' Plot Statistical Keyness
#'
#' @description
#' Creates a horizontal bar plot of distinctive keywords by keyness score.
#'
#' @param keyness_data Data frame from extract_keywords_keyness()
#' @param title Plot title (default: "Top Keywords by Keyness (G-squared)")
#' @param group_label Optional label for the target group (default: NULL)
#'
#' @return A plotly bar chart
#'
#' @concept lexical
#' @export
plot_keyness_keywords <- function(keyness_data,
                                  title = NULL,
                                  group_label = NULL) {

  if (nrow(keyness_data) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "Keyness analysis requires multiple documents",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  keyness_data_sorted <- keyness_data[order(abs(keyness_data$Keyness_Score), decreasing = FALSE), ]

  if (is.null(title)) {
    title <- if (!is.null(group_label)) {
      paste0("Top Keywords by Keyness (G\u00b2) - Grouped by ", group_label)
    } else {
      "Top Keywords by Keyness (G\u00b2)"
    }
  }

  keyness_data_sorted$Keyword_ordered <- factor(keyness_data_sorted$Keyword,
                                                 levels = keyness_data_sorted$Keyword)

  keyness_data_sorted$hover_text <- paste0("Keyword: ", keyness_data_sorted$Keyword,
                                            "\nKeyness Score (G\u00b2): ",
                                            round(keyness_data_sorted$Keyness_Score, 2))

  ggplot2::ggplot(keyness_data_sorted, ggplot2::aes(x = Keyness_Score, y = Keyword_ordered,
                                                     text = hover_text)) +
    ggplot2::geom_col(fill = "#337ab7") +
    ggplot2::labs(x = "Keyness Score (G\u00b2)", y = "", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}


#' Plot Keyword Comparison (TF-IDF vs Frequency)
#'
#' @description
#' Creates a grouped bar plot comparing TF-IDF scores with term frequencies.
#'
#' @param tfidf_data Data frame from extract_keywords_tfidf()
#' @param top_n Number of keywords to display (default: 10)
#' @param title Plot title (default: auto-generated)
#' @param normalized Logical, whether TF-IDF scores are normalized (default: FALSE)
#'
#' @return A plotly grouped bar chart
#'
#' @concept lexical
#' @export
plot_keyword_comparison <- function(tfidf_data,
                                    top_n = 10,
                                    title = NULL,
                                    normalized = FALSE) {

  top_keywords <- head(tfidf_data, top_n)

  score_label <- if (normalized) "TF-IDF Score (Normalized)" else "TF-IDF Score"

  if (is.null(title)) {
    title <- paste0("Top Keywords: ", score_label, " vs Frequency")
  }

  top_keywords$Freq_scaled <- top_keywords$Frequency / max(top_keywords$Frequency) *
    max(top_keywords$TF_IDF_Score)

  plot_long <- rbind(
    data.frame(Keyword = top_keywords$Keyword,
               Score = top_keywords$TF_IDF_Score,
               Metric = "TF-IDF",
               stringsAsFactors = FALSE),
    data.frame(Keyword = top_keywords$Keyword,
               Score = top_keywords$Freq_scaled,
               Metric = "Frequency",
               stringsAsFactors = FALSE)
  )
  plot_long$Keyword <- factor(plot_long$Keyword, levels = top_keywords$Keyword)
  plot_long$hover_text <- paste("Keyword:", plot_long$Keyword,
                                "<br>Metric:", plot_long$Metric,
                                "<br>Score:", round(plot_long$Score, 4))

  ggplot2::ggplot(plot_long, ggplot2::aes(x = Keyword, y = Score, fill = Metric, text = hover_text)) +
    ggplot2::geom_col(position = "dodge") +
    ggplot2::scale_fill_manual(values = c("TF-IDF" = "#337ab7", "Frequency" = "#5cb85c")) +
    ggplot2::labs(x = "Keywords", y = "Score", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text.x = ggplot2::element_text(size = 11, color = "#3B3B3B", angle = -45, hjust = 0),
      axis.text.y = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      legend.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      legend.text = ggplot2::element_text(size = 12, color = "#3B3B3B"),
      legend.position = "right"
    )
}



# Readability

# Readability Analysis Functions
#
# Functions for calculating and visualizing text readability metrics.

#' Plot Readability Distribution
#'
#' @description
#' Creates a boxplot showing the overall distribution of a readability metric.
#'
#' @param readability_data Data frame from calculate_text_readability()
#' @param metric Metric to plot (e.g., "flesch", "flesch_kincaid", "gunning_fog")
#' @param title Plot title (default: auto-generated)
#'
#' @return A plotly boxplot
#'
#' @concept lexical
#' @export
#'
#' @examples
#' \donttest{
#' data(SpecialEduTech, package = "TextAnalysisR")
#' texts <- SpecialEduTech$abstract[1:20]
#' readability <- calculate_text_readability(texts)
#' plot <- plot_readability_distribution(readability, "flesch")
#' print(plot)
#' }
plot_readability_distribution <- function(readability_data,
                                          metric,
                                          title = NULL) {

  if (!metric %in% names(readability_data)) {
    stop(paste("Metric", metric, "not found in readability data"))
  }

  if (is.null(title)) {
    title <- paste(metric, "- Overall Distribution")
  }

  metric_values <- readability_data[[metric]]
  metric_values <- metric_values[is.finite(metric_values)]

  if (length(metric_values) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "No valid data for selected metric",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  plot_df <- data.frame(
    value = metric_values,
    doc = paste("Doc", seq_along(metric_values)),
    x = ""
  )
  plot_df$hover_text <- paste0("Document: ", plot_df$doc,
                               "<br>", metric, ": ", round(plot_df$value, 2))

  ggplot2::ggplot(plot_df, ggplot2::aes(x = x, y = value)) +
    ggplot2::geom_boxplot(fill = "#4A90E2", alpha = 0.7,
                          color = "#0c1f4a", outlier.shape = NA) +
    suppressWarnings(ggplot2::geom_jitter(ggplot2::aes(text = hover_text),
                         width = 0.15, alpha = 0.5, color = "#0c1f4a", size = 1.8)) +
    ggplot2::labs(y = metric, x = "", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      axis.text.x = ggplot2::element_blank(),
      axis.ticks.x = ggplot2::element_blank()
    )

}


#' Plot Readability by Group
#'
#' @description
#' Creates grouped boxplots comparing readability across categories.
#'
#' @param readability_data Data frame from calculate_text_readability()
#' @param metric Metric to plot
#' @param group_var Name of grouping variable column
#' @param title Plot title (default: auto-generated)
#'
#' @return A plotly boxplot
#'
#' @concept lexical
#' @export
plot_readability_by_group <- function(readability_data,
                                      metric,
                                      group_var,
                                      title = NULL) {

  if (!metric %in% names(readability_data)) {
    stop(paste("Metric", metric, "not found in readability data"))
  }

  if (!group_var %in% names(readability_data)) {
    stop(paste("Group variable", group_var, "not found in data"))
  }

  if (is.null(title)) {
    title <- paste(metric, "by", group_var)
  }

  plot_data <- readability_data[, c(metric, group_var)]
  names(plot_data) <- c("metric_value", "group")

  plot_data$metric_value <- round(plot_data$metric_value, 2)
  plot_data <- plot_data[is.finite(plot_data$metric_value), ]

  colors <- c("#4A90E2", "#E74C3C", "#2ECC71", "#F39C12", "#9B59B6",
              "#1ABC9C", "#E67E22", "#3498DB")

  unique_groups <- unique(plot_data$group)

  plot_data$hover_text <- paste0(group_var, ": ", plot_data$group,
                                 "<br>", metric, ": ", plot_data$metric_value)

  ggplot2::ggplot(plot_data, ggplot2::aes(x = group, y = metric_value, fill = group)) +
    ggplot2::geom_boxplot(alpha = 0.7, outlier.shape = NA) +
    suppressWarnings(ggplot2::geom_jitter(ggplot2::aes(text = hover_text),
                         width = 0.15, alpha = 0.5, color = "#0c1f4a", size = 1.8)) +
    ggplot2::scale_fill_manual(values = colors[seq_along(unique_groups)]) +
    ggplot2::labs(x = group_var, y = metric, title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      legend.position = "none"
    )
}


#' Plot Top Documents by Readability
#'
#' @description
#' Creates a bar plot of documents ranked by readability metric.
#'
#' @param readability_data Data frame from calculate_text_readability()
#' @param metric Metric to plot
#' @param top_n Number of documents to show (default: 15)
#' @param order Direction: "highest" or "lowest" (default: "highest")
#' @param title Plot title (default: auto-generated)
#'
#' @return A plotly bar chart
#'
#' @concept lexical
#' @export
plot_top_readability_documents <- function(readability_data,
                                           metric,
                                           top_n = 15,
                                           order = "highest",
                                           title = NULL) {

  if (!metric %in% names(readability_data)) {
    stop(paste("Metric", metric, "not found in readability data"))
  }

  doc_col <- if ("Document" %in% names(readability_data)) "Document" else "document"

  sorted_data <- readability_data %>%
    dplyr::arrange(if (order == "highest") dplyr::desc(.data[[metric]]) else .data[[metric]]) %>%
    head(top_n)

  if (is.null(title)) {
    title <- paste("Top", top_n, "Documents by", metric)
  }

  sorted_data$doc_label <- sorted_data[[doc_col]]
  sorted_data$doc_ordered <- factor(sorted_data$doc_label,
                                     levels = rev(sorted_data$doc_label))
  sorted_data$metric_val <- sorted_data[[metric]]
  sorted_data$hover_text <- paste0(sorted_data$doc_label, "\n",
                                    metric, ": ", round(sorted_data$metric_val, 2))

  ggplot2::ggplot(sorted_data, ggplot2::aes(x = doc_ordered, y = metric_val,
                                             text = hover_text)) +
    ggplot2::geom_col(fill = "#4A90E2") +
    ggplot2::coord_flip() +
    ggplot2::labs(x = "Document", y = metric, title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}

#' Calculate Text Readability
#'
#' @description
#' Calculates multiple readability metrics for texts including Flesch Reading Ease,
#' Flesch-Kincaid Grade Level, Gunning FOG index, and others. Optionally includes
#' lexical diversity metrics and sentence statistics.
#'
#' @param texts Character vector of texts to analyze
#' @param metrics Character vector of readability metrics to calculate.
#'   Options: "flesch", "flesch_kincaid", "gunning_fog", "smog", "ari", "coleman_liau"
#' @param include_lexical_diversity Logical, include the MTLD lexical diversity
#'   index (default: TRUE)
#' @param include_sentence_stats Logical, include average sentence length (default: TRUE)
#' @param doc_names Optional character vector of document names
#'
#' @return A data frame with document names and readability scores
#'
#' @concept lexical
#' @export
#'
#' @examples
#' \donttest{
#' data(SpecialEduTech, package = "TextAnalysisR")
#' texts <- SpecialEduTech$abstract[1:10]
#' readability <- calculate_text_readability(texts)
#' print(readability)
#' }
#'
#' @importFrom quanteda corpus tokens dfm docnames
#' @importFrom quanteda.textstats textstat_readability textstat_lexdiv
calculate_text_readability <- function(texts,
                                      metrics = c("flesch", "flesch_kincaid", "gunning_fog"),
                                      include_lexical_diversity = TRUE,
                                      include_sentence_stats = TRUE,
                                      doc_names = NULL) {

  if (!requireNamespace("quanteda.textstats", quietly = TRUE)) {
    stop("Package 'quanteda.textstats' is required. Please install it.")
  }

  if (is.null(doc_names)) {
    doc_names <- paste0("Doc ", seq_along(texts))
  }

  corp <- quanteda::corpus(texts)
  quanteda::docnames(corp) <- doc_names

  measure_map <- c(
    "flesch" = "Flesch",
    "flesch_kincaid" = "Flesch.Kincaid",
    "gunning_fog" = "FOG",
    "smog" = "SMOG",
    "ari" = "ARI",
    "coleman_liau" = "Coleman.Liau.short"
  )

  valid_metrics <- intersect(metrics, names(measure_map))
  if (length(valid_metrics) == 0) {
    stop("No valid metrics specified. Available metrics: ",
         paste(names(measure_map), collapse = ", "))
  }

  mapped_metrics <- measure_map[valid_metrics]
  names(mapped_metrics) <- valid_metrics

  # Batch all readability metrics in single call for performance
  all_scores <- list()
  tryCatch({
    # Call textstat_readability once with all measures
    batch_scores <- quanteda.textstats::textstat_readability(corp, measure = unname(mapped_metrics))

    # Extract scores for each metric
    for (i in seq_along(valid_metrics)) {
      metric <- valid_metrics[i]
      measure_name <- mapped_metrics[i]
      if (measure_name %in% names(batch_scores)) {
        all_scores[[metric]] <- batch_scores[[measure_name]]
      } else {
        all_scores[[metric]] <- rep(NA, length(texts))
      }
    }
  }, error = function(e) {
    # Fallback to individual calls if batch fails
    warning("Batch readability calculation failed, falling back to individual metrics: ", e$message)
    for (i in seq_along(valid_metrics)) {
      metric <- valid_metrics[i]
      measure_name <- mapped_metrics[i]
      all_scores[[metric]] <- tryCatch({
        quanteda.textstats::textstat_readability(corp, measure = measure_name)[[2]]
      }, error = function(e2) {
        warning(paste("Could not calculate", metric, ":", e2$message))
        rep(NA, length(texts))
      })
    }
  })

  readability_scores <- data.frame(Document = doc_names, stringsAsFactors = FALSE)
  for (metric in names(all_scores)) {
    readability_scores[[metric]] <- all_scores[[metric]]
  }

  if (include_lexical_diversity) {
    # MTLD needs token order; compute per document, not from a DFM
    toks <- quanteda::tokens(texts, remove_punct = TRUE)
    mtld_values <- vapply(seq_len(quanteda::ndoc(toks)), function(i) {
      .calc_mtld(as.character(toks[[i]]))
    }, numeric(1))
    readability_scores$`Lexical Diversity (MTLD)` <- as.numeric(mtld_values)
  }

  if (include_sentence_stats) {
    avg_sentence_length <- vapply(texts, function(t) {
      sents <- quanteda::tokens(t, what = "sentence")[[1]]
      sents <- sents[nzchar(trimws(sents))]
      if (length(sents) == 0) return(NA_real_)
      words <- lengths(quanteda::tokens(sents, what = "word", remove_punct = TRUE))
      sum(words) / length(sents)
    }, numeric(1))
    readability_scores$`Avg Sentence Length` <- avg_sentence_length
  }

  return(readability_scores)
}




#' Plot Term Frequency Trends by Continuous Variable
#'
#' @description
#' Creates a faceted line plot showing how term frequencies vary across
#' a continuous variable (e.g., year, time period).
#'
#' @param term_data Data frame containing term frequencies with columns:
#'   continuous_var, term, and word_frequency
#' @param continuous_var Name of the continuous variable column
#' @param terms Character vector of terms to display (optional, filters if provided)
#' @param title Plot title (default: NULL, auto-generated)
#' @param height Plot height in pixels (default: 600)
#' @param width Plot width in pixels (default: NULL, auto)
#'
#' @return A plotly object with faceted line plots
#'
#' @concept visualization
#' @export
#'
#' @examples
#' \donttest{
#' term_df <- data.frame(
#'   year = rep(2010:2020, each = 3),
#'   term = rep(c("learning", "education", "technology"), 11),
#'   word_frequency = sample(10:100, 33, replace = TRUE)
#' )
#' plot_term_trends_continuous(term_df, "year", c("learning", "education"))
#' }
plot_term_trends_continuous <- function(term_data,
                                         continuous_var,
                                         terms = NULL,
                                         title = NULL,
                                         height = 600,
                                         width = NULL) {

  if (!requireNamespace("scales", quietly = TRUE)) {
    stop("Package 'scales' is required. Please install it.")
  }

  if (!continuous_var %in% names(term_data)) {
    stop("Continuous variable '", continuous_var, "' not found in data")
  }

  if (!"term" %in% names(term_data) && !"word" %in% names(term_data)) {
    stop("term or word column not found in data")
  }

  if ("word" %in% names(term_data) && !"term" %in% names(term_data)) {
    term_data$term <- term_data$word
  }

  if (!"word_frequency" %in% names(term_data) && !"count" %in% names(term_data)) {
    stop("word_frequency or count column not found in data")
  }

  if ("count" %in% names(term_data) && !"word_frequency" %in% names(term_data)) {
    term_data$word_frequency <- term_data$count
  }

  if (!is.null(terms)) {
    term_data <- term_data %>%
      dplyr::filter(term %in% terms) %>%
      dplyr::mutate(term = factor(term, levels = terms))
  }

  if (is.null(title)) {
    title <- paste("Term Frequency by", continuous_var)
  }

  term_data$hover_text <- paste("Term:", term_data$term,
                               paste0("<br>", continuous_var, ":"), term_data[[continuous_var]],
                               "<br>Frequency:", term_data$word_frequency)

  ggplot2::ggplot(
    term_data,
    ggplot2::aes(
      x = .data[[continuous_var]],
      y = word_frequency,
      group = term,
      text = hover_text
    )
  ) +
    ggplot2::geom_point(color = "#337ab7", alpha = 0.6, size = 2.5) +
    ggplot2::geom_line(color = "#337ab7", alpha = 0.6, linewidth = 0.5) +
    ggplot2::facet_wrap(~term, scales = "free") +
    ggplot2::scale_y_continuous(labels = scales::number_format(accuracy = 1)) +
    ggplot2::labs(y = "Word Frequency", x = continuous_var, title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      legend.position = "none",
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.line = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      axis.ticks = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      strip.text.x = ggplot2::element_text(size = 11, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      axis.title.x = ggplot2::element_text(margin = ggplot2::margin(t = 15)),
      axis.title.y = ggplot2::element_text(margin = ggplot2::margin(r = 15)),
      plot.margin = ggplot2::margin(t = 5, r = 10, b = 25, l = 15)
    )
}



#' Plot Part-of-Speech Tag Frequencies
#'
#' @description
#' Creates a bar plot showing the frequency distribution of part-of-speech tags.
#'
#' @param pos_data Data frame containing POS data with columns:
#'   \itemize{
#'     \item \code{pos}: Part-of-speech tag
#'     \item \code{n}: (optional) Pre-computed frequency count
#'   }
#'   If \code{n} is not present, frequencies will be computed from the data.
#' @param top_n Number of top POS tags to display (default: 20)
#' @param title Plot title (default: "Part-of-Speech Tag Frequency")
#' @param color Bar color (default: "#337ab7")
#' @param height Plot height in pixels (default: 500)
#' @param width Plot width in pixels (default: NULL for auto)
#'
#' @return A plotly object
#'
#' @concept visualization
#' @export
#'
#' @examples
#' if (interactive()) {
#'   pos_df <- data.frame(
#'     pos = c("NOUN", "VERB", "ADJ", "ADV", "PRON"),
#'     n = c(500, 400, 250, 150, 100)
#'   )
#'   plot_pos_frequencies(pos_df)
#' }
plot_pos_frequencies <- function(pos_data,
                                  top_n = 20,
                                  title = "Part-of-Speech Tag Frequency",
                                  color = "#337ab7",
                                  height = 500,
                                  width = NULL) {

  if (is.null(pos_data) || nrow(pos_data) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "No POS data available",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  if (!"n" %in% names(pos_data)) {
    pos_freq <- pos_data %>%
      dplyr::count(pos, sort = TRUE) %>%
      dplyr::slice_head(n = top_n)
  } else {
    pos_freq <- pos_data %>%
      dplyr::arrange(dplyr::desc(n)) %>%
      dplyr::slice_head(n = top_n)
  }

  pos_freq$pos_ordered <- stats::reorder(pos_freq$pos, pos_freq$n)

  ggplot2::ggplot(pos_freq, ggplot2::aes(x = pos_ordered, y = n,
                                          text = paste0(pos, "\nFrequency: ", n))) +
    ggplot2::geom_col(fill = color) +
    ggplot2::labs(x = "POS Tag", y = "Frequency", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}


#' Plot Named Entity Frequencies
#'
#' @description
#' Creates a bar plot showing the frequency distribution of named entity types.
#'
#' @param entity_data Data frame containing entity data with columns:
#'   \itemize{
#'     \item \code{entity}: Named entity type (e.g., "PERSON", "ORG", "GPE")
#'     \item \code{n}: (optional) Pre-computed frequency count
#'   }
#'   If \code{n} is not present, frequencies will be computed from the data.
#' @param top_n Number of top entity types to display (default: 20)
#' @param title Plot title (default: "Named Entity Type Frequency")
#' @param color Bar color (default: "#10B981")
#' @param height Plot height in pixels (default: 500)
#' @param width Plot width in pixels (default: NULL for auto)
#' @param custom_colors Named vector of custom entity type colors (e.g.,
#'   c(CONCEPT = "#00acc1", THEME = "#7c4dff")). Custom colors override defaults.
#'
#' @return A plotly object
#'
#' @concept visualization
#' @export
#'
#' @examples
#' if (interactive()) {
#'   entity_df <- data.frame(
#'     entity = c("PERSON", "ORG", "GPE", "DATE", "MONEY"),
#'     n = c(300, 250, 200, 150, 100)
#'   )
#'   plot_entity_frequencies(entity_df)
#'
#'   # With custom colors
#'   plot_entity_frequencies(entity_df, custom_colors = c(PERSON = "#ff0000"))
#' }
plot_entity_frequencies <- function(entity_data,
                                     top_n = 20,
                                     title = "Named Entity Type Frequency",
                                     color = NULL,
                                     height = 500,
                                     width = NULL,
                                     custom_colors = NULL) {

  if (!requireNamespace("plotly", quietly = TRUE)) {
    stop("Package 'plotly' is required. Please install it.")
  }

  if (is.null(entity_data) || nrow(entity_data) == 0) {
    return(plotly::plot_ly(type = "scatter", mode = "markers") %>%
      plotly::layout(
        xaxis = list(showgrid = FALSE, zeroline = FALSE, showticklabels = FALSE),
        yaxis = list(showgrid = FALSE, zeroline = FALSE, showticklabels = FALSE),
        annotations = list(
          list(
            text = "No named entities found",
            x = 0.5, y = 0.5,
            xref = "paper", yref = "paper",
            showarrow = FALSE,
            font = list(size = 16, color = "#6B7280", family = "Roboto")
          )
        )
      ))
  }

  if (!"n" %in% names(entity_data)) {
    entity_freq <- entity_data %>%
      dplyr::count(entity, sort = TRUE) %>%
      dplyr::slice_head(n = top_n)
  } else {
    entity_freq <- entity_data %>%
      dplyr::arrange(dplyr::desc(n)) %>%
      dplyr::slice_head(n = top_n)
  }

  entity_colors <- c(
    "PERSON" = "#e91e63", "ORG" = "#2196f3", "GPE" = "#4caf50",
    "DATE" = "#ff9800", "MONEY" = "#9c27b0", "CARDINAL" = "#607d8b",
    "ORDINAL" = "#795548", "PERCENT" = "#00bcd4", "PRODUCT" = "#3f51b5",
    "EVENT" = "#f44336", "WORK_OF_ART" = "#673ab7", "LAW" = "#009688",
    "LANGUAGE" = "#8bc34a", "LOC" = "#03a9f4", "FAC" = "#cddc39",
    "NORP" = "#ffc107", "TIME" = "#ff5722", "QUANTITY" = "#9e9e9e",
    "DISABILITY" = "#E91E63", "PROGRAM" = "#2196F3", "TEST" = "#4CAF50",
    "CONCEPT" = "#00acc1", "TOOL" = "#FF9800", "METHOD" = "#00BCD4",
    "THEME" = "#7c4dff", "CODE" = "#546e7a", "CATEGORY" = "#26a69a",
    "CUSTOM" = "#d81b60"
  )

  if (!is.null(custom_colors) && length(custom_colors) > 0) {
    entity_colors[names(custom_colors)] <- custom_colors
  }

  bar_colors <- vapply(entity_freq$entity, function(e) {
    if (e %in% names(entity_colors)) entity_colors[[e]] else "#757575"
  }, character(1))

  plotly::plot_ly(
    data = entity_freq,
    x = ~stats::reorder(entity, n),
    y = ~n,
    type = "bar",
    marker = list(color = bar_colors),
    hoverinfo = "text",
    hovertext = ~paste0(entity, "\nFrequency: ", n),
    height = height,
    width = width
  ) %>%
    plotly::layout(
      title = list(
        text = title,
        font = list(size = 18, color = "#0c1f4a", family = "Roboto, sans-serif")
      ),
      xaxis = list(
        title = "",
        tickangle = -45,
        tickfont = list(size = 16, color = "#3B3B3B", family = "Roboto, sans-serif")
      ),
      yaxis = list(
        title = "Frequency",
        titlefont = list(size = 16, color = "#0c1f4a", family = "Roboto, sans-serif"),
        tickfont = list(size = 16, color = "#3B3B3B", family = "Roboto, sans-serif")
      ),
      margin = list(b = 150, l = 60, r = 20, t = 60),
      hoverlabel = list(
        align = "left",
        font = list(size = 16, color = "white", family = "Roboto, sans-serif"),
        bgcolor = "#0c1f4a"
      )
    )
}


#' Render displaCy Entity Visualization
#'
#' @description
#' Renders spaCy's displaCy entity visualization as HTML.
#' Highlights named entities with colored labels.
#'
#' @param text Character string to visualize.
#' @param model spaCy model name (default: "en_core_web_sm").
#' @param colors Named list of entity type to color mappings (e.g.,
#'   list(PERSON = "#e91e63", ORG = "#2196f3")). If NULL, uses spaCy defaults.
#'
#' @return HTML string with entity highlighting.
#'
#' @concept lexical
#' @export
render_displacy_ent <- function(text, model = "en_core_web_sm", colors = NULL) {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("displaCy entity visualization"))
  }

  tryCatch({
    spacy <- reticulate::import("spacy")
    displacy_module <- reticulate::import("spacy.displacy")

    nlp <- spacy$load(model)
    doc <- nlp(text)

    if (length(doc$ents) == 0) {
      return("<div>No named entities detected in this text.</div>")
    }

    options <- list()
    if (!is.null(colors) && length(colors) > 0) {
      options$colors <- colors
    }

    if (length(options) > 0) {
      html <- displacy_module$render(doc, style = "ent", page = FALSE, options = options)
    } else {
      html <- displacy_module$render(doc, style = "ent", page = FALSE)
    }

    return(as.character(html))
  }, error = function(e) {
    stop("displaCy rendering failed: ", e$message)
  })
}


#' Render displaCy Dependency Visualization
#'
#' @description
#' Renders spaCy's displaCy dependency visualization as SVG.
#' Shows syntactic structure with arrows between words.
#'
#' @param text Character string to visualize.
#' @param compact Logical; use compact mode for space (default: TRUE).
#' @param model spaCy model name (default: "en_core_web_sm").
#'
#' @return SVG string with dependency tree.
#'
#' @concept lexical
#' @export
render_displacy_dep <- function(text, compact = TRUE, model = "en_core_web_sm") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("displaCy dependency visualization"))
  }

  tryCatch({
    spacy <- reticulate::import("spacy")
    displacy_module <- reticulate::import("spacy.displacy")

    nlp <- spacy$load(model)
    doc <- nlp(text)

    options <- list(compact = compact)

    svg <- displacy_module$render(doc, style = "dep", options = options, page = FALSE)

    return(as.character(svg))
  }, error = function(e) {
    stop("displaCy rendering failed: ", e$message)
  })
}


# =============================================================================
# spaCy NLP Interface Functions
# =============================================================================
# R wrapper functions for spaCy NLP via reticulate.
# Provides direct Python spaCy access for full control
# over all spaCy features including morphology.

# Package-level spaCy instance
.spacy_env <- new.env(parent = emptyenv())

.spacy_prepare_texts <- function(x) {
  if (inherits(x, "tokens")) {
    texts <- vapply(as.list(x), function(toks) paste(toks, collapse = " "), character(1))
    doc_names <- quanteda::docnames(x)
  } else if (is.character(x)) {
    texts <- x
    doc_names <- names(x) %||% paste0("text", seq_along(texts))
  } else {
    stop("x must be a character vector or quanteda tokens object")
  }
  list(texts_list = as.list(unname(texts)), doc_names = doc_names)
}

.map_spacy_doc_ids <- function(df, doc_names) {
  if (nrow(df) > 0 && "doc_id" %in% names(df) && length(doc_names) > 0) {
    df$doc_id <- as.character(df$doc_id)
    doc_id_map <- stats::setNames(doc_names, paste0("text", seq_along(doc_names)))
    matched <- doc_id_map[df$doc_id]
    df$doc_id <- ifelse(is.na(matched), df$doc_id, matched)
  }
  df
}

#' Initialize spaCy NLP
#'
#' @description
#' Initialize the spaCy NLP pipeline with the specified model.
#' Uses a cached instance for efficiency.
#'
#' @param model Character; spaCy model name (default: "en_core_web_sm").
#' @param force Logical; force reinitialization even if already initialized.
#'
#' @return Invisibly returns the SpacyNLP Python object.
#'
#' @details
#' Available models:
#' \itemize{
#'   \item \code{en_core_web_sm}: Small English model (fast, no word vectors)
#'   \item \code{en_core_web_md}: Medium English model (word vectors)
#'   \item \code{en_core_web_lg}: Large English model (best accuracy)
#' }
#'
#' @concept lexical
#' @keywords internal
#' @export
#'
#' @examples
#' if (interactive()) {
#' init_spacy_nlp("en_core_web_sm")
#' }
init_spacy_nlp <- function(model = "en_core_web_sm", force = FALSE) {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("spaCy initialization"))
  }

  # Check if already initialized with same model
  if (!force && !is.null(.spacy_env$nlp) && isTRUE(.spacy_env$model == model)) {
    return(invisible(.spacy_env$nlp))
  }

  # Import the Python module
  python_path <- system.file("python", package = "TextAnalysisR")
  if (python_path == "") {
    stop("Cannot find Python module directory in TextAnalysisR package")
  }

  tryCatch({
    spacy_module <- reticulate::import_from_path("spacy_nlp", path = python_path)
    nlp <- spacy_module$SpacyNLP(model)
    .spacy_env$nlp <- nlp
    .spacy_env$model <- model
    .spacy_env$module <- spacy_module
    message("spaCy initialized with model: ", model)
    invisible(nlp)
  }, error = function(e) {
    stop("Failed to initialize spaCy: ", e$message,
         "\nMake sure spaCy is installed: pip install spacy",
         "\nAnd download the model: python -m spacy download ", model)
  })
}

#' Check if spaCy is Initialized
#'
#' @description
#' Check whether spaCy has been initialized.
#'
#' @return Logical; TRUE if initialized, FALSE otherwise.
#'
#' @concept lexical
#' @export
spacy_initialized <- function() {
  !is.null(.spacy_env$nlp)
}

#' Check if Model Has Word Vectors
#'
#' @description
#' Check if the loaded spaCy model has word vectors for similarity calculations.
#'
#' @return Logical; TRUE if model has vectors, FALSE otherwise.
#'
#' @concept lexical
#' @export
spacy_has_vectors <- function() {
  if (!spacy_initialized()) {
    stop("spaCy not initialized. Call init_spacy_nlp() first.")
  }
  .spacy_env$nlp$has_vectors()
}

#' Get spaCy Model Information
#'
#' @description
#' Get information about the currently loaded spaCy model.
#'
#' @return A list with model information including name, language,
#'   pipeline components, and vector availability.
#'
#' @concept lexical
#' @export
get_spacy_model_info <- function() {
  if (!spacy_initialized()) {
    stop("spaCy not initialized. Call init_spacy_nlp() first.")
  }
  info <- .spacy_env$nlp$get_model_info()
  as.list(info)
}

#' Parse Texts with spaCy
#'
#' @description
#' Parse texts using spaCy and return token-level annotations.
#' This is the main parsing function for NLP analysis.
#' Works with character vectors or quanteda tokens objects.
#'
#' @param x Character vector of texts OR a quanteda tokens object.
#' @param pos Logical; include coarse POS tags (default: TRUE).
#' @param tag Logical; include fine-grained tags (default: TRUE).
#' @param lemma Logical; include lemmatized forms (default: TRUE).
#' @param entity Logical; include named entity tags (default: FALSE).
#' @param dependency Logical; include dependency relations (default: FALSE).
#' @param morph Logical; include morphological features (default: FALSE).
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#'
#' @return A data frame with token-level annotations including:
#' \itemize{
#'   \item \code{doc_id}: Document identifier
#'   \item \code{sentence_id}: Sentence number within document
#'   \item \code{token_id}: Token position within sentence
#'   \item \code{token}: Original token text
#'   \item \code{pos}: Coarse POS tag (if pos = TRUE)
#'   \item \code{tag}: Fine-grained tag (if tag = TRUE)
#'   \item \code{lemma}: Lemmatized form (if lemma = TRUE)
#'   \item \code{entity}: Named entity tag (if entity = TRUE)
#'   \item \code{head_token_id}: Head token ID (if dependency = TRUE)
#'   \item \code{dep_rel}: Dependency relation (if dependency = TRUE)
#'   \item \code{morph}: Morphological features string (if morph = TRUE)
#' }
#'
#' @concept lexical
#' @export
#'
#' @examples
#' if (interactive()) {
#' # From SpecialEduTech dataset
#' texts <- TextAnalysisR::SpecialEduTech$abstract[1:5]
#' parsed <- spacy_parse_full(texts, morph = TRUE)
#'
#' # From quanteda tokens
#' united <- unite_cols(TextAnalysisR::SpecialEduTech, c("title", "abstract"))
#' tokens <- prep_texts(united, text_field = "united_texts")
#' parsed <- spacy_parse_full(tokens, morph = TRUE)
#' }
spacy_parse_full <- function(x,
                             pos = TRUE,
                             tag = TRUE,
                             lemma = TRUE,
                             entity = FALSE,
                             dependency = FALSE,
                             morph = FALSE,
                             model = "en_core_web_sm") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("spaCy-backed features"))
  }


  if (!spacy_initialized() || !isTRUE(.spacy_env$model == model)) {
    init_spacy_nlp(model)
  }

  prepared <- .spacy_prepare_texts(x)

  result <- .spacy_env$nlp$parse_to_dataframe(
    prepared$texts_list,
    include_pos = pos,
    include_tag = tag,
    include_lemma = lemma,
    include_entity = entity,
    include_dependency = dependency,
    include_morph = morph
  )

  df <- as.data.frame(reticulate::py_to_r(result))
  .map_spacy_doc_ids(df, prepared$doc_names)
}

#' Lemmatize Texts with spaCy
#'
#' @description
#' Perform lemmatization using spaCy with optimized pipeline settings.
#' Disables unnecessary components (NER, parser) for faster processing.
#'
#' @param x Character vector of texts OR a quanteda tokens object.
#' @param batch_size Integer; batch size for processing (default: 100).
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#'
#' @return A data frame with columns: doc_id, token_id, token, lemma.
#'
#' @details
#' This function disables NER, entity_ruler, and parser components to speed up
#' lemmatization. Use this for lemmas without other annotations.
#'
#' @concept lexical
#' @export
spacy_lemmatize <- function(x, batch_size = 100, model = "en_core_web_sm") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("spaCy-backed features"))
  }

  if (!spacy_initialized() || !isTRUE(.spacy_env$model == model)) {
    init_spacy_nlp(model)
  }

  prepared <- .spacy_prepare_texts(x)

  result <- .spacy_env$nlp$lemmatize(
    prepared$texts_list,
    batch_size = as.integer(batch_size)
  )

  df <- as.data.frame(reticulate::py_to_r(result))
  .map_spacy_doc_ids(df, prepared$doc_names)
}

#' Extract Named Entities with spaCy
#'
#' @description
#' Extract named entities from texts using spaCy NER.
#'
#' @param x Character vector of texts OR a quanteda tokens object.
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#'
#' @return A data frame with entity information:
#' \itemize{
#'   \item \code{doc_id}: Document identifier
#'   \item \code{text}: Entity text
#'   \item \code{label}: Entity type (PERSON, ORG, GPE, etc.)
#'   \item \code{start_char}: Start character position
#'   \item \code{end_char}: End character position
#' }
#'
#' @concept lexical
#' @export
spacy_extract_entities <- function(x, model = "en_core_web_sm") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("spaCy-backed features"))
  }

  if (!spacy_initialized() || !isTRUE(.spacy_env$model == model)) {
    init_spacy_nlp(model)
  }

  prepared <- .spacy_prepare_texts(x)
  result <- .spacy_env$nlp$get_entities(prepared$texts_list)
  df <- as.data.frame(reticulate::py_to_r(result))
  .map_spacy_doc_ids(df, prepared$doc_names)
}

#' Extract Noun Chunks
#'
#' @description
#' Extract noun chunks (base noun phrases) from texts.
#' Useful for keyphrase extraction.
#'
#' @param x Character vector of texts OR a quanteda tokens object.
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#'
#' @return A data frame with noun chunk information.
#'
#' @concept lexical
#' @export
extract_noun_chunks <- function(x, model = "en_core_web_sm") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("spaCy-backed features"))
  }

  if (!spacy_initialized() || !isTRUE(.spacy_env$model == model)) {
    init_spacy_nlp(model)
  }

  prepared <- .spacy_prepare_texts(x)
  result <- .spacy_env$nlp$get_noun_chunks(prepared$texts_list)
  df <- as.data.frame(reticulate::py_to_r(result))
  .map_spacy_doc_ids(df, prepared$doc_names)
}

#' Extract Subjects and Objects
#'
#' @description
#' Extract subject-verb-object (SVO) triples from texts using dependency parsing.
#'
#' @param x Character vector of texts OR a quanteda tokens object.
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#'
#' @return A data frame with SVO information.
#'
#' @concept lexical
#' @export
extract_subjects_objects <- function(x, model = "en_core_web_sm") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("spaCy-backed features"))
  }

  if (!spacy_initialized() || !isTRUE(.spacy_env$model == model)) {
    init_spacy_nlp(model)
  }

  prepared <- .spacy_prepare_texts(x)
  result <- .spacy_env$nlp$get_subjects_objects(prepared$texts_list)
  df <- as.data.frame(reticulate::py_to_r(result))
  .map_spacy_doc_ids(df, prepared$doc_names)
}

#' Get Sentences
#'
#' @description
#' Segment texts into sentences using spaCy's sentence boundary detection.
#'
#' @param x Character vector of texts OR a quanteda tokens object.
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#'
#' @return A data frame with sentence information.
#'
#' @concept lexical
#' @export
get_sentences <- function(x, model = "en_core_web_sm") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("spaCy-backed features"))
  }

  if (!spacy_initialized() || !isTRUE(.spacy_env$model == model)) {
    init_spacy_nlp(model)
  }

  prepared <- .spacy_prepare_texts(x)
  result <- .spacy_env$nlp$get_sentences(prepared$texts_list)
  df <- as.data.frame(reticulate::py_to_r(result))
  .map_spacy_doc_ids(df, prepared$doc_names)
}

#' Calculate Word Similarity
#'
#' @description
#' Calculate semantic similarity between two words using word vectors.
#' Requires a spaCy model with word vectors (en_core_web_md or en_core_web_lg).
#'
#' @param word1 Character; first word.
#' @param word2 Character; second word.
#' @param model Character; spaCy model to use (default: "en_core_web_md").
#'
#' @return A list with similarity score and metadata.
#'
#' @concept lexical
#' @export
get_word_similarity <- function(word1, word2, model = "en_core_web_md") {
  if (!spacy_initialized() || !isTRUE(.spacy_env$model == model)) {
    init_spacy_nlp(model)
  }

  result <- .spacy_env$nlp$get_word_similarity(word1, word2)
  as.list(result)
}

#' Find Similar Words
#'
#' @description
#' Find words most similar to a given word using word vectors.
#' Requires a spaCy model with word vectors (en_core_web_md or en_core_web_lg).
#'
#' @param word Character; target word.
#' @param top_n Integer; number of similar words to return (default: 10).
#' @param model Character; spaCy model to use (default: "en_core_web_md").
#'
#' @return A data frame with similar words and similarity scores.
#'
#' @concept lexical
#' @export
find_similar_words <- function(word, top_n = 10L, model = "en_core_web_md") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("spaCy-backed features"))
  }

  if (!spacy_initialized() || !isTRUE(.spacy_env$model == model)) {
    init_spacy_nlp(model)
  }

  result <- .spacy_env$nlp$find_similar_words(word, as.integer(top_n))
  df <- reticulate::py_to_r(result)

  return(df)
}

#' Get spaCy Word Embeddings
#'
#' @description
#' Get word vector embeddings for words or texts using spaCy.
#' Requires a spaCy model with word vectors.
#'
#' @param texts Character vector of words or texts.
#' @param model Character; spaCy model to use (default: "en_core_web_md").
#'
#' @return A matrix of word embeddings (rows = texts, cols = dimensions).
#'
#' @concept lexical
#' @keywords internal
get_spacy_embeddings <- function(texts, model = "en_core_web_md") {
  if (!spacy_initialized() || !isTRUE(.spacy_env$model == model)) {
    init_spacy_nlp(model)
  }

  if (!spacy_has_vectors()) {
    stop("Model '", model, "' has no word vectors. Use en_core_web_md or en_core_web_lg.")
  }

  # Get vectors via Python - access the underlying nlp object
  nlp <- .spacy_env$nlp$nlp

  vectors <- lapply(texts, function(text) {
    doc <- nlp(text)
    as.numeric(doc$vector)
  })

  # Convert to matrix
  mat <- do.call(rbind, vectors)
  rownames(mat) <- texts

  return(mat)
}

#' Parse Morphology String
#'
#' @description
#' Parse spaCy's morphology string format into individual columns.
#' Used internally by morphology analysis functions.
#' Always extracts all common morphology features (Number, Tense, VerbForm,
#' Person, Case, Mood, Aspect) regardless of the features parameter.
#'
#' @param data Data frame with a 'morph' column from spaCy parsing.
#' @param features Character vector of feature names (ignored, kept for
#'   backwards compatibility). All features are always extracted.
#'
#' @return Data frame with additional morph_* columns for each feature.
#'
#' @concept lexical
#' @keywords internal
parse_morphology_string <- function(data, features = NULL) {
  # Always extract all common morphology features
  all_features <- c("Number", "Tense", "VerbForm", "Person", "Case", "Mood", "Aspect")

  for (feat in all_features) {
    col_name <- paste0("morph_", feat)
    data[[col_name]] <- vapply(data$morph, function(m) {
      if (is.null(m) || length(m) == 0 || !is.atomic(m)) return(NA_character_)
      m_str <- as.character(m)[1]
      if (is.na(m_str) || m_str == "") return(NA_character_)
      parts <- strsplit(m_str, "\\|")[[1]]
      for (part in parts) {
        kv <- strsplit(part, "=")[[1]]
        if (length(kv) == 2 && kv[1] == feat) {
          return(kv[2])
        }
      }
      return(NA_character_)
    }, FUN.VALUE = character(1), USE.NAMES = FALSE)
  }
  return(data)
}


# Log odds ratio analysis

#' Calculate Log Odds Ratio Between Categories
#'
#' @description
#' Compares word usage between categories using the log-odds ratio with a
#' uniform Dirichlet prior, ranked by z-score. Identifies words distinctively
#' used in one category. For the informative-prior weighted log-odds, see
#' [calculate_weighted_log_odds()].
#'
#' @param dfm_object A quanteda dfm object
#' @param group_var Character, name of the grouping variable in docvars
#' @param comparison_mode Character, one of "binary", "one_vs_rest", or "pairwise"
#'   \itemize{
#'     \item binary: Compare two categories directly
#'     \item one_vs_rest: Compare each category against all others combined
#'     \item pairwise: Compare all pairs of categories
#'   }
#' @param reference_level Character, reference category for binary comparison (default: first level)
#' @param top_n Number of top terms per comparison (default: 10)
#' @param min_count Minimum word count to include (default: 5)
#'
#' @return Data frame with columns:
#'   \itemize{
#'     \item term: The word/feature
#'     \item category1: First category in comparison
#'     \item category2: Second category in comparison
#'     \item count1: Count in category 1
#'     \item count2: Count in category 2
#'     \item odds1: Smoothed odds in category 1, (count + 1) / (total + V - count - 1)
#'     \item odds2: Smoothed odds in category 2
#'     \item odds_ratio: Ratio of odds
#'     \item log_odds_ratio: Log of odds ratio (positive = more in compared category)
#'     \item variance: Approximate variance of the log odds ratio
#'     \item z_score: Log odds ratio divided by its standard error
#'     \item significant: TRUE when |z| >= 1.96
#'   }
#'   Terms are ranked by absolute z-score.
#'
#' @concept lexical
#' @export
#'
#' @examples
#' \donttest{
#' articles <- TextAnalysisR::SpecialEduTech[1:20, ]
#' corpus <- quanteda::corpus(
#'   articles$abstract,
#'   docvars = data.frame(reference_type = articles$reference_type)
#' )
#' dfm_object <- quanteda::dfm(quanteda::tokens(corpus))
#' log_odds <- calculate_log_odds_ratio(dfm_object, "reference_type")
#' }
calculate_log_odds_ratio <- function(dfm_object,
                                      group_var,
                                      comparison_mode = c("binary", "one_vs_rest", "pairwise"),
                                      reference_level = NULL,
                                      top_n = 10,
                                      min_count = 5) {

  comparison_mode <- match.arg(comparison_mode)

  if (!inherits(dfm_object, "dfm")) {
    stop("dfm_object must be a quanteda dfm object")
  }

  if (!group_var %in% names(quanteda::docvars(dfm_object))) {
    stop("group_var '", group_var, "' not found in document variables")
  }

  groups <- quanteda::docvars(dfm_object)[[group_var]]
  levels <- unique(groups[!is.na(groups)])

  if (length(levels) < 2) {
    stop(
      "Need at least 2 categories for comparison in '", group_var, "'. ",
      "Found only: ", paste(levels, collapse = ", "),
      call. = FALSE
    )
  }

  # Helper function for pairwise comparison
  compare_two <- function(dfm_obj, level1, level2, groups) {
    idx1 <- which(groups == level1)
    idx2 <- which(groups == level2)

    if (length(idx1) == 0 || length(idx2) == 0) {
      return(NULL)
    }

    # Sum counts per group
    counts1 <- Matrix::colSums(dfm_obj[idx1, , drop = FALSE])
    counts2 <- Matrix::colSums(dfm_obj[idx2, , drop = FALSE])

    # Totals from the full vocabulary, before any filtering
    n1 <- sum(counts1)
    n2 <- sum(counts2)
    vocab_size <- length(counts1)

    # min_count filters which terms are reported, not the totals
    keep <- (counts1 + counts2) >= min_count
    counts1 <- counts1[keep]
    counts2 <- counts2[keep]

    if (length(counts1) == 0) {
      return(NULL)
    }

    # smoothed odds with uniform Dirichlet prior (a_w = 1, a_0 = V)
    odds1 <- (counts1 + 1) / (n1 + vocab_size - counts1 - 1)
    odds2 <- (counts2 + 1) / (n2 + vocab_size - counts2 - 1)

    odds_ratio <- odds1 / odds2
    log_odds <- log(odds_ratio)

    variance <- 1 / (counts1 + 1) + 1 / (n1 + vocab_size - counts1 - 1) +
      1 / (counts2 + 1) + 1 / (n2 + vocab_size - counts2 - 1)
    z_score <- log_odds / sqrt(variance)

    result <- data.frame(
      term = names(counts1),
      category1 = level1,
      category2 = level2,
      count1 = as.numeric(counts1),
      count2 = as.numeric(counts2),
      odds1 = odds1,
      odds2 = odds2,
      odds_ratio = odds_ratio,
      log_odds_ratio = log_odds,
      variance = as.numeric(variance),
      z_score = z_score,
      significant = abs(z_score) >= 1.96,
      stringsAsFactors = FALSE
    )

    result <- result[order(abs(result$z_score), decreasing = TRUE), ]
    utils::head(result, top_n)
  }

  results <- list()


  if (comparison_mode == "binary") {
    if (length(levels) != 2 && is.null(reference_level)) {
      message("More than 2 categories. Using first two: ", levels[1], " vs ", levels[2])
    }

    if (is.null(reference_level)) {
      # Default: second level compared to first (first as reference)
      level1 <- levels[2]  # Compared (numerator)
      level2 <- levels[1]  # Reference (denominator)
    } else {
      # User-specified reference: compare the other level to reference
      level1 <- setdiff(levels, reference_level)[1]  # Compared (numerator)
      level2 <- reference_level  # Reference (denominator)
    }

    results[[1]] <- compare_two(dfm_object, level1, level2, groups)

  } else if (comparison_mode == "one_vs_rest") {
    for (level in levels) {
      # Combine all other categories
      groups_binary <- ifelse(groups == level, level, "Other")
      result <- compare_two(dfm_object, level, "Other", groups_binary)
      if (!is.null(result)) {
        results[[length(results) + 1]] <- result
      }
    }

  } else if (comparison_mode == "pairwise") {
    pairs <- utils::combn(levels, 2, simplify = FALSE)
    for (pair in pairs) {
      result <- compare_two(dfm_object, pair[1], pair[2], groups)
      if (!is.null(result)) {
        results[[length(results) + 1]] <- result
      }
    }
  }

  if (length(results) == 0) {
    return(data.frame(
      term = character(),
      category1 = character(),
      category2 = character(),
      count1 = numeric(),
      count2 = numeric(),
      odds1 = numeric(),
      odds2 = numeric(),
      odds_ratio = numeric(),
      log_odds_ratio = numeric(),
      variance = numeric(),
      z_score = numeric()
    ))
  }

  do.call(rbind, results)
}


#' Calculate Weighted Log Odds Ratio
#'
#' @description
#' Computes weighted log odds ratios using the method from Monroe, Colaresi,
#' and Quinn (2008) "Fightin' Words" via the tidylo package. This method
#' weights log odds by variance (z-score) to identify words that reliably
#' distinguish between groups, accounting for sampling variability.
#'
#' @param dfm_object A quanteda dfm object
#' @param group_var Character, name of the document variable to group by
#' @param top_n Number of top terms to return per group (default: 10)
#' @param min_count Minimum total count for a term to be included (default: 5)
#'
#' @return A data frame with the grouping variable, feature, n,
#'   log_odds_weighted (from tidylo::bind_log_odds), and significant
#'   (TRUE when |log_odds_weighted| >= 1.96)
#'
#' @references
#' Monroe, B. L., Colaresi, M. P., & Quinn, K. M. (2008). Fightin' words:
#' Lexical feature selection and evaluation for identifying the content of
#' political conflict. Political Analysis, 16(4), 372-403.
#'
#' Silge, J., & Robinson, D. (2017). Text mining with R: A tidy approach.
#' O'Reilly Media. https://www.tidytextmining.com/
#'
#' @concept lexical
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("tidylo", quietly = TRUE)) {
#'   articles <- TextAnalysisR::SpecialEduTech[1:20, ]
#'   dfm_object <- quanteda::dfm(quanteda::tokens(articles$abstract))
#'   quanteda::docvars(dfm_object, "reference_type") <- articles$reference_type
#'   weighted_odds <- calculate_weighted_log_odds(dfm_object, "reference_type",
#'                                                 top_n = 5)
#' }
#' }
calculate_weighted_log_odds <- function(dfm_object,
                                        group_var,
                                        top_n = 10,
                                        min_count = 5) {

  if (!requireNamespace("tidylo", quietly = TRUE)) {
    stop("Package 'tidylo' is required. Please install it with: install.packages('tidylo')")
  }

  if (!inherits(dfm_object, "dfm")) {
    stop("dfm_object must be a quanteda dfm object")
  }

  if (!group_var %in% names(quanteda::docvars(dfm_object))) {
    stop("group_var '", group_var, "' not found in document variables")
  }

  # sparse group aggregation keeps memory flat on large corpora
  grouped_dfm <- quanteda::dfm_group(dfm_object,
                                     groups = quanteda::docvars(dfm_object)[[group_var]])
  grouped_counts <- quanteda::convert(grouped_dfm, to = "data.frame")
  names(grouped_counts)[1] <- group_var
  grouped_counts <- tidyr::pivot_longer(
    grouped_counts,
    cols = -dplyr::all_of(group_var),
    names_to = "feature",
    values_to = "n"
  )

  # group totals and the prior require the full vocabulary
  result <- tidylo::bind_log_odds(
    grouped_counts,
    set = !!rlang::sym(group_var),
    feature = feature,
    n = n
  )

  result <- result %>%
    dplyr::filter(n >= min_count) %>%
    dplyr::group_by(.data[[group_var]]) %>%
    dplyr::slice_max(abs(.data$log_odds_weighted), n = top_n, with_ties = FALSE) %>%
    dplyr::ungroup() %>%
    dplyr::arrange(.data[[group_var]], dplyr::desc(abs(.data$log_odds_weighted)))

  result$significant <- abs(result$log_odds_weighted) >= 1.96
  as.data.frame(result)
}


#' Plot Log Odds Ratio
#'
#' @description
#' Creates a horizontal bar plot showing log odds ratios for comparing
#' word usage between categories. Positive values indicate higher usage
#' in the first category, negative in the second.
#'
#' @param log_odds_data Data frame from calculate_log_odds_ratio()
#' @param top_n Number of top terms to show per direction (default: 10)
#' @param facet_by Character, column name to facet by (e.g., "category1" for
#'   one_vs_rest comparisons). NULL for no faceting.
#' @param color_positive Color for positive log odds (default: "#10B981" green)
#' @param color_negative Color for negative log odds (default: "#EF4444" red)
#' @param height Plot height in pixels (default: 600)
#' @param width Plot width in pixels (default: NULL for auto)
#' @param title Plot title (default: "Log Odds Ratio Comparison")
#'
#' @return A plotly object
#'
#' @concept visualization
#' @export
#'
#' @examples
#' \donttest{
#' articles <- TextAnalysisR::SpecialEduTech[1:20, ]
#' corpus <- quanteda::corpus(
#'   articles$abstract,
#'   docvars = data.frame(reference_type = articles$reference_type)
#' )
#' dfm_object <- quanteda::dfm(quanteda::tokens(corpus))
#' log_odds <- calculate_log_odds_ratio(dfm_object, "reference_type",
#'                                       comparison_mode = "binary")
#' plot_log_odds_ratio(log_odds, top_n = 5)
#' }
plot_log_odds_ratio <- function(log_odds_data,
                                 top_n = 10,
                                 facet_by = NULL,
                                 color_positive = "#10B981",
                                 color_negative = "#EF4444",
                                 height = 600,
                                 width = NULL,
                                 title = "Log Odds Ratio Comparison") {

  if (is.null(log_odds_data) || nrow(log_odds_data) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "No log odds data available",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  positive <- log_odds_data[log_odds_data$log_odds_ratio > 0, ]
  negative <- log_odds_data[log_odds_data$log_odds_ratio < 0, ]

  positive <- positive[order(positive$log_odds_ratio, decreasing = TRUE), ]
  negative <- negative[order(negative$log_odds_ratio, decreasing = FALSE), ]

  plot_data <- rbind(
    utils::head(positive, top_n),
    utils::head(negative, top_n)
  )

  if (nrow(plot_data) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "No significant differences found",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  plot_data <- plot_data[order(plot_data$log_odds_ratio), ]
  plot_data$term_ordered <- factor(plot_data$term, levels = plot_data$term)
  plot_data$direction <- ifelse(plot_data$log_odds_ratio > 0, "positive", "negative")

  plot_data$hover_text <- paste0(plot_data$term,
                                  "\nLog Odds: ", round(plot_data$log_odds_ratio, 3),
                                  "\n", plot_data$category1, ": ", plot_data$count1,
                                  "\n", plot_data$category2, ": ", plot_data$count2)

  cat1 <- unique(plot_data$category1)[1]
  cat2 <- unique(plot_data$category2)[1]
  subtitle <- paste0(cat2, " (-) vs ", cat1, " (+)")

  ggplot2::ggplot(plot_data, ggplot2::aes(x = log_odds_ratio, y = term_ordered,
                                           fill = direction, text = hover_text)) +
    ggplot2::geom_col() +
    ggplot2::geom_vline(xintercept = 0, linetype = "dotted", color = "#94A3B8") +
    ggplot2::scale_fill_manual(values = c("positive" = color_positive,
                                           "negative" = color_negative),
                                guide = "none") +
    ggplot2::labs(x = "Log Odds Ratio", y = "",
                  title = title, subtitle = subtitle) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      plot.subtitle = ggplot2::element_text(size = 11, color = "#6B7280"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}


#' Plot Weighted Log Odds
#'
#' @description
#' Creates a faceted horizontal bar plot showing weighted log odds for comparing
#' word usage across categories using the Fightin' Words method (Monroe et al. 2008).
#' Each group is displayed in a separate facet showing its most distinctive terms.
#'
#' @param weighted_data Data frame from calculate_weighted_log_odds()
#' @param top_n Number of top terms to show per group (default: 10)
#' @param color_positive Color for positive log odds (default: "#10B981" green)
#' @param color_negative Color for negative log odds (default: "#EF4444" red)
#' @param height Plot height in pixels (default: 600)
#' @param width Plot width in pixels (default: NULL for auto)
#' @param title Plot title (default: "Weighted Log Odds by Group")
#'
#' @return A plotly object
#'
#' @concept visualization
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("tidylo", quietly = TRUE)) {
#'   articles <- TextAnalysisR::SpecialEduTech[1:20, ]
#'   dfm_object <- quanteda::dfm(quanteda::tokens(articles$abstract))
#'   quanteda::docvars(dfm_object, "reference_type") <- articles$reference_type
#'   weighted_odds <- calculate_weighted_log_odds(dfm_object, "reference_type",
#'                                                 top_n = 5)
#'   plot_weighted_log_odds(weighted_odds)
#' }
#' }
plot_weighted_log_odds <- function(weighted_data,
                                   top_n = 10,
                                   color_positive = "#10B981",
                                   color_negative = "#EF4444",
                                   height = 600,
                                   width = NULL,
                                   title = "Weighted Log Odds by Group") {

  if (is.null(weighted_data) || nrow(weighted_data) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "No weighted log odds data available",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  group_col <- setdiff(names(weighted_data), c("feature", "n", "log_odds_weighted", "log_odds"))[1]

  if (is.null(group_col)) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "Could not identify group column",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  plot_data <- weighted_data %>%
    dplyr::group_by(.data[[group_col]]) %>%
    dplyr::slice_max(abs(log_odds_weighted), n = top_n, with_ties = FALSE) %>%
    dplyr::ungroup() %>%
    dplyr::mutate(
      direction = ifelse(log_odds_weighted > 0, "positive", "negative"),
      hover_text = paste0("Term: ", feature,
                          "<br>Log Odds: ", round(log_odds_weighted, 3),
                          "<br>", group_col, ": ", .data[[group_col]]),
      feature = tidytext::reorder_within(feature, log_odds_weighted, .data[[group_col]])
    )

  ggplot2::ggplot(plot_data, ggplot2::aes(x = log_odds_weighted, y = feature,
                                          fill = direction, text = hover_text)) +
    ggplot2::geom_col() +
    ggplot2::geom_vline(xintercept = 0, linetype = "dotted", color = "#94A3B8") +
    ggplot2::facet_wrap(stats::as.formula(paste("~", group_col)), scales = "free_y") +
    tidytext::scale_y_reordered() +
    ggplot2::scale_fill_manual(values = c("positive" = color_positive,
                                          "negative" = color_negative),
                               guide = "none") +
    ggplot2::labs(x = "Log Odds (Weighted)", y = "", title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a", hjust = 0.5),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      strip.text = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}


# Lexical dispersion analysis

#' Calculate Lexical Dispersion
#'
#' @description
#' Computes lexical dispersion data for specified terms across a corpus.
#' Shows where terms appear within each document, useful for understanding
#' term distribution patterns.
#'
#' @param tokens_object A quanteda tokens object
#' @param terms Character vector of terms to analyze
#' @param scale Character, "relative" (0-1 normalized) or "absolute" (token position)
#'
#' @return Data frame with columns:
#'   \itemize{
#'     \item doc_id: Document identifier
#'     \item term: The search term
#'     \item position: Position in document (relative or absolute)
#'     \item doc_length: Total tokens in document
#'   }
#'
#' @concept lexical
#' @export
#'
#' @examples
#' \donttest{
#' tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1:5])
#' dispersion <- calculate_lexical_dispersion(tokens, c("learning", "instruction"))
#' }
calculate_lexical_dispersion <- function(tokens_object,
                                          terms,
                                          scale = c("relative", "absolute")) {

  scale <- match.arg(scale)


  if (!inherits(tokens_object, "tokens")) {
    stop("tokens_object must be a quanteda tokens object")
  }

  if (is.null(terms) || length(terms) == 0) {
    return(data.frame(
      doc_id = character(),
      term = character(),
      position = numeric(),
      doc_length = integer(),
      stringsAsFactors = FALSE
    ))
  }

  results <- list()

  for (i in seq_along(tokens_object)) {
    doc_tokens <- as.character(tokens_object[[i]])
    doc_tokens_lower <- tolower(doc_tokens)
    doc_length <- length(doc_tokens)
    doc_name <- names(tokens_object)[i]

    if (is.null(doc_name) || doc_name == "") {
      doc_name <- paste0("Doc ", i)
    }

    for (term in terms) {
      term_lower <- tolower(term)
      positions <- which(doc_tokens_lower == term_lower)

      if (length(positions) > 0) {
        if (scale == "relative") {
          positions <- positions / doc_length
        }

        results[[length(results) + 1]] <- data.frame(
          doc_id = rep(doc_name, length(positions)),
          term = rep(term, length(positions)),
          position = positions,
          doc_length = rep(doc_length, length(positions)),
          stringsAsFactors = FALSE
        )
      }
    }
  }

  if (length(results) == 0) {
    return(data.frame(
      doc_id = character(),
      term = character(),
      position = numeric(),
      doc_length = integer(),
      stringsAsFactors = FALSE
    ))
  }

  do.call(rbind, results)
}


#' Plot Lexical Dispersion
#'
#' @description
#' Creates an X-ray plot showing where terms appear across documents.
#' Each row represents a term, and marks indicate occurrences.
#'
#' @param dispersion_data Data frame from calculate_lexical_dispersion()
#' @param scale Character, "relative" or "absolute" (must match calculation)
#' @param title Plot title (default: "Lexical Dispersion")
#' @param colors Named vector of colors for each term, or NULL for auto
#' @param height Plot height in pixels (default: 400)
#' @param width Plot width in pixels (default: NULL for auto)
#' @param marker_size Size of position markers (default: 8)
#'
#' @return A plotly object
#'
#' @concept visualization
#' @export
#'
#' @examples
#' \donttest{
#' tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1:5])
#' dispersion <- calculate_lexical_dispersion(tokens, c("learning", "instruction"))
#' plot_lexical_dispersion(dispersion)
#' }
plot_lexical_dispersion <- function(dispersion_data,
                                     scale = "relative",
                                     title = "Lexical Dispersion",
                                     colors = NULL,
                                     height = 400,
                                     width = NULL,
                                     marker_size = 8) {

  if (is.null(dispersion_data) || nrow(dispersion_data) == 0) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "No term occurrences found in the corpus",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  unique_terms <- unique(dispersion_data$term)

  if (is.null(colors)) {
    default_colors <- c("#3B82F6", "#10B981", "#F59E0B", "#EF4444",
                        "#8B5CF6", "#EC4899", "#06B6D4", "#84CC16")
    colors <- stats::setNames(
      rep(default_colors, length.out = length(unique_terms)),
      unique_terms
    )
  }

  dispersion_data$term <- factor(dispersion_data$term, levels = rev(unique_terms))

  dispersion_data$hover_text <- paste0(
    "Term: ", dispersion_data$term,
    "<br>Document: ", dispersion_data$doc_id,
    "<br>Position: ", round(dispersion_data$position, 3),
    if (scale == "relative") " (relative)" else ""
  )

  x_label <- if (scale == "relative") {
    "Relative Position (0 = start, 1 = end)"
  } else {
    "Token Position"
  }

  term_levels <- levels(dispersion_data$term)
  dispersion_data$term_num <- as.numeric(dispersion_data$term)

  p <- ggplot2::ggplot(dispersion_data, ggplot2::aes(color = term)) +
    suppressWarnings(ggplot2::geom_segment(ggplot2::aes(x = position, xend = position,
                                        y = term_num - 0.4, yend = term_num + 0.4,
                                        text = hover_text),
                          linewidth = 0.3, alpha = 0.7)) +
    suppressWarnings(ggplot2::geom_point(ggplot2::aes(x = position, y = term_num, text = hover_text),
                        size = 4, alpha = 0)) +
    ggplot2::scale_y_continuous(breaks = seq_along(term_levels), labels = term_levels) +
    ggplot2::scale_color_manual(values = colors) +
    ggplot2::labs(x = x_label, y = "", title = title, color = "") +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      legend.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      legend.text = ggplot2::element_text(size = 12, color = "#3B3B3B"),
      legend.position = "bottom"
    )

  if (scale == "relative") {
    p <- p + ggplot2::xlim(0, 1)
  }

  p
}


#' Calculate Dispersion Metrics
#'
#' @description
#' Computes quantitative dispersion metrics for terms, measuring how
#' evenly distributed they are across the corpus.
#'
#' @param tokens_object A quanteda tokens object
#' @param terms Character vector of terms to analyze
#'
#' @return Data frame with columns:
#'   \itemize{
#'     \item term: The search term
#'     \item frequency: Total occurrences
#'     \item doc_count: Number of documents containing term
#'     \item doc_ratio: Proportion of documents containing term
#'     \item juilland_d: Juilland's D dispersion (0-1, higher = more even),
#'       computed on within-document proportions
#'     \item rosengren_s: Rosengren's S dispersion adjusted for document
#'       sizes (1/n_docs to 1, higher = more even)
#'   }
#'
#' @concept lexical
#' @export
calculate_dispersion_metrics <- function(tokens_object, terms) {

  if (!inherits(tokens_object, "tokens")) {
    stop("tokens_object must be a quanteda tokens object")
  }

  n_docs <- length(tokens_object)
  doc_lengths <- quanteda::ntoken(tokens_object)
  corpus_size <- sum(doc_lengths)

  results <- lapply(terms, function(term) {
    term_lower <- tolower(term)

    # Count occurrences in each document
    doc_counts <- vapply(tokens_object, function(doc_tokens) {
      sum(tolower(as.character(doc_tokens)) == term_lower)
    }, integer(1))

    total_freq <- sum(doc_counts)
    doc_count <- sum(doc_counts > 0)
    doc_ratio <- doc_count / n_docs

    # Juilland's D: population SD of within-document proportions
    nonempty <- doc_lengths > 0
    n_parts <- sum(nonempty)
    if (total_freq > 0 && n_parts > 1) {
      p <- doc_counts[nonempty] / doc_lengths[nonempty]
      pop_sd <- sqrt(mean((p - mean(p))^2))
      juilland_d <- 1 - (pop_sd / mean(p)) / sqrt(n_parts - 1)
      juilland_d <- max(0, min(1, juilland_d))
    } else {
      juilland_d <- NA_real_
    }

    # size-adjusted Rosengren's S: (sum sqrt(s_i * v_i))^2 / f
    if (total_freq > 0 && n_docs > 1 && corpus_size > 0) {
      s <- doc_lengths / corpus_size
      rosengren_s <- sum(sqrt(s * doc_counts))^2 / total_freq
    } else {
      rosengren_s <- NA_real_
    }

    data.frame(
      term = term,
      frequency = total_freq,
      doc_count = doc_count,
      doc_ratio = round(doc_ratio, 3),
      juilland_d = round(juilland_d, 3),
      rosengren_s = round(rosengren_s, 3),
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, results)
}

# regression tests for hand-implemented statistics

test_that("c-TF-IDF matches the class-based formula", {
  texts <- c("apple banana apple", "banana cherry", "dog cat dog", "cat mouse")
  assignments <- c(1, 1, 2, 2)

  kw <- TextAnalysisR:::generate_semantic_topic_keywords(
    texts, assignments, n_keywords = 3, method = "c-tfidf"
  )

  expect_length(kw, 2)
  expect_true(all(lengths(kw) > 0))
  expect_true("apple" %in% kw[["1"]])
  expect_true("dog" %in% kw[["2"]])
})

test_that("mmr diversity changes keyword selection", {
  set.seed(7)
  texts <- TextAnalysisR::SpecialEduTech$abstract[1:40]
  assignments <- rep(1:2, each = 20)

  kw_low <- TextAnalysisR:::generate_semantic_topic_keywords(
    texts, assignments, n_keywords = 8, method = "mmr", diversity = 0.1
  )
  kw_high <- TextAnalysisR:::generate_semantic_topic_keywords(
    texts, assignments, n_keywords = 8, method = "mmr", diversity = 0.9
  )

  expect_true(all(lengths(kw_low) == 8))
  expect_false(identical(kw_low, kw_high))
})

test_that("keyword stability matches topics across permuted labels", {
  kws <- list(
    a = c("math", "algebra", "geometry"),
    b = c("reading", "phonics", "fluency"),
    c = c("autism", "behavior", "social")
  )

  expect_equal(calculate_keyword_stability(kws, kws[c(3, 1, 2)]), 1)
  expect_equal(
    calculate_keyword_stability(
      kws,
      list(x = c("zebra", "lion"), y = c("piano", "violin"), z = c("soccer", "tennis"))
    ),
    0
  )
})

test_that("weighted log odds uses full-vocabulary totals", {
  skip_if_not_installed("tidylo")

  abstracts <- TextAnalysisR::SpecialEduTech[1:40, ]
  corp <- quanteda::corpus(
    abstracts$abstract,
    docvars = data.frame(reference_type = abstracts$reference_type)
  )
  dfm_object <- quanteda::dfm(quanteda::tokens(corp, remove_punct = TRUE))

  result <- calculate_weighted_log_odds(dfm_object, "reference_type",
                                        top_n = 5, min_count = 2)

  expect_true(all(result$n >= 2))
  expect_true(all(is.finite(result$log_odds_weighted)))
  expect_type(result$significant, "logical")

  # z-scores must match tidylo computed on the unfiltered complete grid
  tidy_full <- tidytext::tidy(dfm_object) %>%
    dplyr::left_join(
      data.frame(document = quanteda::docnames(dfm_object),
                 reference_type = abstracts$reference_type),
      by = "document"
    ) %>%
    dplyr::group_by(reference_type, term) %>%
    dplyr::summarise(n = sum(count), .groups = "drop") %>%
    tidyr::complete(reference_type, term, fill = list(n = 0))
  expected <- tidylo::bind_log_odds(tidy_full, set = reference_type,
                                    feature = term, n = n)
  merged <- merge(result, expected,
                  by.x = c("reference_type", "feature"),
                  by.y = c("reference_type", "term"))
  expect_equal(merged$log_odds_weighted.x, merged$log_odds_weighted.y,
               tolerance = 1e-10)
})

test_that("log odds ratio returns finite Monroe-style estimates", {
  abstracts <- TextAnalysisR::SpecialEduTech[1:40, ]
  corp <- quanteda::corpus(
    abstracts$abstract,
    docvars = data.frame(reference_type = abstracts$reference_type)
  )
  dfm_object <- quanteda::dfm(quanteda::tokens(corp, remove_punct = TRUE))

  result <- calculate_log_odds_ratio(dfm_object, "reference_type", top_n = 5)

  expect_true(all(result$odds1 > 0))
  expect_true(all(result$odds2 > 0))
  expect_true(all(is.finite(result$z_score)))
  expect_type(result$significant, "logical")
})

test_that("keyness handles every documented measure", {
  abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:30]
  dfm_object <- quanteda::dfm(quanteda::tokens(quanteda::corpus(abstracts),
                                               remove_punct = TRUE))

  for (m in c("lr", "chi2", "exact", "pmi")) {
    kw <- extract_keywords_keyness(dfm_object, target = 1:5, top_n = 5, measure = m)
    expect_true(is.numeric(kw$Keyness_Score), info = m)
    expect_equal(nrow(kw), 5, info = m)
  }
  expect_error(extract_keywords_keyness(dfm_object, target = 1, measure = "bogus"))
})

test_that("dispersion metrics stay in bounds", {
  toks <- quanteda::tokens(quanteda::corpus(TextAnalysisR::SpecialEduTech$abstract[1:30]))
  disp <- calculate_dispersion_metrics(toks, c("students", "the"))

  expect_true(all(disp$rosengren_s <= 1 + 1e-9, na.rm = TRUE))
  expect_true(all(disp$rosengren_s > 0, na.rm = TRUE))
  expect_true(all(disp$juilland_d >= 0 & disp$juilland_d <= 1, na.rm = TRUE))
})

test_that("mtld skips short factors and hdd stays in bounds", {
  toks <- rep(c("the", "the", "cat", "sat", "on", "a", "mat", "with",
                "her", "big", "hat", "today"), 12)

  expect_true(is.finite(TextAnalysisR:::.calc_mtld(toks)))
  expect_true(TextAnalysisR:::.calc_mtld(toks) > 0)

  hdd <- TextAnalysisR:::.calc_hdd(toks)
  expect_true(hdd > 0 && hdd <= 1)
  expect_true(is.na(TextAnalysisR:::.calc_hdd(toks[1:10])))
})

test_that("package loads successfully", {
  expect_true(requireNamespace("TextAnalysisR", quietly = TRUE))
})

test_that("example dataset is accessible", {
  data(SpecialEduTech, package = "TextAnalysisR")
  expect_true(exists("SpecialEduTech"))
  expect_s3_class(SpecialEduTech, "data.frame")
  expect_true(nrow(SpecialEduTech) > 0)
})

test_that("unite_cols works", {
  skip_if_not_installed("quanteda")

  data(SpecialEduTech, package = "TextAnalysisR")
  test_data <- SpecialEduTech[1:3, c("title", "abstract")]

  united <- unite_cols(test_data, c("title", "abstract"))
  expect_true("united_texts" %in% names(united))
  expect_equal(nrow(united), 3)
})

test_that("prep_texts works", {
  skip_if_not_installed("quanteda")

  data(SpecialEduTech, package = "TextAnalysisR")
  test_data <- data.frame(
    united_texts = SpecialEduTech$abstract[1:3],
    stringsAsFactors = FALSE
  )

  tokens <- prep_texts(test_data, text_field = "united_texts")
  expect_true(inherits(tokens, "tokens"))
})

test_that("plot_word_frequency works", {
  skip_if_not_installed("quanteda")
  data(SpecialEduTech, package = "TextAnalysisR")
  texts <- SpecialEduTech$abstract[1:5]
  corp <- quanteda::corpus(texts)
  toks <- quanteda::tokens(corp)
  dfm_obj <- quanteda::dfm(toks)

  plot <- plot_word_frequency(dfm_obj, n = 5)
  expect_s3_class(plot, "gg")
})

test_that("sentiment_lexicon_analysis works with basic texts", {
  skip_if_not_installed("quanteda")
  skip_if_not_installed("tidytext")

  data(SpecialEduTech, package = "TextAnalysisR")
  texts <- SpecialEduTech$abstract[1:5]

  corp <- quanteda::corpus(texts)
  toks <- quanteda::tokens(corp)
  dfm_obj <- quanteda::dfm(toks)

  result <- sentiment_lexicon_analysis(
    dfm_object = dfm_obj,
    lexicon = "bing",
    feature_type = "words"
  )

  expect_type(result, "list")
  expect_true("document_sentiment" %in% names(result))
  expect_true("summary_stats" %in% names(result))
  expect_equal(result$feature_type, "words")
})

test_that("calculate_text_readability works", {
  skip_if_not_installed("quanteda")
  skip_if_not_installed("quanteda.textstats")

  data(SpecialEduTech, package = "TextAnalysisR")
  texts <- SpecialEduTech$abstract[1:5]

  result <- calculate_text_readability(
    texts = texts,
    metrics = c("flesch", "gunning_fog")
  )

  expect_s3_class(result, "data.frame")
  expect_true("Document" %in% names(result))
  expect_equal(nrow(result), 5)
  expect_false("Lexical Diversity (TTR)" %in% names(result))
  expect_true("Lexical Diversity (MTLD)" %in% names(result))
})

test_that("calculate_text_readability sentence count handles abbreviations", {
  skip_if_not_installed("quanteda")
  skip_if_not_installed("quanteda.textstats")

  result <- calculate_text_readability(
    texts = "Results from the U.S. were strong (e.g., Fig. 1). A second sentence follows.",
    include_lexical_diversity = FALSE
  )

  expect_true("Avg Sentence Length" %in% names(result))
  expect_gt(result$`Avg Sentence Length`, 4)
})

test_that("calculate_log_odds_ratio ranks by z-score", {
  skip_if_not_installed("quanteda")

  data(SpecialEduTech, package = "TextAnalysisR")
  articles <- SpecialEduTech[1:20, ]
  corp <- quanteda::corpus(
    articles$abstract,
    docvars = data.frame(reference_type = articles$reference_type)
  )
  dfm_object <- quanteda::dfm(quanteda::tokens(corp))

  result <- calculate_log_odds_ratio(dfm_object, "reference_type", min_count = 2)

  expect_true(all(c("variance", "z_score") %in% names(result)))
  ranked <- abs(result$z_score)
  expect_false(is.unsorted(rev(ranked)))
})

test_that("lexical_frequency_analysis works", {
  skip_if_not_installed("quanteda")
  data(SpecialEduTech, package = "TextAnalysisR")
  texts <- SpecialEduTech$abstract[1:5]

  corp <- quanteda::corpus(texts)
  toks <- quanteda::tokens(corp, remove_punct = TRUE)
  dfm_obj <- quanteda::dfm(toks)

  result <- lexical_frequency_analysis(dfm_obj, n = 5)

  expect_s3_class(result, "gg")
})

test_that("extract_keywords_tfidf works", {
  skip_if_not_installed("quanteda")

  data(SpecialEduTech, package = "TextAnalysisR")
  texts <- SpecialEduTech$abstract[1:5]

  corp <- quanteda::corpus(texts)
  toks <- quanteda::tokens(corp)
  dfm_obj <- quanteda::dfm(toks)

  result <- extract_keywords_tfidf(dfm_obj, top_n = 3)

  expect_s3_class(result, "data.frame")
  expect_true("Keyword" %in% names(result))
  expect_true("TF_IDF_Score" %in% names(result))
  expect_true("Frequency" %in% names(result))
})

test_that("run_app function exists", {
  expect_true(exists("run_app"))
  expect_type(run_app, "closure")
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(message = FALSE, warning = FALSE)

## -----------------------------------------------------------------------------
library(TextAnalysisR)

mydata <- SpecialEduTech[1:150, ]
united_tbl <- unite_cols(mydata, listed_vars = c("title", "keyword", "abstract"))
tokens <- prep_texts(united_tbl, text_field = "united_texts", remove_stopwords = TRUE)
dfm_object <- quanteda::dfm(tokens)

## -----------------------------------------------------------------------------
network <- word_co_occurrence_network(
  dfm_object,
  co_occur_n = 50,
  top_node_n = 30,
  node_label_size = 22,
  community_method = "leiden"
)

network$plot
network$table
network$summary

## -----------------------------------------------------------------------------
corr_network <- word_correlation_network(
  dfm_object,
  common_term_n = 15,
  corr_n = 0.3,
  community_method = "leiden"
)

corr_network$plot
corr_network$table

## -----------------------------------------------------------------------------
subset_texts <- united_tbl$united_texts[1:10]

similarity <- semantic_similarity_analysis(
  subset_texts,
  document_feature_type = "words",
  similarity_method = "cosine",
  verbose = FALSE
)

plot_similarity_heatmap(similarity$similarity_matrix, method_name = "Cosine")

## -----------------------------------------------------------------------------
term_matrix <- as.matrix(dfm_object)[1:30, ]
normalized <- term_matrix / sqrt(rowSums(term_matrix^2))
sim_matrix <- normalized %*% t(normalized)

docs_data <- data.frame(
  display_name = paste0("doc", 1:30),
  reference_type = quanteda::docvars(dfm_object, "reference_type")[1:30]
)
dimnames(sim_matrix) <- list(docs_data$display_name, docs_data$display_name)

cross <- extract_cross_category_similarities(
  sim_matrix,
  docs_data,
  reference_category = "journal_article",
  category_var = "reference_type",
  id_var = "display_name"
)

gaps <- analyze_similarity_gaps(cross)
gaps$summary_stats

## ----eval = FALSE-------------------------------------------------------------
# results <- run_rag_search(
#   query = "math intervention for students with disabilities",
#   documents = united_tbl$united_texts,
#   provider = "openai"
# )

## -----------------------------------------------------------------------------
sentiment <- sentiment_lexicon_analysis(dfm_object, lexicon = "bing")
plot_sentiment_distribution(sentiment$document_sentiment)

## ----eval = FALSE-------------------------------------------------------------
# emotion <- sentiment_lexicon_analysis(dfm_object, lexicon = "nrc")
# plot_emotion_radar(emotion$emotion_scores)

## -----------------------------------------------------------------------------
scored <- sentiment_lexicon_analysis(dfm_object, lexicon = "bing")$document_sentiment
scored$reference_type <- quanteda::docvars(dfm_object, "reference_type")[
  match(scored$document, quanteda::docnames(dfm_object))
]
plot_sentiment_by_category(scored, category_var = "reference_type")

## -----------------------------------------------------------------------------
data_matrix <- as.matrix(dfm_object)

groups <- cluster_embeddings(data_matrix, method = "kmeans", n_clusters = 5, verbose = FALSE)
groups$n_clusters

labels <- generate_cluster_labels_auto(data_matrix, groups$clusters, method = "tfidf", n_terms = 3)
labels

## -----------------------------------------------------------------------------
coords <- reduce_dimensions(data_matrix, method = "PCA", n_components = 2, verbose = FALSE)
head(coords$reduced_data)

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom tibble tibble
#' @importFrom stats as.dist coef kmeans lm median runif sd setNames var
## usethis namespace: end
NULL

#' @title TextAnalysisR: A Text Mining Workflow Tool
#'
#' @description
#' TextAnalysisR provides a set of text analysis tools including:
#' - Text preprocessing and corpus preparation
#' - Structural topic modeling
#' - Word networks and co-occurrence analysis
#' - Semantic similarity analysis
#' - Dimensionality reduction and clustering
#' - Interactive visualizations
#' - Shiny web application interface
#'
#' The package is designed to work both as a standalone R package and as the backend
#' for a Shiny web application, providing flexibility for different user needs.
#'
#' @section Main Functions:
#' 
#' **Data Processing:**
#' - `import_files()`: Process various file formats
#' - `unite_cols()`: Combine text columns
#' - `prep_texts()`: Text preprocessing with validation
#'
#' **Topic Modeling:**
#' - `find_optimal_k()`: Find optimal number of topics
#' - `get_topic_terms()`: Extract top terms per topic
#' - `generate_topic_labels()`: Auto-generate topic labels using OpenAI
#' - `plot_word_probability()`: Visualize topic terms
#' - `plot_topic_probability()`: Visualize topic prevalence
#'
#' **Network Analysis:**
#' - `word_co_occurrence_network()`: Word co-occurrence networks
#' - `word_correlation_network()`: Word correlation networks
#' - `calculate_word_frequency()`: Word frequency over variables
#' 
#' **Semantic Analysis:**
#' - `semantic_similarity_analysis()`: Multi-method document similarity analysis
#' - `reduce_dimensions()`: PCA, t-SNE, UMAP dimensionality reduction
#' - `cluster_embeddings()`: Clustering with k-means, hierarchical, DBSCAN
#' - `plot_semantic_viz()`: Interactive semantic analysis visualizations
#'
#' **Workflow Functions:**
#' - `run_text_workflow()`: End-to-end text preprocessing workflow
#'
#' @section Usage Examples:
#' 
#' **Standalone R Package Usage:**
#' \preformatted{
#' # Load the package
#' library(TextAnalysisR)
#'
#' # Basic preprocessing workflow
#' data <- TextAnalysisR::SpecialEduTech
#'
#' # Unite text columns
#' united_data <- unite_cols(data, c("title", "keyword", "abstract"))
#'
#' # Preprocess texts
#' tokens <- prep_texts(
#'   united_tbl = united_data,
#'   text_field = "united_texts",
#'   min_char = 2,
#'   verbose = TRUE
#' )
#'
#' # Create document-feature matrix
#' dfm_obj <- quanteda::dfm(tokens)
#'
#' # Word frequency visualization
#' freq_plot <- plot_word_frequency(dfm_obj, n = 20)
#'
#' # Network analysis
#' cooccur_network <- word_co_occurrence_network(
#'   dfm_obj,
#'   co_occur_n = 10,
#'   top_node_n = 50
#' )
#'
#' # Topic modeling
#' stm_data <- quanteda::convert(dfm_obj, to = "stm")
#' topic_model <- stm::stm(
#'   documents = stm_data$documents,
#'   vocab = stm_data$vocab,
#'   K = 10,
#'   verbose = TRUE
#' )
#'
#' # Get and visualize topic terms
#' topic_terms <- get_topic_terms(topic_model, top_term_n = 10)
#' topic_plot <- plot_word_probability(topic_terms, ncol = 3)
#'
#' # Document similarity analysis
#' texts <- united_data$united_texts
#' sim_result <- semantic_similarity_analysis(
#'   texts = texts,
#'   document_feature_type = "words",
#'   similarity_method = "cosine"
#' )
#'
#' # Visualize similarity matrix
#' sim_plot <- plot_semantic_viz(
#'   sim_result,
#'   plot_type = "similarity"
#' )
#'
#' # Clustering analysis
#' data_matrix <- as.matrix(dfm_obj)
#' cluster_result <- cluster_embeddings(
#'   data_matrix,
#'   method = "kmeans",
#'   n_clusters = 5
#' )
#'
#' # Visualize clusters
#' cluster_plot <- plot_semantic_viz(
#'   cluster_result,
#'   plot_type = "clustering"
#' )
#' }
#' 
#' **Shiny Application Usage:**
#' \preformatted{
#' # Launch the Shiny application
#' TextAnalysisR::run_app()
#'
#' # The Shiny app provides an interactive interface for all functions
#' # and automatically calls the underlying R package functions
#' }
#'
#' @section Package Features:
#' 
#' **Flexible Input Support:**
#' - Excel files (.xlsx, .xls)
#' - CSV files
#' - PDF documents
#' - Word documents (.docx)
#' - Text files
#' - Copy-paste text input
#' 
#' **Text Analysis:**
#' - Multi-method similarity analysis (words, n-grams, topics, embeddings)
#' - Dimensionality reduction (PCA, t-SNE, UMAP)
#' - Multiple clustering algorithms (k-means, hierarchical, UMAP+DBSCAN)
#' - Network analysis and visualization
#' - Topic modeling with STM
#' 
#' **Interactive Visualizations:**
#' - Plotly-based interactive plots
#' - Network visualizations
#' - Similarity heatmaps
#' - Dimensionality reduction scatter plots
#' - Topic term plots
#' 
#' **Integration Capabilities:**
#' - OpenAI API integration for topic labeling
#' - Python integration for embeddings (sentence-transformers)
#' - Export options
#' - Memory-efficient processing
#'
#' @section Getting Started:
#'
#' 1. **For Text Preprocessing:** Use `run_text_workflow()` for end-to-end preprocessing
#' 2. **For Interactive Use:** Launch the Shiny app with `run_app()`
#' 3. **For Custom Workflows:** Use individual functions as building blocks
#' 4. **For Power Users:** Combine with custom R code and other packages
#'

#' @title Launch the TextAnalysisR app
#'
#' @name run_app
#'
#' @description
#' Launch the TextAnalysisR Shiny application.
#'
#' @param launch.browser Logical. Whether to open the app in a browser.
#'   Defaults to `interactive()`, which is FALSE in non-interactive sessions
#'   (e.g., Docker containers, servers).
#'
#' @return No return value, called for side effects (launching Shiny app).
#'
#' @examples
#' if (interactive()) {
#'   run_app()
#' }
#'
#' @export
#'
#' @import dplyr
#' @import ggplot2
#' @import shiny
#' @import tidyr
#' @importFrom magrittr %>%

run_app <- function(launch.browser = interactive()) {
  appDir <- system.file("TextAnalysisR.app", package = "TextAnalysisR")

  if (appDir == "") {
    stop("Error: TextAnalysisR.app directory not found.")
  }

  shiny::runApp(appDir, display.mode = "normal", launch.browser = launch.browser)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(message = FALSE, warning = FALSE)

## -----------------------------------------------------------------------------
library(TextAnalysisR)

mydata <- SpecialEduTech[seq_len(20), c("title", "abstract")]
united <- unite_cols(mydata, listed_vars = c("title", "abstract"))
toks   <- prep_texts(united, text_field = "united_texts")
dfm    <- quanteda::dfm(toks)
plot_word_frequency(dfm, n = 10)

#' Pipe operator
#'
#' See \code{magrittr::\link[magrittr:pipe]{\%>\%}} for details.
#'
#' @name %>%
#' @rdname pipe
#' @keywords internal
#' @export
#' @importFrom magrittr %>%
#' @usage lhs \%>\% rhs
#' @param lhs A value or the magrittr placeholder.
#' @param rhs A function call using the magrittr semantics.
#' @return The result of calling `rhs(lhs)`.
NULL



globalVariables(names = c(
  ".", ".max_idx", ".row_idx", "categorical_var", "centrality", "col_idx",
  "collocation", "community", "continuous_var", "cooccur_count", "correlation",
  "correlation_rounded", "degree_log", "document", "eigenvector", "emotion",
  "entity", "estimate", "feature", "frequency", "from", "generated_content", "group", "size_metric_log",
  "interview_question", "item1", "item2", "label", "labeled_topic", "line_group",
  "lower", "max_corr", "max_count", "max_similarity", "metric_value", "min_corr",
  "min_count", "model_type", "n", "n_sentiment_words", "name", "negative", "rate ratio",
  "odds.ratio", "ord", "other_category", "other_id", "other_idx", "other_name",
  "p.value", "percent", "policy_recommendation", "pos", "positive", "proportion",
  "ref_id", "ref_idx", "ref_name", "research_question", "row_idx", "score",
  "sentiment", "sentiment_score", "similarity", "statistic", "std.error",
  "std.error (rate ratio)", "survey_item", "term", "term_proportion",
  "terms", "text", "theme_description", "to",
  "topic", "topic_display", "topic_label", "total_count", "total_score", "tt",
  "united_texts", "upper", "value", "word", "word_frequency", "x", "xend",
  "y", "yend",
  "Count", "Keyness_Score", "Keyword", "Keyword_ordered", "Metric",
  "Percentage", "Score", "TF_IDF_Score", "Value", "collocation_ordered",
  "direction", "doc_ordered", "entity_ordered", "feature_ordered", "k",
  "log_odds_ratio", "log_odds_weighted", "metric_val", "pos_ordered",
  "position", "term_ordered", "count", "term_num"
))



#' @importFrom utils modifyList
#' @importFrom stats cor
NULL


# Utility and Helper Functions
# General-purpose utility functions for analysis and visualization

#
# Deployment Detection Utilities
#

#' Check Docker Deployment
#'
#' @description
#' Detects whether the app is running in a Docker container.
#' Docker deployments have full Python/spaCy capability.
#'
#' @return Logical TRUE if running in Docker
#'
#' @keywords internal
check_docker_deployment <- function() {
  has_dockerenv <- file.exists("/.dockerenv")
  has_docker_env_var <- nzchar(Sys.getenv("TEXTANALYSISR_DOCKER"))
  has_docker_container <- nzchar(Sys.getenv("DOCKER_CONTAINER"))

  return(has_dockerenv || has_docker_env_var || has_docker_container)
}

#' Check Deployment Environment
#'
#' @description
#' Detects whether the app is running on a web server (shinyapps.io, Posit Connect)
#' versus locally via `run_app()`.
#'
#' @return Logical TRUE if running on web server, FALSE if local
#'
#' @keywords internal
#'
check_web_deployment <- function() {
  # Docker has Python/spaCy available - not a restricted deployment
  if (check_docker_deployment()) {
    return(FALSE)
  }

  # Check for restricted web servers (no Python)
  shinyapps <- Sys.getenv("R_CONFIG_ACTIVE") == "shinyapps"
  shinyapps_io <- grepl("shinyapps", Sys.getenv("SHINY_SERVER_URL", ""), ignore.case = TRUE)
  connect <- nzchar(Sys.getenv("RSTUDIO_CONNECT_HASTE"))

  return(shinyapps || shinyapps_io || connect)
}

#' Check Feature Status
#'
#' @description
#' Checks if a specific optional feature is available in the current environment.
#'
#' @param feature Character: "python", "pdf_tables", "embeddings", "sentiment_transformer"
#'
#' @return Logical TRUE if feature is available
#'
#' @keywords internal
#'
check_feature <- function(feature) {
  feature <- tolower(feature)

  if (check_web_deployment()) {
    return(feature %in% c("core", "lexical", "stm"))
  }

  switch(feature,
    "python" = tryCatch({
      status <- check_python_env()
      isTRUE(status$available)
    }, error = function(e) FALSE),
    "pdf_tables" = tryCatch({
      status <- check_python_env()
      isTRUE(status$available) && isTRUE(status$packages$pdfplumber)
    }, error = function(e) FALSE),
    "embeddings" = tryCatch({
      requireNamespace("reticulate", quietly = TRUE) &&
        reticulate::py_module_available("sentence_transformers")
    }, error = function(e) FALSE),
    "sentiment_transformer" = tryCatch({
      requireNamespace("reticulate", quietly = TRUE) &&
        reticulate::py_module_available("transformers")
    }, error = function(e) FALSE),
    TRUE
  )
}

#' Get Feature Status
#'
#' @description
#' Returns availability status for all optional features.
#'
#' @return Named list with feature availability
#'
#' @keywords internal
#'
get_feature_status <- function() {
  features <- c("python", "pdf_tables", "embeddings", "sentiment_transformer")
  result <- lapply(features, function(f) {
    tryCatch(check_feature(f), error = function(e) FALSE)
  })
  names(result) <- features
  result$web <- check_web_deployment()
  result$local <- !check_web_deployment()
  return(result)
}

#' Show Web Deployment Banner
#'
#' @param disabled Optional character vector naming features to mark as
#'   disabled in the banner; \code{NULL} (default) shows the standard set.
#' @return A \code{shiny.tag} object containing the banner HTML for
#'   inclusion in a Shiny UI, or \code{NULL} when not running in a web
#'   deployment context.
#' @keywords internal
#' @export
show_web_banner <- function(disabled = NULL) {
  if (!check_web_deployment()) return(NULL)

  if (is.null(disabled)) {
    disabled <- c("Python PDF processing", "Local AI",
                  "Embedding analysis", "Large files (>10MB)")
  }

  feature_list <- paste0("<li>", htmltools::htmlEscape(disabled), "</li>", collapse = "")

  shiny::tagList(
    shiny::tags$details(
      class = "web-version-note",
      style = "margin: 10px; padding: 8px 12px; background-color: #EFF6FF; border: 1px solid #BFDBFE; border-radius: 6px; font-size: 14px;",
      shiny::tags$summary(
        style = "cursor: pointer; color: #1E40AF; font-weight: 500;",
        shiny::icon("info-circle"),
        " Web version - some features limited"
      ),
      shiny::tags$div(
        style = "margin-top: 8px; padding-left: 5px; color: #374151;",
        shiny::tags$p(
          style = "margin: 5px 0;",
          "For full features: ",
          shiny::tags$code(
            style = "background: #F3F4F6; padding: 2px 6px; border-radius: 3px; font-size: 13px;",
            "install.packages('TextAnalysisR', repos = c('https://mshin77.r-universe.dev', 'https://cloud.r-project.org'))"
          )
        ),
        shiny::HTML(paste0("<ul style='margin: 5px 0; padding-left: 20px; color: #6B7280;'>", feature_list, "</ul>"))
      )
    )
  )
}

#' Require Feature
#'
#' @description
#' Checks feature availability and shows notification if unavailable.
#'
#' @param feature Character: feature name to check
#' @param session Shiny session object (optional)
#'
#' @return Logical TRUE if available, FALSE if not
#'
#' @keywords internal
#'
require_feature <- function(feature, session = NULL) {
  if (check_feature(feature)) return(TRUE)

  msg <- switch(feature,
    "python" = "Python not configured. Run setup_python_env().",
    "pdf_tables" = "PDF tables require Python. Run setup_python_env().",
    "embeddings" = "Embeddings require sentence-transformers. Run: pip install sentence-transformers torch",
    "sentiment_transformer" = "Neural sentiment requires transformers. Run: pip install transformers torch",
    paste0("Feature '", feature, "' not available.")
  )

  if (check_web_deployment()) {
    msg <- paste(msg, "Only available in R package version.")
  }

  if (!is.null(session)) {
    shiny::showNotification(msg, type = "warning", duration = 8)
  } else {
    message(msg)
  }

  return(FALSE)
}

#' Create Formatted Analysis Data Table
#'
#' @description Creates a consistently formatted DT::datatable for analysis results with
#' export buttons and optional numeric formatting.
#'
#' @param data Data frame to display
#' @param colnames Optional character vector of column names for display
#' @param numeric_cols Optional character vector of numeric columns to round
#' @param digits Number of digits for rounding numeric columns (default: 3)
#' @param export_formats Character vector of export formats (default: c('copy', 'csv', 'excel', 'pdf', 'print'))
#' @param page_length Number of rows per page (default: 25)
#' @param font_size Font size for table cells (default: "16px")
#'
#' @return A DT::datatable object
#'
#' @keywords internal
#'
create_analysis_datatable <- function(data,
                                      colnames = NULL,
                                      numeric_cols = NULL,
                                      digits = 3,
                                      export_formats = c("copy", "csv", "excel", "pdf", "print"),
                                      page_length = 25,
                                      font_size = "16px") {

  dt <- DT::datatable(
    data,
    colnames = colnames,
    rownames = FALSE,
    extensions = "Buttons",
    options = list(
      scrollX = TRUE,
      pageLength = page_length,
      dom = "Bfrtip",
      buttons = export_formats
    )
  )

  if (!is.null(numeric_cols) && length(numeric_cols) > 0) {
    valid_cols <- intersect(numeric_cols, names(data))
    if (length(valid_cols) > 0) {
      dt <- DT::formatRound(dt, columns = valid_cols, digits = digits)
    }
  }

  dt <- DT::formatStyle(dt, columns = names(data), `font-size` = font_size)

  return(dt)
}

#' Calculate Cosine Similarity Matrix
#'
#' @description Calculates the cosine similarity between all pairs of rows in a matrix.
#' @param matrix_data A numeric matrix where rows represent documents/observations
#' @return A square similarity matrix with values between -1 and 1
#' @keywords internal
#' @export
calculate_cosine_similarity <- function(matrix_data) {
  if (is.null(matrix_data) || nrow(matrix_data) == 0 || ncol(matrix_data) == 0) {
    return(matrix(1, nrow = 1, ncol = 1))
  }

  matrix_data[!is.finite(matrix_data)] <- 0

  normalized_matrix <- t(apply(matrix_data, 1, function(row) {
    norm <- sqrt(sum(row^2, na.rm = TRUE))
    if (norm > 0 && is.finite(norm)) {
      return(row / norm)
    } else {
      return(rep(0, length(row)))
    }
  }))

  normalized_matrix[!is.finite(normalized_matrix)] <- 0

  similarity_matrix <- normalized_matrix %*% t(normalized_matrix)

  similarity_matrix[!is.finite(similarity_matrix)] <- 0
  diag(similarity_matrix) <- 1
  similarity_matrix[similarity_matrix > 1] <- 1
  similarity_matrix[similarity_matrix < -1] <- -1

  return(similarity_matrix)
}

#' Create Error Plot for Plotly
#'
#' @description Creates a plotly error/status plot with a message
#' @param message The message to display
#' @param color Color for the text (default: "#ef4444")
#' @return A plotly plot object displaying the message
#' @keywords internal
plot_error <- function(message, color = "#ef4444") {
  if (!requireNamespace("plotly", quietly = TRUE)) {
    stop("plotly package is required for this function. ",
         "Please install it with: install.packages('plotly')")
  }

  plotly::plot_ly(type = "scatter", mode = "markers") %>%
    plotly::add_annotations(
      text = message,
      xref = "paper",
      yref = "paper",
      x = 0.5,
      y = 0.5,
      showarrow = FALSE,
      font = list(size = 14, color = color),
      xanchor = "center",
      yanchor = "middle"
    ) %>%
    plotly::layout(
      xaxis = list(visible = FALSE),
      yaxis = list(visible = FALSE),
      margin = list(t = 40, r = 40, b = 40, l = 40),
      plot_bgcolor = "white",
      paper_bgcolor = "white"
    )
}


#' @title Complete Text Mining Workflow
#'
#' @description
#' This function provides a complete text mining workflow that follows the same sequence
#' as the Shiny application: file processing -> text uniting -> preprocessing ->
#' DFM creation -> analysis.
#' It serves as a convenience function for users who want to execute the entire
#' pipeline programmatically.
#'
#' @param dataset_choice A character string indicating the dataset choice:
#'   "Upload an Example Dataset", "Upload Your File", "Copy and Paste Text".
#' @param file_info A data frame containing file information (for file upload).
#' @param text_input A character string containing text input (for copy-paste).
#' @param listed_vars A character vector of column names to unite into text.
#' @param min_char The minimum number of characters for tokens (default: 2).
#' @param remove_punct Logical; remove punctuation (default: TRUE).
#' @param remove_symbols Logical; remove symbols (default: TRUE).
#' @param remove_numbers Logical; remove numbers (default: TRUE).
#' @param remove_url Logical; remove URLs (default: TRUE).
#' @param detect_compounds Logical; detect multi-word expressions (default: FALSE).
#' @param compound_size Size range for compound detection (default: 2:3).
#' @param compound_min_count Minimum count for compounds (default: 2).
#' @param verbose Logical; print progress messages (default: TRUE).
#'
#' @return A list containing processed data, tokens, DFM, and metadata.
#'
#' @export
#'
#' @examples
#' \donttest{
#'   workflow_result <- TextAnalysisR::run_text_workflow(
#'     dataset_choice = "Upload an Example Dataset",
#'     listed_vars = c("title", "keyword", "abstract")
#'   )
#' }
#' if (interactive()) {
#'   file_info <- data.frame(filepath = "path/to/file.xlsx")
#'   workflow_result <- TextAnalysisR::run_text_workflow(
#'     dataset_choice = "Upload Your File",
#'     file_info = file_info,
#'     listed_vars = c("column1", "column2")
#'   )
#'
#'   workflow_result <- TextAnalysisR::run_text_workflow(
#'     dataset_choice = "Copy and Paste Text",
#'     text_input = "Your text content here",
#'     listed_vars = "text"
#'   )
#' }
run_text_workflow <- function(dataset_choice,
                                         file_info = NULL,
                                         text_input = NULL,
                                         listed_vars,
                                         min_char = 2,
                                         remove_punct = TRUE,
                                         remove_symbols = TRUE,
                                         remove_numbers = TRUE,
                                         remove_url = TRUE,
                                         detect_compounds = FALSE,
                                         compound_size = 2:3,
                                         compound_min_count = 2,
                                         verbose = TRUE) {

  if (verbose) message("Starting complete text mining workflow...")
  start_time <- Sys.time()

  if (verbose) message("Step 1: Processing files...")
  processed_data <- import_files(
    dataset_choice = dataset_choice,
    file_info = file_info,
    text_input = text_input
  )

  if (verbose) message("Step 2: Uniting text columns...")
  united_data <- unite_cols(processed_data, listed_vars = listed_vars)

  if (verbose) message("Step 3: Preprocessing texts...")
  tokens <- prep_texts(
    united_tbl = united_data,
    text_field = "united_texts",
    min_char = min_char,
    remove_punct = remove_punct,
    remove_symbols = remove_symbols,
    remove_numbers = remove_numbers,
    remove_url = remove_url,
    verbose = verbose
  )

  compounds <- NULL
  if (detect_compounds) {
    if (verbose) message("Step 4: Detecting multi-word expressions...")
    compounds <- detect_multi_words(
      tokens = tokens,
      size = compound_size,
      min_count = compound_min_count
    )

    if (length(compounds) > 0) {
      tokens <- quanteda::tokens_compound(tokens, compounds)
    }
  }

  if (verbose) message("Step 5: Creating document-feature matrix...")
  dfm_object <- quanteda::dfm(tokens)

  execution_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  if (verbose) {
    message("Complete text mining workflow finished in ", round(execution_time, 2), " seconds")
    message("Documents processed: ", quanteda::ndoc(dfm_object))
    message("Features identified: ", quanteda::nfeat(dfm_object))
  }

  return(list(
    raw_data = processed_data,
    united_data = united_data,
    tokens = tokens,
    compounds = compounds,
    dfm = dfm_object,
    workflow_info = list(
      dataset_choice = dataset_choice,
      listed_vars = listed_vars,
      compounds_detected = !is.null(compounds),
      execution_time = execution_time,
      timestamp = Sys.time()
    )
  ))
}

# broom has no tidy method for zeroinfl; extract the count component
.tidy_count_model <- function(model) {
  if (!inherits(model, "zeroinfl")) {
    return(broom::tidy(model))
  }
  cf <- summary(model)$coefficients$count
  tibble::tibble(
    term = rownames(cf),
    estimate = cf[, 1],
    std.error = cf[, 2],
    statistic = cf[, 3],
    p.value = cf[, 4]
  )
}

#' @title Analyze and Visualize Word Frequencies Across a Continuous Variable
#'
#' @description
#' This function analyzes and visualizes word frequencies across a continuous variable.
#'
#' @param dfm_object A quanteda document-feature matrix (dfm).
#' @param continuous_variable A continuous variable in the metadata.
#' @param selected_terms A vector of terms to analyze trends for.
#' @param height The height of the resulting Plotly plot, in pixels (default: 500).
#' @param width The width of the resulting Plotly plot, in pixels (default: 900).
#'
#' @return A list containing Plotly objects and tables with the results.
#'
#' @details This function requires a fitted STM model object and a quanteda dfm object.
#' The continuous variable should be a column in the metadata of the dfm object.
#' The selected terms should be a vector of terms to analyze trends for.
#' The required packages are 'htmltools', 'splines', and 'broom' (plus
#' additional ones loaded internally).
#'
#' @importFrom stats glm reformulate binomial
#'
#' @seealso [extract_keywords_tfidf()] and [extract_keywords_keyness()] for ranking terms by importance; [plot_word_frequency()] for the standard frequency-by-doc plot
#' @export
#'
#' @examples
#' \donttest{
#'   mydata <- TextAnalysisR::SpecialEduTech
#'
#'   united_tbl <- TextAnalysisR::unite_cols(
#'     mydata,
#'     listed_vars = c("title", "keyword", "abstract")
#'   )
#'
#'   tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")
#'
#'   dfm_object <- quanteda::dfm(tokens)
#'
#'   word_freq_results <- TextAnalysisR::calculate_word_frequency(
#'     dfm_object,
#'     continuous_variable = "year",
#'     selected_terms = c("calculator", "computer"),
#'     height = 500,
#'     width = 900
#'   )
#'   print(word_freq_results$plot)
#'   print(word_freq_results$table)
#' }
calculate_word_frequency <- function(dfm_object,
                                 continuous_variable,
                                 selected_terms,
                                 height = 500,
                                 width = 900) {

  if (!requireNamespace("htmltools", quietly = TRUE) ||
      !requireNamespace("MASS", quietly = TRUE) ||
      !requireNamespace("pscl", quietly = TRUE) ||
      !requireNamespace("broom", quietly = TRUE)) {
    stop(
      "The 'htmltools', 'pscl', 'MASS', and 'broom' packages are required for this functionality. ",
      "Please install them using install.packages(c('htmltools', 'MASS', 'pscl', 'broom'))."
    )
  }

  dfm_outcome_obj <- dfm_object
  # dense conversion keeps zero counts so regressions include all documents
  dfm_selected <- quanteda::dfm_select(dfm_outcome_obj, pattern = selected_terms,
                                       selection = "keep", valuetype = "fixed")
  dfm_td <- quanteda::convert(dfm_selected, to = "data.frame")
  names(dfm_td)[1] <- "document"
  dfm_td <- tidyr::pivot_longer(dfm_td, cols = -"document",
                                names_to = "term", values_to = "count")

  docvars_df <- quanteda::docvars(dfm_outcome_obj)
  docvars_df$document <- quanteda::docnames(dfm_outcome_obj)

  dfm_td <- dfm_td %>%
    left_join(docvars_df,
              by = c("document" = "document"))

  con_var_term_counts <- dfm_td %>%
    tibble::as_tibble() %>%
    mutate(word_frequency = count)

  con_var_term_gg <- con_var_term_counts %>%
    mutate(term = factor(term, levels = selected_terms)) %>%
    mutate(across(where(is.numeric), ~ round(., 3))) %>%
    filter(term %in% selected_terms) %>%
    ggplot(aes(
      x = !!rlang::sym(continuous_variable),
      y = word_frequency,
      group = term
    )) +
    geom_point(color = "#337ab7", alpha = 0.6, size = 1) +
    geom_line(color = "#337ab7", alpha = 0.6, linewidth = 0.5) +
    facet_wrap(~ term, scales = "free") +
    ggplot2::scale_y_continuous(labels = scales::number_format(accuracy = 1)) +
    labs(y = "Word Frequency") +
    theme_minimal(base_size = 11) +
    theme(
      legend.position = "none",
      panel.grid.major = element_blank(),
      panel.grid.minor = element_blank(),
      axis.line = element_line(color = "#3B3B3B", linewidth = 0.3),
      axis.ticks = element_line(color = "#3B3B3B", linewidth = 0.3),
      strip.text.x = element_text(size = 11, color = "#3B3B3B", face = "bold"),
      axis.text.x = element_text(size = 11, color = "#3B3B3B"),
      axis.text.y = element_text(size = 11, color = "#3B3B3B"),
      axis.title = element_text(size = 11, color = "#3B3B3B"),
      axis.title.x = element_text(margin = margin(t = 9)),
      axis.title.y = element_text(margin = margin(r = 11))
    )

  con_var_term_plot <- con_var_term_gg

  significance_results <- con_var_term_counts %>%
    mutate(word = term) %>%
    filter(word %in% selected_terms) %>%
    group_by(word) %>%
    group_modify(~ {
      continuous_var <- if (is.null(continuous_variable) ||
                            length(continuous_variable) == 0) {
        stop("No continuous variable selected.")
      } else {
        continuous_variable[1]
      }

      df <- .x %>%
        dplyr::mutate(
          word_frequency = as.numeric(word_frequency),
          !!continuous_var := as.numeric(!!rlang::sym(continuous_var))
        ) %>%
        dplyr::filter(is.finite(word_frequency) &
                        is.finite(!!rlang::sym(continuous_var)))

      if (length(unique(df$word_frequency)) <= 1) {
        return(tibble::tibble(term = NA, estimate = NA, std.error = NA,
                              statistic = NA, p.value = NA,
                              `rate ratio` = NA, var.diag = NA,
                              `std.error (rate ratio)` = NA,
                              model_type = "Insufficient data"))
      }

      if (length(unique(df[[continuous_var]])) <= 1) {
        return(tibble::tibble(term = NA, estimate = NA, std.error = NA,
                              statistic = NA, p.value = NA,
                              `rate ratio` = NA, var.diag = NA,
                              `std.error (rate ratio)` = NA,
                              model_type = "Insufficient data"))
      }

      if (nrow(df) < 2) {
        return(tibble::tibble(term = NA, estimate = NA, std.error = NA,
                              statistic = NA, p.value = NA,
                              `rate ratio` = NA, var.diag = NA,
                              `std.error (rate ratio)` = NA,
                              model_type = "Insufficient data"))
      }

      formula_simple <- as.formula(paste0("word_frequency ~ ", continuous_var))

      mean_count <- mean(df$word_frequency, na.rm = TRUE)
      var_count <- var(df$word_frequency, na.rm = TRUE)
      dispersion_ratio <- ifelse(mean_count != 0, var_count / mean_count, NA)
      prop_zero <- mean(df$word_frequency == 0, na.rm = TRUE)

      model <- NULL

      if (prop_zero > 0.5) {
        model <- tryCatch(
          pscl::zeroinfl(formula_simple, data = df, dist = "negbin", link = "logit"),
          error = function(e) {
            return(NULL)
          }
        )
        if (!is.null(model)) {
          model_type <- "Zero-Inflated Negative Binomial"
        }
      }

      if (is.null(model)) {
        model <- tryCatch(
          MASS::glm.nb(formula_simple, data = df, control = glm.control(maxit = 200)),
          error = function(e) {
            return(NULL)
          }
        )
        if (is.null(model)) {
          model <- glm(formula_simple, family = poisson(link = "log"), data = df)
          model_type <- "Poisson"
        } else {
          model_type <- "Negative Binomial"
        }
      }

      tidy_result <- .tidy_count_model(model) %>%
        dplyr::mutate(
          `rate ratio` = exp(estimate),
          var.diag = std.error^2,
          `std.error (rate ratio)` = ifelse(var.diag >= 0,
                                            sqrt(`rate ratio`^2 * var.diag),
                                            NA),
          model_type = model_type
        )

      return(tidy_result)
    }) %>%
    ungroup() %>%
    dplyr::select(word, model_type, term, estimate, std.error,
                  `rate ratio`, `std.error (rate ratio)`, statistic, p.value) %>%
    rename(
      `log(rate ratio)` = estimate,
      `z-statistic` = statistic
    )


  significance_results_tables <- significance_results %>%
    mutate(word = factor(word, levels = selected_terms)) %>%
    arrange(word) %>%
    group_by(word) %>%
    group_map(~ {
      htmltools::tagList(
        htmltools::tags$div(
          style = "margin-bottom: 20px;",
          htmltools::tags$p(
            .y$word,
            style = "font-weight: bold; text-align: center; font-size: 11pt;"
          )
        ),
        .x %>%
          mutate_if(is.numeric, ~ round(., 3)) %>%
          DT::datatable(
            rownames = FALSE,
            extensions = "Buttons",
            options = list(
              scrollX = TRUE,
              width = "80%",
              dom = "Bfrtip",
              buttons = c("copy", "csv", "excel", "pdf", "print")
            )
          ) %>%
          DT::formatStyle(
            columns = names(.x),
            `font-size` = "16px"
          )
      )
    })

  list(
    plot = con_var_term_plot,
    table = htmltools::tagList(significance_results_tables) %>% htmltools::browsable()
  )
}


#' @title Calculate Similarity Metrics
#'
#' @description
#' Calculates similarity metrics including statistical measures
#' and network properties.
#' Internal function used by document_similarity_analysis.
#'
#' @param similarity_matrix A similarity matrix.
#' @param labels Optional vector of labels for clustering metrics.
#' @param method_info Optional method information.
#'
#' @return A list of metrics.
#'
#' @keywords internal
calculate_metrics <- function(similarity_matrix, labels = NULL, method_info = NULL) {
  off_diagonal <- similarity_matrix[upper.tri(similarity_matrix) | lower.tri(similarity_matrix)]

  metrics <- list(
    n_docs = nrow(similarity_matrix),
    mean_similarity = round(mean(off_diagonal, na.rm = TRUE), 4),
    median_similarity = round(median(off_diagonal, na.rm = TRUE), 4),
    std_similarity = round(sd(off_diagonal, na.rm = TRUE), 4),
    min_similarity = round(min(off_diagonal, na.rm = TRUE), 4),
    max_similarity = round(max(off_diagonal, na.rm = TRUE), 4),
    similarity_range = paste(round(min(off_diagonal, na.rm = TRUE), 3), "to",
                             round(max(off_diagonal, na.rm = TRUE), 3)),
    sparsity = round(sum(off_diagonal < 0.1, na.rm = TRUE) / length(off_diagonal), 4),
    connectivity = round(sum(off_diagonal > 0.3, na.rm = TRUE) / length(off_diagonal), 4),
    skewness = ifelse(requireNamespace("moments", quietly = TRUE),
                     round(moments::skewness(off_diagonal, na.rm = TRUE), 4), NA),
    kurtosis = ifelse(requireNamespace("moments", quietly = TRUE),
                     round(moments::kurtosis(off_diagonal, na.rm = TRUE), 4), NA),
    silhouette_score = NA,
    modularity = NA
  )

  if (!is.null(labels) && is.atomic(labels) && is.vector(labels) &&
      length(labels) > 0 && length(unique(labels)) > 1 &&
      length(unique(labels)) < length(labels) * 0.8) {

    tryCatch({
      if (requireNamespace("cluster", quietly = TRUE)) {
        dist_matrix <- as.dist(1 - similarity_matrix)
        sil_result <- cluster::silhouette(as.numeric(as.factor(labels)), dist_matrix)
        metrics$silhouette_score <- if (is.numeric(sil_result[, 3]) && !is.na(sil_result[, 3]))
                                   round(mean(sil_result[, 3]), 3) else NA
      }
    }, error = function(e) {
      metrics$silhouette_score <- NA
    })

    tryCatch({
      if (requireNamespace("igraph", quietly = TRUE)) {
        threshold <- quantile(similarity_matrix[upper.tri(similarity_matrix)], 0.75, na.rm = TRUE)
        adj_matrix <- similarity_matrix > threshold
        diag(adj_matrix) <- FALSE
        graph <- igraph::graph_from_adjacency_matrix(adj_matrix, mode = "undirected")
        communities <- igraph::cluster_louvain(graph)
        metrics$modularity <- if (is.numeric(igraph::modularity(communities)) &&
                                  !is.na(igraph::modularity(communities)))
                             round(igraph::modularity(communities), 3) else NA
      }
    }, error = function(e) {
      metrics$modularity <- NA
    })
  }


  if (!is.null(method_info)) {
    metrics$method <- method_info$method
    metrics$model_name <- method_info$model_name %||% "N/A"
  }

  return(metrics)
}


# Python environment setup and utilities

#' Setup Python Environment
#'
#' @description
#' Sets up a tiered Python virtual environment.
#'
#' @param envname Character string name for the virtual environment
#'   (default: "textanalysisr-env")
#' @param tier Which feature tier to install. One or more of:
#'   `"core"` (spacy + pdfplumber, ~200 MB; default),
#'   `"embeddings"` (adds sentence-transformers + transformers + torch, ~1 GB),
#'   `"topics"` (adds BERTopic + UMAP + HDBSCAN, ~300 MB on top of embeddings).
#' @param force Logical, whether to recreate environment if it exists
#'   (default: FALSE)
#'
#' @return Invisible TRUE if successful, stops with error message if failed
#'
#' @details
#' Default `tier = "core"` keeps the install light -- spaCy NLP and PDF text
#' extraction only. Add `"embeddings"` for sentence-transformers-based
#' similarity/sentiment, and `"topics"` for BERTopic.
#'
#' The virtual environment is isolated; system Python is not modified.
#'
#' @export
#'
#' @examples
#' if (interactive()) {
#'   setup_python_env()                              # core only (~200 MB)
#'   setup_python_env(tier = c("core", "embeddings"))  # +1 GB
#'   setup_python_env(tier = c("core", "embeddings", "topics"))  # full stack
#' }
setup_python_env <- function(envname = "textanalysisr-env",
                             tier = "core",
                             force = FALSE) {
  tier <- match.arg(tier, c("core", "embeddings", "topics"), several.ok = TRUE)
  if (!"core" %in% tier) tier <- c("core", tier)
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("Python environment setup"))
  }

  env_exists <- tryCatch(
    envname %in% reticulate::virtualenv_list(),
    error = function(e) FALSE
  )

  if (!interactive() && !env_exists) {
    message("setup_python_env() skipped: non-interactive session and ",
            "virtualenv '", envname, "' not found. ",
            "Run in an interactive R session to create it.")
    return(invisible(FALSE))
  }

  message("\nPython Environment Setup")


  # Check if Python is available
  python_available <- tryCatch({
    py_config <- reticulate::py_discover_config()
    !is.null(py_config$python)
  }, error = function(e) FALSE)

  if (!python_available) {
    if (!interactive()) {
      message("Python not found. Install Python (e.g. from python.org) and rerun ",
              "setup_python_env() in an interactive R session.")
      return(invisible(FALSE))
    }
    message("No Python found. Install Miniconda? (y/n): ")
    response <- readline(prompt = "")

    if (tolower(trimws(response)) == "y") {
      message("Installing Miniconda...")
      reticulate::install_miniconda()
      message("Done.")
    } else {
      message("Python required. Install from python.org and retry.")
      return(invisible(FALSE))
    }
  } else {
    py_info <- reticulate::py_discover_config()
    message("Python: ", py_info$python, " (v", py_info$version, ")")
  }

  tryCatch({
    env_exists <- envname %in% reticulate::virtualenv_list()

    if (env_exists && !force) {
      message("Environment '", envname, "' exists. Use force=TRUE to recreate.")
      reticulate::use_virtualenv(envname, required = TRUE)
      return(invisible(TRUE))
    }

    if (env_exists && force) {
      message("Removing existing environment...")
      reticulate::virtualenv_remove(envname, confirm = FALSE)
    }

    message("Creating environment '", envname, "'...")
    reticulate::virtualenv_create(envname, python = NULL)
    reticulate::use_virtualenv(envname, required = TRUE)

    tier_files <- paste0("requirements-", tier, ".txt")
    req_packages <- unlist(lapply(tier_files, function(f) {
      path <- system.file("python", f, package = "TextAnalysisR")
      if (!nzchar(path) || !file.exists(path)) return(character(0))
      lines <- readLines(path)
      lines[!grepl("^#|^\\s*$", lines)]
    }))

    if (length(req_packages) == 0) {
      req_packages <- c("spacy>=3.5.0", "pdfplumber>=0.10.0")
    }

    message("Installing packages for tier(s): ", paste(tier, collapse = ", "))
    reticulate::virtualenv_install(envname = envname, packages = req_packages, ignore_installed = FALSE)

    message("Testing imports...")
    test_result <- tryCatch({
      reticulate::py_run_string("import spacy")
      reticulate::py_run_string("import pdfplumber")
      TRUE
    }, error = function(e) {
      message("Import failed: ", e$message)
      FALSE
    })

    if (!test_result) {
      message("Package imports failed. Check Python logs.")
      return(invisible(FALSE))
    }

    # Download spaCy English model
    message("Downloading spaCy English model (en_core_web_sm)...")
    spacy_model_result <- tryCatch({
      reticulate::py_run_string("
import spacy
try:
    nlp = spacy.load('en_core_web_sm')
    print('Model already installed')
except OSError:
    print('Downloading model...')
    spacy.cli.download('en_core_web_sm')
    print('Model downloaded')
")
      TRUE
    }, error = function(e) {
      message("Note: spaCy model download failed: ", e$message)
      message("You can install manually: python -m spacy download en_core_web_sm")
      FALSE
    })

    message("\nSetup complete!")
    if (spacy_model_result) {
      message("- spaCy model: en_core_web_sm installed")
    }
    message("Restart R session to activate.")
    return(invisible(TRUE))

  }, error = function(e) {
    message("Failed to set up Python environment: ", e$message)
    invisible(FALSE)
  })
}


#' Check Python Environment Status
#'
#' @description
#' Checks if Python environment is available and properly configured.
#'
#' @param envname Character string name of the virtual environment
#'   (default: "textanalysisr-env")
#'
#' @return List with status information:
#'   - available: Logical, TRUE if environment exists
#'   - active: Logical, TRUE if environment is currently active
#'   - packages: List of installed package versions
#'
#' @keywords internal
#' @export
#'
#' @examples
#' if (interactive()) {
#' status <- check_python_env()
#' print(status)
#' }
check_python_env <- function(envname = "textanalysisr-env") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    message("Package 'reticulate' is required. Install with install.packages('reticulate').")
    return(list(available = FALSE, active = FALSE, packages = list()))
  }

  env_list <- reticulate::virtualenv_list()
  available <- envname %in% env_list

  if (!available) {
    return(list(
      available = FALSE,
      active = FALSE,
      packages = NULL,
      message = paste("Environment", envname, "not found. Run setup_python_env() to create it.")
    ))
  }

  tryCatch({
    reticulate::use_virtualenv(envname, required = TRUE)

    packages <- tryCatch({
      spacy_version <- reticulate::py_run_string("import spacy; print(spacy.__version__)")
      pdfplumber_check <- reticulate::py_run_string("import pdfplumber; print(pdfplumber.__version__)")

      list(
        spacy = spacy_version,
        pdfplumber = pdfplumber_check
      )
    }, error = function(e) NULL)

    return(list(
      available = TRUE,
      active = TRUE,
      packages = packages,
      message = "Python environment is ready"
    ))

  }, error = function(e) {
    return(list(
      available = TRUE,
      active = FALSE,
      packages = NULL,
      message = paste("Failed to activate environment:", e$message)
    ))
  })
}




# Cloud LLM API utilities (OpenAI, Gemini)

#' Call OpenAI Chat Completion API
#'
#' @description
#' Makes a chat completion request to OpenAI's API.
#'
#' @param system_prompt Character string with system instructions
#' @param user_prompt Character string with user message
#' @param model Character string specifying the model (default: "gpt-4.1-mini")
#' @param temperature Numeric temperature for response randomness (default: 0)
#' @param max_tokens Maximum number of tokens to generate (default: 150)
#' @param api_key Character string with OpenAI API key
#'
#' @return Character string with the model's response
#'
#' @concept ai
#' @keywords internal
#'
call_openai_chat <- function(system_prompt,
                              user_prompt,
                              model = "gpt-4.1-mini",
                              temperature = 0,
                              max_tokens = 150,
                              api_key) {

  if (!requireNamespace("httr", quietly = TRUE)) {
    stop("httr package is required for OpenAI API calls")
  }
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop("jsonlite package is required for OpenAI API calls")
  }

  body_list <- list(
    model = model,
    messages = list(
      list(role = "system", content = system_prompt),
      list(role = "user", content = user_prompt)
    ),
    temperature = temperature,
    max_tokens = max_tokens
  )

  response <- httr::POST(
    url = "https://api.openai.com/v1/chat/completions",
    httr::add_headers(
      `Authorization` = paste("Bearer", api_key)
    ),
    body = body_list,
    encode = "json"
  )

  if (httr::status_code(response) != 200) {
    .stop_api_error("openai", "chat", httr::status_code(response),
                    httr::content(response, "text", encoding = "UTF-8"))
  }

  res_json <- jsonlite::fromJSON(httr::content(response, "text", encoding = "UTF-8"))

  if (!is.null(res_json$choices) && length(res_json$choices) > 0) {
    return(res_json$choices$message$content[1])
  }

  stop("Unexpected response structure from OpenAI API")
}


#' Call Gemini Chat API
#'
#' @description
#' Makes a chat completion request to Google's Gemini API.
#'
#' @param system_prompt Character string with system instructions
#' @param user_prompt Character string with user message
#' @param model Character string specifying the Gemini model (default: "gemini-2.5-flash")
#' @param temperature Numeric temperature for response randomness (default: 0)
#' @param max_tokens Maximum number of tokens to generate (default: 150)
#' @param api_key Character string with Gemini API key
#'
#' @return Character string with the model's response
#'
#' @concept ai
#' @keywords internal
#'
call_gemini_chat <- function(system_prompt,
                              user_prompt,
                              model = "gemini-2.5-flash",
                              temperature = 0,
                              max_tokens = 8192,
                              api_key) {

  if (!requireNamespace("httr", quietly = TRUE)) {
    stop("httr package is required for Gemini API calls")
  }
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop("jsonlite package is required for Gemini API calls")
  }

  body_list <- list(
    systemInstruction = list(
      parts = list(
        list(text = system_prompt)
      )
    ),
    contents = list(
      list(
        role = "user",
        parts = list(
          list(text = user_prompt)
        )
      )
    ),
    generationConfig = list(
      temperature = temperature,
      maxOutputTokens = max_tokens,
      thinkingConfig = list(
        thinkingBudget = 1024
      )
    )
  )

  url <- paste0(
    "https://generativelanguage.googleapis.com/v1beta/models/",
    model, ":generateContent"
  )

  response <- httr::POST(
    url = url,
    httr::add_headers(
      `x-goog-api-key` = api_key
    ),
    body = body_list,
    encode = "json"
  )

  if (httr::status_code(response) != 200) {
    .stop_api_error("gemini", "chat", httr::status_code(response),
                    httr::content(response, "text", encoding = "UTF-8"))
  }

  res_json <- jsonlite::fromJSON(
    httr::content(response, "text", encoding = "UTF-8"),
    simplifyVector = FALSE
  )

  if (!is.null(res_json$candidates) && length(res_json$candidates) > 0) {
    parts <- res_json$candidates[[1]]$content$parts
    if (!is.null(parts) && length(parts) > 0) {
      answer_texts <- character(0)
      for (part in parts) {
        if (isTRUE(part$thought)) next
        if (!is.null(part$text) && nzchar(part$text)) {
          answer_texts <- c(answer_texts, part$text)
        }
      }
      if (length(answer_texts) > 0) {
        return(paste(answer_texts, collapse = "\n"))
      }
      return(parts[[length(parts)]]$text)
    }
  }

  stop("Unexpected response structure from Gemini API")
}


#' Call LLM API (Unified Wrapper)
#'
#' @description
#' Unified wrapper for calling different LLM providers (OpenAI, Gemini).
#' Automatically routes to the appropriate provider-specific function.
#'
#' @param provider Character string: "openai" or "gemini"
#' @param system_prompt Character string with system instructions
#' @param user_prompt Character string with user message
#' @param model Character string specifying the model (provider-specific defaults apply)
#' @param temperature Numeric temperature for response randomness (default: 0)
#' @param max_tokens Maximum number of tokens to generate (default: 150)
#' @param api_key Character string with API key (required for openai/gemini)
#'
#' @return Character string with the model's response
#'
#' @concept ai
#' @seealso [sanitize_llm_input()] to clean text before prompting; [get_best_embeddings()] for vector embeddings; [run_rag_search()] for RAG search (retrieval + generation)
#' @export
#'
#' @examples
#' if (interactive()) {
#' # Using OpenAI
#' response <- call_llm_api(
#'   provider = "openai",
#'   system_prompt = "You are a helpful assistant.",
#'   user_prompt = "Generate a topic label",
#'   api_key = Sys.getenv("OPENAI_API_KEY")
#' )
#'
#' # Using Gemini
#' response <- call_llm_api(
#'   provider = "gemini",
#'   system_prompt = "You are a helpful assistant.",
#'   user_prompt = "Generate a topic label",
#'   api_key = Sys.getenv("GEMINI_API_KEY")
#' )
#' }
call_llm_api <- function(provider = c("openai", "gemini"),
                         system_prompt,
                         user_prompt,
                         model = NULL,
                         temperature = 0,
                         max_tokens = 150,
                         api_key = NULL) {

  provider <- match.arg(provider)

  # Set default models based on provider
  if (is.null(model)) {
    model <- switch(provider,
      "openai" = "gpt-4.1-mini",
      "gemini" = "gemini-2.5-flash"
    )
  }

  if (is.null(api_key) || !nzchar(api_key)) {
    api_key <- switch(provider,
      "openai" = Sys.getenv("OPENAI_API_KEY"),
      "gemini" = Sys.getenv("GEMINI_API_KEY")
    )
  }

  if (!nzchar(api_key)) {
    return(.notify_missing_api_key(provider))
  }

  result <- switch(provider,
    "openai" = call_openai_chat(
      system_prompt = system_prompt,
      user_prompt = user_prompt,
      model = model,
      temperature = temperature,
      max_tokens = max_tokens,
      api_key = api_key
    ),
    "gemini" = call_gemini_chat(
      system_prompt = system_prompt,
      user_prompt = user_prompt,
      model = model,
      temperature = temperature,
      max_tokens = max_tokens,
      api_key = api_key
    )
  )

  return(result)
}


#' Describe Image with OpenAI Vision API
#'
#' @param image_base64 Character string of base64-encoded PNG image
#' @param prompt Character string describing what to extract
#' @param model Character string, OpenAI model name (default: "gpt-4.1")
#' @param max_tokens Integer, maximum tokens in response (default: 500)
#' @param api_key Character string, OpenAI API key
#'
#' @return Character string description, or NULL on failure
#' @keywords internal
describe_image_openai <- function(image_base64,
                                  prompt = "Describe this image: charts, diagrams, tables, and text. Extract visible text.",
                                  model = "gpt-4.1",
                                  max_tokens = 500,
                                  api_key) {
  if (!requireNamespace("httr", quietly = TRUE) ||
      !requireNamespace("jsonlite", quietly = TRUE)) {
    return(NULL)
  }

  tryCatch({
    body <- list(
      model = model,
      messages = list(
        list(
          role = "user",
          content = list(
            list(type = "text", text = prompt),
            list(
              type = "image_url",
              image_url = list(
                url = paste0("data:image/png;base64,", image_base64)
              )
            )
          )
        )
      ),
      max_tokens = max_tokens
    )

    response <- httr::POST(
      url = "https://api.openai.com/v1/chat/completions",
      httr::add_headers(
        Authorization = paste("Bearer", api_key),
        `Content-Type` = "application/json"
      ),
      body = jsonlite::toJSON(body, auto_unbox = TRUE),
      httr::timeout(120)
    )

    status <- httr::status_code(response)
    if (status != 200) {
      resp_text <- tryCatch(httr::content(response, "text", encoding = "UTF-8"), error = function(e) "")
      message(.format_api_error_message("openai", "vision", status, resp_text))
      return(NULL)
    }

    res_json <- jsonlite::fromJSON(httr::content(response, "text", encoding = "UTF-8"))
    result <- res_json$choices$message$content[[1]]
    if (is.null(result) || !nzchar(trimws(result))) return(NULL)
    return(trimws(result))
  }, error = function(e) {
    message("[Vision OpenAI] Error: ", e$message)
    NULL
  })
}


#' Describe Image with Gemini Vision API
#'
#' @param image_base64 Character string of base64-encoded PNG image
#' @param prompt Character string describing what to extract
#' @param model Character string, Gemini model name (default: "gemini-2.5-flash")
#' @param max_tokens Integer, maximum tokens in response (default: 500)
#' @param api_key Character string, Gemini API key
#'
#' @return Character string description, or NULL on failure
#' @keywords internal
describe_image_gemini <- function(image_base64,
                                  prompt = "Describe this image: charts, diagrams, tables, and text. Extract visible text.",
                                  model = "gemini-2.5-flash",
                                  max_tokens = 500,
                                  api_key) {
  if (!requireNamespace("httr", quietly = TRUE) ||
      !requireNamespace("jsonlite", quietly = TRUE)) {
    return(NULL)
  }

  tryCatch({
    body <- list(
      contents = list(
        list(
          parts = list(
            list(text = prompt),
            list(
              inline_data = list(
                mime_type = "image/png",
                data = image_base64
              )
            )
          )
        )
      ),
      generationConfig = list(
        maxOutputTokens = max_tokens
      )
    )

    url <- paste0(
      "https://generativelanguage.googleapis.com/v1beta/models/",
      model, ":generateContent"
    )

    response <- httr::POST(
      url = url,
      httr::add_headers(
        `Content-Type` = "application/json",
        `x-goog-api-key` = api_key
      ),
      body = jsonlite::toJSON(body, auto_unbox = TRUE),
      httr::timeout(120)
    )

    status <- httr::status_code(response)
    if (status != 200) {
      resp_text <- tryCatch(httr::content(response, "text", encoding = "UTF-8"), error = function(e) "")
      message(.format_api_error_message("gemini", "vision", status, resp_text))
      return(NULL)
    }

    res_json <- jsonlite::fromJSON(httr::content(response, "text", encoding = "UTF-8"))

    if (!is.null(res_json$candidates) && length(res_json$candidates) > 0) {
      parts <- tryCatch({
        if (is.data.frame(res_json$candidates)) {
          res_json$candidates$content$parts[[1]]
        } else {
          res_json$candidates[[1]]$content$parts
        }
      }, error = function(e) NULL)

      if (!is.null(parts) && length(parts) > 0) {
        result <- if (is.data.frame(parts)) parts$text[1] else parts[[1]]$text
        if (!is.null(result) && nzchar(trimws(result))) return(trimws(result))
      }
    }
    return(NULL)
  }, error = function(e) {
    message("[Vision Gemini] Error: ", e$message)
    NULL
  })
}


#' Describe Image Using Vision LLM
#'
#' @description
#' Unified dispatcher for image description using vision LLMs.
#' Routes to the appropriate provider (OpenAI or Gemini).
#'
#' @param image_base64 Character string of base64-encoded PNG image
#' @param provider Character: "openai" or "gemini"
#' @param model Character: Model name (uses provider default if NULL)
#' @param api_key Character: API key (required for openai/gemini)
#' @param prompt Character: Description prompt
#' @param timeout Numeric: Request timeout in seconds (default: 120)
#'
#' @return Character string description, or NULL on failure
#'
#' @concept ai
#' @export
describe_image <- function(image_base64,
                           provider = "gemini",
                           model = NULL,
                           api_key = NULL,
                           prompt = "Describe this image: charts, diagrams, tables, and text. Extract visible text.",
                           timeout = 120) {
  if (is.null(model)) {
    model <- switch(provider,
      "openai" = "gpt-4.1",
      "gemini" = "gemini-2.5-flash",
      "gemini-2.5-flash"
    )
  }

  switch(provider,
    "openai" = describe_image_openai(image_base64, prompt, model, api_key = api_key),
    "gemini" = describe_image_gemini(image_base64, prompt, model, api_key = api_key),
    NULL
  )
}


#' Get Embeddings from API
#'
#' @description
#' Generates text embeddings using OpenAI or Gemini embedding APIs.
#'
#' @param texts Character vector of texts to embed
#' @param provider Character string: "openai" or "gemini"
#' @param model Character string specifying the embedding model. Defaults:
#'   - openai: "text-embedding-3-small"
#'   - gemini: "gemini-embedding-001"
#' @param api_key Character string with API key
#' @param batch_size Integer, number of texts to embed per API call (default: 100)
#'
#' @return Matrix with embeddings (rows = texts, columns = dimensions)
#'
#' @concept ai
#' @keywords internal
#'
get_api_embeddings <- function(texts,
                           provider = c("openai", "gemini"),
                           model = NULL,
                           api_key = NULL,
                           batch_size = 100) {

  provider <- match.arg(provider)

  if (is.null(model)) {
    model <- switch(provider,
      "openai" = "text-embedding-3-small",
      "gemini" = "gemini-embedding-001"
    )
  }

  if (is.null(api_key) || !nzchar(api_key)) {
    api_key <- switch(provider,
      "openai" = Sys.getenv("OPENAI_API_KEY"),
      "gemini" = Sys.getenv("GEMINI_API_KEY")
    )
  }

  if (!nzchar(api_key)) {
    return(.notify_missing_api_key(provider))
  }

  if (!requireNamespace("httr", quietly = TRUE)) {
    stop("httr package is required for embedding API calls")
  }
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop("jsonlite package is required for embedding API calls")
  }

  n_texts <- length(texts)
  all_embeddings <- list()

  for (start in seq(1, n_texts, by = batch_size)) {
    end <- min(start + batch_size - 1, n_texts)
    batch_texts <- texts[start:end]

    embeddings <- switch(provider,
      "openai" = get_openai_embeddings(batch_texts, model, api_key),
      "gemini" = get_gemini_embeddings(batch_texts, model, api_key)
    )

    all_embeddings[[length(all_embeddings) + 1]] <- embeddings
  }

  do.call(rbind, all_embeddings)
}


#' Get OpenAI Embeddings (Internal)
#' @keywords internal
get_openai_embeddings <- function(texts, model, api_key) {
  body_list <- list(
    input = texts,
    model = model
  )

  response <- httr::POST(
    url = "https://api.openai.com/v1/embeddings",
    httr::add_headers(
      `Content-Type` = "application/json",
      `Authorization` = paste("Bearer", api_key)
    ),
    body = jsonlite::toJSON(body_list, auto_unbox = TRUE),
    encode = "json"
  )

  if (httr::status_code(response) != 200) {
    .stop_api_error("openai", "embeddings", httr::status_code(response),
                    httr::content(response, "text", encoding = "UTF-8"))
  }

  res_json <- jsonlite::fromJSON(httr::content(response, "text", encoding = "UTF-8"))

  if (!is.null(res_json$data)) {
    # Sort by index to ensure correct order
    embeddings <- res_json$data[order(res_json$data$index), ]
    do.call(rbind, embeddings$embedding)
  } else {
    stop("Unexpected response structure from OpenAI Embeddings API")
  }
}


#' Get Gemini Embeddings (Internal)
#' @keywords internal
get_gemini_embeddings <- function(texts, model, api_key) {
  # Gemini requires individual requests per text
  embeddings_list <- lapply(texts, function(text) {
    body_list <- list(
      model = paste0("models/", model),
      content = list(
        parts = list(
          list(text = text)
        )
      )
    )

    url <- paste0(
      "https://generativelanguage.googleapis.com/v1beta/models/",
      model, ":embedContent"
    )

    response <- httr::POST(
      url = url,
      httr::add_headers(
        `Content-Type` = "application/json",
        `x-goog-api-key` = api_key
      ),
      body = jsonlite::toJSON(body_list, auto_unbox = TRUE),
      encode = "json"
    )

    if (httr::status_code(response) != 200) {
      .stop_api_error("gemini", "embeddings", httr::status_code(response),
                      httr::content(response, "text", encoding = "UTF-8"))
    }

    res_json <- jsonlite::fromJSON(httr::content(response, "text", encoding = "UTF-8"))

    if (!is.null(res_json$embedding$values)) {
      return(res_json$embedding$values)
    } else {
      stop("Unexpected response structure from Gemini Embeddings API")
    }
  })

  do.call(rbind, embeddings_list)
}


#' Get Best Available Embeddings
#'
#' @description
#' Auto-detects and uses the best available embedding provider with the following priority:
#' 1. sentence-transformers (local Python) - if Python environment is set up
#' 2. OpenAI API - if OPENAI_API_KEY is set
#' 3. Gemini API - if GEMINI_API_KEY is set
#'
#' @param texts Character vector of texts to embed
#' @param provider Character string: "auto" (default), "sentence-transformers",
#'   "openai", or "gemini". Use "auto" for automatic detection.
#' @param model Character string specifying the embedding model. If NULL, uses default
#'   model for the selected provider.
#' @param api_key Optional API key for OpenAI or Gemini providers. If NULL, falls back
#'   to environment variables (OPENAI_API_KEY, GEMINI_API_KEY).
#' @param verbose Logical, whether to print progress messages (default: TRUE)
#'
#' @return Matrix with embeddings (rows = texts, columns = dimensions)
#'
#' @concept ai
#' @export
#'
#' @examples
#' if (interactive()) {
#' data(SpecialEduTech)
#' texts <- SpecialEduTech$abstract[1:5]
#'
#' # Auto-detect best available provider
#' embeddings <- get_best_embeddings(texts)
#'
#' # Force specific provider
#' embeddings <- get_best_embeddings(texts, provider = "openai")
#'
#' dim(embeddings)
#' }
get_best_embeddings <- function(texts,
                                 provider = "auto",
                                 model = NULL,
                                 api_key = NULL,
                                 verbose = TRUE) {

  if (!is.character(texts) || length(texts) == 0) {
    stop("texts must be a non-empty character vector")
  }

  if (provider == "auto") {
    if (check_feature("python")) {
      provider <- "sentence-transformers"
      if (verbose) message("Using sentence-transformers embeddings (local Python)")
    } else if ((!is.null(api_key) && nzchar(api_key)) || nzchar(Sys.getenv("OPENAI_API_KEY"))) {
      provider <- "openai"
      if (verbose) message("Using OpenAI embeddings (API)")
    } else if (nzchar(Sys.getenv("GEMINI_API_KEY"))) {
      provider <- "gemini"
      if (verbose) message("Using Gemini embeddings (API)")
    } else {
      message(
        "No embedding provider available. Options:\n",
        "  1. Set up Python with: TextAnalysisR::setup_python_env()\n",
        "  2. Set OPENAI_API_KEY or GEMINI_API_KEY environment variable"
      )
      return(invisible(NULL))
    }
  }

  embeddings <- switch(provider,
    "sentence-transformers" = {
      if (!check_feature("python")) {
        return(.notify_missing_python("Embedding generation via sentence-transformers"))
      }
      model <- model %||% "all-MiniLM-L6-v2"
      generate_embeddings(texts, model = model, verbose = verbose)
    },
    "openai" = {
      model <- model %||% "text-embedding-3-small"
      get_api_embeddings(texts, provider = "openai", model = model, api_key = api_key)
    },
    "gemini" = {
      model <- model %||% "gemini-embedding-001"
      get_api_embeddings(texts, provider = "gemini", model = model, api_key = api_key)
    },
    stop(paste0(
      "Unknown provider: ", provider, ". ",
      "Valid options: auto, sentence-transformers, openai, gemini"
    ))
  )

  embeddings
}


# Security and validation utilities

#' Cybersecurity Utility Functions
#'
#' @description Functions for input validation, sanitization, and security logging
#'
#' @section NIST Compliance:
#' This package follows NIST security standards (based on NIST SP 800-53):
#' - SC-8: Transmission Confidentiality and Integrity (HTTPS encryption)
#' - SC-28: Protection of Information at Rest (secure API key storage)
#' - IA-5: Authenticator Management (API key validation and format checking)
#' - AC-3: Access Enforcement (rate limiting, input validation, file type restrictions)
#' - SI-10: Information Input Validation (malicious content detection)
#' - AU-2: Audit Events (security logging and monitoring)

#' Validate File Upload
#'
#' @param file_info File info object from Shiny fileInput
#' @return TRUE if valid, stops with error message if invalid
#' @keywords internal
validate_file_upload <- function(file_info) {
  if (is.null(file_info)) {
    stop("No file provided")
  }

  allowed_extensions <- c(".csv", ".xlsx", ".txt", ".rds", ".pdf", ".docx")
  ext <- tools::file_ext(file_info$name)

  if (!paste0(".", tolower(ext)) %in% allowed_extensions) {
    stop("Invalid file type. Allowed types: CSV, XLSX, TXT, RDS, PDF, DOCX")
  }

  max_size <- 50 * 1024 * 1024
  if (file_info$size > max_size) {
    stop("File size exceeds maximum limit of 50MB")
  }

  if (ext %in% c("csv", "txt")) {
    content <- tryCatch({
      suppressWarnings(readLines(file_info$datapath, n = 10, warn = FALSE))
    }, error = function(e) {
      stop("Unable to read file contents")
    })

    suspicious_patterns <- c("<script", "javascript:", "onerror=", "onclick=")
    if (any(grepl(paste(suspicious_patterns, collapse = "|"), content, ignore.case = TRUE))) {
      stop("File contains potentially malicious content")
    }
  }

  return(TRUE)
}

#' Sanitize Text Input
#'
#' @param text Text input from user
#' @return Sanitized text
#' @keywords internal
sanitize_text_input <- function(text) {
  if (is.null(text) || nchar(text) == 0) {
    return(text)
  }

  text <- gsub("<script.*?>.*?</script>", "", text, ignore.case = TRUE)
  text <- gsub("javascript:", "", text, ignore.case = TRUE)
  text <- gsub("on\\w+\\s*=", "", text, ignore.case = TRUE)

  max_chars <- 1000000
  if (nchar(text) > max_chars) {
    stop("Text input exceeds maximum length of 1 million characters")
  }

  return(text)
}

#' Sanitize LLM Input
#'
#' @description
#' Sanitizes user input before inclusion in LLM prompts to mitigate prompt
#' injection attacks. Filters common injection patterns such as instruction
#' overrides, system prompt markers, and role-switching attempts.
#' Distinct from \code{sanitize_text_input()} which targets XSS.
#'
#' @param text Character string of user input destined for an LLM prompt
#' @param max_length Maximum allowed character length (default: 2000)
#' @return Sanitized character string
#'
#' @section NIST Compliance:
#' Implements NIST SI-10 (Information Input Validation) for AI/LLM contexts.
#'
#' @keywords internal
sanitize_llm_input <- function(text, max_length = 2000) {
  if (is.null(text) || !is.character(text) || !nzchar(text)) {
    return(text)
  }

  if (nchar(text) > max_length) {
    text <- substr(text, 1, max_length)
  }

  injection_patterns <- c(
    "ignore (all |any )?previous instructions",
    "ignore (all |any )?prior instructions",
    "disregard (all |any )?previous",
    "forget (all |any )?previous",
    "you are now",
    "act as if",
    "pretend you are",
    "new instructions:",
    "override:",
    "system prompt:",
    "\\[INST\\]",
    "\\[/INST\\]",
    "<<SYS>>",
    "<</SYS>>",
    "<\\|im_start\\|>",
    "<\\|im_end\\|>",
    "### (Human|Assistant|System):",
    "\\bBEGIN INSTRUCTION\\b",
    "\\bEND INSTRUCTION\\b"
  )

  combined_pattern <- paste(injection_patterns, collapse = "|")
  text <- gsub(combined_pattern, "", text, ignore.case = TRUE, perl = TRUE)

  text <- trimws(text)

  return(text)
}

#' Check Rate Limit
#'
#' @param session_token Shiny session token
#' @param user_requests Reactive value storing request history
#' @param max_requests Maximum requests allowed in time window
#' @param window_seconds Time window in seconds (default: 3600 = 1 hour)
#' @return TRUE if within limit, stops with error if exceeded
#' @keywords internal
check_rate_limit <- function(session_token, user_requests, max_requests = 100, window_seconds = 3600) {
  current_time <- Sys.time()
  requests <- user_requests()

  if (is.null(requests[[session_token]])) {
    requests[[session_token]] <- list()
  }

  requests[[session_token]] <- Filter(function(x) {
    difftime(current_time, x, units = "secs") < window_seconds
  }, requests[[session_token]])

  if (length(requests[[session_token]]) >= max_requests) {
    stop("Rate limit exceeded. You have made too many requests. Please wait before trying again.")
  }

  requests[[session_token]] <- c(requests[[session_token]], current_time)
  user_requests(requests)

  return(TRUE)
}

#' Log Security Event
#'
#' @param event_type Type of security event
#' @param details Additional details about the event
#' @param session_info Shiny session object
#' @param level Log level ("info", "warning", "error")
#' @keywords internal
log_security_event <- function(event_type, details, session_info, level = "info") {
  timestamp <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")

  session_token <- if (!is.null(session_info$token)) {
    substr(session_info$token, 1, 8)
  } else {
    "unknown"
  }

  log_message <- paste0(
    "[", timestamp, "] ",
    "[", level, "] ",
    "SECURITY: ", event_type, " | ",
    "Session: ", session_token, "... | ",
    "Details: ", details
  )

  log_dir <- Sys.getenv("TEXTANALYSISR_LOG_DIR", unset = "")
  if (nzchar(log_dir) && dir.exists(log_dir)) {
    log_file <- file.path(log_dir, "security.log")
    tryCatch({
      cat(log_message, "\n", file = log_file, append = TRUE)
    }, error = function(e) {
      warning("Failed to write security log: ", e$message)
    })
  }

  if (Sys.getenv("ENVIRONMENT") != "production") {
    message(log_message)
  }

  return(invisible(NULL))
}


.parse_provider_error <- function(provider, response_body) {
  tryCatch({
    parsed <- jsonlite::fromJSON(response_body)
    detail <- if (provider == "openai") {
      parsed$error$message
    } else if (provider == "gemini") {
      parsed$error$message %||% parsed$error$status
    }
    if (is.null(detail) || !nzchar(detail)) return(NULL)
    if (nchar(detail) > 300) detail <- paste0(substr(detail, 1, 297), "...")
    return(detail)
  }, error = function(e) NULL)
}


.format_api_error_message <- function(provider, endpoint, status_code, response_body) {
  status_meaning <- switch(as.character(status_code),
    "400" = "Bad request",
    "401" = "Authentication failed",
    "403" = "Forbidden",
    "404" = "Not found",
    "429" = "Rate limit exceeded",
    "500" = "Server error",
    "502" = "Bad gateway",
    "503" = "Service unavailable",
    paste0("HTTP error ", status_code)
  )

  action <- switch(as.character(status_code),
    "401" = {
      key_var <- if (provider == "openai") "OPENAI_API_KEY" else "GEMINI_API_KEY"
      paste0("Verify your ", key_var, ".")
    },
    "429" = "Wait and retry, or check your usage limits.",
    "500" = "Provider-side issue. Try again later.",
    "502" = "Provider-side issue. Try again later.",
    "503" = "Provider-side issue. Try again later.",
    "Check your request and try again."
  )

  provider_label <- if (provider == "openai") "OpenAI" else "Gemini"
  msg <- sprintf("%s %s API request failed (HTTP %d: %s).\n%s",
                 provider_label, endpoint, status_code, status_meaning, action)

  detail <- .parse_provider_error(provider, response_body)
  if (!is.null(detail)) {
    msg <- paste0(msg, "\nProvider message: ", detail)
  }

  msg
}


.stop_api_error <- function(provider, endpoint, status_code, response_body) {
  tryCatch(
    log_security_event("api_error", sprintf("%s %s status %d: %s",
      if (provider == "openai") "OpenAI" else "Gemini",
      endpoint, status_code, response_body), list(token = NULL), "error"),
    error = function(e) NULL
  )
  stop(.format_api_error_message(provider, endpoint, status_code, response_body),
       call. = FALSE)
}


#' @keywords internal
.ensure_python <- function(required_module = NULL, envname = NULL) {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    stop("Package 'reticulate' is required.")
  }
  if (!is.null(required_module) && !reticulate::py_module_available(required_module)) {
    stop(sprintf(
      "Python module '%s' not found. Run setup_python_env() or install with: pip install %s",
      required_module, required_module
    ))
  }
  invisible(TRUE)
}

.missing_api_key_message <- function(provider, context = "package") {
  env_var <- if (provider == "openai") "OPENAI_API_KEY" else "GEMINI_API_KEY"
  provider_label <- if (provider == "openai") "OpenAI" else "Gemini"

  if (context == "shiny") {
    return(sprintf("%s API key required. Enter in the API Key field or set %s in .Renviron.",
                   provider_label, env_var))
  }

  paste0(
    sprintf("No %s API key found. Provide your key using one of these methods:\n", provider_label),
    sprintf("  1. Sys.setenv(%s = \"your-key-here\")\n", env_var),
    sprintf("  2. Add %s=your-key-here to your .Renviron file\n", env_var),
    "  3. Pass directly via the api_key parameter"
  )
}

.notify_missing_api_key <- function(provider) {
  message(.missing_api_key_message(provider, "package"))
  invisible(NULL)
}

.notify_missing_python <- function(feature = "this feature") {
  hint <- if (!requireNamespace("reticulate", quietly = TRUE)) {
    "Install reticulate (install.packages('reticulate')), then run TextAnalysisR::setup_python_env()."
  } else {
    "Run TextAnalysisR::setup_python_env() to install Python and required modules."
  }
  message(sprintf("%s requires Python. %s", feature, hint))
  invisible(NULL)
}


#' Validate API Key Format
#'
#' @description
#' Validates API key format for OpenAI or Gemini according to NIST IA-5(1).
#' Auto-detects provider from key prefix and validates format requirements.
#'
#' @param api_key Character string containing the API key
#' @param strict Logical, if TRUE performs additional validation checks
#'
#' @return List with valid (logical), provider (character), and error (character if invalid)
#' @keywords internal
#'
#' @section NIST Compliance:
#' Implements NIST IA-5(1): Authenticator Management - Password-Based Authentication.
#' Validates format, length, and character composition to prevent weak or malformed keys.
#'
validate_api_key <- function(api_key, strict = TRUE) {
  if (is.null(api_key) || !is.character(api_key) || !nzchar(api_key)) {
    return(list(valid = FALSE, provider = NULL, error = "API key is NULL, empty, or not a character string"))
  }

  # Detect provider from prefix
  if (grepl("^sk-", api_key)) {
    provider <- "openai"
    min_length <- 40
    length_msg <- "OpenAI keys are typically 48+ characters"
  } else if (grepl("^AIza", api_key)) {
    provider <- "gemini"
    min_length <- 39
    length_msg <- "Gemini keys are typically 39 characters"
  } else {
    return(list(
      valid = FALSE,
      provider = NULL,
      error = "Unknown API key format. OpenAI keys start with 'sk-', Gemini keys start with 'AIza'"
    ))
  }

  if (nchar(api_key) < min_length) {
    return(list(valid = FALSE, provider = provider, error = paste("API key appears too short:", length_msg)))
  }

  if (strict) {
    if (grepl("\\s", api_key)) {
      return(list(valid = FALSE, provider = provider, error = "API key contains whitespace characters"))
    }

    if (grepl("[^A-Za-z0-9_-]", api_key)) {
      return(list(valid = FALSE, provider = provider, error = "API key contains unexpected special characters"))
    }
  }

  return(list(valid = TRUE, provider = provider, error = NULL))
}

#' Validate Column Name
#'
#' @description
#' Validates column names to prevent code injection through formula construction.
#' Ensures column names follow R naming conventions and contain no malicious patterns.
#'
#' @param col_name Character string containing the column name
#'
#' @return TRUE if valid, stops with error if invalid
#' @keywords internal
#'
#' @section Security:
#' Protects against formula injection attacks where malicious column names could
#' execute arbitrary code when used in model formulas. Part of NIST SI-10 input validation.
#'
validate_column_name <- function(col_name) {
  if (is.null(col_name) || !is.character(col_name) || !nzchar(col_name)) {
    stop("Column name is NULL, empty, or not a character string")
  }

  if (length(col_name) != 1) {
    stop("Column name must be a single value, not a vector")
  }

  if (!grepl("^[A-Za-z][A-Za-z0-9_\\.]*$", col_name)) {
    stop("Invalid column name format. Column names must start with a letter and contain only letters, numbers, underscores, or periods.")
  }

  if (nchar(col_name) > 255) {
    stop("Column name exceeds maximum length of 255 characters")
  }

  backticks <- grepl("`", col_name, fixed = TRUE)
  if (backticks) {
    stop("Column name contains backticks which are not allowed")
  }

  return(TRUE)
}


# Web accessibility utilities

#' Web Accessibility Utility Functions
#'
#' @description Functions for ensuring WCAG 2.1 Level AA compliance in the Shiny application
#'
#' @section WCAG 2.1 Level AA Compliance:
#' This package follows Web Content Accessibility Guidelines (WCAG) 2.1 Level AA:
#' - 1.1.1 Non-text Content (Level A): Alt text for images and visualizations
#' - 1.4.3 Contrast Minimum (Level AA): 4.5:1 ratio for normal text, 3:1 for large text/UI
#' - 2.1.1 Keyboard (Level A): Full keyboard navigation support
#' - 2.4.1 Bypass Blocks (Level A): Skip navigation links
#' - 3.1.1 Language of Page (Level A): Page language identification
#' - 4.1.2 Name, Role, Value (Level A): ARIA labels and roles

#' Calculate Color Contrast Ratio
#'
#' @description
#' Calculates the contrast ratio between two colors according to WCAG 2.1 standards
#' using the relative luminance formula from W3C guidelines.
#' Used to verify text/background color combinations meet accessibility requirements.
#'
#' @param foreground Foreground color (hex format, e.g., "#111827")
#' @param background Background color (hex format, e.g., "#ffffff")
#'
#' @return Numeric contrast ratio (1-21)
#' @keywords internal
#'
#' @section WCAG Requirements:
#' - Normal text: Minimum 4.5:1 (Level AA)
#' - Large text (18pt+ or 14pt+ bold): Minimum 3:1 (Level AA)
#' - UI components and graphics: Minimum 3:1 (Level AA)
#'
calculate_contrast_ratio <- function(foreground, background) {
  l1 <- .relative_luminance(.hex_to_rgb(foreground))
  l2 <- .relative_luminance(.hex_to_rgb(background))
  lighter <- max(l1, l2)
  darker <- min(l1, l2)
  round((lighter + 0.05) / (darker + 0.05), 2)
}

.hex_to_rgb <- function(hex) {
  hex <- gsub("#", "", hex)
  c(
    strtoi(substr(hex, 1, 2), 16L),
    strtoi(substr(hex, 3, 4), 16L),
    strtoi(substr(hex, 5, 6), 16L)
  ) / 255
}

.relative_luminance <- function(rgb) {
  rgb <- ifelse(rgb <= 0.03928, rgb / 12.92, ((rgb + 0.055) / 1.055)^2.4)
  0.2126 * rgb[1] + 0.7152 * rgb[2] + 0.0722 * rgb[3]
}

#' Check WCAG Contrast Compliance
#'
#' @description
#' Validates if color combination meets WCAG 2.1 Level AA contrast requirements.
#'
#' @param foreground Foreground color (hex format)
#' @param background Background color (hex format)
#' @param large_text Logical, TRUE if text is large (18pt+ or 14pt+ bold)
#'
#' @return Logical TRUE if compliant, FALSE if not
#' @keywords internal
#'
check_wcag_contrast <- function(foreground, background, large_text = FALSE) {
  ratio <- calculate_contrast_ratio(foreground, background)
  min_ratio <- if (large_text) 3.0 else 4.5

  if (ratio >= min_ratio) {
    return(TRUE)
  } else {
    warning(
      "WCAG contrast failure: ", ratio, ":1 ratio (requires ", min_ratio, ":1)\n",
      "  Foreground: ", foreground, "\n",
      "  Background: ", background
    )
    return(FALSE)
  }
}

#' Generate ARIA Label
#'
#' @description
#' Creates accessible ARIA label for UI elements.
#'
#' @param element_type Type of element (e.g., "button", "input", "plot")
#' @param action Action or purpose (e.g., "analyze", "download", "visualize")
#' @param context Additional context (optional)
#'
#' @return Character string with ARIA label
#' @keywords internal
#'
create_aria_label <- function(element_type, action, context = NULL) {
  if (!is.null(context)) {
    label <- paste(tools::toTitleCase(action), context, element_type)
  } else {
    label <- paste(tools::toTitleCase(action), element_type)
  }
  return(label)
}

#' Create Screen Reader Text
#'
#' @description
#' Generates visually hidden text for screen readers (WCAG 4.1.2).
#'
#' @param text Text to be read by screen readers
#'
#' @return HTML span with sr-only class
#' @keywords internal
#'
create_sr_text <- function(text) {
  return(
    paste0(
      "<span class=\"sr-only\" role=\"status\" aria-live=\"polite\">",
      text,
      "</span>"
    )
  )
}

#' Validate Keyboard Navigation
#'
#' @description
#' Checks if interactive elements have proper tabindex and keyboard handlers.
#' Used for WCAG 2.1.1 (Keyboard) compliance.
#'
#' @param tabindex Integer, tab order (-1 for no tab, 0 for natural order, 1+ for specific order)
#'
#' @return Logical TRUE if valid, FALSE with warning if invalid
#' @keywords internal
#'
validate_keyboard_navigation <- function(tabindex = 0) {
  if (!is.numeric(tabindex)) {
    warning("Tabindex must be numeric")
    return(FALSE)
  }

  if (tabindex > 100) {
    warning("Tabindex > 100 creates unpredictable tab order (WCAG 2.1.1)")
    return(FALSE)
  }

  return(TRUE)
}

#' Check Alt Text Presence
#'
#' @description
#' Validates that images and visualizations have alternative text descriptions.
#' Required for WCAG 1.1.1 (Non-text Content).
#'
#' Note: Decorative images should use empty alt text (alt="") to indicate
#' they should be ignored by assistive technology.
#'
#' @param alt_text Alternative text description
#' @param element_type Type of element (e.g., "plot", "image", "icon")
#' @param decorative Logical, TRUE if element is purely decorative
#'
#' @return Logical TRUE if valid, FALSE with warning if missing/inadequate
#' @keywords internal
#'
check_alt_text <- function(alt_text, element_type = "image", decorative = FALSE) {
  if (decorative) {
    return(TRUE)
  }

  if (is.null(alt_text) || !nzchar(alt_text)) {
    warning("Missing alt text for ", element_type, " (WCAG 1.1.1)")
    return(FALSE)
  }

  if (nchar(alt_text) < 10) {
    warning("Alt text too short for ", element_type, " (consider more descriptive text)")
    return(FALSE)
  }

  return(TRUE)
}


# Plot helper functions

#' Apply Standard Plotly Layout
#'
#' @description
#' Applies consistent layout styling to plotly plots following TextAnalysisR design standards.
#' This ensures all plots have uniform fonts, colors, margins, and interactive features.
#'
#' @param plot A plotly plot object
#' @param title Plot title text (optional)
#' @param xaxis_title X-axis title (optional)
#' @param yaxis_title Y-axis title (optional)
#' @param margin List of margins: list(t, b, l, r) in pixels (default: list(t = 60, b = 80, l = 80, r = 40))
#' @param show_legend Logical, whether to show legend (default: FALSE)
#'
#' @return A plotly plot object with standardized layout
#'
#' @details
#' Design standards applied:
#' - Title: 14px Roboto, #0c1f4a
#' - Axis titles: 13px Roboto, #0c1f4a
#' - Axis tick labels: 12px Roboto, #3B3B3B
#' @keywords internal
apply_standard_plotly_layout <- function(plot,
                                         title = NULL,
                                         xaxis_title = NULL,
                                         yaxis_title = NULL,
                                         margin = list(t = 60, b = 80, l = 80, r = 40),
                                         show_legend = FALSE) {

  if (!requireNamespace("plotly", quietly = TRUE)) {
    stop("Package 'plotly' is required. Please install it.")
  }

  layout_config <- list(
    font = list(family = "Roboto, sans-serif", size = 16, color = "#3B3B3B"),
    hoverlabel = list(
      font = list(size = 16, family = "Roboto, sans-serif"),
      align = "left"
    ),
    margin = margin,
    showlegend = show_legend,
    xaxis = list(
      tickfont = list(size = 16, color = "#3B3B3B", family = "Roboto, sans-serif"),
      titlefont = list(size = 16, color = "#0c1f4a", family = "Roboto, sans-serif")
    ),
    yaxis = list(
      tickfont = list(size = 16, color = "#3B3B3B", family = "Roboto, sans-serif"),
      titlefont = list(size = 16, color = "#0c1f4a", family = "Roboto, sans-serif")
    )
  )

  if (!is.null(title)) {
    layout_config$title <- list(
      text = title,
      font = list(size = 18, color = "#0c1f4a", family = "Roboto, sans-serif")
    )
  }

  if (!is.null(xaxis_title)) {
    layout_config$xaxis$title <- list(
      text = xaxis_title,
      font = list(size = 16, color = "#0c1f4a", family = "Roboto, sans-serif")
    )
  }

  if (!is.null(yaxis_title)) {
    layout_config$yaxis$title <- list(
      text = yaxis_title,
      font = list(size = 16, color = "#0c1f4a", family = "Roboto, sans-serif")
    )
  }

  do.call(plotly::layout, c(list(p = plot), layout_config)) %>%
    plotly::config(displayModeBar = TRUE)
}


#' Plotly hover tooltip config
#' @param bgcolor Tooltip background color.
#' @param fontcolor Tooltip text color.
#' @return A list for `plotly::layout(hoverlabel = ...)`.
#' @keywords internal
get_plotly_hover_config <- function(bgcolor = "#ffffff", fontcolor = "#0c1f4a") {
  list(
    bgcolor = bgcolor,
    bordercolor = bgcolor,
    font = list(
      family = "Roboto, sans-serif",
      size = 16,
      color = fontcolor
    ),
    align = "left",
    namelength = -1
  )
}


#' Create Standard ggplot2 Theme
#'
#' @description
#' Returns a standardized ggplot2 theme matching TextAnalysisR design standards.
#'
#' @param base_size Base font size (default: 14)
#'
#' @return A ggplot2 theme object
#'
#' @concept visualization
#' @keywords internal
#'
create_standard_ggplot_theme <- function(base_size = 11) {

  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required. Please install it.")
  }

  ggplot2::theme_minimal(base_size = base_size) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(
        size = 13,
        color = "#0c1f4a",
        hjust = 0.5
      ),
      axis.title = ggplot2::element_text(
        size = 12,
        color = "#0c1f4a"
      ),
      axis.text = ggplot2::element_text(
        size = 11,
        color = "#3B3B3B"
      ),
      strip.text = ggplot2::element_text(
        size = 11,
        color = "#0c1f4a"
      ),
      legend.text = ggplot2::element_text(
        size = 11,
        color = "#3B3B3B"
      ),
      legend.title = ggplot2::element_text(
        size = 12,
        color = "#0c1f4a"
      )
    )
}


#' Get Sentiment Color Palette
#'
#' @description
#' Returns standardized color mapping for sentiment analysis.
#'
#' @return Named vector of colors
#'
#' @concept visualization
#' @keywords internal
#' @export
get_sentiment_colors <- function() {
  c(
    "positive" = "#10B981",
    "negative" = "#EF4444",
    "neutral" = "#6B7280"
  )
}


#' Generate Sentiment Color Gradient
#'
#' @description
#' Generates a color based on sentiment score using a gradient from red (negative)
#' through gray (neutral) to green (positive).
#'
#' @param score Numeric sentiment score (typically -1 to 1)
#'
#' @return Hex color string
#'
#' @concept visualization
#' @keywords internal
#' @export
#'
#' @examples
#' get_sentiment_color(-0.8)  # Red
#' get_sentiment_color(0)     # Gray
#' get_sentiment_color(0.8)   # Green
get_sentiment_color <- function(score) {
  normalized_score <- (score + 1) / 2
  normalized_score <- pmax(0, pmin(1, normalized_score))

  if (normalized_score < 0.5) {
    t <- normalized_score * 2
    r <- round(185 * (1 - t) + 75 * t)
    g <- round(67 * (1 - t) + 181 * t)
    b <- round(68 * (1 - t) + 67 * t)
  } else {
    t <- (normalized_score - 0.5) * 2
    r <- round(75 * (1 - t) + 16 * t)
    g <- round(181 * (1 - t) + 185 * t)
    b <- round(67 * (1 - t) + 129 * t)
  }

  sprintf("#%02X%02X%02X", r, g, b)
}


#' Single-cell message DT table
#' @param message Character message to display.
#' @param font_size CSS font size.
#' @param color CSS color.
#' @return A `DT::datatable` htmlwidget.
#' @keywords internal
create_message_table <- function(message,
                                 font_size = "16px",
                                 color = "#6c757d") {

  if (!requireNamespace("DT", quietly = TRUE)) {
    stop("Package 'DT' is required. Please install it.")
  }

  DT::datatable(
    data.frame(Message = message),
    rownames = FALSE,
    options = list(
      dom = "t",
      ordering = FALSE,
      columnDefs = list(
        list(className = "dt-center", targets = "_all")
      ),
      initComplete = htmlwidgets::JS(
        sprintf(
          "function(settings, json) {
            $(this.api().table().container()).find('td').css({
              'font-size': '%s',
              'color': '%s',
              'padding': '40px',
              'text-align': 'center'
            });
          }",
          font_size,
          color
        )
      )
    ),
    class = "cell-border stripe"
  )
}


#' Empty plotly placeholder
#' @param message Message to display.
#' @param color Text color.
#' @param font_size Font size in px.
#' @return A `plotly` htmlwidget.
#' @keywords internal
create_empty_plot_message <- function(message,
                                       color = "#6B7280",
                                       font_size = 16) {

  if (!requireNamespace("plotly", quietly = TRUE)) {
    stop("Package 'plotly' is required. Please install it.")
  }

  plotly::plot_ly(type = "scatter", mode = "markers") %>%
    plotly::layout(
      xaxis = list(
        showgrid = FALSE,
        zeroline = FALSE,
        showticklabels = FALSE,
        title = ""
      ),
      yaxis = list(
        showgrid = FALSE,
        zeroline = FALSE,
        showticklabels = FALSE,
        title = ""
      ),
      annotations = list(
        list(
          text = message,
          x = 0.5,
          y = 0.5,
          xref = "paper",
          yref = "paper",
          showarrow = FALSE,
          font = list(
            size = font_size,
            color = color,
            family = "Roboto, sans-serif"
          )
        )
      )
    ) %>%
    plotly::config(displayModeBar = FALSE)
}


#' Default DT datatable options
#' @param scroll_y CSS height for vertical scrolling.
#' @param page_length Default page length.
#' @param show_buttons Include copy/csv/excel/pdf/print buttons.
#' @return A list for `DT::datatable(options = ...)`.
#' @keywords internal
get_dt_options <- function(scroll_y = "400px",
                            page_length = 25,
                            show_buttons = TRUE) {
  opts <- list(
    scrollX = TRUE,
    scrollY = scroll_y,
    pageLength = page_length
  )

  if (show_buttons) {
    opts$dom <- "Bfrtip"
    opts$buttons <- c("copy", "csv", "excel", "pdf", "print")
  }

  opts
}


# Shiny UI helper functions

#' Show persistent loading notification
#' @param message Notification text.
#' @param id Optional unique id for later removal.
#' @param session Shiny session. Defaults to the current reactive domain.
#' @return Invisibly the notification id.
#' @keywords internal
show_loading_notification <- function(message, id = NULL, session = NULL) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  if (is.null(session)) session <- shiny::getDefaultReactiveDomain()
  if (is.null(session)) return(invisible(NULL))

  tryCatch(
    shiny::showNotification(message, type = "message", duration = NULL, id = id, session = session),
    error = function(e) NULL
  )

  invisible(NULL)
}

#' Show completion notification
#' @param message Notification text.
#' @param duration Seconds until auto-dismiss.
#' @param session Shiny session. Defaults to the current reactive domain.
#' @return Invisibly the notification id.
#' @keywords internal
show_completion_notification <- function(message, duration = 5, session = NULL) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  if (is.null(session)) session <- shiny::getDefaultReactiveDomain()
  if (is.null(session)) return(invisible(NULL))

  tryCatch(
    shiny::showNotification(message, type = "message", duration = duration, session = session),
    error = function(e) NULL
  )

  invisible(NULL)
}

#' Show error notification
#' @param message Notification text.
#' @param duration Seconds until auto-dismiss.
#' @param session Shiny session. Defaults to the current reactive domain.
#' @return Invisibly the notification id.
#' @keywords internal
show_error_notification <- function(message, duration = 7, session = NULL) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  if (is.null(session)) session <- shiny::getDefaultReactiveDomain()
  if (is.null(session)) return(invisible(NULL))

  tryCatch(
    shiny::showNotification(message, type = "error", duration = duration, session = session),
    error = function(e) NULL
  )

  invisible(NULL)
}

#' Show warning notification
#' @param message Notification text.
#' @param duration Seconds until auto-dismiss.
#' @param session Shiny session. Defaults to the current reactive domain.
#' @return Invisibly the notification id.
#' @keywords internal
show_warning_notification <- function(message, duration = 5, session = NULL) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  if (is.null(session)) session <- shiny::getDefaultReactiveDomain()
  if (is.null(session)) return(invisible(NULL))

  tryCatch(
    shiny::showNotification(message, type = "warning", duration = duration, session = session),
    error = function(e) NULL
  )

  invisible(NULL)
}

#' Remove Shiny notification by id
#' @param id Notification id.
#' @param session Shiny session. Defaults to the current reactive domain.
#' @return Invisibly `NULL`.
#' @keywords internal
remove_notification_by_id <- function(id, session = NULL) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  if (is.null(session)) session <- shiny::getDefaultReactiveDomain()
  if (is.null(session)) return(invisible(NULL))

  tryCatch(
    shiny::removeNotification(id, session = session),
    error = function(e) NULL
  )

  invisible(NULL)
}

#' Show no-DFM notification
#' @param feature_name Feature description in the message.
#' @param duration Seconds until auto-dismiss.
#' @return Invisibly the notification id.
#' @keywords internal
show_no_dfm_notification <- function(feature_name = "this feature", duration = 7) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  message <- paste0(
    "No document-feature matrix available. ",
    "Please complete preprocessing (at least Step 4: DFM) first."
  )

  shiny::showNotification(
    message,
    type = "error",
    duration = duration
  )

  invisible(NULL)
}

#' Show no-feature-matrix notification
#' @param duration Seconds until auto-dismiss.
#' @return Invisibly the notification id.
#' @keywords internal
show_no_feature_matrix_notification <- function(duration = 7) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  shiny::showNotification(
    "No feature matrix available. Please complete preprocessing (at least Step 4: DFM) first.",
    type = "error",
    duration = duration
  )

  invisible(NULL)
}

#' Show united-texts-required notification
#' @param duration Seconds until auto-dismiss.
#' @return Invisibly the notification id.
#' @keywords internal
show_unite_texts_required_notification <- function(duration = 5) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  shiny::showNotification(
    "Please create united texts first in the preprocessing steps (Step 1: Unite texts).",
    type = "error",
    duration = duration
  )

  invisible(NULL)
}

#' Show guide modal
#' @param guide_name Guide identifier (filename stem).
#' @param title Modal title.
#' @param size Modal size ("s", "m", "l", "xl").
#' @return Invisibly `NULL`.
#' @keywords internal
show_guide_modal <- function(guide_name, title, size = "l") {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  if (!requireNamespace("htmltools", quietly = TRUE)) {
    stop("The 'htmltools' package is required for this function.")
  }

  # strip any directory components from guide_name to block path traversal
  guide_name <- basename(as.character(guide_name))
  if (!grepl("^[A-Za-z0-9_-]+$", guide_name)) {
    stop("Invalid guide_name; allowed: letters, digits, underscore, hyphen.")
  }

  guide_path <- system.file(
    "TextAnalysisR.app", "markdown", "guides",
    paste0(guide_name, ".html"),
    package = "TextAnalysisR"
  )

  if (!file.exists(guide_path) || guide_path == "") {
    guide_path <- file.path(
      "inst", "TextAnalysisR.app", "markdown", "guides",
      paste0(guide_name, ".html")
    )
  }

  if (!file.exists(guide_path)) {
    guide_path <- file.path(
      "markdown", "guides",
      paste0(guide_name, ".html")
    )
  }

  if (file.exists(guide_path)) {
    content <- htmltools::HTML(paste(readLines(guide_path, warn = FALSE), collapse = "\n"))
  } else {
    content <- htmltools::tags$p(
      paste0("Guide content not found: ", guide_name, ".html"),
      style = "color: #DC2626;"
    )
  }

  shiny::showModal(
    shiny::modalDialog(
      title = title,
      size = size,
      content,
      footer = shiny::modalButton("Close"),
      easyClose = TRUE
    )
  )

  invisible(NULL)
}

#' Show DFM-required modal
#' @param feature_name Feature description in the message.
#' @param additional_message Optional extra context appended to the body.
#' @return Invisibly `NULL`.
#' @keywords internal
show_dfm_required_modal <- function(feature_name = "this feature", additional_message = NULL) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  message_content <- list(
    shiny::p("No document-feature matrix (DFM) found."),
    shiny::p("Please complete the required preprocessing steps:")
  )

  if (!is.null(additional_message)) {
    message_content <- c(message_content, list(shiny::p(additional_message)))
  }

  message_content <- c(
    message_content,
    list(
      shiny::tags$div(
        style = "margin-left: 20px; margin-top: 10px;",
        shiny::tags$p(
          shiny::tags$strong(style = "color: #DC2626;", "Required:"),
          style = "margin-bottom: 5px;"
        ),
        shiny::tags$ul(
          shiny::tags$li(shiny::tags$strong("Step 1:"), " Unite Texts"),
          shiny::tags$li(shiny::tags$strong("Step 4:"), " Document-Feature Matrix (DFM)")
        ),
        shiny::tags$p(
          shiny::tags$strong(style = "color: #6B7280;", "Optional:"),
          " Steps 2, 3, 5, and 6",
          style = "margin-top: 10px; font-size: 12px;"
        )
      )
    )
  )

  shiny::showModal(
    shiny::modalDialog(
      title = "Preprocessing Required",
      message_content,
      easyClose = TRUE,
      footer = shiny::modalButton("OK")
    )
  )

  invisible(NULL)
}

#' Show preprocessing-steps modal
#' @param title Modal title.
#' @param message Lead message.
#' @param required_steps Character vector of required step labels.
#' @param optional_steps Character vector of optional step labels.
#' @param additional_note Optional note appended to the body.
#' @return Invisibly `NULL`.
#' @keywords internal
show_preprocessing_steps_modal <- function(title = "Preprocessing Required",
                                          message,
                                          required_steps,
                                          optional_steps = NULL,
                                          additional_note = NULL) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  content <- list(shiny::p(message))

  steps_div <- shiny::tags$div(
    style = "margin-left: 20px; margin-top: 10px;",
    shiny::tags$p(
      shiny::tags$strong(style = "color: #DC2626;", "Required:"),
      style = "margin-bottom: 5px;"
    ),
    shiny::tags$ul(
      lapply(required_steps, function(step) shiny::tags$li(step))
    )
  )

  if (!is.null(optional_steps)) {
    steps_div <- shiny::tagAppendChild(
      steps_div,
      shiny::tags$p(
        shiny::tags$strong(style = "color: #6B7280;", "Optional:"),
        paste(optional_steps, collapse = ", "),
        style = "margin-top: 10px; font-size: 12px;"
      )
    )
  }

  content <- c(content, list(steps_div))

  if (!is.null(additional_note)) {
    content <- c(content, list(shiny::p(additional_note, style = "margin-top: 10px; font-size: 12px; color: #6B7280;")))
  }

  shiny::showModal(
    shiny::modalDialog(
      title = title,
      content,
      easyClose = TRUE,
      footer = shiny::modalButton("OK")
    )
  )

  invisible(NULL)
}

#' Generate DFM Setup Instructions Text
#'
#' @description
#' Generates standardized text instructions for creating a DFM.
#' Used in console output or verbatim text displays.
#'
#' @param feature_name Name of the feature requiring DFM (default: "this feature")
#'
#' @return Character vector of instruction lines
#'
#' @keywords internal
#'
get_dfm_setup_instructions <- function(feature_name = "this feature") {
  c(
    "Warning: DFM Processing Required\n",
    "Please complete the following steps first:\n",
    "1. Go to the 'Preprocess' tab",
    "2. Navigate to Step 4: Document-Feature Matrix",
    "3. Click the 'Process' button\n",
    paste0("Once the DFM is created, you can return here to use ", feature_name, ".")
  )
}

#' Show DFM instructions modal
#' @param output_id Output id to reset after modal closes.
#' @param feature_name Feature description in the instructions.
#' @param session Shiny session.
#' @return Invisibly `NULL`.
#' @keywords internal
show_dfm_instructions_modal <- function(output_id, feature_name = "this feature", session = NULL) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  if (is.null(session)) {
    session <- shiny::getDefaultReactiveDomain()
  }

  shiny::showModal(
    shiny::modalDialog(
      title = "DFM Required",
      shiny::verbatimTextOutput(output_id),
      easyClose = TRUE,
      footer = shiny::modalButton("Close")
    )
  )

  invisible(NULL)
}

#' Show preprocessing-required modal
#' @param message Modal body message.
#' @param title Modal title.
#' @return Invisibly `NULL`.
#' @keywords internal
show_preprocessing_required_modal <- function(message = "Please complete preprocessing steps first.",
                                             title = "Preprocessing Required") {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop("The 'shiny' package is required for this function.")
  }

  shiny::showModal(
    shiny::modalDialog(
      title = title,
      shiny::p(message),
      easyClose = TRUE,
      footer = shiny::modalButton("OK")
    )
  )

  invisible(NULL)
}


# Text Formatting Utilities ----

#' Truncate Text with Ellipsis
#'
#' @description Truncates text to a maximum number of characters and adds
#'   ellipsis if truncated.
#'
#' @param text Character string to truncate.
#' @param max_chars Maximum number of characters (default: 50).
#'
#' @return Truncated text with "..." appended if truncated.
#'
#' @concept text-utilities
#' @keywords internal
truncate_text_with_ellipsis <- function(text, max_chars = 50) {
  text <- as.character(text)
  if (nchar(text) <= max_chars) {
    return(text)
  }
  paste0(substr(text, 1, max_chars), "...")
}

#' Truncate Text to Word Count
#'
#' @description Truncates text to a maximum number of words and adds
#'   ellipsis if truncated.
#'
#' @param text Character string to truncate.
#' @param max_words Maximum number of words (default: 150).
#'
#' @return Truncated text with "..." appended if truncated.
#'
#' @concept text-utilities
#' @keywords internal
truncate_text_to_words <- function(text, max_words = 150) {
  text <- as.character(text)
  words <- strsplit(text, "\\s+")[[1]]

  if (length(words) > max_words) {
    truncated_text <- paste(words[1:max_words], collapse = " ")
    return(paste0(truncated_text, "..."))
  } else {
    return(text)
  }
}

#' Wrap Long Text with Line Breaks
#'
#' @description Wraps long text by inserting line breaks at word boundaries.
#'   Handles both spaced text and continuous text (like URLs).
#'
#' @param text Character string to wrap.
#' @param chars_per_line Maximum characters per line (default: 50).
#' @param max_lines Maximum number of lines (default: 3).
#'
#' @return Text with line breaks inserted.
#'
#' @concept text-utilities
#' @keywords internal
wrap_long_text <- function(text, chars_per_line = 50, max_lines = 3) {
  text <- as.character(text)

  if (nchar(text) <= chars_per_line) return(text)

  # Handle text without spaces (like URLs)
  if (!grepl(" ", text) && nchar(text) > chars_per_line) {
    lines <- character()
    remaining_text <- text
    while (nchar(remaining_text) > chars_per_line && length(lines) < max_lines - 1) {
      lines <- c(lines, substr(remaining_text, 1, chars_per_line))
      remaining_text <- substr(remaining_text, chars_per_line + 1, nchar(remaining_text))
    }
    if (nchar(remaining_text) > 0) {
      if (nchar(remaining_text) > chars_per_line) {
        lines <- c(lines, paste0(substr(remaining_text, 1, chars_per_line - 3), "..."))
      } else {
        lines <- c(lines, remaining_text)
      }
    }
    return(paste(lines, collapse = "\n"))
  }

  # Handle normal text with spaces
  words <- strsplit(text, " ")[[1]]
  lines <- character()
  current_line <- ""

  for (word in words) {
    if (length(lines) >= max_lines - 1 && current_line != "") {
      lines <- c(lines, paste0(current_line, "..."))
      break
    }

    if (current_line == "") {
      test_line <- word
    } else {
      test_line <- paste(current_line, word)
    }

    if (nchar(test_line) <= chars_per_line) {
      current_line <- test_line
    } else {
      if (current_line != "") {
        lines <- c(lines, current_line)
        current_line <- word
      } else {
        if (length(lines) < max_lines - 1) {
          lines <- c(lines, paste0(substr(word, 1, chars_per_line - 3), "..."))
          current_line <- ""
        }
      }
    }
  }

  if (nchar(current_line) > 0) {
    if (length(lines) < max_lines) {
      lines <- c(lines, current_line)
    }
  }

  paste(lines[seq_len(min(length(lines), max_lines))], collapse = "\n")
}

#' Wrap text for tooltip display
#' @param text Input string.
#' @param max_words Maximum words to keep.
#' @param chars_per_line Approximate line width.
#' @param max_lines Maximum lines.
#' @return A string with HTML `<br>` breaks.
#' @keywords internal
wrap_text_for_tooltip <- function(text, max_words = 150, chars_per_line = 50, max_lines = 3) {
  text <- as.character(text)

  # Limit to 150 characters first
  if (nchar(text) > 150) {
    text_to_use <- substr(text, 1, 150)
    needs_ellipsis <- TRUE
  } else {
    text_to_use <- text
    needs_ellipsis <- FALSE
  }

  if (nchar(text_to_use) <= chars_per_line) {
    return(text_to_use)
  }

  result <- ""
  lines_created <- 0
  current_pos <- 1

  while (current_pos <= nchar(text_to_use) && lines_created < max_lines) {
    end_pos <- min(current_pos + chars_per_line - 1, nchar(text_to_use))
    line_text <- substr(text_to_use, current_pos, end_pos)

    if (end_pos < nchar(text_to_use) && lines_created < max_lines - 1) {
      last_space <- regexpr(" [^ ]*$", line_text)
      if (last_space > 0) {
        line_text <- substr(line_text, 1, last_space - 1)
        end_pos <- current_pos + last_space - 2
      }
    }

    if (result == "") {
      result <- line_text
    } else {
      result <- paste0(result, "\n", line_text)
    }

    lines_created <- lines_created + 1
    current_pos <- end_pos + 2
  }

  if (needs_ellipsis || current_pos <= nchar(text_to_use)) {
    result <- paste0(result, "...")
  }

  return(result)
}


# Matrix Utilities ----

#' Clean Similarity Matrix
#'
#' @description Cleans a similarity matrix by handling NA/Inf values,
#'   ensuring symmetry, and setting diagonal to 1.
#'
#' @param similarity_matrix A numeric matrix of similarity values.
#'
#' @return Cleaned similarity matrix.
#'
#' @concept matrix-utilities
#' @keywords internal
clean_similarity_matrix <- function(similarity_matrix) {
  # Replace non-finite values with 0
  similarity_matrix[!is.finite(similarity_matrix)] <- 0

  # Ensure symmetry
  if (nrow(similarity_matrix) == ncol(similarity_matrix)) {
    similarity_matrix <- (similarity_matrix + t(similarity_matrix)) / 2
  }

  # Set diagonal to 1
  if (nrow(similarity_matrix) == ncol(similarity_matrix)) {
    diag(similarity_matrix) <- 1
  }

  return(similarity_matrix)
}

#' Renumber Clusters Sequentially
#'
#' @description Renumbers cluster assignments to sequential integers
#'   starting from 1.
#'
#' @param clusters A vector of cluster assignments.
#'
#' @return Vector with clusters renumbered sequentially (1, 2, 3, ...).
#'
#' @concept matrix-utilities
#' @keywords internal
renumber_clusters_sequentially <- function(clusters) {
  if (is.null(clusters) || length(clusters) == 0) {
    return(clusters)
  }

  unique_clusters <- sort(unique(clusters))
  cluster_mapping <- stats::setNames(seq_along(unique_clusters), unique_clusters)
  return(cluster_mapping[as.character(clusters)])
}

test_that(".parse_provider_error extracts OpenAI error message", {
  body <- '{"error": {"message": "Incorrect API key provided.", "type": "invalid_request_error"}}'
  result <- TextAnalysisR:::.parse_provider_error("openai", body)
  expect_equal(result, "Incorrect API key provided.")
})

test_that(".parse_provider_error extracts Gemini error message", {
  body <- '{"error": {"message": "API key not valid.", "status": "INVALID_ARGUMENT"}}'
  result <- TextAnalysisR:::.parse_provider_error("gemini", body)
  expect_equal(result, "API key not valid.")
})

test_that(".parse_provider_error falls back to Gemini status field", {
  body <- '{"error": {"status": "PERMISSION_DENIED"}}'
  result <- TextAnalysisR:::.parse_provider_error("gemini", body)
  expect_equal(result, "PERMISSION_DENIED")
})

test_that(".parse_provider_error handles malformed JSON", {
  result <- TextAnalysisR:::.parse_provider_error("openai", "not json at all")
  expect_null(result)
})

test_that(".parse_provider_error handles empty body", {
  result <- TextAnalysisR:::.parse_provider_error("openai", "")
  expect_null(result)
})

test_that(".parse_provider_error truncates long messages", {
  long_msg <- paste(rep("x", 400), collapse = "")
  body <- sprintf('{"error": {"message": "%s"}}', long_msg)
  result <- TextAnalysisR:::.parse_provider_error("openai", body)
  expect_true(nchar(result) <= 300)
  expect_true(grepl("\\.\\.\\.$", result))
})

test_that(".format_api_error_message returns correct status meaning for 401", {
  msg <- TextAnalysisR:::.format_api_error_message("openai", "chat", 401, '{}')
  expect_match(msg, "Authentication failed")
  expect_match(msg, "OPENAI_API_KEY")
  expect_match(msg, "HTTP 401")
})

test_that(".format_api_error_message returns correct status meaning for 429", {
  msg <- TextAnalysisR:::.format_api_error_message("gemini", "chat", 429, '{}')
  expect_match(msg, "Rate limit exceeded")
  expect_match(msg, "Wait and retry")
})

test_that(".format_api_error_message returns correct status meaning for 500", {
  msg <- TextAnalysisR:::.format_api_error_message("openai", "embeddings", 500, '{}')
  expect_match(msg, "Server error")
  expect_match(msg, "Provider-side issue")
})

test_that(".format_api_error_message includes provider error detail", {
  body <- '{"error": {"message": "You exceeded your current quota."}}'
  msg <- TextAnalysisR:::.format_api_error_message("openai", "chat", 429, body)
  expect_match(msg, "Provider message: You exceeded your current quota.")
})

test_that(".format_api_error_message uses correct provider label", {
  msg_openai <- TextAnalysisR:::.format_api_error_message("openai", "chat", 401, '{}')
  msg_gemini <- TextAnalysisR:::.format_api_error_message("gemini", "chat", 401, '{}')
  expect_match(msg_openai, "^OpenAI")
  expect_match(msg_gemini, "^Gemini")
})

test_that(".stop_api_error raises error with formatted message", {
  expect_error(
    TextAnalysisR:::.stop_api_error("openai", "chat", 401, '{"error": {"message": "Bad key"}}'),
    "Authentication failed"
  )
  expect_error(
    TextAnalysisR:::.stop_api_error("openai", "chat", 401, '{"error": {"message": "Bad key"}}'),
    "Provider message: Bad key"
  )
})

test_that(".missing_api_key_message returns package format for openai", {
  msg <- TextAnalysisR:::.missing_api_key_message("openai", "package")
  expect_match(msg, "No OpenAI API key found")
  expect_match(msg, "OPENAI_API_KEY")
  expect_match(msg, "Sys.setenv")
  expect_match(msg, "\\.Renviron")
  expect_match(msg, "api_key parameter")
})

test_that(".missing_api_key_message returns package format for gemini", {
  msg <- TextAnalysisR:::.missing_api_key_message("gemini", "package")
  expect_match(msg, "No Gemini API key found")
  expect_match(msg, "GEMINI_API_KEY")
})

test_that(".missing_api_key_message returns shiny format for openai", {
  msg <- TextAnalysisR:::.missing_api_key_message("openai", "shiny")
  expect_match(msg, "OpenAI API key required")
  expect_match(msg, "API Key field")
  expect_match(msg, "OPENAI_API_KEY")
  expect_match(msg, "\\.Renviron")
})

test_that(".missing_api_key_message returns shiny format for gemini", {
  msg <- TextAnalysisR:::.missing_api_key_message("gemini", "shiny")
  expect_match(msg, "Gemini API key required")
  expect_match(msg, "GEMINI_API_KEY")
})

#' @importFrom utils modifyList str
#' @importFrom stats cor as.formula glm.control poisson vcov
NULL

# Suppress R CMD check notes for NSE variables
utils::globalVariables(c("K", "metric", "value", "label", "hover_text"))

# Topic Modeling Functions
# Functions for topic modeling, analysis, and evaluation

.number_labeller <- function(digits) {
  if (requireNamespace("numform", quietly = TRUE)) {
    numform::ff_num(zero = 0, digits = digits)
  } else {
    scales::label_number(accuracy = 10^(-digits))
  }
}

.embedding_silhouette <- function(model) {
  coords <- model$embeddings
  clusters <- model$topic_assignments
  if (is.null(coords) || is.null(clusters)) return(NA_real_)
  if (length(unique(clusters)) < 2 || nrow(coords) < 2) return(NA_real_)
  if (!requireNamespace("cluster", quietly = TRUE)) return(NA_real_)
  # L2-normalize so euclidean dist matches the cosine space clustering used
  norms <- pmax(sqrt(rowSums(coords^2)), 1e-12)
  coords <- coords / norms
  tryCatch(
    mean(cluster::silhouette(as.integer(as.factor(clusters)), stats::dist(coords))[, 3]),
    error = function(e) NA_real_
  )
}


#' @title Find Optimal Number of Topics
#' @description Searches for the optimal number of topics (K) using stm::searchK.
#'   Produces diagnostic plots to help select the best K value. The default
#'   spectral initialization is deterministic; with LDA or random
#'   initialization, fit several seeds per K before selecting a model.
#' @param dfm_object A quanteda dfm object to be used for topic modeling.
#' @param topic_range A vector of K values to test (e.g., 2:10).
#' @param max.em.its Maximum number of EM iterations (default: 75).
#' @param emtol Convergence tolerance for EM algorithm (default: 1e-04).
#'   Higher values (e.g., 1e-03) speed up fitting but may reduce precision.
#' @param cores Number of CPU cores to use for parallel processing (default: 1).
#'   Set to higher values for faster searchK on multi-core systems.
#' @param categorical_var Optional categorical variable(s) for prevalence.
#' @param continuous_var Optional continuous variable(s) for prevalence.
#' @param height Plot height in pixels (default: 600).
#' @param width Plot width in pixels (default: 800).
#' @param verbose Logical indicating whether to print progress (default: TRUE).
#' @param ... Additional arguments passed to stm::searchK.
#' @return A list containing search results and diagnostic plots.
#' @concept topic-modeling
#' @seealso [plot_quality_metrics()] to visualize topic-count diagnostics; `stm::stm()` to fit the chosen model; [fit_embedding_model()] for an embedding-based alternative to STM
#' @export
find_optimal_k <- function(dfm_object,
                           topic_range,
                           max.em.its = 75,
                           emtol = 1e-04,
                           cores = 1,
                           categorical_var = NULL,
                           continuous_var = NULL,
                           height = 600,
                           width = 800,
                           verbose = TRUE, ...) {
  out <- quanteda::convert(dfm_object, to = "stm")
  if (is.null(out$meta) || is.null(out$documents) || is.null(out$vocab)) {
    stop("Conversion to STM format failed. Please ensure your dfm_object is correctly formatted.")
  }
  categorical_var <- if (!is.null(categorical_var)) as.character(categorical_var) else NULL
  continuous_var <- if (!is.null(continuous_var)) as.character(continuous_var) else NULL
  if (!is.null(categorical_var)) {
    categorical_var <- unlist(strsplit(categorical_var, ",\\s*"))
  }
  if (!is.null(continuous_var)) {
    continuous_var <- unlist(strsplit(continuous_var, ",\\s*"))
  }
  missing_vars <- setdiff(c(categorical_var, continuous_var), names(out$meta))
  if (length(missing_vars) > 0) {
    stop("The following variables are missing in the metadata: ",
         paste(missing_vars, collapse = ", "))
  }
  terms <- c()
  if (!is.null(categorical_var) && length(categorical_var) > 0) {
    terms <- c(terms, categorical_var)
  }
  if (!is.null(continuous_var) && length(continuous_var) > 0) {
    terms <- c(terms, continuous_var)
  }
  prevalence_formula <- if (length(terms) > 0) {
    as.formula(paste("~", paste(terms, collapse = " + ")))
  } else {
    NULL
  }
  search_result <- tryCatch({
    stm::searchK(
      data = out$meta,
      documents = out$documents,
      vocab = out$vocab,
      max.em.its = max.em.its,
      emtol = emtol,
      cores = cores,
      init.type = "Spectral",
      K = topic_range,
      prevalence = prevalence_formula,
      verbose = verbose,
      ...
    )
  }, error = function(spectral_error) {
    # Fallback to LDA initialization if Spectral fails
    if (grepl("chol|decomposition|singular", spectral_error$message, ignore.case = TRUE)) {
      if (verbose) message("Spectral initialization failed. Trying LDA initialization...")
      stm::searchK(
        data = out$meta,
        documents = out$documents,
        vocab = out$vocab,
        max.em.its = max.em.its,
        emtol = emtol,
        cores = cores,
        init.type = "LDA",
        K = topic_range,
        prevalence = prevalence_formula,
        verbose = verbose,
        ...
      )
    } else {
      stop("Error in stm::searchK: ", spectral_error$message)
    }
  })
  
  # Clean and prepare results data
  results_clean <- search_result$results
  for (col in names(results_clean)) {
    if (is.list(results_clean[[col]])) {
      results_clean[[col]] <- unlist(results_clean[[col]])
    }
    if (col %in% c("residual", "lbound", "heldout", "semcoh", "K")) {
      results_clean[[col]] <- as.numeric(results_clean[[col]])
    }
  }
  
  # Return the same structure as stm::searchK for compatibility
  list(
    results = results_clean,
    call = match.call(),
    settings = list(
      topic_range = topic_range,
      max.em.its = max.em.its
    )
  )
}


#' Select Top Terms for Each Topic
#'
#' This function selects the top terms for each topic based on their word
#' probability distribution (beta).
#'
#' @param stm_model An STM model object.
#' @param top_term_n The number of top terms to display for each topic (default: 10).
#' @param verbose Logical, if TRUE, prints progress messages.
#' @param ... Further arguments passed to tidytext::tidy.
#'
#' @return A data frame containing the top terms for each topic.
#'
#' @concept topic-modeling
#' @export
#'
#' @examples
#' if (interactive() && requireNamespace("stm", quietly = TRUE)) {
#'   mydata <- TextAnalysisR::SpecialEduTech
#'
#'   united_tbl <- TextAnalysisR::unite_cols(
#'     mydata,
#'     listed_vars = c("title", "keyword", "abstract")
#'   )
#'
#'   tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")
#'
#'   dfm_object <- quanteda::dfm(tokens)
#'
#'   out <- quanteda::convert(dfm_object, to = "stm")
#'
#'   stm_15 <- stm::stm(
#'     data = out$meta,
#'     documents = out$documents,
#'     vocab = out$vocab,
#'     max.em.its = 75,
#'     init.type = "Spectral",
#'     K = 15,
#'     prevalence = ~ reference_type + s(year),
#'     verbose = TRUE
#'   )
#'
#'   top_topic_terms <- TextAnalysisR::get_topic_terms(
#'     stm_model = stm_15,
#'     top_term_n = 10,
#'     verbose = TRUE
#'   )
#'   print(top_topic_terms)
#' }
get_topic_terms <- function(stm_model,
                             top_term_n = 10,
                             verbose = TRUE,
                             ...) {

  beta_td <- tidytext::tidy(stm_model, matrix = "beta", ...)

  top_topic_terms <- beta_td %>%
    dplyr::group_by(topic) %>%
    dplyr::slice_max(order_by = beta, n = top_term_n) %>%
    dplyr::ungroup()

  return(top_topic_terms)
}


#' Get Topic Prevalence (Gamma) from STM Model
#'
#' Extracts topic prevalence values (gamma/theta) from a fitted STM model,
#' returning mean prevalence for each topic as a data frame.
#'
#' @param stm_model A fitted STM model object from stm::stm().
#' @param category Optional character string to add as a category column.
#' @param include_theta Logical, if TRUE includes document-topic matrix (default: FALSE).
#'
#' @return A data frame with columns:
#'   \item{topic}{Topic number}
#'   \item{gamma}{Mean topic prevalence across documents}
#'   \item{category}{Category label (if provided)}
#'
#' @concept topic-modeling
#' @export
#'
#' @examples
#' if (interactive() && requireNamespace("stm", quietly = TRUE)) {
#'   # Requires fitting an STM model first; uses 'stm::gadarian' for demo
#'   data("gadarian", package = "stm")
#'   proc <- stm::textProcessor(gadarian$open.ended.response, metadata = gadarian)
#'   prep <- stm::prepDocuments(proc$documents, proc$vocab, proc$meta)
#'   topic_model <- stm::stm(prep$documents, prep$vocab, K = 3,
#'                            data = prep$meta, max.em.its = 5,
#'                            verbose = FALSE)
#'   prevalence <- get_topic_prevalence(topic_model)
#'   prevalence_label <- get_topic_prevalence(topic_model, category = "demo")
#' }
get_topic_prevalence <- function(stm_model,
                                  category = NULL,
                                  include_theta = FALSE) {

  if (!inherits(stm_model, "STM")) {
    stop("stm_model must be a fitted STM model object")
  }

  theta <- stm_model$theta
  n_topics <- ncol(theta)

  result <- data.frame(
    topic = seq_len(n_topics),
    gamma = colMeans(theta)
  )

  if (!is.null(category)) {
    result$category <- category
  }

  if (include_theta) {
    attr(result, "theta") <- theta
  }

  result
}


#' Convert Topic Terms to Text Strings
#'
#' Concatenates top terms for each topic into text strings suitable for
#' embedding generation. Useful for creating topic representations for
#' semantic similarity analysis.
#'
#' @param top_terms_df A data frame containing top terms for topics, typically
#'   output from \code{\link{get_topic_terms}}.
#' @param topic_var Name of the column containing topic identifiers (default: "topic").
#' @param term_var Name of the column containing terms (default: "term").
#' @param weight_var Optional name of column with term weights (e.g., "beta").
#'   If provided, terms are ordered by weight before concatenation.
#' @param sep Separator between terms (default: " ").
#' @param top_n Optional number of top terms to include per topic (default: NULL, uses all).
#'
#' @return A character vector of topic text strings, one per topic, ordered by topic number.
#'
#' @concept topic-modeling
#' @export
#'
#' @examples
#' # Topic-term frame as produced by get_topic_terms()
#' top_terms <- data.frame(
#'   topic = c(1, 1, 1, 2, 2, 2),
#'   term  = c("calculator", "arithmetic", "elementary",
#'             "computer", "instruction", "multiplication"),
#'   prob  = c(0.10, 0.08, 0.07, 0.12, 0.09, 0.06)
#' )
#' get_topic_texts(top_terms)
#' get_topic_texts(top_terms, weight_var = "prob", top_n = 2)
get_topic_texts <- function(top_terms_df,
                             topic_var = "topic",
                             term_var = "term",
                             weight_var = NULL,
                             sep = " ",
                             top_n = NULL) {

  if (!topic_var %in% names(top_terms_df)) {
    stop("topic_var '", topic_var, "' not found in top_terms_df")
  }
  if (!term_var %in% names(top_terms_df)) {
    stop("term_var '", term_var, "' not found in top_terms_df")
  }

  result <- top_terms_df %>%
    dplyr::group_by(.data[[topic_var]])

  if (!is.null(weight_var) && weight_var %in% names(top_terms_df)) {
    result <- result %>%
      dplyr::arrange(dplyr::desc(.data[[weight_var]]), .by_group = TRUE)
  }

  if (!is.null(top_n)) {
    result <- result %>%
      dplyr::slice_head(n = top_n)
  }

  result <- result %>%
    dplyr::summarise(
      text = paste(.data[[term_var]], collapse = sep),
      .groups = "drop"
    ) %>%
    dplyr::arrange(.data[[topic_var]]) %>%
    dplyr::pull(text)

  result
}


#' Build a topic-term data frame from any supported topic model
#'
#' Unified helper that produces the long-format `data.frame(topic, term, beta)`
#' expected by [generate_topic_labels()] from an STM model or an embedding result.
#' Dispatches on the object's structure:
#' - STM model (has `$beta$logbeta` and `$vocab`) -> top terms via [stm::labelTopics()] FREX
#' - Embedding result (has `$topic_keywords`) -> c-TF-IDF keywords with rank-derived pseudo-beta
#'
#' @param model A topic model object (STM fit or embedding result).
#' @param n Number of top terms per topic (default 7).
#' @return `data.frame(topic, term, beta)` in long format.
#' @concept topic-modeling
#' @export
extract_topic_terms_df <- function(model, n = 7) {
  if (is.null(model)) return(data.frame(topic = integer(0), term = character(0), beta = numeric(0)))

  # STM model
  if (!is.null(model$beta) && !is.null(model$vocab)) {
    label_result <- stm::labelTopics(model, n = n)
    frex_terms <- label_result$frex
    beta_matrix <- exp(model$beta$logbeta[[1]])
    rows <- lapply(seq_len(nrow(frex_terms)), function(i) {
      terms <- as.character(frex_terms[i, ])
      vocab_idx <- match(terms, model$vocab)
      data.frame(topic = i, term = terms, beta = beta_matrix[i, vocab_idx],
                 stringsAsFactors = FALSE)
    })
    return(do.call(rbind, rows))
  }

  if (!is.null(model$topic_keywords)) {
    rows <- lapply(seq_along(model$topic_keywords), function(i) {
      words <- utils::head(as.character(model$topic_keywords[[i]]), n)
      data.frame(topic = i, term = words,
                 beta = seq(length(words), 1) / length(words),
                 stringsAsFactors = FALSE)
    })
    return(do.call(rbind, rows))
  }

  data.frame(topic = integer(0), term = character(0), beta = numeric(0))
}


#' Generate Topic Labels Using AI
#'
#' This function generates descriptive labels for each topic based on their
#' top terms using AI providers (OpenAI or Gemini).
#'
#' @param top_topic_terms A data frame containing the top terms for each topic.
#' @param provider AI provider to use: "auto" (default), "openai", or "gemini".
#'   "auto" picks the first provider with an available API key.
#' @param model A character string specifying which model to use. If NULL, uses
#'   provider defaults: "gpt-4.1-mini" (OpenAI), "gemini-2.5-flash-lite" (Gemini).
#' @param system A character string containing the system prompt for the API.
#'   If NULL, the function uses the default system prompt.
#' @param user A character string containing the user prompt for the API.
#'   If NULL, the function uses the default user prompt.
#' @param temperature A numeric value controlling the randomness of the output (default: 0.5).
#' @param api_key API key for OpenAI or Gemini. If NULL, uses environment variable.
#' @param openai_api_key Deprecated. Use `api_key` instead. Kept for backward compatibility.
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A data frame containing the top terms for each topic along with their generated labels.
#'
#' @concept topic-modeling
#' @seealso [get_topic_terms()] to extract top terms first; [generate_topic_content()] for survey items / RQs / themes grounded in the same terms; [call_llm_api()] for the direct AI provider call
#' @export
#'
#' @examples
#' if (interactive()) {
#' top_topic_terms <- get_topic_terms(stm_model, top_term_n = 10)
#'
#' # Auto-detect provider (tries OpenAI -> Gemini)
#' labels <- generate_topic_labels(top_topic_terms)
#'
#' # Use specific provider
#' labels_openai <- generate_topic_labels(top_topic_terms, provider = "openai")
#' labels_gemini <- generate_topic_labels(top_topic_terms, provider = "gemini")
#' }
generate_topic_labels <- function(top_topic_terms,
                                  provider = "auto",
                                  model = NULL,
                                  system = NULL,
                                  user = NULL,
                                  temperature = 0.5,
                                  api_key = NULL,
                                  openai_api_key = NULL,
                                  verbose = TRUE) {

  if (!requireNamespace("httr", quietly = TRUE) ||
      !requireNamespace("jsonlite", quietly = TRUE)) {
    stop(
      "The 'httr' and 'jsonlite' packages are required for this functionality. ",
      "Please install them using install.packages(c('httr', 'jsonlite'))."
    )
  }

  # Handle backward compatibility: openai_api_key -> api_key

  if (!is.null(openai_api_key) && is.null(api_key)) {
    api_key <- openai_api_key
    if (provider == "auto") provider <- "openai"
  }

  if (provider == "auto") {
    if (nzchar(Sys.getenv("OPENAI_API_KEY")) || (!is.null(api_key) && grepl("^sk-", api_key))) {
      provider <- "openai"
      if (verbose) message("Using OpenAI for topic label generation")
    } else if (nzchar(Sys.getenv("GEMINI_API_KEY")) || (!is.null(api_key) && grepl("^AIza", api_key))) {
      provider <- "gemini"
      if (verbose) message("Using Gemini for topic label generation")
    } else {
      message("No AI provider available. Set OPENAI_API_KEY or GEMINI_API_KEY.")
      return(invisible(NULL))
    }
  }

  if (is.null(model)) {
    model <- switch(provider,
      "openai" = "gpt-4.1-mini",
      "gemini" = "gemini-2.5-flash-lite"
    )
  }

  if (is.null(api_key)) {
    api_key <- switch(provider,
      "openai" = Sys.getenv("OPENAI_API_KEY"),
      "gemini" = Sys.getenv("GEMINI_API_KEY")
    )
  }

  if (!nzchar(api_key)) {
    return(.notify_missing_api_key(provider))
  }

  validation <- validate_api_key(api_key, strict = FALSE)
  if (!validation$valid) {
    message(sprintf("Invalid API key format: %s", validation$error))
    return(invisible(NULL))
  }

  if (verbose) {
    message("Generating topic labels for ", dplyr::n_distinct(top_topic_terms$topic),
            " topics using ", provider, " (", model, ")...")
  }

  system_prompt <- "You generate short, descriptive topic labels from keyword lists.

Rules:
- Max 5 words per label. Title Case. No trailing punctuation.
- Prioritize the keywords listed earliest (they have the highest weight).
- Use person-first language (e.g., 'students with learning disabilities', not 'disabled students').
- Each label must be distinct from other labels in the same model.
- Return ONLY the label text. No quotes, numbering, or explanations.

Example
Keywords: virtual manipulatives, manipulatives, mathematical, app, solving, learning disability, algebra
Label: Virtual Math Tools for Students with Disabilities"

  # Use custom system prompt if provided
  if (!is.null(system)) {
    system_prompt <- system
  }

  top_topic_terms <- top_topic_terms %>%
    dplyr::group_by(topic) %>%
    dplyr::arrange(dplyr::desc(beta)) %>%
    dplyr::ungroup()

  unique_topics <- top_topic_terms %>%
    dplyr::distinct(topic) %>%
    dplyr::arrange(as.numeric(topic)) %>%
    dplyr::mutate(topic_label = NA)

  pb <- NULL
  if (verbose && requireNamespace("progress", quietly = TRUE)) {
    pb <- progress::progress_bar$new(
      format = " Processing [:bar] :percent ETA: :eta",
      total = nrow(unique_topics),
      clear = FALSE, width = 60)
  }

  rate_limit_delay <- 1

  for (i in seq_len(nrow(unique_topics))) {
    if (!is.null(pb)) pb$tick()

    current_topic <- unique_topics$topic[i]

    selected_terms <- top_topic_terms %>%
      dplyr::filter(topic == current_topic) %>%
      dplyr::pull(term)

    user_prompt <- paste0(
      "I have a topic described by the following keywords (highest weight first): '",
      paste(selected_terms, collapse = ", "),
      "'. Based on the keywords, create a short label (max 5 words). Return only the label."
    )

    # Use custom user prompt if provided
    if (!is.null(user)) {
      user_prompt <- user
    }

    topic_label <- tryCatch({
      response <- call_llm_api(
        provider = provider,
        system_prompt = system_prompt,
        user_prompt = user_prompt,
        model = model,
        temperature = temperature,
        max_tokens = 50,
        api_key = api_key
      )
      label <- trimws(response)
      label <- gsub("^\"(.*)\"$", "\\1", label)
      label
    }, error = function(e) {
      warning(sprintf("API request failed for topic '%s': %s", current_topic, e$message))
      NA_character_
    })

    unique_topics$topic_label[i] <- topic_label

    Sys.sleep(rate_limit_delay)
  }

  top_labeled_topic_terms <- top_topic_terms %>%
    dplyr::left_join(unique_topics, by = "topic") %>%
    dplyr::select(topic_label, topic, term, beta) %>%
    dplyr::arrange(topic, dplyr::desc(beta))

  return(top_labeled_topic_terms)
}

#' @title Calculate Topic Probabilities
#'
#' @description
#' Extracts and summarizes topic probabilities (gamma values) from an STM model,
#' returning a formatted data table of mean topic prevalence.
#'
#' @param stm_model A fitted STM model object from stm::stm().
#' @param top_n Number of top topics to display by prevalence (default: 10).
#' @param verbose Logical, if TRUE prints progress messages (default: TRUE).
#' @param ... Additional arguments passed to tidytext::tidy().
#'
#' @return A DT::datatable showing topics and their mean gamma (prevalence) values,
#'   rounded to 3 decimal places.
#'
#' @concept topic-modeling
#' @export
#'
#' @examples
#' if (interactive()) {
#'   data <- TextAnalysisR::SpecialEduTech
#'   united <- unite_cols(data, c("title", "keyword", "abstract"))
#'   tokens <- prep_texts(united, text_field = "united_texts")
#'   dfm_obj <- quanteda::dfm(tokens)
#'   stm_data <- quanteda::convert(dfm_obj, to = "stm")
#'
#'   topic_model <- stm::stm(
#'     documents = stm_data$documents,
#'     vocab = stm_data$vocab,
#'     K = 10,
#'     verbose = FALSE
#'   )
#'
#'   prob_table <- calculate_topic_probability(topic_model, top_n = 10)
#'   print(prob_table)
#' }
calculate_topic_probability <- function(stm_model,
                                    top_n = 10,
                                    verbose = TRUE,
                                    ...) {

  gamma_td <- tidytext::tidy(stm_model, matrix = "gamma", ...)

  gamma_td %>%
    dplyr::group_by(topic) %>%
    dplyr::summarise(gamma = mean(gamma), .groups = "drop") %>%
    dplyr::arrange(dplyr::desc(gamma)) %>%
    dplyr::top_n(top_n, gamma) %>%
    dplyr::mutate(gamma = round(gamma, 3))
}

#' @title Neural Topic Modeling
#'
#' @description
#' Implements neural topic modeling using deep learning architectures for improved
#' topic discovery and representation learning.
#'
#' @param texts Character vector of documents
#' @param n_topics Number of topics to discover
#' @param hidden_layers Number of hidden layers in neural network
#' @param hidden_units Number of units per hidden layer
#' @param dropout_rate Dropout rate for regularization
#' @param embedding_model Transformer model for initial embeddings
#' @param seed Random seed for reproducibility
#'
#' @return List containing neural topic model and diagnostics
#' @concept topic-modeling
#' @keywords internal
#' @export
run_neural_topics_internal <- function(texts, n_topics = 10, hidden_layers = 2,
                                           hidden_units = 100, dropout_rate = 0.2,
                                           embedding_model = "all-MiniLM-L6-v2", seed = 123) {

  tryCatch({
    base_result <- fit_embedding_model(
      texts = texts,
      method = "embedding_clustering",
      n_topics = n_topics,
      embedding_model = embedding_model,
      seed = seed
    )

    coherence <- calculate_coherence(base_result$embeddings, base_result$topic_assignments)

    diagnostics <- list(
      architecture = list(
        hidden_layers = hidden_layers,
        hidden_units = hidden_units,
        dropout_rate = dropout_rate
      ),
      topic_quality = list(
        neural_coherence = coherence$coherence_scores,
        mean_coherence = coherence$mean_coherence
      )
    )

    result <- base_result
    result$method <- "neural_topic_model"
    result$diagnostics <- diagnostics
    result$architecture <- list(hidden_layers = hidden_layers,
                                hidden_units = hidden_units,
                                dropout_rate = dropout_rate)

    return(result)

  }, error = function(e) {
    warning("Neural topic modeling failed, falling back to base method: ", e$message)
    return(fit_embedding_model(texts = texts, method = "embedding_clustering",
                                   n_topics = n_topics,
                                 embedding_model = embedding_model, seed = seed))
  })
}

#' @title Fit Embedding-based Topic Model
#'
#' @description
#' This function performs embedding-based topic modeling using transformer embeddings
#' and specialized clustering techniques. Supports two backends:
#'
#' - **Python backend** (default): Uses BERTopic library which combines transformer
#'   embeddings with UMAP dimensionality reduction and HDBSCAN clustering for optimal
#'   topic discovery.
#' - **R backend**: Uses R-native packages (umap, dbscan, Rtsne) for users without
#'   Python/BERTopic installed. Provides similar functionality with c-TF-IDF keyword
#'   extraction.
#'
#' @param texts A character vector of texts to analyze.
#' @param method The topic modeling method:
#'   - For Python backend: "umap_hdbscan" (uses BERTopic)
#'   - For R backend: "umap_dbscan", "umap_kmeans", "umap_hierarchical",
#'     "tsne_dbscan", "tsne_kmeans", "pca_kmeans", "pca_hierarchical"
#'   - For both: "embedding_clustering", "hierarchical_semantic"
#' @param n_topics The number of topics to identify. For UMAP+HDBSCAN, use NULL or "auto" for automatic determination, or specify an integer.
#' @param embedding_model The embedding model to use (default: "all-MiniLM-L6-v2").
#' @param backend The backend to use: "auto" (default, tries Python then R),
#'   "python" (requires BERTopic), or "r" (R-native packages only).
#' @param clustering_method The clustering method for embedding-based approach: "kmeans", "hierarchical", "dbscan", "hdbscan".
#' @param similarity_threshold The similarity threshold for topic assignment (default: 0.7).
#' @param min_topic_size The minimum number of documents per topic (default: 3).
#' @param min_cluster_size HDBSCAN density threshold (default `NULL` falls back to `min_topic_size`). Setting this independently lets fine-grained clusters merge into broader topics.
#' @param cluster_selection_method HDBSCAN cluster selection method: "eom" (Excess of Mass, default) or "leaf" (finer-grained topics).
#' @param umap_neighbors The number of neighbors for UMAP dimensionality reduction (default: 15).
#' @param umap_min_dist The minimum distance for UMAP (default: 0.0). Use 0.0 for tight, well-separated clusters. Use 0.1+ for visualization purposes. Range: 0.0-0.99.
#' @param umap_n_components The number of UMAP components (default: 5).
#' @param umap_metric Distance metric for UMAP: "cosine" (recommended for text) or "euclidean" (default: "cosine").
#' @param tsne_perplexity Perplexity parameter for t-SNE (default: 30). Only used when method includes "tsne"; t-SNE output is 2-dimensional.
#' @param pca_dims Number of PCA components kept for clustering (default: 50). Only used when method includes "pca".
#' @param dbscan_eps Epsilon parameter for DBSCAN (default: 0.5). Neighborhood size for density-based clustering.
#' @param dbscan_minpts Minimum points for DBSCAN core points (default: 5).
#' @param representation_method The method for topic representation: "c-tfidf", "tfidf", "mmr", "frequency" (default: "c-tfidf").
#'   Applies only to the R fallback backend and the "embedding_clustering",
#'   "semantic_lda", and "hierarchical_semantic" methods; the Python BERTopic
#'   backend ("umap_hdbscan") uses BERTopic's native c-TF-IDF representation.
#' @param diversity Diversity weight between 0 and 1 for the "mmr"
#'   representation (default: 0.5). Higher values penalize redundant terms
#'   more strongly. Applies to the R backend; ignored by the Python BERTopic
#'   backend.
#' @param reduce_outliers Logical, if TRUE, reduces outliers in HDBSCAN clustering (default: TRUE).
#' @param outlier_strategy Strategy for outlier reduction using BERTopic:
#'   "probabilities" (default, uses topic probabilities), "c-tf-idf" (uses
#'   c-TF-IDF similarity), "embeddings" (uses cosine similarity in embedding
#'   space), or "distributions" (uses topic distributions). Ignored if
#'   reduce_outliers = FALSE.
#' @param outlier_threshold Minimum threshold for outlier reassignment (default: 0.0).
#'   Higher values require stronger evidence for reassignment.
#' @param seed Random seed for reproducibility (default: 123).
#' @param verbose Logical, if TRUE, prints progress messages.
#' @param precomputed_embeddings Optional matrix of pre-computed document embeddings.
#'   If provided, skips embedding generation for improved performance. Must have
#'   the same number of rows as the length of texts.
#'
#' @return A list containing topic assignments, topic keywords, and quality metrics.
#'
#' @concept topic-modeling
#' @seealso [get_best_embeddings()] to supply precomputed embeddings; [generate_topic_labels()] for AI-suggested topic names; [find_optimal_k()] for an STM-based alternative
#' @export
#'
#' @examples
#' if (interactive()) {
#'   mydata <- TextAnalysisR::SpecialEduTech
#'   united_tbl <- TextAnalysisR::unite_cols(
#'     mydata,
#'     listed_vars = c("title", "keyword", "abstract")
#'   )
#'   texts <- united_tbl$united_texts
#'
#'   # Embedding-based topic modeling (powered by BERTopic)
#'   result <- TextAnalysisR::fit_embedding_model(
#'     texts = texts,
#'     method = "umap_hdbscan",
#'     n_topics = 8,
#'     min_topic_size = 3
#'   )
#'
#'   print(result$topic_assignments)
#'   print(result$topic_keywords)
#' }
fit_embedding_model <- function(texts,
                                   method = "umap_hdbscan",
                                   n_topics = 10,
                                   embedding_model = "all-MiniLM-L6-v2",
                                   backend = "auto",
                                   clustering_method = "kmeans",
                                   similarity_threshold = 0.7,
                                   min_topic_size = 10,
                                   min_cluster_size = NULL,
                                   cluster_selection_method = "eom",
                                   umap_neighbors = 15,
                                   umap_min_dist = 0.0,
                                   umap_n_components = 5,
                                   umap_metric = "cosine",
                                   tsne_perplexity = 30,
                                   pca_dims = 50,
                                   dbscan_eps = 0.5,
                                   dbscan_minpts = 5,
                                   representation_method = "c-tfidf",
                                   diversity = 0.5,
                                   reduce_outliers = TRUE,
                                   outlier_strategy = "probabilities",
                                   outlier_threshold = 0.0,
                                   seed = 123,
                                   verbose = TRUE,
                                   precomputed_embeddings = NULL) {

  if (is.null(texts) || length(texts) == 0) {
    stop("No texts provided for analysis")
  }

  backend <- match.arg(backend, c("auto", "python", "r"))

  if (backend == "auto") {
    python_available <- tryCatch({
      requireNamespace("reticulate", quietly = TRUE) &&
        reticulate::py_module_available("bertopic")
    }, error = function(e) FALSE)

    backend <- if (python_available) "python" else "r"
    if (verbose) {
      message("Auto-detected backend: ", backend)
    }
  }

  if (backend == "r") {
    return(.fit_embedding_model_r(
      texts = texts,
      method = method,
      n_topics = n_topics,
      embedding_model = embedding_model,
      umap_neighbors = umap_neighbors,
      umap_min_dist = umap_min_dist,
      umap_n_components = umap_n_components,
      umap_metric = umap_metric,
      tsne_perplexity = tsne_perplexity,
      pca_dims = pca_dims,
      dbscan_eps = dbscan_eps,
      dbscan_minpts = dbscan_minpts,
      min_topic_size = min_topic_size,
      representation_method = representation_method,
      diversity = diversity,
      reduce_outliers = reduce_outliers,
      seed = seed,
      verbose = verbose,
      precomputed_embeddings = precomputed_embeddings
    ))
  }

  if (verbose) {
    message("Starting semantic-based topic modeling...")
    message("Method: ", method)
    message("Number of topics: ", n_topics)
    message("Backend: Python (BERTopic)")
  }

  valid_texts <- texts[nchar(trimws(texts)) > 0]
  if (length(valid_texts) < min_topic_size) {
    stop("Need at least ", min_topic_size, " non-empty texts for analysis")
  }

  withr::local_seed(seed)
  start_time <- Sys.time()

  tryCatch({
    if (!is.null(precomputed_embeddings)) {
      if (verbose) message("Step 1: Using precomputed embeddings...")
      if (!is.matrix(precomputed_embeddings)) {
        precomputed_embeddings <- as.matrix(precomputed_embeddings)
      }
      if (nrow(precomputed_embeddings) != length(valid_texts)) {
        stop("Precomputed embeddings dimension mismatch: ",
             nrow(precomputed_embeddings), " embeddings vs ",
             length(valid_texts), " texts")
      }
      embeddings <- precomputed_embeddings
    } else {
      if (verbose) message("Step 1: Generating document embeddings...")

      if (!requireNamespace("reticulate", quietly = TRUE)) {
        return(.notify_missing_python("Semantic topic modeling"))
      }

      python_available <- tryCatch({
        reticulate::py_config()
        TRUE
      }, error = function(e) FALSE)

      if (!python_available) {
        return(.notify_missing_python("Semantic topic modeling (sentence-transformers)"))
      }
      sentence_transformers <- reticulate::import("sentence_transformers")
      model <- sentence_transformers$SentenceTransformer(embedding_model)

      n_docs <- length(valid_texts)
      batch_size <- if (n_docs > 100) 25 else if (n_docs > 50) 50 else n_docs

      if (verbose) message("Processing ", n_docs, " documents with embeddings...")

      embeddings_list <- list()
      for (i in seq(1, n_docs, by = batch_size)) {
        end_idx <- min(i + batch_size - 1, n_docs)
        batch_texts <- valid_texts[i:end_idx]
        batch_embeddings <- model$encode(batch_texts, show_progress_bar = FALSE, normalize_embeddings = TRUE)
        embeddings_list[[length(embeddings_list) + 1]] <- batch_embeddings
      }

      embeddings <- do.call(rbind, embeddings_list)
    }


    result <- switch(method,
      "umap_hdbscan" = {
        if (verbose) message("Step 2: Performing BERTopic-based topic modeling...")

        bertopic_available <- tryCatch({
          reticulate::import("bertopic")
          TRUE
        }, error = function(e) FALSE)

        if (!bertopic_available) {
          stop("BERTopic library not found. Please install: pip install bertopic\n",
               "Or in R: reticulate::py_install('bertopic')")
        }

        bertopic <- reticulate::import("bertopic")

        # "auto" requests BERTopic's automatic topic merging
        nr_topics <- if (is.null(n_topics)) NULL
          else if (identical(n_topics, "auto")) "auto"
          else as.integer(n_topics)

        if (verbose) message("Initializing BERTopic model...")

        topic_model <- bertopic$BERTopic(
          embedding_model = embedding_model,
          nr_topics = nr_topics,
          min_topic_size = as.integer(min_topic_size),
          umap_model = reticulate::import("umap")$UMAP(
            n_neighbors = as.integer(umap_neighbors),
            n_components = as.integer(umap_n_components),
            min_dist = umap_min_dist,
            metric = umap_metric,
            random_state = as.integer(seed)
          ),
          hdbscan_model = reticulate::import("hdbscan")$HDBSCAN(
            min_cluster_size = as.integer(min_cluster_size %||% min_topic_size),
            cluster_selection_method = cluster_selection_method,
            metric = "euclidean",
            prediction_data = TRUE
          ),
          calculate_probabilities = TRUE,
          verbose = verbose
        )

        if (verbose) message("Fitting BERTopic model to ", length(valid_texts), " documents...")

        # cluster on the already computed embeddings
        fit_result <- topic_model$fit_transform(valid_texts, embeddings = embeddings)
        topic_assignments_raw <- fit_result[[1]]
        topic_probs <- fit_result[[2]]

        topic_assignments <- as.vector(topic_assignments_raw) + 1

        outlier_docs <- which(topic_assignments == 0)
        outliers_reassigned <- 0
        outlier_strategy_used <- outlier_strategy
        if (length(outlier_docs) > 0 && reduce_outliers) {
          if (verbose) message("Reassigning ", length(outlier_docs), " outlier documents using '", outlier_strategy, "' strategy...")

          if (outlier_strategy == "probabilities") {
            for (idx in outlier_docs) {
              if (!is.null(topic_probs) && nrow(topic_probs) >= idx) {
                probs <- topic_probs[idx, ]
                if (max(probs) > outlier_threshold) {
                  topic_assignments[idx] <- which.max(probs)
                  outliers_reassigned <- outliers_reassigned + 1
                }
              }
            }
          } else if (outlier_strategy %in% c("c-tf-idf", "embeddings", "distributions")) {
            new_topics <- topic_model$reduce_outliers(
              valid_texts,
              as.integer(topic_assignments_raw),
              strategy = outlier_strategy,
              threshold = outlier_threshold
            )
            # sync topic info with the new assignments
            topic_model$update_topics(valid_texts, topics = new_topics)
            topic_assignments <- as.vector(new_topics) + 1
            outliers_reassigned <- length(outlier_docs) - sum(topic_assignments == 0)
          }
        }

        if (verbose) message("Extracting topic information...")
        topic_info <- topic_model$get_topic_info()

        topic_keywords_list <- list()
        topic_keyword_scores_list <- list()
        for (topic_id in unique(topic_assignments)) {
          if (topic_id == 0) next

          topic_words <- topic_model$get_topic(as.integer(topic_id - 1))

          if (!is.null(topic_words) && length(topic_words) > 0 && is.list(topic_words)) {
            keywords <- vapply(topic_words, function(x) as.character(x[[1]]), character(1))
            scores <- vapply(topic_words, function(x) as.numeric(x[[2]]), numeric(1))
          } else {
            keywords <- character(0)
            scores <- numeric(0)
          }

          topic_keywords_list[[as.character(topic_id)]] <- keywords
          topic_keyword_scores_list[[as.character(topic_id)]] <- scores
        }

        embeddings_matrix <- embeddings

        reduced_embeddings_obj <- topic_model$umap_model$embedding_
        reduced_embeddings <- as.matrix(reduced_embeddings_obj)

        list(
          topic_assignments = topic_assignments,
          topic_keywords = topic_keywords_list,
          topic_keyword_scores = topic_keyword_scores_list,
          embeddings = embeddings_matrix,
          reduced_embeddings = reduced_embeddings,
          bertopic_model = topic_model,
          topic_info = topic_info,
          probabilities = topic_probs,
          method = "umap_hdbscan",
          outlier_reduction = list(
            enabled = reduce_outliers,
            strategy = if (reduce_outliers) outlier_strategy_used else NA_character_,
            threshold = if (reduce_outliers) outlier_threshold else NA_real_,
            outliers_found = length(outlier_docs),
            outliers_reassigned = outliers_reassigned
          )
        )
      },
      "embedding_clustering" = {
        if (verbose) message("Step 2: Performing embedding-based clustering...")

        sklearn_metrics <- reticulate::import("sklearn.metrics.pairwise")
        similarity_matrix <- sklearn_metrics$cosine_similarity(embeddings)
        similarity_matrix <- as.matrix(similarity_matrix)

        clustering_result <- cluster_embeddings(
          data_matrix = similarity_matrix,
          method = clustering_method,
          n_clusters = n_topics,
          seed = seed,
          verbose = verbose
        )


        topic_assignments <- clustering_result$clusters

        topic_keywords <- generate_semantic_topic_keywords(
          texts = valid_texts,
          topic_assignments = topic_assignments,
          n_keywords = 10,
          method = representation_method,
          diversity = diversity
        )

        list(
          topic_assignments = topic_assignments,
          topic_keywords = topic_keywords,
          similarity_matrix = similarity_matrix,
          embeddings = embeddings,
          clustering_result = clustering_result,
          method = "embedding_clustering"
        )
      },
      "semantic_lda" = {
        if (verbose) message("Step 2: Performing semantic LDA...")

        pca_result <- stats::prcomp(embeddings, center = TRUE, scale. = FALSE, rank. = min(50, ncol(embeddings)))
        reduced_embeddings <- pca_result$x

        kmeans_result <- stats::kmeans(reduced_embeddings, centers = n_topics, nstart = 25)
        topic_assignments <- kmeans_result$cluster

        topic_keywords <- generate_semantic_topic_keywords(
          texts = valid_texts,
          topic_assignments = topic_assignments,
          n_keywords = 10,
          method = representation_method,
          diversity = diversity
        )

        list(
          topic_assignments = topic_assignments,
          topic_keywords = topic_keywords,
          embeddings = embeddings,
          reduced_embeddings = reduced_embeddings,
          kmeans_result = kmeans_result,
          method = "semantic_lda"
        )
      },
      "hierarchical_semantic" = {
        if (verbose) message("Step 2: Performing hierarchical semantic clustering...")

        dist_matrix <- stats::dist(embeddings, method = "euclidean")

        hclust_result <- stats::hclust(dist_matrix, method = "ward.D2")

        topic_assignments <- stats::cutree(hclust_result, k = n_topics)

        topic_keywords <- generate_semantic_topic_keywords(
          texts = valid_texts,
          topic_assignments = topic_assignments,
          n_keywords = 10,
          method = representation_method,
          diversity = diversity
        )

        list(
          topic_assignments = topic_assignments,
          topic_keywords = topic_keywords,
          embeddings = embeddings,
          hclust_result = hclust_result,
          method = "hierarchical_semantic"
        )
      },
      stop("Unsupported semantic topic modeling method: ", method)
    )

    if (verbose) message("Step 3: Calculating quality metrics...")

    quality_metrics <- calculate_topic_quality(
      embeddings = embeddings,
      topic_assignments = result$topic_assignments,
      similarity_matrix = result$similarity_matrix
    )

    execution_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

    if (verbose) {
      message("Semantic topic modeling completed in ", round(execution_time, 2), " seconds")
      message("Topics identified: ", length(unique(result$topic_assignments)))
    }

    final_result <- c(result, list(
      quality_metrics = quality_metrics,
      execution_time = execution_time,
      n_documents = length(valid_texts),
      n_topics = length(unique(result$topic_assignments)),
      embedding_model = embedding_model,
      timestamp = Sys.time()
    ))

    return(final_result)

  }, error = function(e) {
    stop("Error in semantic topic modeling: ", e$message)
  })
}


.fit_embedding_model_r <- function(texts,
                                   method = "umap_dbscan",
                                   n_topics = 10,
                                   embedding_model = "all-MiniLM-L6-v2",
                                   umap_neighbors = 15,
                                   umap_min_dist = 0.0,
                                   umap_n_components = 5,
                                   umap_metric = "cosine",
                                   tsne_perplexity = 30,
                                   pca_dims = 50,
                                   dbscan_eps = 0.5,
                                   dbscan_minpts = 5,
                                   min_topic_size = 10,
                                   representation_method = "c-tfidf",
                                   diversity = 0.5,
                                   reduce_outliers = TRUE,
                                   seed = 123,
                                   verbose = TRUE,
                                   precomputed_embeddings = NULL) {

  if (verbose) {
    message("Starting R-native embedding-based topic modeling...")
    message("Method: ", method)
    message("Number of topics: ", n_topics)
    message("Backend: R (no Python required)")
  }

  valid_texts <- texts[nchar(trimws(texts)) > 0]
  if (length(valid_texts) < min_topic_size) {
    stop("Need at least ", min_topic_size, " non-empty texts for analysis")
  }

  withr::local_seed(seed)
  start_time <- Sys.time()

  tryCatch({
    if (!is.null(precomputed_embeddings)) {
      if (verbose) message("Step 1: Using precomputed embeddings...")
      if (!is.matrix(precomputed_embeddings)) {
        precomputed_embeddings <- as.matrix(precomputed_embeddings)
      }
      if (nrow(precomputed_embeddings) != length(valid_texts)) {
        stop("Precomputed embeddings dimension mismatch: ",
             nrow(precomputed_embeddings), " embeddings vs ",
             length(valid_texts), " texts")
      }
      embeddings <- precomputed_embeddings
    } else {
      if (verbose) message("Step 1: Generating document embeddings...")

      if (!requireNamespace("reticulate", quietly = TRUE)) {
        return(.notify_missing_python("Embedding generation"))
      }

      python_available <- tryCatch({
        reticulate::py_config()
        TRUE
      }, error = function(e) FALSE)

      if (!python_available) {
        return(.notify_missing_python("Embedding generation (sentence-transformers)"))
      }
      sentence_transformers <- reticulate::import("sentence_transformers")
      model <- sentence_transformers$SentenceTransformer(embedding_model)

      n_docs <- length(valid_texts)
      batch_size <- if (n_docs > 100) 25 else if (n_docs > 50) 50 else n_docs

      if (verbose) message("Processing ", n_docs, " documents with embeddings...")

      embeddings_list <- list()
      for (i in seq(1, n_docs, by = batch_size)) {
        end_idx <- min(i + batch_size - 1, n_docs)
        batch_texts <- valid_texts[i:end_idx]
        batch_embeddings <- model$encode(batch_texts, show_progress_bar = FALSE, normalize_embeddings = TRUE)
        embeddings_list[[length(embeddings_list) + 1]] <- batch_embeddings
      }

      embeddings <- do.call(rbind, embeddings_list)
    }

    dimred_method <- gsub("_.*", "", method)
    cluster_method <- gsub(".*_", "", method)

    if (verbose) message("Step 2: Applying dimensionality reduction (", toupper(dimred_method), ")...")

    reduced_embeddings <- switch(dimred_method,
      "umap" = {
        if (!requireNamespace("umap", quietly = TRUE)) {
          stop("umap package required. Install with: install.packages('umap')")
        }
        umap_config <- umap::umap.defaults
        umap_config$n_neighbors <- umap_neighbors
        umap_config$min_dist <- max(umap_min_dist, 0.001)
        umap_config$n_components <- min(umap_n_components, ncol(embeddings))
        umap_config$metric <- umap_metric
        umap_result <- umap::umap(embeddings, config = umap_config)
        umap_result$layout
      },
      "tsne" = {
        if (!requireNamespace("Rtsne", quietly = TRUE)) {
          stop("Rtsne package required. Install with: install.packages('Rtsne')")
        }
        tsne_result <- Rtsne::Rtsne(embeddings,
                                    dims = 2,
                                    perplexity = min(tsne_perplexity, floor((nrow(embeddings) - 1) / 3)),
                                    check_duplicates = FALSE)
        tsne_result$Y
      },
      "pca" = {
        pca_result <- stats::prcomp(embeddings, center = TRUE, scale. = FALSE,
                                    rank. = min(pca_dims, ncol(embeddings)))
        pca_result$x[, seq_len(min(pca_dims, ncol(pca_result$x))), drop = FALSE]
      },
      {
        if (!requireNamespace("umap", quietly = TRUE)) {
          stop("umap package required. Install with: install.packages('umap')")
        }
        umap_config <- umap::umap.defaults
        umap_config$n_neighbors <- umap_neighbors
        umap_config$min_dist <- max(umap_min_dist, 0.001)
        umap_config$n_components <- min(umap_n_components, ncol(embeddings))
        umap_config$metric <- umap_metric
        umap_result <- umap::umap(embeddings, config = umap_config)
        umap_result$layout
      }
    )

    if (verbose) message("Step 3: Clustering documents (", cluster_method, ")...")

    clusters <- switch(cluster_method,
      "dbscan" = {
        if (!requireNamespace("dbscan", quietly = TRUE)) {
          stop("dbscan package required. Install with: install.packages('dbscan')")
        }
        db_result <- dbscan::dbscan(reduced_embeddings, eps = dbscan_eps, minPts = dbscan_minpts)
        clusters <- db_result$cluster

        if (reduce_outliers && any(clusters == 0) && length(unique(clusters[clusters > 0])) > 0) {
          if (verbose) message("Reassigning ", sum(clusters == 0), " outlier documents...")
          noise_idx <- which(clusters == 0)
          valid_clusters <- unique(clusters[clusters > 0])

          centroids <- sapply(valid_clusters, function(cl) {
            colMeans(reduced_embeddings[clusters == cl, , drop = FALSE])
          })
          if (is.vector(centroids)) centroids <- matrix(centroids, nrow = 1)
          centroids <- t(centroids)

          for (idx in noise_idx) {
            point <- reduced_embeddings[idx, ]
            distances <- apply(centroids, 1, function(c) sqrt(sum((point - c)^2)))
            clusters[idx] <- valid_clusters[which.min(distances)]
          }
        }
        clusters
      },
      "kmeans" = {
        k <- if (is.null(n_topics) || n_topics == "auto") {
          min(10, nrow(reduced_embeddings) - 1)
        } else {
          as.integer(n_topics)
        }
        km_result <- stats::kmeans(reduced_embeddings, centers = k, nstart = 25)
        km_result$cluster
      },
      "hierarchical" = {
        k <- if (is.null(n_topics) || n_topics == "auto") {
          min(10, nrow(reduced_embeddings) - 1)
        } else {
          as.integer(n_topics)
        }
        dist_matrix <- stats::dist(reduced_embeddings)
        hc_result <- stats::hclust(dist_matrix, method = "ward.D2")
        stats::cutree(hc_result, k = k)
      },
      "hdbscan" = {
        if (!requireNamespace("dbscan", quietly = TRUE)) {
          stop("dbscan package required. Install with: install.packages('dbscan')")
        }
        hdb_result <- dbscan::hdbscan(reduced_embeddings, minPts = min_topic_size)
        clusters <- hdb_result$cluster

        if (reduce_outliers && any(clusters == 0) && length(unique(clusters[clusters > 0])) > 0) {
          if (verbose) message("Reassigning ", sum(clusters == 0), " outlier documents...")
          noise_idx <- which(clusters == 0)
          valid_clusters <- unique(clusters[clusters > 0])

          centroids <- sapply(valid_clusters, function(cl) {
            colMeans(reduced_embeddings[clusters == cl, , drop = FALSE])
          })
          if (is.vector(centroids)) centroids <- matrix(centroids, nrow = 1)
          centroids <- t(centroids)

          for (idx in noise_idx) {
            point <- reduced_embeddings[idx, ]
            distances <- apply(centroids, 1, function(c) sqrt(sum((point - c)^2)))
            clusters[idx] <- valid_clusters[which.min(distances)]
          }
        }
        clusters
      },
      {
        if (!requireNamespace("dbscan", quietly = TRUE)) {
          stop("dbscan package required. Install with: install.packages('dbscan')")
        }
        db_result <- dbscan::dbscan(reduced_embeddings, eps = dbscan_eps, minPts = dbscan_minpts)
        db_result$cluster
      }
    )

    if (verbose) message("Step 4: Generating topic keywords via ", representation_method, "...")
    topic_keywords <- generate_semantic_topic_keywords(
      texts = valid_texts,
      topic_assignments = clusters,
      n_keywords = 10,
      method = representation_method,
      diversity = diversity
    )

    if (verbose) message("Step 5: Calculating quality metrics...")
    quality_metrics <- tryCatch({
      calculate_topic_quality(
        embeddings = embeddings,
        topic_assignments = clusters,
        similarity_matrix = NULL
      )
    }, error = function(e) list(overall_quality = NA))

    execution_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

    if (verbose) {
      message("R-native topic modeling completed in ", round(execution_time, 2), " seconds")
      message("Topics identified: ", length(unique(clusters[clusters > 0])))
    }

    list(
      topic_assignments = clusters,
      topic_keywords = topic_keywords,
      embeddings = embeddings,
      reduced_embeddings = reduced_embeddings,
      method = method,
      backend = "r",
      dimred_method = dimred_method,
      cluster_method = cluster_method,
      quality_metrics = quality_metrics,
      execution_time = execution_time,
      n_documents = length(valid_texts),
      n_topics = length(unique(clusters[clusters > 0])),
      embedding_model = embedding_model,
      timestamp = Sys.time()
    )

  }, error = function(e) {
    stop("Error in R-native topic modeling: ", e$message)
  })
}


#' @title Embedding-based Topic Modeling (Deprecated)
#' @description
#' This function is deprecated. Please use [fit_embedding_model()] instead.
#' @inheritParams fit_embedding_model
#' @return A list containing topic assignments, topic keywords, and quality metrics.
#' @keywords internal
#' @export
fit_embedding_topics <- function(texts,
                                   method = "umap_hdbscan",
                                   n_topics = 10,
                                   embedding_model = "all-MiniLM-L6-v2",
                                   clustering_method = "kmeans",
                                   similarity_threshold = 0.7,
                                   min_topic_size = 10,
                                   cluster_selection_method = "eom",
                                   umap_neighbors = 15,
                                   umap_min_dist = 0.0,
                                   umap_n_components = 5,
                                   representation_method = "c-tfidf",
                                   diversity = 0.5,
                                   reduce_outliers = TRUE,
                                   outlier_strategy = "probabilities",
                                   outlier_threshold = 0.0,
                                   seed = 123,
                                   verbose = TRUE,
                                   precomputed_embeddings = NULL) {
  .Deprecated("fit_embedding_model")
  fit_embedding_model(
    texts = texts,
    method = method,
    n_topics = n_topics,
    embedding_model = embedding_model,
    clustering_method = clustering_method,
    similarity_threshold = similarity_threshold,
    min_topic_size = min_topic_size,
    cluster_selection_method = cluster_selection_method,
    umap_neighbors = umap_neighbors,
    umap_min_dist = umap_min_dist,
    umap_n_components = umap_n_components,
    representation_method = representation_method,
    diversity = diversity,
    reduce_outliers = reduce_outliers,
    outlier_strategy = outlier_strategy,
    outlier_threshold = outlier_threshold,
    seed = seed,
    verbose = verbose,
    precomputed_embeddings = precomputed_embeddings
  )
}

#' @title Find Similar Topics
#'
#' @description
#' This function finds the most similar topics to a given query using semantic similarity analysis.
#' It works with both semantic topic models and traditional STM models by creating topic representations
#' using transformer embeddings and calculating cosine similarity scores.
#'
#' @param topic_model A topic model object (semantic topic model or STM model).
#' @param query A character string representing the query topic.
#' @param top_n The number of similar topics to return (default: 10).
#' @param method The similarity method: "cosine", "euclidean", "embedding".
#' @param embedding_model The embedding model to use for query encoding (default: "all-MiniLM-L6-v2").
#' @param include_terms Logical, whether to include topic terms in the similarity calculation (default: TRUE).
#'
#' @return A list containing similar topics and their similarity scores.
#'
#' @concept topic-modeling
#' @export
#'
#' @examples
#' if (interactive()) {
#'   mydata <- TextAnalysisR::SpecialEduTech
#'   united_tbl <- TextAnalysisR::unite_cols(
#'     mydata,
#'     listed_vars = c("title", "keyword", "abstract")
#'   )
#'   texts <- united_tbl$united_texts
#'
#'   topic_model <- TextAnalysisR::fit_embedding_model(
#'     texts = texts,
#'     method = "umap_hdbscan",
#'     n_topics = 8
#'   )
#'
#'   similar_topics <- TextAnalysisR::find_topic_matches(
#'     topic_model = topic_model,
#'     query = "mathematical learning",
#'     top_n = 5
#'   )
#'
#'   print(similar_topics)
#' }
find_topic_matches <- function(topic_model,
                               query,
                               top_n = 10,
                               method = "cosine",
                               embedding_model = "all-MiniLM-L6-v2",
                               include_terms = TRUE) {

  if (is.null(query) || nchar(trimws(query)) == 0) {
    stop("Query cannot be empty")
  }

  tryCatch({
    if (!requireNamespace("reticulate", quietly = TRUE)) {
      stop("reticulate package is required for topic similarity")
    }
    sentence_transformers <- reticulate::import("sentence_transformers")
    model <- sentence_transformers$SentenceTransformer(embedding_model)
    query_embedding <- model$encode(query, show_progress_bar = FALSE, normalize_embeddings = TRUE)

    if ("beta" %in% names(topic_model) || "theta" %in% names(topic_model)) {
      message("Using STM model...")

      beta_td <- tidytext::tidy(topic_model, matrix = "beta")
      top_terms_by_topic <- beta_td %>%
        dplyr::group_by(topic) %>%
        dplyr::slice_max(order_by = beta, n = 10) %>%
        dplyr::ungroup()

      unique_topics <- unique(top_terms_by_topic$topic)
      topic_representations <- list()

      for (topic in unique_topics) {
        topic_terms <- top_terms_by_topic %>%
          dplyr::filter(topic == !!topic) %>%
          dplyr::pull(term)

        topic_text <- paste(topic_terms, collapse = " ")
        topic_embedding <- model$encode(topic_text, show_progress_bar = FALSE, normalize_embeddings = TRUE)
        topic_representations[[as.character(topic)]] <- topic_embedding
      }

      sklearn_metrics <- reticulate::import("sklearn.metrics.pairwise")
      similarities <- numeric(length(unique_topics))

      for (i in seq_along(unique_topics)) {
        topic <- unique_topics[i]
        topic_embedding <- topic_representations[[as.character(topic)]]
        similarity <- sklearn_metrics$cosine_similarity(
          matrix(query_embedding, nrow = 1),
          matrix(topic_embedding, nrow = 1)
        )[1, 1]
        similarities[i] <- similarity
      }

      topic_terms <- NULL
      if (include_terms) {
        topic_terms <- top_terms_by_topic %>%
          dplyr::group_by(topic) %>%
          dplyr::slice_max(order_by = beta, n = 5) %>%
          dplyr::ungroup() %>%
          dplyr::group_by(topic) %>%
          dplyr::summarise(top_terms = paste(term, collapse = ", "), .groups = "drop")
      }

    } else if ("embeddings" %in% names(topic_model) && "topic_assignments" %in% names(topic_model)) {
      message("Using semantic topic model...")

      unique_topics <- unique(topic_model$topic_assignments)
      topic_centroids <- matrix(0, nrow = length(unique_topics), ncol = ncol(topic_model$embeddings))

      for (i in seq_along(unique_topics)) {
        topic <- unique_topics[i]
        topic_docs <- which(topic_model$topic_assignments == topic)
        topic_centroids[i, ] <- colMeans(topic_model$embeddings[topic_docs, , drop = FALSE])
      }

      sklearn_metrics <- reticulate::import("sklearn.metrics.pairwise")
      similarities <- sklearn_metrics$cosine_similarity(
        matrix(query_embedding, nrow = 1),
        topic_centroids
      )[1, ]

      topic_terms <- NULL
      if (include_terms && "topic_keywords" %in% names(topic_model)) {
        topic_terms <- data.frame(
          topic = names(topic_model$topic_keywords),
          top_terms = sapply(topic_model$topic_keywords, function(x) {
            if (length(x) > 0) paste(head(x, 5), collapse = ", ") else "No terms"
          }),
          stringsAsFactors = FALSE
        )
      }

    } else {
      stop("Unsupported topic model type. Expected STM model (with 'beta'/'theta') or semantic model (with 'embeddings'/'topic_assignments')")
    }

    topic_similarities <- data.frame(
      topic = unique_topics,
      similarity = similarities,
      stringsAsFactors = FALSE
    )

    topic_similarities <- topic_similarities[order(topic_similarities$similarity, decreasing = TRUE), ]

    result <- list(
      similar_topics = topic_similarities$topic[seq_len(min(top_n, nrow(topic_similarities)))],
      similarity_scores = topic_similarities$similarity[seq_len(min(top_n, nrow(topic_similarities)))],
      query = query,
      method = method,
      topic_terms = topic_terms
    )

    return(result)

  }, error = function(e) {
    stop("Error finding similar topics: ", e$message)
  })
}


#' @title Calculate NPMI for Topic Top-Terms
#'
#' @description
#' NPMI (Normalized Pointwise Mutual Information) measures coherence of top-K
#' terms per topic using internal document co-occurrence from the supplied DFM
#' (boolean presence).
#'
#' @param top_terms_list Named or unnamed list of character vectors, one per topic.
#' @param dfm A quanteda dfm providing the reference doc-term frequencies.
#' @param top_k Number of top terms per topic to include (default 10).
#' @param epsilon Numerical stabilizer (default 1e-12).
#' @return list with `mean_npmi` (scalar) and `per_topic_npmi` (numeric vector).
#' @importFrom methods as
#' @concept topic-modeling
#' @keywords internal
calculate_npmi <- function(top_terms_list, dfm, top_k = 10, epsilon = 1e-12) {
  if (is.null(top_terms_list) || length(top_terms_list) == 0 || is.null(dfm)) {
    return(list(mean_npmi = NA_real_, per_topic_npmi = numeric(0)))
  }
  dtm <- methods::as(dfm > 0, "CsparseMatrix")
  n_docs <- nrow(dtm)
  if (n_docs < 2) return(list(mean_npmi = NA_real_, per_topic_npmi = numeric(0)))
  vocab <- colnames(dtm)

  per_topic <- vapply(top_terms_list, function(terms) {
    terms <- utils::head(as.character(terms), top_k)
    terms <- terms[terms %in% vocab]
    if (length(terms) < 2) return(NA_real_)
    sub_dtm <- dtm[, terms, drop = FALSE]
    doc_freq <- Matrix::colSums(sub_dtm)
    pairs <- utils::combn(seq_along(terms), 2)
    npmis <- vapply(seq_len(ncol(pairs)), function(k) {
      i <- pairs[1, k]; j <- pairs[2, k]
      p_i <- doc_freq[i] / n_docs
      p_j <- doc_freq[j] / n_docs
      co <- sum(sub_dtm[, i] * sub_dtm[, j])
      p_ij <- co / n_docs
      # NPMI lower bound: words that never co-occur score -1
      if (p_ij < epsilon) return(-1)
      pmi <- log(p_ij / (p_i * p_j + epsilon))
      pmi / (-log(p_ij + epsilon))
    }, numeric(1))
    mean(npmis, na.rm = TRUE)
  }, numeric(1))

  list(mean_npmi = mean(per_topic, na.rm = TRUE), per_topic_npmi = per_topic)
}

#' Calculate Topic Diversity (Unique-Word Proportion)
#'
#' Diversity = unique top-K terms across all topics / (n_topics * top_k).
#' Range 0 (all topics identical) to 1 (fully disjoint). Complementary to NPMI
#' coherence: a good model scores high on both.
#'
#' @param top_terms_list Named or unnamed list of character vectors, one per topic.
#' @param top_k Number of top terms per topic to include (default 25).
#' @return Numeric between 0 and 1, or NA if input empty.
#' @concept topic-modeling
#' @keywords internal
calculate_topic_diversity <- function(top_terms_list, top_k = 25) {
  if (is.null(top_terms_list) || length(top_terms_list) == 0) return(NA_real_)
  terms_per_topic <- lapply(top_terms_list, function(t) utils::head(as.character(t), top_k))
  actual_total <- sum(lengths(terms_per_topic))
  if (actual_total == 0) return(NA_real_)
  length(unique(unlist(terms_per_topic))) / actual_total
}


#' @title Auto-tune BERTopic Hyperparameters
#'
#' @description
#' Automatically searches for optimal hyperparameters for embedding-based topic modeling.
#' Evaluates multiple configurations of UMAP and HDBSCAN parameters and returns the best
#' model based on the specified metric. Embeddings are generated once and reused across
#' all configurations for efficiency.
#'
#' @param texts Character vector of documents to analyze.
#' @param embeddings Precomputed embeddings matrix (optional). If NULL, embeddings are generated.
#' @param embedding_model Embedding model name (default: "all-MiniLM-L6-v2").
#' @param n_trials Maximum number of configurations to try (default: 12).
#' @param metric Optimization metric: "silhouette", "coherence", or "combined" (default: "silhouette").
#' @param seed Random seed for reproducibility.
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A list containing:
#'   - best_config: Data frame with the optimal hyperparameter configuration
#'   - best_model: The topic model fitted with optimal parameters
#'   - all_results: List of all evaluated configurations with metrics
#'   - n_trials_completed: Number of configurations successfully evaluated
#'
#' @details
#' The function searches over these parameters:
#' - n_neighbors: UMAP neighborhood size (5, 10, 15, 25)
#' - min_cluster_size: HDBSCAN minimum cluster size (3, 5, 10)
#' - cluster_selection_method: "eom" (broader) or "leaf" (finer-grained)
#'
#' @concept topic-modeling
#' @export
#' @examples
#' if (interactive()) {
#'   texts <- c("Machine learning for image recognition",
#'              "Deep learning neural networks",
#'              "Natural language processing models",
#'              "Computer vision applications")
#'
#'   tuning_result <- auto_tune_embedding_topics(
#'     texts = texts,
#'     n_trials = 6,
#'     metric = "silhouette",
#'     verbose = TRUE
#'   )
#'
#'   # View best configuration
#'   tuning_result$best_config
#'
#'   # Use the best model
#'   best_model <- tuning_result$best_model
#' }
auto_tune_embedding_topics <- function(
    texts,
    embeddings = NULL,
    embedding_model = "all-MiniLM-L6-v2",
    n_trials = 12,
    metric = "silhouette",
    seed = 123,
    verbose = TRUE
) {

  if (verbose) message("Starting hyperparameter auto-tuning for embedding topics...")

  # Validate inputs
  if (is.null(texts) || length(texts) == 0) {
    stop("No texts provided for analysis")
  }

  valid_metrics <- c("silhouette", "coherence", "combined")
  if (!metric %in% valid_metrics) {
    stop("metric must be one of: ", paste(valid_metrics, collapse = ", "))
  }

  # Define parameter grid
  param_grid <- expand.grid(
    n_neighbors = c(5, 10, 15, 25),
    min_cluster_size = c(3, 5, 10),
    cluster_selection_method = c("eom", "leaf"),
    stringsAsFactors = FALSE
  )

  # Sample configurations if grid is larger than n_trials
  if (nrow(param_grid) > n_trials) {
    withr::local_seed(seed)
    sampled_rows <- sample(nrow(param_grid), n_trials)
    param_grid <- param_grid[sampled_rows, ]
  }

  if (verbose) {
    message("  Testing ", nrow(param_grid), " hyperparameter configurations")
  }

  # Generate embeddings once if not provided
  if (is.null(embeddings)) {
    if (verbose) message("  Generating embeddings (one-time cost)...")

    embeddings <- tryCatch({
      generate_embeddings(texts, model = embedding_model, verbose = FALSE)
    }, error = function(e) {
      stop("Failed to generate embeddings: ", e$message)
    })

    if (is.null(embeddings)) {
      stop("Embedding generation unavailable. ",
           "Run setup_python_env() or pass precomputed embeddings.",
           call. = FALSE)
    }
  }

  # Evaluate each configuration
  results <- list()
  successful <- 0

  for (i in seq_len(nrow(param_grid))) {
    config <- param_grid[i, ]

    if (verbose) {
      message("  [", i, "/", nrow(param_grid), "] Testing: ",
              "n_neighbors=", config$n_neighbors,
              ", min_cluster_size=", config$min_cluster_size,
              ", method=", config$cluster_selection_method)
    }

    model_result <- tryCatch({
      fit_embedding_model(
        texts = texts,
        method = "umap_hdbscan",
        embedding_model = embedding_model,
        precomputed_embeddings = embeddings,
        umap_neighbors = config$n_neighbors,
        min_topic_size = config$min_cluster_size,
        cluster_selection_method = config$cluster_selection_method,
        seed = seed,
        verbose = FALSE
      )
    }, error = function(e) {
      if (verbose) message("    Configuration failed: ", e$message)
      NULL
    })

    if (!is.null(model_result)) {
      silhouette_score <- .embedding_silhouette(model_result)
      coherence_score <- model_result$quality_metrics$mean_topic_coherence %||% 0
      combined_score <- mean(c(silhouette_score, coherence_score), na.rm = TRUE)

      results[[length(results) + 1]] <- list(
        config = config,
        silhouette = silhouette_score,
        coherence = coherence_score,
        combined = combined_score,
        n_topics = model_result$n_topics,
        model = model_result
      )

      successful <- successful + 1

      if (verbose) {
        message("    -> Topics: ", model_result$n_topics,
                ", Silhouette: ", round(silhouette_score, 3),
                ", Coherence: ", round(coherence_score, 3))
      }
    }
  }

  if (successful == 0) {
    stop("All configurations failed. Check your data and Python dependencies.")
  }

  # Select best configuration by metric
  scores <- sapply(results, function(r) r[[metric]])
  best_idx <- which.max(scores)

  if (verbose) {
    message("\nBest configuration (by ", metric, "):")
    message("  n_neighbors: ", results[[best_idx]]$config$n_neighbors)
    message("  min_cluster_size: ", results[[best_idx]]$config$min_cluster_size)
    message("  cluster_selection_method: ", results[[best_idx]]$config$cluster_selection_method)
    message("  ", metric, " score: ", round(results[[best_idx]][[metric]], 3))
    message("  Number of topics: ", results[[best_idx]]$n_topics)
  }

  list(
    best_config = results[[best_idx]]$config,
    best_model = results[[best_idx]]$model,
    all_results = results,
    n_trials_completed = successful
  )
}


#' @title Assess Embedding Topic Model Stability
#'
#' @description
#' Evaluates the stability of embedding-based topic modeling by running multiple models
#' with different random seeds and comparing their results. High stability (high ARI,
#' consistent keywords) indicates stable topic structure in the data.
#'
#' @param texts Character vector of documents to analyze.
#' @param n_runs Number of model runs with different seeds (default: 5).
#'   More runs give a steadier stability estimate.
#' @param embedding_model Embedding model name (default: "all-MiniLM-L6-v2").
#' @param select_best Logical, if TRUE, returns the best model by quality (default: TRUE).
#' @param base_seed Base random seed; each run uses base_seed + (run - 1).
#' @param verbose Logical, if TRUE, prints progress messages.
#' @param ... Additional arguments passed to fit_embedding_model().
#'
#' @return A list containing:
#'   - stability_metrics: List with mean_ari, sd_ari, mean_jaccard, quality_variance
#'   - best_model: Best model by silhouette score (if select_best = TRUE)
#'   - all_models: List of all fitted models
#'   - is_stable: Logical, TRUE if mean ARI >= 0.6 (considered stable)
#'   - recommendation: Text recommendation based on stability
#'
#' @details
#' Stability is assessed via:
#' - Adjusted Rand Index (ARI): Measures agreement in topic assignments across runs
#' - Keyword Jaccard similarity: Measures overlap in top keywords per topic
#' - Quality variance: Variance in silhouette scores across runs
#'
#' @concept topic-modeling
#' @export
#' @examples
#' if (interactive()) {
#'   texts <- c("Machine learning for image recognition",
#'              "Deep learning neural networks",
#'              "Natural language processing models")
#'
#'   stability <- assess_embedding_stability(
#'     texts = texts,
#'     n_runs = 3,
#'     verbose = TRUE
#'   )
#'
#'   # Check if results are stable
#'   stability$is_stable
#'   stability$stability_metrics$mean_ari
#'
#'   # Use the best model
#'   best_model <- stability$best_model
#' }
assess_embedding_stability <- function(
    texts,
    n_runs = 5,
    embedding_model = "all-MiniLM-L6-v2",
    select_best = TRUE,
    base_seed = 123,
    verbose = TRUE,
    ...
) {

  if (verbose) message("Assessing embedding topic model stability across ", n_runs, " runs...")

  # Validate inputs
  if (is.null(texts) || length(texts) == 0) {
    stop("No texts provided for analysis")
  }

  if (n_runs < 2) {
    stop("n_runs must be at least 2 to assess stability")
  }

  # Generate embeddings once for efficiency
  if (verbose) message("  Generating embeddings (one-time cost)...")
  embeddings <- tryCatch({
    generate_embeddings(texts, model = embedding_model, verbose = FALSE)
  }, error = function(e) {
    stop("Failed to generate embeddings: ", e$message)
  })

  if (is.null(embeddings)) {
    stop("Embedding generation unavailable. Run setup_python_env() first.",
         call. = FALSE)
  }

  # Run models with different seeds
  models <- list()
  for (i in seq_len(n_runs)) {
    seed_i <- base_seed + i - 1

    if (verbose) message("  Running model ", i, "/", n_runs, " (seed=", seed_i, ")...")

    model <- tryCatch({
      fit_embedding_model(
        texts = texts,
        embedding_model = embedding_model,
        precomputed_embeddings = embeddings,
        seed = seed_i,
        verbose = FALSE,
        ...
      )
    }, error = function(e) {
      if (verbose) message("    Run ", i, " failed: ", e$message)
      NULL
    })

    if (!is.null(model)) {
      models[[length(models) + 1]] <- model
    }
  }

  if (length(models) < 2) {
    stop("Need at least 2 successful runs to assess stability. Only ", length(models), " succeeded.")
  }

  if (verbose) message("  Computing stability metrics...")

  # degenerate single-cluster runs distort mean ARI
  degenerate <- vapply(models, function(m) {
    length(unique(m$topic_assignments[m$topic_assignments >= 0])) <= 1
  }, logical(1))
  if (any(degenerate) && sum(!degenerate) >= 2) {
    if (verbose) message("  Excluding ", sum(degenerate), " degenerate run(s)")
    models <- models[!degenerate]
  }

  # Calculate pairwise Adjusted Rand Index
  n_successful <- length(models)
  ari_values <- c()

  for (i in 1:(n_successful - 1)) {
    for (j in (i + 1):n_successful) {
      assignments_i <- models[[i]]$topic_assignments
      assignments_j <- models[[j]]$topic_assignments

      # Calculate ARI using contingency table approach
      ari <- tryCatch({
        if (requireNamespace("aricode", quietly = TRUE)) {
          aricode::ARI(assignments_i, assignments_j)
        } else {
          # Fallback: simple agreement rate
          mean(assignments_i == assignments_j)
        }
      }, error = function(e) NA)

      if (!is.na(ari)) {
        ari_values <- c(ari_values, ari)
      }
    }
  }

  # Calculate keyword stability (Jaccard similarity)
  jaccard_values <- c()

  for (i in 1:(n_successful - 1)) {
    for (j in (i + 1):n_successful) {
      keywords_i <- unlist(models[[i]]$topic_keywords)
      keywords_j <- unlist(models[[j]]$topic_keywords)

      if (length(keywords_i) > 0 && length(keywords_j) > 0) {
        intersection <- length(intersect(keywords_i, keywords_j))
        union_size <- length(union(keywords_i, keywords_j))
        jaccard <- if (union_size > 0) intersection / union_size else 0
        jaccard_values <- c(jaccard_values, jaccard)
      }
    }
  }

  silhouette_scores <- vapply(models, .embedding_silhouette, numeric(1))

  stability_metrics <- list(
    mean_ari = if (length(ari_values) > 0) mean(ari_values) else NA,
    sd_ari = if (length(ari_values) > 1) sd(ari_values) else NA,
    mean_jaccard = if (length(jaccard_values) > 0) mean(jaccard_values) else NA,
    sd_jaccard = if (length(jaccard_values) > 1) sd(jaccard_values) else NA,
    mean_silhouette = if (any(!is.na(silhouette_scores))) mean(silhouette_scores, na.rm = TRUE) else NA,
    quality_variance = if (sum(!is.na(silhouette_scores)) > 1) var(silhouette_scores, na.rm = TRUE) else NA,
    n_successful_runs = n_successful
  )

  best_idx <- if (any(!is.na(silhouette_scores))) which.max(silhouette_scores) else 1L
  best_model <- if (select_best) models[[best_idx]] else NULL

  # Determine stability status and recommendation
  # 0.8 / 0.6 / 0.4 tiers are package heuristics, not published cutoffs
  is_stable <- !is.na(stability_metrics$mean_ari) && stability_metrics$mean_ari >= 0.6

  recommendation <- if (is.na(stability_metrics$mean_ari)) {
    "Could not compute stability metrics."
  } else if (stability_metrics$mean_ari >= 0.8) {
    "Highly stable: Results are very consistent across runs."
  } else if (stability_metrics$mean_ari >= 0.6) {
    "Moderately stable: Results are reasonably consistent."
  } else if (stability_metrics$mean_ari >= 0.4) {
    "Low stability: Consider adjusting parameters or checking data quality."
  } else {
    "Unstable: Topic structure may not be well-defined. Try different parameters."
  }

  if (verbose) {
    message("\nStability Results:")
    message("  Mean ARI: ", round(stability_metrics$mean_ari, 3))
    message("  Mean Keyword Jaccard: ", round(stability_metrics$mean_jaccard, 3))
    message("  Quality Variance: ", round(stability_metrics$quality_variance, 4))
    message("  Status: ", if (is_stable) "STABLE" else "UNSTABLE")
    message("  ", recommendation)
  }

  list(
    stability_metrics = stability_metrics,
    best_model = best_model,
    all_models = models,
    is_stable = is_stable,
    recommendation = recommendation
  )
}


#' @title Generate Semantic Topic Keywords (c-TF-IDF)
#'
#' @description
#' Generate keywords for topics using c-TF-IDF (class-based TF-IDF), similar to BERTopic.
#' This method treats all documents in a topic as a single document and calculates TF-IDF
#' scores relative to other topics.
#'
#' @param texts A character vector of texts.
#' @param topic_assignments A vector of topic assignments.
#' @param n_keywords The number of keywords to extract per topic (default: 10).
#' @param method The representation method: "c-tfidf" (default), "tfidf", "mmr", or "frequency".
#' @param diversity Diversity weight for "mmr" between 0 and 1 (default: 0.5).
#'   MMR selects terms by (1 - diversity) * relevance - diversity * redundancy.
#'
#' @return A list of keywords for each topic.
#'
#' @keywords internal
generate_semantic_topic_keywords <- function(texts,
                                            topic_assignments,
                                            n_keywords = 10,
                                            method = "c-tfidf",
                                            diversity = 0.5) {

  tryCatch({
    unique_topics <- sort(unique(topic_assignments[topic_assignments >= 0]))
    topic_keywords <- list()

    if (method %in% c("c-tfidf", "mmr")) {
      topic_docs <- vapply(unique_topics, function(topic) {
        paste(texts[topic_assignments == topic], collapse = " ")
      }, character(1))

      class_tokens <- quanteda::tokens(quanteda::corpus(topic_docs),
                                       remove_punct = TRUE,
                                       remove_numbers = TRUE,
                                       remove_symbols = TRUE)
      class_tokens <- quanteda::tokens_tolower(class_tokens)
      class_tokens <- quanteda::tokens_remove(class_tokens, quanteda::stopwords("english"))

      # c-TF-IDF: W_tc = tf_tc * log(1 + A / tf_t), A = mean words per class
      tf_mat <- as.matrix(quanteda::dfm(class_tokens))
      avg_class_size <- mean(rowSums(tf_mat))
      term_totals <- colSums(tf_mat)
      ctfidf_mat <- sweep(tf_mat, 2, log(1 + avg_class_size / term_totals), `*`)

      if (method == "c-tfidf") {
        for (i in seq_along(unique_topics)) {
          scores <- ctfidf_mat[i, ]
          scores <- scores[is.finite(scores)]
          top_idx <- order(scores, decreasing = TRUE)[seq_len(min(n_keywords, length(scores)))]
          topic_keywords[[as.character(unique_topics[i])]] <- names(scores)[top_idx]
        }
      } else {
        doc_tokens <- quanteda::tokens(quanteda::corpus(texts),
                                       remove_punct = TRUE,
                                       remove_numbers = TRUE,
                                       remove_symbols = TRUE)
        doc_tokens <- quanteda::tokens_tolower(doc_tokens)
        doc_tokens <- quanteda::tokens_remove(doc_tokens, quanteda::stopwords("english"))
        doc_dfm <- quanteda::dfm(doc_tokens)

        lambda <- 1 - diversity
        for (i in seq_along(unique_topics)) {
          topic <- unique_topics[i]
          relevance <- ctfidf_mat[i, ]
          relevance <- relevance[relevance > 0 & is.finite(relevance)]

          # candidate pool of 3x keywords bounds the pairwise similarity cost
          n_candidates <- min(3 * n_keywords, length(relevance))
          candidates <- names(sort(relevance, decreasing = TRUE))[seq_len(n_candidates)]
          candidates <- intersect(candidates, colnames(doc_dfm))

          if (length(candidates) == 0) {
            topic_keywords[[as.character(topic)]] <- character(0)
            next
          }

          # redundancy = cosine between term occurrence vectors over topic docs
          occurrence <- as.matrix(doc_dfm[which(topic_assignments == topic), candidates, drop = FALSE])
          unit_cols <- sweep(occurrence, 2, pmax(sqrt(colSums(occurrence^2)), 1e-12), `/`)
          term_sim <- crossprod(unit_cols)

          rel <- relevance[candidates] / max(relevance[candidates])
          selected <- candidates[which.max(rel)]
          remaining <- setdiff(candidates, selected)

          while (length(selected) < min(n_keywords, length(candidates)) && length(remaining) > 0) {
            max_sim <- apply(term_sim[remaining, selected, drop = FALSE], 1, max)
            mmr_score <- lambda * rel[remaining] - (1 - lambda) * max_sim
            pick <- remaining[which.max(mmr_score)]
            selected <- c(selected, pick)
            remaining <- setdiff(remaining, pick)
          }
          topic_keywords[[as.character(topic)]] <- selected
        }
      }

    } else if (method == "tfidf") {
      for (topic in unique_topics) {
        topic_texts <- texts[topic_assignments == topic]

        if (length(topic_texts) < 1) {
          topic_keywords[[as.character(topic)]] <- character(0)
          next
        }

        corpus <- quanteda::corpus(topic_texts)
        tokens <- quanteda::tokens(corpus,
                                   remove_punct = TRUE,
                                   remove_numbers = TRUE)
        tokens <- quanteda::tokens_tolower(tokens)
        tokens <- quanteda::tokens_remove(tokens, quanteda::stopwords("english"))

        dfm <- quanteda::dfm(tokens)

        tfidf <- quanteda::dfm_tfidf(dfm)

        mean_tfidf <- colMeans(as.matrix(tfidf))
        mean_tfidf <- sort(mean_tfidf, decreasing = TRUE)

        top_terms <- names(head(mean_tfidf, n_keywords))
        topic_keywords[[as.character(topic)]] <- top_terms
      }

    } else {
      return(generate_topic_keywords(texts, topic_assignments, n_keywords))
    }

    return(topic_keywords)

  }, error = function(e) {
    warning("Error in c-TF-IDF calculation: ", e$message)
    return(generate_topic_keywords(texts, topic_assignments, n_keywords))
  })
}


#' @title Generate Topic Keywords
#'
#' @description
#' Internal function to generate keywords for topics using TF-IDF analysis.
#'
#' @param texts A character vector of texts.
#' @param topic_assignments A vector of topic assignments.
#' @param n_keywords The number of keywords to extract per topic.
#'
#' @return A list of keywords for each topic.
#'
#' @keywords internal
generate_topic_keywords <- function(texts, topic_assignments, n_keywords = 10) {
  tryCatch({
    topic_keywords <- list()
    unique_topics <- unique(topic_assignments)

    for (topic in unique_topics) {
      topic_texts <- texts[topic_assignments == topic]

      if (length(topic_texts) < 2) {
        topic_keywords[[as.character(topic)]] <- character(0)
        next
      }

      corpus <- quanteda::corpus(topic_texts)
      tokens <- quanteda::tokens(corpus,
                                 remove_punct = TRUE,
                                 remove_numbers = TRUE,
                                 remove_symbols = TRUE,
                                 remove_separators = TRUE)
      tokens <- quanteda::tokens_tolower(tokens)
      tokens <- quanteda::tokens_remove(tokens, quanteda::stopwords("english"))

      dfm <- quanteda::dfm(tokens)
      dfm <- quanteda::dfm_trim(dfm, min_termfreq = 2, min_docfreq = 1)

      if (quanteda::nfeat(dfm) == 0) {
        topic_keywords[[as.character(topic)]] <- character(0)
        next
      }

      top_terms <- quanteda.textstats::textstat_frequency(dfm, n = n_keywords)
      topic_keywords[[as.character(topic)]] <- top_terms$feature
    }

    return(topic_keywords)

  }, error = function(e) {
    warning("Error generating topic keywords: ", e$message)
    return(list())
  })
}

#' @title Calculate Semantic Topic Quality Metrics
#'
#' @description
#' Internal function to calculate quality metrics for semantic topic modeling results.
#'
#' @param embeddings Document embeddings matrix.
#' @param topic_assignments Vector of topic assignments.
#' @param similarity_matrix Optional similarity matrix.
#'
#' @return A list of quality metrics.
#'
#' @keywords internal
calculate_topic_quality <- function(embeddings, topic_assignments, similarity_matrix = NULL) {
  tryCatch({
    metrics <- list()

    unique_topics <- unique(topic_assignments)

    # Vectorized topic coherence calculation using vapply
    topic_coherence <- vapply(unique_topics, function(topic) {
      topic_docs <- which(topic_assignments == topic)
      if (length(topic_docs) > 1) {
        if (!is.null(similarity_matrix)) {
          topic_sim <- similarity_matrix[topic_docs, topic_docs]
          mean(topic_sim[upper.tri(topic_sim)], na.rm = TRUE)
        } else {
          topic_embeddings <- embeddings[topic_docs, , drop = FALSE]
          topic_sim <- as.matrix(stats::dist(topic_embeddings, method = "euclidean"))
          1 - mean(topic_sim[upper.tri(topic_sim)], na.rm = TRUE)
        }
      } else {
        NA_real_
      }
    }, FUN.VALUE = numeric(1))

    metrics$mean_topic_coherence <- mean(topic_coherence, na.rm = TRUE)
    metrics$topic_coherence_sd <- sd(topic_coherence, na.rm = TRUE)

    if (length(unique_topics) > 1) {
      # Vectorized centroid calculation using vapply
      topic_centroids <- t(vapply(unique_topics, function(topic) {
        topic_docs <- which(topic_assignments == topic)
        colMeans(embeddings[topic_docs, , drop = FALSE])
      }, FUN.VALUE = numeric(ncol(embeddings))))

      centroid_distances <- as.matrix(stats::dist(topic_centroids, method = "euclidean"))
      metrics$mean_topic_separation <- mean(centroid_distances[upper.tri(centroid_distances)], na.rm = TRUE)
    } else {
      metrics$mean_topic_separation <- NA
    }

    topic_sizes <- table(topic_assignments)
    metrics$topic_size_mean <- mean(topic_sizes)
    metrics$topic_size_sd <- sd(topic_sizes)
    metrics$topic_size_min <- min(topic_sizes)
    metrics$topic_size_max <- max(topic_sizes)

    if (!is.na(metrics$mean_topic_coherence) && !is.na(metrics$mean_topic_separation)) {
      metrics$overall_quality <- metrics$mean_topic_coherence * (1 / (1 + metrics$mean_topic_separation))
    } else {
      metrics$overall_quality <- NA
    }

    return(metrics)

  }, error = function(e) {
    warning("Error calculating quality metrics: ", e$message)
    return(list())
  })
}

#' @title Fit Temporal Topic Model
#' @description Analyzes how topics evolve over time by fitting topic models to
#'   different time periods and tracking semantic changes.
#' @param texts A character vector of text documents to analyze.
#' @param dates A vector of dates corresponding to each document (will be converted to Date).
#' @param time_windows Time grouping strategy: "yearly", "monthly", or "quarterly" (default: "yearly").
#' @param embeddings Optional pre-computed embeddings matrix. If NULL, embeddings will be generated.
#' @param verbose Logical indicating whether to print progress messages (default: TRUE).
#' @return A list containing temporal analysis results with topic evolution patterns.
#' @concept topic-modeling
#' @export
fit_temporal_model <- function(texts,
                                     dates,
                                     time_windows = "yearly",
                                     embeddings = NULL,
                                     verbose = TRUE) {

  if (verbose) message("Starting temporal semantic analysis...")

  dates <- as.Date(dates)

  if (time_windows == "yearly") {
    time_groups <- format(dates, "%Y")
  } else if (time_windows == "monthly") {
    time_groups <- format(dates, "%Y-%m")
  } else if (time_windows == "quarterly") {
    time_groups <- paste(format(dates, "%Y"), "Q", (as.numeric(format(dates, "%m")) - 1) %/% 3 + 1, sep = "")
  }

  time_periods <- unique(time_groups)
  temporal_results <- list()

  for (period in time_periods) {
    period_indices <- which(time_groups == period)
    period_texts <- texts[period_indices]

    if (length(period_texts) < 5) {
      if (verbose) message("Skipping period ", period, " (insufficient documents)")
      next
    }

    if (verbose) message("Analyzing period: ", period)

    period_embeddings <- if (!is.null(embeddings)) {
      embeddings[period_indices, , drop = FALSE]
    } else {
      pe <- generate_embeddings(period_texts, verbose = FALSE)
      if (is.null(pe)) {
        stop("Embedding generation unavailable. Run setup_python_env() first.",
             call. = FALSE)
      }
      pe
    }

    period_results <- fit_embedding_model(
      texts = period_texts,
      method = "umap_hdbscan",
      n_topics = 10,
      embedding_model = "all-MiniLM-L6-v2",
      verbose = FALSE
    )

    temporal_results[[period]] <- list(
      texts = period_texts,
      embeddings = period_embeddings,
      topic_assignments = period_results$topic_assignments,
      topic_keywords = period_results$topic_keywords,
      n_documents = length(period_texts),
      n_topics = period_results$n_topics
    )
  }

  evolution_patterns <- analyze_semantic_evolution(temporal_results, verbose = verbose)

  result <- list(
    temporal_results = temporal_results,
    evolution_patterns = evolution_patterns,
    time_windows = time_windows,
    periods_analyzed = names(temporal_results)
  )

  if (verbose) message("Temporal analysis completed. Analyzed ", length(temporal_results), " periods")

  return(result)
}

calculate_topic_correspondence <- function(semantic_keywords, stm_keywords) {

  correspondence_matrix <- matrix(0,
                                 nrow = length(semantic_keywords),
                                 ncol = length(stm_keywords))

  for (i in seq_along(semantic_keywords)) {
    for (j in seq_along(stm_keywords)) {
      semantic_terms <- semantic_keywords[[i]]
      stm_terms <- stm_keywords[[j]]

      intersection <- length(intersect(semantic_terms, stm_terms))
      union <- length(union(semantic_terms, stm_terms))

      correspondence_matrix[i, j] <- if (union > 0) intersection / union else 0
    }
  }

  return(list(
    correspondence_matrix = correspondence_matrix,
    mean_correspondence = mean(correspondence_matrix),
    max_correspondence = max(correspondence_matrix)
  ))
}

#' @title Validate Semantic Coherence
#'
#' @description
#' Validates semantic coherence of topic assignments.
#'
#' @param embeddings Document embeddings.
#' @param topic_assignments Topic assignments.
#'
#' @return Coherence metrics.
#'
#' @keywords internal
calculate_coherence <- function(embeddings, topic_assignments) {

  unique_topics <- unique(topic_assignments)
  coherence_scores <- numeric(length(unique_topics))

  for (i in seq_along(unique_topics)) {
    topic_docs <- which(topic_assignments == unique_topics[i])

    if (length(topic_docs) > 1) {
      topic_embeddings <- embeddings[topic_docs, , drop = FALSE]
      norms <- sqrt(rowSums(topic_embeddings^2))
      keep <- norms > 0
      if (sum(keep) > 1) {
        norm_emb <- topic_embeddings[keep, , drop = FALSE] / norms[keep]
        sim <- norm_emb %*% t(norm_emb)
        coherence_scores[i] <- mean(sim[upper.tri(sim)], na.rm = TRUE)
      } else {
        coherence_scores[i] <- NA
      }
    } else {
      coherence_scores[i] <- NA
    }
  }

  list(
    coherence_scores = coherence_scores,
    mean_coherence = mean(coherence_scores, na.rm = TRUE),
    topic_coherence = setNames(coherence_scores, unique_topics)
  )
}

#' @title Calculate Topic Stability
#'
#' @description
#' Calculates stability of topics across consecutive time periods. Each
#' period is fitted independently, so topics are matched one-to-one on
#' keyword Jaccard similarity before scoring
#' (see [calculate_keyword_stability()]).
#'
#' @param temporal_results Results from temporal analysis.
#'
#' @return Stability metrics: matched-pair stability per consecutive-period
#'   transition and their mean.
#'
#' @concept topic-modeling
#' @export
calculate_topic_stability <- function(temporal_results) {

  periods <- names(temporal_results)
  stability_scores <- numeric(length(periods) - 1)

  for (i in 1:(length(periods) - 1)) {
    period1 <- periods[i]
    period2 <- periods[i + 1]

    keywords1 <- temporal_results[[period1]]$topic_keywords
    keywords2 <- temporal_results[[period2]]$topic_keywords

    stability_scores[i] <- calculate_keyword_stability(keywords1, keywords2)
  }

  return(list(
    stability_scores = stability_scores,
    mean_stability = mean(stability_scores, na.rm = TRUE),
    periods = paste(periods[-length(periods)], periods[-1], sep = " -> ")
  ))
}

#' @title Calculate Keyword Stability
#'
#' @description
#' Calculates stability between the topic keyword sets of two independently
#' fitted models. Topic numbering is arbitrary across fits, so topics are
#' matched one-to-one by greedy assignment on the pairwise keyword Jaccard
#' matrix before averaging matched-pair similarity.
#'
#' @param keywords1 List of keyword vectors, one per topic, from the first model.
#' @param keywords2 List of keyword vectors, one per topic, from the second model.
#'
#' @return Mean Jaccard similarity of matched topic pairs (0-1).
#'
#' @concept topic-modeling
#' @export
calculate_keyword_stability <- function(keywords1, keywords2) {

  if (length(keywords1) == 0 || length(keywords2) == 0) {
    return(0)
  }

  jaccard <- calculate_topic_correspondence(keywords1, keywords2)$correspondence_matrix

  # greedy one-to-one matching: repeatedly pair the most similar topics
  n_matches <- min(nrow(jaccard), ncol(jaccard))
  matched <- numeric(n_matches)
  for (k in seq_len(n_matches)) {
    best <- arrayInd(which.max(jaccard), dim(jaccard))
    matched[k] <- jaccard[best[1], best[2]]
    jaccard[best[1], ] <- -Inf
    jaccard[, best[2]] <- -Inf
  }

  mean(matched)
}

#' @title Calculate Topic-Cluster Correspondence
#' @description
#' Computes Jaccard-based correspondence between topic keywords and cluster keywords.
#'
#' @param topic_keywords Topic keywords list.
#' @param cluster_keywords Cluster keywords list.
#' @param ... Additional parameters (currently unused).
#'
#' @return List with correspondence metrics.
#'
#' @keywords internal
calculate_topic_cluster_correspondence <- function(topic_keywords, cluster_keywords, ...) {
  result <- calculate_topic_correspondence(topic_keywords, cluster_keywords)
  list(
    match_score = result$mean_correspondence,
    n_topics = length(topic_keywords),
    n_clusters = length(cluster_keywords),
    correspondence_matrix = result$correspondence_matrix
  )
}

#' @title Validate Semantic Coherence
#' @description Validates the semantic coherence of topic assignments using
#'   intra-cluster distance in embedding space.
#' @param embeddings Document embeddings matrix.
#' @param topic_assignments Vector of topic assignments for documents.
#' @param ... Additional parameters (currently unused).
#' @return List containing coherence score and metrics.
#' @concept topic-modeling
#' @export
validate_semantic_coherence <- function(embeddings, topic_assignments, ...) {
  result <- calculate_coherence(embeddings, topic_assignments)
  list(
    score = result$mean_coherence,
    n_topics = length(unique(topic_assignments)),
    per_topic = result$topic_coherence
  )
}

#' @title Calculate Assignment Consistency
#' @description Calculates consistency between two sets of assignments
#' @param assignments1 First set of assignments
#' @param assignments2 Second set of assignments
#' @param ... Additional parameters
#' @return List containing consistency metrics
#' @concept topic-modeling
#' @export
calculate_assignment_consistency <- function(assignments1, assignments2, ...) {
  if (length(assignments1) != length(assignments2)) {
    return(list(consistency = NA, message = "Assignment lengths differ"))
  }

  if (!requireNamespace("aricode", quietly = TRUE)) {
    return(list(
      consistency = NA,
      n_total = length(assignments1),
      message = "Install 'aricode' for permutation-invariant consistency (ARI)."
    ))
  }

  ari <- aricode::ARI(assignments1, assignments2)
  nmi <- aricode::NMI(assignments1, assignments2)

  list(
    consistency = ari,
    ari = ari,
    nmi = nmi,
    n_total = length(assignments1)
  )
}

#' @title Analyze Semantic Evolution
#' @description Analyzes semantic evolution patterns in temporal results
#' @param temporal_results Temporal analysis results
#' @param verbose Logical indicating whether to print progress messages
#' @param ... Additional parameters
#' @return List containing evolution analysis
#' @concept topic-modeling
#' @export
analyze_semantic_evolution <- function(temporal_results, verbose = FALSE, ...) {
  if (verbose) message("Analyzing semantic evolution...")

  periods <- names(temporal_results)
  if (length(periods) < 2) {
    return(list(
      periods_analyzed = length(periods),
      topic_stability = NA,
      evolution_patterns = list(emergence = character(), decline = character(), stable = character())
    ))
  }

  topic_stability <- calculate_topic_stability(temporal_results)
  semantic_drift <- calculate_semantic_drift(temporal_results)

  list(
    periods_analyzed = length(periods),
    topic_stability = topic_stability,
    semantic_drift = semantic_drift
  )
}

#' @title Calculate Semantic Drift
#' @description Calculates semantic drift across time periods
#' @param temporal_results Temporal analysis results
#' @param ... Additional parameters
#' @return List containing drift metrics
#' @concept topic-modeling
#' @export
calculate_semantic_drift <- function(temporal_results, ...) {
  if (is.null(temporal_results) || length(temporal_results) < 2) {
    return(list(drift_score = NA, message = "Insufficient temporal data"))
  }

  centroids <- lapply(temporal_results, function(p) {
    if (is.null(p$embeddings) || !is.matrix(p$embeddings) || nrow(p$embeddings) == 0) return(NULL)
    colMeans(p$embeddings)
  })

  drift_scores <- vapply(seq_len(length(centroids) - 1), function(i) {
    a <- centroids[[i]]; b <- centroids[[i + 1]]
    if (is.null(a) || is.null(b)) return(NA_real_)
    1 - .cosine_sim(a, b)
  }, numeric(1))

  if (all(is.na(drift_scores))) {
    return(list(drift_scores = drift_scores, mean_drift = NA_real_, max_drift = NA_real_,
                message = "No periods had embeddings to compare"))
  }

  list(
    drift_scores = drift_scores,
    mean_drift = mean(drift_scores, na.rm = TRUE),
    max_drift = max(drift_scores, na.rm = TRUE)
  )
}

#' @title Identify Topic Trends
#' @description Identifies trending topics in temporal results
#' @param temporal_results Temporal analysis results
#' @param ... Additional parameters
#' @return List containing identified trends
#' @concept topic-modeling
#' @export
identify_topic_trends <- function(temporal_results, ...) {
  if (is.null(temporal_results) || length(temporal_results) < 2) {
    return(list(trends = NULL, message = "Insufficient temporal data"))
  }

  all_topics <- unique(unlist(lapply(temporal_results, function(r) {
    unique(r$topic_assignments)
  })))

  prevalence_matrix <- sapply(temporal_results, function(r) {
    counts <- table(factor(r$topic_assignments, levels = all_topics))
    counts / sum(counts)
  })

  increasing <- character()
  decreasing <- character()
  stable <- character()

  for (i in seq_len(nrow(prevalence_matrix))) {
    trend <- prevalence_matrix[i, ]
    if (length(trend) < 2) next
    slope <- coef(lm(trend ~ seq_along(trend)))[2]
    topic_label <- rownames(prevalence_matrix)[i]
    if (slope > 0.01) increasing <- c(increasing, topic_label)
    else if (slope < -0.01) decreasing <- c(decreasing, topic_label)
    else stable <- c(stable, topic_label)
  }

  list(
    increasing = increasing,
    decreasing = decreasing,
    stable = stable,
    prevalence_matrix = prevalence_matrix
  )
}


#' Plot Topic Model Quality Metrics
#'
#' @description
#' Creates individual diagnostic metric plots across different K values
#' from stm::searchK results.
#'
#' @param search_results Results from stm::searchK or find_optimal_k()
#'
#' @return A named list of ggplot objects, one per available metric
#'   (possible keys: semcoh, residual, heldout, lbound).
#'
#' @concept topic-modeling
#' @export
plot_quality_metrics <- function(search_results) {

  results_data <- if ("results" %in% names(search_results)) {
    search_results$results
  } else {
    search_results
  }

  for (col in names(results_data)) {
    if (is.list(results_data[[col]])) {
      results_data[[col]] <- unlist(results_data[[col]])
    }
    if (col %in% c("residual", "lbound", "semcoh", "exclus", "heldout", "K")) {
      results_data[[col]] <- as.numeric(results_data[[col]])
    }
  }

  metric_info <- list(
    semcoh = list(name = "Semantic Coherence", color = "#4A90E2"),
    residual = list(name = "Residuals", color = "#E74C3C"),
    heldout = list(name = "Held-out Likelihood", color = "#9B59B6"),
    lbound = list(name = "Lower Bound", color = "#F39C12")
  )

  available_metrics <- intersect(names(metric_info), names(results_data))

  if (length(available_metrics) == 0) {
    stop("No valid metrics found in search results")
  }

  plots <- stats::setNames(vector("list", length(available_metrics)), available_metrics)
  for (metric in available_metrics) {
    info <- metric_info[[metric]]

    plot_df <- data.frame(
      K = results_data$K,
      value = results_data[[metric]]
    )
    plot_df$hover_text <- paste("K:", plot_df$K, "<br>", info$name, ":", round(plot_df$value, 4))

    plots[[metric]] <- ggplot2::ggplot(plot_df, ggplot2::aes(x = K, y = value, text = .data$hover_text)) +
      ggplot2::geom_line(color = info$color, linewidth = 0.8) +
      ggplot2::geom_point(color = info$color, size = 2.5) +
      ggplot2::labs(x = "Number of Topics (K)", y = info$name, title = info$name) +
      ggplot2::theme_minimal(base_size = 11) +
      ggplot2::theme(
        plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a", hjust = 0.5),
        panel.grid.minor = ggplot2::element_blank(),
        panel.grid.major = ggplot2::element_line(color = "#E5E7EB", linewidth = 0.3),
        axis.line = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.5),
        axis.text = ggplot2::element_text(color = "#3B3B3B"),
        axis.title = ggplot2::element_text(color = "#0c1f4a")
      )
  }

  plots
}


#' Plot Topic Model Comparison Scatter
#'
#' @description
#' Creates a scatter plot comparing topic model metrics across K values.
#' Automatically selects the best available metric combination.
#'
#' @param search_results Results from stm::searchK or find_optimal_k()
#' @param title Plot title (default `NULL` selects an auto-title based on which metric columns are present in `search_results`)
#' @param height Plot height in pixels (default: 600)
#' @param width Plot width in pixels (default: 800)
#'
#' @return A plotly scatter plot
#'
#' @concept topic-modeling
#' @export
plot_model_comparison <- function(search_results,
                                  title = NULL,
                                  height = 600,
                                  width = 800) {

  comparison_data <- if ("results" %in% names(search_results)) {
    search_results$results
  } else {
    search_results
  }

  for (col in names(comparison_data)) {
    if (is.list(comparison_data[[col]])) {
      comparison_data[[col]] <- unlist(comparison_data[[col]])
    }
    if (col %in% c("residual", "lbound", "semcoh", "exclus", "K")) {
      comparison_data[[col]] <- as.numeric(comparison_data[[col]])
    }
  }

  if ("semcoh" %in% names(comparison_data) && "exclus" %in% names(comparison_data)) {
    x_values <- comparison_data$semcoh
    y_values <- comparison_data$exclus
    x_label <- "Semantic Coherence"
    y_label <- "Exclusivity"
    auto_title <- "Coherence-Exclusivity Frontier (choose K in the upper-right)"
  } else if ("semcoh" %in% names(comparison_data)) {
    x_values <- comparison_data$semcoh
    y_values <- comparison_data$residual
    x_label <- "Semantic Coherence"
    y_label <- "Residual"
    auto_title <- "Coherence-Residual Trade-off (exclusivity unavailable for content-covariate models)"
  } else {
    x_values <- comparison_data$lbound
    y_values <- comparison_data$residual
    x_label <- "Lower Bound"
    y_label <- "Residual"
    auto_title <- "Lower Bound vs Residual"
  }

  if (is.null(title)) title <- auto_title

  k_values <- comparison_data$K
  k_normalized <- (k_values - min(k_values)) / max(1, max(k_values) - min(k_values))
  marker_sizes <- 5 + k_normalized * 8

  colors <- grDevices::colorRampPalette(c("#0066CC", "#CC3300"))(length(k_values))

  plot_df <- data.frame(x = x_values, y = y_values, k = k_values, sz = marker_sizes)
  plot_df$hover_text <- paste("K:", plot_df$k,
                               paste0("<br>", x_label, ":"), round(plot_df$x, 4),
                               paste0("<br>", y_label, ":"), round(plot_df$y, 4))

  ggplot2::ggplot(plot_df, ggplot2::aes(x = x, y = y, text = hover_text)) +
    ggplot2::geom_point(size = plot_df$sz, color = colors, alpha = 0.9) +
    ggplot2::geom_text(ggplot2::aes(label = k), color = "white", size = 3.5, fontface = "bold") +
    ggplot2::labs(x = x_label, y = y_label, title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a", hjust = 0.5),
      axis.line = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.5),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      panel.grid = ggplot2::element_blank()
    )
}


# Topic modeling visualization functions

#' @title Plot Word Probabilities by Topic
#'
#' @description
#' Creates a faceted bar plot showing the top terms and their probabilities (beta values)
#' for each topic in a topic model.
#'
#' @param top_topic_terms A data frame containing topic terms with columns: topic, term, and beta.
#' @param topic_label Optional topic labels. Can be either a named vector mapping topic numbers
#'   to labels, or a character string specifying a column name in top_topic_terms (default: NULL).
#' @param ncol Number of columns for facet wrap layout (default: 3).
#' @param height Plot height for responsive spacing adjustments (default: 1200).
#' @param width Plot width for responsive spacing adjustments (default: 800).
#' @param ylab Y-axis label (default: "Word probability").
#' @param title Plot title (default: NULL for auto-generated title).
#' @param colors Color palette for topics (default: NULL for auto-generated colors).
#' @param measure_label Label for the probability measure (default: "Beta").
#' @param base_font_size Base font size in points for the plot theme (default: 11). Axis text and strip text will be base_font_size + 2.
#' @param ... Additional arguments (currently unused, kept for compatibility).
#'
#' @return A ggplot2 object showing word probabilities faceted by topic.
#'
#' @concept topic-modeling
#' @export
plot_word_probability <- function(top_topic_terms,
                                   topic_label = NULL,
                                   ncol = 3,
                                   height = 1200,
                                   width = 800,
                                   ylab = "Word probability",
                                   title = NULL,
                                   colors = NULL,
                                   measure_label = "Beta",
                                   base_font_size = 11,
                                   ...) {

  if (!"topic" %in% colnames(top_topic_terms)) {
    stop("The data frame must contain a 'topic' column.")
  }

  top_topic_terms <- top_topic_terms %>%
    dplyr::mutate(topic = as.character(topic))

  if (!is.null(topic_label)) {
    if (is.vector(topic_label) && !is.null(names(topic_label))) {
      manual_labels_df <- data.frame(
        topic = names(topic_label),
        label = unname(topic_label),
        stringsAsFactors = FALSE
      )
      top_topic_terms <- top_topic_terms %>%
        dplyr::left_join(manual_labels_df, by = "topic") %>%
        dplyr::mutate(labeled_topic = ifelse(!is.na(label), label, paste("Topic", topic))) %>%
        dplyr::select(-label)
    } else if (is.character(topic_label) && length(topic_label) == 1) {
      if (!topic_label %in% colnames(top_topic_terms)) {
        stop(paste("Column", topic_label, "not found in top_topic_terms."))
      }
      top_topic_terms <- top_topic_terms %>%
        dplyr::mutate(labeled_topic = as.character(.data[[topic_label]]))
    } else {
      top_topic_terms <- top_topic_terms %>%
        dplyr::mutate(labeled_topic = paste("Topic", topic))
    }
  } else {
    top_topic_terms <- top_topic_terms %>%
      dplyr::mutate(labeled_topic = paste("Topic", topic))
  }

  top_topic_terms <- top_topic_terms %>%
    dplyr::mutate(
      ord = factor(topic, levels = sort(as.numeric(unique(topic)))),
      term = tidytext::reorder_within(term, beta, labeled_topic)
    ) %>%
    dplyr::arrange(ord) %>%
    dplyr::ungroup()

  levelt <- top_topic_terms %>%
    dplyr::arrange(as.numeric(topic)) %>%
    dplyr::distinct(labeled_topic) %>%
    dplyr::pull(labeled_topic)

  top_topic_terms$labeled_topic <- factor(top_topic_terms$labeled_topic, levels = levelt)

  ggplot_obj <- ggplot2::ggplot(
    top_topic_terms,
    ggplot2::aes(term, beta, fill = labeled_topic,
                 text = paste0("Term: ", gsub("___.*$", "", term), "<br>",
                              "Topic: ", labeled_topic, "<br>",
                              measure_label, ": ", sprintf("%.3f", beta)))
  ) +
    ggplot2::geom_col(show.legend = FALSE, alpha = 0.8) +
    ggplot2::facet_wrap(~ labeled_topic, scales = "free", ncol = ncol, strip.position = "top") +
    tidytext::scale_x_reordered() +
    ggplot2::scale_y_continuous(labels = .number_labeller(3)) +
    ggplot2::coord_flip() +
    ggplot2::xlab("") +
    ggplot2::ylab(ylab) +
    ggplot2::theme_minimal(base_size = base_font_size) +
    ggplot2::theme(
      legend.position = "none",
      panel.grid.major = ggplot2::element_blank(),
      panel.grid.minor = ggplot2::element_blank(),
      axis.line = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      axis.ticks = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      strip.text.x = ggplot2::element_text(
        size = base_font_size,
        color = "#0c1f4a",
        lineheight = ifelse(width > 1000, 1.1, 1.2),
        margin = ggplot2::margin(l = 10, r = 10)
      ),
      panel.spacing.x = ggplot2::unit(ifelse(width > 1000, 2.2, 1.6), "lines"),
      panel.spacing.y = ggplot2::unit(ifelse(width > 1000, 2.2, 1.6), "lines"),
      axis.text.x = ggplot2::element_text(size = base_font_size, color = "#3B3B3B", hjust = 1, margin = ggplot2::margin(r = 20)),
      axis.text.y = ggplot2::element_text(size = base_font_size, color = "#3B3B3B", margin = ggplot2::margin(t = 20)),
      axis.title = ggplot2::element_text(size = base_font_size, color = "#0c1f4a"),
      axis.title.x = ggplot2::element_text(margin = ggplot2::margin(t = 25)),
      axis.title.y = ggplot2::element_text(margin = ggplot2::margin(r = 25))
    )

  if (!is.null(colors)) {
    ggplot_obj <- ggplot_obj + ggplot2::scale_fill_manual(values = colors)
  }

  ggplot_obj
}


#' @title Plot Per-Document Per-Topic Probabilities
#'
#' @description
#' Generates a bar plot showing the prevalence of each topic across all documents.
#'
#' @param gamma_data A data frame with gamma values from calculate_topic_probability().
#' @param top_n The number of topics to display (default: 10).
#' @param use_topic_labels Logical. If TRUE, use the `topic_label` column from `gamma_data` for axis labels (falls back to topic number when the column is absent). If FALSE (default), labels are formatted as "Topic N".
#' @param colors Optional color palette for topics (default: NULL).
#' @param ylab Y-axis label (default: "Topic Proportion").
#' @param base_font_size Base font size in points for the plot theme (default: 11). Axis text and strip text will be base_font_size + 2.
#'
#' @return A ggplot2 object showing a bar plot of topic prevalence.
#'
#' @concept topic-modeling
#' @export
plot_topic_probability <- function(gamma_data,
                                   top_n = 10,
                                   use_topic_labels = FALSE,
                                   colors = NULL,
                                   ylab = "Topic Proportion",
                                   base_font_size = 11) {

    gamma_terms <- gamma_data
    if (!is.null(top_n) && top_n < nrow(gamma_terms)) {
      gamma_terms <- gamma_terms %>%
        dplyr::top_n(top_n, gamma)
    }

    if (isTRUE(use_topic_labels)) {
      if ("topic_label" %in% names(gamma_terms)) {
        gamma_terms <- gamma_terms %>%
          dplyr::mutate(topic_display = topic_label)
      } else {
        gamma_terms <- gamma_terms %>%
          dplyr::mutate(topic_display = topic)
      }
    } else {
      gamma_terms <- gamma_terms %>%
        dplyr::mutate(topic_display = paste("Topic", topic))
    }

    if ("tt" %in% names(gamma_terms)) {
      gamma_terms <- gamma_terms %>%
        dplyr::arrange(tt) %>%
        dplyr::mutate(topic_display = factor(topic_display, levels = unique(topic_display)))
    } else {
      gamma_terms <- gamma_terms %>%
        dplyr::mutate(topic_display = factor(topic_display, levels = unique(topic_display)))
    }

    hover_text <- if ("terms" %in% names(gamma_terms)) {
      paste0("Topic: ", gamma_terms$topic_display, "<br>Terms: ", gamma_terms$terms, "<br>Gamma: ", sprintf("%.3f", gamma_terms$gamma))
    } else {
      paste0("Topic: ", gamma_terms$topic_display, "<br>Gamma: ", sprintf("%.3f", gamma_terms$gamma))
    }

    ggplot_obj <- ggplot2::ggplot(gamma_terms, ggplot2::aes(x = topic_display, y = gamma, fill = topic_display,
                                          text = hover_text)) +
      ggplot2::geom_col(alpha = 0.8) +
      ggplot2::coord_flip() +
      ggplot2::scale_y_continuous(labels = .number_labeller(2)) +
      ggplot2::xlab("") +
      ggplot2::ylab(ylab) +
      ggplot2::theme_minimal(base_size = base_font_size) +
      ggplot2::theme(
        legend.position = "none",
        panel.grid.major = ggplot2::element_blank(),
        panel.grid.minor = ggplot2::element_blank(),
        axis.line = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
        axis.ticks = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
        strip.text.x = ggplot2::element_text(size = base_font_size, color = "#0c1f4a"),
        axis.text.x = ggplot2::element_text(size = base_font_size, color = "#3B3B3B"),
        axis.text.y = ggplot2::element_text(size = base_font_size, color = "#3B3B3B"),
        axis.title = ggplot2::element_text(size = base_font_size, color = "#0c1f4a"),
        axis.title.x = ggplot2::element_text(margin = ggplot2::margin(t = 10)),
        axis.title.y = ggplot2::element_text(margin = ggplot2::margin(r = 10))
      )

    if (!is.null(colors)) {
      ggplot_obj <- ggplot_obj + ggplot2::scale_fill_manual(values = colors)
    }

    ggplot_obj
}


#' @title Plot Topic Effects for Categorical Variables
#'
#' @description
#' Creates a faceted plot showing how categorical variables affect topic proportions.
#'
#' @param effects_data Data frame with columns: topic, value, proportion, lower, upper
#' @param ncol Number of columns for faceting (default: 2)
#' @param height Plot height in pixels (default: 800)
#' @param width Plot width in pixels (default: 1000)
#' @param title Plot title (default: "Category Effects")
#' @param base_font_size Base font size in points for the plot theme (default: 11). Axis text and strip text will be base_font_size + 2.
#'
#' @return A plotly object
#'
#' @concept topic-modeling
#' @export
plot_topic_effects_categorical <- function(effects_data,
                                           ncol = 2,
                                           height = 800,
                                           width = 1000,
                                           title = "Category Effects",
                                           base_font_size = 11) {

  if (is.null(effects_data) || nrow(effects_data) == 0) {
    return(ggplot2::ggplot() +
      ggplot2::annotate("text", x = 0.5, y = 0.5,
        label = "No categorical effects available.\nPlease run the effect estimation first.",
        size = 5, color = "#ef4444") +
      ggplot2::theme_void())
  }

  effects_data <- effects_data %>%
    dplyr::mutate(topic_label = paste("Topic", topic))

  effects_data$hover_text <- paste("Topic:", effects_data$topic_label,
                                   "<br>Value:", effects_data$value,
                                   "<br>Proportion:", sprintf("%.3f", effects_data$proportion),
                                   "<br>95% CI:", sprintf("[%.3f, %.3f]", effects_data$lower, effects_data$upper))

  ggplot_obj <- ggplot2::ggplot(effects_data, ggplot2::aes(x = value, y = proportion, text = hover_text)) +
    ggplot2::facet_wrap(~topic_label, ncol = ncol, scales = "free") +
    ggplot2::scale_y_continuous(labels = .number_labeller(3)) +
    ggplot2::xlab("") +
    ggplot2::ylab("Topic proportion") +
    ggplot2::geom_errorbar(
      ggplot2::aes(ymin = lower, ymax = upper),
      width = 0.1,
      linewidth = 0.5,
      color = "#337ab7"
    ) +
    ggplot2::geom_point(color = "#337ab7", size = 1.5) +
    ggplot2::theme_minimal(base_size = base_font_size) +
    ggplot2::theme(
      legend.position = "none",
      panel.grid.major = ggplot2::element_blank(),
      panel.grid.minor = ggplot2::element_blank(),
      axis.line = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      axis.ticks = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      strip.text.x = ggplot2::element_text(size = base_font_size, color = "#0c1f4a", margin = ggplot2::margin(b = 30, t = 15)),
      axis.text.x = ggplot2::element_text(size = base_font_size, color = "#3B3B3B", hjust = 1, margin = ggplot2::margin(t = 20)),
      axis.text.y = ggplot2::element_text(size = base_font_size, color = "#3B3B3B", margin = ggplot2::margin(r = 20)),
      axis.title = ggplot2::element_text(size = base_font_size, color = "#0c1f4a"),
      axis.title.x = ggplot2::element_text(margin = ggplot2::margin(t = 25)),
      axis.title.y = ggplot2::element_text(margin = ggplot2::margin(r = 25)),
      plot.margin = ggplot2::margin(t = 40, b = 40)
    )

  ggplot_obj +
    ggplot2::ggtitle(title) +
    ggplot2::theme(plot.title = ggplot2::element_text(
      size = base_font_size + 2, color = "#0c1f4a", hjust = 0.5))
}


#' @title Plot Topic Effects for Continuous Variables
#'
#' @description
#' Creates a faceted plot showing how continuous variables affect topic proportions.
#'
#' @param effects_data Data frame with columns: topic, value, proportion, lower, upper
#' @param ncol Number of columns for faceting (default: 2)
#' @param height Plot height in pixels (default: 800)
#' @param width Plot width in pixels (default: 1000)
#' @param title Plot title (default: "Continuous Variable Effects")
#' @param base_font_size Base font size in points for the plot theme (default: 11). Axis text and strip text will be base_font_size + 2.
#'
#' @return A plotly object
#'
#' @concept topic-modeling
#' @export
plot_topic_effects_continuous <- function(effects_data,
                                          ncol = 2,
                                          height = 800,
                                          width = 1000,
                                          title = "Continuous Variable Effects",
                                          base_font_size = 11) {

  effects_data <- effects_data %>%
    dplyr::mutate(topic_label = paste("Topic", topic))

  effects_data$hover_text <- paste("Topic:", effects_data$topic_label,
                                   "<br>Value:", round(effects_data$value, 3),
                                   "<br>Proportion:", sprintf("%.3f", effects_data$proportion))

  ggplot_obj <- ggplot2::ggplot(effects_data, ggplot2::aes(x = value, y = proportion, text = hover_text)) +
    ggplot2::facet_wrap(~topic_label, ncol = ncol, scales = "free") +
    ggplot2::scale_y_continuous(labels = .number_labeller(3)) +
    ggplot2::geom_ribbon(ggplot2::aes(ymin = lower, ymax = upper), fill = "#337ab7", alpha = 0.2) +
    ggplot2::geom_line(linewidth = 0.5, color = "#337ab7") +
    ggplot2::xlab("") +
    ggplot2::ylab("Topic proportion") +
    ggplot2::theme_minimal(base_size = base_font_size) +
    ggplot2::theme(
      legend.position = "none",
      panel.grid.major = ggplot2::element_blank(),
      panel.grid.minor = ggplot2::element_blank(),
      axis.line = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      axis.ticks = ggplot2::element_line(color = "#3B3B3B", linewidth = 0.3),
      strip.text.x = ggplot2::element_text(size = base_font_size, color = "#0c1f4a", margin = ggplot2::margin(b = 30, t = 15)),
      axis.text.x = ggplot2::element_text(size = base_font_size, color = "#3B3B3B", hjust = 1, margin = ggplot2::margin(t = 20)),
      axis.text.y = ggplot2::element_text(size = base_font_size, color = "#3B3B3B", margin = ggplot2::margin(r = 20)),
      axis.title = ggplot2::element_text(size = base_font_size, color = "#0c1f4a"),
      axis.title.x = ggplot2::element_text(margin = ggplot2::margin(t = 25)),
      axis.title.y = ggplot2::element_text(margin = ggplot2::margin(r = 25)),
      plot.margin = ggplot2::margin(t = 40, b = 40)
    )

  ggplot_obj +
    ggplot2::ggtitle(title) +
    ggplot2::theme(plot.title = ggplot2::element_text(
      size = base_font_size + 2, color = "#0c1f4a", hjust = 0.5))
}



#' Plot Cluster Top Terms
#'
#' @description
#' Creates a horizontal bar plot showing the top terms in a cluster or document group.
#'
#' @param terms Named numeric vector of term frequencies, or data frame with
#'   'term' and 'frequency' columns
#' @param cluster_id Cluster identifier for the title (default: NULL)
#' @param title Custom title (default: NULL, auto-generated from cluster_id)
#' @param n_terms Number of top terms to display (default: 10)
#' @param color Bar color (default: "#337ab7")
#' @param height Plot height in pixels (default: 500)
#' @param width Plot width in pixels (default: NULL for auto)
#'
#' @return A plotly object
#'
#' @concept visualization
#' @export
plot_cluster_terms <- function(terms,
                                cluster_id = NULL,
                                title = NULL,
                                n_terms = 10,
                                color = "#337ab7",
                                height = 500,
                                width = NULL) {

  if (is.null(terms) || length(terms) == 0) {
    return(ggplot2::ggplot() +
      ggplot2::annotate("text", x = 0.5, y = 0.5,
        label = "No terms available for this cluster",
        size = 5, color = "#ef4444") +
      ggplot2::theme_void())
  }

  if (is.data.frame(terms)) {
    if (!all(c("term", "frequency") %in% names(terms))) {
      stop("Data frame must have 'term' and 'frequency' columns")
    }
    terms <- terms %>%
      dplyr::arrange(dplyr::desc(frequency)) %>%
      dplyr::slice_head(n = n_terms)
    term_names <- terms$term
    term_values <- terms$frequency
  } else {
    top_terms <- utils::head(sort(terms, decreasing = TRUE), n_terms)
    term_names <- names(top_terms)
    term_values <- as.numeric(top_terms)
  }

  if (is.null(title)) {
    title <- if (!is.null(cluster_id)) {
      paste("Top Terms in Cluster", cluster_id)
    } else {
      "Top Terms"
    }
  }

  plot_df <- data.frame(
    term = factor(term_names, levels = term_names[order(term_values)]),
    frequency = term_values
  )

  plot_df$hover_text <- paste("Term:", plot_df$term, "<br>Frequency:", plot_df$frequency)

  ggplot2::ggplot(plot_df, ggplot2::aes(x = frequency, y = term, text = hover_text)) +
    ggplot2::geom_col(fill = color) +
    ggplot2::labs(x = "Frequency", y = NULL, title = title) +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a", hjust = 0.5),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      panel.grid.minor = ggplot2::element_blank(),
      panel.grid.major.y = ggplot2::element_blank()
    )
}


# Topic-based content generation

#' Get Default System Prompt for Content Type
#'
#' Returns the default system prompt for a given content type.
#'
#' @param content_type Type of content to generate.
#'
#' @return Character string with the system prompt.
#'
#' @concept ai
#' @export
get_content_type_prompt <- function(content_type) {
  prompts <- list(
    survey_item = "
You are a survey design expert specializing in creating Likert-scale items for research.
Your task is to generate clear, concise survey statements that can be rated on a 5-point scale (1=Strongly Disagree to 5=Strongly Agree).

Guidelines:
1. Create statements focused on a single concept
2. Use active voice and present tense
3. Avoid double-barreled questions
4. Use simple, direct language
5. Ensure items work well on an agree-disagree scale
6. Frame items to capture the essence of the provided keywords

Return ONLY the survey item text, without numbering, quotes, or explanations.",

    research_question = "
You are a research methodology expert specializing in formulating research questions.
Your task is to generate clear, focused research questions based on topic keywords.

Guidelines:
1. Create questions that are specific and answerable
2. Use appropriate question words (How, What, Why, To what extent)
3. Ensure questions are neither too broad nor too narrow
4. Frame questions to guide empirical investigation
5. Avoid yes/no questions - aim for open-ended inquiry

Return ONLY the research question, without numbering, quotes, or explanations.",

    theme_description = "
You are a qualitative research expert specializing in thematic analysis.
Your task is to generate descriptive summaries of themes based on topic keywords.

Guidelines:
1. Write in third person, academic style
2. Describe what the theme encompasses
3. Use language like 'This theme captures...', 'Participants discussed...'
4. Be concise but complete
5. Avoid interpretation - focus on description

Return ONLY the theme description, without numbering, quotes, or explanations.",

    policy_recommendation = "
You are a policy analysis expert specializing in evidence-based recommendations.
Your task is to generate actionable policy recommendations based on topic keywords.

Guidelines:
1. Begin with action verbs (Implement, Establish, Develop, Ensure)
2. Be specific and actionable
3. Consider feasibility and impact
4. Use clear, professional language
5. Focus on one recommendation per topic

Return ONLY the policy recommendation, without numbering, quotes, or explanations.",

    interview_question = "
You are a qualitative research expert specializing in interview methodology.
Your task is to generate open-ended interview questions based on topic keywords.

Guidelines:
1. Create questions that encourage detailed responses
2. Use open-ended phrasing (Can you describe..., Tell me about..., How do you...)
3. Avoid leading or biased questions
4. Make questions conversational yet focused
5. Ensure questions are appropriate for semi-structured interviews

Return ONLY the interview question, without numbering, quotes, or explanations.",

    custom = "
You are an expert content generator.
Your task is to generate content based on the provided keywords.

Return ONLY the requested content, without numbering, quotes, or explanations."
  )

  prompts[[content_type]] %||% prompts[["custom"]]
}


#' Get Default User Prompt Template for Content Type
#'
#' Returns the default user prompt template for a given content type.
#'
#' @param content_type Type of content to generate.
#'
#' @return Character string with the user prompt template containing \code{\{terms\}} placeholder.
#'
#' @concept ai
#' @export
get_content_type_user_template <- function(content_type) {
  templates <- list(
    survey_item = "Generate a single survey item based on these keywords (ordered by importance): {terms}

The survey item should:
- Capture the main concept from these keywords
- Be rateable on a 5-point Likert scale
- Be clear and concise

Survey item:",

    research_question = "Generate a research question based on these keywords (ordered by importance): {terms}

The research question should:
- Be specific and empirically answerable
- Capture the key concepts from these keywords
- Guide meaningful investigation

Research question:",

    theme_description = "Generate a theme description based on these keywords (ordered by importance): {terms}

The theme description should:
- Summarize what this theme encompasses
- Be written in academic style
- Capture the essence of the keywords

Theme description:",

    policy_recommendation = "Generate a policy recommendation based on these keywords (ordered by importance): {terms}

The recommendation should:
- Be specific and actionable
- Address the key concepts from these keywords
- Be feasible to implement

Policy recommendation:",

    interview_question = "Generate an interview question based on these keywords (ordered by importance): {terms}

The interview question should:
- Be open-ended and encourage detailed responses
- Explore the concepts represented by these keywords
- Be appropriate for a semi-structured interview

Interview question:",

    custom = "Generate content based on these keywords (ordered by importance): {terms}

Content:"
  )

  templates[[content_type]] %||% templates[["custom"]]
}


#' Generate Content from Topic Terms
#'
#' Uses Large Language Models (LLMs) to generate various types of content
#' based on topic model terms. Supports multiple content types with optimized
#' default prompts, or fully custom prompts.
#'
#' @param topic_terms_df A data frame with topic terms, containing columns for
#'   topic identifier, term, and optionally term weight (beta).
#' @param content_type Type of content to generate. One of:
#'   \describe{
#'     \item{"survey_item"}{Likert-scale survey items for scale development}
#'     \item{"research_question"}{Research questions for literature review}
#'     \item{"theme_description"}{Theme descriptions for qualitative analysis}
#'     \item{"policy_recommendation"}{Policy recommendations for policy analysis}
#'     \item{"interview_question"}{Interview questions for qualitative research}
#'     \item{"custom"}{Custom content using user-provided prompts}
#'   }
#' @param topic_var Name of the column containing topic identifiers (default: "topic").
#' @param term_var Name of the column containing terms (default: "term").
#' @param weight_var Name of the column containing term weights (default: "beta").
#' @param provider LLM provider: "openai" or "gemini" (default: "openai").
#' @param model Model name. For OpenAI: "gpt-4.1-mini", "gpt-4", etc.
#'   For Gemini: "gemini-2.5-flash-lite", "gemini-2.5-flash", etc.
#' @param temperature Sampling temperature (0-2). Lower = more deterministic (default: 0).
#' @param system_prompt Custom system prompt. If NULL, uses default for content_type.
#' @param user_prompt_template Custom user prompt template with \{terms\} placeholder.
#'   If NULL, uses default for content_type.
#' @param max_tokens Maximum tokens for response (default: 150).
#' @param api_key API key for the selected provider. If NULL, reads from
#'   OPENAI_API_KEY or GEMINI_API_KEY environment variable.
#' @param output_var Name of the output column (default: based on content_type).
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A data frame with generated content joined to original topic terms.
#'
#' @details
#' The function generates one piece of content per unique topic. Each content type
#' has optimized default prompts, but these can be overridden with custom prompts.
#'
#' Requires an API key set via the \code{api_key} parameter or the relevant
#' environment variable (OPENAI_API_KEY or GEMINI_API_KEY).
#'
#' @concept ai
#' @seealso [generate_topic_labels()] for the step that creates topic labels; [get_content_type_prompt()] and [get_content_type_user_template()] to inspect or override default prompts
#' @export
#'
#' @examples
#' if (interactive()) {
#' # Generate survey items
#' survey_items <- generate_topic_content(
#'   topic_terms_df = top_terms,
#'   content_type = "survey_item",
#'   provider = "openai",
#'   model = "gpt-4.1-mini"
#' )
#'
#' # Generate research questions
#' research_qs <- generate_topic_content(
#'   topic_terms_df = top_terms,
#'   content_type = "research_question",
#'   provider = "gemini",
#'   model = "gemini-2.5-flash-lite"
#' )
#'
#' # Generate with custom prompt
#' custom_content <- generate_topic_content(
#'   topic_terms_df = top_terms,
#'   content_type = "custom",
#'   system_prompt = "You are an expert in educational policy...",
#'   user_prompt_template = "Based on {terms}, generate a learning objective:"
#' )
#' }
generate_topic_content <- function(topic_terms_df,
                                    content_type = c("survey_item", "research_question",
                                                     "theme_description", "policy_recommendation",
                                                     "interview_question", "custom"),
                                    topic_var = "topic",
                                    term_var = "term",
                                    weight_var = "beta",
                                    provider = c("openai", "gemini"),
                                    model = "gpt-4.1-mini",
                                    temperature = 0,
                                    system_prompt = NULL,
                                    user_prompt_template = NULL,
                                    max_tokens = 150,
                                    api_key = NULL,
                                    output_var = NULL,
                                    verbose = TRUE) {

  content_type <- match.arg(content_type)
  provider <- match.arg(provider)

  if (!topic_var %in% names(topic_terms_df)) {
    stop("topic_var '", topic_var, "' not found in topic_terms_df")
  }
  if (!term_var %in% names(topic_terms_df)) {
    stop("term_var '", term_var, "' not found in topic_terms_df")
  }

  # Set default output variable name based on content type
  if (is.null(output_var)) {
    output_var <- switch(content_type,
      "survey_item" = "survey_item",
      "research_question" = "research_question",
      "theme_description" = "theme_description",
      "policy_recommendation" = "policy_recommendation",
      "interview_question" = "interview_question",
      "custom" = "generated_content"
    )
  }

  # Get default prompts if not provided
  if (is.null(system_prompt)) {
    system_prompt <- get_content_type_prompt(content_type)
  }
  if (is.null(user_prompt_template)) {
    user_prompt_template <- get_content_type_user_template(content_type)
  }

  # Prepare topic data
  has_weights <- weight_var %in% names(topic_terms_df)

  if (has_weights) {
    top_terms <- topic_terms_df %>%
      dplyr::group_by(.data[[topic_var]]) %>%
      dplyr::arrange(dplyr::desc(.data[[weight_var]])) %>%
      dplyr::ungroup()
  } else {
    top_terms <- topic_terms_df %>%
      dplyr::group_by(.data[[topic_var]]) %>%
      dplyr::ungroup()
  }

  unique_topics <- top_terms %>%
    dplyr::distinct(.data[[topic_var]]) %>%
    dplyr::arrange(.data[[topic_var]])

  unique_topics[[output_var]] <- NA_character_

  if (is.null(api_key)) {
    env_var <- switch(provider, "openai" = "OPENAI_API_KEY", "gemini" = "GEMINI_API_KEY")
    api_key <- Sys.getenv(env_var)
    if (api_key == "") {
      return(.notify_missing_api_key(provider))
    }
  }

  # Progress bar
  if (verbose && requireNamespace("progress", quietly = TRUE)) {
    pb <- progress::progress_bar$new(
      format = paste0(" Generating ", content_type, "s [:bar] :percent (:current/:total) ETA: :eta"),
      total = nrow(unique_topics),
      clear = FALSE, width = 60
    )
  }

  for (i in seq_len(nrow(unique_topics))) {
    if (verbose && exists("pb")) {
      pb$tick()
    }

    current_topic <- unique_topics[[topic_var]][i]

    # Get terms for this topic
    topic_data <- top_terms %>%
      dplyr::filter(.data[[topic_var]] == current_topic)

    terms_text <- paste(topic_data[[term_var]], collapse = ", ")

    # Create user prompt
    user_prompt <- gsub("\\{terms\\}", terms_text, user_prompt_template)

    generated_content <- tryCatch(
      call_llm_api(
        provider = provider,
        system_prompt = system_prompt,
        user_prompt = user_prompt,
        model = model,
        temperature = temperature,
        max_tokens = max_tokens,
        api_key = api_key
      ),
      error = function(e) {
        if (verbose) {
          warning(sprintf("Error generating content for topic %s: %s", current_topic, e$message))
        }
        NA_character_
      }
    )

    # Clean up response
    if (!is.na(generated_content)) {
      generated_content <- trimws(generated_content)
      generated_content <- gsub("^[\"'](.*)[\"']$", "\\1", generated_content)
    }

    unique_topics[[output_var]][i] <- generated_content

    Sys.sleep(0.5)
  }

  # Join generated content back to original data
  result <- top_terms %>%
    dplyr::left_join(
      unique_topics %>% dplyr::select(dplyr::all_of(c(topic_var, output_var))),
      by = topic_var
    )

  if (verbose) {
    n_generated <- sum(!is.na(unique_topics[[output_var]]))
    message(sprintf("Generated %d/%d %ss", n_generated, nrow(unique_topics), content_type))
  }

  result
}

#' @importFrom utils modifyList
#' @importFrom stats cor
NULL

# Text Preprocessing Functions
# Functions for data import, text preprocessing, and feature extraction

#' Get Available Document-Feature Matrix with Fallback
#'
#' @description Returns the first non-NULL DFM from a priority fallback chain.
#' Useful when multiple DFM processing stages exist and the most processed available version is needed.
#'
#' @param dfm_lemma Optional lemmatized DFM (highest priority)
#' @param dfm_outcome Optional preprocessed DFM (medium priority)
#' @param dfm_final Optional final processed DFM (medium-low priority)
#' @param dfm_init Optional initial DFM (lowest priority)
#'
#' @return The first non-NULL DFM from the priority chain, or NULL if all are NULL
#'
#' @details
#' Priority order (highest to lowest):
#' 1. dfm_lemma - Lemmatized tokens (most processed)
#' 2. dfm_outcome - Preprocessed tokens
#' 3. dfm_final - Final processed version
#' 4. dfm_init - Initial unprocessed tokens
#'
#' @concept preprocessing
#' @keywords internal
#' @export
#'
#' @examples
#' \donttest{
#'   dfm1 <- quanteda::dfm(quanteda::tokens("assistive technology supports learning"))
#'   result <- get_available_dfm(dfm_init = dfm1)
#' }
get_available_dfm <- function(dfm_lemma = NULL, dfm_outcome = NULL, dfm_final = NULL, dfm_init = NULL) {
  if (!is.null(dfm_lemma)) {
    return(dfm_lemma)
  }

  if (!is.null(dfm_outcome)) {
    return(dfm_outcome)
  }

  if (!is.null(dfm_final)) {
    return(dfm_final)
  }

  if (!is.null(dfm_init)) {
    return(dfm_init)
  }

  return(NULL)
}


#' Get Available Tokens with Fallback
#'
#' @description Returns the first non-NULL tokens object from a priority fallback chain.
#' Useful when multiple token processing stages exist and the most processed available version is needed.
#'
#' @param final_tokens Optional fully processed tokens (highest priority)
#' @param processed_tokens Optional partially processed tokens
#' @param preprocessed_tokens Optional early-stage preprocessed tokens
#' @param united_tbl Optional data frame with united_texts column (lowest priority, will be tokenized)
#'
#' @return The first non-NULL tokens from the priority chain, or NULL if all are NULL
#'
#' @details
#' Priority order (highest to lowest):
#' 1. final_tokens - Fully processed tokens
#' 2. processed_tokens - Partially processed tokens
#' 3. preprocessed_tokens - Early stage preprocessed tokens
#' 4. united_tbl - Raw text (will be tokenized if used)
#'
#' @concept preprocessing
#' @keywords internal
#' @export
#'
#' @examples
#' \donttest{
#'   toks <- quanteda::tokens("assistive technology supports learning")
#'   result <- get_available_tokens(processed_tokens = toks)
#' }
get_available_tokens <- function(final_tokens = NULL,
                                  processed_tokens = NULL,
                                  preprocessed_tokens = NULL,
                                  united_tbl = NULL) {
  tokens <- Find(Negate(is.null), list(final_tokens, processed_tokens, preprocessed_tokens))
  if (!is.null(tokens)) return(tokens)

  if (!is.null(united_tbl) && "united_texts" %in% names(united_tbl)) {
    toks <- quanteda::tokens(united_tbl$united_texts, what = "word")
    other_cols <- united_tbl[, !names(united_tbl) %in% "united_texts", drop = FALSE]
    if (ncol(other_cols) > 0) {
      quanteda::docvars(toks) <- other_cols
    }
    return(toks)
  }

  NULL
}

#' @title Process Files
#'
#' @description
#' This function processes different types of files and text input based on the dataset choice.
#'
#' @param dataset_choice A character string indicating the dataset choice.
#' @param file_info A data frame containing file information with a column
#'   named 'filepath' (default: NULL).
#' @param text_input A character string containing text input (default: NULL).
#'
#' @return A data frame containing the processed data.
#'
#' @importFrom utils read.csv
#'
#' @concept preprocessing
#' @export
#'
#' @examples
#' if (interactive()) {
#'   mydata <- TextAnalysisR::SpecialEduTech
#'   mydata <- TextAnalysisR::import_files(dataset_choice = "Upload an Example Dataset")
#'   head(mydata)
#'
#'   xlsx_path <- system.file("extdata", "SpecialEduTech.xlsx",
#'                            package = "TextAnalysisR")
#'   file_info <- data.frame(filepath = xlsx_path)
#'   mydata <- TextAnalysisR::import_files(dataset_choice = "Upload Your File",
#'                                           file_info = file_info)
#'   head(mydata)
#'
#'
#'   text_input <- paste("Virtual manipulatives for algebra instruction",
#'                       "manipulatives mathematics learning disability",
#'                       "This study examined virtual manipulatives effects on",
#'                       "students with learning disabilities")
#'   mydata <- TextAnalysisR::import_files(dataset_choice = "Copy and Paste Text",
#'                                           text_input = text_input)
#'   head(mydata)
#' }
import_files <- function(dataset_choice, file_info = NULL, text_input = NULL) {

  if (!requireNamespace("readxl", quietly = TRUE) ||
      !requireNamespace("pdftools", quietly = TRUE) ||
      !requireNamespace("officer", quietly = TRUE)) {
    stop(
      "The 'readxl', 'pdftools' and 'officer' packages are required for this functionality. ",
      "Please install them using install.packages(c('readxl', 'pdftools', 'officer'))."
    )
  }

  if (dataset_choice == "Upload an Example Dataset") {
    data <- TextAnalysisR::SpecialEduTech
    data <- tibble::as_tibble(data)
  } else if (dataset_choice == "Copy and Paste Text") {
    if (is.null(text_input)) stop("No text provided")
    data <- tibble::tibble(text = text_input)
  } else if (dataset_choice == "Upload Your File") {
    if (is.null(file_info)) stop("No file provided")

    data_list <- lapply(seq_len(nrow(file_info)), function(i) {
      filepath <- file_info$filepath[i]
      ext <- tolower(tools::file_ext(filepath))

      df <- tryCatch({
        if (ext %in% c("xlsx", "xls", "xlsm")) {
          readxl::read_excel(filepath, col_names = TRUE)
        } else if (ext == "csv") {
          read.csv(filepath, header = TRUE, stringsAsFactors = FALSE)
        } else if (ext == "pdf") {
          tryCatch({
            pages <- pdftools::pdf_text(filepath)
            lines <- unlist(lapply(pages, function(page) {
              lines <- strsplit(page, "\n")[[1]]
              trimws(lines)
            }))
            lines <- lines[lines != ""]
            data.frame(text = lines, stringsAsFactors = FALSE)
          }, error = function(e) {
            message("Error processing PDF file: ", filepath, ": ", e$message)
            data.frame(text = "", stringsAsFactors = FALSE)
          })
        } else if (ext == "docx") {
          # Note: This extracts text only. For image extraction from DOCX,
          # use process_docx_multimodal() function (requires Python + Vision API)
          doc <- officer::read_docx(filepath)
          doc_summary <- officer::docx_summary(doc)
          lines <- unlist(lapply(doc_summary$text, function(x) {
            trimws(unlist(strsplit(x, "\n")))
          }))
          lines <- lines[lines != ""]
          data.frame(text = lines, stringsAsFactors = FALSE)
        } else if (ext == "txt") {
          tryCatch({
            lines <- readLines(filepath, warn = FALSE, encoding = "UTF-8")
            lines <- trimws(lines)
            lines <- lines[lines != ""]
            data.frame(text = lines, stringsAsFactors = FALSE)
          }, error = function(e) {
            message("Error processing TXT file: ", filepath, ": ", e$message)
            data.frame(text = "", stringsAsFactors = FALSE)
          })
        } else {
          stop("Unsupported file extension: ", ext)
        }
      }, error = function(e) {
        message("Error processing file: ", filepath, ": ", e$message)
        NULL
      })

      if (is.null(df)) return(NULL)
      tibble::as_tibble(df)
    })

    data_list <- Filter(Negate(is.null), data_list)
    data <- dplyr::bind_rows(data_list)
  } else {
    stop("Unrecognized dataset_choice: ", dataset_choice)
  }

  return(data)
}


#' Render PDF Pages to Base64 PNG
#'
#' @description
#' Renders each page of a PDF as a PNG image and returns base64-encoded strings.
#' Uses pdftools (R-native, no Python required).
#'
#' @param file_path Character string path to PDF file
#' @param dpi Numeric, rendering resolution (default: 150)
#'
#' @return List of base64-encoded PNG strings, one per page
#' @keywords internal
render_pdf_pages_to_base64 <- function(file_path, dpi = 150) {
  if (!requireNamespace("pdftools", quietly = TRUE)) {
    stop("pdftools package required for PDF rendering")
  }

  n_pages <- pdftools::pdf_info(file_path)$pages
  if (n_pages == 0) return(list())

  pages <- lapply(seq_len(n_pages), function(i) {
    tryCatch({
      tmp <- tempfile(fileext = ".png")
      pdftools::pdf_convert(file_path, format = "png", pages = i,
                            dpi = dpi, filenames = tmp, verbose = FALSE)
      raw_bytes <- readBin(tmp, "raw", file.info(tmp)$size)
      unlink(tmp)
      jsonlite::base64_enc(raw_bytes)
    }, error = function(e) NULL)
  })
  Filter(Negate(is.null), pages)
}


.extract_multimodal_pdf <- function(file_path, vision_provider,
                                    vision_model, api_key,
                                    describe_images) {
  if (is.null(vision_model)) {
    vision_model <- switch(vision_provider,
      "openai" = "gpt-4.1",
      "gemini" = "gemini-2.5-flash",
      "gemini-2.5-flash"
    )
  }

  text_pages <- tryCatch(pdftools::pdf_text(file_path), error = function(e) character(0))

  page_images <- render_pdf_pages_to_base64(file_path)

  image_descriptions <- list()
  num_described <- 0

  if (describe_images && length(page_images) > 0) {
    for (i in seq_along(page_images)) {
      page_text_len <- if (i <= length(text_pages)) nchar(trimws(text_pages[i])) else 0
      if (page_text_len > 500) next

      desc <- describe_image(
        image_base64 = page_images[[i]],
        provider = vision_provider,
        model = vision_model,
        api_key = api_key
      )
      if (!is.null(desc)) {
        image_descriptions[[length(image_descriptions) + 1]] <- desc
        num_described <- num_described + 1
      }
    }
  }

  all_text <- paste(trimws(text_pages), collapse = "\n\n")
  if (length(image_descriptions) > 0) {
    all_text <- paste0(all_text, "\n\n", paste(image_descriptions, collapse = "\n\n"))
  }

  combined_lines <- strsplit(all_text, "\n")[[1]]
  combined_lines <- combined_lines[nchar(trimws(combined_lines)) > 0]

  list(
    success = TRUE,
    data = data.frame(text = combined_lines, stringsAsFactors = FALSE),
    type = "multimodal",
    method = "multimodal",
    message = paste("Extracted text and", num_described, "page descriptions"),
    num_images = num_described,
    vision_provider = vision_provider
  )
}

.extract_python_pdf <- function(file_path) {
  env_check <- check_python_env()
  if (!env_check$available) {
    stop("Python environment not available")
  }

  pdf_data <- process_pdf_file_py(file_path, content_type = "auto")

  if (!pdf_data$success || is.null(pdf_data$data)) {
    stop(pdf_data$message)
  }

  list(
    success = TRUE,
    data = pdf_data$data,
    type = pdf_data$type,
    method = "python",
    message = paste("Python extraction:", pdf_data$message)
  )
}

.extract_r_pdf <- function(file_path) {
  if (!requireNamespace("pdftools", quietly = TRUE)) {
    stop("pdftools package required")
  }

  pages <- pdftools::pdf_text(file_path)
  lines <- unlist(lapply(pages, function(page) {
    page_lines <- strsplit(page, "\n")[[1]]
    trimws(page_lines)
  }))
  lines <- lines[lines != ""]

  if (length(lines) == 0) {
    stop("No extractable text found")
  }

  list(
    success = TRUE,
    data = data.frame(text = lines, stringsAsFactors = FALSE),
    type = "text",
    method = "r",
    message = paste("R extraction: Extracted", length(lines), "lines")
  )
}

#' @title Process PDF File (Unified Entry Point)
#'
#' @description
#' Unified PDF processing:
#' 1. Multimodal (R-native pdftools + Vision LLM) if enabled
#' 2. R pdftools text extraction as fallback
#'
#' @param file_path Character string path to PDF file
#' @param use_multimodal Logical, enable multimodal extraction
#' @param vision_provider Character, "openai" or "gemini"
#' @param vision_model Character, model name
#' @param api_key Character, API key (if using openai/gemini)
#' @param describe_images Logical, generate image descriptions
#'
#' @return List: success, data, type, method, message
#'
#' @concept preprocessing
#' @seealso [import_files()] for the generic file-loading entry point; [describe_image()] for standalone image captioning
#' @export
process_pdf_unified <- function(file_path,
                                use_multimodal = FALSE,
                                vision_provider = "gemini",
                                vision_model = NULL,
                                api_key = NULL,
                                describe_images = TRUE) {

  if (!file.exists(file_path)) {
    return(list(success = FALSE, data = NULL, type = "error",
                method = "none", message = "File not found"))
  }

  if (use_multimodal) {
    prereq_check <- check_multimodal_prerequisites(
      vision_provider = vision_provider,
      vision_model = vision_model,
      api_key = api_key
    )

    if (!prereq_check$ready) {
      return(list(
        success = FALSE,
        data = NULL,
        type = "prerequisite_error",
        method = "multimodal",
        message = prereq_check$instructions,
        missing = prereq_check$missing
      ))
    }

    result <- tryCatch(
      .extract_multimodal_pdf(file_path, vision_provider, vision_model,
                              api_key, describe_images),
      error = function(e) list(
        success = FALSE,
        type = "extraction_error",
        message = paste("Multimodal extraction failed:", e$message)
      )
    )

    if (result$success) return(result)

    return(list(
      success = FALSE,
      data = NULL,
      type = "extraction_error",
      method = "multimodal",
      message = paste0(
        "Multimodal extraction encountered an error.\n\n",
        "Error: ", result$message
      )
    ))
  }

  tryCatch(
    .extract_r_pdf(file_path),
    error = function(e) {
      list(success = FALSE, data = NULL, type = "error", method = "none",
           message = paste("PDF extraction failed:", e$message))
    }
  )
}



#' @title Unite Text Columns
#'
#' @description
#' This function unites specified text columns in a data frame into a single
#' column named "united_texts" while retaining the original columns.
#'
#' @param df A data frame that contains text data.
#' @param listed_vars A character vector of column names to be united into
#'   "united_texts".
#'
#' @return A data frame with a new column "united_texts" created by uniting
#'   the specified variables.
#'
#' @concept preprocessing
#' @seealso [prep_texts()] to tokenize and clean the united text; [import_files()] to load source data first
#' @export
#'
#' @examples
#' \donttest{
#'   mydata <- TextAnalysisR::SpecialEduTech
#'
#'   united_tbl <- TextAnalysisR::unite_cols(
#'     mydata,
#'     listed_vars = c("title", "keyword", "abstract")
#'   )
#'   print(united_tbl)
#' }
unite_cols <- function(df, listed_vars) {
  united_texts_tbl <- df %>%
    dplyr::select(all_of(unname(listed_vars))) %>%
    dplyr::mutate(dplyr::across(dplyr::everything(), ~ tidyr::replace_na(as.character(.x), ""))) %>%
    tidyr::unite(col = "united_texts", sep = " ", remove = FALSE) %>%
    dplyr::mutate(united_texts = trimws(gsub("\\s+", " ", united_texts)))

  docvar_tbl <- df %>%
    dplyr::select(-all_of(unname(listed_vars)))

  united_tbl <- dplyr::bind_cols(united_texts_tbl, docvar_tbl)

  return(united_tbl)
}


#' @title Preprocess Text Data
#'
#' @description
#' Preprocesses text data following the complete workflow implemented in the Shiny application:
#' - Constructing a corpus from united texts
#' - Tokenizing text into words with configurable options
#' - Converting to lowercase with acronym preservation option
#' - Applying character length filtering
#' - Optional multi-word expression detection and compound term creation
#' - Stopword removal and lemmatization capabilities
#'
#' This function serves as the foundation for all subsequent text analysis workflows.
#'
#' @param united_tbl A data frame that contains text data.
#' @param text_field The name of the column that contains the text data.
#' @param min_char The minimum number of characters for a token to be included (default: 2).
#' @param lowercase Logical; convert all tokens to lowercase (default: TRUE). Recommended for most text analysis tasks.
#' @param remove_punct Logical; remove punctuation from the text (default: TRUE).
#' @param remove_symbols Logical; remove symbols from the text (default: TRUE).
#' @param remove_numbers Logical; remove numbers from the text (default: TRUE).
#' @param remove_url Logical; remove URLs from the text (default: TRUE).
#' @param remove_separators Logical; remove separators from the text (default: TRUE).
#' @param split_hyphens Logical; split hyphenated words into separate tokens (default: TRUE).
#' @param split_tags Logical; split tags into separate tokens (default: TRUE).
#' @param include_docvars Logical; include document variables in the tokens object (default: TRUE).
#' @param keep_acronyms Logical; keep acronyms in the text (default: FALSE).
#' @param padding Logical; add padding to the tokens object (default: FALSE).
#' @param remove_stopwords Logical; remove stopwords from the text (default: FALSE).
#' @param stopwords_source Character; source for stopwords, e.g., "snowball", "stopwords-iso" (default: "snowball").
#' @param stopwords_language Character; language for stopwords (default: "en").
#' @param custom_stopwords Character vector; additional words to remove (default: NULL).
#' @param custom_valuetype Character; valuetype for custom_stopwords pattern matching, one of "glob", "regex", or "fixed" (default: "glob").
#' @param math_mode Logical; if \code{TRUE}, preserve math content (numbers, operators, symbols) by forcing \code{remove_punct}, \code{remove_symbols}, and \code{remove_numbers} all to \code{FALSE}, then strip only sentence-end punctuation such as periods, commas, question marks, exclamation marks, colons, semicolons, parentheses, brackets, braces, quotation marks, em dashes, and en dashes. The \code{min_char} default of 2 still applies, so noisy single-character tokens are dropped; pass \code{min_char = 1} to keep them. Use for math or STEM corpora where multi-character operators and numerals carry meaning (default: FALSE).
#' @param verbose Logical; print verbose output (default: FALSE).
#' @param ... Additional arguments passed to \code{quanteda::tokens}.
#'
#' @return A \code{tokens} object that contains the preprocessed text data.
#'
#' @import quanteda
#'
#' @concept preprocessing
#' @seealso [unite_cols()] to combine text columns first; [lemmatize_tokens()] to reduce words to base form (e.g., running -> run); `quanteda::dfm()` to build a document-feature matrix from the result
#' @export
#'
#' @examples
#' \donttest{
#' mydata <- TextAnalysisR::SpecialEduTech
#'
#' united_tbl <- TextAnalysisR::unite_cols(
#'   mydata,
#'   listed_vars = c("title", "keyword", "abstract")
#' )
#'
#' tokens <- TextAnalysisR::prep_texts(united_tbl,
#'                                          text_field = "united_texts",
#'                                          min_char = 2,
#'                                          lowercase = TRUE,
#'                                          remove_punct = TRUE,
#'                                          remove_symbols = TRUE,
#'                                          remove_numbers = TRUE,
#'                                          remove_url = TRUE,
#'                                          remove_separators = TRUE,
#'                                          split_hyphens = TRUE,
#'                                          split_tags = TRUE,
#'                                          include_docvars = TRUE,
#'                                          keep_acronyms = FALSE,
#'                                          padding = FALSE,
#'                                          verbose = FALSE)
#' print(tokens)
#' }
prep_texts <- function(united_tbl,
                             text_field = "united_texts",
                             min_char = 2,
                             lowercase = TRUE,
                             remove_punct = TRUE,
                             remove_symbols = TRUE,
                             remove_numbers = TRUE,
                             remove_url = TRUE,
                             remove_separators = TRUE,
                             split_hyphens = TRUE,
                             split_tags = TRUE,
                             include_docvars = TRUE,
                             keep_acronyms = FALSE,
                             padding = FALSE,
                             remove_stopwords = FALSE,
                             stopwords_source = "snowball",
                             stopwords_language = "en",
                             custom_stopwords = NULL,
                             custom_valuetype = "glob",
                             math_mode = FALSE,
                             verbose = FALSE,
                             ...) {

  if (math_mode) {
    remove_punct   <- FALSE
    remove_symbols <- FALSE
    remove_numbers <- FALSE
    min_char       <- 1L
  }

  start_time <- Sys.time()

  tryCatch({
    if (verbose) message("Creating corpus...")
    corp <- quanteda::corpus(united_tbl, text_field = text_field)

    if (verbose) message("Tokenizing texts...")
    toks <- quanteda::tokens(corp,
                   what = "word",
                   remove_punct = remove_punct,
                   remove_symbols = remove_symbols,
                   remove_numbers = remove_numbers,
                   remove_url = remove_url,
                   remove_separators = remove_separators,
                   split_hyphens = split_hyphens,
                   split_tags = split_tags,
                   include_docvars = include_docvars,
                   padding = padding,
                   verbose = verbose,
                   ...)

    if (lowercase) {
      if (verbose) message("Converting to lowercase...")
      toks <- quanteda::tokens_tolower(toks, keep_acronyms = keep_acronyms)
    }

    if (verbose) message("Applying minimum character filter...")
    tokens <- quanteda::tokens_select(toks,
                                      min_nchar = min_char,
                                      verbose = FALSE)

    if (remove_stopwords) {
      if (!requireNamespace("stopwords", quietly = TRUE)) {
        stop("Package 'stopwords' is required for stopword removal. Install with: install.packages('stopwords')")
      }
      if (verbose) message("Removing stopwords (", stopwords_language, ", ", stopwords_source, ")...")
      sw <- stopwords::stopwords(stopwords_language, source = stopwords_source)
      tokens <- quanteda::tokens_remove(tokens, pattern = sw, verbose = FALSE)
    }

    if (!is.null(custom_stopwords) && length(custom_stopwords) > 0) {
      if (verbose) message("Removing custom stopwords (", length(custom_stopwords), " words)...")
      tokens <- quanteda::tokens_remove(tokens, pattern = custom_stopwords,
                                        valuetype = custom_valuetype, verbose = FALSE)
    }

    if (math_mode) {
      if (verbose) message("Math mode: stripping sentence-end punctuation (keeping math operators)...")
      sentence_punct <- c(".", ",", "?", "!", ":", ";", "(", ")", "[", "]",
                          "{", "}", "\"", "'", "`", "''", "``", "...",
                          "\u2014", "\u2013")
      tokens <- quanteda::tokens_remove(tokens, pattern = sentence_punct,
                                        valuetype = "fixed", verbose = FALSE)
    }

    processing_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

    if (verbose) {
      message("Text preprocessing completed in ", round(processing_time, 2), " seconds")
      message("Processed ", quanteda::ndoc(tokens), " documents with ",
              sum(quanteda::ntoken(tokens)), " total tokens")
    }

    return(tokens)

  }, error = function(e) {
    stop("Error in text preprocessing: ", e$message)
  })
}


#' Lemmatize Tokens with Batch Processing
#'
#' @description
#' Converts tokens to their lemmatized forms using spaCy, with batch processing
#' to handle large document collections without timeout issues.
#'
#' @param tokens A quanteda tokens object to lemmatize.
#' @param batch_size Integer; number of documents to process per batch (default: 50).
#' @param model Character; spaCy model to use (default: "en_core_web_sm").
#' @param verbose Logical; print progress messages (default: TRUE).
#'
#' @return A quanteda tokens object containing lemmatized tokens.
#'
#' @details
#' Uses spaCy for linguistic lemmatization producing proper dictionary forms
#' (e.g., "studies" -> "study", "better" -> "good").
#' Batch processing prevents timeout errors with large document collections.
#'
#' @concept preprocessing
#' @export
#'
#' @examples
#' if (interactive()) {
#' tokens <- quanteda::tokens(c("The studies showed better results"))
#' lemmatized <- lemmatize_tokens(tokens, batch_size = 50)
#' }
lemmatize_tokens <- function(tokens,
                             batch_size = 50,
                             model = "en_core_web_sm",
                             verbose = TRUE) {

  if (!inherits(tokens, "tokens")) {
    stop("Input must be a quanteda tokens object")
  }

  texts <- sapply(tokens, paste, collapse = " ")
  doc_names <- names(texts)
  n_docs <- length(texts)

  if (n_docs == 0) {
    if (verbose) message("No documents to process")
    return(tokens)
  }

  start_time <- Sys.time()
  all_lemmas <- list()
  n_batches <- ceiling(n_docs / batch_size)

  for (i in seq_len(n_batches)) {
    start_idx <- (i - 1) * batch_size + 1
    end_idx <- min(i * batch_size, n_docs)

    if (verbose) {
      message(sprintf("Lemmatizing batch %d/%d (documents %d-%d)...",
                      i, n_batches, start_idx, end_idx))
    }

    batch_texts <- texts[start_idx:end_idx]
    batch_names <- doc_names[start_idx:end_idx]

    tryCatch({
      parsed <- spacy_parse_full(batch_texts, pos = FALSE, tag = FALSE,
                                 lemma = TRUE, entity = FALSE, model = model)

      for (j in seq_along(batch_names)) {
        doc_id <- batch_names[j]
        doc_lemmas <- parsed$lemma[parsed$doc_id == doc_id]
        all_lemmas[[doc_id]] <- doc_lemmas
      }
    }, error = function(e) {
      warning(sprintf("Error in batch %d: %s", i, e$message))
    })
  }

  if (length(all_lemmas) == 0) {
    stop("All batches failed. Check spaCy installation.")
  }

  lemmatized_tokens <- quanteda::as.tokens(all_lemmas)

  if (verbose) {
    processing_time <- difftime(Sys.time(), start_time, units = "secs")
    message(sprintf("Lemmatization completed in %.2f seconds (%d documents)",
                    as.numeric(processing_time), n_docs))
  }

  return(lemmatized_tokens)
}


#' Extract Text from PDF
#'
#' @description
#' Extracts text content from a PDF file using pdftools package.
#'
#' @param file_path Character string path to PDF file
#'
#' @return Data frame with columns: page (integer), text (character)
#'   Returns NULL if extraction fails or PDF is empty
#'
#' @details
#' Uses `pdftools::pdf_text()` to extract text from each page.
#' Preserves page structure and cleans whitespace.
#' Works best with text-based PDFs (not scanned images).
#'
#' @concept pdf
#' @keywords internal
#'
extract_text_from_pdf <- function(file_path) {
  if (!requireNamespace("pdftools", quietly = TRUE)) {
    stop("Package 'pdftools' is required. Install it with: install.packages('pdftools')")
  }

  tryCatch({
    text_pages <- pdftools::pdf_text(file_path)

    if (length(text_pages) == 0) {
      message("PDF file is empty or contains no extractable text")
      return(NULL)
    }

    text_pages <- trimws(text_pages)
    text_pages <- gsub("\\s+", " ", text_pages)

    non_empty_pages <- which(nchar(text_pages) > 0)

    if (length(non_empty_pages) == 0) {
      message("PDF contains no readable text")
      return(NULL)
    }

    df <- data.frame(
      page = seq_along(text_pages),
      text = text_pages,
      stringsAsFactors = FALSE
    )

    df <- df[nchar(df$text) > 0, ]

    attr(df, "pdf_type") <- "text"
    attr(df, "total_pages") <- length(text_pages)
    attr(df, "file_name") <- basename(file_path)

    return(df)
  }, error = function(e) {
    warning(paste("Error extracting text from PDF:", e$message))
    return(NULL)
  })
}


#' Detect PDF Content Type
#'
#' @description
#' Analyzes PDF to determine if it contains readable text.
#'
#' @param file_path Character string path to PDF file
#'
#' @return Character string: "text" or "unknown"
#'
#' @details
#' Attempts text extraction using pdftools. Returns "text" if successful,
#' or "unknown" if extraction fails or PDF is empty.
#'
#' For table extraction from PDFs, use \code{\link{extract_tables_from_pdf_py}}.
#'
#' @concept pdf
#' @export
#'
#' @examples
#' \donttest{
#' pdf_path <- "path/to/document.pdf"
#' content_type <- detect_pdf_content_type(pdf_path)
#' print(content_type)
#' }
detect_pdf_content_type <- function(file_path) {
  if (requireNamespace("pdftools", quietly = TRUE)) {
    text <- tryCatch({
      pdftools::pdf_text(file_path)
    }, error = function(e) NULL)

    if (!is.null(text) && length(text) > 0) {
      combined_text <- paste(text, collapse = " ")
      if (nchar(trimws(combined_text)) > 50) {
        return("text")
      }
    }
  }

  return("unknown")
}


#' Process PDF File
#'
#' @description
#' Main function to process PDF files - extracts text content using pdftools.
#' For table extraction, use \code{\link{process_pdf_file_py}}.
#'
#' @param file_path Character string path to PDF file
#' @param content_type Character string: "auto" or "text" (default: "auto")
#'
#' @return List with:
#'   - data: Data frame with extracted content
#'   - type: Character string indicating content type ("text" or "error")
#'   - success: Logical indicating success
#'   - message: Character string with status message
#'
#' @details
#' This function extracts text content from PDFs using pdftools package.
#' Works best with text-based PDFs (not scanned images).
#'
#' For PDFs containing tables or complex layouts, use the Python-based
#' \code{\link{process_pdf_file_py}} which provides better table extraction.
#'
#' @concept pdf
#' @keywords internal
#'
process_pdf_file <- function(file_path, content_type = "auto") {
  if (!file.exists(file_path)) {
    return(list(
      data = NULL,
      type = "error",
      success = FALSE,
      message = "File not found"
    ))
  }

  data <- extract_text_from_pdf(file_path)

  if (!is.null(data)) {
    return(list(
      data = data,
      type = "text",
      success = TRUE,
      message = paste("Successfully extracted text from", nrow(data), "pages")
    ))
  }

  return(list(
    data = NULL,
    type = "error",
    success = FALSE,
    message = "Could not extract content from PDF. File may be password-protected, corrupted, or scanned without OCR."
  ))
}



#' Extract Text from PDF using Python
#'
#' @description
#' Extracts text content from a PDF file using pdfplumber (Python).
#' No Java required - uses Python environment.
#'
#' @param file_path Character string path to PDF file
#' @param envname Character string, name of Python virtual environment
#'   (default: "textanalysisr-env")
#'
#' @return Data frame with columns: page (integer), text (character)
#'   Returns NULL if extraction fails or PDF is empty
#'
#' @details
#' Uses pdfplumber Python library through reticulate.
#' Requires Python environment setup. See \code{setup_python_env()}.
#'
#' @concept pdf
#' @keywords internal
#'
extract_text_from_pdf_py <- function(file_path, envname = NULL) {
  .ensure_python("pdfplumber", envname)

  tryCatch({
    pdf_module <- reticulate::import_from_path(
      "pdf_extraction",
      path = system.file("python", package = "TextAnalysisR")
    )

    result <- pdf_module$extract_text_from_pdf(file_path)

    if (!result$success) {
      message(result$message)
      return(NULL)
    }

    df <- do.call(rbind, lapply(result$data, as.data.frame))
    df$page <- as.integer(df$page)
    df$text <- as.character(df$text)

    attr(df, "pdf_type") <- "text"
    attr(df, "total_pages") <- result$total_pages
    attr(df, "file_name") <- basename(file_path)

    return(df)

  }, error = function(e) {
    warning(paste("Error extracting text from PDF:", e$message))
    return(NULL)
  })
}


#' Extract Tables from PDF using Python
#'
#' @description
#' Extracts tabular data from PDF using pdfplumber (Python).
#' No Java required - pure Python solution.
#'
#' @param file_path Character string path to PDF file
#' @param pages Integer vector of page numbers to process (NULL = all pages)
#' @param envname Character string, name of Python virtual environment
#'   (default: "textanalysisr-env")
#'
#' @return Data frame with extracted table data
#'   Returns NULL if no tables found or extraction fails
#'
#' @details
#' Uses pdfplumber Python library through reticulate.
#' Works with complex table layouts without Java dependency.
#'
#' @concept pdf
#' @keywords internal
#' @export
#'
#' @examples
#' if (interactive()) {
#' setup_python_env()
#'
#' pdf_path <- "path/to/table_document.pdf"
#' table_data <- extract_tables_from_pdf_py(pdf_path)
#' head(table_data)
#' }
extract_tables_from_pdf_py <- function(file_path, pages = NULL, envname = NULL) {
  .ensure_python("pdfplumber", envname)

  tryCatch({
    pdf_module <- reticulate::import_from_path(
      "pdf_extraction",
      path = system.file("python", package = "TextAnalysisR")
    )

    pages_list <- if (!is.null(pages)) as.list(as.integer(pages)) else NULL

    result <- pdf_module$extract_tables_from_pdf(file_path, pages = pages_list)

    if (!result$success) {
      message(result$message)
      return(NULL)
    }

    df <- as.data.frame(result$data, stringsAsFactors = FALSE)

    if (nrow(df) == 0) {
      message("Table extraction resulted in empty data frame")
      return(NULL)
    }

    attr(df, "pdf_type") <- "tabular"
    attr(df, "file_name") <- basename(file_path)
    attr(df, "num_tables") <- result$num_tables

    return(df)

  }, error = function(e) {
    warning(paste("Error extracting tables from PDF:", e$message))
    return(NULL)
  })
}


#' Detect PDF Content Type using Python
#'
#' @description
#' Analyzes PDF to determine if it contains primarily tabular data or text.
#'
#' @param file_path Character string path to PDF file
#' @param envname Character string, name of Python virtual environment
#'   (default: "textanalysisr-env")
#'
#' @return Character string: "tabular", "text", or "unknown"
#'
#' @concept pdf
#' @keywords internal
#' @export
#'
#' @examples
#' if (interactive()) {
#' setup_python_env()
#'
#' pdf_path <- "path/to/document.pdf"
#' content_type <- detect_pdf_content_type_py(pdf_path)
#' print(content_type)
#' }
detect_pdf_content_type_py <- function(file_path, envname = NULL) {
  ok <- tryCatch({ .ensure_python("pdfplumber", envname); TRUE }, error = function(e) FALSE)
  if (!ok) return("unknown")

  tryCatch({
    pdf_module <- reticulate::import_from_path(
      "pdf_extraction",
      path = system.file("python", package = "TextAnalysisR")
    )

    result <- pdf_module$detect_pdf_content_type(file_path)

    return(result$content_type)

  }, error = function(e) {
    return("unknown")
  })
}


#' Process PDF File using Python
#'
#' @description
#' Main function to process PDF files using pdfplumber (Python).
#' Automatically detects content type and extracts data accordingly.
#' No Java required.
#'
#' @param file_path Character string path to PDF file
#' @param content_type Character string: "auto", "text", or "tabular"
#'   If "auto", will detect content type automatically
#' @param envname Character string, name of Python virtual environment
#'   (default: "textanalysisr-env")
#'
#' @return List with:
#'   - data: Data frame with extracted content
#'   - type: Character string indicating content type
#'   - success: Logical indicating success
#'   - message: Character string with status message
#'
#' @details
#' This function uses Python's pdfplumber library which:
#' - Handles both text and tables
#' - No Java dependency
#' - Better accuracy than tabulizer for complex tables
#' - Uses TextAnalysisR Python environment
#'
#' @concept pdf
#' @keywords internal
#'
process_pdf_file_py <- function(file_path, content_type = "auto", envname = NULL) {
  if (!file.exists(file_path)) {
    return(list(data = NULL, type = "error", success = FALSE, message = "File not found"))
  }

  ok <- tryCatch({ .ensure_python("pdfplumber", envname); TRUE }, error = function(e) e$message)
  if (!isTRUE(ok)) {
    return(list(data = NULL, type = "error", success = FALSE, message = as.character(ok)))
  }

  tryCatch({
    pdf_module <- reticulate::import_from_path(
      "pdf_extraction",
      path = system.file("python", package = "TextAnalysisR")
    )

    result <- pdf_module$process_pdf_file(file_path, content_type = content_type)

    if (!result$success) {
      return(list(
        data = NULL,
        type = result$type,
        success = FALSE,
        message = result$message
      ))
    }

    if (result$type == "tabular") {
      df <- as.data.frame(result$data, stringsAsFactors = FALSE)
    } else {
      df <- do.call(rbind, lapply(result$data, as.data.frame))
      df$page <- as.integer(df$page)
      df$text <- as.character(df$text)

      if (!"category" %in% names(df)) {
        df$category <- "Uploaded"
      }
    }

    return(list(
      data = df,
      type = result$type,
      success = TRUE,
      message = result$message
    ))

  }, error = function(e) {
    return(list(
      data = NULL,
      type = "error",
      success = FALSE,
      message = paste("Error processing PDF:", e$message)
    ))
  })
}




#' Check Multimodal Prerequisites
#'
#' @description
#' Checks all prerequisites for multimodal PDF extraction.
#' Uses R-native pdftools for rendering (no Python required).
#'
#' @param vision_provider Character: "openai" or "gemini"
#' @param vision_model Character: Model name (optional)
#' @param api_key Character: API key for OpenAI/Gemini
#' @param envname Character: Kept for backward compatibility, ignored
#'
#' @return List with:
#'   - ready: Logical - TRUE if all prerequisites met
#'   - missing: Character vector of missing components
#'   - instructions: Character - Detailed setup instructions
#'   - details: List with component-specific status
#'
#' @keywords internal
check_multimodal_prerequisites <- function(
  vision_provider = "gemini",
  vision_model = NULL,
  api_key = NULL,
  envname = "textanalysisr-env"
) {
  missing <- character(0)
  details <- list()
  instructions <- character(0)

  pdftools_ok <- requireNamespace("pdftools", quietly = TRUE)
  if (!pdftools_ok) {
    missing <- c(missing, "pdftools package")
    instructions <- c(instructions,
      "pdftools package required.\nInstall: install.packages('pdftools')"
    )
  }
  details$pdftools <- list(available = pdftools_ok)

  if (vision_provider == "openai") {
    if (is.null(api_key) || nchar(api_key) == 0) {
      missing <- c(missing, "OpenAI API key")
      instructions <- c(instructions, .missing_api_key_message("openai", "shiny"))
    }
    details$openai <- list(api_key_provided = !is.null(api_key) && nchar(api_key) > 0)

  } else if (vision_provider == "gemini") {
    if (is.null(api_key) || nchar(api_key) == 0) {
      missing <- c(missing, "Gemini API key")
      instructions <- c(instructions, .missing_api_key_message("gemini", "shiny"))
    }
    details$gemini <- list(api_key_provided = !is.null(api_key) && nchar(api_key) > 0)
  }

  ready <- length(missing) == 0

  if (!ready) {
    full_instructions <- paste0(
      "Multimodal extraction requires:\n\n",
      paste(paste0(seq_along(instructions), ". ", instructions), collapse = "\n\n")
    )
  } else {
    full_instructions <- "All prerequisites met"
  }

  return(list(
    ready = ready,
    missing = missing,
    instructions = full_instructions,
    details = details
  ))
}
#' Extract PDF with Multimodal Analysis
#'
#' @description
#' Extract both text and visual content from PDFs using R-native pdftools
#' and vision LLM APIs. No Python required.
#'
#' @param file_path Character string path to PDF file
#' @param vision_provider Character: "openai" or "gemini" (default)
#' @param vision_model Character: Model name
#'   - For OpenAI: "gpt-4.1", "gpt-4.1-mini"
#'   - For Gemini: "gemini-2.5-flash", "gemini-2.5-pro"
#' @param api_key Character: API key (required for openai/gemini providers)
#' @param describe_images Logical: Convert page images to text descriptions (default: TRUE)
#' @param envname Character: Kept for backward compatibility, ignored
#'
#' @return List with:
#'   - success: Logical
#'   - combined_text: Character string with all content for text analysis
#'   - text_content: List of text chunks
#'   - image_descriptions: List of image descriptions
#'   - num_images: Integer count of described pages
#'   - vision_provider: Character indicating provider used
#'   - message: Character status message
#'
#' @details
#' **Workflow:**
#' 1. Extracts text using pdftools (R-native)
#' 2. Renders each page as an image
#' 3. Sends sparse-text pages to vision LLM for description
#' 4. Merges text + descriptions into a single text corpus
#'
#' @concept pdf
#' @keywords internal
#'
extract_pdf_multimodal <- function(
  file_path,
  vision_provider = "gemini",
  vision_model = NULL,
  api_key = NULL,
  describe_images = TRUE,
  envname = "textanalysisr-env"
) {
  if (!file.exists(file_path)) {
    return(list(success = FALSE, message = "File not found"))
  }

  if (is.null(vision_model)) {
    vision_model <- switch(vision_provider,
      "openai" = "gpt-4.1",
      "gemini" = "gemini-2.5-flash",
      "gemini-2.5-flash"
    )
  }

  if (vision_provider == "openai") {
    if (is.null(api_key) || !nzchar(api_key)) {
      return(list(success = FALSE, message = .missing_api_key_message("openai", "shiny")))
    }
  } else if (vision_provider == "gemini") {
    if (is.null(api_key) || !nzchar(api_key)) {
      return(list(success = FALSE, message = .missing_api_key_message("gemini", "shiny")))
    }
  }

  tryCatch({
    text_pages <- pdftools::pdf_text(file_path)
    text_content <- list(paste(trimws(text_pages), collapse = "\n\n"))

    image_descriptions <- list()
    num_described <- 0

    if (describe_images) {
      page_images <- render_pdf_pages_to_base64(file_path)

      for (i in seq_along(page_images)) {
        page_text_len <- if (i <= length(text_pages)) nchar(trimws(text_pages[i])) else 0
        if (page_text_len > 500) next

        desc <- describe_image(
          image_base64 = page_images[[i]],
          provider = vision_provider,
          model = vision_model,
          api_key = api_key
        )
        if (!is.null(desc)) {
          image_descriptions[[length(image_descriptions) + 1]] <- desc
          num_described <- num_described + 1
        }
      }
    }

    combined_text <- paste(trimws(text_pages), collapse = "\n\n")
    if (length(image_descriptions) > 0) {
      combined_text <- paste0(combined_text, "\n\n", paste(image_descriptions, collapse = "\n\n"))
    }

    return(list(
      success = TRUE,
      text_content = text_content,
      image_descriptions = image_descriptions,
      combined_text = combined_text,
      total_pages = length(text_pages),
      num_images = num_described,
      vision_provider = vision_provider,
      message = paste("Extracted text and", num_described, "page descriptions")
    ))

  }, error = function(e) {
    return(list(success = FALSE, message = paste("Error:", e$message)))
  })
}


#' Smart PDF Extraction with Auto-Detection
#'
#' @description
#' Extracts text and visual content from PDFs using R-native pdftools
#' and vision LLM APIs. Routes directly to multimodal extraction.
#'
#' @param file_path Character string path to PDF file
#' @param doc_type Character: "auto" (default), "academic", or "general" (kept for compatibility)
#' @param vision_provider Character: "openai" or "gemini" (default)
#' @param vision_model Character: Model name for vision analysis
#' @param api_key Character: API key for cloud providers
#' @param envname Character: Kept for backward compatibility, ignored
#'
#' @return List with extracted content ready for text analysis
#'
#' @concept pdf
#' @keywords internal
#'
extract_pdf_smart <- function(
  file_path,
  doc_type = "auto",
  vision_provider = "gemini",
  vision_model = NULL,
  api_key = NULL,
  envname = "textanalysisr-env"
) {
  if (!file.exists(file_path)) {
    return(list(success = FALSE, message = "File not found"))
  }

  extract_pdf_multimodal(
    file_path = file_path,
    vision_provider = vision_provider,
    vision_model = vision_model,
    api_key = api_key,
    describe_images = TRUE
  )
}


#' Check Vision Model Availability
#'
#' @description
#' Check if required vision models are available for multimodal processing.
#'
#' @param provider Character: "openai" or "gemini"
#' @param api_key Character: API key (for OpenAI/Gemini)
#'
#' @return List with availability status and recommendations
#'
#' @concept pdf
#' @keywords internal
#' @export
#'
#' @examples
#' \donttest{
#' status <- check_vision_models("openai", api_key = Sys.getenv("OPENAI_API_KEY"))
#' status <- check_vision_models("gemini", api_key = Sys.getenv("GEMINI_API_KEY"))
#' }
check_vision_models <- function(provider = "gemini", api_key = NULL) {
  if (provider == "openai") {
    if (is.null(api_key) || !nzchar(api_key)) {
      return(list(available = FALSE, message = "OpenAI API key required"))
    }

    valid <- nchar(api_key) > 20 && grepl("^sk-", api_key)
    return(list(
      available = valid,
      message = if (valid) "API key format valid" else "Invalid API key format"
    ))

  } else if (provider == "gemini") {
    if (is.null(api_key) || !nzchar(api_key)) {
      return(list(available = FALSE, message = "Gemini API key required"))
    }

    valid <- nchar(api_key) > 10
    return(list(
      available = valid,
      message = if (valid) "API key provided" else "Invalid API key"
    ))
  }

  return(list(
    available = FALSE,
    message = paste("Unknown provider:", provider)
  ))
}

test_that("prep_texts works with default settings", {
  df <- data.frame(united_texts = c(
    "Solve 2 + 3 = 5 for the value of x.",
    "Find the area of a circle with radius 4 cm."
  ))
  toks <- prep_texts(df, text_field = "united_texts")
  expect_s3_class(toks, "tokens")
  expect_equal(length(toks), 2)
})

test_that("math_mode = TRUE preserves numbers", {
  df <- data.frame(united_texts = c("Compute 3 + 4 = 7 quickly."))
  toks_default <- prep_texts(df, text_field = "united_texts")
  toks_math    <- prep_texts(df, text_field = "united_texts", math_mode = TRUE)

  flat_default <- as.character(unlist(as.list(toks_default)))
  flat_math    <- as.character(unlist(as.list(toks_math)))

  expect_false(any(c("3", "4", "7") %in% flat_default))
  expect_true(all(c("3", "4", "7") %in% flat_math))
})

test_that("math_mode = TRUE preserves math operators", {
  df <- data.frame(united_texts = c("Solve x + y = 10 and x - y = 2."))
  toks <- prep_texts(df, text_field = "united_texts", math_mode = TRUE)
  flat <- as.character(unlist(as.list(toks)))
  expect_true("+" %in% flat)
  expect_true("=" %in% flat)
  expect_true("-" %in% flat)
})

test_that("math_mode = TRUE strips sentence-end punctuation", {
  df <- data.frame(united_texts = c("Solve 2 + 3, then check the answer."))
  toks <- prep_texts(df, text_field = "united_texts", math_mode = TRUE)
  flat <- as.character(unlist(as.list(toks)))
  expect_false("." %in% flat)
  expect_false("," %in% flat)
})

test_that("math_mode overrides remove_* flags even when set TRUE", {
  df <- data.frame(united_texts = c("3 + 4 = 7"))
  toks <- prep_texts(df, text_field = "united_texts",
                     math_mode = TRUE,
                     remove_punct   = TRUE,
                     remove_symbols = TRUE,
                     remove_numbers = TRUE)
  flat <- as.character(unlist(as.list(toks)))
  expect_true("3" %in% flat)
  expect_true("+" %in% flat)
})

test_that("process_pdf_unified validates file existence", {
  result <- process_pdf_unified("nonexistent.pdf")

  expect_false(result$success)
  expect_equal(result$type, "error")
  expect_equal(result$method, "none")
  expect_match(result$message, "File not found")
})

test_that("process_pdf_unified falls back to R when Python unavailable", {
  skip_if_not_installed("pdftools")
  skip_if(!file.exists("test-sample.pdf"))

  result <- process_pdf_unified(
    "test-sample.pdf",
    use_multimodal = FALSE
  )

  expect_type(result, "list")
  expect_true("success" %in% names(result))
  expect_true("method" %in% names(result))
  expect_true("data" %in% names(result))
})

test_that("import_files handles DOCX text extraction", {
  skip_if_not_installed("officer")
  skip_if(!file.exists("test-sample.docx"))

  file_info <- data.frame(
    filepath = "test-sample.docx",
    stringsAsFactors = FALSE
  )

  result <- import_files(
    dataset_choice = "Upload Your File",
    file_info = file_info
  )

  expect_s3_class(result, "tbl_df")
  expect_true("text" %in% names(result))
  expect_true(nrow(result) > 0)
})

test_that("import_files handles CSV files", {
  temp_csv <- tempfile(fileext = ".csv")
  write.csv(
    data.frame(text = c("Test1", "Test2"), stringsAsFactors = FALSE),
    temp_csv,
    row.names = FALSE
  )

  file_info <- data.frame(filepath = temp_csv, stringsAsFactors = FALSE)
  result <- import_files("Upload Your File", file_info = file_info)

  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 2)

  unlink(temp_csv)
})

test_that("import_files handles TXT files", {
  temp_txt <- tempfile(fileext = ".txt")
  writeLines(c("Line 1", "Line 2"), temp_txt)

  file_info <- data.frame(filepath = temp_txt, stringsAsFactors = FALSE)
  result <- import_files("Upload Your File", file_info = file_info)

  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 2)

  unlink(temp_txt)
})

test_that("import_files validates empty files", {
  temp_txt <- tempfile(fileext = ".txt")
  writeLines("", temp_txt)

  file_info <- data.frame(filepath = temp_txt, stringsAsFactors = FALSE)
  result <- import_files("Upload Your File", file_info = file_info)

  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 0)

  unlink(temp_txt)
})

test_that("unite_cols pastes selected columns with single-space separator", {
  df <- data.frame(
    title = c("Hello", "Foo"),
    body  = c("World", "Bar"),
    other = c("x", "y"),
    stringsAsFactors = FALSE
  )
  out <- unite_cols(df, listed_vars = c("title", "body"))
  expect_true("united_texts" %in% names(out))
  expect_equal(out$united_texts, c("Hello World", "Foo Bar"))
})

test_that("unite_cols replaces NA cells with empty string", {
  df <- data.frame(
    title = c("Alpha", NA, "Gamma"),
    body  = c(NA, "Beta", "Delta"),
    stringsAsFactors = FALSE
  )
  out <- unite_cols(df, listed_vars = c("title", "body"))
  expect_equal(out$united_texts, c("Alpha", "Beta", "Gamma Delta"))
  expect_false(any(grepl("\\bNA\\b", out$united_texts)))
})

test_that("unite_cols collapses internal whitespace from NA gaps", {
  df <- data.frame(
    a = c("one", NA),
    b = c(NA, "two"),
    c = c("three", "four"),
    stringsAsFactors = FALSE
  )
  out <- unite_cols(df, listed_vars = c("a", "b", "c"))
  expect_equal(out$united_texts, c("one three", "two four"))
})

test_that("unite_cols preserves docvar columns alongside united_texts", {
  df <- data.frame(
    title = c("A", "B"),
    body  = c("a", "b"),
    year  = c(2020L, 2021L),
    group = c("x", "y"),
    stringsAsFactors = FALSE
  )
  out <- unite_cols(df, listed_vars = c("title", "body"))
  expect_true(all(c("year", "group") %in% names(out)))
  expect_equal(out$year, c(2020L, 2021L))
})

suppressPackageStartupMessages({
  library(shiny)
  library(shinyBS)
  library(shinyjs)
})

# Detect web/Docker deployment (hide GPU option in these environments)
is_web <- tryCatch({
  TextAnalysisR:::check_web_deployment()
}, error = function(e) FALSE)
is_docker <- tryCatch({
  TextAnalysisR:::check_docker_deployment()
}, error = function(e) FALSE)
is_remote_ui <- is_web || is_docker


ui <- fluidPage(
  lang = "en",
  useShinyjs(),
  shinybusy::add_busy_spinner(
    spin = "fading-circle",
    position = "full-page",
    color = "#337ab7",
    height = "100px",
    width = "100px"
  ),
  tags$head(
    tags$meta(charset = "UTF-8"),
    tags$meta(name = "viewport", content = "width=device-width, initial-scale=1.0"),
    tags$meta(name = "description", content = "TextAnalysisR: A text mining workflow tool"),
    tags$link(rel = "icon", type = "image/png", href = "logo.png"),
    tags$script(HTML(
      "document.querySelectorAll('title').forEach(function(t, i) {
         if (i > 0 || !t.textContent.trim()) t.remove();
       });"
    )),
    tags$meta(name = "keywords", content = "text mining, topic modeling, semantic analysis, R Shiny"),

    tags$meta(
      `http-equiv` = "Content-Security-Policy",
      content = paste(
        "default-src 'self';",
        "script-src 'self' 'unsafe-inline' 'unsafe-eval' https://cdn.jsdelivr.net https://cdn.plot.ly https://translate.google.com https://translate.googleapis.com https://translate.googleusercontent.com https://translate-pa.googleapis.com https://static.cloudflareinsights.com;",
        "style-src 'self' 'unsafe-inline' https://fonts.googleapis.com https://translate.googleapis.com https://translate.googleusercontent.com https://www.gstatic.com https://use.fontawesome.com https://cdnjs.cloudflare.com https://ka-f.fontawesome.com;",
        "font-src 'self' https://fonts.gstatic.com https://use.fontawesome.com https://cdnjs.cloudflare.com https://ka-f.fontawesome.com data:;",
        "img-src 'self' data: https: https://www.gstatic.com https://translate.google.com;",
        "connect-src 'self' http://127.0.0.1:* http://localhost:* https://translate.googleapis.com https://translate-pa.googleapis.com https://cloudflareinsights.com;",
        "frame-src 'self' http: https: https://translate.google.com https://translate.googleusercontent.com;"
      )
    ),
    tags$meta(`http-equiv` = "X-Content-Type-Options", content = "nosniff"),
    tags$meta(`http-equiv` = "X-XSS-Protection", content = "1; mode=block"),
    tags$meta(`http-equiv` = "Referrer-Policy", content = "no-referrer-when-downgrade"),

    tags$meta(name = "theme-color", content = "#337ab7"),

    # Font Awesome icons (preload to avoid render-blocking)
    tags$link(
      rel = "preload",
      href = "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.4/css/all.min.css",
      as = "style",
      onload = "this.onload=null;this.rel='stylesheet'",
      integrity = "sha512-1ycn6IcaQQ40/MKBW2W4Rhis/DbILU74C1vSrLJxCq57o941Ym01SwNsOMqvEBFlcgUa6xLiPY/NS5R+E6ztJQ==",
      crossorigin = "anonymous",
      referrerpolicy = "no-referrer"
    ),

    # Google Translate API loads on demand the first time the user opens the
    # language picker. Avoids ~600 ms of render-blocking 3rd-party JS for every
    # session, since most researchers never touch the translation widget.
    tags$script(HTML("
      window.loadGoogleTranslate = function() {
        if (window.__gtLoaded) return;
        window.__gtLoaded = true;
        var s = document.createElement('script');
        s.src = 'https://translate.google.com/translate_a/element.js?cb=googleTranslateElementInit';
        s.async = true;
        document.head.appendChild(s);
      };
    ")),

    tags$style(HTML("
      .password-wrapper { position: relative; }
      .password-wrapper input[type='password'],
      .password-wrapper input[type='text'] { padding-right: 36px; }
      .password-toggle {
        position: absolute; right: 8px; top: 50%; transform: translateY(-50%);
        background: none; border: none; cursor: pointer; color: #888;
        padding: 2px 4px; font-size: 14px; line-height: 1; z-index: 2;
      }
      .password-toggle:hover { color: #333; }
    ")),
    tags$script(HTML("
      $(document).on('shiny:value shiny:inputchanged shiny:bound', function() {
        setTimeout(function() {
          $('input[type=\"password\"]').each(function() {
            if ($(this).parent().hasClass('password-wrapper')) return;
            $(this).wrap('<div class=\"password-wrapper\"></div>');
            var btn = $('<button type=\"button\" class=\"password-toggle\" title=\"Show/hide\" aria-label=\"Toggle visibility\"><i class=\"fa fa-eye\"></i></button>');
            $(this).after(btn);
          });
        }, 100);
      });
      $(document).on('click', '.password-toggle', function(e) {
        e.preventDefault();
        var input = $(this).siblings('input');
        var icon = $(this).find('i');
        if (input.attr('type') === 'password') {
          input.attr('type', 'text');
          icon.removeClass('fa-eye').addClass('fa-eye-slash');
        } else {
          input.attr('type', 'password');
          icon.removeClass('fa-eye-slash').addClass('fa-eye');
        }
      });
      $(document).ready(function() {
        $('input[type=\"password\"]').each(function() {
          if ($(this).parent().hasClass('password-wrapper')) return;
          $(this).wrap('<div class=\"password-wrapper\"></div>');
          var btn = $('<button type=\"button\" class=\"password-toggle\" title=\"Show/hide\" aria-label=\"Toggle visibility\"><i class=\"fa fa-eye\"></i></button>');
          $(this).after(btn);
        });
      });
    "))
  ),

  tags$div(
    class = "top-right-controls",
    tags$a(
      id = "tts_toggle",
      href = "javascript:void(0);",
      style = "cursor: pointer; text-decoration: none; color: #5F7088; font-size: 26px; user-select: none;",
      `aria-label` = "Text to speech",
      title = "Text to Speech (Alt+S)",
      tags$i(class = "fa fa-volume-up", id = "tts_icon", `aria-hidden` = "true", style = "pointer-events: none;")
    ),
    tags$a(
      id = "dark_mode_toggle",
      href = "javascript:void(0);",
      onclick = "toggleDarkMode()",
      style = "cursor: pointer; text-decoration: none; color: #5F7088; font-size: 30px;",
      `aria-label` = "Toggle dark mode",
      title = "Toggle dark mode",
      tags$i(class = "fa fa-moon", `aria-hidden` = "true")
    ),
    tags$div(
      style = "position: relative; display: flex; align-items: center; gap: 8px;",
      tags$a(
        id = "translate_icon",
        href = "javascript:void(0);",
        style = "cursor: pointer; text-decoration: none; color: #5F7088; font-size: 26px; user-select: none;",
        `aria-label` = "Select language",
        title = "Select Language",
        tags$i(class = "fa fa-globe", `aria-hidden` = "true", style = "pointer-events: none;")
      ),
      # Hidden Google Translate element (accessed programmatically)
      tags$div(
        id = "google_translate_element",
        style = "position: absolute; left: -9999px; top: -9999px;"
      ),
      # Custom language menu dropdown
      tags$div(
        id = "translate_dropdown",
        class = "notranslate",
        translate = "no",
        style = "display: none; position: absolute; top: 100%; right: 0; margin-top: 8px; background: white; border: 1px solid #e5e7eb; border-radius: 8px; box-shadow: 0 4px 16px rgba(0,0,0,0.15); z-index: 10000; min-width: 200px; max-height: 360px; overflow-y: auto;",
        # Language buttons
        tags$button(class = "lang-btn active", `data-lang` = "en", lang = "en", tags$i(class = "fa fa-check"), " English"),
        tags$button(class = "lang-btn", `data-lang` = "de", lang = "de", tags$i(class = "fa fa-check"), " Deutsch"),
        tags$button(class = "lang-btn", `data-lang` = "es", lang = "es", tags$i(class = "fa fa-check"), " Español"),
        tags$button(class = "lang-btn", `data-lang` = "fr", lang = "fr", tags$i(class = "fa fa-check"), " Français"),
        tags$button(class = "lang-btn", `data-lang` = "hi", lang = "hi", tags$i(class = "fa fa-check"), " हिन्दी"),
        tags$button(class = "lang-btn", `data-lang` = "id", lang = "id", tags$i(class = "fa fa-check"), " Bahasa Indonesia"),
        tags$button(class = "lang-btn", `data-lang` = "it", lang = "it", tags$i(class = "fa fa-check"), " Italiano"),
        tags$button(class = "lang-btn", `data-lang` = "ja", lang = "ja", tags$i(class = "fa fa-check"), " 日本語"),
        tags$button(class = "lang-btn", `data-lang` = "ko", lang = "ko", tags$i(class = "fa fa-check"), " 한국어"),
        tags$button(class = "lang-btn", `data-lang` = "nl", lang = "nl", tags$i(class = "fa fa-check"), " Nederlands"),
        tags$button(class = "lang-btn", `data-lang` = "pl", lang = "pl", tags$i(class = "fa fa-check"), " Polski"),
        tags$button(class = "lang-btn", `data-lang` = "pt", lang = "pt", tags$i(class = "fa fa-check"), " Português"),
        tags$button(class = "lang-btn", `data-lang` = "ru", lang = "ru", tags$i(class = "fa fa-check"), " Русский"),
        tags$button(class = "lang-btn", `data-lang` = "zh-CN", lang = "zh-Hans", tags$i(class = "fa fa-check"), " 中文 (简体)"),
        tags$button(class = "lang-btn", `data-lang` = "zh-TW", lang = "zh-Hant", tags$i(class = "fa fa-check"), " 中文 (繁體)"),
        # Attribution at bottom
        tags$div(
          class = "translate-attribution",
          style = "display: flex; align-items: center; justify-content: center; gap: 4px; font-size: 16px; color: #9ca3af; padding: 8px 12px; border-top: 1px solid #e5e7eb; background: #f9fafb; border-radius: 0 0 8px 8px; white-space: nowrap; flex-wrap: nowrap;",
          tags$span("Powered by"),
          tags$img(src = "https://www.gstatic.com/images/branding/googlelogo/1x/googlelogo_color_42x16dp.png", alt = "Google", height = "12"),
          tags$span("Translate")
        )
      )
    )
  ),
  includeCSS("www/styles.css"),
  includeCSS("www/mobile.css"),
  includeScript("www/script.js"),

  tags$script(HTML(
    "$(document).on('shiny:connected', function(){",
    "  $('.navbar-brand').css({cursor:'pointer', 'user-select':'none'}).on('click', function(e){",
    "    e.preventDefault();",
    "    $('a[data-value=\"Home\"]').tab('show');",
    "  });",
    "});"
  )),

  tags$a(
    href = "#main-content",
    class = "skip-link",
    "Skip to main content"
  ),

  tags$div(
    role = "status",
    `aria-live` = "polite",
    `aria-atomic` = "true",
    id = "status_region",
    class = "sr-only"
  ),

  tags$head(
    tags$title("TextAnalysisR")
  ),
  tags$main(
    id = "main-content",
    role = "main",
    tabindex = "-1",
    tags$h1("TextAnalysisR", class = "sr-only"),
    navbarPage(
      "TextAnalysisR",
      id = "main_navbar",
      collapsible = TRUE,
      tabPanel(
        "Home",
        fluidRow(
          column(
            width = 3,
            wellPanel(
              style = "padding: 0; border: none; box-shadow: none; background: transparent;",
              tags$ul(
                class = "nav nav-pills nav-stacked",
                id = "home_nav_menu",
                tags$li(tags$a(href = "#about-tab", "data-toggle" = "tab", "About")),
                tags$li(tags$a(href = "#semantic-tab", "data-toggle" = "tab", "Semantic Analysis")),
                tags$li(tags$a(href = "#lexicon-tab", "data-toggle" = "tab", "Sentiment Lexicons")),
                tags$li(tags$a(href = "#cyber-tab", "data-toggle" = "tab", "Cybersecurity")),
                tags$li(tags$a(href = "#access-tab", "data-toggle" = "tab", "Web Accessibility")),
                tags$li(tags$a(href = "#resources-tab", "data-toggle" = "tab", "Resources")),
                tags$li(tags$a(href = "#support-tab", "data-toggle" = "tab", "Support"))
              )
            )
          ),
          column(
            width = 9,
            tags$div(
              class = "tab-content",
              tags$div(id = "about-tab", class = "tab-pane", div(id = "about-content", class = "markdown-content", uiOutput("about_content"))),
              tags$div(id = "semantic-tab", class = "tab-pane", div(id = "installation-semantic-content", class = "markdown-content", uiOutput("installation_semantic_content"))),
              tags$div(id = "lexicon-tab", class = "tab-pane", div(id = "installation-lexical-content", class = "markdown-content", uiOutput("installation_lexical_content"))),
              tags$div(id = "cyber-tab", class = "tab-pane", div(id = "cybersecurity-content", class = "markdown-content", uiOutput("cybersecurity_content"))),
              tags$div(id = "access-tab", class = "tab-pane", div(id = "web-accessibility-content", class = "markdown-content", uiOutput("web_accessibility_content"))),
              tags$div(id = "resources-tab", class = "tab-pane", div(id = "links-content", class = "markdown-content", uiOutput("links_content"))),
              tags$div(id = "support-tab", class = "tab-pane", div(id = "support-content", class = "markdown-content", uiOutput("support_content")))
            )
          )
        )
      ),
    tabPanel(
      "AI Setup",
      sidebarLayout(
        sidebarPanel(
          width = 3,
          class = "sidebar-panel",
          tags$h5(
            HTML("<strong>AI Configuration</strong>"),
            tags$span("OPTIONAL", style = "background-color: #6c757d; color: white; padding: 2px 8px; border-radius: 3px; font-size: 13px; margin-left: 8px;"),
            style = "color: #4269BF; margin-bottom: 10px;"
          ),
          tags$p(style = "font-size: 16px; color: #666;", "API keys entered here apply to all AI features. You can also enter keys per-feature."),
          .password_input("global_openai_api_key", "OpenAI API Key:", placeholder = "sk-..."),
          .password_input("global_gemini_api_key", "Gemini API Key:", placeholder = "AIza..."),
          tags$hr(),
          tags$h5(
            HTML("<strong>Usage Log</strong>"),
            style = "color: #4269BF; margin-bottom: 10px;"
          ),
          tags$p(style = "font-size: 16px; color: #666;", "Track which AI models were used (for reproducibility reporting)."),
          downloadButton("download_ai_log", "Download Log as CSV", class = "btn-secondary btn-block")
        ),
        mainPanel(
          width = 9,
          conditionalPanel(
            condition = "output.has_ai_usage_log",
            DT::DTOutput("ai_usage_log_table")
          ),
          conditionalPanel(
            condition = "!output.has_ai_usage_log",
            div(
              style = "padding: 30px 20px;",
              tags$h5(HTML("<strong>AI Features Overview</strong>"), style = "color: #4269BF; margin-bottom: 16px;"),
              tags$p(style = "font-size: 16px; color: #475569; margin-bottom: 20px;",
                "Configure API keys in the sidebar, then use AI features across the app. Usage is logged here for reproducibility."),
              tags$table(
                class = "table table-bordered",
                style = "font-size: 16px;",
                tags$thead(
                  tags$tr(
                    tags$th("Feature", style = "width: 30%;"),
                    tags$th("Providers"),
                    tags$th("Tab", style = "width: 22%;")
                  )
                ),
                tags$tbody(
                  tags$tr(
                    tags$td("Document Similarity"),
                    tags$td("Sentence Transformers, OpenAI, Gemini"),
                    tags$td("Semantic Analysis")
                  ),
                  tags$tr(
                    tags$td("Semantic Search"),
                    tags$td("Sentence Transformers, OpenAI, Gemini"),
                    tags$td("Semantic Analysis")
                  ),
                  tags$tr(
                    tags$td("RAG Q&A"),
                    tags$td("OpenAI, Gemini"),
                    tags$td("Semantic Analysis")
                  ),
                  tags$tr(
                    tags$td("LLM Sentiment"),
                    tags$td("OpenAI, Gemini"),
                    tags$td("Semantic Analysis")
                  ),
                  tags$tr(
                    tags$td("Topic Modeling Embeddings"),
                    tags$td("Sentence Transformers, OpenAI, Gemini"),
                    tags$td("Topic Modeling")
                  ),
                  tags$tr(
                    tags$td("Topic Labels & Content"),
                    tags$td("OpenAI, Gemini"),
                    tags$td("Topic Modeling")
                  ),
                  tags$tr(
                    tags$td("Vision OCR"),
                    tags$td("OpenAI, Gemini"),
                    tags$td("Upload")
                  )
                )
              ),
              tags$div(
                style = "margin-top: 20px; padding: 12px 16px; background-color: #F1F5F9; border-radius: 6px;",
                tags$p(style = "margin: 0 0 6px 0; font-size: 16px; color: #5C6E88;",
                  tags$strong("Sentence Transformers"), " \u2014 Free, runs locally. Requires Python + sentence-transformers."),
                tags$p(style = "margin: 0 0 6px 0; font-size: 16px; color: #5C6E88;",
                  tags$strong("OpenAI"), " \u2014 Cloud API. Enter key above or set ", tags$code("OPENAI_API_KEY"), " in .Renviron."),
                tags$p(style = "margin: 0; font-size: 16px; color: #5C6E88;",
                  tags$strong("Gemini"), " \u2014 Cloud API. Enter key above or set ", tags$code("GEMINI_API_KEY"), " in .Renviron.")
              )
            )
          )
        )
      )
    ),
    tabPanel(
      "Upload",
      sidebarLayout(
        sidebarPanel(
          width = 3,
          class = "sidebar-panel",
          selectizeInput(
            "dataset_choice",
            "Dataset",
            choices = c(
              "Select a dataset" = "",
              "Upload an Example Dataset",
              "Upload Your File",
              "Copy and Paste Text"
            ),
            selected = "",
            options = list(placeholder = "Select dataset", loadThrottle = 0)
          ),
          fileInput("file", "File upload",
            multiple = FALSE,
            accept = c(".xlsx", ".xls", ".xlsm", ".csv", ".pdf", ".docx", ".txt")
          ),
          conditionalPanel(
            condition = "input.dataset_choice == 'Upload Your File'",
            uiOutput("pdf_status_indicator")
          ),
          conditionalPanel(
            condition = "input.dataset_choice == 'Upload Your File'",
            uiOutput("multimodal_options_ui")
          ),
          conditionalPanel(
            condition = "input.dataset_choice == 'Copy and Paste Text'",
            tags$div(
              class = "text-input-white-placeholder",
              textAreaInput("text_input", "Text input", "",
                rows = 10, placeholder = "Paste your text here...

Supports:
• Plain text
• Tabular data (Excel, CSV, web tables)"
              )
            )
          ),
          tags$div(
            class = "limits-info-box status-step-purple",
            style = "margin-top: 0;",
            tags$i(class = "fa fa-info-circle limits-icon status-icon status-icon-purple"),
            tags$strong("Limits:", class = "limits-title"), " Max 100MB file upload, 50MB paste. Optimal: 1K-5K documents"
          )
        ),
        mainPanel(
          width = 9,
          DT::dataTableOutput("data_table")
        )
      )
    ),
    tabPanel(
      "Preprocess",
      sidebarLayout(
        sidebarPanel(
          width = 3,
          class = "sidebar-panel",
          conditionalPanel(
            condition = "input.conditioned == 1",
            tags$h5(
              HTML("<strong>Select columns</strong> <a href='https://tidyr.tidyverse.org/reference/unite.html' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"),
              tags$span("REQUIRED", style = "background-color: #dc3545; color: white; padding: 2px 8px; border-radius: 3px; font-size: 13px; margin-left: 8px;"),
              style = "color: #4269BF; margin-bottom: 10px;"
            ),
            div(class = "checkbox-margin", checkboxGroupInput("show_vars",
              label = NULL,
              choices = NULL
            )),
            actionButton("apply", "Apply", class = "btn-primary btn-block")
          ),
          conditionalPanel(
            condition = "input.conditioned == 2",
            tags$h5(
              HTML("<strong>Segment corpus into tokens</strong> <a href='https://quanteda.io/reference/tokens.html' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"),
              tags$span("OPTIONAL", style = "background-color: #6c757d; color: white; padding: 2px 8px; border-radius: 3px; font-size: 13px; margin-left: 8px;"),
              style = "color: #4269BF; margin-bottom: 10px;"
            ),
            tags$div(
              class = "warning-box status-sidebar-warning",
              style = "padding: 0 12px; margin-top: 0;",
              tags$div(
                style = "margin: 0; padding: 0;",
                checkboxInput("math_mode",
                  HTML("<strong>Math Mode:</strong> Keep numbers, symbols, and punctuation"),
                  value = FALSE)
              )
            ),
            div(class = "checkbox-margin", checkboxGroupInput("segment_options",
              label = NULL,
              choices = list(
                "Convert to lowercase" = "lowercase",
                "Remove punctuation" = "remove_punct",
                "Remove symbols" = "remove_symbols",
                "Remove numbers" = "remove_numbers",
                "Remove URLs" = "remove_url",
                "Remove separators" = "remove_separators",
                "Split hyphens" = "split_hyphens",
                "Split tags" = "split_tags",
                "Include document variables" = "include_docvars",
                "Keep acronyms" = "keep_acronyms",
                "Keep padding" = "padding"
              ),
              selected = c("lowercase", "remove_punct", "remove_symbols", "remove_numbers", "remove_url",
                          "remove_separators", "split_hyphens", "split_tags", "include_docvars")
            )),
            numericInput("min_char",
              label = "Minimum characters per token",
              value = 2,
              min = 1,
              max = 10,
              step = 1),
            div(
              style = "display: flex; gap: 10px; margin-bottom: 15px;",
              div(
                style = "flex: 1;",
                actionButton("preprocess", "Apply", class = "btn-primary btn-block")
              ),
              div(
                style = "flex: 1;",
                actionButton("skip_segment", "Skip", class = "btn-secondary btn-block")
              )
            )
          ),
          conditionalPanel(
            condition = "input.conditioned == 4",
            tags$h5(
              HTML("<strong>Detect multi-words</strong> <a href='https://www.tidytextmining.com/ngrams' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"),
              tags$span("OPTIONAL", style = "background-color: #6c757d; color: white; padding: 2px 8px; border-radius: 3px; font-size: 13px; margin-left: 8px;"),
              style = "color: #4269BF; margin-bottom: 10px;"
            ),
            checkboxGroupInput(
              "ngram_sizes",
              "N-gram sizes",
              choices = list(
                "Bigrams (2 words)" = "2",
                "Trigrams (3 words)" = "3",
                "4-grams (4 words)" = "4",
                "5-grams (5 words)" = "5"
              ),
              selected = c("2", "3")
            ),
            sliderInput(
              "ngram_min_count",
              "Minimum frequency",
              min = 2,
              max = 20,
              value = 3,
              step = 1
            ),
            sliderInput(
              "ngram_min_lambda",
              "Minimum lambda (collocation strength)",
              min = 0,
              max = 10,
              value = 3,
              step = 0.5
            ),
            div(
              style = "display: flex; gap: 10px; margin-bottom: 15px;",
              div(
                style = "flex: 1;",
                actionButton("detect_ngrams", "Apply", class = "btn-primary btn-block")
              ),
              div(
                style = "flex: 1;",
                actionButton("skip_ngram_detection", "Skip", class = "btn-secondary btn-block")
              )
            ),
            tags$hr(),
            tags$h5(strong("Compound selected n-grams"), style = "color: #4269BF; margin-bottom: 10px;"),
            selectizeInput(
              "multi_word_expressions",
              label = "N-grams to compound",
              choices = NULL,
              multiple = TRUE,
              options = list(
                create = TRUE,
                placeholder = "Select from detected n-grams"
              )
            ),
            div(class = "checkbox-margin", checkboxGroupInput("stopword_options",
              label = "Clean stopwords from compound edges",
              choices = list(
                "Remove leading stopwords" = "leading_stopwords",
                "Remove trailing stopwords" = "trailing_stopwords"
              ),
              selected = c("leading_stopwords", "trailing_stopwords")
            )),
            div(
              style = "display: flex; gap: 10px; margin-bottom: 15px;",
              div(
                style = "flex: 1;",
                actionButton("dictionary", "Apply", class = "btn-primary btn-block")
              ),
              div(
                style = "flex: 1;",
                actionButton("skip_dictionary", "Skip", class = "btn-secondary btn-block")
              )
            )
          ),
          conditionalPanel(
            condition = "input.conditioned == 5",
            tags$div(
              style = "display: flex; align-items: center; justify-content: space-between; margin-bottom: 10px;",
              tags$h5(
                HTML("<strong>Document-feature matrix</strong> <a href='https://quanteda.io/reference/dfm.html' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"),
                tags$span("REQUIRED", style = "background-color: #dc3545; color: white; padding: 2px 8px; border-radius: 3px; font-size: 13px; margin-left: 8px;"),
                style = "color: #4269BF; margin: 0;"
              ),
              actionLink("showDFMInfo", tags$i(class = "fas fa-info-circle"),
                        style = "color: #337ab7; font-size: 16px;",
                        title = "Click for DFM guide")
            ),
            div(
              style = "display: flex; gap: 10px; margin-bottom: 15px;",
              div(
                style = "flex: 1;",
                actionButton("dfm_btn", "Process", class = "btn-primary btn-block")
              ),
              div(
                style = "flex: 1;",
                downloadButton("download_preprocessing_report", "Report", class = "btn-secondary btn-block")
              )
            )
          ),
          conditionalPanel(
            condition = "input.conditioned == 3",
            tags$h5(
              strong("Remove stopwords"),
              tags$span("OPTIONAL", style = "background-color: #6c757d; color: white; padding: 2px 8px; border-radius: 3px; font-size: 13px; margin-left: 8px;"),
              style = "color: #4269BF; margin-bottom: 10px;"
            ),
            selectizeInput(
              "common_words",
              label = "Top 10 frequent words pre-selected",
              choices = NULL,
              multiple = TRUE,
              options = list(
                create = TRUE,
                placeholder = "Type to add more or modify"
              )
            ),
            div(
              class = "stopwords-container",
              style = "max-height: 400px; overflow-y: auto;",
              selectizeInput(
                "custom_stopwords",
                label = HTML("<strong>Predefined stopwords</strong> <a href='https://search.r-project.org/CRAN/refmans/stopwords/html/stopwords.html' target='_blank' rel='noopener noreferrer' onclick='window.open(this.href); return false;' style='font-size: 16px;'>Source</a>"),
                choices = stopwords::stopwords("en", source = "snowball"),
                selected = stopwords::stopwords("en", source = "snowball"),
                multiple = TRUE,
                options = list(create = TRUE)
              )
            ),
            br(),
            div(
              style = "display: flex; gap: 10px; margin-bottom: 15px;",
              div(
                style = "flex: 1;",
                actionButton("remove", "Apply", class = "btn-primary btn-block")
              ),
              div(
                style = "flex: 1;",
                actionButton("skip_stopwords", "Skip", class = "btn-secondary btn-block")
              )
            )
          )
        ),
        mainPanel(
          width = 9,
          tabsetPanel(
            id = "conditioned",
            tabPanel(
              "1. Unite Texts",
              value = 1,
              conditionalPanel(
                condition = "output.has_united_table_results",
                br(),
                DT::dataTableOutput("united_table")
              ),
              conditionalPanel(
                condition = "!output.has_united_table_results",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-columns", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Select columns and click ",
                      tags$strong("'Apply'", style = "color: #4269BF;"),
                      " to unite text columns",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "2. Segment Texts",
              value = 2,
              conditionalPanel(
                condition = "output.has_preprocess_results",
                shiny::verbatimTextOutput("dict_print_preprocess")
              ),
              conditionalPanel(
                condition = "!output.has_preprocess_results",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-cut", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Configure options and click ",
                      tags$strong("'Apply'", style = "color: #4269BF;"),
                      " to tokenize texts",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "3. Remove Stopwords",
              value = 3,
              conditionalPanel(
                condition = "output.step_3_outdated == true",
                div(
                  class = "status-outdated",
                  tags$i(class = "fa fa-exclamation-triangle status-icon-lg status-icon-warning"),
                  tags$span(
                    tags$strong("Outdated Results: "),
                    "Based on previous Step 2 settings. Click ",
                    tags$strong("Apply"),
                    " to update with latest Step 2 output."
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_stopword_results",
                br(),
                div(
                  style = "margin-bottom: 20px; overflow: visible;",
                  plotly::plotlyOutput("stopword_plot", height = 500, width = "100%")
                ),
                br(),
                DT::dataTableOutput("stopword_table")
              ),
              conditionalPanel(
                condition = "!output.has_stopword_results",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-filter", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Select stopwords and click ",
                      tags$strong("'Apply'", style = "color: #4269BF;"),
                      " to remove common words",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            ),
            tabPanel(
              "4. Multi-Word Dictionary",
              value = 4,
              br(),
              tabsetPanel(
                id = "multiword_subtabs",
                tabPanel(
                  "Detected N-grams",
                  conditionalPanel(
                    condition = "output.has_ngram_detection_results",
                    br(),
                    uiOutput("ngram_detection_plot_uiOutput"),
                    br(),
                    DT::dataTableOutput("ngram_detection_table")
                  ),
                  conditionalPanel(
                    condition = "!output.has_ngram_detection_results",
                    div(
                      style = "padding: 60px 40px; text-align: center;",
                      div(
                        style = "max-width: 400px; margin: 0 auto;",
                        tags$i(class = "fa fa-search", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                        tags$p(
                          "Configure settings and click ",
                          tags$strong("'Apply'", style = "color: #4269BF;"),
                          " to detect multi-words",
                          style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                        )
                      )
                    )
                  )
                ),
                tabPanel(
                  "Construct Multi-Words",
                  conditionalPanel(
                    condition = "output.step_4_outdated == true",
                    div(
                      class = "status-outdated",
                      tags$i(class = "fa fa-exclamation-triangle status-icon-lg status-icon-warning"),
                      tags$span(
                        tags$strong("Outdated Results: "),
                        "Based on previous Step 3 settings. Click ",
                        tags$strong("Apply"),
                        " to update with latest Step 3 output."
                      )
                    )
                  ),
                  conditionalPanel(
                    condition = "output.has_dictionary_results",
                    br(),
                    uiOutput("selected_ngrams_plot_uiOutput"),
                    br(),
                    DT::dataTableOutput("dictionary_table")
                  ),
                  conditionalPanel(
                    condition = "!output.has_dictionary_results",
                    div(
                      style = "padding: 60px 40px; text-align: center;",
                      div(
                        style = "max-width: 400px; margin: 0 auto;",
                        tags$i(class = "fa fa-link", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                        tags$p(
                          "Select n-grams and click ",
                          tags$strong("'Apply'", style = "color: #4269BF;"),
                          " to compound multi-words",
                          style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                        )
                      )
                    )
                  )
                )
              )
            ),
            tabPanel(
              "5. Document-Feature Matrix",
              value = 5,
              conditionalPanel(
                condition = "output.step_5_outdated == true",
                div(
                  class = "status-outdated",
                  tags$i(class = "fa fa-exclamation-triangle status-icon-lg status-icon-warning"),
                  tags$span(
                    tags$strong("Outdated Results: "),
                    "Based on previous Step 4 settings. Click ",
                    tags$strong("Process"),
                    " to update with latest Step 4 output."
                  )
                )
              ),
              conditionalPanel(
                condition = "output.has_dfm_results",
                br(),
                div(
                  style = "margin-bottom: 20px; overflow: visible;",
                  plotly::plotlyOutput("dfm_plot", height = 500, width = "100%")
                ),
                br(),
                DT::dataTableOutput("dfm_table")
              ),
              conditionalPanel(
                condition = "!output.has_dfm_results",
                div(
                  style = "padding: 60px 40px; text-align: center;",
                  div(
                    style = "max-width: 400px; margin: 0 auto;",
                    tags$i(class = "fa fa-table", style = "font-size: 48px; color: #CBD5E1; margin-bottom: 20px; display: block;", "aria-hidden" = "true"),
                    tags$p(
                      "Click ",
                      tags$strong("'Process'", style = "color: #4269BF;"),
                      " to create document-feature matrix",
                      style = "font-size: 18px; font-weight: 400; line-height: 1.7; color: #475569; margin: 0;"
                    )
                  )
                )
              )
            )
          )
        )
      )
    ),
    tabPanel(
      "Lexical Analysis",
      uiOutput("lexical_analysis_ui")
    ),
    tabPanel(
      "Semantic Analysis",
      uiOutput("semantic_analysis_ui")
    ),
    tabPanel(
      "Topic Modeling",
      uiOutput("topic_modeling_ui")
    )
  )
  )
)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(message = FALSE, warning = FALSE)

## ----eval = FALSE-------------------------------------------------------------
# install.packages("TextAnalysisR",
#   repos = c("https://mshin77.r-universe.dev", "https://cloud.r-project.org"))

## ----eval = FALSE-------------------------------------------------------------
# library(TextAnalysisR)
# run_app()

## -----------------------------------------------------------------------------
library(TextAnalysisR)

mydata <- SpecialEduTech
united_tbl <- unite_cols(mydata, listed_vars = c("title", "keyword", "abstract"))

tokens <- prep_texts(united_tbl, text_field = "united_texts")
dfm_object <- quanteda::dfm(tokens)

plot_word_frequency(dfm_object, n = 20)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(message = FALSE, warning = FALSE)

## -----------------------------------------------------------------------------
library(TextAnalysisR)

mydata <- SpecialEduTech

united_tbl <- unite_cols(mydata, listed_vars = c("title", "keyword", "abstract"))

tokens <- prep_texts(
  united_tbl,
  text_field = "united_texts",
  remove_punct = TRUE,
  remove_numbers = TRUE
)

tokens_clean <- quanteda::tokens_remove(tokens, quanteda::stopwords("en"))

dfm_object <- quanteda::dfm(tokens_clean)

## ----eval = FALSE-------------------------------------------------------------
# compounds <- detect_multi_words(tokens, min_count = 10)
# tokens <- quanteda::tokens_compound(tokens, compounds)

#' Special education technology bibliographic data
#'
#' Contains bibliographic data of journal articles and dissertations on the use of technology in teaching mathematics to students with disabilities.
#'
#' @source Bibliographic records compiled and coded by the author for
#'   Shin, M., Ok, M. W., Choo, S., Hossain, G., Bryant, D. P., & Kang, E.
#'   (2023). A content analysis of research on technology use for teaching
#'   mathematics to students with disabilities: Word networks and topic
#'   modeling. \emph{International Journal of STEM Education}, 10, Article 23.
#'   \doi{10.1186/s40594-023-00414-x}. Shared as open data at
#'   \url{https://github.com/mshin77/math-tech-sped}.
#'
#' @examples
#' SpecialEduTech
#'
"SpecialEduTech"

#' An example structure of a structural topic model
#'
#' Contains 15 topics, topic prevalences, and etc. from \code{stm::stm}.
#'
#' @examples
#' \donttest{
#' stm_15
#' }
"stm_15"

#' Acronym List
#'
#' A dataset containing common acronyms used in text processing
#'
#' @format A character vector of acronyms
#' @name acronym
#' @docType data
#' @keywords datasets
NULL

#' Stopwords List
#'
#' A dataset containing stopwords for text preprocessing
#'
#' @format A character vector of stopwords
#' @name stopwords_list
#' @docType data
#' @keywords datasets
NULL

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(message = FALSE, warning = FALSE)

## -----------------------------------------------------------------------------
library(TextAnalysisR)

mydata <- SpecialEduTech[1:150, ]
united_tbl <- unite_cols(mydata, listed_vars = c("title", "keyword", "abstract"))
tokens <- prep_texts(united_tbl, text_field = "united_texts", remove_stopwords = TRUE)
dfm_object <- quanteda::dfm(tokens)

## -----------------------------------------------------------------------------
k_search <- find_optimal_k(dfm_object, topic_range = 5:6)
names(k_search)

## -----------------------------------------------------------------------------
out <- quanteda::convert(dfm_object, to = "stm")
out$meta$year <- as.numeric(out$meta$year)

model <- stm::stm(
  documents = out$documents,
  vocab = out$vocab,
  K = 5,
  prevalence = ~ reference_type + year,
  data = out$meta,
  max.em.its = 15,
  init.type = "Spectral",
  verbose = FALSE
)

## -----------------------------------------------------------------------------
terms <- get_topic_terms(model, top_term_n = 10)
head(terms, 15)

## ----eval = FALSE-------------------------------------------------------------
# labels <- generate_topic_labels(terms, provider = "openai")
# 
# content <- generate_topic_content(terms, content_type = "research_question", provider = "openai")

## -----------------------------------------------------------------------------
doc_topic <- calculate_topic_probability(model, top_n = 10, verbose = FALSE)
doc_topic

## -----------------------------------------------------------------------------
quotes <- stm::findThoughts(model, texts = united_tbl$united_texts, topics = 1, n = 3)
quotes

## -----------------------------------------------------------------------------
prep <- stm::estimateEffect(1:5 ~ reference_type + year, model, metadata = out$meta, uncertainty = "Global")
head(tidytext::tidy(prep), 10)

## -----------------------------------------------------------------------------
ec <- stm::plot.estimateEffect(prep, "reference_type", method = "pointestimate",
                               model = model, omit.plot = TRUE)
effects_cat <- do.call(rbind, lapply(seq_along(ec$topics), function(i) {
  data.frame(topic = ec$topics[i], value = as.character(ec$uvals),
             proportion = as.numeric(ec$means[[i]]),
             lower = as.numeric(ec$cis[[i]][1, ]), upper = as.numeric(ec$cis[[i]][2, ]))
}))
plot_topic_effects_categorical(effects_cat)

## -----------------------------------------------------------------------------
en <- stm::plot.estimateEffect(prep, "year", method = "continuous",
                               model = model, omit.plot = TRUE)
effects_cont <- do.call(rbind, lapply(seq_along(en$topics), function(i) {
  data.frame(topic = en$topics[i], value = en$x,
             proportion = as.numeric(en$means[[i]]),
             lower = as.numeric(en$ci[[i]][1, ]), upper = as.numeric(en$ci[[i]][2, ]))
}))
plot_topic_effects_continuous(effects_cont)

## ----eval = FALSE-------------------------------------------------------------
# embeddings <- get_best_embeddings(united_tbl$united_texts, provider = "openai")
# 
# topics <- fit_embedding_model(
#   united_tbl$united_texts,
#   method = "umap_hdbscan",
#   n_topics = 10,
#   precomputed_embeddings = embeddings
# )
# 
# embedding_labels <- generate_topic_labels(topics$topic_terms, provider = "openai")

#' @importFrom utils modifyList
#' @importFrom stats cor hclust dist cutree
NULL

# Semantic Analysis Functions
# Functions for semantic analysis, embeddings, and document clustering

.cosine_sim <- function(a, b) {
  sum(a * b) / (sqrt(sum(a^2)) * sqrt(sum(b^2)))
}

.resolve_llm_setup <- function(provider, model, api_key, defaults, strict_validate = FALSE) {
  if (is.null(model)) model <- defaults[[provider]]
  if (is.null(api_key)) {
    env_var <- switch(provider, openai = "OPENAI_API_KEY", gemini = "GEMINI_API_KEY")
    api_key <- Sys.getenv(env_var)
  }
  if (!nzchar(api_key)) {
    message(.missing_api_key_message(provider, "package"))
    return(NULL)
  }
  if (strict_validate) {
    validation <- validate_api_key(api_key, strict = FALSE)
    if (!validation$valid) {
      message(sprintf("Invalid API key format: %s", validation$error))
      return(NULL)
    }
  }
  list(provider = provider, model = model, api_key = api_key)
}

.empty_sentiment_row <- function(doc_name) {
  n <- length(doc_name)
  data.frame(
    document = doc_name,
    sentiment = rep(NA_character_, n),
    sentiment_score = rep(NA_real_, n),
    confidence = rep(NA_real_, n),
    explanation = rep(NA_character_, n),
    stringsAsFactors = FALSE
  )
}

#' @title Calculate Document Similarity
#'
#' @description
#' Calculates similarity between documents using traditional NLP methods or
#' modern embedding-based approaches. Metrics are automatically
#' computed unless disabled.
#'
#' @param texts A character vector of texts to compare.
#' @param document_feature_type Feature extraction type: "words", "ngrams", or "embeddings".
#' @param semantic_ngram_range Integer, n-gram range for ngram features (default: 2).
#' @param similarity_method Similarity calculation method: "cosine", "jaccard", "euclidean", "manhattan".
#' @param use_embeddings Logical, use embedding-based similarity (default: FALSE).
#' @param embedding_model Sentence transformer model name (default: "all-MiniLM-L6-v2").
#' @param calculate_metrics Logical, compute similarity metrics (default: TRUE).
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A list containing:
#'   \describe{
#'     \item{similarity_matrix}{N x N similarity matrix}
#'     \item{feature_matrix}{Document feature matrix used for calculation}
#'     \item{method_info}{Information about the method used}
#'     \item{metrics}{Similarity metrics (if calculate_metrics = TRUE)}
#'     \item{execution_time}{Time taken for analysis}
#'   }
#'
#' @concept semantic
#' @keywords internal
#' @export
#'
#' @examples
#' \donttest{
#'   data(SpecialEduTech)
#'   texts <- SpecialEduTech$abstract[1:5]
#'
#'   result <- calculate_document_similarity(
#'     texts = texts,
#'     document_feature_type = "words",
#'     similarity_method = "cosine"
#'   )
#'
#'   print(result$similarity_matrix)
#'   print(result$metrics)
#' }
calculate_document_similarity <- function(texts,
                            document_feature_type = "words",
                            semantic_ngram_range = 2,
                            similarity_method = "cosine",
                            use_embeddings = FALSE,
                            embedding_model = "all-MiniLM-L6-v2",
                            calculate_metrics = TRUE,
                            verbose = TRUE) {

  if (verbose) {
    message("Starting document similarity analysis...")
    message("Feature type: ", document_feature_type)
    message("Similarity method: ", similarity_method)
    message("Use embeddings: ", use_embeddings)
  }

  if (is.null(texts) || length(texts) == 0) {
    stop("No texts provided for analysis")
  }

  valid_texts <- texts[nchar(trimws(texts)) > 0]
  if (length(valid_texts) < 2) {
    stop("Need at least 2 non-empty texts for analysis")
  }

  start_time <- Sys.time()

  tryCatch({
    if (verbose) message("Step 1: Generating feature matrix...")

    feature_matrix <- switch(document_feature_type,
      "words" = {
        if (verbose) message("Using word-based features...")
        corpus <- quanteda::corpus(valid_texts)
        tokens <- quanteda::tokens(corpus,
                                   remove_punct = TRUE,
                                   remove_numbers = TRUE,
                                   remove_symbols = TRUE,
                                   remove_separators = TRUE)
        tokens <- quanteda::tokens_tolower(tokens)
        tokens <- quanteda::tokens_remove(tokens, quanteda::stopwords("english"))
        dfm <- quanteda::dfm(tokens)
        dfm <- quanteda::dfm_trim(dfm, min_termfreq = 2, min_docfreq = 1)
        as.matrix(dfm)
      },
      "ngrams" = {
        if (verbose) message("Using n-gram features (n=", semantic_ngram_range, ")...")
        corpus <- quanteda::corpus(valid_texts)
        tokens <- quanteda::tokens(corpus,
                                   remove_punct = TRUE,
                                   remove_numbers = TRUE,
                                   remove_symbols = TRUE,
                                   remove_separators = TRUE)
        tokens <- quanteda::tokens_tolower(tokens)
        tokens <- quanteda::tokens_remove(tokens, quanteda::stopwords("english"))
        tokens_ngrams <- quanteda::tokens_ngrams(tokens, n = semantic_ngram_range,
                                                  concatenator = "_")
        dfm_ngrams <- quanteda::dfm(tokens_ngrams)
        as.matrix(dfm_ngrams)
      },
      "embeddings" = {
        if (verbose) message("Using embedding features...")
        if (!requireNamespace("reticulate", quietly = TRUE)) {
          return(.notify_missing_python("Embedding-based document similarity"))
        }

        python_available <- tryCatch({
          reticulate::py_config()
          TRUE
        }, error = function(e) FALSE)

        if (!python_available) {
          return(.notify_missing_python("Embedding-based document similarity (sentence-transformers)"))
        }
        sentence_transformers <- reticulate::import("sentence_transformers")
        model <- sentence_transformers$SentenceTransformer(embedding_model)

        n_docs <- length(valid_texts)
        batch_size <- if (n_docs > 100) 25 else if (n_docs > 50) 50 else n_docs

        if (verbose) message("Processing ", n_docs, " documents with embeddings...")

        embeddings_list <- list()
        for (i in seq(1, n_docs, by = batch_size)) {
          end_idx <- min(i + batch_size - 1, n_docs)
          batch_texts <- valid_texts[i:end_idx]
          batch_embeddings <- model$encode(batch_texts, show_progress_bar = FALSE, normalize_embeddings = TRUE)
          embeddings_list[[length(embeddings_list) + 1]] <- batch_embeddings
        }

        do.call(rbind, embeddings_list)
      },
      "topics" = {
        if (verbose) message("Using topic features...")
        stop("Topic features require a pre-fitted STM model. ",
             "Please use words, ngrams, or embeddings.")
      },
      stop("Unsupported feature type: ", document_feature_type)
    )

    if (verbose) message("Step 2: Calculating similarity matrix...")

    if (use_embeddings && document_feature_type == "embeddings") {
      if (!requireNamespace("reticulate", quietly = TRUE)) {
        return(.notify_missing_python("Embedding-based similarity"))
      }

      sklearn_metrics <- reticulate::import("sklearn.metrics.pairwise")
      similarity_matrix <- sklearn_metrics$cosine_similarity(feature_matrix)
      similarity_matrix <- as.matrix(similarity_matrix)

      method_info <- list(
        method = "embedding_cosine",
        model_name = embedding_model,
        n_docs = length(valid_texts)
      )
    } else {
      if (ncol(feature_matrix) == 0) {
        if (verbose) message("No features remaining. Using basic character similarity...")
        similarity_matrix <- outer(valid_texts, valid_texts, function(x, y) {
          mapply(function(a, b) {
            if (nchar(a) == 0 || nchar(b) == 0) return(0)
            common_chars <- sum(utf8ToInt(a) %in% utf8ToInt(b))
            max(common_chars / max(nchar(a), nchar(b)), 0)
          }, x, y)
        })
        method_info <- list(method = "character_similarity", n_docs = length(valid_texts))
      } else {
        if (similarity_method == "cosine") {
          similarity_matrix <- calculate_cosine_similarity(feature_matrix)
        } else {
          stop("Non-cosine similarity methods require dfm input. Please use cosine similarity or provide a dfm object.")
        }
        method_info <- list(
          method = paste0("traditional_", similarity_method),
          n_docs = length(valid_texts)
        )
      }
    }

    diag(similarity_matrix) <- 1
    similarity_matrix[similarity_matrix > 1] <- 1
    similarity_matrix[similarity_matrix < -1] <- -1

    metrics <- NULL
    if (calculate_metrics) {
      if (verbose) message("Step 3: Calculating metrics...")
      metrics <- calculate_metrics(similarity_matrix, method_info = method_info)
    }

    execution_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

    if (verbose) {
      message("Document similarity analysis completed in ", round(execution_time, 2), " seconds")
      message("Documents analyzed: ", length(valid_texts))
      message("Feature dimensions: ", ncol(feature_matrix))
    }

    result <- list(
      similarity_matrix = similarity_matrix,
      feature_matrix = feature_matrix,
      method_info = method_info,
      metrics = metrics,
      execution_time = execution_time,
      n_documents = length(valid_texts),
      feature_type = document_feature_type,
      similarity_method = similarity_method,
      use_embeddings = use_embeddings
    )

    return(result)

  }, error = function(e) {
    stop("Error in document similarity analysis: ", e$message)
  })
}

#' @title Fit Semantic Model
#'
#' @description
#' Performs semantic analysis including similarity, dimensionality reduction,
#' and clustering. This is a high-level wrapper function.
#'
#' @param texts A character vector of texts to analyze.
#' @param analysis_types Types of analysis to perform: "similarity", "dimensionality_reduction", "clustering".
#' @param document_feature_type Feature extraction type (default: "embeddings").
#' @param similarity_method Similarity calculation method (default: "cosine").
#' @param use_embeddings Logical, use embedding-based approaches (default: TRUE).
#' @param embedding_model Sentence transformer model name (default: "all-MiniLM-L6-v2").
#' @param dimred_method Dimensionality reduction method: "PCA", "t-SNE", "UMAP" (default: "UMAP").
#' @param clustering_method Clustering method: "kmeans", "hierarchical", "umap_dbscan" (default: "umap_dbscan").
#' @param n_components Number of dimensions for reduction (default: 2).
#' @param n_clusters Number of clusters (default: 5).
#' @param seed Random seed for reproducibility (default: 123).
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A list containing results from requested analyses.
#'
#' @concept semantic
#' @export
#'
#' @examples
#' if (interactive()) {
#'   data(SpecialEduTech)
#'   texts <- SpecialEduTech$abstract[1:10]
#'
#'   results <- fit_semantic_model(
#'     texts = texts,
#'     analysis_types = c("similarity", "clustering")
#'   )
#'
#'   print(results)
#' }
fit_semantic_model <- function(texts,
                             analysis_types = c("similarity", "dimensionality_reduction",
                                                "clustering"),
                             document_feature_type = "embeddings",
                             similarity_method = "cosine",
                             use_embeddings = TRUE,
                             embedding_model = "all-MiniLM-L6-v2",
                             dimred_method = "UMAP",
                             clustering_method = "umap_dbscan",
                             n_components = 2,
                             n_clusters = 5,
                             seed = 123,
                             verbose = TRUE) {

  if (verbose) {
    message("Starting semantic analysis...")
    message("Analysis types: ", paste(analysis_types, collapse = ", "))
  }

  results <- list()
  start_time <- Sys.time()

  if ("similarity" %in% analysis_types || length(analysis_types) > 1) {
    if (verbose) message("Step 1: Document similarity analysis...")

    results$similarity <- calculate_document_similarity(
      texts = texts,
      document_feature_type = document_feature_type,
      similarity_method = similarity_method,
      use_embeddings = use_embeddings,
      embedding_model = embedding_model,
      verbose = verbose
    )
  }

  if ("dimensionality_reduction" %in% analysis_types) {
    if (verbose) message("Step 2: Dimensionality reduction analysis...")

    data_matrix <- if (!is.null(results$similarity)) {
      results$similarity$feature_matrix
    } else {
      similarity_result <- calculate_document_similarity(
        texts = texts,
        document_feature_type = document_feature_type,
        use_embeddings = use_embeddings,
        embedding_model = embedding_model,
        verbose = FALSE
      )
      similarity_result$feature_matrix
    }

    results$dimensionality_reduction <- reduce_dimensions(
      data_matrix = data_matrix,
      method = dimred_method,
      n_components = n_components,
      seed = seed,
      verbose = verbose
    )
  }

  if ("clustering" %in% analysis_types) {
    if (verbose) message("Step 3: Clustering analysis...")

    data_matrix <- if (!is.null(results$similarity)) {
      results$similarity$feature_matrix
    } else {
      similarity_result <- calculate_document_similarity(
        texts = texts,
        document_feature_type = document_feature_type,
        use_embeddings = use_embeddings,
        embedding_model = embedding_model,
        verbose = FALSE
      )
      similarity_result$feature_matrix
    }

    if (clustering_method != "none") {
      results$clustering <- cluster_embeddings(
        data_matrix,
        method = clustering_method,
        n_clusters = n_clusters,
        seed = seed,
        verbose = verbose
      )
    }
  }

  execution_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  if (verbose) {
    message("Semantic analysis completed in ", round(execution_time, 2), " seconds")
  }

  results$execution_time <- execution_time
  results$analysis_types <- analysis_types
  results$timestamp <- Sys.time()

  return(results)
}

#' @title Dimensionality Reduction Analysis
#'
#' @description
#' This function performs dimensionality reduction using various methods
#' including PCA, t-SNE, and UMAP.
#' UMAP runs directly on inputs with 1000 or fewer features (e.g. dense
#' embeddings); PCA pre-reduction applies to wider inputs and to the PCA and
#' t-SNE methods.
#'
#' @param data_matrix A numeric matrix where rows represent documents and
#'   columns represent features.
#' @param method The dimensionality reduction method. Options: "PCA", "t-SNE", "UMAP".
#' @param n_components The number of components/dimensions to reduce to (default: 2).
#' @param pca_dims The number of dimensions for PCA preprocessing (default: 50).
#' @param tsne_perplexity The perplexity parameter for t-SNE (default: 30).
#' @param tsne_max_iter The maximum number of iterations for t-SNE (default: 1000).
#' @param umap_neighbors The number of neighbors for UMAP (default: 15).
#' @param umap_min_dist The minimum distance for UMAP (default: 0.1).
#' @param umap_metric The metric for UMAP (default: "cosine").
#' @param seed Random seed for reproducibility (default: 123).
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A list containing the reduced dimensions, method used, and additional metadata.
#'
#' @concept semantic
#' @export
#'
#' @examples
#' if (interactive()) {
#'   mydata <- TextAnalysisR::SpecialEduTech
#'
#'   united_tbl <- TextAnalysisR::unite_cols(
#'     mydata,
#'     listed_vars = c("title", "keyword", "abstract")
#'   )
#'
#'   tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")
#'
#'   dfm_object <- quanteda::dfm(tokens)
#'
#'   data_matrix <- as.matrix(dfm_object)
#'
#'   pca_result <- TextAnalysisR::reduce_dimensions(
#'     data_matrix,
#'     method = "PCA"
#'   )
#'   print(pca_result)
#'
#'   tsne_result <- TextAnalysisR::reduce_dimensions(
#'     data_matrix,
#'     method = "t-SNE"
#'   )
#'   print(tsne_result)
#'
#'   umap_result <- TextAnalysisR::reduce_dimensions(
#'     data_matrix,
#'     method = "UMAP"
#'   )
#'   print(umap_result)
#' }
reduce_dimensions <- function(data_matrix,
                                            method = "PCA",
                                            n_components = 2,
                                            pca_dims = 50,
                                            tsne_perplexity = 30,
                                            tsne_max_iter = 1000,
                                            umap_neighbors = 15,
                                            umap_min_dist = 0.1,
                                            umap_metric = "cosine",
                                            seed = 123,
                                            verbose = TRUE) {

  if (verbose) {
    message("Starting dimensionality reduction with method: ", method)
  }

  start_time <- Sys.time()

  withr::with_seed(seed, tryCatch({
    col_vars <- apply(data_matrix, 2, var, na.rm = TRUE)
    constant_cols <- which(col_vars == 0 | is.na(col_vars))

    if (length(constant_cols) > 0) {
      if (verbose) message(paste("Removing", length(constant_cols),
                                 "constant/zero columns"))
      data_matrix <- data_matrix[, -constant_cols, drop = FALSE]
    }

    if (ncol(data_matrix) < 2) {
      stop("Insufficient non-constant features for analysis")
    }

    max_components <- min(nrow(data_matrix) - 1, ncol(data_matrix), pca_dims)
    pca_dims_actual <- min(pca_dims, max_components)

    # z-scored PCA distorts cosine geometry; UMAP takes dense embeddings
    # directly, PCA pre-reduction only speeds up wide sparse inputs
    needs_pca <- method %in% c("PCA", "t-SNE") || ncol(data_matrix) > 1000
    pca_result <- NULL
    if (needs_pca) {
      if (verbose) message("Performing PCA preprocessing...")
      pca_result <- stats::prcomp(data_matrix, center = TRUE, scale. = FALSE,
                                  rank. = pca_dims_actual)
    }


    result <- switch(method,
      "PCA" = {
        if (verbose) message("Returning PCA results...")

        final_components <- min(n_components, ncol(pca_result$x))
        reduced_data <- pca_result$x[, 1:final_components, drop = FALSE]

        list(
          reduced_data = reduced_data,
          method = "PCA",
          pca_result = pca_result,
          variance_explained = summary(pca_result)$importance[2, 1:final_components]
        )
      },
      "t-SNE" = {
        if (verbose) message("Performing t-SNE on PCA results...")

        if (!requireNamespace("Rtsne", quietly = TRUE)) {
          stop("Rtsne package is required for t-SNE analysis. ",
               "Please install it with: install.packages('Rtsne')")
        }

        data_for_tsne <- pca_result$x
        adjusted_perplexity <- min(tsne_perplexity,
                                   floor((nrow(data_matrix) - 1) / 3))

        tsne_result <- Rtsne::Rtsne(data_for_tsne,
                                   dims = n_components,
                                   perplexity = adjusted_perplexity,
                                   max_iter = tsne_max_iter,
                                   check_duplicates = FALSE,
                                   pca = FALSE,
                                   verbose = verbose)

        list(
          reduced_data = tsne_result$Y,
          method = "t-SNE",
          tsne_result = tsne_result,
          pca_result = pca_result,
          perplexity = adjusted_perplexity
        )
      },
      "UMAP" = {
        umap_input <- if (is.null(pca_result)) data_matrix else pca_result$x
        if (verbose) message("Performing UMAP on ",
                             if (is.null(pca_result)) "original features..." else "PCA results...")

        if (requireNamespace("umap", quietly = TRUE)) {
          safe_n_neighbors <- min(umap_neighbors, nrow(data_matrix) - 1, 15)

          umap_config <- umap::umap.defaults
          umap_config$n_neighbors <- safe_n_neighbors
          umap_config$min_dist <- umap_min_dist
          umap_config$n_components <- n_components
          umap_config$metric <- umap_metric
          umap_config$random_state <- seed

          umap_result <- umap::umap(umap_input, config = umap_config)

          list(
            reduced_data = umap_result$layout,
            method = "UMAP",
            umap_result = umap_result,
            pca_result = pca_result,
            umap_params = list(
              n_neighbors = safe_n_neighbors,
              min_dist = umap_min_dist,
              metric = umap_metric
            )
          )
        } else if (requireNamespace("reticulate", quietly = TRUE)) {
          if (!reticulate::py_module_available("umap")) {
            stop("umap-learn Python package is required. ",
                 "Please install it with: pip install umap-learn")
          }

          umap_module <- reticulate::import("umap")
          reducer <- umap_module$UMAP(
            n_neighbors = umap_neighbors,
            min_dist = umap_min_dist,
            n_components = as.integer(n_components),
            metric = umap_metric,
            random_state = as.integer(seed)
          )

          embedding <- reducer$fit_transform(umap_input)

          list(
            reduced_data = embedding,
            method = "UMAP",
            pca_result = pca_result,
            umap_params = list(
              n_neighbors = umap_neighbors,
              min_dist = umap_min_dist,
              metric = umap_metric
            )
          )
        } else {
          stop("Either 'umap' R package or 'reticulate' with Python 'umap-learn' ",
               "is required for UMAP analysis")
        }
      },
      stop("Unsupported dimensionality reduction method: ", method)
    )

    execution_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

    if (verbose) {
      message("Dimensionality reduction completed in ", round(execution_time, 2), " seconds")
    }

    result$execution_time <- execution_time
    result$timestamp <- Sys.time()
    result$seed <- seed

    return(result)

  }, error = function(e) {
    stop("Error in dimensionality reduction analysis: ", e$message)
  }))
}

#' @title Embedding-based Document Clustering
#'
#' @description
#' This function performs clustering analysis using various methods, ordered
#' from simple to detailed:
#' k-means (simplest), hierarchical (intermediate), and UMAP+DBSCAN
#' (most detailed).
#'
#' @param data_matrix A numeric matrix where rows represent documents and
#'   columns represent features.
#' @param method The clustering method. Options: "kmeans", "hierarchical",
#'   "umap_dbscan".
#' @param n_clusters The number of clusters (for k-means and hierarchical).
#'   If 0, optimal number is determined automatically.
#' @param umap_neighbors The number of neighbors for UMAP (default: 15).
#' @param umap_min_dist The minimum distance for UMAP (default: 0.1).
#' @param umap_n_components The number of UMAP components (default: 10).
#' @param umap_metric The metric for UMAP (default: "cosine").
#' @param dbscan_eps The eps parameter for DBSCAN. If 0, optimal value is determined automatically.
#' @param dbscan_min_samples The minimum samples for DBSCAN. Default NULL sets
#'   \code{umap_n_components + 1} (dimensions-plus-one rule of thumb).
#' @param reduce_outliers Logical, if TRUE, reassigns noise points (cluster 0) to nearest cluster (default: TRUE).
#'   Set FALSE to keep density-based noise labels.
#' @param outlier_strategy Strategy for outlier reduction: "centroid" (default,
#'   Euclidean nearest-centroid in UMAP space; package-specific heuristic) or
#'   "embeddings" (cosine similarity to cluster mean in original embedding
#'   space, matching the BERTopic "embeddings" strategy).
#' @param outlier_threshold Minimum similarity for reassigning a noise point
#'   (default: 0, reassign all). For "embeddings" this is cosine similarity;
#'   for "centroid" it applies to 1 / (1 + euclidean distance). Points below
#'   the threshold keep cluster label 0.
#' @param seed Random seed for reproducibility (default: 123).
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A list containing cluster assignments, method used, and quality metrics.
#'   For \code{method = "umap_dbscan"} with \code{reduce_outliers = TRUE}, the list
#'   also reports how noise points were reassigned: \code{outlier_strategy}
#'   ("embeddings" or "centroid"), \code{outlier_metric} ("cosine" or "euclidean"),
#'   \code{outlier_space} ("original embedding space" or "UMAP space"), and
#'   \code{outliers_reassigned} (count).
#'
#' @concept semantic
#' @export
#'
#' @examples
#' if (interactive()) {
#'   mydata <- TextAnalysisR::SpecialEduTech
#'
#'   united_tbl <- TextAnalysisR::unite_cols(
#'     mydata,
#'     listed_vars = c("title", "keyword", "abstract")
#'   )
#'
#'   tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")
#'
#'   dfm_object <- quanteda::dfm(tokens)
#'
#'   data_matrix <- as.matrix(dfm_object)
#'
#'   kmeans_result <- TextAnalysisR::cluster_embeddings(
#'     data_matrix,
#'     method = "kmeans",
#'     n_clusters = 5
#'   )
#'   print(kmeans_result)
#'
#'   hierarchical_result <- TextAnalysisR::cluster_embeddings(
#'     data_matrix,
#'     method = "hierarchical",
#'     n_clusters = 5
#'   )
#'   print(hierarchical_result)
#'
#'   umap_dbscan_result <- TextAnalysisR::cluster_embeddings(
#'     data_matrix,
#'     method = "umap_dbscan",
#'     umap_neighbors = 15,
#'     umap_min_dist = 0.1
#'   )
#'   print(umap_dbscan_result)
#' }
cluster_embeddings <- function(data_matrix,
                                       method = "kmeans",
                                       n_clusters = 0,
                                       umap_neighbors = 15,
                                       umap_min_dist = 0.1,
                                       umap_n_components = 10,
                                       umap_metric = "cosine",
                                       dbscan_eps = 0,
                                       dbscan_min_samples = NULL,
                                       reduce_outliers = TRUE,
                                       outlier_strategy = "centroid",
                                       outlier_threshold = 0,
                                       seed = 123,
                                       verbose = TRUE) {

  # minPts rule of thumb: dimensions + 1
  if (is.null(dbscan_min_samples)) {
    dbscan_min_samples <- umap_n_components + 1
  }

  if (verbose) {
    message("Starting clustering analysis with method: ", method)
  }

  # scale() yields NaN on zero-variance columns, crashing kmeans/hclust
  col_vars <- apply(data_matrix, 2, stats::var)
  constant_cols <- is.na(col_vars) | col_vars == 0
  if (any(constant_cols)) {
    if (all(constant_cols)) stop("data_matrix has no columns with non-zero variance")
    data_matrix <- data_matrix[, !constant_cols, drop = FALSE]
    if (verbose) message("Removed ", sum(constant_cols), " zero-variance columns")
  }

  start_time <- Sys.time()

  withr::with_seed(seed, tryCatch({
    result <- switch(method,
      "umap_dbscan" = {
        if (verbose) message("Performing UMAP + DBSCAN clustering...")

        umap_result <- reduce_dimensions(
          data_matrix,
          method = "UMAP",
          n_components = umap_n_components,
          umap_neighbors = umap_neighbors,
          umap_min_dist = umap_min_dist,
          umap_metric = umap_metric,
          seed = seed,
          verbose = verbose
        )

        if (!requireNamespace("dbscan", quietly = TRUE)) {
          stop("dbscan package is required for DBSCAN clustering. ",
               "Please install it with: install.packages('dbscan')")
        }

        if (dbscan_eps == 0) {
          if (verbose) message("Determining optimal eps parameter...")
          # kNNdist convention: k = minPts - 1 (a point is its own neighbor)
          knn_dist <- dbscan::kNNdist(umap_result$reduced_data, k = max(1, dbscan_min_samples - 1))
          dbscan_eps <- stats::quantile(knn_dist, 0.9)
          auto_eps <- TRUE
        } else {
          auto_eps <- FALSE
        }

        dbscan_result <- dbscan::dbscan(umap_result$reduced_data,
                                       eps = dbscan_eps,
                                       minPts = dbscan_min_samples)

        clusters <- dbscan_result$cluster
        n_clusters_found <- length(unique(clusters)) - (0 %in% clusters)
        noise_ratio <- sum(clusters == 0) / length(clusters)

        outliers_reassigned <- 0
        outlier_strategy_used <- outlier_strategy
        outlier_metric <- if (outlier_strategy == "embeddings") "cosine" else "euclidean"
        outlier_space  <- if (outlier_strategy == "embeddings") "original embedding" else "UMAP"
        if (reduce_outliers && any(clusters == 0) && n_clusters_found > 0) {
          if (verbose) message("Reassigning up to ", sum(clusters == 0), " noise points to nearest cluster (",
                               outlier_metric, " in ", outlier_space, " space, threshold ",
                               outlier_threshold, ")...")

          noise_idx <- which(clusters == 0)
          valid_clusters <- unique(clusters[clusters > 0])

          # points below outlier_threshold keep noise label 0
          if (outlier_strategy == "embeddings") {
            embedding_centroids <- sapply(valid_clusters, function(cl) {
              colMeans(data_matrix[clusters == cl, , drop = FALSE])
            })
            if (is.vector(embedding_centroids)) embedding_centroids <- matrix(embedding_centroids, nrow = 1)
            embedding_centroids <- t(embedding_centroids)

            for (idx in noise_idx) {
              point <- data_matrix[idx, ]
              similarities <- apply(embedding_centroids, 1, function(c) .cosine_sim(point, c))
              if (max(similarities) >= outlier_threshold) {
                clusters[idx] <- valid_clusters[which.max(similarities)]
                outliers_reassigned <- outliers_reassigned + 1
              }
            }
          } else {
            centroids <- sapply(valid_clusters, function(cl) {
              colMeans(umap_result$reduced_data[clusters == cl, , drop = FALSE])
            })
            if (is.vector(centroids)) centroids <- matrix(centroids, nrow = 1)
            centroids <- t(centroids)

            for (idx in noise_idx) {
              point <- umap_result$reduced_data[idx, ]
              distances <- apply(centroids, 1, function(c) sqrt(sum((point - c)^2)))
              # bounded similarity transform of euclidean distance
              if (1 / (1 + min(distances)) >= outlier_threshold) {
                clusters[idx] <- valid_clusters[which.min(distances)]
                outliers_reassigned <- outliers_reassigned + 1
              }
            }
          }
        }

        list(
          clusters = clusters,
          method = "umap_dbscan",
          n_clusters = n_clusters_found,
          umap_embedding = umap_result$reduced_data,
          dbscan_result = dbscan_result,
          auto_detected = auto_eps,
          detection_method = if (auto_eps) "90th percentile of kNN distances" else "Manual",
          noise_ratio = noise_ratio,
          reduce_outliers = reduce_outliers,
          outlier_strategy = if (reduce_outliers) outlier_strategy_used else NA_character_,
          outlier_metric = if (reduce_outliers) outlier_metric else NA_character_,
          outlier_space = if (reduce_outliers) paste0(outlier_space, " space") else NA_character_,
          outliers_reassigned = outliers_reassigned,
          parameters = list(
            eps = dbscan_eps,
            min_samples = dbscan_min_samples,
            umap_neighbors = umap_neighbors,
            umap_min_dist = umap_min_dist,
            umap_metric = umap_metric
          )
        )
      },
      "kmeans" = {
        if (verbose) message("Performing k-means clustering...")

        if (n_clusters == 0) {
          if (verbose) message("Determining optimal number of clusters...")
          max_k <- min(10, nrow(data_matrix) - 1)
          wss <- sapply(1:max_k, function(k) {
            kmeans(scale(data_matrix), k, nstart = 10, iter.max = 100)$tot.withinss
          })

          n_clusters <- which.max(diff(diff(wss))) + 1
          if (n_clusters < 2) n_clusters <- 2
          auto_detected <- TRUE
        } else {
          auto_detected <- FALSE
        }

        kmeans_result <- stats::kmeans(scale(data_matrix),
                                     centers = n_clusters,
                                     nstart = 25,
                                     iter.max = 100)

        list(
          clusters = kmeans_result$cluster,
          method = "kmeans",
          n_clusters = n_clusters,
          kmeans_result = kmeans_result,
          auto_detected = auto_detected,
          detection_method = if (auto_detected) "Elbow Method" else "Manual"
        )
      },
      "hierarchical" = {
        if (verbose) message("Performing hierarchical clustering...")

        dist_matrix <- stats::dist(scale(data_matrix))
        hclust_result <- stats::hclust(dist_matrix, method = "ward.D2")

        if (n_clusters == 0) {
          if (verbose) message("Determining optimal number of clusters...")
          sil_scores <- sapply(2:min(10, nrow(data_matrix) - 1), function(k) {
            clusters <- stats::cutree(hclust_result, k = k)
            if (!requireNamespace("cluster", quietly = TRUE)) {
              stop("cluster package is required for silhouette analysis. ",
                   "Please install it with: install.packages('cluster')")
            }
            mean(cluster::silhouette(clusters, dist_matrix)[, 3])
          })
          n_clusters <- which.max(sil_scores) + 1
          auto_detected <- TRUE
        } else {
          auto_detected <- FALSE
        }

        clusters <- stats::cutree(hclust_result, k = n_clusters)

        list(
          clusters = clusters,
          method = "hierarchical",
          n_clusters = n_clusters,
          hclust_result = hclust_result,
          auto_detected = auto_detected,
          detection_method = if (auto_detected) "Silhouette Method" else "Manual"
        )
      },
      stop("Unsupported clustering method: ", method)
    )

    if (result$n_clusters > 1 && !all(result$clusters == 0)) {
      if (verbose) message("Calculating quality metrics...")

      # umap_dbscan clusters form in UMAP space, so score metrics there
      eval_matrix <- if (method == "umap_dbscan") result$umap_embedding else scale(data_matrix)

      if (requireNamespace("cluster", quietly = TRUE)) {
        tryCatch({
          sil <- cluster::silhouette(result$clusters, stats::dist(eval_matrix))
          result$silhouette <- mean(sil[, 3])
        }, error = function(e) {
          result$silhouette <- NA
        })
      }

      if (requireNamespace("clusterCrit", quietly = TRUE)) {
        tryCatch({
          result$davies_bouldin <- clusterCrit::intCriteria(
            as.matrix(eval_matrix),
            result$clusters,
            "Davies_Bouldin"
          )$davies_bouldin
        }, error = function(e) {
          result$davies_bouldin <- NA
        })
      }

      if (requireNamespace("clusterCrit", quietly = TRUE)) {
        tryCatch({
          result$calinski_harabasz <- clusterCrit::intCriteria(
            as.matrix(eval_matrix),
            result$clusters,
            "Calinski_Harabasz"
          )$calinski_harabasz
        }, error = function(e) {
          result$calinski_harabasz <- NA
        })
      }
    }

    execution_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

    if (verbose) {
      message("Clustering analysis completed in ", round(execution_time, 2), " seconds")
      if (result$n_clusters > 0) {
        message("Found ", result$n_clusters, " clusters")
      }
    }

    result$execution_time <- execution_time
    result$timestamp <- Sys.time()
    result$seed <- seed

    return(result)

  }, error = function(e) {
    stop("Error in clustering analysis: ", e$message)
  }))
}

#' @title Generate Embeddings
#'
#' @description
#' Generates L2-normalized embeddings for texts using sentence transformers.
#'
#' @param texts A character vector of texts.
#' @param model Sentence transformer model name (default: "all-MiniLM-L6-v2").
#'   Newer small models (e.g. "BAAI/bge-small-en-v1.5") often score higher on
#'   retrieval benchmarks but expect instruction prefixes this function does
#'   not add.
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A matrix of L2-normalized embeddings.
#'
#' @concept semantic
#' @export
generate_embeddings <- function(texts, model = "all-MiniLM-L6-v2", verbose = TRUE) {

  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("Embedding generation via sentence-transformers"))
  }

  if (verbose) message("Generating embeddings using model: ", model)

  tryCatch({
    sentence_transformers <- reticulate::import("sentence_transformers")
    embedding_model <- sentence_transformers$SentenceTransformer(model)
    embeddings <- embedding_model$encode(texts, show_progress_bar = verbose, normalize_embeddings = TRUE)

    if (verbose) message("Embeddings generated successfully")

    return(embeddings)

  }, error = function(e) {
    stop("Error generating embeddings: ", e$message)
  })
}

#' @title Semantic Similarity Analysis
#' @description Wrapper for calculate_document_similarity
#' @param ... Arguments passed to calculate_document_similarity
#' @return Similarity analysis results
#' @concept semantic
#' @seealso [calculate_document_similarity()] for the core computation; [plot_similarity_heatmap()] to render the resulting matrix
#' @export
semantic_similarity_analysis <- function(...) {
  return(calculate_document_similarity(...))
}

#' @title Semantic Document Clustering
#' @description Creates a unified visualization of document clustering with optional clustering
#' @param embeddings Document embeddings matrix
#' @param method Dimensionality reduction method ("PCA", "t-SNE", "UMAP")
#' @param clusters Optional cluster assignments
#' @param ... Additional arguments
#'
#' @return A ggplot2 visualization of document clusters
#'
#' @concept semantic
#' @export
semantic_document_clustering <- function(embeddings, method = "UMAP", clusters = NULL, ...) {
  reduced_dims <- reduce_dimensions(embeddings, method = method, ...)

  if (!is.null(clusters)) {
    return(list(
      coordinates = reduced_dims,
      clusters = clusters,
      method = method
    ))
  }

  return(list(
    coordinates = reduced_dims,
    method = method
  ))
}

#' @title Analyze Document Clustering
#' @description Complete document clustering analysis with dimensionality reduction and optional clustering
#' @param feature_matrix Feature matrix (documents x features)
#' @param method Dimensionality reduction method ("PCA", "t-SNE", "UMAP")
#' @param clustering_method Clustering method ("none", "kmeans", "hierarchical", "dbscan", "hdbscan")
#' @param ... Additional parameters for methods
#' @return List containing coordinates, clusters, method info, and quality metrics
#' @concept semantic
#' @export
analyze_document_clustering <- function(feature_matrix,
                                  method = "UMAP",
                                  clustering_method = "none",
                                  ...) {

  coords <- reduce_dimensions(feature_matrix, method = method, ...)

  clusters <- NULL
  quality_metrics <- list()

  if (clustering_method != "none") {
    if (clustering_method == "kmeans") {
      k <- list(...)$k %||% 5
      km_result <- kmeans(coords, centers = k, nstart = 25)
      clusters <- km_result$cluster
    } else if (clustering_method == "hierarchical") {
      k <- list(...)$k %||% 5
      hc_result <- hclust(dist(coords))
      clusters <- cutree(hc_result, k = k)
    } else if (clustering_method == "dbscan") {
      eps <- list(...)$eps %||% 0.5
      minPts <- list(...)$minPts %||% 5
      dbscan_result <- dbscan::dbscan(coords, eps = eps, minPts = minPts)
      clusters <- dbscan_result$cluster
    } else if (clustering_method == "hdbscan") {
      min_size <- list(...)$min_cluster_size %||% 5
      hdbscan_result <- dbscan::hdbscan(coords, minPts = min_size)
      clusters <- hdbscan_result$cluster
    }

    if (!is.null(clusters) && length(unique(clusters)) > 1) {
      sil <- cluster::silhouette(clusters, dist(coords))
      quality_metrics$silhouette <- mean(sil[, 3])
      quality_metrics$n_clusters <- length(unique(clusters[clusters > 0]))

      if (min(clusters) == 0) {
        quality_metrics$outlier_pct <- sum(clusters == 0) / length(clusters) * 100
      }
    }
  }

  return(list(
    coordinates = coords,
    clusters = clusters,
    method = method,
    clustering_method = clustering_method,
    quality_metrics = quality_metrics
  ))
}

#' @title Generate Cluster Labels
#' @description Generate descriptive labels for document clusters
#' @param feature_matrix Feature matrix used for clustering
#' @param clusters Cluster assignments
#' @param method Label generation method ("tfidf", "representative", "frequent")
#' @param n_terms Number of terms per label
#' @return Named list of cluster labels
#' @concept semantic
#' @export
generate_cluster_labels_auto <- function(feature_matrix,
                                         clusters,
                                         method = "tfidf",
                                         n_terms = 3) {

  unique_clusters <- sort(unique(clusters[clusters > 0]))
  labels <- list()

  for (cluster_id in unique_clusters) {
    cluster_docs <- which(clusters == cluster_id)
    cluster_features <- feature_matrix[cluster_docs, , drop = FALSE]

    if (method == "tfidf") {
      tf <- colSums(cluster_features)
      idf <- log(nrow(feature_matrix) / colSums(feature_matrix > 0))
      tfidf <- tf * idf
      top_terms <- names(head(sort(tfidf, decreasing = TRUE), n_terms))
    } else if (method == "representative") {
      cluster_mean <- colMeans(cluster_features)
      overall_mean <- colMeans(feature_matrix)
      diff <- cluster_mean - overall_mean
      top_terms <- names(head(sort(diff, decreasing = TRUE), n_terms))
    } else {
      term_freq <- colSums(cluster_features)
      top_terms <- names(head(sort(term_freq, decreasing = TRUE), n_terms))
    }

    labels[[as.character(cluster_id)]] <- paste(top_terms, collapse = ", ")
  }

  return(labels)
}

#' @title Generate Cluster Label Suggestions (Human-in-the-Loop)
#'
#' @description
#' Suggests descriptive labels for clusters using AI. Labels are suggestions
#' for human review - users should edit and approve before using.
#' Supports OpenAI or Gemini for AI generation.
#'
#' @param cluster_keywords List of keywords for each cluster.
#' @param provider AI provider to use: "auto" (default), "openai", or "gemini".
#'   "auto" picks the first provider with an available API key.
#' @param model Model name. If NULL, uses provider defaults: "gpt-4.1-mini" (OpenAI),
#'   "gemini-2.5-flash-lite" (Gemini).
#' @param temperature Temperature parameter (default: 0.3).
#' @param max_tokens Maximum tokens for response (default: 50).
#' @param api_key API key for OpenAI or Gemini. If NULL, uses environment variable.
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A list of generated labels.
#'
#' @concept semantic
#' @export
#'
#' @examples
#' if (interactive()) {
#'   cluster_keywords <- list(
#'     "1" = c("calculator", "arithmetic", "elementary", "remedial"),
#'     "2" = c("computer", "instruction", "multiplication", "drill")
#'   )
#'   labels_openai <- generate_cluster_labels(cluster_keywords, provider = "openai")
#'   labels_gemini <- generate_cluster_labels(cluster_keywords, provider = "gemini")
#' }
generate_cluster_labels <- function(cluster_keywords,
                                   provider = "auto",
                                   model = NULL,
                                   temperature = 0.3,
                                   max_tokens = 50,
                                   api_key = NULL,
                                   verbose = TRUE) {

  if (!requireNamespace("httr", quietly = TRUE) ||
      !requireNamespace("jsonlite", quietly = TRUE)) {
    stop(
      "The 'httr' and 'jsonlite' packages are required for this functionality. ",
      "Please install them using install.packages(c('httr', 'jsonlite'))."
    )
  }

  if (provider == "auto") {
    if (nzchar(Sys.getenv("OPENAI_API_KEY")) || (!is.null(api_key) && grepl("^sk-", api_key))) {
      provider <- "openai"
      if (verbose) message("Using OpenAI for label generation")
    } else if (nzchar(Sys.getenv("GEMINI_API_KEY")) || (!is.null(api_key) && grepl("^AIza", api_key))) {
      provider <- "gemini"
      if (verbose) message("Using Gemini for label generation")
    } else {
      message("No AI provider available. Set OPENAI_API_KEY or GEMINI_API_KEY.")
      return(invisible(NULL))
    }
  }

  setup <- .resolve_llm_setup(
    provider, model, api_key,
    defaults = list(openai = "gpt-4.1-mini", gemini = "gemini-2.5-flash-lite"),
    strict_validate = TRUE
  )
  if (is.null(setup)) return(invisible(NULL))
  model <- setup$model
  api_key <- setup$api_key

  if (verbose) {
    message("Generating AI labels for ", length(cluster_keywords), " clusters using ", provider, " (", model, ")...")
  }

  gen_names <- list()

  for (cluster_id in names(cluster_keywords)) {
    if (cluster_id == "0" || cluster_id == 0) {
      gen_names[[cluster_id]] <- "Outlier"
      next
    }

    keywords <- cluster_keywords[[cluster_id]]
    if (length(keywords) == 0) {
      gen_names[[cluster_id]] <- paste("Cluster", cluster_id)
      next
    }

    top_keywords <- paste(keywords[seq_len(min(10, length(keywords)))], collapse = ", ")

    prompt <- paste0(
      "You are a highly skilled data scientist specializing in generating concise and
descriptive topic labels based on provided top terms for each topic.

Your objective is to create precise and concise labels that capture the essence of each topic by following these guidelines:

1. Use Person-First Language:
   - Prioritize respectful and inclusive language.
   - Avoid terms that may be considered offensive or stigmatizing.
   - For example, use 'students with learning disabilities' instead of 'disabled students'.

2. Analyze the significance of the top terms:
   - Focus primarily on the most significant terms.
   - Include additional terms if they add essential context.

3. Synthesize the Topic Label:
   - Ensure clarity and conciseness (aim for 4-5 words).
   - Reflect the collective meaning of the most influential terms.
   - Use descriptive yet precise phrasing.

4. Maintain consistency:
   - Capitalize the first word using title case.
   - Use uniform formatting and avoid ambiguity.
   - Make concise and complete expressions.

Top 10 Keywords: ", top_keywords, "

Respond with ONLY the topic label, nothing else. Do not include quotes or explanations.

Generated Topic Label:"
    )

    tryCatch({
      # Call LLM
      response_text <- call_llm_api(
        provider = provider,
        system_prompt = "You are a data scientist specializing in generating concise cluster labels.",
        user_prompt = prompt,
        model = model,
        temperature = temperature,
        max_tokens = max_tokens,
        api_key = api_key
      )

      if (!is.null(response_text) && nzchar(response_text)) {
        label <- trimws(response_text)
        label <- gsub("^\"(.*)\"$", "\\1", label)
        label <- gsub("^Generated Topic Label:\\s*", "", label, ignore.case = TRUE)
        label <- trimws(label)
        gen_names[[cluster_id]] <- label
      } else {
        gen_names[[cluster_id]] <- paste("Cluster", cluster_id)
      }

    }, error = function(e) {
      warning("AI call failed for cluster ", cluster_id, ": ", e$message)
      gen_names[[cluster_id]] <- paste("Cluster", cluster_id)
    })

    Sys.sleep(1)
  }

  if (verbose) message("AI label generation completed")

  return(gen_names)
}

#' @title Export Document Clustering Analysis
#' @description Export document clustering analysis results to CSV
#' @param coordinates Document coordinates
#' @param clusters Cluster assignments (optional)
#' @param labels Cluster labels (optional)
#' @param doc_ids Document IDs
#' @param file_path Path to save the CSV file
#'
#' @return Invisible data frame of the exported data
#'
#' @concept semantic
#' @export
export_document_clustering <- function(coordinates,
                                 clusters = NULL,
                                 labels = NULL,
                                 doc_ids = NULL,
                                 file_path) {

  export_data <- data.frame(
    doc_id = doc_ids %||% paste0("doc_", seq_len(nrow(coordinates))),
    x = coordinates[, 1],
    y = coordinates[, 2]
  )

  if (!is.null(clusters)) {
    export_data$cluster <- clusters

    if (!is.null(labels)) {
      export_data$cluster_label <- labels[match(clusters, names(labels))]
    }
  }

  utils::write.csv(export_data, file_path, row.names = FALSE)
  invisible(export_data)
}


#' @title Cross-Analysis Validation
#'
#' @description
#' Performs cross-validation between different analysis types (STM, semantic, clustering).
#'
#' @param semantic_results Results from semantic analysis.
#' @param stm_results Optional STM results for comparison.
#' @param verbose Logical, if TRUE, prints progress messages.
#'
#' @return A list containing cross-validation metrics.
#'
#' @concept semantic
#' @export
validate_cross_models <- function(semantic_results,
                                    stm_results = NULL,
                                    verbose = TRUE) {

  if (verbose) message("Running cross-analysis validation...")

  validation_results <- list()

  if (!is.null(semantic_results$clustering$topic_keywords) &&
      !is.null(stm_results$topic_keywords)) {

    if (verbose) message("Calculating topic-cluster correspondence...")

    correspondence <- calculate_topic_cluster_correspondence(
      semantic_results$clustering$topic_keywords,
      stm_results$topic_keywords
    )
    validation_results$topic_cluster_correspondence <- correspondence
  }

  if (!is.null(semantic_results$embeddings) &&
      !is.null(semantic_results$clustering$topic_assignments)) {

    if (verbose) message("Validating semantic coherence...")

    coherence <- validate_semantic_coherence(
      embeddings = semantic_results$embeddings,
      topic_assignments = semantic_results$clustering$topic_assignments
    )
    validation_results$semantic_coherence <- coherence
  }

  if (!is.null(semantic_results$clustering$topic_assignments) &&
      !is.null(stm_results$topic_assignments)) {

    if (verbose) message("Calculating assignment consistency...")

    consistency <- calculate_assignment_consistency(
      semantic_results$clustering$topic_assignments,
      stm_results$topic_assignments
    )
    validation_results$assignment_consistency <- consistency
  }

  if (verbose) message("Cross-analysis validation completed")

  return(validation_results)
}

#' @title Calculate Document Similarity with Fallbacks
#'
#' @description
#' Calculates document similarity with fallback methods and diagnostics.
#' Attempts embeddings first, falls back to Jaccard similarity if needed.
#'
#' @param texts Character vector of texts
#' @param method Similarity method ("embeddings" or "jaccard")
#' @param embedding_model Model name for embeddings (default: "all-MiniLM-L6-v2")
#' @param cache_embeddings Logical, cache embeddings (default: TRUE)
#' @param min_word_length Minimum word length for Jaccard (default: 3)
#' @param doc_names Optional document names
#'
#' @return List containing similarity matrix, method used, embeddings, and diagnostics
#'
#' @concept semantic
#' @export
#'
#' @examples
#' if (interactive()) {
#'   abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:5]
#'   similarity_result <- calculate_similarity_robust(abstracts)
#'   print(similarity_result$similarity_matrix)
#'   print(similarity_result$diagnostics)
#' }
calculate_similarity_robust <- function(texts,
                                       method = "embeddings",
                                       embedding_model = "all-MiniLM-L6-v2",
                                       cache_embeddings = TRUE,
                                       min_word_length = 3,
                                       doc_names = NULL) {

  diagnostics <- list(
    attempted_methods = character(),
    warnings = character(),
    computation_time = NULL
  )

  start_time <- Sys.time()

  if (is.null(doc_names)) {
    doc_names <- paste0("doc", seq_along(texts))
  }

  if (method == "embeddings" && requireNamespace("reticulate", quietly = TRUE)) {
    diagnostics$attempted_methods <- c(diagnostics$attempted_methods, "embeddings")

    embedding_result <- tryCatch({
      embeddings <- generate_embeddings(
        texts = texts,
        model = embedding_model,
        verbose = FALSE
      )

      if (is.null(dim(embeddings))) embeddings <- matrix(embeddings, nrow = 1)
      sklearn_metrics <- reticulate::import("sklearn.metrics.pairwise")
      similarity_matrix <- sklearn_metrics$cosine_similarity(embeddings)
      similarity_matrix <- as.matrix(similarity_matrix)

      rownames(similarity_matrix) <- doc_names
      colnames(similarity_matrix) <- doc_names

      list(
        success = TRUE,
        similarity_matrix = similarity_matrix,
        embeddings = embeddings,
        method = "embeddings"
      )
    }, error = function(e) {
      list(success = FALSE, error = e$message)
    })

    if (!embedding_result$success) {
      diagnostics$warnings <- c(diagnostics$warnings,
                                paste("Embeddings failed:", embedding_result$error))
    }

    if (embedding_result$success) {
      diagnostics$computation_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

      return(list(
        similarity_matrix = embedding_result$similarity_matrix,
        method_used = "embeddings",
        embeddings = embedding_result$embeddings,
        diagnostics = diagnostics
      ))
    }
  }

  diagnostics$attempted_methods <- c(diagnostics$attempted_methods, "jaccard")

  get_words <- function(text) {
    text <- tolower(text)
    text <- gsub("[^[:alnum:]\\s]", "", text)
    words <- unlist(strsplit(text, "\\s+"))
    words <- words[nchar(words) >= min_word_length]
    unique(words)
  }

  word_lists <- lapply(texts, get_words)
  n_docs <- length(texts)
  similarity_matrix <- matrix(0, nrow = n_docs, ncol = n_docs)

  for (i in seq_len(n_docs)) {
    for (j in seq_len(n_docs)) {
      if (i == j) {
        similarity_matrix[i, j] <- 1.0
      } else {
        words_i <- word_lists[[i]]
        words_j <- word_lists[[j]]

        if (length(words_i) == 0 && length(words_j) == 0) {
          similarity_matrix[i, j] <- 1.0
        } else if (length(words_i) == 0 || length(words_j) == 0) {
          similarity_matrix[i, j] <- 0.0
        } else {
          intersection <- length(intersect(words_i, words_j))
          union <- length(union(words_i, words_j))
          similarity_matrix[i, j] <- intersection / union
        }
      }
    }
  }

  rownames(similarity_matrix) <- doc_names
  colnames(similarity_matrix) <- doc_names

  diagnostics$computation_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  return(list(
    similarity_matrix = similarity_matrix,
    method_used = "jaccard",
    embeddings = NULL,
    diagnostics = diagnostics
  ))
}


#' Calculate Clustering Quality Metrics
#'
#' @description
#' Calculates common clustering evaluation metrics including Silhouette Score,
#' Davies-Bouldin Index, and Calinski-Harabasz Index.
#'
#' @param clusters Integer vector of cluster assignments
#' @param data_matrix Numeric matrix of data points (rows = observations, cols = features)
#' @param dist_matrix Optional distance matrix. If NULL, computed from data_matrix
#' @param metrics Character vector of metrics to calculate.
#'   Options: "silhouette", "davies_bouldin", "calinski_harabasz", or "all" (default)
#'
#' @return A named list containing:
#'   \describe{
#'     \item{silhouette}{Silhouette score (-1 to 1, higher is better)}
#'     \item{davies_bouldin}{Davies-Bouldin index (lower is better)}
#'     \item{calinski_harabasz}{Calinski-Harabasz index (higher is better)}
#'     \item{n_clusters}{Number of clusters}
#'     \item{cluster_sizes}{Table of cluster sizes}
#'   }
#'
#' @details
#' - Silhouette Score: Measures how similar an object is to its own cluster compared
#'   to other clusters. Range: -1 to 1, higher is better.
#' - Davies-Bouldin Index: Average similarity between each cluster and its most
#'   similar cluster. Lower values indicate better clustering.
#' - Calinski-Harabasz Index: Ratio of between-cluster to within-cluster variance.
#'   Higher values indicate better-defined clusters.
#'
#' @concept semantic
#' @export
#'
#' @examples
#' \donttest{
#' abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:20]
#' term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(abstracts)))
#' kmeans_result <- stats::kmeans(term_matrix, centers = 2)
#' metrics <- calculate_clustering_metrics(kmeans_result$cluster, term_matrix)
#' print(metrics)
#' }
calculate_clustering_metrics <- function(clusters,
                                            data_matrix,
                                            dist_matrix = NULL,
                                            metrics = "all") {
  if (is.null(clusters) || length(clusters) == 0) {
    return(list(
      silhouette = NA,
      davies_bouldin = NA,
      calinski_harabasz = NA,
      n_clusters = 0,
      cluster_sizes = NULL
    ))
  }

  data_matrix <- as.matrix(data_matrix)
  clusters <- as.numeric(as.factor(clusters))
  unique_clusters <- unique(clusters)
  n_clusters <- length(unique_clusters)
  n_points <- nrow(data_matrix)

  if (n_clusters <= 1) {
    return(list(
      silhouette = NA,
      davies_bouldin = NA,
      calinski_harabasz = NA,
      n_clusters = n_clusters,
      cluster_sizes = table(clusters)
    ))
  }

  if (metrics == "all") {
    metrics <- c("silhouette", "davies_bouldin", "calinski_harabasz")
  }

  result <- list(
    n_clusters = n_clusters,
    cluster_sizes = table(clusters)
  )

  if (is.null(dist_matrix)) {
    dist_matrix <- stats::dist(data_matrix)
  }

  if ("silhouette" %in% metrics) {
    result$silhouette <- tryCatch({
      if (requireNamespace("cluster", quietly = TRUE)) {
        sil_result <- cluster::silhouette(clusters, dist_matrix)
        mean(sil_result[, 3])
      } else {
        NA
      }
    }, error = function(e) NA)
  }

  if ("davies_bouldin" %in% metrics) {
    result$davies_bouldin <- tryCatch({
      centers <- matrix(0, nrow = n_clusters, ncol = ncol(data_matrix))
      for (i in seq_len(n_clusters)) {
        cluster_id <- unique_clusters[i]
        cluster_points <- data_matrix[clusters == cluster_id, , drop = FALSE]
        if (nrow(cluster_points) > 0) {
          centers[i, ] <- colMeans(cluster_points)
        }
      }

      within_distances <- numeric(n_clusters)
      for (i in seq_len(n_clusters)) {
        cluster_id <- unique_clusters[i]
        cluster_points <- data_matrix[clusters == cluster_id, , drop = FALSE]
        if (nrow(cluster_points) > 0) {
          distances <- apply(cluster_points, 1, function(x) sqrt(sum((x - centers[i, ])^2)))
          within_distances[i] <- mean(distances)
        }
      }

      db_values <- numeric(n_clusters)
      for (i in seq_len(n_clusters)) {
        max_ratio <- 0
        for (j in seq_len(n_clusters)) {
          if (i != j) {
            between_distance <- sqrt(sum((centers[i, ] - centers[j, ])^2))
            if (between_distance > 0) {
              ratio <- (within_distances[i] + within_distances[j]) / between_distance
              max_ratio <- max(max_ratio, ratio)
            }
          }
        }
        db_values[i] <- max_ratio
      }

      mean(db_values)
    }, error = function(e) NA)
  }

  if ("calinski_harabasz" %in% metrics) {
    result$calinski_harabasz <- tryCatch({
      if (n_clusters >= n_points) {
        return(NA)
      }

      overall_centroid <- colMeans(data_matrix)

      cluster_centroids <- matrix(0, nrow = n_clusters, ncol = ncol(data_matrix))
      cluster_sizes <- numeric(n_clusters)

      for (i in seq_len(n_clusters)) {
        cluster_id <- unique_clusters[i]
        cluster_points <- data_matrix[clusters == cluster_id, , drop = FALSE]
        if (nrow(cluster_points) > 0) {
          cluster_centroids[i, ] <- colMeans(cluster_points)
          cluster_sizes[i] <- nrow(cluster_points)
        }
      }

      ssb <- 0
      for (i in seq_len(n_clusters)) {
        if (cluster_sizes[i] > 0) {
          distance_to_overall <- sum((cluster_centroids[i, ] - overall_centroid)^2)
          ssb <- ssb + cluster_sizes[i] * distance_to_overall
        }
      }

      ssw <- 0
      for (i in seq_len(n_clusters)) {
        cluster_id <- unique_clusters[i]
        cluster_points <- data_matrix[clusters == cluster_id, , drop = FALSE]
        if (nrow(cluster_points) > 0) {
          for (j in seq_len(nrow(cluster_points))) {
            distance_to_centroid <- sum((cluster_points[j, ] - cluster_centroids[i, ])^2)
            ssw <- ssw + distance_to_centroid
          }
        }
      }

      if (ssw > 0 && (n_points - n_clusters) > 0) {
        (ssb / (n_clusters - 1)) / (ssw / (n_points - n_clusters))
      } else {
        NA
      }
    }, error = function(e) NA)
  }

  result
}


#' Calculate Cross-Matrix Cosine Similarity
#'
#' Calculates cosine similarity between two different embedding matrices,
#' useful for comparing documents/topics across different categories or groups.
#'
#' @param embeddings1 A numeric matrix where rows are items and columns are embedding dimensions.
#' @param embeddings2 A numeric matrix where rows are items and columns are embedding dimensions.
#'   Must have the same number of columns as embeddings1.
#' @param labels1 Optional character vector of labels for items in embeddings1.
#' @param labels2 Optional character vector of labels for items in embeddings2.
#' @param normalize Logical, whether to L2-normalize embeddings before computing similarity (default: TRUE).
#'
#' @return A list containing:
#'   \item{similarity_matrix}{Matrix of cosine similarities (nrow(embeddings1) x nrow(embeddings2))}
#'   \item{similarity_df}{Long-format data frame with columns: row_idx, col_idx, similarity, and optionally label1, label2}
#'
#' @concept semantic
#' @export
#'
#' @examples
#' \donttest{
#' abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:6]
#' term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(abstracts)))
#' similarity_result <- calculate_cross_similarity(
#'   term_matrix[1:3, ], term_matrix[4:6, ],
#'   labels1 = paste("Doc", 1:3),
#'   labels2 = paste("Doc", 4:6)
#' )
#' similarity_result$similarity_matrix
#' }
calculate_cross_similarity <- function(embeddings1,
                                        embeddings2,
                                        labels1 = NULL,
                                        labels2 = NULL,
                                        normalize = TRUE) {

  embeddings1 <- as.matrix(embeddings1)
  embeddings2 <- as.matrix(embeddings2)

  if (ncol(embeddings1) != ncol(embeddings2)) {
    stop("Embedding matrices must have the same number of columns (dimensions)")
  }

  if (normalize) {
    norms1 <- sqrt(rowSums(embeddings1^2))
    norms1[norms1 == 0] <- 1
    norm_emb1 <- embeddings1 / norms1

    norms2 <- sqrt(rowSums(embeddings2^2))
    norms2[norms2 == 0] <- 1
    norm_emb2 <- embeddings2 / norms2
  } else {
    norm_emb1 <- embeddings1
    norm_emb2 <- embeddings2
  }

  similarity_matrix <- norm_emb1 %*% t(norm_emb2)

  similarity_df <- tidyr::expand_grid(
    row_idx = seq_len(nrow(embeddings1)),
    col_idx = seq_len(nrow(embeddings2))
  ) %>%
    dplyr::mutate(
      similarity = as.vector(similarity_matrix)
    )

  if (!is.null(labels1)) {
    if (length(labels1) != nrow(embeddings1)) {
      warning("labels1 length does not match number of rows in embeddings1")
    } else {
      similarity_df <- similarity_df %>%
        dplyr::mutate(label1 = labels1[row_idx])
    }
  }

  if (!is.null(labels2)) {
    if (length(labels2) != nrow(embeddings2)) {
      warning("labels2 length does not match number of rows in embeddings2")
    } else {
      similarity_df <- similarity_df %>%
        dplyr::mutate(label2 = labels2[col_idx])
    }
  }

  list(
    similarity_matrix = similarity_matrix,
    similarity_df = similarity_df
  )
}


#' Extract Cross-Category Similarities from Full Similarity Matrix
#'
#' Given a full similarity matrix and category information, extracts pairwise
#' similarities between a reference category and other categories into a long-format
#' data frame suitable for visualization and analysis.
#'
#' @param similarity_matrix A square similarity matrix (n x n).
#' @param docs_data A data frame containing document metadata with at least:
#'   \describe{
#'     \item{category_var}{Column indicating category membership}
#'     \item{id_var}{Column with unique document identifiers}
#'   }
#' @param reference_category Character string specifying the reference category to compare against.
#' @param compare_categories Character vector of categories to compare with the reference.
#'   If NULL, compares with all categories except reference.
#' @param category_var Name of the column containing category information (default: "category").
#' @param id_var Name of the column containing document IDs (default: "display_name").
#' @param name_var Optional name of column with display names (default: NULL, uses id_var).
#'
#' @return A data frame with columns:
#'   \item{ref_id}{Reference document ID}
#'   \item{ref_name}{Reference document name (if name_var provided)}
#'   \item{other_id}{Comparison document ID}
#'   \item{other_name}{Comparison document name (if name_var provided)}
#'   \item{other_category}{Category of comparison document}
#'   \item{similarity}{Cosine similarity value}
#'
#' @concept semantic
#' @export
#'
#' @examples
#' \donttest{
#' articles <- TextAnalysisR::SpecialEduTech[1:6, ]
#' articles$display_name <- paste0("d", seq_len(nrow(articles)))
#' term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(articles$abstract)))
#' normalized_matrix <- term_matrix / sqrt(rowSums(term_matrix ^ 2))
#' similarity_matrix <- normalized_matrix %*% t(normalized_matrix)
#' dimnames(similarity_matrix) <- list(articles$display_name, articles$display_name)
#' cross_similarities <- extract_cross_category_similarities(
#'   similarity_matrix  = similarity_matrix,
#'   docs_data          = articles,
#'   reference_category = "thesis",
#'   compare_categories = "journal_article",
#'   category_var       = "reference_type",
#'   id_var             = "display_name",
#'   name_var           = "title"
#' )
#' }
extract_cross_category_similarities <- function(similarity_matrix,
                                                 docs_data,
                                                 reference_category,
                                                 compare_categories = NULL,
                                                 category_var = "category",
                                                 id_var = "display_name",
                                                 name_var = NULL) {

  if (!requireNamespace("purrr", quietly = TRUE)) {
    stop("Package 'purrr' is required. Install with: install.packages('purrr')")
  }

  if (!category_var %in% names(docs_data)) {
    stop("category_var '", category_var, "' not found in docs_data")
  }
  if (!id_var %in% names(docs_data)) {
    stop("id_var '", id_var, "' not found in docs_data")
  }

  docs_data <- docs_data %>%
    dplyr::mutate(.row_idx = dplyr::row_number())

  ref_indices <- docs_data %>%
    dplyr::filter(.data[[category_var]] == reference_category) %>%
    dplyr::pull(.row_idx)

  if (length(ref_indices) == 0) {
    stop("No documents found in reference category: ", reference_category)
  }

  if (is.null(compare_categories)) {
    compare_categories <- unique(docs_data[[category_var]])
    compare_categories <- compare_categories[compare_categories != reference_category]
  }

  result_list <- list()

  for (comp_cat in compare_categories) {
    comp_indices <- docs_data %>%
      dplyr::filter(.data[[category_var]] == comp_cat) %>%
      dplyr::pull(.row_idx)

    if (length(comp_indices) == 0) next

    pairs <- tidyr::expand_grid(
      ref_idx = ref_indices,
      other_idx = comp_indices
    ) %>%
      dplyr::mutate(
        similarity = purrr::map2_dbl(ref_idx, other_idx, ~similarity_matrix[.x, .y]),
        other_category = comp_cat
      )

    result_list[[comp_cat]] <- pairs
  }

  result <- dplyr::bind_rows(result_list)

  ref_lookup <- docs_data %>%
    dplyr::select(.row_idx, !!rlang::sym(id_var)) %>%
    dplyr::rename(ref_id = !!rlang::sym(id_var))

  other_lookup <- docs_data %>%
    dplyr::select(.row_idx, !!rlang::sym(id_var)) %>%
    dplyr::rename(other_id = !!rlang::sym(id_var))

  result <- result %>%
    dplyr::left_join(ref_lookup, by = c("ref_idx" = ".row_idx")) %>%
    dplyr::left_join(other_lookup, by = c("other_idx" = ".row_idx"))

  if (!is.null(name_var) && name_var %in% names(docs_data)) {
    ref_names <- docs_data %>%
      dplyr::select(.row_idx, !!rlang::sym(name_var)) %>%
      dplyr::rename(ref_name = !!rlang::sym(name_var))

    other_names <- docs_data %>%
      dplyr::select(.row_idx, !!rlang::sym(name_var)) %>%
      dplyr::rename(other_name = !!rlang::sym(name_var))

    result <- result %>%
      dplyr::left_join(ref_names, by = c("ref_idx" = ".row_idx")) %>%
      dplyr::left_join(other_names, by = c("other_idx" = ".row_idx")) %>%
      dplyr::select(-ref_idx, -other_idx) %>%
      dplyr::select(ref_id, ref_name, other_id, other_name, other_category, similarity)
  } else {
    result <- result %>%
      dplyr::select(-ref_idx, -other_idx) %>%
      dplyr::select(ref_id, other_id, other_category, similarity)
  }

  result %>%
    dplyr::mutate(other_category = factor(other_category, levels = compare_categories))
}


#' Analyze Similarity Gaps Between Categories
#'
#' Identifies unique items, missing content, and cross-category learning opportunities
#' based on similarity thresholds. Useful for gap analysis in policy documents,
#' topic comparisons, or any cross-category similarity study.
#'
#' @param similarity_data A data frame with cross-category similarities, containing:
#'   \describe{
#'     \item{ref_var}{Reference item identifier}
#'     \item{other_var}{Comparison item identifier}
#'     \item{similarity_var}{Similarity score}
#'     \item{category_var}{Category of comparison item}
#'   }
#' @param ref_var Name of column with reference item IDs (default: "ref_id").
#' @param other_var Name of column with comparison item IDs (default: "other_id").
#' @param similarity_var Name of column with similarity values (default: "similarity").
#' @param category_var Name of column with category information (default: "other_category").
#' @param ref_label_var Optional column with reference item labels (for output).
#' @param other_label_var Optional column with comparison item labels (for output).
#' @param unique_threshold Threshold below which reference items are considered unique (default: 0.6).
#' @param cross_policy_min Minimum similarity for cross-policy opportunities (default: 0.6).
#' @param cross_policy_max Maximum similarity for cross-policy opportunities (default: 0.8).
#'
#' @return A list containing:
#'   \item{unique_items}{Data frame of reference items with low similarity (unique content)}
#'   \item{missing_items}{Data frame of comparison items with low similarity (content gaps)}
#'   \item{cross_policy}{Data frame of items with moderate similarity (learning opportunities)}
#'   \item{summary_stats}{Summary statistics by category}
#'
#' @concept semantic
#' @export
#'
#' @examples
#' \donttest{
#' articles <- TextAnalysisR::SpecialEduTech[1:6, ]
#' articles$display_name <- paste0("d", seq_len(nrow(articles)))
#' term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(articles$abstract)))
#' normalized_matrix <- term_matrix / sqrt(rowSums(term_matrix ^ 2))
#' similarity_matrix <- normalized_matrix %*% t(normalized_matrix)
#' dimnames(similarity_matrix) <- list(articles$display_name, articles$display_name)
#' cross_similarities <- extract_cross_category_similarities(
#'   similarity_matrix  = similarity_matrix,
#'   docs_data          = articles,
#'   reference_category = "thesis",
#'   compare_categories = "journal_article",
#'   category_var       = "reference_type",
#'   id_var             = "display_name"
#' )
#' gap_analysis <- analyze_similarity_gaps(
#'   similarity_data = cross_similarities,
#'   ref_var = "ref_id",
#'   other_var = "other_id",
#'   similarity_var = "similarity",
#'   category_var = "other_category",
#'   unique_threshold = 0.6
#' )
#' print(gap_analysis$summary_stats)
#' }
analyze_similarity_gaps <- function(similarity_data,
                                     ref_var = "ref_id",
                                     other_var = "other_id",
                                     similarity_var = "similarity",
                                     category_var = "other_category",
                                     ref_label_var = NULL,
                                     other_label_var = NULL,
                                     unique_threshold = 0.6,
                                     cross_policy_min = 0.6,
                                     cross_policy_max = 0.8) {

  for (v in c(ref_var, other_var, similarity_var, category_var)) {
    if (!v %in% names(similarity_data)) {
      stop("Column '", v, "' not found in similarity_data")
    }
  }

  summary_stats <- similarity_data %>%
    dplyr::group_by(.data[[category_var]]) %>%
    dplyr::summarise(
      mean_similarity = round(mean(.data[[similarity_var]], na.rm = TRUE), 3),
      median_similarity = round(stats::median(.data[[similarity_var]], na.rm = TRUE), 3),
      sd_similarity = round(stats::sd(.data[[similarity_var]], na.rm = TRUE), 3),
      min_similarity = round(min(.data[[similarity_var]], na.rm = TRUE), 3),
      max_similarity = round(max(.data[[similarity_var]], na.rm = TRUE), 3),
      n_pairs = dplyr::n(),
      .groups = "drop"
    )

  group_vars <- ref_var
  if (!is.null(ref_label_var) && ref_label_var %in% names(similarity_data)) {
    group_vars <- c(group_vars, ref_label_var)
  }

  unique_items <- similarity_data %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(group_vars))) %>%
    dplyr::summarise(
      max_similarity = max(.data[[similarity_var]], na.rm = TRUE),
      min_similarity = min(.data[[similarity_var]], na.rm = TRUE),
      .max_idx = which.max(.data[[similarity_var]]),
      best_match = .data[[other_var]][.max_idx],
      best_match_category = .data[[category_var]][.max_idx],
      .groups = "drop"
    ) %>%
    dplyr::select(-.max_idx) %>%
    dplyr::filter(max_similarity < unique_threshold) %>%
    dplyr::mutate(gap_type = "Unique")

  other_group_vars <- c(other_var, category_var)
  if (!is.null(other_label_var) && other_label_var %in% names(similarity_data)) {
    other_group_vars <- c(other_group_vars, other_label_var)
  }

  missing_items <- similarity_data %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(other_group_vars))) %>%
    dplyr::summarise(
      max_similarity = max(.data[[similarity_var]], na.rm = TRUE),
      min_similarity = min(.data[[similarity_var]], na.rm = TRUE),
      .max_idx = which.max(.data[[similarity_var]]),
      best_match = .data[[ref_var]][.max_idx],
      .groups = "drop"
    ) %>%
    dplyr::select(-.max_idx) %>%
    dplyr::filter(max_similarity < unique_threshold) %>%
    dplyr::mutate(gap_type = "Missing")

  cross_policy <- similarity_data %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(group_vars))) %>%
    dplyr::summarise(
      max_similarity = max(.data[[similarity_var]], na.rm = TRUE),
      min_similarity = min(.data[[similarity_var]], na.rm = TRUE),
      .max_idx = which.max(.data[[similarity_var]]),
      best_match = .data[[other_var]][.max_idx],
      best_match_category = .data[[category_var]][.max_idx],
      .groups = "drop"
    ) %>%
    dplyr::select(-.max_idx) %>%
    dplyr::filter(
      max_similarity >= cross_policy_min,
      max_similarity < cross_policy_max
    ) %>%
    dplyr::mutate(gap_type = "Cross-Policy")

  list(
    unique_items = unique_items,
    missing_items = missing_items,
    cross_policy = cross_policy,
    summary_stats = summary_stats
  )
}


# Sentiment analysis

#' Analyze Text Sentiment
#'
#' @description
#' Performs sentiment analysis on text data using the syuzhet package.
#' Classification uses the per-token average score with a +/- 0.05 neutral
#' band, so document length does not drive polarity. Lexicon methods score
#' words in isolation; negation, irony, and idioms are not handled. For
#' negation-sensitive text, see [analyze_sentiment_llm()].
#'
#' @param texts Character vector of texts to analyze
#' @param method Sentiment analysis method: "syuzhet", "bing", "afinn", or "nrc" (default: "syuzhet").
#'   All methods score each matched dictionary word once per document
#'   regardless of repetition.
#' @param doc_ids Optional character vector of document identifiers (default: NULL)
#'
#' @return A data frame with columns:
#'   \describe{
#'     \item{document}{Document identifier}
#'     \item{text}{Original text}
#'     \item{sentiment_score}{Raw summed sentiment score (length-dependent)}
#'     \item{n_tokens}{Number of whitespace-delimited tokens}
#'     \item{avg_sentiment}{sentiment_score / n_tokens}
#'     \item{sentiment}{Classification: "positive", "negative", or "neutral"}
#'   }
#'
#' @concept sentiment
#' @export
#'
#' @examples
#' \donttest{
#' abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
#' sentiment_results <- analyze_sentiment(abstracts)
#' print(sentiment_results)
#' }
analyze_sentiment <- function(texts,
                              method = "syuzhet",
                              doc_ids = NULL) {

  if (!requireNamespace("syuzhet", quietly = TRUE)) {
    stop("Package 'syuzhet' is required. Please install it with: install.packages('syuzhet')")
  }

  if (is.null(doc_ids)) {
    doc_ids <- paste0("doc", seq_along(texts))
  }

  sentiment_scores <- syuzhet::get_sentiment(texts, method = method)
  n_tokens <- vapply(strsplit(texts, "\\s+"), function(x) sum(nzchar(x)), integer(1))
  avg_sentiment <- ifelse(n_tokens > 0, sentiment_scores / n_tokens, 0)

  sentiment_classification <- dplyr::case_when(
    avg_sentiment >= 0.05 ~ "positive",
    avg_sentiment <= -0.05 ~ "negative",
    .default = "neutral"
  )

  data.frame(
    document = doc_ids,
    text = texts,
    sentiment_score = sentiment_scores,
    n_tokens = n_tokens,
    avg_sentiment = avg_sentiment,
    sentiment = sentiment_classification,
    stringsAsFactors = FALSE
  )
}


#' Plot Sentiment Distribution
#'
#' @description
#' Creates a bar plot showing the distribution of sentiment classifications.
#'
#' @param sentiment_data Data frame from analyze_sentiment() or with 'sentiment' column
#' @param title Plot title (default: "Sentiment Distribution")
#'
#' @return A ggplot2 bar chart
#'
#' @concept sentiment
#' @export
#'
#' @examples
#' \donttest{
#' abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
#' sentiment_data <- analyze_sentiment(abstracts)
#' sentiment_plot <- plot_sentiment_distribution(sentiment_data)
#' print(sentiment_plot)
#' }
plot_sentiment_distribution <- function(sentiment_data,
                                        title = "Sentiment Distribution") {

  if (!"sentiment" %in% names(sentiment_data)) {
    stop("Data must contain a 'sentiment' column. Use analyze_sentiment() first.")
  }

  sentiment_counts <- table(sentiment_data$sentiment)

  ordered_sentiments <- c("positive", "negative", "neutral")
  sentiment_counts <- sentiment_counts[ordered_sentiments[ordered_sentiments %in% names(sentiment_counts)]]

  colors <- get_sentiment_colors()

  plot_df <- data.frame(
    sentiment = factor(names(sentiment_counts), levels = names(sentiment_counts)),
    count = as.numeric(sentiment_counts)
  )

  plot_df$hover_text <- paste("Sentiment:", plot_df$sentiment, "<br>Count:", plot_df$count)

  ggplot2::ggplot(plot_df, ggplot2::aes(x = .data$sentiment, y = .data$count, fill = .data$sentiment,
                                         text = .data$hover_text)) +
    ggplot2::geom_col() +
    ggplot2::scale_fill_manual(values = colors, guide = "none") +
    ggplot2::labs(title = title, x = "Sentiment", y = "Number of Documents") +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a")
    )
}


#' Plot Sentiment by Category
#'
#' @description
#' Creates a grouped or stacked bar plot showing sentiment distribution across categories.
#'
#' @param sentiment_data Data frame with 'sentiment' column
#' @param category_var Name of the category variable column
#' @param plot_type Type of plot: "bar" or "stacked" (default: "bar")
#' @param title Plot title (default: auto-generated)
#'
#' @return A ggplot2 grouped/stacked bar chart
#'
#' @concept sentiment
#' @export
#'
#' @examples
#' \donttest{
#' articles <- TextAnalysisR::SpecialEduTech[1:20, ]
#' sentiment_results <- analyze_sentiment(articles$abstract)
#' sentiment_data <- cbind(
#'   reference_type = articles$reference_type,
#'   sentiment_results
#' )
#' sentiment_plot <- plot_sentiment_by_category(sentiment_data, "reference_type")
#' print(sentiment_plot)
#' }
plot_sentiment_by_category <- function(sentiment_data,
                                       category_var,
                                       plot_type = "bar",
                                       title = NULL) {

  if (!category_var %in% names(sentiment_data)) {
    stop(paste("Category variable", category_var, "not found in data"))
  }

  if (!"sentiment" %in% names(sentiment_data)) {
    stop("Data must contain a 'sentiment' column. Use analyze_sentiment() first.")
  }

  grouped_data <- sentiment_data %>%
    dplyr::group_by(!!rlang::sym(category_var), sentiment) %>%
    dplyr::summarise(count = dplyr::n(), .groups = "drop") %>%
    dplyr::group_by(!!rlang::sym(category_var)) %>%
    dplyr::mutate(proportion = count / sum(count)) %>%
    dplyr::ungroup()

  names(grouped_data)[1] <- "category_var"

  colors <- get_sentiment_colors()

  if (is.null(title)) {
    title <- paste("Sentiment by", category_var)
  }

  bar_position <- if (plot_type == "stacked") "stack" else "dodge"

  grouped_data$hover_text <- paste("Category:", grouped_data$category_var,
                                   "<br>Sentiment:", grouped_data$sentiment,
                                   "<br>Proportion:", round(grouped_data$proportion, 3))

  ggplot2::ggplot(grouped_data, ggplot2::aes(x = .data$category_var, y = .data$proportion, fill = .data$sentiment,
                                              text = .data$hover_text)) +
    ggplot2::geom_col(position = bar_position) +
    ggplot2::scale_fill_manual(values = colors) +
    ggplot2::labs(title = title, x = category_var, y = "Proportion", fill = "Sentiment") +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
      axis.text.x = ggplot2::element_text(angle = 45, hjust = 1, size = 11, color = "#3B3B3B"),
      axis.text.y = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      legend.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      legend.text = ggplot2::element_text(size = 12, color = "#3B3B3B")
    )
}


#' Plot Document Sentiment Trajectory
#'
#' @description
#' Creates a line chart showing sentiment scores across documents with color gradient.
#'
#' @param sentiment_data Data frame from analyze_sentiment() with sentiment_score column
#' @param top_n Number of documents to display (default: NULL for all)
#' @param doc_ids Optional vector of custom document IDs for display (default: NULL)
#' @param text_preview Optional vector of text snippets for tooltips (default: NULL)
#' @param title Plot title (default: "Document Sentiment Scores")
#'
#' @return A ggplot2 line chart with color gradient
#'
#' @concept sentiment
#' @export
plot_document_sentiment_trajectory <- function(sentiment_data,
                                               top_n = NULL,
                                               doc_ids = NULL,
                                               text_preview = NULL,
                                               title = "Document Sentiment Scores") {

  doc_data <- sentiment_data %>%
    dplyr::arrange(document) %>%
    dplyr::mutate(doc_index = dplyr::row_number())

  if (!is.null(top_n)) {
    doc_data <- doc_data %>%
      dplyr::slice_head(n = top_n)
  }

  if (!is.null(doc_ids)) {
    doc_data$display_id <- doc_ids[doc_data$document]
  } else {
    doc_data$display_id <- paste("Doc", doc_data$document)
  }

  doc_data$hover_text <- paste("Document:", doc_data$display_id,
                               "<br>Score:", round(doc_data$sentiment_score, 3))

  ggplot2::ggplot(doc_data, ggplot2::aes(x = .data$doc_index, y = .data$sentiment_score, text = .data$hover_text)) +
    ggplot2::geom_line(color = "#6B7280") +
    ggplot2::geom_point(ggplot2::aes(color = .data$sentiment_score), size = 2) +
    ggplot2::scale_color_gradient2(
      low = "#EF4444", mid = "#6B7280", high = "#10B981", midpoint = 0,
      name = "Sentiment Score"
    ) +
    ggplot2::geom_hline(yintercept = 0, linetype = "dashed", color = "#9CA3AF") +
    ggplot2::labs(title = title, x = "Document Index", y = "Sentiment Score") +
    ggplot2::theme_minimal(base_size = 11)
}

#' Analyze Sentiment Using Tidytext Lexicons
#'
#' @description
#' Performs lexicon-based sentiment analysis on a DFM object using tidytext lexicons.
#' Supports AFINN, Bing, and NRC lexicons with scoring and emotion analysis.
#'
#' Tokens absent from the lexicon are ignored (not treated as neutral);
#' \code{summary_stats$token_match_rate} reports the share of corpus tokens
#' the lexicon covered. AFINN classification uses a +/- 0.5 band on the
#' per-sentiment-word average (\code{avg_sentiment}); its
#' \code{n_sentiment_words} column counts matched sentiment tokens, not
#' document length.
#'
#' @param dfm_object A quanteda DFM object (unigram)
#' @param lexicon Lexicon to use: "afinn", "bing", or "nrc" (default: "bing")
#' @param texts_df Optional data frame with original texts and metadata (default: NULL)
#' @param feature_type Feature space: "words" (unigrams). The AFINN/Bing/NRC lexicons are
#'   unigram lexicons; "ngrams" falls back to unigram scoring with a warning (default: "words").
#' @param ngram_range Retained for backward compatibility; not used for scoring (default: 2)
#' @param texts Optional character vector of texts used to rebuild a unigram DFM (default: NULL)
#'
#' @return A list containing:
#'   \describe{
#'     \item{document_sentiment}{Data frame with sentiment scores per document}
#'     \item{emotion_scores}{Data frame with emotion scores (NRC only)}
#'     \item{summary_stats}{List of summary statistics}
#'     \item{feature_type}{Feature type used for analysis}
#'   }
#'
#' @importFrom tidytext tidy get_sentiments
#' @importFrom dplyr inner_join group_by summarise mutate case_when ungroup n_distinct
#' @importFrom tidyr pivot_wider pivot_longer
#' @concept sentiment
#' @export
#'
#' @examples
#' \donttest{
#' abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
#' corpus <- quanteda::corpus(abstracts)
#' dfm_object <- quanteda::dfm(quanteda::tokens(corpus))
#' lexicon_results <- sentiment_lexicon_analysis(dfm_object, lexicon = "bing")
#' print(lexicon_results$document_sentiment)
#' }
sentiment_lexicon_analysis <- function(dfm_object,
                                       lexicon = "bing",
                                       texts_df = NULL,
                                       feature_type = "words",
                                       ngram_range = 2,
                                       texts = NULL) {

  if (!requireNamespace("tidytext", quietly = TRUE)) {
    stop("Package 'tidytext' is required. Please install it.")
  }

  if (feature_type == "ngrams" && !is.null(texts)) {
    if (!requireNamespace("quanteda", quietly = TRUE)) {
      stop("Package 'quanteda' is required to rebuild the DFM from texts.")
    }

    warning("AFINN/Bing/NRC are unigram lexicons and cannot match n-grams; ",
            "scoring falls back to unigrams.")

    corp <- quanteda::corpus(texts)
    toks <- quanteda::tokens(corp, remove_punct = TRUE, remove_symbols = TRUE)
    dfm_object <- quanteda::dfm(toks)
  }

  tidy_dfm <- tidytext::tidy(dfm_object)
  lexicon_name <- tolower(lexicon)

  if (!lexicon_name %in% c("afinn", "bing", "nrc")) {
    stop("Invalid lexicon. Choose 'afinn', 'bing', or 'nrc'.")
  }

  # afinn/nrc download through textdata behind an interactive license prompt
  sentiment_lexicon <- tryCatch(
    tidytext::get_sentiments(lexicon_name),
    error = function(e) {
      stop("Lexicon '", lexicon_name, "' is unavailable: ", conditionMessage(e),
           "\nThe 'afinn' and 'nrc' lexicons require a one-time interactive ",
           "download. Run tidytext::get_sentiments(\"", lexicon_name,
           "\") in an interactive session to accept the license.",
           call. = FALSE)
    }
  )

  doc_names <- quanteda::docnames(dfm_object)

  if (lexicon_name == "afinn") {
    doc_sentiment <- tidy_dfm %>%
      dplyr::inner_join(sentiment_lexicon, by = c("term" = "word"), relationship = "many-to-many") %>%
      dplyr::group_by(document) %>%
      dplyr::summarise(
        sentiment_score = sum(value * count, na.rm = TRUE),
        n_sentiment_words = sum(count),
        avg_sentiment = sentiment_score / n_sentiment_words,
        # +/- 0.5 marks mild valence on the AFINN -5..5 scale
        sentiment = dplyr::case_when(
          avg_sentiment > 0.5 ~ "positive",
          avg_sentiment < -0.5 ~ "negative",
          TRUE ~ "neutral"
        ),
        .groups = "drop"
      )
  } else if (lexicon_name == "bing") {
    doc_sentiment <- tidy_dfm %>%
      dplyr::inner_join(sentiment_lexicon, by = c("term" = "word"), relationship = "many-to-many") %>%
      dplyr::group_by(document, sentiment) %>%
      dplyr::summarise(n = sum(count), .groups = "drop") %>%
      tidyr::pivot_wider(names_from = sentiment, values_from = n, values_fill = 0)

    if (!("positive" %in% names(doc_sentiment))) {
      doc_sentiment$positive <- 0
    }
    if (!("negative" %in% names(doc_sentiment))) {
      doc_sentiment$negative <- 0
    }

    doc_sentiment <- doc_sentiment %>%
      dplyr::mutate(
        total_sentiment_words = .data$positive + .data$negative,
        sentiment_score = ifelse(.data$total_sentiment_words > 0,
                                 (.data$positive - .data$negative) / .data$total_sentiment_words,
                                 0),
        sentiment_raw = .data$positive - .data$negative,
        sentiment = dplyr::case_when(
          .data$sentiment_score > 0 ~ "positive",
          .data$sentiment_score < 0 ~ "negative",
          TRUE ~ "neutral"
        )
      )
  } else if (lexicon_name == "nrc") {
    doc_sentiment <- tidy_dfm %>%
      dplyr::inner_join(sentiment_lexicon, by = c("term" = "word"), relationship = "many-to-many") %>%
      dplyr::group_by(document, sentiment) %>%
      dplyr::summarise(n = sum(count), .groups = "drop") %>%
      tidyr::pivot_wider(names_from = sentiment, values_from = n, values_fill = 0)

    if ("positive" %in% names(doc_sentiment) && "negative" %in% names(doc_sentiment)) {
      doc_sentiment <- doc_sentiment %>%
        dplyr::mutate(
          total_sentiment_words = .data$positive + .data$negative,
          sentiment_score = ifelse(.data$total_sentiment_words > 0,
                                   (.data$positive - .data$negative) / .data$total_sentiment_words,
                                   0),
          sentiment = dplyr::case_when(
            .data$sentiment_score > 0 ~ "positive",
            .data$sentiment_score < 0 ~ "negative",
            TRUE ~ "neutral"
          )
        )
    }
  }

  emotion_data <- NULL
  if (lexicon_name == "nrc") {
    emotion_cols <- c("anger", "anticipation", "disgust", "fear", "joy", "sadness", "surprise", "trust")
    available_emotions <- intersect(emotion_cols, names(doc_sentiment))
    if (length(available_emotions) > 0) {
      emotion_data <- doc_sentiment %>%
        dplyr::select(document, dplyr::all_of(available_emotions)) %>%
        tidyr::pivot_longer(cols = -document, names_to = "emotion", values_to = "score") %>%
        dplyr::group_by(emotion) %>%
        dplyr::summarise(total_score = sum(score, na.rm = TRUE), .groups = "drop")
    }
  }

  total_docs_in_corpus <- length(doc_names)
  docs_analyzed <- dplyr::n_distinct(doc_sentiment$document)

  # token match rate
  total_tokens <- sum(tidy_dfm$count)
  matched_tokens <- tidy_dfm %>%
    dplyr::semi_join(sentiment_lexicon, by = c("term" = "word")) %>%
    dplyr::pull(count) %>%
    sum()

  summary_stats <- list(
    total_documents = total_docs_in_corpus,
    documents_analyzed = docs_analyzed,
    documents_without_sentiment = total_docs_in_corpus - docs_analyzed,
    coverage_percentage = round((docs_analyzed / total_docs_in_corpus) * 100, 1),
    token_match_rate = round(if (total_tokens > 0) matched_tokens / total_tokens else NA_real_, 3),
    positive_docs = sum(doc_sentiment$sentiment == "positive", na.rm = TRUE),
    negative_docs = sum(doc_sentiment$sentiment == "negative", na.rm = TRUE),
    neutral_docs = sum(doc_sentiment$sentiment == "neutral", na.rm = TRUE),
    avg_sentiment_score = mean(doc_sentiment$sentiment_score, na.rm = TRUE)
  )

  list(
    document_sentiment = doc_sentiment,
    emotion_scores = emotion_data,
    summary_stats = summary_stats,
    lexicon_used = lexicon_name,
    feature_type = feature_type
  )
}


#' Embedding-based Sentiment Analysis
#'
#' @description
#' Performs sentiment analysis using transformer-based embeddings and neural models.
#' This approach uses pre-trained language models for contextual sentiment detection
#' without requiring sentiment lexicons. Particularly effective for handling:
#' - Complex contextual sentiment
#' - Implicit sentiment and sarcasm
#' - Domain-specific sentiment
#' - Negation and intensifiers (automatically handled by the model)
#'
#' @param texts Character vector of texts to analyze
#' @param embeddings Optional pre-computed embedding matrix (from generate_embeddings)
#' @param model_name Sentiment model name (default: "distilbert-base-uncased-finetuned-sst-2-english")
#' @param doc_names Optional document names/IDs
#' @param use_gpu Whether to use GPU if available (default: FALSE)
#'
#' @return A list containing:
#'   \describe{
#'     \item{document_sentiment}{Data frame with document-level sentiment scores and classifications}
#'     \item{emotion_scores}{NULL (emotion detection not currently supported for embeddings)}
#'     \item{summary_stats}{Summary statistics including document counts and average scores}
#'     \item{model_used}{Name of the transformer model used}
#'     \item{feature_type}{"embeddings"}
#'   }
#'
#' @concept sentiment
#' @export
#'
#' @examples
#' if (interactive()) {
#'   abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
#'   embedding_sentiment <- sentiment_embedding_analysis(abstracts)
#'   print(embedding_sentiment$document_sentiment)
#'   print(embedding_sentiment$summary_stats)
#' }
sentiment_embedding_analysis <- function(texts,
                                        embeddings = NULL,
                                        model_name = "distilbert-base-uncased-finetuned-sst-2-english",
                                        doc_names = NULL,
                                        use_gpu = FALSE) {

  if (!requireNamespace("reticulate", quietly = TRUE)) {
    return(.notify_missing_python("Embedding-based sentiment analysis"))
  }

  if (is.null(doc_names)) {
    doc_names <- paste0("text", seq_along(texts))
  }

  if (length(texts) != length(doc_names)) {
    stop("Length of texts and doc_names must match")
  }

  tryCatch({
    transformers <- reticulate::import("transformers")

    message("Loading sentiment model: ", model_name)

    device <- if (use_gpu) 0L else -1L

    sentiment_pipeline <- transformers$pipeline(
      "sentiment-analysis",
      model = model_name,
      device = device
    )

    message("Analyzing sentiment for ", length(texts), " documents...")

    results <- sentiment_pipeline(texts, truncation = TRUE, max_length = 512L)

    doc_sentiment <- data.frame(
      document = doc_names,
      label = sapply(results, function(x) tolower(x$label)),
      confidence = sapply(results, function(x) x$score),
      stringsAsFactors = FALSE
    )

    doc_sentiment$sentiment_score <- ifelse(
      doc_sentiment$label == "positive",
      doc_sentiment$confidence,
      -doc_sentiment$confidence
    )

    doc_sentiment$sentiment <- doc_sentiment$label

    if ("neg" %in% doc_sentiment$label || "negative" %in% doc_sentiment$label) {
      doc_sentiment$sentiment <- ifelse(
        doc_sentiment$label %in% c("neg", "negative"),
        "negative",
        "positive"
      )
    }

    summary_stats <- list(
      total_documents = length(texts),
      documents_analyzed = nrow(doc_sentiment),
      documents_without_sentiment = 0,
      coverage_percentage = 100,
      positive_docs = sum(doc_sentiment$sentiment == "positive", na.rm = TRUE),
      negative_docs = sum(doc_sentiment$sentiment == "negative", na.rm = TRUE),
      neutral_docs = sum(doc_sentiment$sentiment == "neutral", na.rm = TRUE),
      avg_sentiment_score = mean(doc_sentiment$sentiment_score, na.rm = TRUE),
      avg_confidence = mean(doc_sentiment$confidence, na.rm = TRUE)
    )

    message("Sentiment analysis completed successfully")

    return(list(
      document_sentiment = doc_sentiment,
      emotion_scores = NULL,
      summary_stats = summary_stats,
      model_used = model_name,
      feature_type = "embeddings"
    ))

  }, error = function(e) {
    stop(
      "Error in embedding-based sentiment analysis: ", e$message, "\n",
      "Please ensure Python transformers library is installed:\n",
      "  reticulate::py_install('transformers')\n",
      "  reticulate::py_install('torch')"
    )
  })
}


#' LLM-based Sentiment Analysis
#'
#' @description
#' Analyzes sentiment using Large Language Models (OpenAI or Gemini).
#' Provides nuanced sentiment understanding including sarcasm detection,
#' mixed emotions, and contextual interpretation that lexicon-based methods miss.
#'
#' Requests use temperature 0, which makes outputs near-deterministic but
#' not fully reproducible across runs or model versions; record the returned
#' \code{model_used} field when reporting results.
#'
#' @param texts Character vector of texts to analyze.
#' @param doc_names Optional character vector of document names (default: text1, text2, ...).
#' @param provider AI provider to use: "openai" (default) or "gemini".
#' @param model Model name. If NULL, uses provider defaults: "gpt-4.1-mini" (OpenAI),
#'   "gemini-2.5-flash-lite" (Gemini).
#' @param api_key API key for OpenAI or Gemini. If NULL, uses environment variable.
#' @param batch_size Number of texts to process per API call (default: 5).
#'   Larger batches are more efficient but may hit token limits.
#' @param include_explanation Logical, if TRUE includes natural language explanation
#'   for each sentiment classification (default: FALSE).
#' @param verbose Logical, if TRUE prints progress messages (default: TRUE).
#'
#' @return A list containing:
#' \describe{
#'   \item{document_sentiment}{Data frame with document-level sentiment scores}
#'   \item{summary_stats}{Summary statistics of the analysis}
#'   \item{model_used}{Model name used for analysis}
#'   \item{provider}{AI provider used}
#' }
#'
#' @details
#' LLM-based sentiment analysis offers several advantages over lexicon methods:
#' \itemize{
#'   \item Understands context and nuance
#'   \item Detects sarcasm and irony
#'   \item Handles mixed emotions
#'   \item Works across domains without retraining
#' }
#'
#' @concept sentiment
#' @export
#'
#' @examples
#' if (interactive()) {
#'   abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:5]
#'
#'   sentiment_openai <- analyze_sentiment_llm(abstracts, provider = "openai")
#'
#'   sentiment_gemini <- analyze_sentiment_llm(abstracts, provider = "gemini",
#'                                              include_explanation = TRUE)
#' }
analyze_sentiment_llm <- function(texts,
                                  doc_names = NULL,
                                  provider = c("openai", "gemini"),
                                  model = NULL,
                                  api_key = NULL,
                                  batch_size = 5,
                                  include_explanation = FALSE,
                                  verbose = TRUE) {

  provider <- match.arg(provider)

  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop("Package 'jsonlite' is required. Install with: install.packages('jsonlite')")
  }

  # Set document names if not provided
  doc_names <- doc_names %||% paste0("text", seq_along(texts))
  if (length(texts) != length(doc_names)) {
    stop("Length of texts and doc_names must match")
  }

  # Resolve provider, model, and API key
  setup <- .resolve_llm_setup(
    provider, model, api_key,
    defaults = list(openai = "gpt-4.1-mini", gemini = "gemini-2.5-flash-lite")
  )
  if (is.null(setup)) return(invisible(NULL))
  model <- setup$model
  api_key <- setup$api_key

  if (verbose) {
    message(sprintf("Analyzing sentiment for %d texts using %s (%s)...",
                    length(texts), provider, model))
  }

  # System prompt for sentiment analysis
  system_prompt <- "You are an expert sentiment analyst. Analyze the sentiment of each text and respond in valid JSON format only.

For each text, provide:
- sentiment: 'positive', 'negative', or 'neutral'
- score: a number from -1.0 (very negative) to 1.0 (very positive)
- confidence: a number from 0.0 to 1.0 indicating your confidence
"

  if (include_explanation) {
    system_prompt <- paste0(system_prompt, "- explanation: a brief explanation of why you classified it this way\n")
  }

  system_prompt <- paste0(system_prompt, "
Respond with a JSON array. Example:
[{\"sentiment\": \"positive\", \"score\": 0.8, \"confidence\": 0.95}]

Important:
- Detect sarcasm and irony (e.g., 'Oh great, another meeting' is likely negative)
- Consider context and nuance
- Respond ONLY with valid JSON, no additional text")

  # Process texts in batches
  results <- .empty_sentiment_row(character(0))

  n_batches <- ceiling(length(texts) / batch_size)

  for (i in seq_len(n_batches)) {
    start_idx <- (i - 1) * batch_size + 1
    end_idx <- min(i * batch_size, length(texts))
    batch_texts <- texts[start_idx:end_idx]
    batch_names <- doc_names[start_idx:end_idx]

    if (verbose) {
      message(sprintf("Processing batch %d/%d...", i, n_batches))
    }

    # Build user prompt with numbered texts
    user_prompt <- paste0(
      "Analyze the sentiment of these texts:\n\n",
      paste(sprintf("Text %d: %s\n", seq_along(batch_texts), batch_texts), collapse = "\n")
    )

    # Call LLM
    response <- tryCatch(
      call_llm_api(
        provider = provider, system_prompt = system_prompt, user_prompt = user_prompt,
        model = model, temperature = 0, max_tokens = 500, api_key = api_key
      ),
      error = function(e) {
        warning(sprintf("Batch %d failed: %s", i, e$message))
        NULL
      }
    )

    if (is.null(response)) {
      # Failed batch: NA rows
      results <- rbind(results, .empty_sentiment_row(batch_names))
      next
    }

    # Parse JSON response
    parsed <- tryCatch({
      json_str <- response
      start_bracket <- regexpr("\\[", json_str)
      end_bracket <- regexpr("\\](?=[^\\]]*$)", json_str, perl = TRUE)
      if (start_bracket > 0 && end_bracket > 0) {
        json_str <- substr(json_str, start_bracket, end_bracket + attr(end_bracket, "match.length") - 1)
      }
      jsonlite::fromJSON(json_str)
    }, error = function(e) {
      warning(sprintf("Failed to parse JSON response for batch %d: %s", i, e$message))
      NULL
    })

    if (is.null(parsed) || length(parsed) == 0) {
      results <- rbind(results, .empty_sentiment_row(batch_names))
      next
    }

    # Map parsed results to batch
    for (j in seq_along(batch_texts)) {
      if (j <= nrow(parsed)) {
        row <- parsed[j, ]
        results <- rbind(results, data.frame(
          document = batch_names[j],
          sentiment = as.character(row$sentiment),
          sentiment_score = as.numeric(row$score),
          confidence = as.numeric(row$confidence),
          explanation = if (include_explanation && "explanation" %in% names(row)) as.character(row$explanation) else NA_character_,
          stringsAsFactors = FALSE
        ))
      } else {
        results <- rbind(results, .empty_sentiment_row(batch_names[j]))
      }
    }

    # Rate limiting
    Sys.sleep(1)
  }

  # Remove explanation column if not requested
  if (!include_explanation) {
    results$explanation <- NULL
  }

  # Summary statistics
  valid_results <- results[!is.na(results$sentiment), ]
  summary_stats <- list(
    total_documents = length(texts),
    documents_analyzed = nrow(valid_results),
    documents_without_sentiment = sum(is.na(results$sentiment)),
    coverage_percentage = round(nrow(valid_results) / length(texts) * 100, 1),
    positive_docs = sum(valid_results$sentiment == "positive", na.rm = TRUE),
    negative_docs = sum(valid_results$sentiment == "negative", na.rm = TRUE),
    neutral_docs = sum(valid_results$sentiment == "neutral", na.rm = TRUE),
    avg_sentiment_score = mean(valid_results$sentiment_score, na.rm = TRUE),
    avg_confidence = mean(valid_results$confidence, na.rm = TRUE)
  )

  if (verbose) {
    message(sprintf("Sentiment analysis completed. Analyzed %d/%d documents.",
                    summary_stats$documents_analyzed, summary_stats$total_documents))
    message(sprintf("Distribution: %d positive, %d negative, %d neutral",
                    summary_stats$positive_docs, summary_stats$negative_docs,
                    summary_stats$neutral_docs))
  }

  return(list(
    document_sentiment = results,
    summary_stats = summary_stats,
    model_used = model,
    provider = provider,
    feature_type = "llm"
  ))
}


#' Plot Emotion Radar Chart
#'
#' @description
#' Creates a polar/radar chart for NRC emotion analysis with optional grouping.
#'
#' @param emotion_data Data frame with emotion scores (columns: emotion, total_score)
#' @param group_var Optional grouping variable column name for overlaid radars (default: NULL)
#' @param normalize Logical, whether to normalize scores to 0-100 scale (default: FALSE)
#' @param title Plot title (default: "Emotion Analysis")
#'
#' @return A ggplot2 polar chart
#'
#' @concept sentiment
#' @export
plot_emotion_radar <- function(emotion_data,
                               group_var = NULL,
                               normalize = FALSE,
                               title = "Emotion Analysis") {

  if (!is.null(group_var) && group_var %in% names(emotion_data)) {

    plot_data <- emotion_data

    if (normalize) {
      plot_data <- plot_data %>%
        dplyr::group_by(!!rlang::sym(group_var)) %>%
        dplyr::mutate(
          total_score = if (max(total_score, na.rm = TRUE) > 0) {
            (total_score / max(total_score, na.rm = TRUE)) * 100
          } else {
            total_score
          }
        ) %>%
        dplyr::ungroup()
    }

    categories <- unique(plot_data[[group_var]])

    if (length(categories) == 0) {
      return(
        ggplot2::ggplot() +
          ggplot2::annotate("text", x = 0.5, y = 0.5, label = "No data available", size = 5, color = "#ef4444") +
          ggplot2::theme_void()
      )
    }

    plot_data$group_col <- plot_data[[group_var]]
    plot_data$hover_text <- paste("Emotion:", plot_data$emotion,
                                  "<br>Score:", round(plot_data$total_score, 2),
                                  paste0("<br>", group_var, ":"), plot_data$group_col)

    ggplot2::ggplot(plot_data, ggplot2::aes(x = .data$emotion, y = .data$total_score,
                                             group = .data$group_col, color = .data$group_col, fill = .data$group_col,
                                             text = .data$hover_text)) +
      ggplot2::geom_polygon(alpha = 0.1) +
      ggplot2::geom_point(size = 2) +
      ggplot2::coord_polar() +
      ggplot2::labs(title = title, x = NULL, y = NULL, color = group_var, fill = group_var) +
      ggplot2::theme_minimal(base_size = 11)

  } else {

    scores <- emotion_data$total_score

    if (normalize) {
      max_score <- max(scores, na.rm = TRUE)
      if (max_score > 0) {
        scores <- (scores / max_score) * 100
      }
    }

    radar_df <- data.frame(
      emotion = emotion_data$emotion,
      score = scores
    )

    radar_df$hover_text <- paste("Emotion:", radar_df$emotion, "<br>Score:", round(radar_df$score, 2))

    ggplot2::ggplot(radar_df, ggplot2::aes(x = .data$emotion, y = .data$score, group = 1, text = .data$hover_text)) +
      ggplot2::geom_polygon(fill = "#8B5CF6", alpha = 0.1, color = "#8B5CF6") +
      ggplot2::geom_point(color = "#8B5CF6", size = 2) +
      ggplot2::coord_polar() +
      ggplot2::labs(title = title, x = NULL, y = NULL) +
      ggplot2::theme_minimal(base_size = 11)
  }
}


#' Plot Sentiment Box Plot by Category
#'
#' @description
#' Creates a box plot showing sentiment score distribution by category.
#'
#' @param sentiment_data Data frame from analyze_sentiment() containing sentiment_score
#'   and category columns
#' @param category_var Name of the category variable column (default: "category_var")
#' @param title Plot title (default: "Sentiment Score Distribution")
#'
#' @return A ggplot2 box plot
#'
#' @concept sentiment
#' @export
plot_sentiment_boxplot <- function(sentiment_data,
                                   category_var = "category_var",
                                   title = "Sentiment Score Distribution") {

  if (!category_var %in% names(sentiment_data)) {
    stop("Category variable '", category_var, "' not found in data")
  }

  if (!"sentiment_score" %in% names(sentiment_data)) {
    stop("sentiment_score column not found in data")
  }

  sentiment_data$hover_text <- paste(category_var, ":", sentiment_data[[category_var]],
                                     "<br>Score:", round(sentiment_data$sentiment_score, 3))

  ggplot2::ggplot(sentiment_data, ggplot2::aes(x = .data[[category_var]], y = .data$sentiment_score,
                                                fill = .data[[category_var]], text = .data$hover_text)) +
    ggplot2::geom_boxplot(show.legend = FALSE) +
    ggplot2::labs(title = title, x = category_var, y = "Sentiment Score") +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
}


#' Plot Sentiment Violin Plot by Category
#'
#' @description
#' Creates a violin plot showing sentiment score distribution by category.
#'
#' @param sentiment_data Data frame from analyze_sentiment() containing sentiment_score
#'   and category columns
#' @param category_var Name of the category variable column (default: "category_var")
#' @param title Plot title (default: "Sentiment Score Distribution")
#'
#' @return A ggplot2 violin plot
#'
#' @concept sentiment
#' @export
plot_sentiment_violin <- function(sentiment_data,
                                  category_var = "category_var",
                                  title = "Sentiment Score Distribution") {

  if (!category_var %in% names(sentiment_data)) {
    stop("Category variable '", category_var, "' not found in data")
  }

  if (!"sentiment_score" %in% names(sentiment_data)) {
    stop("sentiment_score column not found in data")
  }

  sentiment_data$hover_text <- paste(category_var, ":", sentiment_data[[category_var]],
                                     "<br>Score:", round(sentiment_data$sentiment_score, 3))

  ggplot2::ggplot(sentiment_data, ggplot2::aes(x = .data[[category_var]], y = .data$sentiment_score,
                                                fill = .data[[category_var]], text = .data$hover_text)) +
    ggplot2::geom_violin(show.legend = FALSE) +
    ggplot2::labs(title = title, x = category_var, y = "Sentiment Score") +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
}


# Semantic network analysis (co-occurrence and correlation)

#' @importFrom utils modifyList
#' @importFrom stats cor
#' @importFrom igraph graph_from_data_frame V vcount degree betweenness closeness eigen_centrality layout_with_fr
#' @importFrom dplyr count filter mutate select group_by summarise ungroup
#' @importFrom tibble as_tibble
#' @importFrom tidytext tidy
#' @importFrom widyr pairwise_count pairwise_cor
#' @importFrom stats quantile
#' @importFrom shiny showNotification
#' @importFrom rlang sym %||%
#' @importFrom utils head
#' @importFrom grDevices colorRampPalette
NULL

# Network Analysis Functions

#' @title Plot Semantic Analysis Visualization
#'
#' @description
#' Creates interactive visualizations for semantic analysis results including
#' similarity heatmaps, dimensionality reduction plots, and clustering visualizations.
#'
#' @param analysis_result A list containing semantic analysis results from functions like
#'   semantic_similarity_analysis(), semantic_document_clustering(), or reduce_dimensions().
#' @param plot_type Type of visualization: "similarity" for heatmap, "dimensionality_reduction"
#'   for scatter plot, or "clustering" for cluster visualization (default: "similarity").
#' @param data_labels Optional character vector of labels for data points (default: NULL).
#' @param color_by Optional variable to color points by in scatter plots (default: NULL).
#' @param coords Optional pre-computed coordinates for dimensionality reduction plots (default: NULL).
#' @param clusters Optional cluster assignments vector (default: NULL).
#' @param hover_text Optional custom hover text for points (default: NULL).
#' @param hover_config Optional hover configuration list (default: NULL).
#' @param cluster_colors Optional color palette for clusters (default: NULL).
#' @param height The height of the resulting Plotly plot, in pixels (default: 600).
#' @param width The width of the resulting Plotly plot, in pixels (default: 800).
#' @param title Optional custom title for the plot (default: NULL).
#'
#' @return A ggplot2 object showing the specified visualization.
#'
#' @concept visualization
#' @export
#'
#' @examples
#' \donttest{
#'   data(SpecialEduTech)
#'   texts <- SpecialEduTech$abstract[1:5]
#'   result <- semantic_similarity_analysis(texts)
#'   plot <- plot_semantic_viz(result, plot_type = "similarity")
#'   print(plot)
#' }
plot_semantic_viz <- function(analysis_result = NULL,
                                       plot_type = "similarity",
                                       data_labels = NULL,
                                       color_by = NULL,
                                       height = 600,
                                       width = 800,
                                       title = NULL,
                                       coords = NULL,
                                       clusters = NULL,
                                       hover_text = NULL,
                                       hover_config = NULL,
                                       cluster_colors = NULL) {

  tryCatch({
    plot_obj <- switch(plot_type,
      "similarity" = {
        similarity_matrix <- analysis_result$similarity_matrix

        if (is.null(data_labels)) {
          data_labels <- paste0("Doc ", seq_len(nrow(similarity_matrix)))
        }

        plot_title <- title %||% paste("Similarity Heatmap -", analysis_result$method)

        heat_df <- expand.grid(
          x = seq_len(ncol(similarity_matrix)),
          y = seq_len(nrow(similarity_matrix))
        )
        heat_df$similarity <- as.vector(similarity_matrix)
        heat_df$x_label <- factor(data_labels[heat_df$x], levels = data_labels)
        heat_df$y_label <- factor(data_labels[heat_df$y], levels = data_labels)

        heat_df$hover_text <- paste("X:", heat_df$x_label, "<br>Y:", heat_df$y_label,
                                    "<br>Similarity:", round(heat_df$similarity, 3))

        ggplot2::ggplot(heat_df, ggplot2::aes(x = .data$x_label, y = .data$y_label, fill = .data$similarity,
                                               text = .data$hover_text)) +
          ggplot2::geom_tile() +
          ggplot2::scale_fill_viridis_c(name = "Similarity") +
          ggplot2::labs(title = plot_title, x = "Documents", y = "Documents") +
          ggplot2::theme_minimal(base_size = 11) +
          ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
      },
      "dimensionality_reduction" = {
        reduced_data <- if (!is.null(coords)) {
          coords
        } else if (!is.null(analysis_result$reduced_data)) {
          analysis_result$reduced_data
        } else {
          stop("No dimensionality reduction data available")
        }

        if (is.null(data_labels)) {
          data_labels <- paste0("Doc ", seq_len(nrow(reduced_data)))
        }

        plot_clusters <- !is.null(clusters) || (!is.null(analysis_result) && !is.null(analysis_result$clusters))
        cluster_data <- clusters %||% (if (!is.null(analysis_result)) analysis_result$clusters else NULL)

        scatter_df <- data.frame(
          x = reduced_data[, 1],
          y = if (ncol(reduced_data) > 1) reduced_data[, 2] else rep(0, nrow(reduced_data)),
          label = data_labels
        )

        if (plot_clusters) {
          scatter_df$color_var <- as.factor(cluster_data)
        } else if (!is.null(color_by)) {
          scatter_df$color_var <- color_by
        }

        plot_title <- title %||% paste("Dimensionality Reduction -",
                                        if (!is.null(analysis_result)) analysis_result$method else "Custom")

        x_title <- paste("Component 1",
                          if (!is.null(analysis_result) && !is.null(analysis_result$variance_explained))
                            paste0("(", round(analysis_result$variance_explained[1] * 100, 1), "%)")
                          else "")
        y_title <- paste("Component 2",
                          if (!is.null(analysis_result) && !is.null(analysis_result$variance_explained) &&
                              length(analysis_result$variance_explained) > 1)
                            paste0("(", round(analysis_result$variance_explained[2] * 100, 1), "%)")
                          else "")

        scatter_df$hover_text <- paste("Document:", scatter_df$label,
                                       "<br>X:", round(scatter_df$x, 3),
                                       "<br>Y:", round(scatter_df$y, 3))

        p <- ggplot2::ggplot(scatter_df, ggplot2::aes(x = .data$x, y = .data$y, text = .data$hover_text))

        if ("color_var" %in% names(scatter_df)) {
          p <- p + ggplot2::geom_point(ggplot2::aes(color = .data$color_var), size = 2, alpha = 0.7)
        } else {
          p <- p + ggplot2::geom_point(color = "steelblue", size = 2, alpha = 0.7)
        }

        p + ggplot2::labs(title = plot_title, x = x_title, y = y_title, color = NULL) +
          ggplot2::theme_minimal(base_size = 11)
      },
      "clustering" = {
        plot_data_mat <- if (!is.null(coords)) {
          coords
        } else if (!is.null(analysis_result)) {
          analysis_result$umap_embedding %||% analysis_result$reduced_data
        } else {
          stop("No clustering visualization data available")
        }

        cluster_data <- clusters %||% (if (!is.null(analysis_result)) analysis_result$clusters else NULL)

        plot_title <- title %||% paste("Clustering Results -",
                                        if (!is.null(analysis_result)) analysis_result$method else "Custom")

        if (is.null(plot_data_mat)) {
          if (is.null(data_labels)) {
            data_labels <- paste0("Doc ", seq_along(cluster_data))
          }

          scatter_df <- data.frame(
            x = seq_along(cluster_data),
            y = cluster_data,
            cluster = as.factor(cluster_data),
            label = data_labels
          )
          scatter_df$hover_text <- paste("Document:", scatter_df$label, "<br>Cluster:", scatter_df$cluster)

          ggplot2::ggplot(scatter_df, ggplot2::aes(x = .data$x, y = .data$y, color = .data$cluster, text = .data$hover_text)) +
            ggplot2::geom_point(size = 2, alpha = 0.7) +
            ggplot2::labs(title = plot_title, x = "Document Index", y = "Cluster", color = "Cluster") +
            ggplot2::theme_minimal(base_size = 11)
        } else {
          if (is.null(data_labels)) {
            data_labels <- paste0("Doc ", seq_len(nrow(plot_data_mat)))
          }

          scatter_df <- data.frame(
            x = plot_data_mat[, 1],
            y = if (ncol(plot_data_mat) > 1) plot_data_mat[, 2] else rep(0, nrow(plot_data_mat)),
            cluster = as.factor(cluster_data),
            label = data_labels
          )
          scatter_df$hover_text <- paste("Document:", scatter_df$label, "<br>Cluster:", scatter_df$cluster)

          ggplot2::ggplot(scatter_df, ggplot2::aes(x = .data$x, y = .data$y, color = .data$cluster, text = .data$hover_text)) +
            ggplot2::geom_point(size = 2, alpha = 0.7) +
            ggplot2::labs(title = plot_title, x = "Component 1", y = "Component 2", color = "Cluster") +
            ggplot2::theme_minimal(base_size = 11)
        }
      },
      stop("Unsupported plot type: ", plot_type)
    )

    return(plot_obj)

  }, error = function(e) {
    stop("Error creating semantic visualization: ", e$message)
  })
}


#' Plot Cross-Category Similarity Comparison
#'
#' @description
#' Creates a faceted ggplot heatmap for cross-category document similarity
#' comparison. Accepts either a pre-built long-format data frame or extracts
#' from a similarity matrix.
#'
#' @param similarity_data Either a similarity matrix (square numeric matrix) or
#'   a data frame in long format with columns for row labels, column labels,
#'   similarity values, and category.
#' @param docs_data Data frame with document metadata (required if similarity_data is a matrix)
#' @param row_var Column name for row document labels (default: "ld_doc_name")
#' @param col_var Column name for column document labels (default: "other_doc_name")
#' @param value_var Column name for similarity values (default: "cosine_similarity")
#' @param category_var Column name for category in long-format data or docs_data (default: "other_category")
#' @param row_category Category for row documents (used with matrix input)
#' @param col_categories Categories for column documents (used with matrix input)
#' @param row_display_var Column name for row display labels in tooltip (default: NULL, uses row_var)
#' @param col_display_var Column name for column display labels in tooltip (default: NULL, uses col_var)
#' @param method_name Similarity method name for legend (default: "Cosine")
#' @param title Plot title (default: NULL)
#' @param show_values Logical; show similarity values as text on tiles (default: TRUE)
#' @param row_label Label for y-axis (default: "Documents")
#' @param label_max_chars Maximum characters for axis labels before truncation (default: 25)
#' @param order_by_numeric Logical; order by numeric ID extracted from labels (default: TRUE)
#' @param height Plot height (default: 600)
#' @param width Plot width (default: NULL)
#'
#' @return A ggplot object
#'
#' @concept visualization
#' @export
#'
#' @examples
#' \donttest{
#' articles <- TextAnalysisR::SpecialEduTech[1:7, ]
#' term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(articles$abstract)))
#' normalized_matrix <- term_matrix / sqrt(rowSums(term_matrix ^ 2))
#' similarity_matrix <- normalized_matrix %*% t(normalized_matrix)
#' thesis_rows <- which(articles$reference_type == "thesis")[1:3]
#' article_cols <- which(articles$reference_type == "journal_article")[1:4]
#' similarity_data <- expand.grid(
#'   ld_doc_name    = paste("Thesis", seq_along(thesis_rows)),
#'   other_doc_name = paste("Article", seq_along(article_cols)),
#'   stringsAsFactors = FALSE
#' )
#' similarity_data$cosine_similarity <- as.vector(
#'   similarity_matrix[thesis_rows, article_cols]
#' )
#' similarity_data$other_category <- articles$reference_type[article_cols]
#' plot_cross_category_heatmap(
#'   similarity_data = similarity_data,
#'   row_var = "ld_doc_name",
#'   col_var = "other_doc_name",
#'   value_var = "cosine_similarity",
#'   category_var = "other_category",
#'   row_label = "Theses"
#' )
#' }
plot_cross_category_heatmap <- function(similarity_data,
                                         docs_data = NULL,
                                         row_var = "ld_doc_name",
                                         col_var = "other_doc_name",
                                         value_var = "cosine_similarity",
                                         category_var = "other_category",
                                         row_category = NULL,
                                         col_categories = NULL,
                                         row_display_var = NULL,
                                         col_display_var = NULL,
                                         method_name = "Cosine",
                                         title = NULL,
                                         show_values = TRUE,
                                         row_label = "Documents",
                                         label_max_chars = 25,
                                         order_by_numeric = TRUE,
                                         height = 600,
                                         width = NULL) {

  if (!requireNamespace("stringr", quietly = TRUE)) {
    stop("Package 'stringr' is required. Install with: install.packages('stringr')")
  }

  # Branch on input type: long-format data frame or matrix
  if (is.data.frame(similarity_data)) {
    plot_data <- similarity_data

    # Validate required columns
    required_cols <- c(row_var, col_var, value_var, category_var)
    missing_cols <- setdiff(required_cols, names(plot_data))
    if (length(missing_cols) > 0) {
      stop("Missing required columns: ", paste(missing_cols, collapse = ", "))
    }

    # Rename to internal names
    plot_data <- plot_data %>%
      dplyr::rename(
        row_doc = !!rlang::sym(row_var),
        col_doc = !!rlang::sym(col_var),
        similarity = !!rlang::sym(value_var),
        col_category = !!rlang::sym(category_var)
      )

    # Display columns for tooltips
    if (!is.null(row_display_var) && row_display_var %in% names(similarity_data)) {
      plot_data$row_display <- similarity_data[[row_display_var]]
    } else {
      plot_data$row_display <- plot_data$row_doc
    }

    if (!is.null(col_display_var) && col_display_var %in% names(similarity_data)) {
      plot_data$col_display <- similarity_data[[col_display_var]]
    } else {
      plot_data$col_display <- plot_data$col_doc
    }

    # Truncated labels
    plot_data <- plot_data %>%
      dplyr::mutate(
        row_label_trunc = stringr::str_trunc(.data$row_doc, label_max_chars),
        col_label_trunc = stringr::str_trunc(.data$col_doc, label_max_chars)
      )

    # Order by numeric ID if requested
    if (order_by_numeric) {
      plot_data <- plot_data %>%
        dplyr::mutate(
          row_numeric_id = as.numeric(stringr::str_extract(.data$row_doc, "\\d+")),
          col_numeric_id = as.numeric(stringr::str_extract(.data$col_doc, "\\d+"))
        )

      row_order <- plot_data %>%
        dplyr::arrange(.data$row_numeric_id) %>%
        dplyr::pull(.data$row_label_trunc) %>%
        unique()

      col_order <- plot_data %>%
        dplyr::arrange(.data$col_numeric_id) %>%
        dplyr::pull(.data$col_label_trunc) %>%
        unique()
    } else {
      row_order <- unique(plot_data$row_label_trunc)
      col_order <- unique(plot_data$col_label_trunc)
    }

    # Build final plot data with factor levels and tooltips
    cat_levels <- unique(plot_data$col_category)

    plot_data <- plot_data %>%
      dplyr::mutate(
        row_label_trunc = factor(.data$row_label_trunc, levels = rev(row_order)),
        col_label_trunc = factor(.data$col_label_trunc, levels = col_order),
        col_category = factor(.data$col_category, levels = cat_levels),
        tooltip_text = paste0(
          row_label, ": ", dplyr::coalesce(as.character(.data$row_display), as.character(.data$row_doc)),
          "<br>", .data$col_category, ": ", dplyr::coalesce(as.character(.data$col_display), as.character(.data$col_doc)),
          "<br>", method_name, " Similarity: ", round(.data$similarity, 3)
        )
      )

  } else if (is.matrix(similarity_data)) {
    # Matrix input: extract cross-category data
    if (is.null(docs_data) || is.null(row_category) || is.null(col_categories)) {
      stop("For matrix input, docs_data, row_category, and col_categories are required")
    }

    if (!category_var %in% names(docs_data)) {
      stop("category_var '", category_var, "' not found in docs_data")
    }

    row_indices <- which(docs_data[[category_var]] == row_category)
    if (length(row_indices) == 0) {
      return(create_empty_plot_message(paste("No documents found for category:", row_category)))
    }

    row_docs <- docs_data[row_indices, ]
    row_short_labels <- row_docs$document_number %||% paste("Doc", row_indices)
    row_full_ids <- row_docs$document_id_display %||% row_short_labels

    plot_data_list <- list()

    for (col_cat in col_categories) {
      col_indices <- which(docs_data[[category_var]] == col_cat)
      if (length(col_indices) == 0) next

      col_docs <- docs_data[col_indices, ]
      col_short_labels <- col_docs$document_number %||% paste("Doc", col_indices)
      col_full_ids <- col_docs$document_id_display %||% col_short_labels

      sub_matrix <- similarity_data[row_indices, col_indices, drop = FALSE]

      for (i in seq_along(row_indices)) {
        for (j in seq_along(col_indices)) {
          plot_data_list[[length(plot_data_list) + 1]] <- data.frame(
            row_label_trunc = row_short_labels[i],
            col_label_trunc = col_short_labels[j],
            row_display = row_full_ids[i],
            col_display = col_full_ids[j],
            similarity = sub_matrix[i, j],
            col_category = col_cat,
            stringsAsFactors = FALSE
          )
        }
      }
    }

    if (length(plot_data_list) == 0) {
      return(create_empty_plot_message("No matching documents found for specified categories"))
    }

    plot_data <- do.call(rbind, plot_data_list)

    row_order <- unique(plot_data$row_label_trunc)
    col_order <- unique(plot_data$col_label_trunc)

    plot_data <- plot_data %>%
      dplyr::mutate(
        row_label_trunc = factor(.data$row_label_trunc, levels = rev(row_order)),
        col_label_trunc = factor(.data$col_label_trunc, levels = col_order),
        col_category = factor(.data$col_category, levels = col_categories),
        tooltip_text = paste0(
          as.character(.data$row_label_trunc), ": ", .data$row_display,
          "<br>", as.character(.data$col_label_trunc), ": ", .data$col_display,
          "<br>Category: ", row_category, " / ", .data$col_category,
          "<br>", method_name, " Similarity: ", round(.data$similarity, 3)
        )
      )

    if (is.null(row_label) || row_label == "Documents") {
      row_label <- paste(row_category, "Documents")
    }

  } else {
    stop("similarity_data must be a data frame or matrix")
  }

  # Build the plot
  p <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(x = .data$col_label_trunc, y = .data$row_label_trunc, fill = .data$similarity, text = .data$tooltip_text)
  ) +
    ggplot2::geom_tile(color = "white", linewidth = 0.1)

  if (show_values) {
    q75 <- stats::quantile(plot_data$similarity, 0.75, na.rm = TRUE)
    p <- p + ggplot2::geom_text(
      ggplot2::aes(
        label = round(.data$similarity, 2),
        color = ifelse(.data$similarity > q75, "black", "white")
      ),
      size = 3.5,
      fontface = "bold",
      show.legend = FALSE
    ) +
      ggplot2::scale_color_identity()
  }

  p <- p +
    ggplot2::scale_fill_viridis_c(name = paste0(method_name, "\nSimilarity")) +
    ggplot2::facet_wrap(~ col_category, scales = "free_x") +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      strip.text.x = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      axis.text.x = ggplot2::element_text(angle = 45, hjust = 1, size = 11, color = "#3B3B3B"),
      axis.text.y = ggplot2::element_text(size = 11, color = "#3B3B3B"),
      axis.title.x = ggplot2::element_blank(),
      legend.title = ggplot2::element_text(size = 12, color = "#0c1f4a"),
      legend.text = ggplot2::element_text(size = 12, color = "#3B3B3B"),
      plot.title = ggplot2::element_text(size = 13, color = "#0c1f4a", hjust = 0.5)
    ) +
    ggplot2::labs(y = row_label, title = title)

  return(p)
}


#' Plot Document Similarity Heatmap
#'
#' @description
#' Creates an interactive heatmap visualization of document similarity matrices
#' with support for document metadata, feature-specific colorscales, and rich tooltips.
#' Supports both symmetric (all-vs-all) and cross-category comparison modes.
#'
#' @param similarity_matrix A square numeric matrix of similarity scores
#' @param docs_data Optional data frame with document metadata containing:
#'   \itemize{
#'     \item \code{document_number}: Document identifiers for axis labels
#'     \item \code{document_id_display}: Document IDs for hover text
#'     \item \code{category_display}: Category labels for hover text
#'   }
#' @param feature_type Feature space type: "words", "topics", "ngrams", or "embeddings"
#'   (determines colorscale and display name)
#' @param method_name Similarity method name for display (default: "Cosine")
#' @param title Plot title (default: NULL, auto-generated from feature_type)
#' @param category_filter Optional category filter label for title (default: NULL)
#' @param doc_id_var Name of document ID variable (affects label text, default: NULL)
#' @param colorscale Plotly colorscale override (default: NULL, uses feature_type default)
#' @param height Plot height in pixels (default: 600)
#' @param width Plot width in pixels (default: NULL for auto)
#' @param row_category Category for row documents in cross-category mode (default: NULL)
#' @param col_categories Character vector of categories for column documents (default: NULL)
#' @param category_var Name of category variable in docs_data (default: "category_display")
#' @param show_values Logical; show similarity values as text on tiles (default: FALSE)
#' @param facet Logical; facet by column categories (default: TRUE when col_categories specified)
#' @param row_label Label for row axis (default: NULL, uses row_category)
#' @param output_type Output type: "plotly" or "ggplot" (default: "plotly", auto-switches to "ggplot" for faceting)
#'
#' @return A ggplot2 heatmap object
#'
#' @concept visualization
#' @export
#'
#' @examples
#' \donttest{
#' articles <- TextAnalysisR::SpecialEduTech[1:5, ]
#' term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(articles$abstract)))
#' normalized_matrix <- term_matrix / sqrt(rowSums(term_matrix ^ 2))
#' similarity_matrix <- normalized_matrix %*% t(normalized_matrix)
#' plot_similarity_heatmap(similarity_matrix)
#'
#' document_metadata <- data.frame(
#'   document_number     = paste("Doc", 1:5),
#'   document_id_display = articles$title,
#'   category_display    = articles$reference_type
#' )
#' plot_similarity_heatmap(similarity_matrix, docs_data = document_metadata,
#'                         feature_type = "embeddings")
#'
#' plot_similarity_heatmap(
#'   similarity_matrix,
#'   docs_data      = document_metadata,
#'   row_category   = "thesis",
#'   col_categories = "journal_article",
#'   show_values    = TRUE,
#'   facet          = TRUE
#' )
#' }
plot_similarity_heatmap <- function(similarity_matrix,
                                     docs_data = NULL,
                                     feature_type = "words",
                                     method_name = "Cosine",
                                     title = NULL,
                                     category_filter = NULL,
                                     doc_id_var = NULL,
                                     colorscale = NULL,
                                     height = 600,
                                     width = NULL,
                                     row_category = NULL,
                                     col_categories = NULL,
                                     category_var = "category_display",
                                     show_values = FALSE,
                                     facet = NULL,
                                     row_label = NULL,
                                     output_type = "plotly") {

  if (is.null(similarity_matrix) || nrow(similarity_matrix) < 2) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5,
                          label = "Need at least 2 documents for similarity analysis",
                          size = 5, color = "#ef4444") +
        ggplot2::theme_void()
    )
  }

  if (!is.null(row_category) && !is.null(col_categories) && !is.null(docs_data)) {
    if (is.null(facet)) facet <- TRUE
    if (facet || output_type == "ggplot") {
      return(plot_cross_category_heatmap(
        similarity_data = similarity_matrix,
        docs_data = docs_data,
        row_category = row_category,
        col_categories = col_categories,
        category_var = category_var,
        method_name = method_name,
        title = title,
        show_values = show_values,
        row_label = row_label,
        height = height,
        width = width
      ))
    }
  }

  n_docs <- nrow(similarity_matrix)

  feature_config <- switch(feature_type,
    "words" = list(display_name = "Word Co-occurrence", viridis_option = "C"),
    "topics" = list(display_name = "Topic Distribution", viridis_option = "B"),
    "ngrams" = list(display_name = "N-gram Pattern", viridis_option = "D"),
    "embeddings" = list(display_name = "Semantic Embedding", viridis_option = "A"),
    list(display_name = feature_type, viridis_option = "H")
  )

  if (!is.null(docs_data) && nrow(docs_data) >= n_docs) {
    docs_data <- docs_data[1:n_docs, ]
    x_labels <- docs_data$document_number %||% paste("Doc", 1:n_docs)
    y_labels <- x_labels
  } else {
    x_labels <- paste("Doc", 1:n_docs)
    y_labels <- x_labels
  }

  if (is.null(title)) {
    title <- if (!is.null(category_filter) && category_filter != "all") {
      paste("Document", feature_config$display_name, "Similarity:", category_filter)
    } else {
      paste("Document", feature_config$display_name, "Similarity Heatmap")
    }
  }

  heat_df <- expand.grid(
    col_idx = seq_len(n_docs),
    row_idx = seq_len(n_docs)
  )
  heat_df$similarity <- as.vector(similarity_matrix)
  heat_df$x_label <- factor(x_labels[heat_df$col_idx], levels = x_labels)
  heat_df$y_label <- factor(y_labels[heat_df$row_idx], levels = y_labels)

  if (!is.null(docs_data) && nrow(docs_data) >= n_docs) {
    doc_ids <- docs_data$document_id_display %||% x_labels
    cats <- docs_data$category_display %||% rep("", n_docs)
    heat_df$tooltip_text <- paste0(
      y_labels[heat_df$row_idx], ": ", doc_ids[heat_df$row_idx],
      "\n", x_labels[heat_df$col_idx], ": ", doc_ids[heat_df$col_idx],
      "\nCategory: ", cats[heat_df$row_idx], " / ", cats[heat_df$col_idx],
      "\n", method_name, " Similarity: ", round(heat_df$similarity, 3)
    )
  } else {
    heat_df$tooltip_text <- paste0(
      y_labels[heat_df$row_idx], " vs ", x_labels[heat_df$col_idx],
      "\nSimilarity: ", round(heat_df$similarity, 3)
    )
  }

  p <- ggplot2::ggplot(heat_df, ggplot2::aes(x = .data$x_label, y = .data$y_label,
                                               fill = .data$similarity, text = .data$tooltip_text)) +
    ggplot2::geom_tile()

  if (show_values) {
    p <- p + ggplot2::geom_text(ggplot2::aes(label = round(.data$similarity, 2)), size = 3)
  }

  p +
    ggplot2::scale_fill_viridis_c(name = "Similarity\nScore", option = feature_config$viridis_option) +
    ggplot2::labs(title = title, x = "Documents", y = "Documents") +
    ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
}


#' RAG Semantic Search
#'
#' @description
#' Simple in-memory RAG (Retrieval Augmented Generation) for question-answering
#' over document corpus with source attribution. Uses OpenAI or Gemini
#' embeddings for semantic search and LLM for answer generation.
#'
#' @param query Character string, user question
#' @param documents Character vector, corpus to search
#' @param provider Character string, provider: "openai" or "gemini"
#' @param api_key Character string, API key (or from
#'   OPENAI_API_KEY/GEMINI_API_KEY env).
#' @param embedding_model Character string, embedding model. Defaults:
#'   "text-embedding-3-small" (openai), "gemini-embedding-001" (gemini)
#' @param chat_model Character string, chat model. Defaults:
#'   "gpt-4.1-mini" (openai), "gemini-2.5-flash-lite" (gemini)
#' @param top_k Integer, number of documents to retrieve (default: 5)
#'
#' @return List with:
#'   - success: Logical
#'   - answer: Generated answer
#'   - confidence: Confidence score (0-1)
#'   - sources: Vector of source document indices
#'   - retrieved_docs: Retrieved document chunks
#'   - scores: Similarity scores
#'
#' @details
#' Simple RAG workflow:
#' 1. Generate embeddings for documents and query
#' 2. Find top-k similar documents via cosine similarity
#' 3. Generate answer using LLM with retrieved context
#'
#' @concept ai
#' @seealso [get_best_embeddings()] for the retrieval step alone; [call_llm_api()] for the answer-generation step alone; [sanitize_llm_input()] for an input safety check before calling
#' @export
#'
#' @examples
#' if (interactive()) {
#' documents <- c(
#'   "Assistive technology helps students with disabilities access curriculum.",
#'   "Universal Design for Learning provides multiple means of engagement.",
#'   "Response to Intervention uses tiered support systems."
#' )
#'
#' # Using OpenAI (requires API key)
#' result <- run_rag_search(
#'   query = "How does assistive technology support learning?",
#'   documents = documents,
#'   provider = "openai"
#' )
#'
#' if (result$success) {
#'   cat("Answer:", result$answer, "\n")
#'   cat("Sources:", paste(result$sources, collapse = ", "), "\n")
#' }
#' }
run_rag_search <- function(
  query,
  documents,
  provider = c("openai", "gemini"),
  api_key = NULL,
  embedding_model = NULL,
  chat_model = NULL,
  top_k = 5
) {
  provider <- match.arg(provider)

  if (is.null(api_key)) {
    api_key <- switch(provider,
      "openai" = Sys.getenv("OPENAI_API_KEY"),
      "gemini" = Sys.getenv("GEMINI_API_KEY")
    )
  }

  if (!nzchar(api_key)) {
    return(list(
      success = FALSE,
      error = .missing_api_key_message(provider, "package"),
      answer = "",
      confidence = 0.0,
      sources = c()
    ))
  }

  if (length(documents) < 1) {
    return(list(
      success = FALSE,
      error = "No documents provided",
      answer = "",
      confidence = 0.0,
      sources = c()
    ))
  }

  if (is.null(embedding_model)) {
    embedding_model <- switch(provider,
      "openai" = "text-embedding-3-small",
      "gemini" = "gemini-embedding-001"
    )
  }

  if (is.null(chat_model)) {
    chat_model <- switch(provider,
      "openai" = "gpt-4.1-mini",
      "gemini" = "gemini-2.5-flash-lite"
    )
  }

  # Cap top_k at available documents
  top_k <- min(top_k, length(documents))

  doc_embeddings <- tryCatch({
    get_api_embeddings(
      texts = documents,
      provider = provider,
      model = embedding_model,
      api_key = api_key
    )
  }, error = function(e) {
    return(list(error = e$message))
  })

  if (is.list(doc_embeddings) && !is.null(doc_embeddings$error)) {
    return(list(
      success = FALSE,
      error = paste("Failed to generate document embeddings:", doc_embeddings$error),
      answer = "",
      confidence = 0.0,
      sources = c()
    ))
  }

  if (is.null(doc_embeddings)) {
    return(list(
      success = FALSE,
      error = .missing_api_key_message(provider, "package"),
      answer = "",
      confidence = 0.0,
      sources = character(0)
    ))
  }

  query_embedding <- tryCatch({
    get_api_embeddings(
      texts = query,
      provider = provider,
      model = embedding_model,
      api_key = api_key
    )
  }, error = function(e) {
    return(list(error = e$message))
  })

  if (is.list(query_embedding) && !is.null(query_embedding$error)) {
    return(list(
      success = FALSE,
      error = paste("Failed to generate query embedding:", query_embedding$error),
      answer = "",
      confidence = 0.0,
      sources = c()
    ))
  }

  if (is.null(query_embedding)) {
    return(list(
      success = FALSE,
      error = .missing_api_key_message(provider, "package"),
      answer = "",
      confidence = 0.0,
      sources = character(0)
    ))
  }

  # Step 3: Calculate cosine similarity between query and all documents
  query_vec <- as.numeric(query_embedding[1, ])
  doc_matrix <- as.matrix(doc_embeddings)

  similarities <- apply(doc_matrix, 1, function(doc_vec) {
    .cosine_sim(query_vec, doc_vec)
  })

  # Step 4: Get top-k most similar documents
  top_indices <- order(similarities, decreasing = TRUE)[1:top_k]
  top_scores <- similarities[top_indices]
  retrieved_docs <- documents[top_indices]

  query <- sanitize_llm_input(query)

  # Step 5: Generate answer using LLM with retrieved context
  context <- paste(
    sapply(seq_along(retrieved_docs), function(i) {
      sprintf("[Document %d] %s", top_indices[i], retrieved_docs[i])
    }),
    collapse = "\n\n"
  )

  system_prompt <- "You are a helpful research assistant. Answer the user's question based on the provided context documents. Be concise and accurate. If the context doesn't contain relevant information, say so."

  user_prompt <- sprintf(
    "Context Documents:\n%s\n\nQuestion: %s\n\nPlease provide a concise answer based on the context above.",
    context,
    query
  )

  answer <- tryCatch({
    call_llm_api(
      provider = provider,
      system_prompt = system_prompt,
      user_prompt = user_prompt,
      model = chat_model,
      temperature = 0.3,
      max_tokens = 500,
      api_key = api_key
    )
  }, error = function(e) {
    return(NULL)
  })

  if (is.null(answer)) {
    return(list(
      success = FALSE,
      error = "Failed to generate answer from LLM",
      answer = "",
      confidence = 0.0,
      sources = top_indices,
      retrieved_docs = retrieved_docs,
      scores = top_scores
    ))
  }

  # Confidence from top similarity scores
  avg_similarity <- mean(top_scores)
  confidence <- min(1.0, max(0.0, avg_similarity))

  return(list(
    success = TRUE,
    answer = answer,
    confidence = round(confidence, 3),
    sources = top_indices,
    retrieved_docs = retrieved_docs,
    scores = round(top_scores, 4),
    models = list(
      embedding = embedding_model,
      chat = chat_model
    ),
    provider = provider
  ))
}


# Network Analysis Functions

#' @title Analyze and Visualize Word Co-occurrence Networks
#'
#' @description
#' This function creates a word co-occurrence network based on a document-feature matrix (dfm).
#'
#' @param dfm_object A quanteda document-feature matrix (dfm).
#' @param doc_var A document-level metadata variable (default: NULL).
#' @param co_occur_n Minimum co-occurrence count (default: 50). A heuristic
#'   default; vary it to check edge-count sensitivity.
#' @param edge_metric Edge weight: "count" (default, raw co-occurrence) or
#'   "pmi" (pointwise mutual information, correcting the bias toward generic
#'   high-frequency words; only positive-PMI pairs are kept).
#' @param top_node_n Number of top nodes to display (default: 30).
#' @param nrows Number of rows to display in the table (default: 1).
#' @param height The height of the resulting Plotly plot, in pixels (default: 800).
#' @param width The width of the resulting Plotly plot, in pixels (default: 900).
#' @param node_label_size Maximum font size for node labels in pixels (default: 22).
#' @param community_method Community detection method: "leiden" (default) or "louvain".
#' @param node_size_by Node sizing method: "degree", "betweenness", "closeness", "eigenvector", or "fixed" (default: "degree").
#' @param node_color_by Node coloring method: "community" or "centrality" (default: "community").
#' @param seed Integer seed for the force-directed layout, so the plot is reproducible (default: 123).
#' @param category_params Optional named list of category-specific parameters. Each element should be a list with `co_occur_n` and `top_node_n` values for that category (default: NULL).
#'
#' @return A list containing the ggplot2 plot, a table, and a summary.
#'
#' @importFrom igraph graph_from_data_frame V vcount ecount degree betweenness closeness eigen_centrality layout_with_fr induced_subgraph cluster_leiden cluster_louvain edge_density diameter transitivity modularity assortativity_degree distances
#' @importFrom ggplot2 ggplot aes geom_segment geom_point scale_size_identity theme_void labs
#' @importFrom dplyr count filter mutate select group_by summarise ungroup left_join arrange desc group_map pull
#' @importFrom tibble as_tibble
#' @importFrom tidytext tidy
#' @importFrom widyr pairwise_count pairwise_pmi
#' @importFrom scales rescale
#' @importFrom stats quantile setNames
#' @importFrom DT datatable formatStyle
#' @importFrom rlang sym
#' @importFrom utils head
#' @importFrom grDevices colorRampPalette
#' @importFrom htmltools tagList tags browsable
#' @importFrom RColorBrewer brewer.pal
#'
#' @family semantic
#' @export
#'
#' @examples
#' if (interactive()) {
#'   df <- TextAnalysisR::SpecialEduTech
#'
#'   united_tbl <- TextAnalysisR::unite_cols(df, listed_vars = c("title", "abstract"))
#'
#'   tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")
#'
#'   dfm_object <- quanteda::dfm(tokens)
#'
#'   word_co_occurrence_network_results <- TextAnalysisR::word_co_occurrence_network(
#'                                         dfm_object,
#'                                         doc_var = NULL,
#'                                         co_occur_n = 50,
#'                                         top_node_n = 30,
#'                                         nrows = 1,
#'                                         height = 800,
#'                                         width = 900,
#'                                         community_method = "leiden")
#'   print(word_co_occurrence_network_results$plot)
#'   print(word_co_occurrence_network_results$table)
#'   print(word_co_occurrence_network_results$summary)
#' }
word_co_occurrence_network <- function(dfm_object,
                                       doc_var = NULL,
                                       co_occur_n = 50,
                                       edge_metric = c("count", "pmi"),
                                       top_node_n = 30,
                                       nrows = 1,
                                       height = 800,
                                       width = 900,
                                       node_label_size = 22,
                                       community_method = "leiden",
                                       node_size_by = "degree",
                                       node_color_by = "community",
                                       seed = 123,
                                       category_params = NULL) {

  edge_metric <- match.arg(edge_metric)

  if (!requireNamespace("htmltools", quietly = TRUE) ||
      !requireNamespace("RColorBrewer", quietly = TRUE)) {
    stop(
      "The 'htmltools' and 'RColorBrewer' packages are required for this functionality. ",
      "Please install them using install.packages(c('htmltools', 'RColorBrewer'))."
    )
  }

  dfm_td <- tidytext::tidy(dfm_object)
  docvars_df <- quanteda::docvars(dfm_object)
  docvars_df$document <- quanteda::docnames(dfm_object)
  dfm_td <- dplyr::left_join(dfm_td, docvars_df, by = "document")

  if (!is.null(doc_var) && doc_var != "" && !doc_var %in% colnames(dfm_td)) {
    message("Document-level metadata variable '", doc_var, "' was not selected or not found.")
    doc_var <- NULL
  }

  if (!is.null(doc_var) && doc_var %in% colnames(dfm_td)) {
    docvar_levels <- unique(dfm_td[[doc_var]])
    message(paste("doc_var has", length(docvar_levels), "levels:", paste(docvar_levels, collapse = ", ")))
  } else {
    docvar_levels <- NULL
  }

  build_table <- function(net, group_label) {
    layout_dff <- net$layout_df %>%
      dplyr::select(-c("x", "y")) %>%
      dplyr::mutate(dplyr::across(dplyr::where(is.numeric), ~ round(., 3)))

    table <- DT::datatable(layout_dff, rownames = FALSE,
                           extensions = 'Buttons',
                           options = list(scrollX = TRUE,
                                          width = "80%",
                                          dom = 'Bfrtip',
                                          buttons = c('copy', 'csv', 'excel', 'pdf', 'print'))) %>%
      DT::formatStyle(columns = colnames(layout_dff), `font-size` = "16px")

    htmltools::tagList(
      htmltools::tags$div(
        style = "margin-bottom: 20px;",
        htmltools::tags$p(
          group_label,
          style = "font-weight: bold; text-align: center; font-size: 16px;"
        )
      ),
      table
    )
  }

  build_summary <- function(net, group_label) {
    g <- net$graph
    summary_df <- data.frame(
      Metric = c("Nodes", "Edges", "Density", "Diameter",
                 "Global Clustering Coefficient", "Local Clustering Coefficient (Mean)",
                 "Modularity", "Assortativity", "Geodesic Distance (Mean)"),
      Value = c(
        igraph::vcount(g),
        igraph::ecount(g),
        igraph::edge_density(g),
        igraph::diameter(g),
        igraph::transitivity(g, type = "global"),
        mean(igraph::transitivity(g, type = "local"), na.rm = TRUE),
        igraph::modularity(g, membership = igraph::V(g)$community),
        igraph::assortativity_degree(g),
        mean(igraph::distances(g)[igraph::distances(g) != Inf], na.rm = TRUE)
      )
    ) %>%
      dplyr::mutate(dplyr::across(dplyr::where(is.numeric), ~ round(., 3)))

    summary_table <- DT::datatable(summary_df, rownames = FALSE,
                                   extensions = 'Buttons',
                                   options = list(scrollX = TRUE,
                                                  width = "80%",
                                                  dom = 'Bfrtip',
                                                  buttons = c('copy', 'csv', 'excel', 'pdf', 'print'))) %>%
      DT::formatStyle(columns = colnames(summary_df), `font-size` = "16px")

    htmltools::tagList(
      htmltools::tags$div(
        style = "margin-bottom: 20px;",
        htmltools::tags$p(
          group_label,
          style = "font-weight: bold; text-align: center; font-size: 16px;"
        )
      ),
      summary_table
    )
  }

  build_network_plot <- function(data, group_level = NULL, local_co_occur_n = NULL, local_top_node_n = NULL) {
    # Use category-specific params if provided, otherwise fall back to global params
    effective_co_occur_n <- if (!is.null(local_co_occur_n)) local_co_occur_n else co_occur_n
    effective_top_node_n <- if (!is.null(local_top_node_n)) local_top_node_n else top_node_n

    term_co_occur <- data %>%
      widyr::pairwise_count(term, document, sort = TRUE, upper = FALSE) %>%
      dplyr::filter(n >= effective_co_occur_n)

    # count floor controls rare-pair PMI inflation
    if (edge_metric == "pmi") {
      term_pmi <- data %>%
        widyr::pairwise_pmi(term, document, sort = FALSE)
      term_co_occur <- term_co_occur %>%
        dplyr::inner_join(term_pmi, by = c("item1", "item2")) %>%
        dplyr::filter(.data$pmi > 0) %>%
        dplyr::mutate(n = .data$pmi)
    }

    graph <- igraph::graph_from_data_frame(term_co_occur, directed = FALSE)
    if (igraph::vcount(graph) == 0) {
      message("No co-occurrence relationships meet the threshold.")
      return(NULL)
    }

    # Cap nodes before layout: force-directed placement stalls on dense graphs
    node_cap <- min(max(effective_top_node_n, 1L), 300L)
    if (igraph::vcount(graph) > node_cap) {
      keep <- utils::head(order(igraph::degree(graph), decreasing = TRUE), node_cap)
      message(sprintf(
        "Network has %d nodes; rendering the top %d by degree to keep the layout responsive.",
        igraph::vcount(graph), node_cap))
      graph <- igraph::induced_subgraph(graph, keep)
    }

    igraph::V(graph)$degree      <- igraph::degree(graph)
    igraph::V(graph)$betweenness <- igraph::betweenness(graph)
    igraph::V(graph)$closeness   <- igraph::closeness(graph)
    igraph::V(graph)$eigenvector <- igraph::eigen_centrality(graph)$vector
    community_result <- if (community_method == "louvain") {
      igraph::cluster_louvain(graph)
    } else {
      # modularity objective matches the reported modularity statistic
      igraph::cluster_leiden(graph, objective_function = "modularity")
    }
    igraph::V(graph)$community <- community_result$membership

    layout_mat <- withr::with_seed(seed, igraph::layout_with_fr(graph))
    layout_df <- as.data.frame(layout_mat) %>% stats::setNames(c("x", "y"))
    layout_df <- layout_df %>%
      dplyr::mutate(label       = igraph::V(graph)$name,
                    degree      = igraph::V(graph)$degree,
                    betweenness = igraph::V(graph)$betweenness,
                    closeness   = igraph::V(graph)$closeness,
                    eigenvector = igraph::V(graph)$eigenvector,
                    community   = igraph::V(graph)$community)

    # Calculate word frequency from the original data
    word_freq <- data %>%
      dplyr::group_by(term) %>%
      dplyr::summarise(frequency = sum(count), .groups = "drop")
    layout_df <- layout_df %>%
      dplyr::left_join(word_freq, by = c("label" = "term")) %>%
      dplyr::mutate(frequency = ifelse(is.na(frequency), 1, frequency))

    edge_data <- igraph::as_data_frame(graph, what = "edges") %>%
      dplyr::mutate(x    = layout_df$x[match(from, layout_df$label)],
                    y    = layout_df$y[match(from, layout_df$label)],
                    xend = layout_df$x[match(to, layout_df$label)],
                    yend = layout_df$y[match(to, layout_df$label)],
                    cooccur_count = n) %>%
      dplyr::select(from, to, x, y, xend, yend, cooccur_count) %>%
      dplyr::mutate(line_width = scales::rescale(cooccur_count, to = c(0.3, 2)),
                    alpha      = scales::rescale(cooccur_count, to = c(0.2, 0.9)))

    # Node size
    size_metric <- switch(node_size_by,
      "degree" = layout_df$degree,
      "betweenness" = layout_df$betweenness,
      "frequency" = layout_df$frequency,
      "fixed" = rep(20, nrow(layout_df)),
      layout_df$degree
    )

    node_data <- layout_df %>%
      dplyr::mutate(
        size = if (node_size_by == "fixed") 6 else scales::rescale(sqrt(size_metric), to = c(3, 10)),
        text_size = scales::rescale(log1p(size_metric), to = c(node_label_size - 8, node_label_size)),
        alpha = scales::rescale(log1p(size_metric), to = c(0.2, 1)),
        hover_text = paste("Word:", label,
                           "<br>Degree:", degree,
                           "<br>Betweenness:", round(betweenness, 2),
                           "<br>Closeness:", round(closeness, 2),
                           "<br>Eigenvector:", round(eigenvector, 2),
                           "<br>Frequency:", frequency,
                           "<br>Community:", community,
                           if (!is.null(doc_var)) {
                             if (length(docvar_levels) > 1) {
                               paste0("<br>", doc_var, ": ", group_level)
                             } else {
                               paste0("<br>", doc_var)
                             }
                           } else ""
        )
      )

    # Create community palette
    n_communities <- length(unique(node_data$community))
    if (n_communities >= 3 && n_communities <= 8) {
      palette <- RColorBrewer::brewer.pal(n_communities, "Set2")
    } else if (n_communities > 8) {
      palette <- grDevices::colorRampPalette(RColorBrewer::brewer.pal(8, "Set2"))(n_communities)
    } else if (n_communities > 0 && n_communities < 3) {
      palette <- RColorBrewer::brewer.pal(3, "Set2")[1:n_communities]
    } else {
      palette <- rep("#000000", n_communities)
    }

    node_data$community <- factor(node_data$community, levels = unique(node_data$community))
    community_levels <- levels(node_data$community)
    names(palette) <- community_levels

    # Determine node color based on node_color_by parameter
    if (node_color_by == "frequency") {
      node_data$color <- scales::col_numeric("viridis", domain = range(node_data$frequency, na.rm = TRUE))(node_data$frequency)
    } else {
      node_data$color <- palette[as.character(node_data$community)]
    }

    sort_col <- switch(node_size_by,
      "betweenness" = "betweenness",
      "frequency"   = "frequency",
      "degree"
    )
    top_nodes <- dplyr::arrange(node_data, dplyr::desc(.data[[sort_col]])) %>%
      utils::head(effective_top_node_n) %>%
      dplyr::mutate(dplyr::across(dplyr::where(is.numeric), ~ round(., 3)))

    p <- ggplot2::ggplot() +
      ggplot2::geom_segment(data = edge_data,
                            ggplot2::aes(x = .data$x, y = .data$y, xend = .data$xend, yend = .data$yend,
                                         linewidth = .data$line_width, alpha = .data$alpha),
                            color = "#5C5CFF", show.legend = FALSE) +
      ggplot2::scale_linewidth_identity() +
      ggplot2::scale_alpha_identity()

    if (node_color_by == "frequency") {
      p <- p +
        ggplot2::geom_point(data = node_data,
                            ggplot2::aes(x = .data$x, y = .data$y, size = .data$size, color = .data$frequency,
                                         text = .data$hover_text),
                            stroke = 0.5) +
        ggplot2::scale_color_viridis_c(name = "Frequency")
    } else {
      p <- p +
        ggplot2::geom_point(data = node_data,
                            ggplot2::aes(x = .data$x, y = .data$y, size = .data$size, fill = .data$community,
                                         text = .data$hover_text),
                            shape = 21, color = "white", stroke = 1) +
        ggplot2::scale_fill_manual(values = palette, name = "Community")
    }

    p <- p +
      ggplot2::geom_text(data = top_nodes,
                         ggplot2::aes(x = .data$x, y = .data$y, label = .data$label),
                         size = node_label_size / 5, color = "#0c1f4a",
                         vjust = -0.7, show.legend = FALSE) +
      ggplot2::scale_size_identity() +
      ggplot2::theme_void(base_size = 12) +
      ggplot2::labs(title = "Word Co-occurrence Network") +
      ggplot2::theme(
        plot.title = ggplot2::element_text(face = "bold", hjust = 0.5, size = 16, color = "#0c1f4a"),
        legend.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
        legend.text = ggplot2::element_text(size = 12, color = "#3B3B3B")
      )

    list(plot = p, layout_df = layout_df, graph = graph, top_nodes = top_nodes)
  }

  if (!is.null(doc_var) && length(docvar_levels) > 1) {
    plots_list <- dfm_td %>%
      dplyr::ungroup() %>%
      dplyr::group_by(!!rlang::sym(doc_var)) %>%
      dplyr::group_map(~ {
        group_level <- .y[[doc_var]]
        message(paste("Processing group level:", group_level))

        if (is.null(group_level)) {
          stop("doc_var is missing or not found in the current group")
        }

        local_co_occur_n <- NULL
        local_top_node_n <- NULL
        if (!is.null(category_params) && group_level %in% names(category_params)) {
          cat_params <- category_params[[group_level]]
          if (!is.null(cat_params$co_occur_n)) local_co_occur_n <- cat_params$co_occur_n
          if (!is.null(cat_params$top_node_n)) local_top_node_n <- cat_params$top_node_n
          message(paste("  Using category-specific params - co_occur_n:", local_co_occur_n, ", top_node_n:", local_top_node_n))
        }

        net <- build_network_plot(.x, group_level, local_co_occur_n, local_top_node_n)
        if (!is.null(net)) {
          net$plot + ggplot2::labs(title = group_level)
        } else {
          NULL
        }
      })

    plots_list <- Filter(Negate(is.null), plots_list)
    if (!requireNamespace("patchwork", quietly = TRUE)) {
      stop("Install the 'patchwork' package to combine category-faceted networks.")
    }
    combined_plot <- patchwork::wrap_plots(plots_list, nrow = nrows)

    table_list <- lapply(docvar_levels, function(level) {
      message(paste("Generating table for level:", level))
      group_data <- dplyr::filter(dfm_td, !!rlang::sym(doc_var) == level)
      # Look up category-specific parameters
      local_co_occur_n <- NULL
      local_top_node_n <- NULL
      if (!is.null(category_params) && level %in% names(category_params)) {
        cat_params <- category_params[[level]]
        if (!is.null(cat_params$co_occur_n)) local_co_occur_n <- cat_params$co_occur_n
        if (!is.null(cat_params$top_node_n)) local_top_node_n <- cat_params$top_node_n
      }
      net <- build_network_plot(group_data, level, local_co_occur_n, local_top_node_n)
      if (!is.null(net)) build_table(net, level) else NULL
    })

    summary_list <- lapply(docvar_levels, function(level) {
      message(paste("Generating summary for level:", level))
      group_data <- dplyr::filter(dfm_td, !!rlang::sym(doc_var) == level)
      # Look up category-specific parameters
      local_co_occur_n <- NULL
      local_top_node_n <- NULL
      if (!is.null(category_params) && level %in% names(category_params)) {
        cat_params <- category_params[[level]]
        if (!is.null(cat_params$co_occur_n)) local_co_occur_n <- cat_params$co_occur_n
        if (!is.null(cat_params$top_node_n)) local_top_node_n <- cat_params$top_node_n
      }
      net <- build_network_plot(group_data, level, local_co_occur_n, local_top_node_n)
      if (!is.null(net)) build_summary(net, level) else NULL
    })

    return(list(
      plot = combined_plot,
      table = table_list %>% htmltools::tagList() %>% htmltools::browsable(),
      summary = summary_list %>% htmltools::tagList() %>% htmltools::browsable()
    ))
  } else {
    net <- build_network_plot(dfm_td)
    if (is.null(net)) {
      message("No network generated.")
      return(NULL)
    }
    return(list(
      plot = net$plot,
      top_nodes = net$top_nodes,
      table = build_table(net, if (!is.null(doc_var)) paste("Network Centrality Table for", doc_var) else "Network Centrality Table") %>% htmltools::browsable(),
      summary = build_summary(net, if (!is.null(doc_var)) paste("Network Summary for", doc_var) else "Network Summary") %>% htmltools::browsable()
    ))
  }
}


#' @title Analyze and Visualize Word Correlation Networks
#'
#' @description
#' This function creates a word correlation network based on a document-feature matrix (dfm).
#'
#' @param dfm_object A quanteda document-feature matrix (dfm).
#' @param doc_var A document-level metadata variable (default: NULL).
#' @param common_term_n Minimum number of documents a term must appear in
#'   (default: 130). This prefilter prevents spurious high correlations from
#'   rare words; lower it for small corpora.
#' @param corr_n Minimum phi correlation for keeping an edge (default: 0.4).
#'   A heuristic default; report edge counts across nearby thresholds to
#'   check sensitivity.
#' @param top_node_n Number of top nodes to display (default: 40).
#' @param nrows Number of rows to display in the table (default: 1).
#' @param height The height of the resulting Plotly plot, in pixels (default: 1000).
#' @param width The width of the resulting Plotly plot, in pixels (default: 900).
#' @param node_label_size Maximum font size for node labels in pixels (default: 22).
#' @param community_method Community detection method: "leiden" (default) or "louvain".
#' @param node_size_by Node sizing method: "degree", "betweenness", "closeness", "eigenvector", or "fixed" (default: "degree").
#' @param node_color_by Node coloring method: "community" or "centrality" (default: "community").
#' @param seed Integer seed for the force-directed layout, so the plot is reproducible (default: 123).
#' @param category_params Optional named list of category-specific parameters. Each element should be a list with `common_term_n`, `corr_n`, and `top_node_n` values for that category (default: NULL).
#'
#' @return A list containing the ggplot2 plot, a table, and a summary.
#'
#' @importFrom igraph graph_from_data_frame V vcount ecount degree betweenness closeness eigen_centrality layout_with_fr induced_subgraph cluster_leiden cluster_louvain edge_density diameter transitivity modularity assortativity_degree distances
#' @importFrom ggplot2 ggplot aes geom_segment geom_point scale_size_identity theme_void labs
#' @importFrom dplyr count filter mutate select group_by summarise ungroup left_join arrange desc group_map pull
#' @importFrom tibble as_tibble
#' @importFrom tidytext tidy
#' @importFrom widyr pairwise_cor
#' @importFrom scales rescale
#' @importFrom stats quantile setNames
#' @importFrom DT datatable formatStyle
#' @importFrom rlang sym
#' @importFrom utils head
#' @importFrom grDevices colorRampPalette
#' @importFrom htmltools tagList tags browsable
#' @importFrom RColorBrewer brewer.pal
#'
#' @family semantic
#' @export
#'
#' @examples
#' if (interactive()) {
#'   df <- TextAnalysisR::SpecialEduTech
#'
#'   united_tbl <- TextAnalysisR::unite_cols(df, listed_vars = c("title", "abstract"))
#'
#'   tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")
#'
#'   dfm_object <- quanteda::dfm(tokens)
#'
#'   word_correlation_network_results <- TextAnalysisR::word_correlation_network(
#'                                       dfm_object,
#'                                       doc_var = NULL,
#'                                       common_term_n = 30,
#'                                       corr_n = 0.4,
#'                                       top_node_n = 40,
#'                                       nrows = 1,
#'                                       height = 1000,
#'                                       width = 900,
#'                                       community_method = "leiden")
#'   print(word_correlation_network_results$plot)
#'   print(word_correlation_network_results$table)
#'   print(word_correlation_network_results$summary)
#' }
word_correlation_network <- function(dfm_object,
                                     doc_var = NULL,
                                     common_term_n = 130,
                                     corr_n = 0.4,
                                     top_node_n = 40,
                                     nrows = 1,
                                     height = 1000,
                                     width = 900,
                                     node_label_size = 22,
                                     community_method = "leiden",
                                     node_size_by = "degree",
                                     node_color_by = "community",
                                     seed = 123,
                                     category_params = NULL) {

  if (!requireNamespace("htmltools", quietly = TRUE) ||
      !requireNamespace("RColorBrewer", quietly = TRUE)) {
    stop(
      "The 'htmltools' and 'RColorBrewer' packages are required for this functionality. ",
      "Please install them using install.packages(c('htmltools', 'RColorBrewer'))."
    )
  }

  dfm_td <- tidytext::tidy(dfm_object)
  docvars_df <- quanteda::docvars(dfm_object)
  docvars_df$document <- quanteda::docnames(dfm_object)
  dfm_td <- dplyr::left_join(dfm_td, docvars_df, by = "document")

  if (!is.null(doc_var) && doc_var != "" && !doc_var %in% colnames(dfm_td)) {
    message("Document-level metadata variable '", doc_var, "' was not selected or not found.")
    doc_var <- NULL
  }

  if (!is.null(doc_var) && doc_var %in% colnames(dfm_td)) {
    docvar_levels <- unique(dfm_td[[doc_var]])
    message(paste("doc_var has", length(docvar_levels), "levels:", paste(docvar_levels, collapse = ", ")))
  } else {
    docvar_levels <- NULL
  }

  build_table <- function(net, group_label) {
    layout_dff <- net$layout_df %>%
      dplyr::select(-c("x", "y")) %>%
      dplyr::mutate(dplyr::across(dplyr::where(is.numeric), ~ round(., 3)))

    table <- DT::datatable(layout_dff, rownames = FALSE,
                           extensions = 'Buttons',
                           options = list(scrollX = TRUE,
                                          width = "80%",
                                          dom = 'Bfrtip',
                                          buttons = c('copy', 'csv', 'excel', 'pdf', 'print'))) %>%
      DT::formatStyle(columns = colnames(layout_dff), `font-size` = "16px")

    htmltools::tagList(
      htmltools::tags$div(
        style = "margin-bottom: 20px;",
        htmltools::tags$p(
          group_label,
          style = "font-weight: bold; text-align: center; font-size: 16px;"
        )
      ),
      table
    )
  }

  build_summary <- function(net, group_label) {
    g <- net$graph
    summary_df <- data.frame(
      Metric = c("Nodes", "Edges", "Density", "Diameter",
                 "Global Clustering Coefficient", "Local Clustering Coefficient (Mean)",
                 "Modularity", "Assortativity", "Geodesic Distance (Mean)"),
      Value = c(
        igraph::vcount(g),
        igraph::ecount(g),
        igraph::edge_density(g),
        igraph::diameter(g),
        igraph::transitivity(g, type = "global"),
        mean(igraph::transitivity(g, type = "local"), na.rm = TRUE),
        igraph::modularity(g, membership = igraph::V(g)$community),
        igraph::assortativity_degree(g),
        mean(igraph::distances(g)[igraph::distances(g) != Inf], na.rm = TRUE)
      )
    ) %>%
      dplyr::mutate(dplyr::across(dplyr::where(is.numeric), ~ round(., 3)))

    summary_table <- DT::datatable(summary_df, rownames = FALSE,
                                   extensions = 'Buttons',
                                   options = list(scrollX = TRUE,
                                                  width = "80%",
                                                  dom = 'Bfrtip',
                                                  buttons = c('copy', 'csv', 'excel', 'pdf', 'print'))) %>%
      DT::formatStyle(columns = colnames(summary_df), `font-size` = "16px")

    htmltools::tagList(
      htmltools::tags$div(
        style = "margin-bottom: 20px;",
        htmltools::tags$p(
          group_label,
          style = "font-weight: bold; text-align: center; font-size: 16px;"
        )
      ),
      summary_table
    )
  }

  build_network_plot <- function(data, group_level = NULL, local_common_term_n = NULL, local_corr_n = NULL, local_top_node_n = NULL) {
    # Use category-specific params if provided, otherwise fall back to global params
    effective_common_term_n <- if (!is.null(local_common_term_n)) local_common_term_n else common_term_n
    effective_corr_n <- if (!is.null(local_corr_n)) local_corr_n else corr_n
    effective_top_node_n <- if (!is.null(local_top_node_n)) local_top_node_n else top_node_n

    term_cor <- data %>%
      dplyr::group_by(term) %>%
      dplyr::filter(dplyr::n() >= effective_common_term_n) %>%
      widyr::pairwise_cor(term, document, sort = TRUE, upper = FALSE) %>%
      dplyr::ungroup() %>%
      dplyr::filter(correlation > effective_corr_n)

    graph <- igraph::graph_from_data_frame(term_cor, directed = FALSE)
    if(igraph::vcount(graph) == 0) {
      message("No correlation relationships meet the threshold.")
      return(NULL)
    }

    # Cap nodes before layout: force-directed placement stalls on dense graphs
    node_cap <- min(max(effective_top_node_n, 1L), 300L)
    if (igraph::vcount(graph) > node_cap) {
      keep <- utils::head(order(igraph::degree(graph), decreasing = TRUE), node_cap)
      message(sprintf(
        "Network has %d nodes; rendering the top %d by degree to keep the layout responsive.",
        igraph::vcount(graph), node_cap))
      graph <- igraph::induced_subgraph(graph, keep)
    }

    igraph::V(graph)$degree      <- igraph::degree(graph)
    igraph::V(graph)$betweenness <- igraph::betweenness(graph)
    igraph::V(graph)$closeness   <- igraph::closeness(graph)
    igraph::V(graph)$eigenvector <- igraph::eigen_centrality(graph)$vector
    community_result <- if (community_method == "louvain") {
      igraph::cluster_louvain(graph)
    } else {
      # modularity objective matches the reported modularity statistic
      igraph::cluster_leiden(graph, objective_function = "modularity")
    }
    igraph::V(graph)$community <- community_result$membership

    layout_mat <- withr::with_seed(seed, igraph::layout_with_fr(graph))
    layout_df <- as.data.frame(layout_mat) %>% stats::setNames(c("x", "y"))
    layout_df <- layout_df %>%
      dplyr::mutate(label       = igraph::V(graph)$name,
                    degree      = igraph::V(graph)$degree,
                    betweenness = igraph::V(graph)$betweenness,
                    closeness   = igraph::V(graph)$closeness,
                    eigenvector = igraph::V(graph)$eigenvector,
                    community   = igraph::V(graph)$community)

    # Calculate word frequency from the original data
    word_freq <- data %>%
      dplyr::group_by(term) %>%
      dplyr::summarise(frequency = sum(count), .groups = "drop")
    layout_df <- layout_df %>%
      dplyr::left_join(word_freq, by = c("label" = "term")) %>%
      dplyr::mutate(frequency = ifelse(is.na(frequency), 1, frequency))

    edge_data <- igraph::as_data_frame(graph, what = "edges") %>%
      dplyr::mutate(x    = layout_df$x[match(from, layout_df$label)],
                    y    = layout_df$y[match(from, layout_df$label)],
                    xend = layout_df$x[match(to, layout_df$label)],
                    yend = layout_df$y[match(to, layout_df$label)],
                    correlation = correlation) %>%
      dplyr::select(from, to, x, y, xend, yend, correlation) %>%
      dplyr::mutate(line_width = scales::rescale(correlation, to = c(0.3, 2)),
                    alpha      = scales::rescale(correlation, to = c(0.2, 0.9)))

    # Node size
    size_metric <- switch(node_size_by,
      "degree" = layout_df$degree,
      "betweenness" = layout_df$betweenness,
      "frequency" = layout_df$frequency,
      "fixed" = rep(20, nrow(layout_df)),
      layout_df$degree
    )

    node_data <- layout_df %>%
      dplyr::mutate(
        size = if (node_size_by == "fixed") 6 else scales::rescale(sqrt(size_metric), to = c(3, 10)),
        text_size = scales::rescale(log1p(size_metric), to = c(node_label_size - 8, node_label_size)),
        alpha = scales::rescale(log1p(size_metric), to = c(0.2, 1)),
        hover_text = paste(
          "Word:", label,
          "<br>Degree:", degree,
          "<br>Betweenness:", round(betweenness, 2),
          "<br>Closeness:", round(closeness, 2),
          "<br>Eigenvector:", round(eigenvector, 2),
          "<br>Frequency:", frequency,
          "<br>Community:", community,
          if (!is.null(doc_var)) {
            if (length(docvar_levels) > 1) {
              paste0("<br>", doc_var, ": ", group_level)
            } else {
              paste0("<br>", doc_var)
            }
          } else ""
        )
      )

    # Create community palette
    n_communities <- length(unique(node_data$community))
    if (n_communities >= 3 && n_communities <= 8) {
      palette <- RColorBrewer::brewer.pal(n_communities, "Set2")
    } else if (n_communities > 8) {
      palette <- grDevices::colorRampPalette(RColorBrewer::brewer.pal(8, "Set2"))(n_communities)
    } else if (n_communities > 0 && n_communities < 3) {
      palette <- RColorBrewer::brewer.pal(3, "Set2")[1:n_communities]
    } else {
      palette <- rep("#000000", n_communities)
    }

    node_data$community <- factor(node_data$community, levels = unique(node_data$community))
    community_levels <- levels(node_data$community)
    names(palette) <- community_levels

    # Determine node color based on node_color_by parameter
    if (node_color_by == "frequency") {
      node_data$color <- scales::col_numeric("viridis", domain = range(node_data$frequency, na.rm = TRUE))(node_data$frequency)
    } else {
      node_data$color <- palette[as.character(node_data$community)]
    }

    sort_col <- switch(node_size_by,
      "betweenness" = "betweenness",
      "frequency"   = "frequency",
      "degree"
    )
    top_nodes <- dplyr::arrange(node_data, dplyr::desc(.data[[sort_col]])) %>%
      utils::head(effective_top_node_n) %>%
      dplyr::mutate(dplyr::across(dplyr::where(is.numeric), ~ round(., 3)))

    p <- ggplot2::ggplot() +
      ggplot2::geom_segment(data = edge_data,
                            ggplot2::aes(x = .data$x, y = .data$y, xend = .data$xend, yend = .data$yend,
                                         linewidth = .data$line_width, alpha = .data$alpha),
                            color = "#5C5CFF", show.legend = FALSE) +
      ggplot2::scale_linewidth_identity() +
      ggplot2::scale_alpha_identity()

    if (node_color_by == "frequency") {
      p <- p +
        ggplot2::geom_point(data = node_data,
                            ggplot2::aes(x = .data$x, y = .data$y, size = .data$size, color = .data$frequency,
                                         text = .data$hover_text),
                            stroke = 0.5) +
        ggplot2::scale_color_viridis_c(name = "Frequency")
    } else {
      p <- p +
        ggplot2::geom_point(data = node_data,
                            ggplot2::aes(x = .data$x, y = .data$y, size = .data$size, fill = .data$community,
                                         text = .data$hover_text),
                            shape = 21, color = "white", stroke = 1) +
        ggplot2::scale_fill_manual(values = palette, name = "Community")
    }

    p <- p +
      ggplot2::geom_text(data = top_nodes,
                         ggplot2::aes(x = .data$x, y = .data$y, label = .data$label),
                         size = node_label_size / 5, color = "#0c1f4a",
                         vjust = -0.7, show.legend = FALSE) +
      ggplot2::scale_size_identity() +
      ggplot2::theme_void(base_size = 12) +
      ggplot2::labs(title = "Word Correlation Network") +
      ggplot2::theme(
        plot.title = ggplot2::element_text(face = "bold", hjust = 0.5, size = 16, color = "#0c1f4a"),
        legend.title = ggplot2::element_text(size = 13, color = "#0c1f4a"),
        legend.text = ggplot2::element_text(size = 12, color = "#3B3B3B")
      )

    list(plot = p, layout_df = layout_df, graph = graph, top_nodes = top_nodes)
  }

  if (!is.null(doc_var) && length(docvar_levels) > 1) {
    plots_list <- dfm_td %>%
      dplyr::ungroup() %>%
      dplyr::group_by(!!rlang::sym(doc_var)) %>%
      dplyr::group_map(~ {
        group_level <- .y[[doc_var]]
        message(paste("Processing group level:", group_level))

        if (is.null(group_level)) {
          stop("doc_var is missing or not found in the current group")
        }

        local_common_term_n <- NULL
        local_corr_n <- NULL
        local_top_node_n <- NULL
        if (!is.null(category_params) && group_level %in% names(category_params)) {
          cat_params <- category_params[[group_level]]
          if (!is.null(cat_params$common_term_n)) local_common_term_n <- cat_params$common_term_n
          if (!is.null(cat_params$corr_n)) local_corr_n <- cat_params$corr_n
          if (!is.null(cat_params$top_node_n)) local_top_node_n <- cat_params$top_node_n
          message(paste("  Using category-specific params - common_term_n:", local_common_term_n,
                        ", corr_n:", local_corr_n, ", top_node_n:", local_top_node_n))
        }

        net <- build_network_plot(.x, group_level, local_common_term_n, local_corr_n, local_top_node_n)
        if (!is.null(net)) {
          net$plot + ggplot2::labs(title = group_level)
        } else {
          NULL
        }
      })

    plots_list <- Filter(Negate(is.null), plots_list)
    if (!requireNamespace("patchwork", quietly = TRUE)) {
      stop("Install the 'patchwork' package to combine category-faceted networks.")
    }
    combined_plot <- patchwork::wrap_plots(plots_list, nrow = nrows)

    table_list <- lapply(docvar_levels, function(level) {
      message(paste("Generating table for level:", level))
      group_data <- dplyr::filter(dfm_td, !!rlang::sym(doc_var) == level)
      # Look up category-specific parameters
      local_common_term_n <- NULL
      local_corr_n <- NULL
      local_top_node_n <- NULL
      if (!is.null(category_params) && level %in% names(category_params)) {
        cat_params <- category_params[[level]]
        if (!is.null(cat_params$common_term_n)) local_common_term_n <- cat_params$common_term_n
        if (!is.null(cat_params$corr_n)) local_corr_n <- cat_params$corr_n
        if (!is.null(cat_params$top_node_n)) local_top_node_n <- cat_params$top_node_n
      }
      net <- build_network_plot(group_data, level, local_common_term_n, local_corr_n, local_top_node_n)
      if (!is.null(net)) build_table(net, level) else NULL
    })

    summary_list <- lapply(docvar_levels, function(level) {
      message(paste("Generating summary for level:", level))
      group_data <- dplyr::filter(dfm_td, !!rlang::sym(doc_var) == level)
      # Look up category-specific parameters
      local_common_term_n <- NULL
      local_corr_n <- NULL
      local_top_node_n <- NULL
      if (!is.null(category_params) && level %in% names(category_params)) {
        cat_params <- category_params[[level]]
        if (!is.null(cat_params$common_term_n)) local_common_term_n <- cat_params$common_term_n
        if (!is.null(cat_params$corr_n)) local_corr_n <- cat_params$corr_n
        if (!is.null(cat_params$top_node_n)) local_top_node_n <- cat_params$top_node_n
      }
      net <- build_network_plot(group_data, level, local_common_term_n, local_corr_n, local_top_node_n)
      if (!is.null(net)) build_summary(net, level) else NULL
    })

    return(list(
      plot = combined_plot,
      table = table_list %>% htmltools::tagList() %>% htmltools::browsable(),
      summary = summary_list %>% htmltools::tagList() %>% htmltools::browsable()
    ))
  } else {
    net <- build_network_plot(dfm_td)
    if (is.null(net)) {
      message("No network generated.")
      return(NULL)
    }
    return(list(
      plot = net$plot,
      top_nodes = net$top_nodes,
      table = build_table(net,
                          if (!is.null(doc_var)) paste("Network Centrality Table for", doc_var)
                          else "Network Centrality Table") %>% htmltools::browsable(),
      summary = build_summary(net,
                              if (!is.null(doc_var)) paste("Network Summary for", doc_var)
                              else "Network Summary") %>% htmltools::browsable()
    ))
  }
}

test_that("reduce_dimensions does not leak .Random.seed to caller", {
  skip_if_not_installed("withr")
  withr::with_seed(42, {
    m <- matrix(rnorm(40), nrow = 10, ncol = 4)
    pre_seed <- .Random.seed
    suppressMessages(
      reduce_dimensions(
        m,
        method = "PCA",
        n_components = 2,
        seed = 7,
        verbose = FALSE
      )
    )
    expect_identical(.Random.seed, pre_seed)
  })
})

test_that("reduce_dimensions is reproducible across seeded calls", {
  skip_if_not_installed("withr")
  m <- matrix(rnorm(40), nrow = 10, ncol = 4)
  a <- suppressMessages(
    reduce_dimensions(m, method = "PCA", n_components = 2, seed = 11, verbose = FALSE)
  )
  b <- suppressMessages(
    reduce_dimensions(m, method = "PCA", n_components = 2, seed = 11, verbose = FALSE)
  )
  expect_equal(a$reduced_data, b$reduced_data)
})

test_that("run_rag_search validates document input", {
  skip_on_cran()
  result <- run_rag_search(
    query = "Test question?",
    documents = character(0),
    provider = "openai"
  )

  expect_false(result$success)
  expect_match(result$error, "No documents provided|API key")
})

test_that("run_rag_search requires API key", {
  skip_on_cran()
  result <- run_rag_search(
    query = "What is assistive technology?",
    documents = c("Assistive tech helps students", "Technology supports learning"),
    provider = "openai",
    api_key = ""
  )

  expect_type(result, "list")
  expect_true("success" %in% names(result))

  if (!result$success) {
    expect_match(result$error, "API key|No documents provided")
  }
})

test_that("run_rag_search returns expected structure", {
  skip_on_cran()
  skip_if(
    !nzchar(Sys.getenv("OPENAI_API_KEY")),
    "OPENAI_API_KEY not set"
  )

  result <- run_rag_search(
    query = "What is special education?",
    documents = c(
      "Special education provides support for students with disabilities.",
      "Individualized Education Programs guide special education services.",
      "Assistive technology is used in special education."
    ),
    provider = "openai",
    top_k = 2
  )

  expect_type(result, "list")
  expect_true("success" %in% names(result))
  expect_true("answer" %in% names(result))
  expect_true("confidence" %in% names(result))
  expect_true("sources" %in% names(result))

  if (result$success) {
    expect_type(result$answer, "character")
    expect_type(result$confidence, "double")
    expect_gte(result$confidence, 0)
    expect_lte(result$confidence, 1)
  }
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(message = FALSE, warning = FALSE)

## -----------------------------------------------------------------------------
library(TextAnalysisR)

tokens <- quanteda::tokens(SpecialEduTech$abstract[1:5])
dispersion <- calculate_lexical_dispersion(
  tokens,
  terms = c("learning", "instruction")
)
head(dispersion)

---
title: "Security"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Security}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

```{r}
library(TextAnalysisR)

mydata <- SpecialEduTech[seq_len(5), c("title", "abstract")]
united <- unite_cols(mydata, listed_vars = c("title", "abstract"))
toks   <- prep_texts(united, text_field = "united_texts")
quanteda::ndoc(toks)
```

TextAnalysisR includes built-in security features.

## Input Validation

| Feature | Description |
|---------|-------------|
| File uploads | Extension whitelist, 50MB limit, malicious content scanning |
| Text and LLM inputs | XSS and prompt injection filtering |
| Column names | Regex validation to prevent formula injection |

## API Key Security

- Stored via `.env` or environment variables (never logged or persisted)
- Masked input field with format validation
- Transmitted via secure headers only

Environment Variable: add `OPENAI_API_KEY=...` to `.Renviron`, or set in-session:

```{r, eval = FALSE}
Sys.setenv(OPENAI_API_KEY = "sk-...")
```

## Network Security

- Content Security Policy, X-Frame-Options, SRI for CDN resources
- HTTPS with TLS 1.2+ via Nginx/Cloudflare

## Data Protection

- Session-scoped with no persistent storage, cookies, or identifiers
- Rate limiting: 100 requests/hour per session
- Security event logging with sanitized error messages
- Local processing option (FERPA/HIPAA compatible)

## Infrastructure

- Cloudflare DNS with DDoS protection
- Docker + Nginx deployment

## Requirements

- `textdata` package
- Internet connection (first download only)

## Installation

```r
install.packages("textdata")

library(tidytext)
get_sentiments("bing")     # Positive/negative
get_sentiments("afinn")    # Scores -5 to +5
get_sentiments("nrc")      # Emotions + sentiment
```

Lexicons are cached locally after first download.

## Available Lexicons

- **Bing**: Binary positive/negative classification
- **AFINN**: Numeric sentiment intensity (-5 to +5)
- **NRC**: Multi-label with 8 emotions + sentiment

## Features Enabled

- Sentiment analysis with multiple lexicons
- Emotion detection (joy, anger, fear, etc.)
- Document-level sentiment scoring

---
title: "Multimodal Analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Multimodal Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

```{r}
library(TextAnalysisR)

sample_text <- c(
  "Figure 1 shows the distribution of student outcomes.",
  "Table 2 reports the effect sizes for each intervention."
)
toks <- prep_texts(
  data.frame(united_texts = sample_text),
  text_field = "united_texts"
)
quanteda::ntoken(toks)
```

Extract text from PDFs with charts, diagrams, and images using vision AI. R-native pipeline -- no Python required.

## How It Works

1. Extracts text from each page using `pdftools::pdf_text()` (R-native)
2. Renders each page as a PNG image via `pdftools::pdf_render_page()`
3. Identifies sparse-text pages (< 500 characters) that likely contain figures
4. Sends only those pages to a vision LLM for description
5. Merges extracted text + image descriptions into a single text corpus

## Functions

`process_pdf_unified()` runs the full pipeline with automatic fallback:

1. **Multimodal** (pdftools + vision LLM) -- extracts text and describes visual content
2. **Text-only** (pdftools) -- fallback when no vision provider is set

`describe_image()` describes a single base64-encoded PNG. Both require a vision-provider API key (OpenAI/Gemini) and network access; see their reference pages for usage.

## Provider Comparison

| Provider | Cost | Privacy | Accuracy | Setup |
|----------|------|---------|----------|-------|
| OpenAI | Per use | Cloud | Best | API key |
| Gemini | Free on hosted app (Google Cloud Research); otherwise per use | Cloud | Best | API key |

---
title: "Getting Started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

TextAnalysisR provides text analysis through an interactive Shiny app or R code.

## Install

```{r, eval = FALSE}
install.packages("TextAnalysisR",
  repos = c("https://mshin77.r-universe.dev", "https://cloud.r-project.org"))
```

## Launch App

```{r, eval = FALSE}
library(TextAnalysisR)
run_app()
```

Or visit [textanalysisr.org](https://www.textanalysisr.org) for the web version.

## Quick Example

```{r}
library(TextAnalysisR)

mydata <- SpecialEduTech
united_tbl <- unite_cols(mydata, listed_vars = c("title", "keyword", "abstract"))

tokens <- prep_texts(united_tbl, text_field = "united_texts")
dfm_object <- quanteda::dfm(tokens)

plot_word_frequency(dfm_object, n = 20)
```

## Features

| Category | Analyses |
|----------|----------|
| Lexical | Word frequency, keywords, networks |
| Semantic | Similarity, clustering, sentiment |
| Topic Modeling | STM and BERTopic |

## Author

Mikyung Shin, Ph.D.\
Assistant Professor, Department of Special Education \
Illinois State University

## Contact

- Email: [mshin2@ilstu.edu](mailto:mshin2@ilstu.edu)
- GitHub: [github.com/mshin77](https://github.com/mshin77)
- Homepage: [mshin77.net](https://mshin77.net/)

## Selected Publications

Shin, M., Deniz, F., Watson, L., Dieterich, C., Ewoldt, K., Johnson, F., Kong, J., Lee, S. H., & Whitehurst, A. (2026). Addressing the void of AI policies in education for students with specific learning disabilities. *Learning Disability Quarterly, 49*(3), 134-146. [DOI](https://doi.org/10.1177/07319487251412879)

Shin, M., Park, H., & Kang, E. (2025). Development of education policies and practices for students with learning disabilities in South Korea using Delphi surveys and topic modeling. *Learning Disability Quarterly, 48*(1), 13-25. [DOI](https://doi.org/10.1177/07319487241280393)

Shin, M., Ok, M. W., Choo, S., Hossain, G., Bryant, D. P., & Kang, E. (2023). A content analysis of research on technology use for teaching mathematics to students with disabilities: Word networks and topic modeling. *International Journal of STEM Education, 10*, Article 23. [DOI](https://doi.org/10.1186/s40594-023-00414-x)

---
title: "Python Environment"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Python Environment}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

```{r}
library(TextAnalysisR)

tokens <- quanteda::tokens(SpecialEduTech$abstract[1:5])
dispersion <- calculate_lexical_dispersion(
  tokens,
  terms = c("learning", "instruction")
)
head(dispersion)
```

Python enables features: NLP with spaCy, embeddings, and neural sentiment analysis.

## Quick Setup

`setup_python_env()` automatically:

1. Creates virtual environment `textanalysisr-env`
2. Installs `spacy` and `pdfplumber`
3. Downloads spaCy English model (`en_core_web_sm`)

Uses virtualenv (or conda if available).

## Check Status

Run `check_python_env()` to verify the environment.

## Common Issues

### "Another Python already initialized"

Set preferred environment in `.Rprofile` with `Sys.setenv(RETICULATE_PYTHON_ENV = "textanalysisr-env")`, then restart R.

### Environment in OneDrive

Avoid OneDrive paths. Use `setup_python_env(envname = "textanalysisr-env")`.

## spaCy Models

The default `en_core_web_sm` model is installed automatically. For word vectors (similarity), the medium model is 91 MB and the large model is 560 MB:

```bash
python -m spacy download en_core_web_md
python -m spacy download en_core_web_lg
```

## Deep Learning (Optional)

For embeddings and neural sentiment:

```bash
pip install sentence-transformers transformers torch
```

## Diagnostics

Use `reticulate::py_config()` and `reticulate::virtualenv_list()` to inspect the active Python.

## Input Validation

- File uploads: Extension whitelist, 50MB limit, malicious content scanning
- Text and LLM inputs: XSS and prompt injection filtering
- Column names: Regex validation to prevent formula injection

## API Key Security

- Stored via `.env` or environment variables (never logged or persisted)
- Masked input, format validation, transmitted via secure headers only

## Network Security

- Content Security Policy, X-Frame-Options, SRI for CDN resources
- HTTPS with TLS 1.2+ via Nginx/Cloudflare

## Data Protection

- Session-scoped with no persistent storage, cookies, or identifiers
- Rate limiting: 100 requests/hour per session
- Security event logging with sanitized error messages
- Local processing option (FERPA/HIPAA compatible)

## Infrastructure

- Cloudflare DNS with DDoS protection
- Docker + Nginx deployment

## Keyboard Support

- Tab / Shift+Tab: Navigate forward / backward
- Enter / Space: Activate button or link
- Alt+S: Text-to-speech toggle
- Esc: Close modals
- Skip to main content link available

Follows WCAG 2.1 Level AA standards with full keyboard navigation.

## Screen Reader

- ARIA landmarks and labels
- Live announcements for dynamic content
- Descriptive button and image labels

## Visual Features

- **Dark Mode**: Toggle via moon/sun icon
- **High Contrast**: 4.5:1+ ratio (WCAG AA)
- **Resizable Text**: Up to 200%
- **Reduced Motion**: Respects system preferences

## Translation & TTS

- Translation: 100+ languages (globe icon)
- Text-to-Speech: Auto language detection (speaker icon)

---
title: "Installation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Installation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

```{r}
library(TextAnalysisR)
packageVersion("TextAnalysisR")

data("SpecialEduTech")
nrow(SpecialEduTech)
head(SpecialEduTech[, c("title", "year")], 3)
```

## Choose a Setup

| Goal | Setup |
|------|-------|
| Small datasets in the browser | Web app at [textanalysisr.org](https://www.textanalysisr.org) |
| Full analyses and large datasets | `install.packages()` below |
| spaCy annotation and local transformer models | `setup_python_env()` after installing |

Python is optional. Core analyses run in plain R; cloud AI providers
(OpenAI, Gemini) can substitute for local models with an API key.

## R Package

```{r, eval = FALSE}
install.packages("TextAnalysisR",
  repos = c("https://mshin77.r-universe.dev", "https://cloud.r-project.org"))
library(TextAnalysisR)
run_app()
```

Requirements: R >= 4.0, RStudio recommended

## Web App

Visit [textanalysisr.org](https://www.textanalysisr.org) - no installation needed.

Note: Web version has limited features (no Python, no large files). Cloud AI providers (OpenAI, Gemini) are available with an API key.

## Optional Features

### Linguistic Analysis (spaCy)

For lemmatization, POS tagging, and named entity recognition, set up Python with `TextAnalysisR::setup_python_env()`.

### Python Features

For embeddings and neural sentiment analysis, run `setup_python_env()`. Requires Python 3.9+.

## Troubleshooting

| Issue | Solution |
|-------|----------|
| Package install failed | `remotes::install_github("mshin77/TextAnalysisR", dependencies = TRUE)` |
| Browser doesn't open | Navigate to URL shown in R console |
| Python errors | See [Python Environment](python_environment.html) |

---
title: "Lexical Analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Lexical Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

Lexical analysis examines word patterns, distinctiveness, and complexity. The sections below follow the
Shiny app's **Lexical Analysis** tabs in order.

## Setup

A 150-document subset of `SpecialEduTech` keeps the build fast; the full dataset works the same way.

```{r}
library(TextAnalysisR)

mydata <- SpecialEduTech[1:150, ]
united_tbl <- unite_cols(mydata, listed_vars = c("title", "keyword", "abstract"))
tokens <- prep_texts(united_tbl, text_field = "united_texts", remove_stopwords = TRUE)
dfm_object <- quanteda::dfm(tokens)
```

## Linguistic Annotation

Token-level annotation (lemmas, part-of-speech, morphology, dependencies, named entities) uses spaCy
through reticulate, so the examples below require Python and are not run here.

### Part-of-Speech Tags

`extract_pos_tags()` returns one row per token with `doc_id, token, lemma, pos, tag`. Universal POS tags
include NOUN, VERB, ADJ, ADV (content words), PROPN (proper nouns), and DET, ADP, PRON (function words).

```{r, eval = FALSE}
pos <- extract_pos_tags(united_tbl$united_texts)
```

### Morphological Features

`extract_morphology()` extracts grammatical features such as Number (Sing/Plur), Tense (Past/Pres/Fut),
VerbForm, Person, and Case.

```{r, eval = FALSE}
morphology <- extract_morphology(united_tbl$united_texts)
```

### Named Entity Recognition

`extract_named_entities()` tags entities such as PERSON, ORG, GPE/LOC, and DATE/MONEY/PERCENT.

```{r, eval = FALSE}
entities <- extract_named_entities(united_tbl$united_texts)
```

## Frequency Trends

`plot_word_frequency()` shows the most frequent terms in the document-feature matrix.

```{r}
plot_word_frequency(dfm_object, n = 20)
```

## Keywords

### TF-IDF

`extract_keywords_tfidf()` weights terms that are frequent in a document but rare across the corpus,
surfacing distinctive vocabulary.

```{r}
keywords <- extract_keywords_tfidf(dfm_object, top_n = 10)
plot_tfidf_keywords(keywords)
```

### Statistical Keyness

`extract_keywords_keyness()` identifies terms that distinguish one group from the rest using a
log-likelihood (G^2) statistic.

```{r}
keyness <- extract_keywords_keyness(
  dfm_object,
  target = quanteda::docvars(dfm_object, "reference_type") == "journal_article"
)
plot_keyness_keywords(keyness)
```

### Comparison

`plot_keyword_comparison()` places TF-IDF scores next to term frequency for the top keywords.

```{r}
plot_keyword_comparison(keywords, top_n = 10)
```

## Lexical Diversity

`lexical_diversity_analysis()` reports vocabulary-richness indices. MTLD, MATTR, and HDD are stable
across text lengths; TTR and CTTR are length-sensitive.

```{r}
diversity <- lexical_diversity_analysis(dfm_object)
plot_lexical_diversity_distribution(diversity$lexical_diversity, metric = "TTR")
```

| Metric | Description | Note |
|--------|-------------|------|
| TTR | Types / Tokens | Length-sensitive |
| CTTR | Types / sqrt(2 × Tokens) | Partly length-corrected |
| MATTR | Moving-average TTR | Stable across lengths |
| MTLD | Mean length maintaining TTR | Length-independent |
| HDD | Hypergeometric sampling probability | Length-independent, needs 42+ tokens |
| Maas | Log-based index | Lower = more diverse |

## Readability

`calculate_text_readability()` computes grade-level and reading-ease indices from sentence and word
structure.

```{r}
readability <- calculate_text_readability(united_tbl$united_texts)
plot_readability_distribution(readability, metric = "flesch")
```

| Metric | Basis | Output |
|--------|-------|--------|
| Flesch Reading Ease | Sentence length + syllables | 0-100 (higher = easier) |
| Flesch-Kincaid | Sentence length + syllables | Grade level |
| Gunning Fog | Sentence length + complex words | Years of education |
| SMOG | Polysyllabic words | Years of education |
| ARI | Characters per word | Grade level |
| Coleman-Liau | Letters per 100 words | Grade level |

## Log Odds Ratio

`calculate_log_odds_ratio()` compares word odds between categories using a Dirichlet-smoothed
log-odds ratio to find distinctive vocabulary.

```{r}
log_odds <- calculate_log_odds_ratio(
  dfm_object,
  group_var = "reference_type",
  comparison_mode = "binary",
  top_n = 15
)
plot_log_odds_ratio(log_odds)
```

`calculate_weighted_log_odds()` weights the ratio by a z-score (Monroe et al.), so reliably distinctive
terms rank above rare terms with extreme ratios (uses the tidylo package).

```{r}
weighted_odds <- calculate_weighted_log_odds(
  dfm_object,
  group_var = "reference_type",
  top_n = 15
)
plot_weighted_log_odds(weighted_odds)
```

## Lexical Dispersion

`calculate_lexical_dispersion()` shows where selected terms appear across documents (an X-ray plot).

```{r}
dispersion <- calculate_lexical_dispersion(tokens[1:50], terms = c("education", "technology"))
plot_lexical_dispersion(dispersion)
```

## Multi-Word Expressions

Multi-word (n-gram) detection belongs to the **Preprocess → Multi-Word Dictionary** step in the app.
`detect_multi_words()` returns a collocations table to feed `quanteda::tokens_compound()`.

```{r}
compounds <- detect_multi_words(tokens, min_count = 10)
head(compounds, 10)
```

<img src="logo.png" alt="TextAnalysisR Logo" align="right" width="150px"/>

[![R-CMD-check](https://github.com/mshin77/TextAnalysisR/workflows/R-CMD-check/badge.svg)](https://github.com/mshin77/TextAnalysisR/actions)
[![Project Status:
Active](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![License:
GPL-3](https://img.shields.io/badge/License-GPL%20v3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)

Text mining and natural language processing workflow for documents
(`PDF`, `DOCX`, `XLSX`, `CSV`, `TXT`). Includes preprocessing via
[quanteda](https://github.com/quanteda/quanteda), lexical analysis
(term frequency-inverse document frequency, log-odds ratios, lexical
diversity) via [tidytext](https://github.com/juliasilge/tidytext),
topic modeling via [stm](https://github.com/bstewart/stm) and
[BERTopic](https://maartengr.github.io/BERTopic/), semantic similarity
and document clustering on transformer embeddings, an interactive
[Shiny](https://shiny.posit.co/) interface with
[ggplot2](https://ggplot2.tidyverse.org/) visualization, optional
[spaCy](https://spacy.io/) lemmatization, and local
[sentence-transformers](https://www.sbert.net/) or web-based
(OpenAI,
[Gemini](https://ai.google.dev/)) model providers for
retrieval-augmented generation.

## Installation

    install.packages("TextAnalysisR",
      repos = c("https://mshin77.r-universe.dev", "https://cloud.r-project.org"))

## Load the TextAnalysisR Package

    library(TextAnalysisR)

## Alternatively, Launch and Browse the Shiny App

Access the web app at <https://www.textanalysisr.org>.

Launch and browse the app on the local computer:

    run_app()

## Getting Started

See [Quick
Start](https://mshin77.github.io/TextAnalysisR/articles/quickstart.html)
for tutorials.

## Citation

- Shin, M. (2026). *TextAnalysisR: A text mining workflow tool* (R
  package version 0.1.4) \[Computer software\].
  <https://mshin77.github.io/TextAnalysisR/>

- Shin, M. (2026). *TextAnalysisR: A text mining workflow tool* \[Web
  application\]. <https://www.textanalysisr.org>

## AI Integration

Gemini usage is free here, supported by the Google Cloud Research program. OpenAI requires your own API key.

TextAnalysisR provides AI/NLP capabilities via cloud-based providers.

### Providers

<table style="width: 100%; border-collapse: collapse; margin-bottom: 15px;">
<thead>
<tr style="border-bottom: 1px solid #dee2e6;">
<th style="text-align: left; padding: 8px;">Provider</th>
<th style="text-align: left; padding: 8px;">Type</th>
<th style="text-align: left; padding: 8px;">Best For</th>
</tr>
</thead>
<tbody>
<tr style="border-bottom: 1px solid #dee2e6;">
<td style="padding: 8px;">OpenAI</td>
<td style="padding: 8px;">Web-based</td>
<td style="padding: 8px;">Quality, speed</td>
</tr>
<tr style="border-bottom: 1px solid #dee2e6;">
<td style="padding: 8px;"><a href="https://ai.google.dev/" target="_blank">Gemini</a></td>
<td style="padding: 8px;">Web-based</td>
<td style="padding: 8px;">Quality, speed</td>
</tr>
</tbody>
</table>

### Features

- **Topic Labels**: AI-suggested labels from topic model terms
- **Content Generation**: Survey items, research questions, themes
- **Cluster Labels**: AI-suggested names for document clusters
- **RAG Search**: Question-answering over the document corpus
- **Vision AI**: PDF image/chart extraction via OpenAI or Gemini
- **Sentiment Analysis**: LLM-based or local transformer models
- **Linguistic Analysis**: POS, NER, dependency parsing via spaCy

### Setup

```r
Sys.setenv(OPENAI_API_KEY = "<openai-api-key>")
# or
Sys.setenv(GEMINI_API_KEY = "<gemini-api-key>")
```

### Responsible AI

All AI features follow human-in-the-loop design: AI suggests; review and approve.

See [AI Integration vignette](https://mshin77.github.io/TextAnalysisR/articles/ai_integration.html) for detailed documentation.

## Documentation

- [R Package Documentation](https://mshin77.github.io/TextAnalysisR/)
- [GitHub Repository](https://github.com/mshin77/TextAnalysisR)
- [BERTopic Documentation](https://maartengr.github.io/BERTopic/)
- [Sentence Transformers](https://www.sbert.net/)

## Applications

- **R Package** (full features): `install.packages("TextAnalysisR", repos = c("https://mshin77.r-universe.dev", "https://cloud.r-project.org"))`
- **Web App** (core features): [textanalysisr.org](https://www.textanalysisr.org)

## Use Policy

- Academic, research, and non-commercial use only
- Third-party services restricted to non-commercial purposes under GPL-3 license

## Support

- [GitHub Issues](https://github.com/mshin77/TextAnalysisR/issues)
- [Documentation](https://mshin77.github.io/TextAnalysisR)

---
title: "AI Integration"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{AI Integration}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

```{r}
library(TextAnalysisR)
packageVersion("TextAnalysisR")

mydata <- SpecialEduTech[seq_len(20), c("title", "abstract")]
united <- unite_cols(mydata, listed_vars = c("title", "abstract"))
toks   <- prep_texts(united, text_field = "united_texts")
dfm    <- quanteda::dfm(toks)
extract_keywords_tfidf(dfm, top_n = 10)
```

TextAnalysisR provides AI features via cloud-based providers.

On the hosted web app, Gemini usage is free, supported by the Google Cloud Research program. OpenAI calls require a personal API key.

## Providers

| Provider | Type | API Key | Best For |
|----------|------|---------|----------|
| OpenAI | Web-based | OPENAI_API_KEY | Quality, speed |
| [Gemini](https://ai.google.dev/) | Web-based | None on hosted app; otherwise GEMINI_API_KEY | Quality, speed |
| [spaCy](https://spacy.io/) | Local | None | Linguistic analysis |
| Transformers | Local | None | Embeddings, sentiment |

## Setup

Set keys via `.Renviron` (persistent) or `Sys.setenv()` (session). The cloud chat, embedding, and RAG functions (`call_llm_api()`, `call_openai_chat()`, `call_gemini_chat()`, `get_api_embeddings()`, `run_rag_search()`) require an API key and network access; see their reference pages for usage.

## Default Models

| Provider | Chat Model | Embedding Model |
|----------|------------|-----------------|
| OpenAI | gpt-4.1-mini | text-embedding-3-small |
| Gemini | gemini-2.5-flash | gemini-embedding-001 |
| Local | - | all-MiniLM-L6-v2 |

## Responsible AI Design

All AI features follow [NIST AI Risk Management Framework](https://www.nist.gov/itl/ai-risk-management-framework) principles:

| Principle | Implementation |
|-----------|----------------|
| Human oversight | AI suggests; review and approve |
| User control | Edit, regenerate, or override any output |
| Transparency | View prompts and parameters used |
| Privacy | Local sentence-transformers and spaCy options for sensitive data |
| Grounding | Content based on input data, not generic knowledge |

---
title: "Accessibility"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Accessibility}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

```{r}
library(TextAnalysisR)

mydata <- SpecialEduTech[seq_len(20), c("title", "abstract")]
united <- unite_cols(mydata, listed_vars = c("title", "abstract"))
toks   <- prep_texts(united, text_field = "united_texts")
dfm    <- quanteda::dfm(toks)
plot_word_frequency(dfm, n = 10)
```

TextAnalysisR meets WCAG 2.1 Level AA standards.

## Keyboard Navigation

| Key | Action |
|-----|--------|
| Tab / Shift+Tab | Move between elements |
| Enter / Space | Activate buttons |
| Esc | Close dialogs |

## Visual Features

- Dark mode toggle
- High contrast (4.5:1 ratio)
- 200% zoom support
- Reduced motion option

## Screen Reader Support

- Descriptive button labels
- Chart text descriptions
- Status announcements

## Language Support

- 100+ languages via Google Translate
- Auto-detected text-to-speech

---
title: "Topic Modeling"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Topic Modeling}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

Topic modeling discovers latent themes in a text collection. The Shiny app offers two types:

- **Structural Topic Model (STM):** word-probability topics that incorporate document metadata as
  covariates. Runs in R.
- **Embedding-based topics:** transformer embeddings reduced and clustered into topics. Best for short
  texts and multilingual content. Requires Python (BERTopic) or an embedding API key.

The sections below follow the app's **Topic Modeling** tabs in order, first with STM (run live on the
bundled data) and then with the embedding-based type.

## Setup

A 150-document subset of `SpecialEduTech` keeps the build fast; the full dataset works the same way.

```{r}
library(TextAnalysisR)

mydata <- SpecialEduTech[1:150, ]
united_tbl <- unite_cols(mydata, listed_vars = c("title", "keyword", "abstract"))
tokens <- prep_texts(united_tbl, text_field = "united_texts", remove_stopwords = TRUE)
dfm_object <- quanteda::dfm(tokens)
```

## Model Configuration

`find_optimal_k()` compares models across a range of K via `searchK` (semantic coherence, exclusivity,
held-out likelihood, residuals). The example uses a small document subset and a narrow K range to keep
the runtime short.

```{r}
k_search <- find_optimal_k(dfm_object, topic_range = 5:6)
names(k_search)
```

Fit a small STM for the examples below. `prevalence` makes topic proportions depend on document
metadata.

```{r}
out <- quanteda::convert(dfm_object, to = "stm")
out$meta$year <- as.numeric(out$meta$year)

model <- stm::stm(
  documents = out$documents,
  vocab = out$vocab,
  K = 5,
  prevalence = ~ reference_type + year,
  data = out$meta,
  max.em.its = 15,
  init.type = "Spectral",
  verbose = FALSE
)
```

See: [Structural Topic Model](https://www.structuraltopicmodel.com/) · [stm on CRAN](https://CRAN.R-project.org/package=stm)

## Word-Topic

`get_topic_terms()` returns the top terms per topic from the word-topic (beta) distribution.

```{r}
terms <- get_topic_terms(model, top_term_n = 10)
head(terms, 15)
```

## Content Generation

`generate_topic_content()` drafts topic-grounded content (survey items, research questions, theme
descriptions, policy recommendations, interview questions) from `get_topic_terms()` output. It calls an
LLM, so it needs an OpenAI or Gemini API key.

| Content type | Output |
|------|--------|
| `survey_item` | Likert-scale statement |
| `research_question` | Research question |
| `theme_description` | Qualitative theme summary |
| `policy_recommendation` | Action-oriented statement |
| `interview_question` | Open-ended question |

```{r, eval = FALSE}
labels <- generate_topic_labels(terms, provider = "openai")

content <- generate_topic_content(terms, content_type = "research_question", provider = "openai")
```

## Document-Topic

`calculate_topic_probability()` summarizes the corpus-level expected topic proportions (mean of the
per-document theta matrix).

```{r}
doc_topic <- calculate_topic_probability(model, top_n = 10, verbose = FALSE)
doc_topic
```

## Quotes

`stm::findThoughts()` returns the documents most representative of a topic.

```{r}
quotes <- stm::findThoughts(model, texts = united_tbl$united_texts, topics = 1, n = 3)
quotes
```

## Estimated Effects

`stm::estimateEffect()` regresses topic proportions on document covariates; `tidytext::tidy()` returns
the per-topic coefficient table shown in the app's Estimated Effects tab.

```{r}
prep <- stm::estimateEffect(1:5 ~ reference_type + year, model, metadata = out$meta, uncertainty = "Global")
head(tidytext::tidy(prep), 10)
```

## Categorical Covariates

`plot_topic_effects_categorical()` plots topic proportions across the levels of a categorical covariate.
`stm::plot.estimateEffect(..., method = "pointestimate", omit.plot = TRUE)` returns the model-based
estimate and 95% CI per topic and level, reshaped here into the columns the plot expects.

```{r}
ec <- stm::plot.estimateEffect(prep, "reference_type", method = "pointestimate",
                               model = model, omit.plot = TRUE)
effects_cat <- do.call(rbind, lapply(seq_along(ec$topics), function(i) {
  data.frame(topic = ec$topics[i], value = as.character(ec$uvals),
             proportion = as.numeric(ec$means[[i]]),
             lower = as.numeric(ec$cis[[i]][1, ]), upper = as.numeric(ec$cis[[i]][2, ]))
}))
plot_topic_effects_categorical(effects_cat)
```

## Continuous Covariates

`plot_topic_effects_continuous()` plots topic proportions across a continuous covariate. The same
`plot.estimateEffect()` call with `method = "continuous"` returns the estimate and CI over a grid of the
covariate.

```{r}
en <- stm::plot.estimateEffect(prep, "year", method = "continuous",
                               model = model, omit.plot = TRUE)
effects_cont <- do.call(rbind, lapply(seq_along(en$topics), function(i) {
  data.frame(topic = en$topics[i], value = en$x,
             proportion = as.numeric(en$means[[i]]),
             lower = as.numeric(en$ci[[i]][1, ]), upper = as.numeric(en$ci[[i]][2, ]))
}))
plot_topic_effects_continuous(effects_cont)
```

## Embedding-based Topics

The second type embeds documents with a transformer, reduces dimensionality, and clusters the
embeddings into topics. `get_best_embeddings()` produces the embeddings and `fit_embedding_model()`
clusters them; `generate_topic_labels()` drafts labels. These need Python (BERTopic) or an embedding API
key, so they are shown without running.

```{r, eval = FALSE}
embeddings <- get_best_embeddings(united_tbl$united_texts, provider = "openai")

topics <- fit_embedding_model(
  united_tbl$united_texts,
  method = "umap_hdbscan",
  n_topics = 10,
  precomputed_embeddings = embeddings
)

embedding_labels <- generate_topic_labels(topics$topic_terms, provider = "openai")
```

| Provider | Model | Notes |
|----------|-------|-------|
| Sentence Transformers (default) | all-MiniLM-L6-v2, all-mpnet-base-v2 | Requires Python |
| OpenAI | text-embedding-3-small, text-embedding-3-large | API key |
| Gemini | gemini-embedding-001 | API key |

R backend methods follow `{dimred}_{clustering}` (e.g., `"umap_dbscan"`, `"tsne_kmeans"`,
`"pca_hierarchical"`). See: [BERTopic](https://maartengr.github.io/BERTopic/) · [Sentence-BERT](https://www.sbert.net/)

## STM vs Embedding

| Feature | STM | Embedding |
|---------|-----|-----------|
| Speed | Fast | Medium |
| Metadata covariates | Yes | No |
| Short texts | Weaker | Strong |
| Multilingual | No | Yes |
| Dependencies | R only | Python or API |

## Features

### Available in Both Local & Web

- Text preprocessing and tokenization
- Linguistic analysis via spaCy
- Word frequency analysis
- Log odds ratio comparison between categories
- Lexical dispersion visualization
- Sentiment analysis (lexicon and neural)
- Topic modeling (STM)
- Co-occurrence networks with node attribute controls
- Correlation networks with node attribute controls
- CSV, XLSX, TXT, PDF, DOCX import

### Local R Package Only

- Large file support (>10MB)
- AI-suggested topic labels via local and web-based APIs
- Multimodal PDF extraction with vision AI (OpenAI/Gemini)

## Linguistic Analysis Setup

```r
# Python spaCy required
TextAnalysisR::setup_python_env()
```

See the Installation tab for detailed setup instructions.

## Semantic Analysis Setup

Install Python dependencies for BERTopic and transformer-based semantic analysis.

## Requirements

- Python >= 3.7
- pip or conda package manager

## Installation via Python

```bash
# Using pip
pip install bertopic sentence-transformers umap-learn hdbscan

# Using conda
conda install -c conda-forge bertopic sentence-transformers umap-learn hdbscan
```

## Installation via R

```r
library(reticulate)
py_install(c("bertopic", "sentence-transformers", "umap-learn", "hdbscan"))
```

## Features Enabled

- Neural topic modeling (BERTopic)
- Transformer-based embeddings
- Semantic clustering (UMAP + HDBSCAN)
- Document similarity analysis
- Interactive topic visualization

---
title: "Semantic Analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Semantic Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(message = FALSE, warning = FALSE)
```

Semantic analysis examines relationships of meaning between words and documents. The sections below
follow the Shiny app's **Semantic Analysis** tabs in order.

## Setup

A 150-document subset of `SpecialEduTech` keeps the build fast; the full dataset works the same way.

```{r}
library(TextAnalysisR)

mydata <- SpecialEduTech[1:150, ]
united_tbl <- unite_cols(mydata, listed_vars = c("title", "keyword", "abstract"))
tokens <- prep_texts(united_tbl, text_field = "united_texts", remove_stopwords = TRUE)
dfm_object <- quanteda::dfm(tokens)
```

## Word Co-occurrence

`word_co_occurrence_network()` builds a network of words that co-occur across documents, with community
detection and centrality metrics. Edges weight by raw counts; `edge_metric = "pmi"` corrects the bias
toward generic frequent words.

```{r}
network <- word_co_occurrence_network(
  dfm_object,
  co_occur_n = 50,
  top_node_n = 30,
  node_label_size = 22,
  community_method = "leiden"
)

network$plot
network$table
network$summary
```

## Word Correlation

`word_correlation_network()` connects words by phi correlation of their document co-occurrence patterns.

```{r}
corr_network <- word_correlation_network(
  dfm_object,
  common_term_n = 15,
  corr_n = 0.3,
  community_method = "leiden"
)

corr_network$plot
corr_network$table
```

## Document Similarity

`semantic_similarity_analysis()` compares documents by words, n-grams, or embeddings (embeddings require
Python). The example below uses word features on a subset and renders a cosine similarity heatmap.

```{r}
subset_texts <- united_tbl$united_texts[1:10]

similarity <- semantic_similarity_analysis(
  subset_texts,
  document_feature_type = "words",
  similarity_method = "cosine",
  verbose = FALSE
)

plot_similarity_heatmap(similarity$similarity_matrix, method_name = "Cosine")
```

| Method | Description | Requires |
|--------|-------------|----------|
| Words | Word-frequency vectors (bag-of-words) | none |
| N-grams | Word-sequence vectors | none |
| Embeddings | Transformer sentence vectors | Python |

## Comparative Analysis

Comparative analysis scores how similar documents in one category are to a reference category.
`extract_cross_category_similarities()` pulls cross-category pairs from a similarity matrix and
`analyze_similarity_gaps()` summarizes the differences. The example uses the first 30 documents.

```{r}
term_matrix <- as.matrix(dfm_object)[1:30, ]
normalized <- term_matrix / sqrt(rowSums(term_matrix^2))
sim_matrix <- normalized %*% t(normalized)

docs_data <- data.frame(
  display_name = paste0("doc", 1:30),
  reference_type = quanteda::docvars(dfm_object, "reference_type")[1:30]
)
dimnames(sim_matrix) <- list(docs_data$display_name, docs_data$display_name)

cross <- extract_cross_category_similarities(
  sim_matrix,
  docs_data,
  reference_category = "journal_article",
  category_var = "reference_type",
  id_var = "display_name"
)

gaps <- analyze_similarity_gaps(cross)
gaps$summary_stats
```

## Semantic Search

`run_rag_search()` retrieves the documents most relevant to a query using embedding similarity. It
requires an OpenAI or Gemini API key; see [AI Integration](ai_integration.html).

```{r, eval = FALSE}
results <- run_rag_search(
  query = "math intervention for students with disabilities",
  documents = united_tbl$united_texts,
  provider = "openai"
)
```

## Sentiment & Emotion

`sentiment_lexicon_analysis()` scores documents with the Bing, AFINN, or NRC lexicon. The Bing
example runs below.

```{r}
sentiment <- sentiment_lexicon_analysis(dfm_object, lexicon = "bing")
plot_sentiment_distribution(sentiment$document_sentiment)
```

NRC also yields discrete emotions for `plot_emotion_radar()`. NRC downloads through `textdata` behind a
license-gated prompt, so the emotion example is shown but not run.

```{r, eval = FALSE}
emotion <- sentiment_lexicon_analysis(dfm_object, lexicon = "nrc")
plot_emotion_radar(emotion$emotion_scores)
```

`plot_sentiment_by_category()` compares sentiment across a metadata category after joining it to the
scored documents. Transformer (`sentiment_embedding_analysis()`) and LLM (`analyze_sentiment_llm()`)
scoring require Python or an API key and are not run here.

```{r}
scored <- sentiment_lexicon_analysis(dfm_object, lexicon = "bing")$document_sentiment
scored$reference_type <- quanteda::docvars(dfm_object, "reference_type")[
  match(scored$document, quanteda::docnames(dfm_object))
]
plot_sentiment_by_category(scored, category_var = "reference_type")
```

## Document Groups

`cluster_embeddings()` groups documents from a feature matrix. K-means and hierarchical clustering run
in base R; the app's default UMAP + DBSCAN path requires Python. `generate_cluster_labels_auto()` labels
each group with its most distinctive terms.

```{r}
data_matrix <- as.matrix(dfm_object)

groups <- cluster_embeddings(data_matrix, method = "kmeans", n_clusters = 5, verbose = FALSE)
groups$n_clusters

labels <- generate_cluster_labels_auto(data_matrix, groups$clusters, method = "tfidf", n_terms = 3)
labels
```

A 2-D map of the groups uses `reduce_dimensions()` (PCA runs in base R; t-SNE and UMAP need their
packages).

```{r}
coords <- reduce_dimensions(data_matrix, method = "PCA", n_components = 2, verbose = FALSE)
head(coords$reduced_data)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{stm_15}
\alias{stm_15}
\title{An example structure of a structural topic model}
\format{
An object of class \code{STM} of length 11.
}
\usage{
stm_15
}
\description{
Contains 15 topics, topic prevalences, and etc. from \code{stm::stm}.
}
\examples{
\donttest{
stm_15
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{calculate_weighted_log_odds}
\alias{calculate_weighted_log_odds}
\title{Calculate Weighted Log Odds Ratio}
\usage{
calculate_weighted_log_odds(dfm_object, group_var, top_n = 10, min_count = 5)
}
\arguments{
\item{dfm_object}{A quanteda dfm object}

\item{group_var}{Character, name of the document variable to group by}

\item{top_n}{Number of top terms to return per group (default: 10)}

\item{min_count}{Minimum total count for a term to be included (default: 5)}
}
\value{
A data frame with the grouping variable, feature, n,
log_odds_weighted (from tidylo::bind_log_odds), and significant
(TRUE when |log_odds_weighted| >= 1.96)
}
\description{
Computes weighted log odds ratios using the method from Monroe, Colaresi,
and Quinn (2008) "Fightin' Words" via the tidylo package. This method
weights log odds by variance (z-score) to identify words that reliably
distinguish between groups, accounting for sampling variability.
}
\examples{
\donttest{
if (requireNamespace("tidylo", quietly = TRUE)) {
  articles <- TextAnalysisR::SpecialEduTech[1:20, ]
  dfm_object <- quanteda::dfm(quanteda::tokens(articles$abstract))
  quanteda::docvars(dfm_object, "reference_type") <- articles$reference_type
  weighted_odds <- calculate_weighted_log_odds(dfm_object, "reference_type",
                                                top_n = 5)
}
}
}
\references{
Monroe, B. L., Colaresi, M. P., & Quinn, K. M. (2008). Fightin' words:
Lexical feature selection and evaluation for identifying the content of
political conflict. Political Analysis, 16(4), 372-403.

Silge, J., & Robinson, D. (2017). Text mining with R: A tidy approach.
O'Reilly Media. https://www.tidytextmining.com/
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{calculate_clustering_metrics}
\alias{calculate_clustering_metrics}
\title{Calculate Clustering Quality Metrics}
\usage{
calculate_clustering_metrics(
  clusters,
  data_matrix,
  dist_matrix = NULL,
  metrics = "all"
)
}
\arguments{
\item{clusters}{Integer vector of cluster assignments}

\item{data_matrix}{Numeric matrix of data points (rows = observations, cols = features)}

\item{dist_matrix}{Optional distance matrix. If NULL, computed from data_matrix}

\item{metrics}{Character vector of metrics to calculate.
Options: "silhouette", "davies_bouldin", "calinski_harabasz", or "all" (default)}
}
\value{
A named list containing:
\describe{
\item{silhouette}{Silhouette score (-1 to 1, higher is better)}
\item{davies_bouldin}{Davies-Bouldin index (lower is better)}
\item{calinski_harabasz}{Calinski-Harabasz index (higher is better)}
\item{n_clusters}{Number of clusters}
\item{cluster_sizes}{Table of cluster sizes}
}
}
\description{
Calculates common clustering evaluation metrics including Silhouette Score,
Davies-Bouldin Index, and Calinski-Harabasz Index.
}
\details{
\itemize{
\item Silhouette Score: Measures how similar an object is to its own cluster compared
to other clusters. Range: -1 to 1, higher is better.
\item Davies-Bouldin Index: Average similarity between each cluster and its most
similar cluster. Lower values indicate better clustering.
\item Calinski-Harabasz Index: Ratio of between-cluster to within-cluster variance.
Higher values indicate better-defined clusters.
}
}
\examples{
\donttest{
abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:20]
term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(abstracts)))
kmeans_result <- stats::kmeans(term_matrix, centers = 2)
metrics <- calculate_clustering_metrics(kmeans_result$cluster, term_matrix)
print(metrics)
}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_readability_by_group}
\alias{plot_readability_by_group}
\title{Plot Readability by Group}
\usage{
plot_readability_by_group(readability_data, metric, group_var, title = NULL)
}
\arguments{
\item{readability_data}{Data frame from calculate_text_readability()}

\item{metric}{Metric to plot}

\item{group_var}{Name of grouping variable column}

\item{title}{Plot title (default: auto-generated)}
}
\value{
A plotly boxplot
}
\description{
Creates grouped boxplots comparing readability across categories.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{wrap_long_text}
\alias{wrap_long_text}
\title{Wrap Long Text with Line Breaks}
\usage{
wrap_long_text(text, chars_per_line = 50, max_lines = 3)
}
\arguments{
\item{text}{Character string to wrap.}

\item{chars_per_line}{Maximum characters per line (default: 50).}

\item{max_lines}{Maximum number of lines (default: 3).}
}
\value{
Text with line breaks inserted.
}
\description{
Wraps long text by inserting line breaks at word boundaries.
Handles both spaced text and continuous text (like URLs).
}
\concept{text-utilities}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{word_co_occurrence_network}
\alias{word_co_occurrence_network}
\title{Analyze and Visualize Word Co-occurrence Networks}
\usage{
word_co_occurrence_network(
  dfm_object,
  doc_var = NULL,
  co_occur_n = 50,
  edge_metric = c("count", "pmi"),
  top_node_n = 30,
  nrows = 1,
  height = 800,
  width = 900,
  node_label_size = 22,
  community_method = "leiden",
  node_size_by = "degree",
  node_color_by = "community",
  seed = 123,
  category_params = NULL
)
}
\arguments{
\item{dfm_object}{A quanteda document-feature matrix (dfm).}

\item{doc_var}{A document-level metadata variable (default: NULL).}

\item{co_occur_n}{Minimum co-occurrence count (default: 50). A heuristic
default; vary it to check edge-count sensitivity.}

\item{edge_metric}{Edge weight: "count" (default, raw co-occurrence) or
"pmi" (pointwise mutual information, correcting the bias toward generic
high-frequency words; only positive-PMI pairs are kept).}

\item{top_node_n}{Number of top nodes to display (default: 30).}

\item{nrows}{Number of rows to display in the table (default: 1).}

\item{height}{The height of the resulting Plotly plot, in pixels (default: 800).}

\item{width}{The width of the resulting Plotly plot, in pixels (default: 900).}

\item{node_label_size}{Maximum font size for node labels in pixels (default: 22).}

\item{community_method}{Community detection method: "leiden" (default) or "louvain".}

\item{node_size_by}{Node sizing method: "degree", "betweenness", "closeness", "eigenvector", or "fixed" (default: "degree").}

\item{node_color_by}{Node coloring method: "community" or "centrality" (default: "community").}

\item{seed}{Integer seed for the force-directed layout, so the plot is reproducible (default: 123).}

\item{category_params}{Optional named list of category-specific parameters. Each element should be a list with \code{co_occur_n} and \code{top_node_n} values for that category (default: NULL).}
}
\value{
A list containing the ggplot2 plot, a table, and a summary.
}
\description{
This function creates a word co-occurrence network based on a document-feature matrix (dfm).
}
\examples{
if (interactive()) {
  df <- TextAnalysisR::SpecialEduTech

  united_tbl <- TextAnalysisR::unite_cols(df, listed_vars = c("title", "abstract"))

  tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")

  dfm_object <- quanteda::dfm(tokens)

  word_co_occurrence_network_results <- TextAnalysisR::word_co_occurrence_network(
                                        dfm_object,
                                        doc_var = NULL,
                                        co_occur_n = 50,
                                        top_node_n = 30,
                                        nrows = 1,
                                        height = 800,
                                        width = 900,
                                        community_method = "leiden")
  print(word_co_occurrence_network_results$plot)
  print(word_co_occurrence_network_results$table)
  print(word_co_occurrence_network_results$summary)
}
}
\seealso{
Other semantic:
\code{\link[=word_correlation_network]{word_correlation_network()}}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{extract_topic_terms_df}
\alias{extract_topic_terms_df}
\title{Build a topic-term data frame from any supported topic model}
\usage{
extract_topic_terms_df(model, n = 7)
}
\arguments{
\item{model}{A topic model object (STM fit or embedding result).}

\item{n}{Number of top terms per topic (default 7).}
}
\value{
\code{data.frame(topic, term, beta)} in long format.
}
\description{
Unified helper that produces the long-format \code{data.frame(topic, term, beta)}
expected by \code{\link[=generate_topic_labels]{generate_topic_labels()}} from an STM model or an embedding result.
Dispatches on the object's structure:
\itemize{
\item STM model (has \verb{$beta$logbeta} and \verb{$vocab}) -> top terms via \code{\link[stm:labelTopics]{stm::labelTopics()}} FREX
\item Embedding result (has \verb{$topic_keywords}) -> c-TF-IDF keywords with rank-derived pseudo-beta
}
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{semantic_similarity_analysis}
\alias{semantic_similarity_analysis}
\title{Semantic Similarity Analysis}
\usage{
semantic_similarity_analysis(...)
}
\arguments{
\item{...}{Arguments passed to calculate_document_similarity}
}
\value{
Similarity analysis results
}
\description{
Wrapper for calculate_document_similarity
}
\seealso{
\code{\link[=calculate_document_similarity]{calculate_document_similarity()}} for the core computation; \code{\link[=plot_similarity_heatmap]{plot_similarity_heatmap()}} to render the resulting matrix
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{extract_text_from_pdf}
\alias{extract_text_from_pdf}
\title{Extract Text from PDF}
\usage{
extract_text_from_pdf(file_path)
}
\arguments{
\item{file_path}{Character string path to PDF file}
}
\value{
Data frame with columns: page (integer), text (character)
Returns NULL if extraction fails or PDF is empty
}
\description{
Extracts text content from a PDF file using pdftools package.
}
\details{
Uses \code{pdftools::pdf_text()} to extract text from each page.
Preserves page structure and cleans whitespace.
Works best with text-based PDFs (not scanned images).
}
\concept{pdf}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{get_content_type_prompt}
\alias{get_content_type_prompt}
\title{Get Default System Prompt for Content Type}
\usage{
get_content_type_prompt(content_type)
}
\arguments{
\item{content_type}{Type of content to generate.}
}
\value{
Character string with the system prompt.
}
\description{
Returns the default system prompt for a given content type.
}
\concept{ai}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{process_pdf_file}
\alias{process_pdf_file}
\title{Process PDF File}
\usage{
process_pdf_file(file_path, content_type = "auto")
}
\arguments{
\item{file_path}{Character string path to PDF file}

\item{content_type}{Character string: "auto" or "text" (default: "auto")}
}
\value{
List with:
\itemize{
\item data: Data frame with extracted content
\item type: Character string indicating content type ("text" or "error")
\item success: Logical indicating success
\item message: Character string with status message
}
}
\description{
Main function to process PDF files - extracts text content using pdftools.
For table extraction, use \code{\link{process_pdf_file_py}}.
}
\details{
This function extracts text content from PDFs using pdftools package.
Works best with text-based PDFs (not scanned images).

For PDFs containing tables or complex layouts, use the Python-based
\code{\link{process_pdf_file_py}} which provides better table extraction.
}
\concept{pdf}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{parse_morphology_string}
\alias{parse_morphology_string}
\title{Parse Morphology String}
\usage{
parse_morphology_string(data, features = NULL)
}
\arguments{
\item{data}{Data frame with a 'morph' column from spaCy parsing.}

\item{features}{Character vector of feature names (ignored, kept for
backwards compatibility). All features are always extracted.}
}
\value{
Data frame with additional morph_* columns for each feature.
}
\description{
Parse spaCy's morphology string format into individual columns.
Used internally by morphology analysis functions.
Always extracts all common morphology features (Number, Tense, VerbForm,
Person, Case, Mood, Aspect) regardless of the features parameter.
}
\concept{lexical}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{plot_model_comparison}
\alias{plot_model_comparison}
\title{Plot Topic Model Comparison Scatter}
\usage{
plot_model_comparison(search_results, title = NULL, height = 600, width = 800)
}
\arguments{
\item{search_results}{Results from stm::searchK or find_optimal_k()}

\item{title}{Plot title (default \code{NULL} selects an auto-title based on which metric columns are present in \code{search_results})}

\item{height}{Plot height in pixels (default: 600)}

\item{width}{Plot width in pixels (default: 800)}
}
\value{
A plotly scatter plot
}
\description{
Creates a scatter plot comparing topic model metrics across K values.
Automatically selects the best available metric combination.
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_docker_deployment}
\alias{check_docker_deployment}
\title{Check Docker Deployment}
\usage{
check_docker_deployment()
}
\value{
Logical TRUE if running in Docker
}
\description{
Detects whether the app is running in a Docker container.
Docker deployments have full Python/spaCy capability.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_topic_probability}
\alias{calculate_topic_probability}
\title{Calculate Topic Probabilities}
\usage{
calculate_topic_probability(stm_model, top_n = 10, verbose = TRUE, ...)
}
\arguments{
\item{stm_model}{A fitted STM model object from stm::stm().}

\item{top_n}{Number of top topics to display by prevalence (default: 10).}

\item{verbose}{Logical, if TRUE prints progress messages (default: TRUE).}

\item{...}{Additional arguments passed to tidytext::tidy().}
}
\value{
A DT::datatable showing topics and their mean gamma (prevalence) values,
rounded to 3 decimal places.
}
\description{
Extracts and summarizes topic probabilities (gamma values) from an STM model,
returning a formatted data table of mean topic prevalence.
}
\examples{
if (interactive()) {
  data <- TextAnalysisR::SpecialEduTech
  united <- unite_cols(data, c("title", "keyword", "abstract"))
  tokens <- prep_texts(united, text_field = "united_texts")
  dfm_obj <- quanteda::dfm(tokens)
  stm_data <- quanteda::convert(dfm_obj, to = "stm")

  topic_model <- stm::stm(
    documents = stm_data$documents,
    vocab = stm_data$vocab,
    K = 10,
    verbose = FALSE
  )

  prob_table <- calculate_topic_probability(topic_model, top_n = 10)
  print(prob_table)
}
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TextAnalysisR-package.R
\docType{package}
\name{TextAnalysisR-package}
\alias{TextAnalysisR}
\alias{TextAnalysisR-package}
\title{TextAnalysisR: A Text Mining Workflow Tool}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides a text mining and natural language processing workflow for documents. Includes preprocessing via 'quanteda', lexical analysis (term frequency-inverse document frequency, log-odds ratios, lexical diversity) via 'tidytext', topic modeling via 'stm' and the 'BERTopic' approach, semantic similarity and document clustering on transformer representations, an interactive 'Shiny' interface with 'ggplot2' visualization, optional 'spaCy' preprocessing, and local 'sentence-transformers' or web-based ('OpenAI', 'Gemini') model providers for retrieval-augmented generation, as described in Shin et al. (2026) \doi{10.1177/07319487251412879}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://mshin77.github.io/TextAnalysisR/}
  \item \url{https://github.com/mshin77/TextAnalysisR}
  \item Report bugs at \url{https://github.com/mshin77/TextAnalysisR/issues}
}

}
\author{
\strong{Maintainer}: Mikyung Shin \email{shin.mikyung@gmail.com} (\href{https://orcid.org/0000-0001-7907-9193}{ORCID})

Authors:
\itemize{
  \item Mikyung Shin \email{shin.mikyung@gmail.com} (\href{https://orcid.org/0000-0001-7907-9193}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{extract_tables_from_pdf_py}
\alias{extract_tables_from_pdf_py}
\title{Extract Tables from PDF using Python}
\usage{
extract_tables_from_pdf_py(file_path, pages = NULL, envname = NULL)
}
\arguments{
\item{file_path}{Character string path to PDF file}

\item{pages}{Integer vector of page numbers to process (NULL = all pages)}

\item{envname}{Character string, name of Python virtual environment
(default: "textanalysisr-env")}
}
\value{
Data frame with extracted table data
Returns NULL if no tables found or extraction fails
}
\description{
Extracts tabular data from PDF using pdfplumber (Python).
No Java required - pure Python solution.
}
\details{
Uses pdfplumber Python library through reticulate.
Works with complex table layouts without Java dependency.
}
\examples{
if (interactive()) {
setup_python_env()

pdf_path <- "path/to/table_document.pdf"
table_data <- extract_tables_from_pdf_py(pdf_path)
head(table_data)
}
}
\concept{pdf}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{sentiment_embedding_analysis}
\alias{sentiment_embedding_analysis}
\title{Embedding-based Sentiment Analysis}
\usage{
sentiment_embedding_analysis(
  texts,
  embeddings = NULL,
  model_name = "distilbert-base-uncased-finetuned-sst-2-english",
  doc_names = NULL,
  use_gpu = FALSE
)
}
\arguments{
\item{texts}{Character vector of texts to analyze}

\item{embeddings}{Optional pre-computed embedding matrix (from generate_embeddings)}

\item{model_name}{Sentiment model name (default: "distilbert-base-uncased-finetuned-sst-2-english")}

\item{doc_names}{Optional document names/IDs}

\item{use_gpu}{Whether to use GPU if available (default: FALSE)}
}
\value{
A list containing:
\describe{
\item{document_sentiment}{Data frame with document-level sentiment scores and classifications}
\item{emotion_scores}{NULL (emotion detection not currently supported for embeddings)}
\item{summary_stats}{Summary statistics including document counts and average scores}
\item{model_used}{Name of the transformer model used}
\item{feature_type}{"embeddings"}
}
}
\description{
Performs sentiment analysis using transformer-based embeddings and neural models.
This approach uses pre-trained language models for contextual sentiment detection
without requiring sentiment lexicons. Particularly effective for handling:
\itemize{
\item Complex contextual sentiment
\item Implicit sentiment and sarcasm
\item Domain-specific sentiment
\item Negation and intensifiers (automatically handled by the model)
}
}
\examples{
if (interactive()) {
  abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
  embedding_sentiment <- sentiment_embedding_analysis(abstracts)
  print(embedding_sentiment$document_sentiment)
  print(embedding_sentiment$summary_stats)
}
}
\concept{sentiment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_no_feature_matrix_notification}
\alias{show_no_feature_matrix_notification}
\title{Show no-feature-matrix notification}
\usage{
show_no_feature_matrix_notification(duration = 7)
}
\arguments{
\item{duration}{Seconds until auto-dismiss.}
}
\value{
Invisibly the notification id.
}
\description{
Show no-feature-matrix notification
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_semantic_drift}
\alias{calculate_semantic_drift}
\title{Calculate Semantic Drift}
\usage{
calculate_semantic_drift(temporal_results, ...)
}
\arguments{
\item{temporal_results}{Temporal analysis results}

\item{...}{Additional parameters}
}
\value{
List containing drift metrics
}
\description{
Calculates semantic drift across time periods
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_coherence}
\alias{calculate_coherence}
\title{Validate Semantic Coherence}
\usage{
calculate_coherence(embeddings, topic_assignments)
}
\arguments{
\item{embeddings}{Document embeddings.}

\item{topic_assignments}{Topic assignments.}
}
\value{
Coherence metrics.
}
\description{
Validates semantic coherence of topic assignments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{check_vision_models}
\alias{check_vision_models}
\title{Check Vision Model Availability}
\usage{
check_vision_models(provider = "gemini", api_key = NULL)
}
\arguments{
\item{provider}{Character: "openai" or "gemini"}

\item{api_key}{Character: API key (for OpenAI/Gemini)}
}
\value{
List with availability status and recommendations
}
\description{
Check if required vision models are available for multimodal processing.
}
\examples{
\donttest{
status <- check_vision_models("openai", api_key = Sys.getenv("OPENAI_API_KEY"))
status <- check_vision_models("gemini", api_key = Sys.getenv("GEMINI_API_KEY"))
}
}
\concept{pdf}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_dt_options}
\alias{get_dt_options}
\title{Default DT datatable options}
\usage{
get_dt_options(scroll_y = "400px", page_length = 25, show_buttons = TRUE)
}
\arguments{
\item{scroll_y}{CSS height for vertical scrolling.}

\item{page_length}{Default page length.}

\item{show_buttons}{Include copy/csv/excel/pdf/print buttons.}
}
\value{
A list for \code{DT::datatable(options = ...)}.
}
\description{
Default DT datatable options
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_topic_cluster_correspondence}
\alias{calculate_topic_cluster_correspondence}
\title{Calculate Topic-Cluster Correspondence}
\usage{
calculate_topic_cluster_correspondence(topic_keywords, cluster_keywords, ...)
}
\arguments{
\item{topic_keywords}{Topic keywords list.}

\item{cluster_keywords}{Cluster keywords list.}

\item{...}{Additional parameters (currently unused).}
}
\value{
List with correspondence metrics.
}
\description{
Computes Jaccard-based correspondence between topic keywords and cluster keywords.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_wcag_contrast}
\alias{check_wcag_contrast}
\title{Check WCAG Contrast Compliance}
\usage{
check_wcag_contrast(foreground, background, large_text = FALSE)
}
\arguments{
\item{foreground}{Foreground color (hex format)}

\item{background}{Background color (hex format)}

\item{large_text}{Logical, TRUE if text is large (18pt+ or 14pt+ bold)}
}
\value{
Logical TRUE if compliant, FALSE if not
}
\description{
Validates if color combination meets WCAG 2.1 Level AA contrast requirements.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_api_embeddings}
\alias{get_api_embeddings}
\title{Get Embeddings from API}
\usage{
get_api_embeddings(
  texts,
  provider = c("openai", "gemini"),
  model = NULL,
  api_key = NULL,
  batch_size = 100
)
}
\arguments{
\item{texts}{Character vector of texts to embed}

\item{provider}{Character string: "openai" or "gemini"}

\item{model}{Character string specifying the embedding model. Defaults:
\itemize{
\item openai: "text-embedding-3-small"
\item gemini: "gemini-embedding-001"
}}

\item{api_key}{Character string with API key}

\item{batch_size}{Integer, number of texts to embed per API call (default: 100)}
}
\value{
Matrix with embeddings (rows = texts, columns = dimensions)
}
\description{
Generates text embeddings using OpenAI or Gemini embedding APIs.
}
\concept{ai}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{plot_similarity_heatmap}
\alias{plot_similarity_heatmap}
\title{Plot Document Similarity Heatmap}
\usage{
plot_similarity_heatmap(
  similarity_matrix,
  docs_data = NULL,
  feature_type = "words",
  method_name = "Cosine",
  title = NULL,
  category_filter = NULL,
  doc_id_var = NULL,
  colorscale = NULL,
  height = 600,
  width = NULL,
  row_category = NULL,
  col_categories = NULL,
  category_var = "category_display",
  show_values = FALSE,
  facet = NULL,
  row_label = NULL,
  output_type = "plotly"
)
}
\arguments{
\item{similarity_matrix}{A square numeric matrix of similarity scores}

\item{docs_data}{Optional data frame with document metadata containing:
\itemize{
\item \code{document_number}: Document identifiers for axis labels
\item \code{document_id_display}: Document IDs for hover text
\item \code{category_display}: Category labels for hover text
}}

\item{feature_type}{Feature space type: "words", "topics", "ngrams", or "embeddings"
(determines colorscale and display name)}

\item{method_name}{Similarity method name for display (default: "Cosine")}

\item{title}{Plot title (default: NULL, auto-generated from feature_type)}

\item{category_filter}{Optional category filter label for title (default: NULL)}

\item{doc_id_var}{Name of document ID variable (affects label text, default: NULL)}

\item{colorscale}{Plotly colorscale override (default: NULL, uses feature_type default)}

\item{height}{Plot height in pixels (default: 600)}

\item{width}{Plot width in pixels (default: NULL for auto)}

\item{row_category}{Category for row documents in cross-category mode (default: NULL)}

\item{col_categories}{Character vector of categories for column documents (default: NULL)}

\item{category_var}{Name of category variable in docs_data (default: "category_display")}

\item{show_values}{Logical; show similarity values as text on tiles (default: FALSE)}

\item{facet}{Logical; facet by column categories (default: TRUE when col_categories specified)}

\item{row_label}{Label for row axis (default: NULL, uses row_category)}

\item{output_type}{Output type: "plotly" or "ggplot" (default: "plotly", auto-switches to "ggplot" for faceting)}
}
\value{
A ggplot2 heatmap object
}
\description{
Creates an interactive heatmap visualization of document similarity matrices
with support for document metadata, feature-specific colorscales, and rich tooltips.
Supports both symmetric (all-vs-all) and cross-category comparison modes.
}
\examples{
\donttest{
articles <- TextAnalysisR::SpecialEduTech[1:5, ]
term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(articles$abstract)))
normalized_matrix <- term_matrix / sqrt(rowSums(term_matrix ^ 2))
similarity_matrix <- normalized_matrix \%*\% t(normalized_matrix)
plot_similarity_heatmap(similarity_matrix)

document_metadata <- data.frame(
  document_number     = paste("Doc", 1:5),
  document_id_display = articles$title,
  category_display    = articles$reference_type
)
plot_similarity_heatmap(similarity_matrix, docs_data = document_metadata,
                        feature_type = "embeddings")

plot_similarity_heatmap(
  similarity_matrix,
  docs_data      = document_metadata,
  row_category   = "thesis",
  col_categories = "journal_article",
  show_values    = TRUE,
  facet          = TRUE
)
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_sentiment_color}
\alias{get_sentiment_color}
\title{Generate Sentiment Color Gradient}
\usage{
get_sentiment_color(score)
}
\arguments{
\item{score}{Numeric sentiment score (typically -1 to 1)}
}
\value{
Hex color string
}
\description{
Generates a color based on sentiment score using a gradient from red (negative)
through gray (neutral) to green (positive).
}
\examples{
get_sentiment_color(-0.8)  # Red
get_sentiment_color(0)     # Gray
get_sentiment_color(0.8)   # Green
}
\concept{visualization}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{calculate_cosine_similarity}
\alias{calculate_cosine_similarity}
\title{Calculate Cosine Similarity Matrix}
\usage{
calculate_cosine_similarity(matrix_data)
}
\arguments{
\item{matrix_data}{A numeric matrix where rows represent documents/observations}
}
\value{
A square similarity matrix with values between -1 and 1
}
\description{
Calculates the cosine similarity between all pairs of rows in a matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_tfidf_keywords}
\alias{plot_tfidf_keywords}
\title{Plot TF-IDF Keywords}
\usage{
plot_tfidf_keywords(tfidf_data, title = NULL, normalized = FALSE)
}
\arguments{
\item{tfidf_data}{Data frame from extract_keywords_tfidf()}

\item{title}{Plot title (default: "Top Keywords by TF-IDF Score")}

\item{normalized}{Logical, whether scores are normalized (for label) (default: FALSE)}
}
\value{
A plotly bar chart
}
\description{
Creates a horizontal bar plot of top keywords by TF-IDF score.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{render_displacy_dep}
\alias{render_displacy_dep}
\title{Render displaCy Dependency Visualization}
\usage{
render_displacy_dep(text, compact = TRUE, model = "en_core_web_sm")
}
\arguments{
\item{text}{Character string to visualize.}

\item{compact}{Logical; use compact mode for space (default: TRUE).}

\item{model}{spaCy model name (default: "en_core_web_sm").}
}
\value{
SVG string with dependency tree.
}
\description{
Renders spaCy's displaCy dependency visualization as SVG.
Shows syntactic structure with arrows between words.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_sentiment_colors}
\alias{get_sentiment_colors}
\title{Get Sentiment Color Palette}
\usage{
get_sentiment_colors()
}
\value{
Named vector of colors
}
\description{
Returns standardized color mapping for sentiment analysis.
}
\concept{visualization}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{plot_quality_metrics}
\alias{plot_quality_metrics}
\title{Plot Topic Model Quality Metrics}
\usage{
plot_quality_metrics(search_results)
}
\arguments{
\item{search_results}{Results from stm::searchK or find_optimal_k()}
}
\value{
A named list of ggplot objects, one per available metric
(possible keys: semcoh, residual, heldout, lbound).
}
\description{
Creates individual diagnostic metric plots across different K values
from stm::searchK results.
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{calculate_document_similarity}
\alias{calculate_document_similarity}
\title{Calculate Document Similarity}
\usage{
calculate_document_similarity(
  texts,
  document_feature_type = "words",
  semantic_ngram_range = 2,
  similarity_method = "cosine",
  use_embeddings = FALSE,
  embedding_model = "all-MiniLM-L6-v2",
  calculate_metrics = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{texts}{A character vector of texts to compare.}

\item{document_feature_type}{Feature extraction type: "words", "ngrams", or "embeddings".}

\item{semantic_ngram_range}{Integer, n-gram range for ngram features (default: 2).}

\item{similarity_method}{Similarity calculation method: "cosine", "jaccard", "euclidean", "manhattan".}

\item{use_embeddings}{Logical, use embedding-based similarity (default: FALSE).}

\item{embedding_model}{Sentence transformer model name (default: "all-MiniLM-L6-v2").}

\item{calculate_metrics}{Logical, compute similarity metrics (default: TRUE).}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A list containing:
\describe{
\item{similarity_matrix}{N x N similarity matrix}
\item{feature_matrix}{Document feature matrix used for calculation}
\item{method_info}{Information about the method used}
\item{metrics}{Similarity metrics (if calculate_metrics = TRUE)}
\item{execution_time}{Time taken for analysis}
}
}
\description{
Calculates similarity between documents using traditional NLP methods or
modern embedding-based approaches. Metrics are automatically
computed unless disabled.
}
\examples{
\donttest{
  data(SpecialEduTech)
  texts <- SpecialEduTech$abstract[1:5]

  result <- calculate_document_similarity(
    texts = texts,
    document_feature_type = "words",
    similarity_method = "cosine"
  )

  print(result$similarity_matrix)
  print(result$metrics)
}
}
\concept{semantic}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{summarize_morphology}
\alias{summarize_morphology}
\title{Summarize Morphology Features}
\usage{
summarize_morphology(data, features = NULL)
}
\arguments{
\item{data}{Data frame with morph_* columns from extract_morphology().}

\item{features}{Character vector of features to summarize.
If NULL, all available morph_* columns are used.}
}
\value{
A data frame with Feature, Value, Count, and Percentage columns.
}
\description{
Creates a summary table of morphological feature distributions
with counts and percentages for each feature value.
}
\examples{
if (interactive()) {
  tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1])
  morphology_data <- extract_morphology(tokens)
  summary_table <- summarize_morphology(morphology_data)
  print(summary_table)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_rate_limit}
\alias{check_rate_limit}
\title{Check Rate Limit}
\usage{
check_rate_limit(
  session_token,
  user_requests,
  max_requests = 100,
  window_seconds = 3600
)
}
\arguments{
\item{session_token}{Shiny session token}

\item{user_requests}{Reactive value storing request history}

\item{max_requests}{Maximum requests allowed in time window}

\item{window_seconds}{Time window in seconds (default: 3600 = 1 hour)}
}
\value{
TRUE if within limit, stops with error if exceeded
}
\description{
Check Rate Limit
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{fit_temporal_model}
\alias{fit_temporal_model}
\title{Fit Temporal Topic Model}
\usage{
fit_temporal_model(
  texts,
  dates,
  time_windows = "yearly",
  embeddings = NULL,
  verbose = TRUE
)
}
\arguments{
\item{texts}{A character vector of text documents to analyze.}

\item{dates}{A vector of dates corresponding to each document (will be converted to Date).}

\item{time_windows}{Time grouping strategy: "yearly", "monthly", or "quarterly" (default: "yearly").}

\item{embeddings}{Optional pre-computed embeddings matrix. If NULL, embeddings will be generated.}

\item{verbose}{Logical indicating whether to print progress messages (default: TRUE).}
}
\value{
A list containing temporal analysis results with topic evolution patterns.
}
\description{
Analyzes how topics evolve over time by fitting topic models to
different time periods and tracking semantic changes.
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{extract_keywords_tfidf}
\alias{extract_keywords_tfidf}
\title{Extract Keywords Using TF-IDF}
\usage{
extract_keywords_tfidf(dfm, top_n = 20, normalize = FALSE)
}
\arguments{
\item{dfm}{A quanteda dfm object}

\item{top_n}{Number of top keywords to extract (default: 20)}

\item{normalize}{Logical, whether to normalize TF-IDF scores to 0-1 range (default: FALSE)}
}
\value{
Data frame with columns: Keyword, TF_IDF_Score, Frequency
}
\description{
Extracts top keywords from a document-feature matrix using TF-IDF weighting.
Uses quanteda's default scheme (raw term counts, base-10 idf), which
differs from tidytext's proportion-based tf; rankings favor terms in
longer documents.
}
\examples{
\donttest{
  library(quanteda)
  corp <- corpus(c("text analysis", "data mining", "text mining"))
  dfm_obj <- dfm(tokens(corp))
  keywords <- extract_keywords_tfidf(dfm_obj, top_n = 5)
  print(keywords)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{wrap_text_for_tooltip}
\alias{wrap_text_for_tooltip}
\title{Wrap text for tooltip display}
\usage{
wrap_text_for_tooltip(
  text,
  max_words = 150,
  chars_per_line = 50,
  max_lines = 3
)
}
\arguments{
\item{text}{Input string.}

\item{max_words}{Maximum words to keep.}

\item{chars_per_line}{Approximate line width.}

\item{max_lines}{Maximum lines.}
}
\value{
A string with HTML \verb{<br>} breaks.
}
\description{
Wrap text for tooltip display
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{apply_standard_plotly_layout}
\alias{apply_standard_plotly_layout}
\title{Apply Standard Plotly Layout}
\usage{
apply_standard_plotly_layout(
  plot,
  title = NULL,
  xaxis_title = NULL,
  yaxis_title = NULL,
  margin = list(t = 60, b = 80, l = 80, r = 40),
  show_legend = FALSE
)
}
\arguments{
\item{plot}{A plotly plot object}

\item{title}{Plot title text (optional)}

\item{xaxis_title}{X-axis title (optional)}

\item{yaxis_title}{Y-axis title (optional)}

\item{margin}{List of margins: list(t, b, l, r) in pixels (default: list(t = 60, b = 80, l = 80, r = 40))}

\item{show_legend}{Logical, whether to show legend (default: FALSE)}
}
\value{
A plotly plot object with standardized layout
}
\description{
Applies consistent layout styling to plotly plots following TextAnalysisR design standards.
This ensures all plots have uniform fonts, colors, margins, and interactive features.
}
\details{
Design standards applied:
\itemize{
\item Title: 14px Roboto, #0c1f4a
\item Axis titles: 13px Roboto, #0c1f4a
\item Axis tick labels: 12px Roboto, #3B3B3B
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{check_multimodal_prerequisites}
\alias{check_multimodal_prerequisites}
\title{Check Multimodal Prerequisites}
\usage{
check_multimodal_prerequisites(
  vision_provider = "gemini",
  vision_model = NULL,
  api_key = NULL,
  envname = "textanalysisr-env"
)
}
\arguments{
\item{vision_provider}{Character: "openai" or "gemini"}

\item{vision_model}{Character: Model name (optional)}

\item{api_key}{Character: API key for OpenAI/Gemini}

\item{envname}{Character: Kept for backward compatibility, ignored}
}
\value{
List with:
\itemize{
\item ready: Logical - TRUE if all prerequisites met
\item missing: Character vector of missing components
\item instructions: Character - Detailed setup instructions
\item details: List with component-specific status
}
}
\description{
Checks all prerequisites for multimodal PDF extraction.
Uses R-native pdftools for rendering (no Python required).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{plot_topic_effects_categorical}
\alias{plot_topic_effects_categorical}
\title{Plot Topic Effects for Categorical Variables}
\usage{
plot_topic_effects_categorical(
  effects_data,
  ncol = 2,
  height = 800,
  width = 1000,
  title = "Category Effects",
  base_font_size = 11
)
}
\arguments{
\item{effects_data}{Data frame with columns: topic, value, proportion, lower, upper}

\item{ncol}{Number of columns for faceting (default: 2)}

\item{height}{Plot height in pixels (default: 800)}

\item{width}{Plot width in pixels (default: 1000)}

\item{title}{Plot title (default: "Category Effects")}

\item{base_font_size}{Base font size in points for the plot theme (default: 11). Axis text and strip text will be base_font_size + 2.}
}
\value{
A plotly object
}
\description{
Creates a faceted plot showing how categorical variables affect topic proportions.
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{find_topic_matches}
\alias{find_topic_matches}
\title{Find Similar Topics}
\usage{
find_topic_matches(
  topic_model,
  query,
  top_n = 10,
  method = "cosine",
  embedding_model = "all-MiniLM-L6-v2",
  include_terms = TRUE
)
}
\arguments{
\item{topic_model}{A topic model object (semantic topic model or STM model).}

\item{query}{A character string representing the query topic.}

\item{top_n}{The number of similar topics to return (default: 10).}

\item{method}{The similarity method: "cosine", "euclidean", "embedding".}

\item{embedding_model}{The embedding model to use for query encoding (default: "all-MiniLM-L6-v2").}

\item{include_terms}{Logical, whether to include topic terms in the similarity calculation (default: TRUE).}
}
\value{
A list containing similar topics and their similarity scores.
}
\description{
This function finds the most similar topics to a given query using semantic similarity analysis.
It works with both semantic topic models and traditional STM models by creating topic representations
using transformer embeddings and calculating cosine similarity scores.
}
\examples{
if (interactive()) {
  mydata <- TextAnalysisR::SpecialEduTech
  united_tbl <- TextAnalysisR::unite_cols(
    mydata,
    listed_vars = c("title", "keyword", "abstract")
  )
  texts <- united_tbl$united_texts

  topic_model <- TextAnalysisR::fit_embedding_model(
    texts = texts,
    method = "umap_hdbscan",
    n_topics = 8
  )

  similar_topics <- TextAnalysisR::find_topic_matches(
    topic_model = topic_model,
    query = "mathematical learning",
    top_n = 5
  )

  print(similar_topics)
}
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{analyze_similarity_gaps}
\alias{analyze_similarity_gaps}
\title{Analyze Similarity Gaps Between Categories}
\usage{
analyze_similarity_gaps(
  similarity_data,
  ref_var = "ref_id",
  other_var = "other_id",
  similarity_var = "similarity",
  category_var = "other_category",
  ref_label_var = NULL,
  other_label_var = NULL,
  unique_threshold = 0.6,
  cross_policy_min = 0.6,
  cross_policy_max = 0.8
)
}
\arguments{
\item{similarity_data}{A data frame with cross-category similarities, containing:
\describe{
\item{ref_var}{Reference item identifier}
\item{other_var}{Comparison item identifier}
\item{similarity_var}{Similarity score}
\item{category_var}{Category of comparison item}
}}

\item{ref_var}{Name of column with reference item IDs (default: "ref_id").}

\item{other_var}{Name of column with comparison item IDs (default: "other_id").}

\item{similarity_var}{Name of column with similarity values (default: "similarity").}

\item{category_var}{Name of column with category information (default: "other_category").}

\item{ref_label_var}{Optional column with reference item labels (for output).}

\item{other_label_var}{Optional column with comparison item labels (for output).}

\item{unique_threshold}{Threshold below which reference items are considered unique (default: 0.6).}

\item{cross_policy_min}{Minimum similarity for cross-policy opportunities (default: 0.6).}

\item{cross_policy_max}{Maximum similarity for cross-policy opportunities (default: 0.8).}
}
\value{
A list containing:
\item{unique_items}{Data frame of reference items with low similarity (unique content)}
\item{missing_items}{Data frame of comparison items with low similarity (content gaps)}
\item{cross_policy}{Data frame of items with moderate similarity (learning opportunities)}
\item{summary_stats}{Summary statistics by category}
}
\description{
Identifies unique items, missing content, and cross-category learning opportunities
based on similarity thresholds. Useful for gap analysis in policy documents,
topic comparisons, or any cross-category similarity study.
}
\examples{
\donttest{
articles <- TextAnalysisR::SpecialEduTech[1:6, ]
articles$display_name <- paste0("d", seq_len(nrow(articles)))
term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(articles$abstract)))
normalized_matrix <- term_matrix / sqrt(rowSums(term_matrix ^ 2))
similarity_matrix <- normalized_matrix \%*\% t(normalized_matrix)
dimnames(similarity_matrix) <- list(articles$display_name, articles$display_name)
cross_similarities <- extract_cross_category_similarities(
  similarity_matrix  = similarity_matrix,
  docs_data          = articles,
  reference_category = "thesis",
  compare_categories = "journal_article",
  category_var       = "reference_type",
  id_var             = "display_name"
)
gap_analysis <- analyze_similarity_gaps(
  similarity_data = cross_similarities,
  ref_var = "ref_id",
  other_var = "other_id",
  similarity_var = "similarity",
  category_var = "other_category",
  unique_threshold = 0.6
)
print(gap_analysis$summary_stats)
}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_alt_text}
\alias{check_alt_text}
\title{Check Alt Text Presence}
\usage{
check_alt_text(alt_text, element_type = "image", decorative = FALSE)
}
\arguments{
\item{alt_text}{Alternative text description}

\item{element_type}{Type of element (e.g., "plot", "image", "icon")}

\item{decorative}{Logical, TRUE if element is purely decorative}
}
\value{
Logical TRUE if valid, FALSE with warning if missing/inadequate
}
\description{
Validates that images and visualizations have alternative text descriptions.
Required for WCAG 1.1.1 (Non-text Content).

Note: Decorative images should use empty alt text (alt="") to indicate
they should be ignored by assistive technology.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{import_files}
\alias{import_files}
\title{Process Files}
\usage{
import_files(dataset_choice, file_info = NULL, text_input = NULL)
}
\arguments{
\item{dataset_choice}{A character string indicating the dataset choice.}

\item{file_info}{A data frame containing file information with a column
named 'filepath' (default: NULL).}

\item{text_input}{A character string containing text input (default: NULL).}
}
\value{
A data frame containing the processed data.
}
\description{
This function processes different types of files and text input based on the dataset choice.
}
\examples{
if (interactive()) {
  mydata <- TextAnalysisR::SpecialEduTech
  mydata <- TextAnalysisR::import_files(dataset_choice = "Upload an Example Dataset")
  head(mydata)

  xlsx_path <- system.file("extdata", "SpecialEduTech.xlsx",
                           package = "TextAnalysisR")
  file_info <- data.frame(filepath = xlsx_path)
  mydata <- TextAnalysisR::import_files(dataset_choice = "Upload Your File",
                                          file_info = file_info)
  head(mydata)


  text_input <- paste("Virtual manipulatives for algebra instruction",
                      "manipulatives mathematics learning disability",
                      "This study examined virtual manipulatives effects on",
                      "students with learning disabilities")
  mydata <- TextAnalysisR::import_files(dataset_choice = "Copy and Paste Text",
                                          text_input = text_input)
  head(mydata)
}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{render_pdf_pages_to_base64}
\alias{render_pdf_pages_to_base64}
\title{Render PDF Pages to Base64 PNG}
\usage{
render_pdf_pages_to_base64(file_path, dpi = 150)
}
\arguments{
\item{file_path}{Character string path to PDF file}

\item{dpi}{Numeric, rendering resolution (default: 150)}
}
\value{
List of base64-encoded PNG strings, one per page
}
\description{
Renders each page of a PDF as a PNG image and returns base64-encoded strings.
Uses pdftools (R-native, no Python required).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{assess_embedding_stability}
\alias{assess_embedding_stability}
\title{Assess Embedding Topic Model Stability}
\usage{
assess_embedding_stability(
  texts,
  n_runs = 5,
  embedding_model = "all-MiniLM-L6-v2",
  select_best = TRUE,
  base_seed = 123,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{texts}{Character vector of documents to analyze.}

\item{n_runs}{Number of model runs with different seeds (default: 5).
More runs give a steadier stability estimate.}

\item{embedding_model}{Embedding model name (default: "all-MiniLM-L6-v2").}

\item{select_best}{Logical, if TRUE, returns the best model by quality (default: TRUE).}

\item{base_seed}{Base random seed; each run uses base_seed + (run - 1).}

\item{verbose}{Logical, if TRUE, prints progress messages.}

\item{...}{Additional arguments passed to fit_embedding_model().}
}
\value{
A list containing:
\itemize{
\item stability_metrics: List with mean_ari, sd_ari, mean_jaccard, quality_variance
\item best_model: Best model by silhouette score (if select_best = TRUE)
\item all_models: List of all fitted models
\item is_stable: Logical, TRUE if mean ARI >= 0.6 (considered stable)
\item recommendation: Text recommendation based on stability
}
}
\description{
Evaluates the stability of embedding-based topic modeling by running multiple models
with different random seeds and comparing their results. High stability (high ARI,
consistent keywords) indicates stable topic structure in the data.
}
\details{
Stability is assessed via:
\itemize{
\item Adjusted Rand Index (ARI): Measures agreement in topic assignments across runs
\item Keyword Jaccard similarity: Measures overlap in top keywords per topic
\item Quality variance: Variance in silhouette scores across runs
}
}
\examples{
if (interactive()) {
  texts <- c("Machine learning for image recognition",
             "Deep learning neural networks",
             "Natural language processing models")

  stability <- assess_embedding_stability(
    texts = texts,
    n_runs = 3,
    verbose = TRUE
  )

  # Check if results are stable
  stability$is_stable
  stability$stability_metrics$mean_ari

  # Use the best model
  best_model <- stability$best_model
}
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{describe_image_openai}
\alias{describe_image_openai}
\title{Describe Image with OpenAI Vision API}
\usage{
describe_image_openai(
  image_base64,
  prompt =
    "Describe this image: charts, diagrams, tables, and text. Extract visible text.",
  model = "gpt-4.1",
  max_tokens = 500,
  api_key
)
}
\arguments{
\item{image_base64}{Character string of base64-encoded PNG image}

\item{prompt}{Character string describing what to extract}

\item{model}{Character string, OpenAI model name (default: "gpt-4.1")}

\item{max_tokens}{Integer, maximum tokens in response (default: 500)}

\item{api_key}{Character string, OpenAI API key}
}
\value{
Character string description, or NULL on failure
}
\description{
Describe Image with OpenAI Vision API
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{get_content_type_user_template}
\alias{get_content_type_user_template}
\title{Get Default User Prompt Template for Content Type}
\usage{
get_content_type_user_template(content_type)
}
\arguments{
\item{content_type}{Type of content to generate.}
}
\value{
Character string with the user prompt template containing \code{\{terms\}} placeholder.
}
\description{
Returns the default user prompt template for a given content type.
}
\concept{ai}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{spacy_initialized}
\alias{spacy_initialized}
\title{Check if spaCy is Initialized}
\usage{
spacy_initialized()
}
\value{
Logical; TRUE if initialized, FALSE otherwise.
}
\description{
Check whether spaCy has been initialized.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{SpecialEduTech}
\alias{SpecialEduTech}
\title{Special education technology bibliographic data}
\format{
An object of class \code{tbl_df} (inherits from \code{tbl}, \code{data.frame}) with 490 rows and 6 columns.
}
\source{
Bibliographic records compiled and coded by the author for
Shin, M., Ok, M. W., Choo, S., Hossain, G., Bryant, D. P., & Kang, E.
(2023). A content analysis of research on technology use for teaching
mathematics to students with disabilities: Word networks and topic
modeling. \emph{International Journal of STEM Education}, 10, Article 23.
\doi{10.1186/s40594-023-00414-x}. Shared as open data at
\url{https://github.com/mshin77/math-tech-sped}.
}
\usage{
SpecialEduTech
}
\description{
Contains bibliographic data of journal articles and dissertations on the use of technology in teaching mathematics to students with disabilities.
}
\examples{
SpecialEduTech

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{truncate_text_to_words}
\alias{truncate_text_to_words}
\title{Truncate Text to Word Count}
\usage{
truncate_text_to_words(text, max_words = 150)
}
\arguments{
\item{text}{Character string to truncate.}

\item{max_words}{Maximum number of words (default: 150).}
}
\value{
Truncated text with "..." appended if truncated.
}
\description{
Truncates text to a maximum number of words and adds
ellipsis if truncated.
}
\concept{text-utilities}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_dfm_instructions_modal}
\alias{show_dfm_instructions_modal}
\title{Show DFM instructions modal}
\usage{
show_dfm_instructions_modal(
  output_id,
  feature_name = "this feature",
  session = NULL
)
}
\arguments{
\item{output_id}{Output id to reset after modal closes.}

\item{feature_name}{Feature description in the instructions.}

\item{session}{Shiny session.}
}
\value{
Invisibly \code{NULL}.
}
\description{
Show DFM instructions modal
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{truncate_text_with_ellipsis}
\alias{truncate_text_with_ellipsis}
\title{Truncate Text with Ellipsis}
\usage{
truncate_text_with_ellipsis(text, max_chars = 50)
}
\arguments{
\item{text}{Character string to truncate.}

\item{max_chars}{Maximum number of characters (default: 50).}
}
\value{
Truncated text with "..." appended if truncated.
}
\description{
Truncates text to a maximum number of characters and adds
ellipsis if truncated.
}
\concept{text-utilities}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{get_available_tokens}
\alias{get_available_tokens}
\title{Get Available Tokens with Fallback}
\usage{
get_available_tokens(
  final_tokens = NULL,
  processed_tokens = NULL,
  preprocessed_tokens = NULL,
  united_tbl = NULL
)
}
\arguments{
\item{final_tokens}{Optional fully processed tokens (highest priority)}

\item{processed_tokens}{Optional partially processed tokens}

\item{preprocessed_tokens}{Optional early-stage preprocessed tokens}

\item{united_tbl}{Optional data frame with united_texts column (lowest priority, will be tokenized)}
}
\value{
The first non-NULL tokens from the priority chain, or NULL if all are NULL
}
\description{
Returns the first non-NULL tokens object from a priority fallback chain.
Useful when multiple token processing stages exist and the most processed available version is needed.
}
\details{
Priority order (highest to lowest):
\enumerate{
\item final_tokens - Fully processed tokens
\item processed_tokens - Partially processed tokens
\item preprocessed_tokens - Early stage preprocessed tokens
\item united_tbl - Raw text (will be tokenized if used)
}
}
\examples{
\donttest{
  toks <- quanteda::tokens("assistive technology supports learning")
  result <- get_available_tokens(processed_tokens = toks)
}
}
\concept{preprocessing}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_guide_modal}
\alias{show_guide_modal}
\title{Show guide modal}
\usage{
show_guide_modal(guide_name, title, size = "l")
}
\arguments{
\item{guide_name}{Guide identifier (filename stem).}

\item{title}{Modal title.}

\item{size}{Modal size ("s", "m", "l", "xl").}
}
\value{
Invisibly \code{NULL}.
}
\description{
Show guide modal
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_keyword_stability}
\alias{calculate_keyword_stability}
\title{Calculate Keyword Stability}
\usage{
calculate_keyword_stability(keywords1, keywords2)
}
\arguments{
\item{keywords1}{List of keyword vectors, one per topic, from the first model.}

\item{keywords2}{List of keyword vectors, one per topic, from the second model.}
}
\value{
Mean Jaccard similarity of matched topic pairs (0-1).
}
\description{
Calculates stability between the topic keyword sets of two independently
fitted models. Topic numbering is arbitrary across fits, so topics are
matched one-to-one by greedy assignment on the pairwise keyword Jaccard
matrix before averaging matched-pair similarity.
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{spacy_lemmatize}
\alias{spacy_lemmatize}
\title{Lemmatize Texts with spaCy}
\usage{
spacy_lemmatize(x, batch_size = 100, model = "en_core_web_sm")
}
\arguments{
\item{x}{Character vector of texts OR a quanteda tokens object.}

\item{batch_size}{Integer; batch size for processing (default: 100).}

\item{model}{Character; spaCy model to use (default: "en_core_web_sm").}
}
\value{
A data frame with columns: doc_id, token_id, token, lemma.
}
\description{
Perform lemmatization using spaCy with optimized pipeline settings.
Disables unnecessary components (NER, parser) for faster processing.
}
\details{
This function disables NER, entity_ruler, and parser components to speed up
lemmatization. Use this for lemmas without other annotations.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{plot_cross_category_heatmap}
\alias{plot_cross_category_heatmap}
\title{Plot Cross-Category Similarity Comparison}
\usage{
plot_cross_category_heatmap(
  similarity_data,
  docs_data = NULL,
  row_var = "ld_doc_name",
  col_var = "other_doc_name",
  value_var = "cosine_similarity",
  category_var = "other_category",
  row_category = NULL,
  col_categories = NULL,
  row_display_var = NULL,
  col_display_var = NULL,
  method_name = "Cosine",
  title = NULL,
  show_values = TRUE,
  row_label = "Documents",
  label_max_chars = 25,
  order_by_numeric = TRUE,
  height = 600,
  width = NULL
)
}
\arguments{
\item{similarity_data}{Either a similarity matrix (square numeric matrix) or
a data frame in long format with columns for row labels, column labels,
similarity values, and category.}

\item{docs_data}{Data frame with document metadata (required if similarity_data is a matrix)}

\item{row_var}{Column name for row document labels (default: "ld_doc_name")}

\item{col_var}{Column name for column document labels (default: "other_doc_name")}

\item{value_var}{Column name for similarity values (default: "cosine_similarity")}

\item{category_var}{Column name for category in long-format data or docs_data (default: "other_category")}

\item{row_category}{Category for row documents (used with matrix input)}

\item{col_categories}{Categories for column documents (used with matrix input)}

\item{row_display_var}{Column name for row display labels in tooltip (default: NULL, uses row_var)}

\item{col_display_var}{Column name for column display labels in tooltip (default: NULL, uses col_var)}

\item{method_name}{Similarity method name for legend (default: "Cosine")}

\item{title}{Plot title (default: NULL)}

\item{show_values}{Logical; show similarity values as text on tiles (default: TRUE)}

\item{row_label}{Label for y-axis (default: "Documents")}

\item{label_max_chars}{Maximum characters for axis labels before truncation (default: 25)}

\item{order_by_numeric}{Logical; order by numeric ID extracted from labels (default: TRUE)}

\item{height}{Plot height (default: 600)}

\item{width}{Plot width (default: NULL)}
}
\value{
A ggplot object
}
\description{
Creates a faceted ggplot heatmap for cross-category document similarity
comparison. Accepts either a pre-built long-format data frame or extracts
from a similarity matrix.
}
\examples{
\donttest{
articles <- TextAnalysisR::SpecialEduTech[1:7, ]
term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(articles$abstract)))
normalized_matrix <- term_matrix / sqrt(rowSums(term_matrix ^ 2))
similarity_matrix <- normalized_matrix \%*\% t(normalized_matrix)
thesis_rows <- which(articles$reference_type == "thesis")[1:3]
article_cols <- which(articles$reference_type == "journal_article")[1:4]
similarity_data <- expand.grid(
  ld_doc_name    = paste("Thesis", seq_along(thesis_rows)),
  other_doc_name = paste("Article", seq_along(article_cols)),
  stringsAsFactors = FALSE
)
similarity_data$cosine_similarity <- as.vector(
  similarity_matrix[thesis_rows, article_cols]
)
similarity_data$other_category <- articles$reference_type[article_cols]
plot_cross_category_heatmap(
  similarity_data = similarity_data,
  row_var = "ld_doc_name",
  col_var = "other_doc_name",
  value_var = "cosine_similarity",
  category_var = "other_category",
  row_label = "Theses"
)
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_preprocessing_steps_modal}
\alias{show_preprocessing_steps_modal}
\title{Show preprocessing-steps modal}
\usage{
show_preprocessing_steps_modal(
  title = "Preprocessing Required",
  message,
  required_steps,
  optional_steps = NULL,
  additional_note = NULL
)
}
\arguments{
\item{title}{Modal title.}

\item{message}{Lead message.}

\item{required_steps}{Character vector of required step labels.}

\item{optional_steps}{Character vector of optional step labels.}

\item{additional_note}{Optional note appended to the body.}
}
\value{
Invisibly \code{NULL}.
}
\description{
Show preprocessing-steps modal
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_unite_texts_required_notification}
\alias{show_unite_texts_required_notification}
\title{Show united-texts-required notification}
\usage{
show_unite_texts_required_notification(duration = 5)
}
\arguments{
\item{duration}{Seconds until auto-dismiss.}
}
\value{
Invisibly the notification id.
}
\description{
Show united-texts-required notification
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_entity_frequencies}
\alias{plot_entity_frequencies}
\title{Plot Named Entity Frequencies}
\usage{
plot_entity_frequencies(
  entity_data,
  top_n = 20,
  title = "Named Entity Type Frequency",
  color = NULL,
  height = 500,
  width = NULL,
  custom_colors = NULL
)
}
\arguments{
\item{entity_data}{Data frame containing entity data with columns:
\itemize{
\item \code{entity}: Named entity type (e.g., "PERSON", "ORG", "GPE")
\item \code{n}: (optional) Pre-computed frequency count
}
If \code{n} is not present, frequencies will be computed from the data.}

\item{top_n}{Number of top entity types to display (default: 20)}

\item{title}{Plot title (default: "Named Entity Type Frequency")}

\item{color}{Bar color (default: "#10B981")}

\item{height}{Plot height in pixels (default: 500)}

\item{width}{Plot width in pixels (default: NULL for auto)}

\item{custom_colors}{Named vector of custom entity type colors (e.g.,
c(CONCEPT = "#00acc1", THEME = "#7c4dff")). Custom colors override defaults.}
}
\value{
A plotly object
}
\description{
Creates a bar plot showing the frequency distribution of named entity types.
}
\examples{
if (interactive()) {
  entity_df <- data.frame(
    entity = c("PERSON", "ORG", "GPE", "DATE", "MONEY"),
    n = c(300, 250, 200, 150, 100)
  )
  plot_entity_frequencies(entity_df)

  # With custom colors
  plot_entity_frequencies(entity_df, custom_colors = c(PERSON = "#ff0000"))
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_morphology_feature}
\alias{plot_morphology_feature}
\title{Plot Morphology Feature Distribution}
\usage{
plot_morphology_feature(data, feature, title = NULL, colors = NULL)
}
\arguments{
\item{data}{Data frame with morph_* columns from extract_morphology().}

\item{feature}{Character; feature name (e.g., "Number", "Tense").}

\item{title}{Character; plot title (auto-generated if NULL).}

\item{colors}{Named character vector of custom colors for feature values.}
}
\value{
A plotly object.
}
\description{
Creates a bar chart showing the distribution of a morphological feature
using consistent package styling.
}
\examples{
if (interactive()) {
  tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1])
  morphology_data <- extract_morphology(tokens)
  plot_morphology_feature(morphology_data, "Tense")
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{semantic_document_clustering}
\alias{semantic_document_clustering}
\title{Semantic Document Clustering}
\usage{
semantic_document_clustering(embeddings, method = "UMAP", clusters = NULL, ...)
}
\arguments{
\item{embeddings}{Document embeddings matrix}

\item{method}{Dimensionality reduction method ("PCA", "t-SNE", "UMAP")}

\item{clusters}{Optional cluster assignments}

\item{...}{Additional arguments}
}
\value{
A ggplot2 visualization of document clusters
}
\description{
Creates a unified visualization of document clustering with optional clustering
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_warning_notification}
\alias{show_warning_notification}
\title{Show warning notification}
\usage{
show_warning_notification(message, duration = 5, session = NULL)
}
\arguments{
\item{message}{Notification text.}

\item{duration}{Seconds until auto-dismiss.}

\item{session}{Shiny session. Defaults to the current reactive domain.}
}
\value{
Invisibly the notification id.
}
\description{
Show warning notification
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_dfm_setup_instructions}
\alias{get_dfm_setup_instructions}
\title{Generate DFM Setup Instructions Text}
\usage{
get_dfm_setup_instructions(feature_name = "this feature")
}
\arguments{
\item{feature_name}{Name of the feature requiring DFM (default: "this feature")}
}
\value{
Character vector of instruction lines
}
\description{
Generates standardized text instructions for creating a DFM.
Used in console output or verbatim text displays.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{plot_document_sentiment_trajectory}
\alias{plot_document_sentiment_trajectory}
\title{Plot Document Sentiment Trajectory}
\usage{
plot_document_sentiment_trajectory(
  sentiment_data,
  top_n = NULL,
  doc_ids = NULL,
  text_preview = NULL,
  title = "Document Sentiment Scores"
)
}
\arguments{
\item{sentiment_data}{Data frame from analyze_sentiment() with sentiment_score column}

\item{top_n}{Number of documents to display (default: NULL for all)}

\item{doc_ids}{Optional vector of custom document IDs for display (default: NULL)}

\item{text_preview}{Optional vector of text snippets for tooltips (default: NULL)}

\item{title}{Plot title (default: "Document Sentiment Scores")}
}
\value{
A ggplot2 line chart with color gradient
}
\description{
Creates a line chart showing sentiment scores across documents with color gradient.
}
\concept{sentiment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{acronym}
\alias{acronym}
\title{Acronym List}
\format{
A character vector of acronyms
}
\description{
A dataset containing common acronyms used in text processing
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{analyze_sentiment}
\alias{analyze_sentiment}
\title{Analyze Text Sentiment}
\usage{
analyze_sentiment(texts, method = "syuzhet", doc_ids = NULL)
}
\arguments{
\item{texts}{Character vector of texts to analyze}

\item{method}{Sentiment analysis method: "syuzhet", "bing", "afinn", or "nrc" (default: "syuzhet").
All methods score each matched dictionary word once per document
regardless of repetition.}

\item{doc_ids}{Optional character vector of document identifiers (default: NULL)}
}
\value{
A data frame with columns:
\describe{
\item{document}{Document identifier}
\item{text}{Original text}
\item{sentiment_score}{Raw summed sentiment score (length-dependent)}
\item{n_tokens}{Number of whitespace-delimited tokens}
\item{avg_sentiment}{sentiment_score / n_tokens}
\item{sentiment}{Classification: "positive", "negative", or "neutral"}
}
}
\description{
Performs sentiment analysis on text data using the syuzhet package.
Classification uses the per-token average score with a +/- 0.05 neutral
band, so document length does not drive polarity. Lexicon methods score
words in isolation; negation, irony, and idioms are not handled. For
negation-sensitive text, see \code{\link[=analyze_sentiment_llm]{analyze_sentiment_llm()}}.
}
\examples{
\donttest{
abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
sentiment_results <- analyze_sentiment(abstracts)
print(sentiment_results)
}
}
\concept{sentiment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_plotly_hover_config}
\alias{get_plotly_hover_config}
\title{Plotly hover tooltip config}
\usage{
get_plotly_hover_config(bgcolor = "#ffffff", fontcolor = "#0c1f4a")
}
\arguments{
\item{bgcolor}{Tooltip background color.}

\item{fontcolor}{Tooltip text color.}
}
\value{
A list for \code{plotly::layout(hoverlabel = ...)}.
}
\description{
Plotly hover tooltip config
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_assignment_consistency}
\alias{calculate_assignment_consistency}
\title{Calculate Assignment Consistency}
\usage{
calculate_assignment_consistency(assignments1, assignments2, ...)
}
\arguments{
\item{assignments1}{First set of assignments}

\item{assignments2}{Second set of assignments}

\item{...}{Additional parameters}
}
\value{
List containing consistency metrics
}
\description{
Calculates consistency between two sets of assignments
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{prep_texts}
\alias{prep_texts}
\title{Preprocess Text Data}
\usage{
prep_texts(
  united_tbl,
  text_field = "united_texts",
  min_char = 2,
  lowercase = TRUE,
  remove_punct = TRUE,
  remove_symbols = TRUE,
  remove_numbers = TRUE,
  remove_url = TRUE,
  remove_separators = TRUE,
  split_hyphens = TRUE,
  split_tags = TRUE,
  include_docvars = TRUE,
  keep_acronyms = FALSE,
  padding = FALSE,
  remove_stopwords = FALSE,
  stopwords_source = "snowball",
  stopwords_language = "en",
  custom_stopwords = NULL,
  custom_valuetype = "glob",
  math_mode = FALSE,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{united_tbl}{A data frame that contains text data.}

\item{text_field}{The name of the column that contains the text data.}

\item{min_char}{The minimum number of characters for a token to be included (default: 2).}

\item{lowercase}{Logical; convert all tokens to lowercase (default: TRUE). Recommended for most text analysis tasks.}

\item{remove_punct}{Logical; remove punctuation from the text (default: TRUE).}

\item{remove_symbols}{Logical; remove symbols from the text (default: TRUE).}

\item{remove_numbers}{Logical; remove numbers from the text (default: TRUE).}

\item{remove_url}{Logical; remove URLs from the text (default: TRUE).}

\item{remove_separators}{Logical; remove separators from the text (default: TRUE).}

\item{split_hyphens}{Logical; split hyphenated words into separate tokens (default: TRUE).}

\item{split_tags}{Logical; split tags into separate tokens (default: TRUE).}

\item{include_docvars}{Logical; include document variables in the tokens object (default: TRUE).}

\item{keep_acronyms}{Logical; keep acronyms in the text (default: FALSE).}

\item{padding}{Logical; add padding to the tokens object (default: FALSE).}

\item{remove_stopwords}{Logical; remove stopwords from the text (default: FALSE).}

\item{stopwords_source}{Character; source for stopwords, e.g., "snowball", "stopwords-iso" (default: "snowball").}

\item{stopwords_language}{Character; language for stopwords (default: "en").}

\item{custom_stopwords}{Character vector; additional words to remove (default: NULL).}

\item{custom_valuetype}{Character; valuetype for custom_stopwords pattern matching, one of "glob", "regex", or "fixed" (default: "glob").}

\item{math_mode}{Logical; if \code{TRUE}, preserve math content (numbers, operators, symbols) by forcing \code{remove_punct}, \code{remove_symbols}, and \code{remove_numbers} all to \code{FALSE}, then strip only sentence-end punctuation such as periods, commas, question marks, exclamation marks, colons, semicolons, parentheses, brackets, braces, quotation marks, em dashes, and en dashes. The \code{min_char} default of 2 still applies, so noisy single-character tokens are dropped; pass \code{min_char = 1} to keep them. Use for math or STEM corpora where multi-character operators and numerals carry meaning (default: FALSE).}

\item{verbose}{Logical; print verbose output (default: FALSE).}

\item{...}{Additional arguments passed to \code{quanteda::tokens}.}
}
\value{
A \code{tokens} object that contains the preprocessed text data.
}
\description{
Preprocesses text data following the complete workflow implemented in the Shiny application:
\itemize{
\item Constructing a corpus from united texts
\item Tokenizing text into words with configurable options
\item Converting to lowercase with acronym preservation option
\item Applying character length filtering
\item Optional multi-word expression detection and compound term creation
\item Stopword removal and lemmatization capabilities
}

This function serves as the foundation for all subsequent text analysis workflows.
}
\examples{
\donttest{
mydata <- TextAnalysisR::SpecialEduTech

united_tbl <- TextAnalysisR::unite_cols(
  mydata,
  listed_vars = c("title", "keyword", "abstract")
)

tokens <- TextAnalysisR::prep_texts(united_tbl,
                                         text_field = "united_texts",
                                         min_char = 2,
                                         lowercase = TRUE,
                                         remove_punct = TRUE,
                                         remove_symbols = TRUE,
                                         remove_numbers = TRUE,
                                         remove_url = TRUE,
                                         remove_separators = TRUE,
                                         split_hyphens = TRUE,
                                         split_tags = TRUE,
                                         include_docvars = TRUE,
                                         keep_acronyms = FALSE,
                                         padding = FALSE,
                                         verbose = FALSE)
print(tokens)
}
}
\seealso{
\code{\link[=unite_cols]{unite_cols()}} to combine text columns first; \code{\link[=lemmatize_tokens]{lemmatize_tokens()}} to reduce words to base form (e.g., running -> run); \code{quanteda::dfm()} to build a document-feature matrix from the result
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_log_odds_ratio}
\alias{plot_log_odds_ratio}
\title{Plot Log Odds Ratio}
\usage{
plot_log_odds_ratio(
  log_odds_data,
  top_n = 10,
  facet_by = NULL,
  color_positive = "#10B981",
  color_negative = "#EF4444",
  height = 600,
  width = NULL,
  title = "Log Odds Ratio Comparison"
)
}
\arguments{
\item{log_odds_data}{Data frame from calculate_log_odds_ratio()}

\item{top_n}{Number of top terms to show per direction (default: 10)}

\item{facet_by}{Character, column name to facet by (e.g., "category1" for
one_vs_rest comparisons). NULL for no faceting.}

\item{color_positive}{Color for positive log odds (default: "#10B981" green)}

\item{color_negative}{Color for negative log odds (default: "#EF4444" red)}

\item{height}{Plot height in pixels (default: 600)}

\item{width}{Plot width in pixels (default: NULL for auto)}

\item{title}{Plot title (default: "Log Odds Ratio Comparison")}
}
\value{
A plotly object
}
\description{
Creates a horizontal bar plot showing log odds ratios for comparing
word usage between categories. Positive values indicate higher usage
in the first category, negative in the second.
}
\examples{
\donttest{
articles <- TextAnalysisR::SpecialEduTech[1:20, ]
corpus <- quanteda::corpus(
  articles$abstract,
  docvars = data.frame(reference_type = articles$reference_type)
)
dfm_object <- quanteda::dfm(quanteda::tokens(corpus))
log_odds <- calculate_log_odds_ratio(dfm_object, "reference_type",
                                      comparison_mode = "binary")
plot_log_odds_ratio(log_odds, top_n = 5)
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{call_openai_chat}
\alias{call_openai_chat}
\title{Call OpenAI Chat Completion API}
\usage{
call_openai_chat(
  system_prompt,
  user_prompt,
  model = "gpt-4.1-mini",
  temperature = 0,
  max_tokens = 150,
  api_key
)
}
\arguments{
\item{system_prompt}{Character string with system instructions}

\item{user_prompt}{Character string with user message}

\item{model}{Character string specifying the model (default: "gpt-4.1-mini")}

\item{temperature}{Numeric temperature for response randomness (default: 0)}

\item{max_tokens}{Maximum number of tokens to generate (default: 150)}

\item{api_key}{Character string with OpenAI API key}
}
\value{
Character string with the model's response
}
\description{
Makes a chat completion request to OpenAI's API.
}
\concept{ai}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{get_topic_terms}
\alias{get_topic_terms}
\title{Select Top Terms for Each Topic}
\usage{
get_topic_terms(stm_model, top_term_n = 10, verbose = TRUE, ...)
}
\arguments{
\item{stm_model}{An STM model object.}

\item{top_term_n}{The number of top terms to display for each topic (default: 10).}

\item{verbose}{Logical, if TRUE, prints progress messages.}

\item{...}{Further arguments passed to tidytext::tidy.}
}
\value{
A data frame containing the top terms for each topic.
}
\description{
This function selects the top terms for each topic based on their word
probability distribution (beta).
}
\examples{
if (interactive() && requireNamespace("stm", quietly = TRUE)) {
  mydata <- TextAnalysisR::SpecialEduTech

  united_tbl <- TextAnalysisR::unite_cols(
    mydata,
    listed_vars = c("title", "keyword", "abstract")
  )

  tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")

  dfm_object <- quanteda::dfm(tokens)

  out <- quanteda::convert(dfm_object, to = "stm")

  stm_15 <- stm::stm(
    data = out$meta,
    documents = out$documents,
    vocab = out$vocab,
    max.em.its = 75,
    init.type = "Spectral",
    K = 15,
    prevalence = ~ reference_type + s(year),
    verbose = TRUE
  )

  top_topic_terms <- TextAnalysisR::get_topic_terms(
    stm_model = stm_15,
    top_term_n = 10,
    verbose = TRUE
  )
  print(top_topic_terms)
}
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{plot_topic_probability}
\alias{plot_topic_probability}
\title{Plot Per-Document Per-Topic Probabilities}
\usage{
plot_topic_probability(
  gamma_data,
  top_n = 10,
  use_topic_labels = FALSE,
  colors = NULL,
  ylab = "Topic Proportion",
  base_font_size = 11
)
}
\arguments{
\item{gamma_data}{A data frame with gamma values from calculate_topic_probability().}

\item{top_n}{The number of topics to display (default: 10).}

\item{use_topic_labels}{Logical. If TRUE, use the \code{topic_label} column from \code{gamma_data} for axis labels (falls back to topic number when the column is absent). If FALSE (default), labels are formatted as "Topic N".}

\item{colors}{Optional color palette for topics (default: NULL).}

\item{ylab}{Y-axis label (default: "Topic Proportion").}

\item{base_font_size}{Base font size in points for the plot theme (default: 11). Axis text and strip text will be base_font_size + 2.}
}
\value{
A ggplot2 object showing a bar plot of topic prevalence.
}
\description{
Generates a bar plot showing the prevalence of each topic across all documents.
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_word_frequency}
\alias{plot_word_frequency}
\title{Plot Word Frequency}
\usage{
plot_word_frequency(dfm_object, n = 20, height = NULL, width = NULL, ...)
}
\arguments{
\item{dfm_object}{A document-feature matrix created by quanteda::dfm().}

\item{n}{The number of top words to display (default: 20).}

\item{height}{Plot height in pixels (default: 800). Kept for backward compatibility.}

\item{width}{Plot width in pixels (default: 1000). Kept for backward compatibility.}

\item{...}{Additional arguments (kept for backward compatibility).}
}
\value{
A ggplot object showing word frequency.
}
\description{
Creates a bar plot showing the most frequent words in a document-feature matrix (dfm).
}
\examples{
\donttest{
  data(SpecialEduTech, package = "TextAnalysisR")
  texts <- SpecialEduTech$abstract[1:10]
  dfm <- quanteda::dfm(quanteda::tokens(texts))
  plot <- plot_word_frequency(dfm, n = 10)
  print(plot)
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_topic_stability}
\alias{calculate_topic_stability}
\title{Calculate Topic Stability}
\usage{
calculate_topic_stability(temporal_results)
}
\arguments{
\item{temporal_results}{Results from temporal analysis.}
}
\value{
Stability metrics: matched-pair stability per consecutive-period
transition and their mean.
}
\description{
Calculates stability of topics across consecutive time periods. Each
period is fitted independently, so topics are matched one-to-one on
keyword Jaccard similarity before scoring
(see \code{\link[=calculate_keyword_stability]{calculate_keyword_stability()}}).
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_preprocessing_required_modal}
\alias{show_preprocessing_required_modal}
\title{Show preprocessing-required modal}
\usage{
show_preprocessing_required_modal(
  message = "Please complete preprocessing steps first.",
  title = "Preprocessing Required"
)
}
\arguments{
\item{message}{Modal body message.}

\item{title}{Modal title.}
}
\value{
Invisibly \code{NULL}.
}
\description{
Show preprocessing-required modal
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{plot_semantic_viz}
\alias{plot_semantic_viz}
\title{Plot Semantic Analysis Visualization}
\usage{
plot_semantic_viz(
  analysis_result = NULL,
  plot_type = "similarity",
  data_labels = NULL,
  color_by = NULL,
  height = 600,
  width = 800,
  title = NULL,
  coords = NULL,
  clusters = NULL,
  hover_text = NULL,
  hover_config = NULL,
  cluster_colors = NULL
)
}
\arguments{
\item{analysis_result}{A list containing semantic analysis results from functions like
semantic_similarity_analysis(), semantic_document_clustering(), or reduce_dimensions().}

\item{plot_type}{Type of visualization: "similarity" for heatmap, "dimensionality_reduction"
for scatter plot, or "clustering" for cluster visualization (default: "similarity").}

\item{data_labels}{Optional character vector of labels for data points (default: NULL).}

\item{color_by}{Optional variable to color points by in scatter plots (default: NULL).}

\item{height}{The height of the resulting Plotly plot, in pixels (default: 600).}

\item{width}{The width of the resulting Plotly plot, in pixels (default: 800).}

\item{title}{Optional custom title for the plot (default: NULL).}

\item{coords}{Optional pre-computed coordinates for dimensionality reduction plots (default: NULL).}

\item{clusters}{Optional cluster assignments vector (default: NULL).}

\item{hover_text}{Optional custom hover text for points (default: NULL).}

\item{hover_config}{Optional hover configuration list (default: NULL).}

\item{cluster_colors}{Optional color palette for clusters (default: NULL).}
}
\value{
A ggplot2 object showing the specified visualization.
}
\description{
Creates interactive visualizations for semantic analysis results including
similarity heatmaps, dimensionality reduction plots, and clustering visualizations.
}
\examples{
\donttest{
  data(SpecialEduTech)
  texts <- SpecialEduTech$abstract[1:5]
  result <- semantic_similarity_analysis(texts)
  plot <- plot_semantic_viz(result, plot_type = "similarity")
  print(plot)
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{analyze_sentiment_llm}
\alias{analyze_sentiment_llm}
\title{LLM-based Sentiment Analysis}
\usage{
analyze_sentiment_llm(
  texts,
  doc_names = NULL,
  provider = c("openai", "gemini"),
  model = NULL,
  api_key = NULL,
  batch_size = 5,
  include_explanation = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{texts}{Character vector of texts to analyze.}

\item{doc_names}{Optional character vector of document names (default: text1, text2, ...).}

\item{provider}{AI provider to use: "openai" (default) or "gemini".}

\item{model}{Model name. If NULL, uses provider defaults: "gpt-4.1-mini" (OpenAI),
"gemini-2.5-flash-lite" (Gemini).}

\item{api_key}{API key for OpenAI or Gemini. If NULL, uses environment variable.}

\item{batch_size}{Number of texts to process per API call (default: 5).
Larger batches are more efficient but may hit token limits.}

\item{include_explanation}{Logical, if TRUE includes natural language explanation
for each sentiment classification (default: FALSE).}

\item{verbose}{Logical, if TRUE prints progress messages (default: TRUE).}
}
\value{
A list containing:
\describe{
\item{document_sentiment}{Data frame with document-level sentiment scores}
\item{summary_stats}{Summary statistics of the analysis}
\item{model_used}{Model name used for analysis}
\item{provider}{AI provider used}
}
}
\description{
Analyzes sentiment using Large Language Models (OpenAI or Gemini).
Provides nuanced sentiment understanding including sarcasm detection,
mixed emotions, and contextual interpretation that lexicon-based methods miss.

Requests use temperature 0, which makes outputs near-deterministic but
not fully reproducible across runs or model versions; record the returned
\code{model_used} field when reporting results.
}
\details{
LLM-based sentiment analysis offers several advantages over lexicon methods:
\itemize{
\item Understands context and nuance
\item Detects sarcasm and irony
\item Handles mixed emotions
\item Works across domains without retraining
}
}
\examples{
if (interactive()) {
  abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:5]

  sentiment_openai <- analyze_sentiment_llm(abstracts, provider = "openai")

  sentiment_gemini <- analyze_sentiment_llm(abstracts, provider = "gemini",
                                             include_explanation = TRUE)
}
}
\concept{sentiment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{generate_embeddings}
\alias{generate_embeddings}
\title{Generate Embeddings}
\usage{
generate_embeddings(texts, model = "all-MiniLM-L6-v2", verbose = TRUE)
}
\arguments{
\item{texts}{A character vector of texts.}

\item{model}{Sentence transformer model name (default: "all-MiniLM-L6-v2").
Newer small models (e.g. "BAAI/bge-small-en-v1.5") often score higher on
retrieval benchmarks but expect instruction prefixes this function does
not add.}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A matrix of L2-normalized embeddings.
}
\description{
Generates L2-normalized embeddings for texts using sentence transformers.
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_mwe_frequency}
\alias{plot_mwe_frequency}
\title{Plot Multi-Word Expression Frequency}
\usage{
plot_mwe_frequency(
  mwe_data,
  title = "Multi-Word Expression Frequency",
  color_by_source = TRUE,
  primary_color = "#10B981",
  secondary_color = "#A855F7",
  height = 500,
  width = NULL
)
}
\arguments{
\item{mwe_data}{Data frame containing MWE data with columns:
\itemize{
\item \code{feature}: The multi-word expression text
\item \code{frequency}: Frequency count
\item \code{rank}: (optional) Rank of the expression
\item \code{docfreq}: (optional) Document frequency
\item \code{source}: (optional) Source category (e.g., "Top 20", "Manual")
}}

\item{title}{Plot title (default: "Multi-Word Expression Frequency")}

\item{color_by_source}{Whether to color bars by source column (default: TRUE)}

\item{primary_color}{Color for primary/top expressions (default: "#10B981")}

\item{secondary_color}{Color for secondary/manual expressions (default: "#A855F7")}

\item{height}{Plot height in pixels (default: 500)}

\item{width}{Plot width in pixels (default: NULL for auto)}
}
\value{
A plotly object
}
\description{
Creates a bar plot showing multi-word expression frequencies with optional
source-based coloring to distinguish between detected and manually added expressions.
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{sentiment_lexicon_analysis}
\alias{sentiment_lexicon_analysis}
\title{Analyze Sentiment Using Tidytext Lexicons}
\usage{
sentiment_lexicon_analysis(
  dfm_object,
  lexicon = "bing",
  texts_df = NULL,
  feature_type = "words",
  ngram_range = 2,
  texts = NULL
)
}
\arguments{
\item{dfm_object}{A quanteda DFM object (unigram)}

\item{lexicon}{Lexicon to use: "afinn", "bing", or "nrc" (default: "bing")}

\item{texts_df}{Optional data frame with original texts and metadata (default: NULL)}

\item{feature_type}{Feature space: "words" (unigrams). The AFINN/Bing/NRC lexicons are
unigram lexicons; "ngrams" falls back to unigram scoring with a warning (default: "words").}

\item{ngram_range}{Retained for backward compatibility; not used for scoring (default: 2)}

\item{texts}{Optional character vector of texts used to rebuild a unigram DFM (default: NULL)}
}
\value{
A list containing:
\describe{
\item{document_sentiment}{Data frame with sentiment scores per document}
\item{emotion_scores}{Data frame with emotion scores (NRC only)}
\item{summary_stats}{List of summary statistics}
\item{feature_type}{Feature type used for analysis}
}
}
\description{
Performs lexicon-based sentiment analysis on a DFM object using tidytext lexicons.
Supports AFINN, Bing, and NRC lexicons with scoring and emotion analysis.

Tokens absent from the lexicon are ignored (not treated as neutral);
\code{summary_stats$token_match_rate} reports the share of corpus tokens
the lexicon covered. AFINN classification uses a +/- 0.5 band on the
per-sentiment-word average (\code{avg_sentiment}); its
\code{n_sentiment_words} column counts matched sentiment tokens, not
document length.
}
\examples{
\donttest{
abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
corpus <- quanteda::corpus(abstracts)
dfm_object <- quanteda::dfm(quanteda::tokens(corpus))
lexicon_results <- sentiment_lexicon_analysis(dfm_object, lexicon = "bing")
print(lexicon_results$document_sentiment)
}
}
\concept{sentiment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{plot_word_probability}
\alias{plot_word_probability}
\title{Plot Word Probabilities by Topic}
\usage{
plot_word_probability(
  top_topic_terms,
  topic_label = NULL,
  ncol = 3,
  height = 1200,
  width = 800,
  ylab = "Word probability",
  title = NULL,
  colors = NULL,
  measure_label = "Beta",
  base_font_size = 11,
  ...
)
}
\arguments{
\item{top_topic_terms}{A data frame containing topic terms with columns: topic, term, and beta.}

\item{topic_label}{Optional topic labels. Can be either a named vector mapping topic numbers
to labels, or a character string specifying a column name in top_topic_terms (default: NULL).}

\item{ncol}{Number of columns for facet wrap layout (default: 3).}

\item{height}{Plot height for responsive spacing adjustments (default: 1200).}

\item{width}{Plot width for responsive spacing adjustments (default: 800).}

\item{ylab}{Y-axis label (default: "Word probability").}

\item{title}{Plot title (default: NULL for auto-generated title).}

\item{colors}{Color palette for topics (default: NULL for auto-generated colors).}

\item{measure_label}{Label for the probability measure (default: "Beta").}

\item{base_font_size}{Base font size in points for the plot theme (default: 11). Axis text and strip text will be base_font_size + 2.}

\item{...}{Additional arguments (currently unused, kept for compatibility).}
}
\value{
A ggplot2 object showing word probabilities faceted by topic.
}
\description{
Creates a faceted bar plot showing the top terms and their probabilities (beta values)
for each topic in a topic model.
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_feature}
\alias{check_feature}
\title{Check Feature Status}
\usage{
check_feature(feature)
}
\arguments{
\item{feature}{Character: "python", "pdf_tables", "embeddings", "sentiment_transformer"}
}
\value{
Logical TRUE if feature is available
}
\description{
Checks if a specific optional feature is available in the current environment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{generate_semantic_topic_keywords}
\alias{generate_semantic_topic_keywords}
\title{Generate Semantic Topic Keywords (c-TF-IDF)}
\usage{
generate_semantic_topic_keywords(
  texts,
  topic_assignments,
  n_keywords = 10,
  method = "c-tfidf",
  diversity = 0.5
)
}
\arguments{
\item{texts}{A character vector of texts.}

\item{topic_assignments}{A vector of topic assignments.}

\item{n_keywords}{The number of keywords to extract per topic (default: 10).}

\item{method}{The representation method: "c-tfidf" (default), "tfidf", "mmr", or "frequency".}

\item{diversity}{Diversity weight for "mmr" between 0 and 1 (default: 0.5).
MMR selects terms by (1 - diversity) * relevance - diversity * redundancy.}
}
\value{
A list of keywords for each topic.
}
\description{
Generate keywords for topics using c-TF-IDF (class-based TF-IDF), similar to BERTopic.
This method treats all documents in a topic as a single document and calculates TF-IDF
scores relative to other topics.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{init_spacy_nlp}
\alias{init_spacy_nlp}
\title{Initialize spaCy NLP}
\usage{
init_spacy_nlp(model = "en_core_web_sm", force = FALSE)
}
\arguments{
\item{model}{Character; spaCy model name (default: "en_core_web_sm").}

\item{force}{Logical; force reinitialization even if already initialized.}
}
\value{
Invisibly returns the SpacyNLP Python object.
}
\description{
Initialize the spaCy NLP pipeline with the specified model.
Uses a cached instance for efficiency.
}
\details{
Available models:
\itemize{
\item \code{en_core_web_sm}: Small English model (fast, no word vectors)
\item \code{en_core_web_md}: Medium English model (word vectors)
\item \code{en_core_web_lg}: Large English model (best accuracy)
}
}
\examples{
if (interactive()) {
init_spacy_nlp("en_core_web_sm")
}
}
\concept{lexical}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_lexical_dispersion}
\alias{plot_lexical_dispersion}
\title{Plot Lexical Dispersion}
\usage{
plot_lexical_dispersion(
  dispersion_data,
  scale = "relative",
  title = "Lexical Dispersion",
  colors = NULL,
  height = 400,
  width = NULL,
  marker_size = 8
)
}
\arguments{
\item{dispersion_data}{Data frame from calculate_lexical_dispersion()}

\item{scale}{Character, "relative" or "absolute" (must match calculation)}

\item{title}{Plot title (default: "Lexical Dispersion")}

\item{colors}{Named vector of colors for each term, or NULL for auto}

\item{height}{Plot height in pixels (default: 400)}

\item{width}{Plot width in pixels (default: NULL for auto)}

\item{marker_size}{Size of position markers (default: 8)}
}
\value{
A plotly object
}
\description{
Creates an X-ray plot showing where terms appear across documents.
Each row represents a term, and marks indicate occurrences.
}
\examples{
\donttest{
tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1:5])
dispersion <- calculate_lexical_dispersion(tokens, c("learning", "instruction"))
plot_lexical_dispersion(dispersion)
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{extract_subjects_objects}
\alias{extract_subjects_objects}
\title{Extract Subjects and Objects}
\usage{
extract_subjects_objects(x, model = "en_core_web_sm")
}
\arguments{
\item{x}{Character vector of texts OR a quanteda tokens object.}

\item{model}{Character; spaCy model to use (default: "en_core_web_sm").}
}
\value{
A data frame with SVO information.
}
\description{
Extract subject-verb-object (SVO) triples from texts using dependency parsing.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{unite_cols}
\alias{unite_cols}
\title{Unite Text Columns}
\usage{
unite_cols(df, listed_vars)
}
\arguments{
\item{df}{A data frame that contains text data.}

\item{listed_vars}{A character vector of column names to be united into
"united_texts".}
}
\value{
A data frame with a new column "united_texts" created by uniting
the specified variables.
}
\description{
This function unites specified text columns in a data frame into a single
column named "united_texts" while retaining the original columns.
}
\examples{
\donttest{
  mydata <- TextAnalysisR::SpecialEduTech

  united_tbl <- TextAnalysisR::unite_cols(
    mydata,
    listed_vars = c("title", "keyword", "abstract")
  )
  print(united_tbl)
}
}
\seealso{
\code{\link[=prep_texts]{prep_texts()}} to tokenize and clean the united text; \code{\link[=import_files]{import_files()}} to load source data first
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{extract_text_from_pdf_py}
\alias{extract_text_from_pdf_py}
\title{Extract Text from PDF using Python}
\usage{
extract_text_from_pdf_py(file_path, envname = NULL)
}
\arguments{
\item{file_path}{Character string path to PDF file}

\item{envname}{Character string, name of Python virtual environment
(default: "textanalysisr-env")}
}
\value{
Data frame with columns: page (integer), text (character)
Returns NULL if extraction fails or PDF is empty
}
\description{
Extracts text content from a PDF file using pdfplumber (Python).
No Java required - uses Python environment.
}
\details{
Uses pdfplumber Python library through reticulate.
Requires Python environment setup. See \code{setup_python_env()}.
}
\concept{pdf}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{calculate_contrast_ratio}
\alias{calculate_contrast_ratio}
\title{Web Accessibility Utility Functions}
\usage{
calculate_contrast_ratio(foreground, background)
}
\arguments{
\item{foreground}{Foreground color (hex format, e.g., "#111827")}

\item{background}{Background color (hex format, e.g., "#ffffff")}
}
\value{
Numeric contrast ratio (1-21)
}
\description{
Functions for ensuring WCAG 2.1 Level AA compliance in the Shiny application

Calculates the contrast ratio between two colors according to WCAG 2.1 standards
using the relative luminance formula from W3C guidelines.
Used to verify text/background color combinations meet accessibility requirements.
}
\section{WCAG 2.1 Level AA Compliance}{

This package follows Web Content Accessibility Guidelines (WCAG) 2.1 Level AA:
\itemize{
\item 1.1.1 Non-text Content (Level A): Alt text for images and visualizations
\item 1.4.3 Contrast Minimum (Level AA): 4.5:1 ratio for normal text, 3:1 for large text/UI
\item 2.1.1 Keyboard (Level A): Full keyboard navigation support
\item 2.4.1 Bypass Blocks (Level A): Skip navigation links
\item 3.1.1 Language of Page (Level A): Page language identification
\item 4.1.2 Name, Role, Value (Level A): ARIA labels and roles
Calculate Color Contrast Ratio
}
}

\section{WCAG Requirements}{

\itemize{
\item Normal text: Minimum 4.5:1 (Level AA)
\item Large text (18pt+ or 14pt+ bold): Minimum 3:1 (Level AA)
\item UI components and graphics: Minimum 3:1 (Level AA)
}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{find_optimal_k}
\alias{find_optimal_k}
\title{Find Optimal Number of Topics}
\usage{
find_optimal_k(
  dfm_object,
  topic_range,
  max.em.its = 75,
  emtol = 1e-04,
  cores = 1,
  categorical_var = NULL,
  continuous_var = NULL,
  height = 600,
  width = 800,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{dfm_object}{A quanteda dfm object to be used for topic modeling.}

\item{topic_range}{A vector of K values to test (e.g., 2:10).}

\item{max.em.its}{Maximum number of EM iterations (default: 75).}

\item{emtol}{Convergence tolerance for EM algorithm (default: 1e-04).
Higher values (e.g., 1e-03) speed up fitting but may reduce precision.}

\item{cores}{Number of CPU cores to use for parallel processing (default: 1).
Set to higher values for faster searchK on multi-core systems.}

\item{categorical_var}{Optional categorical variable(s) for prevalence.}

\item{continuous_var}{Optional continuous variable(s) for prevalence.}

\item{height}{Plot height in pixels (default: 600).}

\item{width}{Plot width in pixels (default: 800).}

\item{verbose}{Logical indicating whether to print progress (default: TRUE).}

\item{...}{Additional arguments passed to stm::searchK.}
}
\value{
A list containing search results and diagnostic plots.
}
\description{
Searches for the optimal number of topics (K) using stm::searchK.
Produces diagnostic plots to help select the best K value. The default
spectral initialization is deterministic; with LDA or random
initialization, fit several seeds per K before selecting a model.
}
\seealso{
\code{\link[=plot_quality_metrics]{plot_quality_metrics()}} to visualize topic-count diagnostics; \code{stm::stm()} to fit the chosen model; \code{\link[=fit_embedding_model]{fit_embedding_model()}} for an embedding-based alternative to STM
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{analyze_document_clustering}
\alias{analyze_document_clustering}
\title{Analyze Document Clustering}
\usage{
analyze_document_clustering(
  feature_matrix,
  method = "UMAP",
  clustering_method = "none",
  ...
)
}
\arguments{
\item{feature_matrix}{Feature matrix (documents x features)}

\item{method}{Dimensionality reduction method ("PCA", "t-SNE", "UMAP")}

\item{clustering_method}{Clustering method ("none", "kmeans", "hierarchical", "dbscan", "hdbscan")}

\item{...}{Additional parameters for methods}
}
\value{
List containing coordinates, clusters, method info, and quality metrics
}
\description{
Complete document clustering analysis with dimensionality reduction and optional clustering
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{create_standard_ggplot_theme}
\alias{create_standard_ggplot_theme}
\title{Create Standard ggplot2 Theme}
\usage{
create_standard_ggplot_theme(base_size = 11)
}
\arguments{
\item{base_size}{Base font size (default: 14)}
}
\value{
A ggplot2 theme object
}
\description{
Returns a standardized ggplot2 theme matching TextAnalysisR design standards.
}
\concept{visualization}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{get_sentences}
\alias{get_sentences}
\title{Get Sentences}
\usage{
get_sentences(x, model = "en_core_web_sm")
}
\arguments{
\item{x}{Character vector of texts OR a quanteda tokens object.}

\item{model}{Character; spaCy model to use (default: "en_core_web_sm").}
}
\value{
A data frame with sentence information.
}
\description{
Segment texts into sentences using spaCy's sentence boundary detection.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_dfm_required_modal}
\alias{show_dfm_required_modal}
\title{Show DFM-required modal}
\usage{
show_dfm_required_modal(
  feature_name = "this feature",
  additional_message = NULL
)
}
\arguments{
\item{feature_name}{Feature description in the message.}

\item{additional_message}{Optional extra context appended to the body.}
}
\value{
Invisibly \code{NULL}.
}
\description{
Show DFM-required modal
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{lexical_diversity_analysis}
\alias{lexical_diversity_analysis}
\title{Lexical Diversity Analysis}
\usage{
lexical_diversity_analysis(x, measures = "all", texts = NULL, cache_key = NULL)
}
\arguments{
\item{x}{A quanteda DFM or tokens object. Tokens object is preferred for
accurate MTLD calculation since it preserves token order.}

\item{measures}{Character vector of measures to calculate.
Default is "all" which includes: TTR, C, R, CTTR, U, S, K, I, D, Vm, Maas, MATTR, MSTTR, MTLD, and HDD.
Most recommended: "MTLD", "MATTR", or "HDD" for length-independent measures.}

\item{texts}{Optional character vector of original texts. Required for MTLD
calculation when using DFM input (since DFM loses token order).}

\item{cache_key}{Optional cache key (e.g., from digest::digest) for caching
expensive calculations. Use the same cache_key to retrieve cached results.}
}
\value{
A list containing:
\itemize{
\item \code{lexical_diversity}: Data frame with per-document lexical diversity scores
\item \code{summary_stats}: List of summary statistics (mean, median, sd) for each measure
}
}
\description{
Calculates multiple lexical diversity metrics for a document-feature matrix (DFM)
or tokens object. Supports all quanteda.textstats measures plus MTLD
(Measure of Textual Lexical Diversity), which is the most recommended measure
according to McCarthy & Jarvis (2010) for being independent of text length.
}
\details{
MTLD (Measure of Textual Lexical Diversity) is calculated using the algorithm
from McCarthy & Jarvis (2010). It counts the number of "factors" needed to
reduce TTR below 0.72, then divides the number of tokens by the number of factors.
This provides a length-independent measure of lexical diversity.

Important notes:
\itemize{
\item For MTLD accuracy, pass a tokens object (not DFM) as input
\item If using DFM, provide the 'texts' parameter for MTLD calculation
\item MATTR and MSTTR window sizes are automatically adjusted for short documents
\item Raw TTR falls mechanically as documents lengthen; compare TTR only
across documents of similar length
\item MTLD and MATTR are most reliable at 100+ tokens per document
\item Results are cached when cache_key is provided for repeated analysis
}
}
\examples{
\donttest{
data(SpecialEduTech)
texts <- SpecialEduTech$abstract[1:10]
corp <- quanteda::corpus(texts)
toks <- quanteda::tokens(corp)
# Preferred: pass tokens object for accurate MTLD
lex_div <- lexical_diversity_analysis(toks, texts = texts)
# With caching for repeated analysis
cache_key <- digest::digest(texts)
lex_div <- lexical_diversity_analysis(toks, texts = texts, cache_key = cache_key)
# Alternative: pass DFM with texts for MTLD accuracy
dfm_obj <- quanteda::dfm(toks)
lex_div <- lexical_diversity_analysis(dfm_obj, texts = texts)
print(lex_div)
}

}
\references{
McCarthy, P. M., & Jarvis, S. (2010). MTLD, vocd-D, and HD-D: A validation study
of sophisticated approaches to lexical diversity assessment.
Behavior Research Methods, 42(2), 381-392.
}
\seealso{
\code{\link[=calculate_text_readability]{calculate_text_readability()}} for grade-level / Flesch metrics on the same input; \code{\link[=calculate_lexical_dispersion]{calculate_lexical_dispersion()}} for term spread across documents; \code{\link[=plot_lexical_diversity_distribution]{plot_lexical_diversity_distribution()}} to visualize
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_keyness_keywords}
\alias{plot_keyness_keywords}
\title{Plot Statistical Keyness}
\usage{
plot_keyness_keywords(keyness_data, title = NULL, group_label = NULL)
}
\arguments{
\item{keyness_data}{Data frame from extract_keywords_keyness()}

\item{title}{Plot title (default: "Top Keywords by Keyness (G-squared)")}

\item{group_label}{Optional label for the target group (default: NULL)}
}
\value{
A plotly bar chart
}
\description{
Creates a horizontal bar plot of distinctive keywords by keyness score.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validate_api_key}
\alias{validate_api_key}
\title{Validate API Key Format}
\usage{
validate_api_key(api_key, strict = TRUE)
}
\arguments{
\item{api_key}{Character string containing the API key}

\item{strict}{Logical, if TRUE performs additional validation checks}
}
\value{
List with valid (logical), provider (character), and error (character if invalid)
}
\description{
Validates API key format for OpenAI or Gemini according to NIST IA-5(1).
Auto-detects provider from key prefix and validates format requirements.
}
\section{NIST Compliance}{

Implements NIST IA-5(1): Authenticator Management - Password-Based Authentication.
Validates format, length, and character composition to prevent weak or malformed keys.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{run_text_workflow}
\alias{run_text_workflow}
\title{Complete Text Mining Workflow}
\usage{
run_text_workflow(
  dataset_choice,
  file_info = NULL,
  text_input = NULL,
  listed_vars,
  min_char = 2,
  remove_punct = TRUE,
  remove_symbols = TRUE,
  remove_numbers = TRUE,
  remove_url = TRUE,
  detect_compounds = FALSE,
  compound_size = 2:3,
  compound_min_count = 2,
  verbose = TRUE
)
}
\arguments{
\item{dataset_choice}{A character string indicating the dataset choice:
"Upload an Example Dataset", "Upload Your File", "Copy and Paste Text".}

\item{file_info}{A data frame containing file information (for file upload).}

\item{text_input}{A character string containing text input (for copy-paste).}

\item{listed_vars}{A character vector of column names to unite into text.}

\item{min_char}{The minimum number of characters for tokens (default: 2).}

\item{remove_punct}{Logical; remove punctuation (default: TRUE).}

\item{remove_symbols}{Logical; remove symbols (default: TRUE).}

\item{remove_numbers}{Logical; remove numbers (default: TRUE).}

\item{remove_url}{Logical; remove URLs (default: TRUE).}

\item{detect_compounds}{Logical; detect multi-word expressions (default: FALSE).}

\item{compound_size}{Size range for compound detection (default: 2:3).}

\item{compound_min_count}{Minimum count for compounds (default: 2).}

\item{verbose}{Logical; print progress messages (default: TRUE).}
}
\value{
A list containing processed data, tokens, DFM, and metadata.
}
\description{
This function provides a complete text mining workflow that follows the same sequence
as the Shiny application: file processing -> text uniting -> preprocessing ->
DFM creation -> analysis.
It serves as a convenience function for users who want to execute the entire
pipeline programmatically.
}
\examples{
\donttest{
  workflow_result <- TextAnalysisR::run_text_workflow(
    dataset_choice = "Upload an Example Dataset",
    listed_vars = c("title", "keyword", "abstract")
  )
}
if (interactive()) {
  file_info <- data.frame(filepath = "path/to/file.xlsx")
  workflow_result <- TextAnalysisR::run_text_workflow(
    dataset_choice = "Upload Your File",
    file_info = file_info,
    listed_vars = c("column1", "column2")
  )

  workflow_result <- TextAnalysisR::run_text_workflow(
    dataset_choice = "Copy and Paste Text",
    text_input = "Your text content here",
    listed_vars = "text"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{validate_cross_models}
\alias{validate_cross_models}
\title{Cross-Analysis Validation}
\usage{
validate_cross_models(semantic_results, stm_results = NULL, verbose = TRUE)
}
\arguments{
\item{semantic_results}{Results from semantic analysis.}

\item{stm_results}{Optional STM results for comparison.}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A list containing cross-validation metrics.
}
\description{
Performs cross-validation between different analysis types (STM, semantic, clustering).
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{create_empty_plot_message}
\alias{create_empty_plot_message}
\title{Empty plotly placeholder}
\usage{
create_empty_plot_message(message, color = "#6B7280", font_size = 16)
}
\arguments{
\item{message}{Message to display.}

\item{color}{Text color.}

\item{font_size}{Font size in px.}
}
\value{
A \code{plotly} htmlwidget.
}
\description{
Empty plotly placeholder
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{plot_sentiment_distribution}
\alias{plot_sentiment_distribution}
\title{Plot Sentiment Distribution}
\usage{
plot_sentiment_distribution(sentiment_data, title = "Sentiment Distribution")
}
\arguments{
\item{sentiment_data}{Data frame from analyze_sentiment() or with 'sentiment' column}

\item{title}{Plot title (default: "Sentiment Distribution")}
}
\value{
A ggplot2 bar chart
}
\description{
Creates a bar plot showing the distribution of sentiment classifications.
}
\examples{
\donttest{
abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
sentiment_data <- analyze_sentiment(abstracts)
sentiment_plot <- plot_sentiment_distribution(sentiment_data)
print(sentiment_plot)
}
}
\concept{sentiment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{run_rag_search}
\alias{run_rag_search}
\title{RAG Semantic Search}
\usage{
run_rag_search(
  query,
  documents,
  provider = c("openai", "gemini"),
  api_key = NULL,
  embedding_model = NULL,
  chat_model = NULL,
  top_k = 5
)
}
\arguments{
\item{query}{Character string, user question}

\item{documents}{Character vector, corpus to search}

\item{provider}{Character string, provider: "openai" or "gemini"}

\item{api_key}{Character string, API key (or from
OPENAI_API_KEY/GEMINI_API_KEY env).}

\item{embedding_model}{Character string, embedding model. Defaults:
"text-embedding-3-small" (openai), "gemini-embedding-001" (gemini)}

\item{chat_model}{Character string, chat model. Defaults:
"gpt-4.1-mini" (openai), "gemini-2.5-flash-lite" (gemini)}

\item{top_k}{Integer, number of documents to retrieve (default: 5)}
}
\value{
List with:
\itemize{
\item success: Logical
\item answer: Generated answer
\item confidence: Confidence score (0-1)
\item sources: Vector of source document indices
\item retrieved_docs: Retrieved document chunks
\item scores: Similarity scores
}
}
\description{
Simple in-memory RAG (Retrieval Augmented Generation) for question-answering
over document corpus with source attribution. Uses OpenAI or Gemini
embeddings for semantic search and LLM for answer generation.
}
\details{
Simple RAG workflow:
\enumerate{
\item Generate embeddings for documents and query
\item Find top-k similar documents via cosine similarity
\item Generate answer using LLM with retrieved context
}
}
\examples{
if (interactive()) {
documents <- c(
  "Assistive technology helps students with disabilities access curriculum.",
  "Universal Design for Learning provides multiple means of engagement.",
  "Response to Intervention uses tiered support systems."
)

# Using OpenAI (requires API key)
result <- run_rag_search(
  query = "How does assistive technology support learning?",
  documents = documents,
  provider = "openai"
)

if (result$success) {
  cat("Answer:", result$answer, "\n")
  cat("Sources:", paste(result$sources, collapse = ", "), "\n")
}
}
}
\seealso{
\code{\link[=get_best_embeddings]{get_best_embeddings()}} for the retrieval step alone; \code{\link[=call_llm_api]{call_llm_api()}} for the answer-generation step alone; \code{\link[=sanitize_llm_input]{sanitize_llm_input()}} for an input safety check before calling
}
\concept{ai}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{sanitize_llm_input}
\alias{sanitize_llm_input}
\title{Sanitize LLM Input}
\usage{
sanitize_llm_input(text, max_length = 2000)
}
\arguments{
\item{text}{Character string of user input destined for an LLM prompt}

\item{max_length}{Maximum allowed character length (default: 2000)}
}
\value{
Sanitized character string
}
\description{
Sanitizes user input before inclusion in LLM prompts to mitigate prompt
injection attacks. Filters common injection patterns such as instruction
overrides, system prompt markers, and role-switching attempts.
Distinct from \code{sanitize_text_input()} which targets XSS.
}
\section{NIST Compliance}{

Implements NIST SI-10 (Information Input Validation) for AI/LLM contexts.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{spacy_extract_entities}
\alias{spacy_extract_entities}
\title{Extract Named Entities with spaCy}
\usage{
spacy_extract_entities(x, model = "en_core_web_sm")
}
\arguments{
\item{x}{Character vector of texts OR a quanteda tokens object.}

\item{model}{Character; spaCy model to use (default: "en_core_web_sm").}
}
\value{
A data frame with entity information:
\itemize{
\item \code{doc_id}: Document identifier
\item \code{text}: Entity text
\item \code{label}: Entity type (PERSON, ORG, GPE, etc.)
\item \code{start_char}: Start character position
\item \code{end_char}: End character position
}
}
\description{
Extract named entities from texts using spaCy NER.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{extract_pos_tags}
\alias{extract_pos_tags}
\title{Extract Part-of-Speech Tags from Tokens}
\usage{
extract_pos_tags(
  tokens,
  include_lemma = TRUE,
  include_entity = FALSE,
  include_dependency = FALSE,
  model = "en_core_web_sm"
)
}
\arguments{
\item{tokens}{A quanteda tokens object or character vector of texts.}

\item{include_lemma}{Logical; include lemmatized forms (default: TRUE).}

\item{include_entity}{Logical; include named entity recognition (default: FALSE).}

\item{include_dependency}{Logical; include dependency parsing (default: FALSE).}

\item{model}{Character; spaCy model to use (default: "en_core_web_sm").}
}
\value{
A data frame with columns:
\itemize{
\item \code{doc_id}: Document identifier
\item \code{sentence_id}: Sentence number within document
\item \code{token_id}: Token position within sentence
\item \code{token}: Original token
\item \code{pos}: Universal POS tag (e.g., NOUN, VERB, ADJ)
\item \code{tag}: Detailed POS tag (e.g., NN, VBD, JJ)
\item \code{lemma}: Lemmatized form (if include_lemma = TRUE)
\item \code{entity}: Named entity type (if include_entity = TRUE)
\item \code{head_token_id}: Head token in dependency tree (if include_dependency = TRUE)
\item \code{dep_rel}: Dependency relation type, e.g., nsubj, dobj (if include_dependency = TRUE)
}
}
\description{
Uses spaCy to extract part-of-speech (POS) tags from tokenized text.
Returns a data frame with token-level POS annotations.
}
\details{
This function requires the Python
with spaCy installed. If spaCy is not initialized, this function will
attempt to initialize it with the specified model.
}
\examples{
if (interactive()) {
  tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1])
  pos_data <- extract_pos_tags(tokens)
  print(pos_data)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{call_llm_api}
\alias{call_llm_api}
\title{Call LLM API (Unified Wrapper)}
\usage{
call_llm_api(
  provider = c("openai", "gemini"),
  system_prompt,
  user_prompt,
  model = NULL,
  temperature = 0,
  max_tokens = 150,
  api_key = NULL
)
}
\arguments{
\item{provider}{Character string: "openai" or "gemini"}

\item{system_prompt}{Character string with system instructions}

\item{user_prompt}{Character string with user message}

\item{model}{Character string specifying the model (provider-specific defaults apply)}

\item{temperature}{Numeric temperature for response randomness (default: 0)}

\item{max_tokens}{Maximum number of tokens to generate (default: 150)}

\item{api_key}{Character string with API key (required for openai/gemini)}
}
\value{
Character string with the model's response
}
\description{
Unified wrapper for calling different LLM providers (OpenAI, Gemini).
Automatically routes to the appropriate provider-specific function.
}
\examples{
if (interactive()) {
# Using OpenAI
response <- call_llm_api(
  provider = "openai",
  system_prompt = "You are a helpful assistant.",
  user_prompt = "Generate a topic label",
  api_key = Sys.getenv("OPENAI_API_KEY")
)

# Using Gemini
response <- call_llm_api(
  provider = "gemini",
  system_prompt = "You are a helpful assistant.",
  user_prompt = "Generate a topic label",
  api_key = Sys.getenv("GEMINI_API_KEY")
)
}
}
\seealso{
\code{\link[=sanitize_llm_input]{sanitize_llm_input()}} to clean text before prompting; \code{\link[=get_best_embeddings]{get_best_embeddings()}} for vector embeddings; \code{\link[=run_rag_search]{run_rag_search()}} for RAG search (retrieval + generation)
}
\concept{ai}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{sanitize_text_input}
\alias{sanitize_text_input}
\title{Sanitize Text Input}
\usage{
sanitize_text_input(text)
}
\arguments{
\item{text}{Text input from user}
}
\value{
Sanitized text
}
\description{
Sanitize Text Input
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{fit_semantic_model}
\alias{fit_semantic_model}
\title{Fit Semantic Model}
\usage{
fit_semantic_model(
  texts,
  analysis_types = c("similarity", "dimensionality_reduction", "clustering"),
  document_feature_type = "embeddings",
  similarity_method = "cosine",
  use_embeddings = TRUE,
  embedding_model = "all-MiniLM-L6-v2",
  dimred_method = "UMAP",
  clustering_method = "umap_dbscan",
  n_components = 2,
  n_clusters = 5,
  seed = 123,
  verbose = TRUE
)
}
\arguments{
\item{texts}{A character vector of texts to analyze.}

\item{analysis_types}{Types of analysis to perform: "similarity", "dimensionality_reduction", "clustering".}

\item{document_feature_type}{Feature extraction type (default: "embeddings").}

\item{similarity_method}{Similarity calculation method (default: "cosine").}

\item{use_embeddings}{Logical, use embedding-based approaches (default: TRUE).}

\item{embedding_model}{Sentence transformer model name (default: "all-MiniLM-L6-v2").}

\item{dimred_method}{Dimensionality reduction method: "PCA", "t-SNE", "UMAP" (default: "UMAP").}

\item{clustering_method}{Clustering method: "kmeans", "hierarchical", "umap_dbscan" (default: "umap_dbscan").}

\item{n_components}{Number of dimensions for reduction (default: 2).}

\item{n_clusters}{Number of clusters (default: 5).}

\item{seed}{Random seed for reproducibility (default: 123).}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A list containing results from requested analyses.
}
\description{
Performs semantic analysis including similarity, dimensionality reduction,
and clustering. This is a high-level wrapper function.
}
\examples{
if (interactive()) {
  data(SpecialEduTech)
  texts <- SpecialEduTech$abstract[1:10]

  results <- fit_semantic_model(
    texts = texts,
    analysis_types = c("similarity", "clustering")
  )

  print(results)
}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_web_deployment}
\alias{check_web_deployment}
\title{Check Deployment Environment}
\usage{
check_web_deployment()
}
\value{
Logical TRUE if running on web server, FALSE if local
}
\description{
Detects whether the app is running on a web server (shinyapps.io, Posit Connect)
versus locally via \code{run_app()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_npmi}
\alias{calculate_npmi}
\title{Calculate NPMI for Topic Top-Terms}
\usage{
calculate_npmi(top_terms_list, dfm, top_k = 10, epsilon = 1e-12)
}
\arguments{
\item{top_terms_list}{Named or unnamed list of character vectors, one per topic.}

\item{dfm}{A quanteda dfm providing the reference doc-term frequencies.}

\item{top_k}{Number of top terms per topic to include (default 10).}

\item{epsilon}{Numerical stabilizer (default 1e-12).}
}
\value{
list with \code{mean_npmi} (scalar) and \code{per_topic_npmi} (numeric vector).
}
\description{
NPMI (Normalized Pointwise Mutual Information) measures coherence of top-K
terms per topic using internal document co-occurrence from the supplied DFM
(boolean presence).
}
\concept{topic-modeling}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{get_spacy_model_info}
\alias{get_spacy_model_info}
\title{Get spaCy Model Information}
\usage{
get_spacy_model_info()
}
\value{
A list with model information including name, language,
pipeline components, and vector availability.
}
\description{
Get information about the currently loaded spaCy model.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{describe_image}
\alias{describe_image}
\title{Describe Image Using Vision LLM}
\usage{
describe_image(
  image_base64,
  provider = "gemini",
  model = NULL,
  api_key = NULL,
  prompt =
    "Describe this image: charts, diagrams, tables, and text. Extract visible text.",
  timeout = 120
)
}
\arguments{
\item{image_base64}{Character string of base64-encoded PNG image}

\item{provider}{Character: "openai" or "gemini"}

\item{model}{Character: Model name (uses provider default if NULL)}

\item{api_key}{Character: API key (required for openai/gemini)}

\item{prompt}{Character: Description prompt}

\item{timeout}{Numeric: Request timeout in seconds (default: 120)}
}
\value{
Character string description, or NULL on failure
}
\description{
Unified dispatcher for image description using vision LLMs.
Routes to the appropriate provider (OpenAI or Gemini).
}
\concept{ai}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_term_trends_continuous}
\alias{plot_term_trends_continuous}
\title{Plot Term Frequency Trends by Continuous Variable}
\usage{
plot_term_trends_continuous(
  term_data,
  continuous_var,
  terms = NULL,
  title = NULL,
  height = 600,
  width = NULL
)
}
\arguments{
\item{term_data}{Data frame containing term frequencies with columns:
continuous_var, term, and word_frequency}

\item{continuous_var}{Name of the continuous variable column}

\item{terms}{Character vector of terms to display (optional, filters if provided)}

\item{title}{Plot title (default: NULL, auto-generated)}

\item{height}{Plot height in pixels (default: 600)}

\item{width}{Plot width in pixels (default: NULL, auto)}
}
\value{
A plotly object with faceted line plots
}
\description{
Creates a faceted line plot showing how term frequencies vary across
a continuous variable (e.g., year, time period).
}
\examples{
\donttest{
term_df <- data.frame(
  year = rep(2010:2020, each = 3),
  term = rep(c("learning", "education", "technology"), 11),
  word_frequency = sample(10:100, 33, replace = TRUE)
)
plot_term_trends_continuous(term_df, "year", c("learning", "education"))
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{calculate_metrics}
\alias{calculate_metrics}
\title{Calculate Similarity Metrics}
\usage{
calculate_metrics(similarity_matrix, labels = NULL, method_info = NULL)
}
\arguments{
\item{similarity_matrix}{A similarity matrix.}

\item{labels}{Optional vector of labels for clustering metrics.}

\item{method_info}{Optional method information.}
}
\value{
A list of metrics.
}
\description{
Calculates similarity metrics including statistical measures
and network properties.
Internal function used by document_similarity_analysis.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{plot_sentiment_boxplot}
\alias{plot_sentiment_boxplot}
\title{Plot Sentiment Box Plot by Category}
\usage{
plot_sentiment_boxplot(
  sentiment_data,
  category_var = "category_var",
  title = "Sentiment Score Distribution"
)
}
\arguments{
\item{sentiment_data}{Data frame from analyze_sentiment() containing sentiment_score
and category columns}

\item{category_var}{Name of the category variable column (default: "category_var")}

\item{title}{Plot title (default: "Sentiment Score Distribution")}
}
\value{
A ggplot2 box plot
}
\description{
Creates a box plot showing sentiment score distribution by category.
}
\concept{sentiment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_lexical_diversity_distribution}
\alias{plot_lexical_diversity_distribution}
\title{Plot Lexical Diversity Distribution}
\usage{
plot_lexical_diversity_distribution(lexdiv_data, metric, title = NULL)
}
\arguments{
\item{lexdiv_data}{Data frame from lexical_diversity_analysis()}

\item{metric}{Metric to plot. Recommended: "MTLD" or "MATTR" (text-length independent)}

\item{title}{Plot title (default: auto-generated)}
}
\value{
A plotly boxplot
}
\description{
Creates a boxplot showing the distribution of a lexical diversity metric.
}
\examples{
\donttest{
abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
tokens <- quanteda::tokens(quanteda::corpus(abstracts))
diversity_result <- lexical_diversity_analysis(tokens, texts = abstracts)
diversity_plot <- plot_lexical_diversity_distribution(
  diversity_result$lexical_diversity, "MTLD"
)
print(diversity_plot)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{plot_emotion_radar}
\alias{plot_emotion_radar}
\title{Plot Emotion Radar Chart}
\usage{
plot_emotion_radar(
  emotion_data,
  group_var = NULL,
  normalize = FALSE,
  title = "Emotion Analysis"
)
}
\arguments{
\item{emotion_data}{Data frame with emotion scores (columns: emotion, total_score)}

\item{group_var}{Optional grouping variable column name for overlaid radars (default: NULL)}

\item{normalize}{Logical, whether to normalize scores to 0-100 scale (default: FALSE)}

\item{title}{Plot title (default: "Emotion Analysis")}
}
\value{
A ggplot2 polar chart
}
\description{
Creates a polar/radar chart for NRC emotion analysis with optional grouping.
}
\concept{sentiment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_app.R
\name{run_app}
\alias{run_app}
\title{Launch the TextAnalysisR app}
\usage{
run_app(launch.browser = interactive())
}
\arguments{
\item{launch.browser}{Logical. Whether to open the app in a browser.
Defaults to \code{interactive()}, which is FALSE in non-interactive sessions
(e.g., Docker containers, servers).}
}
\value{
No return value, called for side effects (launching Shiny app).
}
\description{
Launch the TextAnalysisR Shiny application.
}
\examples{
if (interactive()) {
  run_app()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{setup_python_env}
\alias{setup_python_env}
\title{Setup Python Environment}
\usage{
setup_python_env(envname = "textanalysisr-env", tier = "core", force = FALSE)
}
\arguments{
\item{envname}{Character string name for the virtual environment
(default: "textanalysisr-env")}

\item{tier}{Which feature tier to install. One or more of:
\code{"core"} (spacy + pdfplumber, ~200 MB; default),
\code{"embeddings"} (adds sentence-transformers + transformers + torch, ~1 GB),
\code{"topics"} (adds BERTopic + UMAP + HDBSCAN, ~300 MB on top of embeddings).}

\item{force}{Logical, whether to recreate environment if it exists
(default: FALSE)}
}
\value{
Invisible TRUE if successful, stops with error message if failed
}
\description{
Sets up a tiered Python virtual environment.
}
\details{
Default \code{tier = "core"} keeps the install light -- spaCy NLP and PDF text
extraction only. Add \code{"embeddings"} for sentence-transformers-based
similarity/sentiment, and \code{"topics"} for BERTopic.

The virtual environment is isolated; system Python is not modified.
}
\examples{
if (interactive()) {
  setup_python_env()                              # core only (~200 MB)
  setup_python_env(tier = c("core", "embeddings"))  # +1 GB
  setup_python_env(tier = c("core", "embeddings", "topics"))  # full stack
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{lexical_analysis}
\alias{lexical_analysis}
\title{Lexical Analysis Functions}
\description{
Functions for lexical analysis including:
\itemize{
\item Linguistic Annotation (POS tagging, NER)
\item Frequency Analysis (word frequency, n-grams, MWEs)
\item Keywords (TF-IDF, keyness)
\item Lexical Diversity (TTR, MTLD, MATTR)
\item Readability (Flesch, Gunning Fog, etc.)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_no_dfm_notification}
\alias{show_no_dfm_notification}
\title{Show no-DFM notification}
\usage{
show_no_dfm_notification(feature_name = "this feature", duration = 7)
}
\arguments{
\item{feature_name}{Feature description in the message.}

\item{duration}{Seconds until auto-dismiss.}
}
\value{
Invisibly the notification id.
}
\description{
Show no-DFM notification
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{plot_cluster_terms}
\alias{plot_cluster_terms}
\title{Plot Cluster Top Terms}
\usage{
plot_cluster_terms(
  terms,
  cluster_id = NULL,
  title = NULL,
  n_terms = 10,
  color = "#337ab7",
  height = 500,
  width = NULL
)
}
\arguments{
\item{terms}{Named numeric vector of term frequencies, or data frame with
'term' and 'frequency' columns}

\item{cluster_id}{Cluster identifier for the title (default: NULL)}

\item{title}{Custom title (default: NULL, auto-generated from cluster_id)}

\item{n_terms}{Number of top terms to display (default: 10)}

\item{color}{Bar color (default: "#337ab7")}

\item{height}{Plot height in pixels (default: 500)}

\item{width}{Plot width in pixels (default: NULL for auto)}
}
\value{
A plotly object
}
\description{
Creates a horizontal bar plot showing the top terms in a cluster or document group.
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_pos_frequencies}
\alias{plot_pos_frequencies}
\title{Plot Part-of-Speech Tag Frequencies}
\usage{
plot_pos_frequencies(
  pos_data,
  top_n = 20,
  title = "Part-of-Speech Tag Frequency",
  color = "#337ab7",
  height = 500,
  width = NULL
)
}
\arguments{
\item{pos_data}{Data frame containing POS data with columns:
\itemize{
\item \code{pos}: Part-of-speech tag
\item \code{n}: (optional) Pre-computed frequency count
}
If \code{n} is not present, frequencies will be computed from the data.}

\item{top_n}{Number of top POS tags to display (default: 20)}

\item{title}{Plot title (default: "Part-of-Speech Tag Frequency")}

\item{color}{Bar color (default: "#337ab7")}

\item{height}{Plot height in pixels (default: 500)}

\item{width}{Plot width in pixels (default: NULL for auto)}
}
\value{
A plotly object
}
\description{
Creates a bar plot showing the frequency distribution of part-of-speech tags.
}
\examples{
if (interactive()) {
  pos_df <- data.frame(
    pos = c("NOUN", "VERB", "ADJ", "ADV", "PRON"),
    n = c(500, 400, 250, 150, 100)
  )
  plot_pos_frequencies(pos_df)
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{calculate_text_readability}
\alias{calculate_text_readability}
\title{Calculate Text Readability}
\usage{
calculate_text_readability(
  texts,
  metrics = c("flesch", "flesch_kincaid", "gunning_fog"),
  include_lexical_diversity = TRUE,
  include_sentence_stats = TRUE,
  doc_names = NULL
)
}
\arguments{
\item{texts}{Character vector of texts to analyze}

\item{metrics}{Character vector of readability metrics to calculate.
Options: "flesch", "flesch_kincaid", "gunning_fog", "smog", "ari", "coleman_liau"}

\item{include_lexical_diversity}{Logical, include the MTLD lexical diversity
index (default: TRUE)}

\item{include_sentence_stats}{Logical, include average sentence length (default: TRUE)}

\item{doc_names}{Optional character vector of document names}
}
\value{
A data frame with document names and readability scores
}
\description{
Calculates multiple readability metrics for texts including Flesch Reading Ease,
Flesch-Kincaid Grade Level, Gunning FOG index, and others. Optionally includes
lexical diversity metrics and sentence statistics.
}
\examples{
\donttest{
data(SpecialEduTech, package = "TextAnalysisR")
texts <- SpecialEduTech$abstract[1:10]
readability <- calculate_text_readability(texts)
print(readability)
}

}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{calculate_cross_similarity}
\alias{calculate_cross_similarity}
\title{Calculate Cross-Matrix Cosine Similarity}
\usage{
calculate_cross_similarity(
  embeddings1,
  embeddings2,
  labels1 = NULL,
  labels2 = NULL,
  normalize = TRUE
)
}
\arguments{
\item{embeddings1}{A numeric matrix where rows are items and columns are embedding dimensions.}

\item{embeddings2}{A numeric matrix where rows are items and columns are embedding dimensions.
Must have the same number of columns as embeddings1.}

\item{labels1}{Optional character vector of labels for items in embeddings1.}

\item{labels2}{Optional character vector of labels for items in embeddings2.}

\item{normalize}{Logical, whether to L2-normalize embeddings before computing similarity (default: TRUE).}
}
\value{
A list containing:
\item{similarity_matrix}{Matrix of cosine similarities (nrow(embeddings1) x nrow(embeddings2))}
\item{similarity_df}{Long-format data frame with columns: row_idx, col_idx, similarity, and optionally label1, label2}
}
\description{
Calculates cosine similarity between two different embedding matrices,
useful for comparing documents/topics across different categories or groups.
}
\examples{
\donttest{
abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:6]
term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(abstracts)))
similarity_result <- calculate_cross_similarity(
  term_matrix[1:3, ], term_matrix[4:6, ],
  labels1 = paste("Doc", 1:3),
  labels2 = paste("Doc", 4:6)
)
similarity_result$similarity_matrix
}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{clear_lexdiv_cache}
\alias{clear_lexdiv_cache}
\title{Clear Lexical Diversity Cache}
\usage{
clear_lexdiv_cache()
}
\value{
Invisible NULL
}
\description{
Clears the internal cache used for lexical diversity calculations.
Call this function to free memory or force fresh calculations.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{cluster_embeddings}
\alias{cluster_embeddings}
\title{Embedding-based Document Clustering}
\usage{
cluster_embeddings(
  data_matrix,
  method = "kmeans",
  n_clusters = 0,
  umap_neighbors = 15,
  umap_min_dist = 0.1,
  umap_n_components = 10,
  umap_metric = "cosine",
  dbscan_eps = 0,
  dbscan_min_samples = NULL,
  reduce_outliers = TRUE,
  outlier_strategy = "centroid",
  outlier_threshold = 0,
  seed = 123,
  verbose = TRUE
)
}
\arguments{
\item{data_matrix}{A numeric matrix where rows represent documents and
columns represent features.}

\item{method}{The clustering method. Options: "kmeans", "hierarchical",
"umap_dbscan".}

\item{n_clusters}{The number of clusters (for k-means and hierarchical).
If 0, optimal number is determined automatically.}

\item{umap_neighbors}{The number of neighbors for UMAP (default: 15).}

\item{umap_min_dist}{The minimum distance for UMAP (default: 0.1).}

\item{umap_n_components}{The number of UMAP components (default: 10).}

\item{umap_metric}{The metric for UMAP (default: "cosine").}

\item{dbscan_eps}{The eps parameter for DBSCAN. If 0, optimal value is determined automatically.}

\item{dbscan_min_samples}{The minimum samples for DBSCAN. Default NULL sets
\code{umap_n_components + 1} (dimensions-plus-one rule of thumb).}

\item{reduce_outliers}{Logical, if TRUE, reassigns noise points (cluster 0) to nearest cluster (default: TRUE).
Set FALSE to keep density-based noise labels.}

\item{outlier_strategy}{Strategy for outlier reduction: "centroid" (default,
Euclidean nearest-centroid in UMAP space; package-specific heuristic) or
"embeddings" (cosine similarity to cluster mean in original embedding
space, matching the BERTopic "embeddings" strategy).}

\item{outlier_threshold}{Minimum similarity for reassigning a noise point
(default: 0, reassign all). For "embeddings" this is cosine similarity;
for "centroid" it applies to 1 / (1 + euclidean distance). Points below
the threshold keep cluster label 0.}

\item{seed}{Random seed for reproducibility (default: 123).}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A list containing cluster assignments, method used, and quality metrics.
For \code{method = "umap_dbscan"} with \code{reduce_outliers = TRUE}, the list
also reports how noise points were reassigned: \code{outlier_strategy}
("embeddings" or "centroid"), \code{outlier_metric} ("cosine" or "euclidean"),
\code{outlier_space} ("original embedding space" or "UMAP space"), and
\code{outliers_reassigned} (count).
}
\description{
This function performs clustering analysis using various methods, ordered
from simple to detailed:
k-means (simplest), hierarchical (intermediate), and UMAP+DBSCAN
(most detailed).
}
\examples{
if (interactive()) {
  mydata <- TextAnalysisR::SpecialEduTech

  united_tbl <- TextAnalysisR::unite_cols(
    mydata,
    listed_vars = c("title", "keyword", "abstract")
  )

  tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")

  dfm_object <- quanteda::dfm(tokens)

  data_matrix <- as.matrix(dfm_object)

  kmeans_result <- TextAnalysisR::cluster_embeddings(
    data_matrix,
    method = "kmeans",
    n_clusters = 5
  )
  print(kmeans_result)

  hierarchical_result <- TextAnalysisR::cluster_embeddings(
    data_matrix,
    method = "hierarchical",
    n_clusters = 5
  )
  print(hierarchical_result)

  umap_dbscan_result <- TextAnalysisR::cluster_embeddings(
    data_matrix,
    method = "umap_dbscan",
    umap_neighbors = 15,
    umap_min_dist = 0.1
  )
  print(umap_dbscan_result)
}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{spacy_parse_full}
\alias{spacy_parse_full}
\title{Parse Texts with spaCy}
\usage{
spacy_parse_full(
  x,
  pos = TRUE,
  tag = TRUE,
  lemma = TRUE,
  entity = FALSE,
  dependency = FALSE,
  morph = FALSE,
  model = "en_core_web_sm"
)
}
\arguments{
\item{x}{Character vector of texts OR a quanteda tokens object.}

\item{pos}{Logical; include coarse POS tags (default: TRUE).}

\item{tag}{Logical; include fine-grained tags (default: TRUE).}

\item{lemma}{Logical; include lemmatized forms (default: TRUE).}

\item{entity}{Logical; include named entity tags (default: FALSE).}

\item{dependency}{Logical; include dependency relations (default: FALSE).}

\item{morph}{Logical; include morphological features (default: FALSE).}

\item{model}{Character; spaCy model to use (default: "en_core_web_sm").}
}
\value{
A data frame with token-level annotations including:
\itemize{
\item \code{doc_id}: Document identifier
\item \code{sentence_id}: Sentence number within document
\item \code{token_id}: Token position within sentence
\item \code{token}: Original token text
\item \code{pos}: Coarse POS tag (if pos = TRUE)
\item \code{tag}: Fine-grained tag (if tag = TRUE)
\item \code{lemma}: Lemmatized form (if lemma = TRUE)
\item \code{entity}: Named entity tag (if entity = TRUE)
\item \code{head_token_id}: Head token ID (if dependency = TRUE)
\item \code{dep_rel}: Dependency relation (if dependency = TRUE)
\item \code{morph}: Morphological features string (if morph = TRUE)
}
}
\description{
Parse texts using spaCy and return token-level annotations.
This is the main parsing function for NLP analysis.
Works with character vectors or quanteda tokens objects.
}
\examples{
if (interactive()) {
# From SpecialEduTech dataset
texts <- TextAnalysisR::SpecialEduTech$abstract[1:5]
parsed <- spacy_parse_full(texts, morph = TRUE)

# From quanteda tokens
united <- unite_cols(TextAnalysisR::SpecialEduTech, c("title", "abstract"))
tokens <- prep_texts(united, text_field = "united_texts")
parsed <- spacy_parse_full(tokens, morph = TRUE)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{remove_notification_by_id}
\alias{remove_notification_by_id}
\title{Remove Shiny notification by id}
\usage{
remove_notification_by_id(id, session = NULL)
}
\arguments{
\item{id}{Notification id.}

\item{session}{Shiny session. Defaults to the current reactive domain.}
}
\value{
Invisibly \code{NULL}.
}
\description{
Remove Shiny notification by id
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{show_web_banner}
\alias{show_web_banner}
\title{Show Web Deployment Banner}
\usage{
show_web_banner(disabled = NULL)
}
\arguments{
\item{disabled}{Optional character vector naming features to mark as
disabled in the banner; \code{NULL} (default) shows the standard set.}
}
\value{
A \code{shiny.tag} object containing the banner HTML for
inclusion in a Shiny UI, or \code{NULL} when not running in a web
deployment context.
}
\description{
Show Web Deployment Banner
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{lemmatize_tokens}
\alias{lemmatize_tokens}
\title{Lemmatize Tokens with Batch Processing}
\usage{
lemmatize_tokens(
  tokens,
  batch_size = 50,
  model = "en_core_web_sm",
  verbose = TRUE
)
}
\arguments{
\item{tokens}{A quanteda tokens object to lemmatize.}

\item{batch_size}{Integer; number of documents to process per batch (default: 50).}

\item{model}{Character; spaCy model to use (default: "en_core_web_sm").}

\item{verbose}{Logical; print progress messages (default: TRUE).}
}
\value{
A quanteda tokens object containing lemmatized tokens.
}
\description{
Converts tokens to their lemmatized forms using spaCy, with batch processing
to handle large document collections without timeout issues.
}
\details{
Uses spaCy for linguistic lemmatization producing proper dictionary forms
(e.g., "studies" -> "study", "better" -> "good").
Batch processing prevents timeout errors with large document collections.
}
\examples{
if (interactive()) {
tokens <- quanteda::tokens(c("The studies showed better results"))
lemmatized <- lemmatize_tokens(tokens, batch_size = 50)
}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{process_pdf_file_py}
\alias{process_pdf_file_py}
\title{Process PDF File using Python}
\usage{
process_pdf_file_py(file_path, content_type = "auto", envname = NULL)
}
\arguments{
\item{file_path}{Character string path to PDF file}

\item{content_type}{Character string: "auto", "text", or "tabular"
If "auto", will detect content type automatically}

\item{envname}{Character string, name of Python virtual environment
(default: "textanalysisr-env")}
}
\value{
List with:
\itemize{
\item data: Data frame with extracted content
\item type: Character string indicating content type
\item success: Logical indicating success
\item message: Character string with status message
}
}
\description{
Main function to process PDF files using pdfplumber (Python).
Automatically detects content type and extracts data accordingly.
No Java required.
}
\details{
This function uses Python's pdfplumber library which:
\itemize{
\item Handles both text and tables
\item No Java dependency
\item Better accuracy than tabulizer for complex tables
\item Uses TextAnalysisR Python environment
}
}
\concept{pdf}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{generate_cluster_labels_auto}
\alias{generate_cluster_labels_auto}
\title{Generate Cluster Labels}
\usage{
generate_cluster_labels_auto(
  feature_matrix,
  clusters,
  method = "tfidf",
  n_terms = 3
)
}
\arguments{
\item{feature_matrix}{Feature matrix used for clustering}

\item{clusters}{Cluster assignments}

\item{method}{Label generation method ("tfidf", "representative", "frequent")}

\item{n_terms}{Number of terms per label}
}
\value{
Named list of cluster labels
}
\description{
Generate descriptive labels for document clusters
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{calculate_similarity_robust}
\alias{calculate_similarity_robust}
\title{Calculate Document Similarity with Fallbacks}
\usage{
calculate_similarity_robust(
  texts,
  method = "embeddings",
  embedding_model = "all-MiniLM-L6-v2",
  cache_embeddings = TRUE,
  min_word_length = 3,
  doc_names = NULL
)
}
\arguments{
\item{texts}{Character vector of texts}

\item{method}{Similarity method ("embeddings" or "jaccard")}

\item{embedding_model}{Model name for embeddings (default: "all-MiniLM-L6-v2")}

\item{cache_embeddings}{Logical, cache embeddings (default: TRUE)}

\item{min_word_length}{Minimum word length for Jaccard (default: 3)}

\item{doc_names}{Optional document names}
}
\value{
List containing similarity matrix, method used, embeddings, and diagnostics
}
\description{
Calculates document similarity with fallback methods and diagnostics.
Attempts embeddings first, falls back to Jaccard similarity if needed.
}
\examples{
if (interactive()) {
  abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:5]
  similarity_result <- calculate_similarity_robust(abstracts)
  print(similarity_result$similarity_matrix)
  print(similarity_result$diagnostics)
}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{fit_embedding_model}
\alias{fit_embedding_model}
\title{Fit Embedding-based Topic Model}
\usage{
fit_embedding_model(
  texts,
  method = "umap_hdbscan",
  n_topics = 10,
  embedding_model = "all-MiniLM-L6-v2",
  backend = "auto",
  clustering_method = "kmeans",
  similarity_threshold = 0.7,
  min_topic_size = 10,
  min_cluster_size = NULL,
  cluster_selection_method = "eom",
  umap_neighbors = 15,
  umap_min_dist = 0,
  umap_n_components = 5,
  umap_metric = "cosine",
  tsne_perplexity = 30,
  pca_dims = 50,
  dbscan_eps = 0.5,
  dbscan_minpts = 5,
  representation_method = "c-tfidf",
  diversity = 0.5,
  reduce_outliers = TRUE,
  outlier_strategy = "probabilities",
  outlier_threshold = 0,
  seed = 123,
  verbose = TRUE,
  precomputed_embeddings = NULL
)
}
\arguments{
\item{texts}{A character vector of texts to analyze.}

\item{method}{The topic modeling method:
\itemize{
\item For Python backend: "umap_hdbscan" (uses BERTopic)
\item For R backend: "umap_dbscan", "umap_kmeans", "umap_hierarchical",
"tsne_dbscan", "tsne_kmeans", "pca_kmeans", "pca_hierarchical"
\item For both: "embedding_clustering", "hierarchical_semantic"
}}

\item{n_topics}{The number of topics to identify. For UMAP+HDBSCAN, use NULL or "auto" for automatic determination, or specify an integer.}

\item{embedding_model}{The embedding model to use (default: "all-MiniLM-L6-v2").}

\item{backend}{The backend to use: "auto" (default, tries Python then R),
"python" (requires BERTopic), or "r" (R-native packages only).}

\item{clustering_method}{The clustering method for embedding-based approach: "kmeans", "hierarchical", "dbscan", "hdbscan".}

\item{similarity_threshold}{The similarity threshold for topic assignment (default: 0.7).}

\item{min_topic_size}{The minimum number of documents per topic (default: 3).}

\item{min_cluster_size}{HDBSCAN density threshold (default \code{NULL} falls back to \code{min_topic_size}). Setting this independently lets fine-grained clusters merge into broader topics.}

\item{cluster_selection_method}{HDBSCAN cluster selection method: "eom" (Excess of Mass, default) or "leaf" (finer-grained topics).}

\item{umap_neighbors}{The number of neighbors for UMAP dimensionality reduction (default: 15).}

\item{umap_min_dist}{The minimum distance for UMAP (default: 0.0). Use 0.0 for tight, well-separated clusters. Use 0.1+ for visualization purposes. Range: 0.0-0.99.}

\item{umap_n_components}{The number of UMAP components (default: 5).}

\item{umap_metric}{Distance metric for UMAP: "cosine" (recommended for text) or "euclidean" (default: "cosine").}

\item{tsne_perplexity}{Perplexity parameter for t-SNE (default: 30). Only used when method includes "tsne"; t-SNE output is 2-dimensional.}

\item{pca_dims}{Number of PCA components kept for clustering (default: 50). Only used when method includes "pca".}

\item{dbscan_eps}{Epsilon parameter for DBSCAN (default: 0.5). Neighborhood size for density-based clustering.}

\item{dbscan_minpts}{Minimum points for DBSCAN core points (default: 5).}

\item{representation_method}{The method for topic representation: "c-tfidf", "tfidf", "mmr", "frequency" (default: "c-tfidf").
Applies only to the R fallback backend and the "embedding_clustering",
"semantic_lda", and "hierarchical_semantic" methods; the Python BERTopic
backend ("umap_hdbscan") uses BERTopic's native c-TF-IDF representation.}

\item{diversity}{Diversity weight between 0 and 1 for the "mmr"
representation (default: 0.5). Higher values penalize redundant terms
more strongly. Applies to the R backend; ignored by the Python BERTopic
backend.}

\item{reduce_outliers}{Logical, if TRUE, reduces outliers in HDBSCAN clustering (default: TRUE).}

\item{outlier_strategy}{Strategy for outlier reduction using BERTopic:
"probabilities" (default, uses topic probabilities), "c-tf-idf" (uses
c-TF-IDF similarity), "embeddings" (uses cosine similarity in embedding
space), or "distributions" (uses topic distributions). Ignored if
reduce_outliers = FALSE.}

\item{outlier_threshold}{Minimum threshold for outlier reassignment (default: 0.0).
Higher values require stronger evidence for reassignment.}

\item{seed}{Random seed for reproducibility (default: 123).}

\item{verbose}{Logical, if TRUE, prints progress messages.}

\item{precomputed_embeddings}{Optional matrix of pre-computed document embeddings.
If provided, skips embedding generation for improved performance. Must have
the same number of rows as the length of texts.}
}
\value{
A list containing topic assignments, topic keywords, and quality metrics.
}
\description{
This function performs embedding-based topic modeling using transformer embeddings
and specialized clustering techniques. Supports two backends:
\itemize{
\item \strong{Python backend} (default): Uses BERTopic library which combines transformer
embeddings with UMAP dimensionality reduction and HDBSCAN clustering for optimal
topic discovery.
\item \strong{R backend}: Uses R-native packages (umap, dbscan, Rtsne) for users without
Python/BERTopic installed. Provides similar functionality with c-TF-IDF keyword
extraction.
}
}
\examples{
if (interactive()) {
  mydata <- TextAnalysisR::SpecialEduTech
  united_tbl <- TextAnalysisR::unite_cols(
    mydata,
    listed_vars = c("title", "keyword", "abstract")
  )
  texts <- united_tbl$united_texts

  # Embedding-based topic modeling (powered by BERTopic)
  result <- TextAnalysisR::fit_embedding_model(
    texts = texts,
    method = "umap_hdbscan",
    n_topics = 8,
    min_topic_size = 3
  )

  print(result$topic_assignments)
  print(result$topic_keywords)
}
}
\seealso{
\code{\link[=get_best_embeddings]{get_best_embeddings()}} to supply precomputed embeddings; \code{\link[=generate_topic_labels]{generate_topic_labels()}} for AI-suggested topic names; \code{\link[=find_optimal_k]{find_optimal_k()}} for an STM-based alternative
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{create_message_table}
\alias{create_message_table}
\title{Single-cell message DT table}
\usage{
create_message_table(message, font_size = "16px", color = "#6c757d")
}
\arguments{
\item{message}{Character message to display.}

\item{font_size}{CSS font size.}

\item{color}{CSS color.}
}
\value{
A \code{DT::datatable} htmlwidget.
}
\description{
Single-cell message DT table
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validate_file_upload}
\alias{validate_file_upload}
\title{Cybersecurity Utility Functions}
\usage{
validate_file_upload(file_info)
}
\arguments{
\item{file_info}{File info object from Shiny fileInput}
}
\value{
TRUE if valid, stops with error message if invalid
}
\description{
Functions for input validation, sanitization, and security logging
}
\section{NIST Compliance}{

This package follows NIST security standards (based on NIST SP 800-53):
\itemize{
\item SC-8: Transmission Confidentiality and Integrity (HTTPS encryption)
\item SC-28: Protection of Information at Rest (secure API key storage)
\item IA-5: Authenticator Management (API key validation and format checking)
\item AC-3: Access Enforcement (rate limiting, input validation, file type restrictions)
\item SI-10: Information Input Validation (malicious content detection)
\item AU-2: Audit Events (security logging and monitoring)
Validate File Upload
}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{auto_tune_embedding_topics}
\alias{auto_tune_embedding_topics}
\title{Auto-tune BERTopic Hyperparameters}
\usage{
auto_tune_embedding_topics(
  texts,
  embeddings = NULL,
  embedding_model = "all-MiniLM-L6-v2",
  n_trials = 12,
  metric = "silhouette",
  seed = 123,
  verbose = TRUE
)
}
\arguments{
\item{texts}{Character vector of documents to analyze.}

\item{embeddings}{Precomputed embeddings matrix (optional). If NULL, embeddings are generated.}

\item{embedding_model}{Embedding model name (default: "all-MiniLM-L6-v2").}

\item{n_trials}{Maximum number of configurations to try (default: 12).}

\item{metric}{Optimization metric: "silhouette", "coherence", or "combined" (default: "silhouette").}

\item{seed}{Random seed for reproducibility.}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A list containing:
\itemize{
\item best_config: Data frame with the optimal hyperparameter configuration
\item best_model: The topic model fitted with optimal parameters
\item all_results: List of all evaluated configurations with metrics
\item n_trials_completed: Number of configurations successfully evaluated
}
}
\description{
Automatically searches for optimal hyperparameters for embedding-based topic modeling.
Evaluates multiple configurations of UMAP and HDBSCAN parameters and returns the best
model based on the specified metric. Embeddings are generated once and reused across
all configurations for efficiency.
}
\details{
The function searches over these parameters:
\itemize{
\item n_neighbors: UMAP neighborhood size (5, 10, 15, 25)
\item min_cluster_size: HDBSCAN minimum cluster size (3, 5, 10)
\item cluster_selection_method: "eom" (broader) or "leaf" (finer-grained)
}
}
\examples{
if (interactive()) {
  texts <- c("Machine learning for image recognition",
             "Deep learning neural networks",
             "Natural language processing models",
             "Computer vision applications")

  tuning_result <- auto_tune_embedding_topics(
    texts = texts,
    n_trials = 6,
    metric = "silhouette",
    verbose = TRUE
  )

  # View best configuration
  tuning_result$best_config

  # Use the best model
  best_model <- tuning_result$best_model
}
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{get_topic_texts}
\alias{get_topic_texts}
\title{Convert Topic Terms to Text Strings}
\usage{
get_topic_texts(
  top_terms_df,
  topic_var = "topic",
  term_var = "term",
  weight_var = NULL,
  sep = " ",
  top_n = NULL
)
}
\arguments{
\item{top_terms_df}{A data frame containing top terms for topics, typically
output from \code{\link{get_topic_terms}}.}

\item{topic_var}{Name of the column containing topic identifiers (default: "topic").}

\item{term_var}{Name of the column containing terms (default: "term").}

\item{weight_var}{Optional name of column with term weights (e.g., "beta").
If provided, terms are ordered by weight before concatenation.}

\item{sep}{Separator between terms (default: " ").}

\item{top_n}{Optional number of top terms to include per topic (default: NULL, uses all).}
}
\value{
A character vector of topic text strings, one per topic, ordered by topic number.
}
\description{
Concatenates top terms for each topic into text strings suitable for
embedding generation. Useful for creating topic representations for
semantic similarity analysis.
}
\examples{
# Topic-term frame as produced by get_topic_terms()
top_terms <- data.frame(
  topic = c(1, 1, 1, 2, 2, 2),
  term  = c("calculator", "arithmetic", "elementary",
            "computer", "instruction", "multiplication"),
  prob  = c(0.10, 0.08, 0.07, 0.12, 0.09, 0.06)
)
get_topic_texts(top_terms)
get_topic_texts(top_terms, weight_var = "prob", top_n = 2)
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_feature_status}
\alias{get_feature_status}
\title{Get Feature Status}
\usage{
get_feature_status()
}
\value{
Named list with feature availability
}
\description{
Returns availability status for all optional features.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{extract_cross_category_similarities}
\alias{extract_cross_category_similarities}
\title{Extract Cross-Category Similarities from Full Similarity Matrix}
\usage{
extract_cross_category_similarities(
  similarity_matrix,
  docs_data,
  reference_category,
  compare_categories = NULL,
  category_var = "category",
  id_var = "display_name",
  name_var = NULL
)
}
\arguments{
\item{similarity_matrix}{A square similarity matrix (n x n).}

\item{docs_data}{A data frame containing document metadata with at least:
\describe{
\item{category_var}{Column indicating category membership}
\item{id_var}{Column with unique document identifiers}
}}

\item{reference_category}{Character string specifying the reference category to compare against.}

\item{compare_categories}{Character vector of categories to compare with the reference.
If NULL, compares with all categories except reference.}

\item{category_var}{Name of the column containing category information (default: "category").}

\item{id_var}{Name of the column containing document IDs (default: "display_name").}

\item{name_var}{Optional name of column with display names (default: NULL, uses id_var).}
}
\value{
A data frame with columns:
\item{ref_id}{Reference document ID}
\item{ref_name}{Reference document name (if name_var provided)}
\item{other_id}{Comparison document ID}
\item{other_name}{Comparison document name (if name_var provided)}
\item{other_category}{Category of comparison document}
\item{similarity}{Cosine similarity value}
}
\description{
Given a full similarity matrix and category information, extracts pairwise
similarities between a reference category and other categories into a long-format
data frame suitable for visualization and analysis.
}
\examples{
\donttest{
articles <- TextAnalysisR::SpecialEduTech[1:6, ]
articles$display_name <- paste0("d", seq_len(nrow(articles)))
term_matrix <- as.matrix(quanteda::dfm(quanteda::tokens(articles$abstract)))
normalized_matrix <- term_matrix / sqrt(rowSums(term_matrix ^ 2))
similarity_matrix <- normalized_matrix \%*\% t(normalized_matrix)
dimnames(similarity_matrix) <- list(articles$display_name, articles$display_name)
cross_similarities <- extract_cross_category_similarities(
  similarity_matrix  = similarity_matrix,
  docs_data          = articles,
  reference_category = "thesis",
  compare_categories = "journal_article",
  category_var       = "reference_type",
  id_var             = "display_name",
  name_var           = "title"
)
}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{detect_pdf_content_type_py}
\alias{detect_pdf_content_type_py}
\title{Detect PDF Content Type using Python}
\usage{
detect_pdf_content_type_py(file_path, envname = NULL)
}
\arguments{
\item{file_path}{Character string path to PDF file}

\item{envname}{Character string, name of Python virtual environment
(default: "textanalysisr-env")}
}
\value{
Character string: "tabular", "text", or "unknown"
}
\description{
Analyzes PDF to determine if it contains primarily tabular data or text.
}
\examples{
if (interactive()) {
setup_python_env()

pdf_path <- "path/to/document.pdf"
content_type <- detect_pdf_content_type_py(pdf_path)
print(content_type)
}
}
\concept{pdf}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{create_aria_label}
\alias{create_aria_label}
\title{Generate ARIA Label}
\usage{
create_aria_label(element_type, action, context = NULL)
}
\arguments{
\item{element_type}{Type of element (e.g., "button", "input", "plot")}

\item{action}{Action or purpose (e.g., "analyze", "download", "visualize")}

\item{context}{Additional context (optional)}
}
\value{
Character string with ARIA label
}
\description{
Creates accessible ARIA label for UI elements.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{generate_topic_keywords}
\alias{generate_topic_keywords}
\title{Generate Topic Keywords}
\usage{
generate_topic_keywords(texts, topic_assignments, n_keywords = 10)
}
\arguments{
\item{texts}{A character vector of texts.}

\item{topic_assignments}{A vector of topic assignments.}

\item{n_keywords}{The number of keywords to extract per topic.}
}
\value{
A list of keywords for each topic.
}
\description{
Internal function to generate keywords for topics using TF-IDF analysis.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{calculate_word_frequency}
\alias{calculate_word_frequency}
\title{Analyze and Visualize Word Frequencies Across a Continuous Variable}
\usage{
calculate_word_frequency(
  dfm_object,
  continuous_variable,
  selected_terms,
  height = 500,
  width = 900
)
}
\arguments{
\item{dfm_object}{A quanteda document-feature matrix (dfm).}

\item{continuous_variable}{A continuous variable in the metadata.}

\item{selected_terms}{A vector of terms to analyze trends for.}

\item{height}{The height of the resulting Plotly plot, in pixels (default: 500).}

\item{width}{The width of the resulting Plotly plot, in pixels (default: 900).}
}
\value{
A list containing Plotly objects and tables with the results.
}
\description{
This function analyzes and visualizes word frequencies across a continuous variable.
}
\details{
This function requires a fitted STM model object and a quanteda dfm object.
The continuous variable should be a column in the metadata of the dfm object.
The selected terms should be a vector of terms to analyze trends for.
The required packages are 'htmltools', 'splines', and 'broom' (plus
additional ones loaded internally).
}
\examples{
\donttest{
  mydata <- TextAnalysisR::SpecialEduTech

  united_tbl <- TextAnalysisR::unite_cols(
    mydata,
    listed_vars = c("title", "keyword", "abstract")
  )

  tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")

  dfm_object <- quanteda::dfm(tokens)

  word_freq_results <- TextAnalysisR::calculate_word_frequency(
    dfm_object,
    continuous_variable = "year",
    selected_terms = c("calculator", "computer"),
    height = 500,
    width = 900
  )
  print(word_freq_results$plot)
  print(word_freq_results$table)
}
}
\seealso{
\code{\link[=extract_keywords_tfidf]{extract_keywords_tfidf()}} and \code{\link[=extract_keywords_keyness]{extract_keywords_keyness()}} for ranking terms by importance; \code{\link[=plot_word_frequency]{plot_word_frequency()}} for the standard frequency-by-doc plot
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{reduce_dimensions}
\alias{reduce_dimensions}
\title{Dimensionality Reduction Analysis}
\usage{
reduce_dimensions(
  data_matrix,
  method = "PCA",
  n_components = 2,
  pca_dims = 50,
  tsne_perplexity = 30,
  tsne_max_iter = 1000,
  umap_neighbors = 15,
  umap_min_dist = 0.1,
  umap_metric = "cosine",
  seed = 123,
  verbose = TRUE
)
}
\arguments{
\item{data_matrix}{A numeric matrix where rows represent documents and
columns represent features.}

\item{method}{The dimensionality reduction method. Options: "PCA", "t-SNE", "UMAP".}

\item{n_components}{The number of components/dimensions to reduce to (default: 2).}

\item{pca_dims}{The number of dimensions for PCA preprocessing (default: 50).}

\item{tsne_perplexity}{The perplexity parameter for t-SNE (default: 30).}

\item{tsne_max_iter}{The maximum number of iterations for t-SNE (default: 1000).}

\item{umap_neighbors}{The number of neighbors for UMAP (default: 15).}

\item{umap_min_dist}{The minimum distance for UMAP (default: 0.1).}

\item{umap_metric}{The metric for UMAP (default: "cosine").}

\item{seed}{Random seed for reproducibility (default: 123).}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A list containing the reduced dimensions, method used, and additional metadata.
}
\description{
This function performs dimensionality reduction using various methods
including PCA, t-SNE, and UMAP.
UMAP runs directly on inputs with 1000 or fewer features (e.g. dense
embeddings); PCA pre-reduction applies to wider inputs and to the PCA and
t-SNE methods.
}
\examples{
if (interactive()) {
  mydata <- TextAnalysisR::SpecialEduTech

  united_tbl <- TextAnalysisR::unite_cols(
    mydata,
    listed_vars = c("title", "keyword", "abstract")
  )

  tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")

  dfm_object <- quanteda::dfm(tokens)

  data_matrix <- as.matrix(dfm_object)

  pca_result <- TextAnalysisR::reduce_dimensions(
    data_matrix,
    method = "PCA"
  )
  print(pca_result)

  tsne_result <- TextAnalysisR::reduce_dimensions(
    data_matrix,
    method = "t-SNE"
  )
  print(tsne_result)

  umap_result <- TextAnalysisR::reduce_dimensions(
    data_matrix,
    method = "UMAP"
  )
  print(umap_result)
}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_ngram_frequency}
\alias{plot_ngram_frequency}
\title{Plot N-gram Frequency}
\usage{
plot_ngram_frequency(
  ngram_data,
  top_n = 30,
  selected = NULL,
  title = "N-gram Frequency",
  highlight_color = "#10B981",
  default_color = "#6B7280",
  height = 500,
  width = NULL,
  show_stats = TRUE
)
}
\arguments{
\item{ngram_data}{Data frame containing n-gram data with columns:
\itemize{
\item \code{collocation}: The n-gram text
\item \code{count}: Frequency count
\item \code{lambda}: (optional) Lambda statistic
\item \code{z}: (optional) Z-score statistic
}}

\item{top_n}{Number of top n-grams to display (default: 30)}

\item{selected}{Character vector of selected n-grams to highlight (default: NULL)}

\item{title}{Plot title (default: "N-gram Frequency")}

\item{highlight_color}{Color for highlighted bars (default: "#10B981")}

\item{default_color}{Color for non-highlighted bars (default: "#6B7280")}

\item{height}{Plot height in pixels (default: 500)}

\item{width}{Plot width in pixels (default: NULL for auto)}

\item{show_stats}{Whether to show lambda and z-score in hover (default: TRUE)}
}
\value{
A plotly object
}
\description{
Creates a bar plot showing n-gram frequencies with optional highlighting
of selected n-grams. Supports both detected n-grams and selected multi-word expressions.
}
\examples{
\donttest{
  ngram_df <- data.frame(
    collocation = c("machine learning", "deep learning", "neural network"),
    count = c(150, 120, 90),
    lambda = c(5.2, 4.8, 4.1),
    z = c(12.3, 10.5, 9.2)
  )
  plot_ngram_frequency(ngram_df, selected = c("machine learning"))
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{plot_sentiment_by_category}
\alias{plot_sentiment_by_category}
\title{Plot Sentiment by Category}
\usage{
plot_sentiment_by_category(
  sentiment_data,
  category_var,
  plot_type = "bar",
  title = NULL
)
}
\arguments{
\item{sentiment_data}{Data frame with 'sentiment' column}

\item{category_var}{Name of the category variable column}

\item{plot_type}{Type of plot: "bar" or "stacked" (default: "bar")}

\item{title}{Plot title (default: auto-generated)}
}
\value{
A ggplot2 grouped/stacked bar chart
}
\description{
Creates a grouped or stacked bar plot showing sentiment distribution across categories.
}
\examples{
\donttest{
articles <- TextAnalysisR::SpecialEduTech[1:20, ]
sentiment_results <- analyze_sentiment(articles$abstract)
sentiment_data <- cbind(
  reference_type = articles$reference_type,
  sentiment_results
)
sentiment_plot <- plot_sentiment_by_category(sentiment_data, "reference_type")
print(sentiment_plot)
}
}
\concept{sentiment}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{require_feature}
\alias{require_feature}
\title{Require Feature}
\usage{
require_feature(feature, session = NULL)
}
\arguments{
\item{feature}{Character: feature name to check}

\item{session}{Shiny session object (optional)}
}
\value{
Logical TRUE if available, FALSE if not
}
\description{
Checks feature availability and shows notification if unavailable.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{get_spacy_embeddings}
\alias{get_spacy_embeddings}
\title{Get spaCy Word Embeddings}
\usage{
get_spacy_embeddings(texts, model = "en_core_web_md")
}
\arguments{
\item{texts}{Character vector of words or texts.}

\item{model}{Character; spaCy model to use (default: "en_core_web_md").}
}
\value{
A matrix of word embeddings (rows = texts, cols = dimensions).
}
\description{
Get word vector embeddings for words or texts using spaCy.
Requires a spaCy model with word vectors.
}
\concept{lexical}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_top_readability_documents}
\alias{plot_top_readability_documents}
\title{Plot Top Documents by Readability}
\usage{
plot_top_readability_documents(
  readability_data,
  metric,
  top_n = 15,
  order = "highest",
  title = NULL
)
}
\arguments{
\item{readability_data}{Data frame from calculate_text_readability()}

\item{metric}{Metric to plot}

\item{top_n}{Number of documents to show (default: 15)}

\item{order}{Direction: "highest" or "lowest" (default: "highest")}

\item{title}{Plot title (default: auto-generated)}
}
\value{
A plotly bar chart
}
\description{
Creates a bar plot of documents ranked by readability metric.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{plot_error}
\alias{plot_error}
\title{Create Error Plot for Plotly}
\usage{
plot_error(message, color = "#ef4444")
}
\arguments{
\item{message}{The message to display}

\item{color}{Color for the text (default: "#ef4444")}
}
\value{
A plotly plot object displaying the message
}
\description{
Creates a plotly error/status plot with a message
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{validate_semantic_coherence}
\alias{validate_semantic_coherence}
\title{Validate Semantic Coherence}
\usage{
validate_semantic_coherence(embeddings, topic_assignments, ...)
}
\arguments{
\item{embeddings}{Document embeddings matrix.}

\item{topic_assignments}{Vector of topic assignments for documents.}

\item{...}{Additional parameters (currently unused).}
}
\value{
List containing coherence score and metrics.
}
\description{
Validates the semantic coherence of topic assignments using
intra-cluster distance in embedding space.
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{extract_keywords_keyness}
\alias{extract_keywords_keyness}
\title{Extract Keywords Using Statistical Keyness}
\usage{
extract_keywords_keyness(
  dfm,
  target,
  top_n = 20,
  measure = "lr",
  min_count = 0
)
}
\arguments{
\item{dfm}{A quanteda dfm object}

\item{target}{Target document indices or logical vector}

\item{top_n}{Number of top keywords to extract (default: 20)}

\item{measure}{Keyness measure: "lr" (log-likelihood G-squared), "chi2",
"exact" (Fisher's exact odds ratio), or "pmi" (default: "lr")}

\item{min_count}{Minimum total term frequency before computing keyness
(default: 0, no pruning)}
}
\value{
Data frame with columns: Keyword, Keyness_Score
}
\description{
Extracts distinctive keywords by comparing document groups using log-likelihood ratio (G-squared).
}
\examples{
\donttest{
abstracts <- TextAnalysisR::SpecialEduTech$abstract[1:10]
dfm_object <- quanteda::dfm(quanteda::tokens(quanteda::corpus(abstracts)))
keywords <- extract_keywords_keyness(dfm_object, target = 1)
print(keywords)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{spacy_has_vectors}
\alias{spacy_has_vectors}
\title{Check if Model Has Word Vectors}
\usage{
spacy_has_vectors()
}
\value{
Logical; TRUE if model has vectors, FALSE otherwise.
}
\description{
Check if the loaded spaCy model has word vectors for similarity calculations.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_best_embeddings}
\alias{get_best_embeddings}
\title{Get Best Available Embeddings}
\usage{
get_best_embeddings(
  texts,
  provider = "auto",
  model = NULL,
  api_key = NULL,
  verbose = TRUE
)
}
\arguments{
\item{texts}{Character vector of texts to embed}

\item{provider}{Character string: "auto" (default), "sentence-transformers",
"openai", or "gemini". Use "auto" for automatic detection.}

\item{model}{Character string specifying the embedding model. If NULL, uses default
model for the selected provider.}

\item{api_key}{Optional API key for OpenAI or Gemini providers. If NULL, falls back
to environment variables (OPENAI_API_KEY, GEMINI_API_KEY).}

\item{verbose}{Logical, whether to print progress messages (default: TRUE)}
}
\value{
Matrix with embeddings (rows = texts, columns = dimensions)
}
\description{
Auto-detects and uses the best available embedding provider with the following priority:
\enumerate{
\item sentence-transformers (local Python) - if Python environment is set up
\item OpenAI API - if OPENAI_API_KEY is set
\item Gemini API - if GEMINI_API_KEY is set
}
}
\examples{
if (interactive()) {
data(SpecialEduTech)
texts <- SpecialEduTech$abstract[1:5]

# Auto-detect best available provider
embeddings <- get_best_embeddings(texts)

# Force specific provider
embeddings <- get_best_embeddings(texts, provider = "openai")

dim(embeddings)
}
}
\concept{ai}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{calculate_dispersion_metrics}
\alias{calculate_dispersion_metrics}
\title{Calculate Dispersion Metrics}
\usage{
calculate_dispersion_metrics(tokens_object, terms)
}
\arguments{
\item{tokens_object}{A quanteda tokens object}

\item{terms}{Character vector of terms to analyze}
}
\value{
Data frame with columns:
\itemize{
\item term: The search term
\item frequency: Total occurrences
\item doc_count: Number of documents containing term
\item doc_ratio: Proportion of documents containing term
\item juilland_d: Juilland's D dispersion (0-1, higher = more even),
computed on within-document proportions
\item rosengren_s: Rosengren's S dispersion adjusted for document
sizes (1/n_docs to 1, higher = more even)
}
}
\description{
Computes quantitative dispersion metrics for terms, measuring how
evenly distributed they are across the corpus.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_python_env}
\alias{check_python_env}
\title{Check Python Environment Status}
\usage{
check_python_env(envname = "textanalysisr-env")
}
\arguments{
\item{envname}{Character string name of the virtual environment
(default: "textanalysisr-env")}
}
\value{
List with status information:
\itemize{
\item available: Logical, TRUE if environment exists
\item active: Logical, TRUE if environment is currently active
\item packages: List of installed package versions
}
}
\description{
Checks if Python environment is available and properly configured.
}
\examples{
if (interactive()) {
status <- check_python_env()
print(status)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{stopwords_list}
\alias{stopwords_list}
\title{Stopwords List}
\format{
A character vector of stopwords
}
\description{
A dataset containing stopwords for text preprocessing
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{calculate_log_odds_ratio}
\alias{calculate_log_odds_ratio}
\title{Calculate Log Odds Ratio Between Categories}
\usage{
calculate_log_odds_ratio(
  dfm_object,
  group_var,
  comparison_mode = c("binary", "one_vs_rest", "pairwise"),
  reference_level = NULL,
  top_n = 10,
  min_count = 5
)
}
\arguments{
\item{dfm_object}{A quanteda dfm object}

\item{group_var}{Character, name of the grouping variable in docvars}

\item{comparison_mode}{Character, one of "binary", "one_vs_rest", or "pairwise"
\itemize{
\item binary: Compare two categories directly
\item one_vs_rest: Compare each category against all others combined
\item pairwise: Compare all pairs of categories
}}

\item{reference_level}{Character, reference category for binary comparison (default: first level)}

\item{top_n}{Number of top terms per comparison (default: 10)}

\item{min_count}{Minimum word count to include (default: 5)}
}
\value{
Data frame with columns:
\itemize{
\item term: The word/feature
\item category1: First category in comparison
\item category2: Second category in comparison
\item count1: Count in category 1
\item count2: Count in category 2
\item odds1: Smoothed odds in category 1, (count + 1) / (total + V - count - 1)
\item odds2: Smoothed odds in category 2
\item odds_ratio: Ratio of odds
\item log_odds_ratio: Log of odds ratio (positive = more in compared category)
\item variance: Approximate variance of the log odds ratio
\item z_score: Log odds ratio divided by its standard error
\item significant: TRUE when |z| >= 1.96
}
Terms are ranked by absolute z-score.
}
\description{
Compares word usage between categories using the log-odds ratio with a
uniform Dirichlet prior, ranked by z-score. Identifies words distinctively
used in one category. For the informative-prior weighted log-odds, see
\code{\link[=calculate_weighted_log_odds]{calculate_weighted_log_odds()}}.
}
\examples{
\donttest{
articles <- TextAnalysisR::SpecialEduTech[1:20, ]
corpus <- quanteda::corpus(
  articles$abstract,
  docvars = data.frame(reference_type = articles$reference_type)
)
dfm_object <- quanteda::dfm(quanteda::tokens(corpus))
log_odds <- calculate_log_odds_ratio(dfm_object, "reference_type")
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{detect_multi_words}
\alias{detect_multi_words}
\title{Detect Multi-Word Expressions}
\usage{
detect_multi_words(tokens, size = 2:5, min_count = 2)
}
\arguments{
\item{tokens}{A \code{tokens} object from the \code{quanteda} package.}

\item{size}{A numeric vector specifying the sizes of the collocations to detect (default: 2:5).}

\item{min_count}{The minimum number of occurrences for a collocation to be
considered (default: 2).}
}
\value{
A character vector of detected collocations.
}
\description{
This function detects multi-word expressions (collocations) of specified
sizes that appear at least a specified number of times in the provided tokens.
}
\examples{
\donttest{
  mydata <- TextAnalysisR::SpecialEduTech[seq_len(50), ]

  united_tbl <- TextAnalysisR::unite_cols(
    mydata,
    listed_vars = c("title", "keyword", "abstract")
  )

  tokens <- TextAnalysisR::prep_texts(united_tbl, text_field = "united_texts")

  collocations <- TextAnalysisR::detect_multi_words(tokens, size = 2:3, min_count = 2)
  print(collocations)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{extract_pdf_multimodal}
\alias{extract_pdf_multimodal}
\title{Extract PDF with Multimodal Analysis}
\usage{
extract_pdf_multimodal(
  file_path,
  vision_provider = "gemini",
  vision_model = NULL,
  api_key = NULL,
  describe_images = TRUE,
  envname = "textanalysisr-env"
)
}
\arguments{
\item{file_path}{Character string path to PDF file}

\item{vision_provider}{Character: "openai" or "gemini" (default)}

\item{vision_model}{Character: Model name
\itemize{
\item For OpenAI: "gpt-4.1", "gpt-4.1-mini"
\item For Gemini: "gemini-2.5-flash", "gemini-2.5-pro"
}}

\item{api_key}{Character: API key (required for openai/gemini providers)}

\item{describe_images}{Logical: Convert page images to text descriptions (default: TRUE)}

\item{envname}{Character: Kept for backward compatibility, ignored}
}
\value{
List with:
\itemize{
\item success: Logical
\item combined_text: Character string with all content for text analysis
\item text_content: List of text chunks
\item image_descriptions: List of image descriptions
\item num_images: Integer count of described pages
\item vision_provider: Character indicating provider used
\item message: Character status message
}
}
\description{
Extract both text and visual content from PDFs using R-native pdftools
and vision LLM APIs. No Python required.
}
\details{
\strong{Workflow:}
\enumerate{
\item Extracts text using pdftools (R-native)
\item Renders each page as an image
\item Sends sparse-text pages to vision LLM for description
\item Merges text + descriptions into a single text corpus
}
}
\concept{pdf}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_topic_quality}
\alias{calculate_topic_quality}
\title{Calculate Semantic Topic Quality Metrics}
\usage{
calculate_topic_quality(
  embeddings,
  topic_assignments,
  similarity_matrix = NULL
)
}
\arguments{
\item{embeddings}{Document embeddings matrix.}

\item{topic_assignments}{Vector of topic assignments.}

\item{similarity_matrix}{Optional similarity matrix.}
}
\value{
A list of quality metrics.
}
\description{
Internal function to calculate quality metrics for semantic topic modeling results.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validate_column_name}
\alias{validate_column_name}
\title{Validate Column Name}
\usage{
validate_column_name(col_name)
}
\arguments{
\item{col_name}{Character string containing the column name}
}
\value{
TRUE if valid, stops with error if invalid
}
\description{
Validates column names to prevent code injection through formula construction.
Ensures column names follow R naming conventions and contain no malicious patterns.
}
\section{Security}{

Protects against formula injection attacks where malicious column names could
execute arbitrary code when used in model formulas. Part of NIST SI-10 input validation.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_readability_distribution}
\alias{plot_readability_distribution}
\title{Plot Readability Distribution}
\usage{
plot_readability_distribution(readability_data, metric, title = NULL)
}
\arguments{
\item{readability_data}{Data frame from calculate_text_readability()}

\item{metric}{Metric to plot (e.g., "flesch", "flesch_kincaid", "gunning_fog")}

\item{title}{Plot title (default: auto-generated)}
}
\value{
A plotly boxplot
}
\description{
Creates a boxplot showing the overall distribution of a readability metric.
}
\examples{
\donttest{
data(SpecialEduTech, package = "TextAnalysisR")
texts <- SpecialEduTech$abstract[1:20]
readability <- calculate_text_readability(texts)
plot <- plot_readability_distribution(readability, "flesch")
print(plot)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{get_available_dfm}
\alias{get_available_dfm}
\title{Get Available Document-Feature Matrix with Fallback}
\usage{
get_available_dfm(
  dfm_lemma = NULL,
  dfm_outcome = NULL,
  dfm_final = NULL,
  dfm_init = NULL
)
}
\arguments{
\item{dfm_lemma}{Optional lemmatized DFM (highest priority)}

\item{dfm_outcome}{Optional preprocessed DFM (medium priority)}

\item{dfm_final}{Optional final processed DFM (medium-low priority)}

\item{dfm_init}{Optional initial DFM (lowest priority)}
}
\value{
The first non-NULL DFM from the priority chain, or NULL if all are NULL
}
\description{
Returns the first non-NULL DFM from a priority fallback chain.
Useful when multiple DFM processing stages exist and the most processed available version is needed.
}
\details{
Priority order (highest to lowest):
\enumerate{
\item dfm_lemma - Lemmatized tokens (most processed)
\item dfm_outcome - Preprocessed tokens
\item dfm_final - Final processed version
\item dfm_init - Initial unprocessed tokens
}
}
\examples{
\donttest{
  dfm1 <- quanteda::dfm(quanteda::tokens("assistive technology supports learning"))
  result <- get_available_dfm(dfm_init = dfm1)
}
}
\concept{preprocessing}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{create_analysis_datatable}
\alias{create_analysis_datatable}
\title{Create Formatted Analysis Data Table}
\usage{
create_analysis_datatable(
  data,
  colnames = NULL,
  numeric_cols = NULL,
  digits = 3,
  export_formats = c("copy", "csv", "excel", "pdf", "print"),
  page_length = 25,
  font_size = "16px"
)
}
\arguments{
\item{data}{Data frame to display}

\item{colnames}{Optional character vector of column names for display}

\item{numeric_cols}{Optional character vector of numeric columns to round}

\item{digits}{Number of digits for rounding numeric columns (default: 3)}

\item{export_formats}{Character vector of export formats (default: c('copy', 'csv', 'excel', 'pdf', 'print'))}

\item{page_length}{Number of rows per page (default: 25)}

\item{font_size}{Font size for table cells (default: "16px")}
}
\value{
A DT::datatable object
}
\description{
Creates a consistently formatted DT::datatable for analysis results with
export buttons and optional numeric formatting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{extract_pdf_smart}
\alias{extract_pdf_smart}
\title{Smart PDF Extraction with Auto-Detection}
\usage{
extract_pdf_smart(
  file_path,
  doc_type = "auto",
  vision_provider = "gemini",
  vision_model = NULL,
  api_key = NULL,
  envname = "textanalysisr-env"
)
}
\arguments{
\item{file_path}{Character string path to PDF file}

\item{doc_type}{Character: "auto" (default), "academic", or "general" (kept for compatibility)}

\item{vision_provider}{Character: "openai" or "gemini" (default)}

\item{vision_model}{Character: Model name for vision analysis}

\item{api_key}{Character: API key for cloud providers}

\item{envname}{Character: Kept for backward compatibility, ignored}
}
\value{
List with extracted content ready for text analysis
}
\description{
Extracts text and visual content from PDFs using R-native pdftools
and vision LLM APIs. Routes directly to multimodal extraction.
}
\concept{pdf}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{calculate_lexical_dispersion}
\alias{calculate_lexical_dispersion}
\title{Calculate Lexical Dispersion}
\usage{
calculate_lexical_dispersion(
  tokens_object,
  terms,
  scale = c("relative", "absolute")
)
}
\arguments{
\item{tokens_object}{A quanteda tokens object}

\item{terms}{Character vector of terms to analyze}

\item{scale}{Character, "relative" (0-1 normalized) or "absolute" (token position)}
}
\value{
Data frame with columns:
\itemize{
\item doc_id: Document identifier
\item term: The search term
\item position: Position in document (relative or absolute)
\item doc_length: Total tokens in document
}
}
\description{
Computes lexical dispersion data for specified terms across a corpus.
Shows where terms appear within each document, useful for understanding
term distribution patterns.
}
\examples{
\donttest{
tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1:5])
dispersion <- calculate_lexical_dispersion(tokens, c("learning", "instruction"))
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{clean_similarity_matrix}
\alias{clean_similarity_matrix}
\title{Clean Similarity Matrix}
\usage{
clean_similarity_matrix(similarity_matrix)
}
\arguments{
\item{similarity_matrix}{A numeric matrix of similarity values.}
}
\value{
Cleaned similarity matrix.
}
\description{
Cleans a similarity matrix by handling NA/Inf values,
ensuring symmetry, and setting diagonal to 1.
}
\concept{matrix-utilities}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{generate_topic_content}
\alias{generate_topic_content}
\title{Generate Content from Topic Terms}
\usage{
generate_topic_content(
  topic_terms_df,
  content_type = c("survey_item", "research_question", "theme_description",
    "policy_recommendation", "interview_question", "custom"),
  topic_var = "topic",
  term_var = "term",
  weight_var = "beta",
  provider = c("openai", "gemini"),
  model = "gpt-4.1-mini",
  temperature = 0,
  system_prompt = NULL,
  user_prompt_template = NULL,
  max_tokens = 150,
  api_key = NULL,
  output_var = NULL,
  verbose = TRUE
)
}
\arguments{
\item{topic_terms_df}{A data frame with topic terms, containing columns for
topic identifier, term, and optionally term weight (beta).}

\item{content_type}{Type of content to generate. One of:
\describe{
\item{"survey_item"}{Likert-scale survey items for scale development}
\item{"research_question"}{Research questions for literature review}
\item{"theme_description"}{Theme descriptions for qualitative analysis}
\item{"policy_recommendation"}{Policy recommendations for policy analysis}
\item{"interview_question"}{Interview questions for qualitative research}
\item{"custom"}{Custom content using user-provided prompts}
}}

\item{topic_var}{Name of the column containing topic identifiers (default: "topic").}

\item{term_var}{Name of the column containing terms (default: "term").}

\item{weight_var}{Name of the column containing term weights (default: "beta").}

\item{provider}{LLM provider: "openai" or "gemini" (default: "openai").}

\item{model}{Model name. For OpenAI: "gpt-4.1-mini", "gpt-4", etc.
For Gemini: "gemini-2.5-flash-lite", "gemini-2.5-flash", etc.}

\item{temperature}{Sampling temperature (0-2). Lower = more deterministic (default: 0).}

\item{system_prompt}{Custom system prompt. If NULL, uses default for content_type.}

\item{user_prompt_template}{Custom user prompt template with \{terms\} placeholder.
If NULL, uses default for content_type.}

\item{max_tokens}{Maximum tokens for response (default: 150).}

\item{api_key}{API key for the selected provider. If NULL, reads from
OPENAI_API_KEY or GEMINI_API_KEY environment variable.}

\item{output_var}{Name of the output column (default: based on content_type).}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A data frame with generated content joined to original topic terms.
}
\description{
Uses Large Language Models (LLMs) to generate various types of content
based on topic model terms. Supports multiple content types with optimized
default prompts, or fully custom prompts.
}
\details{
The function generates one piece of content per unique topic. Each content type
has optimized default prompts, but these can be overridden with custom prompts.

Requires an API key set via the \code{api_key} parameter or the relevant
environment variable (OPENAI_API_KEY or GEMINI_API_KEY).
}
\examples{
if (interactive()) {
# Generate survey items
survey_items <- generate_topic_content(
  topic_terms_df = top_terms,
  content_type = "survey_item",
  provider = "openai",
  model = "gpt-4.1-mini"
)

# Generate research questions
research_qs <- generate_topic_content(
  topic_terms_df = top_terms,
  content_type = "research_question",
  provider = "gemini",
  model = "gemini-2.5-flash-lite"
)

# Generate with custom prompt
custom_content <- generate_topic_content(
  topic_terms_df = top_terms,
  content_type = "custom",
  system_prompt = "You are an expert in educational policy...",
  user_prompt_template = "Based on {terms}, generate a learning objective:"
)
}
}
\seealso{
\code{\link[=generate_topic_labels]{generate_topic_labels()}} for the step that creates topic labels; \code{\link[=get_content_type_prompt]{get_content_type_prompt()}} and \code{\link[=get_content_type_user_template]{get_content_type_user_template()}} to inspect or override default prompts
}
\concept{ai}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{find_similar_words}
\alias{find_similar_words}
\title{Find Similar Words}
\usage{
find_similar_words(word, top_n = 10L, model = "en_core_web_md")
}
\arguments{
\item{word}{Character; target word.}

\item{top_n}{Integer; number of similar words to return (default: 10).}

\item{model}{Character; spaCy model to use (default: "en_core_web_md").}
}
\value{
A data frame with similar words and similarity scores.
}
\description{
Find words most similar to a given word using word vectors.
Requires a spaCy model with word vectors (en_core_web_md or en_core_web_lg).
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validate_keyboard_navigation}
\alias{validate_keyboard_navigation}
\title{Validate Keyboard Navigation}
\usage{
validate_keyboard_navigation(tabindex = 0)
}
\arguments{
\item{tabindex}{Integer, tab order (-1 for no tab, 0 for natural order, 1+ for specific order)}
}
\value{
Logical TRUE if valid, FALSE with warning if invalid
}
\description{
Checks if interactive elements have proper tabindex and keyboard handlers.
Used for WCAG 2.1.1 (Keyboard) compliance.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_weighted_log_odds}
\alias{plot_weighted_log_odds}
\title{Plot Weighted Log Odds}
\usage{
plot_weighted_log_odds(
  weighted_data,
  top_n = 10,
  color_positive = "#10B981",
  color_negative = "#EF4444",
  height = 600,
  width = NULL,
  title = "Weighted Log Odds by Group"
)
}
\arguments{
\item{weighted_data}{Data frame from calculate_weighted_log_odds()}

\item{top_n}{Number of top terms to show per group (default: 10)}

\item{color_positive}{Color for positive log odds (default: "#10B981" green)}

\item{color_negative}{Color for negative log odds (default: "#EF4444" red)}

\item{height}{Plot height in pixels (default: 600)}

\item{width}{Plot width in pixels (default: NULL for auto)}

\item{title}{Plot title (default: "Weighted Log Odds by Group")}
}
\value{
A plotly object
}
\description{
Creates a faceted horizontal bar plot showing weighted log odds for comparing
word usage across categories using the Fightin' Words method (Monroe et al. 2008).
Each group is displayed in a separate facet showing its most distinctive terms.
}
\examples{
\donttest{
if (requireNamespace("tidylo", quietly = TRUE)) {
  articles <- TextAnalysisR::SpecialEduTech[1:20, ]
  dfm_object <- quanteda::dfm(quanteda::tokens(articles$abstract))
  quanteda::docvars(dfm_object, "reference_type") <- articles$reference_type
  weighted_odds <- calculate_weighted_log_odds(dfm_object, "reference_type",
                                                top_n = 5)
  plot_weighted_log_odds(weighted_odds)
}
}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_gemini_embeddings}
\alias{get_gemini_embeddings}
\title{Get Gemini Embeddings (Internal)}
\usage{
get_gemini_embeddings(texts, model, api_key)
}
\description{
Get Gemini Embeddings (Internal)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{render_displacy_ent}
\alias{render_displacy_ent}
\title{Render displaCy Entity Visualization}
\usage{
render_displacy_ent(text, model = "en_core_web_sm", colors = NULL)
}
\arguments{
\item{text}{Character string to visualize.}

\item{model}{spaCy model name (default: "en_core_web_sm").}

\item{colors}{Named list of entity type to color mappings (e.g.,
list(PERSON = "#e91e63", ORG = "#2196f3")). If NULL, uses spaCy defaults.}
}
\value{
HTML string with entity highlighting.
}
\description{
Renders spaCy's displaCy entity visualization as HTML.
Highlights named entities with colored labels.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{plot_keyword_comparison}
\alias{plot_keyword_comparison}
\title{Plot Keyword Comparison (TF-IDF vs Frequency)}
\usage{
plot_keyword_comparison(
  tfidf_data,
  top_n = 10,
  title = NULL,
  normalized = FALSE
)
}
\arguments{
\item{tfidf_data}{Data frame from extract_keywords_tfidf()}

\item{top_n}{Number of keywords to display (default: 10)}

\item{title}{Plot title (default: auto-generated)}

\item{normalized}{Logical, whether TF-IDF scores are normalized (default: FALSE)}
}
\value{
A plotly grouped bar chart
}
\description{
Creates a grouped bar plot comparing TF-IDF scores with term frequencies.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{extract_morphology}
\alias{extract_morphology}
\title{Extract Morphological Features}
\usage{
extract_morphology(
  tokens,
  features = c("Number", "Tense", "VerbForm", "Person", "Case", "Mood", "Aspect"),
  include_pos = TRUE,
  include_lemma = TRUE,
  model = "en_core_web_sm"
)
}
\arguments{
\item{tokens}{A quanteda tokens object or character vector of texts.}

\item{features}{Character vector of morphological features to extract.
Default includes common Universal Dependencies features.}

\item{include_pos}{Logical; include POS tags (default: TRUE).}

\item{include_lemma}{Logical; include lemmatized forms (default: TRUE).}

\item{model}{Character; spaCy model to use (default: "en_core_web_sm").}
}
\value{
A data frame with token-level morphological annotations including
morph_* columns for each requested feature.
}
\description{
Uses spaCy to extract morphological features from text.
Returns data with Number, Tense, VerbForm, Person, Case, Mood, Aspect, etc.
}
\details{
Morphological features follow Universal Dependencies annotation.
Common features include:
\itemize{
\item \code{Number}: Sing (singular), Plur (plural)
\item \code{Tense}: Past, Pres (present), Fut (future)
\item \code{VerbForm}: Fin (finite), Inf (infinitive), Part (participle), Ger (gerund)
\item \code{Person}: 1, 2, 3 (first, second, third person)
\item \code{Case}: Nom (nominative), Acc (accusative), Gen (genitive), Dat (dative)
\item \code{Mood}: Ind (indicative), Imp (imperative), Sub (subjunctive)
\item \code{Aspect}: Perf (perfective), Imp (imperfective), Prog (progressive)
}
}
\examples{
if (interactive()) {
  tokens <- quanteda::tokens(TextAnalysisR::SpecialEduTech$abstract[1])
  morphology_data <- extract_morphology(tokens)
  print(morphology_data)
}
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{generate_topic_labels}
\alias{generate_topic_labels}
\title{Generate Topic Labels Using AI}
\usage{
generate_topic_labels(
  top_topic_terms,
  provider = "auto",
  model = NULL,
  system = NULL,
  user = NULL,
  temperature = 0.5,
  api_key = NULL,
  openai_api_key = NULL,
  verbose = TRUE
)
}
\arguments{
\item{top_topic_terms}{A data frame containing the top terms for each topic.}

\item{provider}{AI provider to use: "auto" (default), "openai", or "gemini".
"auto" picks the first provider with an available API key.}

\item{model}{A character string specifying which model to use. If NULL, uses
provider defaults: "gpt-4.1-mini" (OpenAI), "gemini-2.5-flash-lite" (Gemini).}

\item{system}{A character string containing the system prompt for the API.
If NULL, the function uses the default system prompt.}

\item{user}{A character string containing the user prompt for the API.
If NULL, the function uses the default user prompt.}

\item{temperature}{A numeric value controlling the randomness of the output (default: 0.5).}

\item{api_key}{API key for OpenAI or Gemini. If NULL, uses environment variable.}

\item{openai_api_key}{Deprecated. Use \code{api_key} instead. Kept for backward compatibility.}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A data frame containing the top terms for each topic along with their generated labels.
}
\description{
This function generates descriptive labels for each topic based on their
top terms using AI providers (OpenAI or Gemini).
}
\examples{
if (interactive()) {
top_topic_terms <- get_topic_terms(stm_model, top_term_n = 10)

# Auto-detect provider (tries OpenAI -> Gemini)
labels <- generate_topic_labels(top_topic_terms)

# Use specific provider
labels_openai <- generate_topic_labels(top_topic_terms, provider = "openai")
labels_gemini <- generate_topic_labels(top_topic_terms, provider = "gemini")
}
}
\seealso{
\code{\link[=get_topic_terms]{get_topic_terms()}} to extract top terms first; \code{\link[=generate_topic_content]{generate_topic_content()}} for survey items / RQs / themes grounded in the same terms; \code{\link[=call_llm_api]{call_llm_api()}} for the direct AI provider call
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{calculate_topic_diversity}
\alias{calculate_topic_diversity}
\title{Calculate Topic Diversity (Unique-Word Proportion)}
\usage{
calculate_topic_diversity(top_terms_list, top_k = 25)
}
\arguments{
\item{top_terms_list}{Named or unnamed list of character vectors, one per topic.}

\item{top_k}{Number of top terms per topic to include (default 25).}
}
\value{
Numeric between 0 and 1, or NA if input empty.
}
\description{
Diversity = unique top-K terms across all topics / (n_topics * top_k).
Range 0 (all topics identical) to 1 (fully disjoint). Complementary to NPMI
coherence: a good model scores high on both.
}
\concept{topic-modeling}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{lexical_frequency_analysis}
\alias{lexical_frequency_analysis}
\title{Lexical Frequency Analysis}
\usage{
lexical_frequency_analysis(...)
}
\arguments{
\item{...}{Arguments passed to plot_word_frequency}
}
\value{
A plotly bar chart of word frequencies
}
\description{
Wrapper function for plot_word_frequency for lexical analysis.
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{renumber_clusters_sequentially}
\alias{renumber_clusters_sequentially}
\title{Renumber Clusters Sequentially}
\usage{
renumber_clusters_sequentially(clusters)
}
\arguments{
\item{clusters}{A vector of cluster assignments.}
}
\value{
Vector with clusters renumbered sequentially (1, 2, 3, ...).
}
\description{
Renumbers cluster assignments to sequential integers
starting from 1.
}
\concept{matrix-utilities}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{create_sr_text}
\alias{create_sr_text}
\title{Create Screen Reader Text}
\usage{
create_sr_text(text)
}
\arguments{
\item{text}{Text to be read by screen readers}
}
\value{
HTML span with sr-only class
}
\description{
Generates visually hidden text for screen readers (WCAG 4.1.2).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{get_topic_prevalence}
\alias{get_topic_prevalence}
\title{Get Topic Prevalence (Gamma) from STM Model}
\usage{
get_topic_prevalence(stm_model, category = NULL, include_theta = FALSE)
}
\arguments{
\item{stm_model}{A fitted STM model object from stm::stm().}

\item{category}{Optional character string to add as a category column.}

\item{include_theta}{Logical, if TRUE includes document-topic matrix (default: FALSE).}
}
\value{
A data frame with columns:
\item{topic}{Topic number}
\item{gamma}{Mean topic prevalence across documents}
\item{category}{Category label (if provided)}
}
\description{
Extracts topic prevalence values (gamma/theta) from a fitted STM model,
returning mean prevalence for each topic as a data frame.
}
\examples{
if (interactive() && requireNamespace("stm", quietly = TRUE)) {
  # Requires fitting an STM model first; uses 'stm::gadarian' for demo
  data("gadarian", package = "stm")
  proc <- stm::textProcessor(gadarian$open.ended.response, metadata = gadarian)
  prep <- stm::prepDocuments(proc$documents, proc$vocab, proc$meta)
  topic_model <- stm::stm(prep$documents, prep$vocab, K = 3,
                           data = prep$meta, max.em.its = 5,
                           verbose = FALSE)
  prevalence <- get_topic_prevalence(topic_model)
  prevalence_label <- get_topic_prevalence(topic_model, category = "demo")
}
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{analyze_semantic_evolution}
\alias{analyze_semantic_evolution}
\title{Analyze Semantic Evolution}
\usage{
analyze_semantic_evolution(temporal_results, verbose = FALSE, ...)
}
\arguments{
\item{temporal_results}{Temporal analysis results}

\item{verbose}{Logical indicating whether to print progress messages}

\item{...}{Additional parameters}
}
\value{
List containing evolution analysis
}
\description{
Analyzes semantic evolution patterns in temporal results
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{process_pdf_unified}
\alias{process_pdf_unified}
\title{Process PDF File (Unified Entry Point)}
\usage{
process_pdf_unified(
  file_path,
  use_multimodal = FALSE,
  vision_provider = "gemini",
  vision_model = NULL,
  api_key = NULL,
  describe_images = TRUE
)
}
\arguments{
\item{file_path}{Character string path to PDF file}

\item{use_multimodal}{Logical, enable multimodal extraction}

\item{vision_provider}{Character, "openai" or "gemini"}

\item{vision_model}{Character, model name}

\item{api_key}{Character, API key (if using openai/gemini)}

\item{describe_images}{Logical, generate image descriptions}
}
\value{
List: success, data, type, method, message
}
\description{
Unified PDF processing:
\enumerate{
\item Multimodal (R-native pdftools + Vision LLM) if enabled
\item R pdftools text extraction as fallback
}
}
\seealso{
\code{\link[=import_files]{import_files()}} for the generic file-loading entry point; \code{\link[=describe_image]{describe_image()}} for standalone image captioning
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{detect_pdf_content_type}
\alias{detect_pdf_content_type}
\title{Detect PDF Content Type}
\usage{
detect_pdf_content_type(file_path)
}
\arguments{
\item{file_path}{Character string path to PDF file}
}
\value{
Character string: "text" or "unknown"
}
\description{
Analyzes PDF to determine if it contains readable text.
}
\details{
Attempts text extraction using pdftools. Returns "text" if successful,
or "unknown" if extraction fails or PDF is empty.

For table extraction from PDFs, use \code{\link{extract_tables_from_pdf_py}}.
}
\examples{
\donttest{
pdf_path <- "path/to/document.pdf"
content_type <- detect_pdf_content_type(pdf_path)
print(content_type)
}
}
\concept{pdf}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{identify_topic_trends}
\alias{identify_topic_trends}
\title{Identify Topic Trends}
\usage{
identify_topic_trends(temporal_results, ...)
}
\arguments{
\item{temporal_results}{Temporal analysis results}

\item{...}{Additional parameters}
}
\value{
List containing identified trends
}
\description{
Identifies trending topics in temporal results
}
\concept{topic-modeling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lexical_analysis.R
\name{get_word_similarity}
\alias{get_word_similarity}
\title{Calculate Word Similarity}
\usage{
get_word_similarity(word1, word2, model = "en_core_web_md")
}
\arguments{
\item{word1}{Character; first word.}

\item{word2}{Character; second word.}

\item{model}{Character; spaCy model to use (default: "en_core_web_md").}
}
\value{
A list with similarity score and metadata.
}
\description{
Calculate semantic similarity between two words using word vectors.
Requires a spaCy model with word vectors (en_core_web_md or en_core_web_lg).
}
\concept{lexical}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{export_document_clustering}
\alias{export_document_clustering}
\title{Export Document Clustering Analysis}
\usage{
export_document_clustering(
  coordinates,
  clusters = NULL,
  labels = NULL,
  doc_ids = NULL,
  file_path
)
}
\arguments{
\item{coordinates}{Document coordinates}

\item{clusters}{Cluster assignments (optional)}

\item{labels}{Cluster labels (optional)}

\item{doc_ids}{Document IDs}

\item{file_path}{Path to save the CSV file}
}
\value{
Invisible data frame of the exported data
}
\description{
Export document clustering analysis results to CSV
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/semantic_analysis.R
\name{generate_cluster_labels}
\alias{generate_cluster_labels}
\title{Generate Cluster Label Suggestions (Human-in-the-Loop)}
\usage{
generate_cluster_labels(
  cluster_keywords,
  provider = "auto",
  model = NULL,
  temperature = 0.3,
  max_tokens = 50,
  api_key = NULL,
  verbose = TRUE
)
}
\arguments{
\item{cluster_keywords}{List of keywords for each cluster.}

\item{provider}{AI provider to use: "auto" (default), "openai", or "gemini".
"auto" picks the first provider with an available API key.}

\item{model}{Model name. If NULL, uses provider defaults: "gpt-4.1-mini" (OpenAI),
"gemini-2.5-flash-lite" (Gemini).}

\item{temperature}{Temperature parameter (default: 0.3).}

\item{max_tokens}{Maximum tokens for response (default: 50).}

\item{api_key}{API key for OpenAI or Gemini. If NULL, uses environment variable.}

\item{verbose}{Logical, if TRUE, prints progress messages.}
}
\value{
A list of generated labels.
}
\description{
Suggests descriptive labels for clusters using AI. Labels are suggestions
for human review - users should edit and approve before using.
Supports OpenAI or Gemini for AI generation.
}
\examples{
if (interactive()) {
  cluster_keywords <- list(
    "1" = c("calculator", "arithmetic", "elementary", "remedial"),
    "2" = c("computer", "instruction", "multiplication", "drill")
  )
  labels_openai <- generate_cluster_labels(cluster_keywords, provider = "openai")
  labels_gemini <- generate_cluster_labels(cluster_keywords, provider = "gemini")
}
}
\concept{semantic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{log_security_event}
\alias{log_security_event}
\title{Log Security Event}
\usage{
log_security_event(event_type, details, session_info, level = "info")
}
\arguments{
\item{event_type}{Type of security event}

\item{details}{Additional details about the event}

\item{session_info}{Shiny session object}

\item{level}{Log level ("info", "warning", "error")}
}
\description{
Log Security Event
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic_modeling.R
\name{run_neural_topics_internal}
\alias{run_neural_topics_internal}
\title{Neural Topic Modeling}
\usage{
run_neural_topics_internal(
  texts,
  n_topics = 10,
  hidden_layers = 2,
  hidden_units = 100,
  dropout_rate = 0.2,
  embedding_model = "all-MiniLM-L6-v2",
  seed = 123
)
}
\arguments{
\item{texts}{Character vector of documents}

\item{n_topics}{Number of topics to discover}

\item{hidden_layers}{Number of hidden layers in neural network}

\item{hidden_units}{Number of units per hidden layer}

\item{dropout_rate}{Dropout rate for regularization}

\item{embedding_model}{Transformer model for initial embeddings}

\item{seed}{Random seed for reproducibility}
}
\value{
List containing neural topic model and diagnostics
}
\description{
Implements neural topic modeling using deep learning architectures for improved
topic discovery and representation learning.
}
\concept{topic-modeling}
\keyword{internal}
